#include "fis.h"
double ** FIS::NormCheckDataDist(char * datafile,int &nbex,int &nbinputs,bool display,bool wordless)
{
  //normalize data with current FPs and normalize FPs for active inputs
  //minimal compatibility check

  double minp, maxp, ratiofp,median,meanO, stdO,minfp,maxfp;
  double ** C=NULL,** C2=NULL;
  int p,cpt;
  double ** normdata=NULL;
  
  nbinputs = nbex =0;
  //allocates normdata, computes nbinputs, nbex
  normdata = ReadSampleFile(datafile, nbinputs, nbex);
  nbinputs=GetNbIn();
  //transfer data to columns
  C=new double * [nbinputs];
  C2=new double * [nbinputs];
  for (p=0;p<nbinputs;p++)
    {
      C[p]=new double [nbex];
      C2[p]=new double [nbex];
    }

  //standardize data (each column) with FISIN range
  //C[p] are normalized data, Copy[p] are unnormalized data
  for (p=0;p<nbinputs;p++)
    {
      GetColumn(normdata, nbex, p, C2[p]);
      GetColumn(normdata, nbex, p, C[p]);
      
      //get min max of data column
      StatArray(C2[p], nbex, 0, median, meanO, stdO, minp, maxp, 0);
      //get min max of corresponding FISIN
      minfp=In[p]->min();
      maxfp=In[p]->max();
      ratiofp=maxfp-minfp;
      if(In[p]->IsActive() && !wordless)
	{
	  printf("column %d, data range: [%f,%f],",p+1,minp,maxp);
	  printf(" fuzzy input range: [%f,%f]\n",minfp,maxfp);
	}
      // check range compatibility of pth partition with data set pth column
      if(In[p]->IsActive())
	if (minp < minfp || maxp > maxfp)
	  throw std::runtime_error ("~Data~range~is~outside~fuzzy~partition~range");
      // then normalize
      if (fabs(ratiofp) >EPSILON)
	{
	  for (cpt=0;cpt<nbex;cpt++)
	    C[p][cpt]=(C[p][cpt]-minfp)/(ratiofp);
	  // then SFP check and normalize pth partition of FIS S
	  In[p]->CheckFuzDist();
	}
      else
	{
	  sprintf( ErrorMsg, "~Constant~input~in~FIS~for~column%d",p+1);
	  throw std::runtime_error (ErrorMsg);
	}
      if (display)
	{
	  printf("Data column #%d after std \n",p+1);
	  for ( int cpt = 0 ; cpt < nbex; cpt++ ) 
	    printf("%f ", C[p][cpt]);
	  printf("\n");
	  printf("FIS input #%d after std \n",p+1);
	  In[p]->Print();
	}
    }
  //transfer normed columns into data array
  for (p=0;p<nbinputs;p++)
    for (cpt=0;cpt<nbex;cpt++)
      normdata[cpt][p]=C[p][cpt];

  //delete temp arrays
   for (p=0;p<nbinputs;p++)
     {
       delete [] C[p];
       delete [] C2[p];
     }
   delete [] C;
   delete [] C2;
  return normdata;
  
}
double** FIS::dist(char * datafile,int &nbex,char*dissoutfile, double* numc,bool norm,double mink,bool display,bool wordless)
//*******************************************************************************************************************
{
  FILE *f=NULL;
  int i,j,p,nbinputs;

  double ** diss=NULL;
  double ** C=NULL, **Copy=NULL,** C2=NULL;
  double tmp,dista;
  double **din=NULL;
  double minp, maxp, ratiofp,median,meanO, stdO,minfp,maxfp;

  nbinputs = nbex =0;
  din = ReadSampleFile(datafile, nbinputs, nbex);
  nbinputs=GetNbIn();
  C=new double * [nbinputs];
  C2=new double * [nbinputs];
  Copy=new double * [nbinputs];
  for (p=0;p<nbinputs;p++)
    {
      C[p]=new double [nbex];
      C2[p]=new double [nbex];
      Copy[p]=new double [nbex];
    }
    
  //standardize data (each column) with FISIN range
  //C[p] are normalized data, Copy[p] are unnormalized data
  for (p=0;p<nbinputs;p++)
    {
      GetColumn(din, nbex, p, C[p]);
      GetColumn(din, nbex, p, C2[p]);
      GetColumn(din, nbex, p, Copy[p]);
      //get min max of data column-have to use C2 because statarray sorts data column
      StatArray(C2[p], nbex, 0, median, meanO, stdO, minp, maxp, 0);
      //get min max of corresponding FISIN
      minfp=In[p]->min();
      maxfp=In[p]->max();
      ratiofp=maxfp-minfp;
      if(In[p]->IsActive() && !wordless)
	{
	  printf("column %d, data range: [%f,%f],",p+1,minp,maxp);
	  printf(" fuzzy input range: [%f,%f]\n",minfp,maxfp);
	}
      // check range compatibility of pth partition with data set pth column
      if(In[p]->IsActive())
	if (minp < minfp || maxp > maxfp)
	  throw std::runtime_error ("~Data~range~is~outside~fuzzy~partition~range");
      // then normalize
      if (fabs(ratiofp) >EPSILON)
	{
	  for (int cpt=0;cpt<nbex;cpt++)
	    C[p][cpt]=(C[p][cpt]-minfp)/(ratiofp);
	  // then SFP check and normalize pth partition of FIS S
	  In[p]->CheckFuzDist();
	}
      else
	{
	  sprintf( ErrorMsg, "~Constant~input~in~FIS~for~column%d",p+1);
	  throw std::runtime_error (ErrorMsg);
	}
      if (display)
	{
	  printf("Data column #%d after std \n",p+1);
	  for ( int cpt = 0 ; cpt < nbex; cpt++ ) 
	    printf("%f ", C[p][cpt]);
	  printf("\n");
	  printf("FIS input #%d after std \n",p+1);
	  In[p]->Print();
	}
    }
  diss=new double*[nbex];
  for (p=0;p<nbex;p++)
    diss[p]=new double [nbex];
   
  for (i = 0; i < nbex; i++)
    {
      diss[i][i]=0.;
      //distances are symmetric, so compute half matrix
      for(j = i+1; j < nbex; j++)
	{
	  tmp=0.0;
	  //loop on active inputs
	  for (p=0;p<nbinputs;p++)
	    {
	      if (In[p]->IsActive())
		{
		  // choice of distance type: 0=Euclidean,1=fp-based
		  if (numc[p]==1)
		    //fp (normalized=default) multidimensional distance
		    dista=In[p]->Distance(C[p][i],C[p][j],norm,0);
		  else
		    if(norm)
		      dista=fabs(C[p][i]-C[p][j]);//Euclidean on normalized data
		    else
		      dista=fabs(Copy[p][i]-Copy[p][j]);//Euclidean on unnormalized data
		  //Minkowski combination
		  tmp+= pow(dista,mink);
		}
	    }
	  //use symmetry to fill other half of matrix
	  diss[i][j]=diss[j][i]=pow(tmp,1/mink);
	}
    }
  //matrix will be written to diss.out file
  f = fopen(dissoutfile, "wt");
  for (i = 0; i < nbex; i++)
    {
      for(j = 0; j < nbex-1; j++)
	fprintf(f, "%12.7f ;", diss[i][j]);
      fprintf(f, "%12.7f \n", diss[i][j]); 
    }
  fclose(f);
    
  for (p=0;p<nbinputs;p++)
    {
      delete [] C[p];
      delete [] Copy[p];
      delete [] C2[p];
    }
  delete [] C;
  delete [] C2;
  delete [] Copy;
	
   if (din !=NULL) 
      {
	for( int i=0 ; i<nbex ; i++ )
	  delete [] din[i];
	delete [] din;
	din=NULL;
      }

   return diss;
}
double** FIS::distWithNormedData(double **normdata,int nbex,int nbinputs,char*dissoutfile, double* numc,bool norm,double mink,bool display,bool wordless)
//*******************************************************************************************************************
{
  FILE *f=NULL;
  int i,j,p;
  double ** C=NULL;
  double ** diss=NULL;
  double tmp,dista;

  //C[p] are normalized data
  C=new double*[nbinputs];
  for (p=0;p<nbinputs;p++)
    {
      C[p]=new double [nbex];
      GetColumn(normdata, nbex, p, C[p]);//normalized data
    }
  
   
  diss=new double*[nbex];
  for (p=0;p<nbex;p++)
    diss[p]=new double [nbex];
   
  for (i = 0; i < nbex; i++)
    {
      diss[i][i]=0.;
      //distances are symmetric, so compute half matrix
      for(j = i+1; j < nbex; j++)
	{
	  tmp=0.0;
	  //loop on active inputs
	  for (p=0;p<nbinputs;p++)
	    {
	      if (In[p]->IsActive())
		{
		  // choice of distance type: 0=Euclidean,1=fp-based
		  if (numc[p]!=0)
		    //fp (normalized=default) multidimensional distance
		    dista=In[p]->Distance(C[p][i],C[p][j],norm,0);
		  else
		    dista=fabs(C[p][i]-C[p][j]);//Euclidean on normalized data
		  //Minkowski combination
		  tmp+= pow(dista,mink);
		}
	    }
	  //use symmetry to fill other half of matrix
	  diss[i][j]=diss[j][i]=pow(tmp,1/mink);
	}
    }
  //matrix will be written to diss.out file
  f = fopen(dissoutfile, "wt");
  for (i = 0; i < nbex; i++)
    {
      for(j = 0; j < nbex-1; j++)
	fprintf(f, "%12.7f ;", diss[i][j]);
      fprintf(f, "%12.7f \n", diss[i][j]); 
    }
  fclose(f);
    
  if (display)
    {
      printf("\ndiss.out file written");
      printf("\nsecond element of dist matrix=%g",diss[0][1]);
    }
    
  for (p=0;p<nbinputs;p++)
    delete [] C[p];
  delete [] C;

   return diss;
}
