#include "fis.h"
#include "algo.h"
#include "distoptim.h"
///////////////////////////////////////////////////////////////////////
double GlobalOptPam(
		    // a pointer to an Object U try to optimize
		    void *  Item  ,
		    // which part of object U optimize
		    vkey * mykey,
		    // array of double ( values )
		    double *Array ,
		    // Array size
		    int sizeArray ,
		    // A set of void * pointers to pass args to feval
		    conteneur * fEvalArgs )
{
  //1 argument in container=ptr to DistOptim object
  DistOptim *fisFP= (DistOptim *)fEvalArgs->get(1);
  return fisFP->optPam(Item,mykey ,Array,sizeArray );
}
///////////////////////////////////////////////////////////////////////
void DistOptim::setNumFpOpt(int numinput,bool display)
{
  NumFpOpt=numinput;
  if (display)
    printf("\nNumFpOpt=%d\n",NumFpOpt);
}
///////////////////////////////////////////////////////////////////////
void DistOptim::AddfEvalArgs(int &nmf, int * &typemf, int &sizeP, double *&paramsFP,bool &display, double &constnear)
{
  // update feval parameters
  fEvalArgs->add(&nmf);
  fEvalArgs->add(typemf);
  fEvalArgs->add(&sizeP);
  fEvalArgs->add(paramsFP);
  fEvalArgs->add(&display);
  fEvalArgs->add(&constnear);
}
///////////////////////////////////////////////////////////////////////
FISIN * DistOptim::GetFuzzyInputParams(int &nmf,int *& typemf, int &sizeparams, double *& params, bool display)
{
  FISIN *fuzinput=FisFPOpt->In[NumFpOpt];
  sizeparams=0;
  params=NULL;
    
  if (fuzinput==NULL) return NULL;

  if (display)
    {
      cout<<"in GetFuzzyInputParams, standardized fuzzy input";
      fuzinput->Print();
    }
  nmf=fuzinput->GetNbMf();
  fuzinput->GetSFPparams(params, typemf, sizeparams,display);//allocates and modifies params and typemf
  return fuzinput;
}
///////////////////////////////////////////////////////////////////////

int DistOptim::UpdateFuzzyInputFromParams(int nmf,int * typemf, int sizeparams, double * params, bool display, double cnear)
// return value: 0 if success in update, 1 if cnear not satisfied 
{
  // cnear is in [0,1] and represents the minimu distance between adjacent kernels
  // for one fuzzy input and one set of params
  FISIN *fuzinput=NULL;
  int k;
  //clone gPar.fd:  ensures the update of MF linearization
  fuzinput=new FISIN(*(FisFPOpt->In[NumFpOpt]));

  int notnear = true;
 
  double near;
  //if(cnear <0. || cnear > (1./nmf)) near = 0.;
  if(cnear <0.) near = 0.;
  else if (cnear > (1./nmf)) near=(1./nmf)-EPSILON;
  else near = cnear;

  if (display)
    {
      printf("new FP params=\t");
      for ( int i=0;i<sizeparams;i++)
	printf("%g\t",params[i]);
      printf("\n");
      cout << "UpdateFuzzyInputFromParams near:  " << near << endl;
    }
  // build SFP with triangles or trapezoids
  k=0;
  fuzinput->ReplaceMF(0,new MFTRAPINF(0.0,params[0],params[1]));
  if (nmf==2)
    {
      if((params[1] - params[0]) < near)
	    { 
	      notnear = false; 
	      if(display) cout << "near:  " << near << "  p[k+1]:  " << params[k+1] << "  p[k]:  " << params[k] << endl; 
	    }
      fuzinput->ReplaceMF(1,new MFTRAPSUP(params[0],params[1],1.0));
    }
  else
    {
      for (int imf=1;imf<(nmf-1);imf++)
	{
	  if((params[k+1] - params[k]) < near)
	    { 
	      notnear = false; 
	      if(display) cout << "near:  " << near << "  p[k+1]:  " << params[k+1] << "  p[k]:  " << params[k] << endl; 
	      break; 
	    }

	  if (typemf[imf]==1) //trap
	    {
	      fuzinput->ReplaceMF(imf,new MFTRAP(params[k],params[k+1],params[k+2],params[k+3]));
	      k=k+2;
	    }
	  else //tri
	    {
	      fuzinput->ReplaceMF(imf,new MFTRI(params[k],params[k+1],params[k+2]));
	      k=k+1;
	    }
	}
      if((params[k+1] - params[k]) < near)
	{ 
	  notnear = false; 
	  if(display) 
	    cout << "near:  " << near << "  p[k+1]:  " << params[k+1] << "  p[k]:  " << params[k] << endl; 
	}
      
      fuzinput->ReplaceMF(nmf-1,new MFTRAPSUP(params[k],params[k+1],1.0));
    }
  if(notnear)
    {
      //update  ptr on fuzzy input
      delete FisFPOpt->In[NumFpOpt];
      FisFPOpt->In[NumFpOpt]=fuzinput;
    }
  else delete fuzinput;

  return (! notnear);

  /*if (display)
    {
    cout <<"\nafter update, fuzzy input number "<<NumFpOpt+1<<" is:";
    FisFPOpt->In[NumFpOpt]->Print();
    }
  */
}
int DistOptim::launchEntry(double &qual, bool display)
{
  // launch sw with eval fn
  //params are to be optimized-returns NULL if no fuzzy input
  // paramsFP will be modified by optim, T will not be modified
	 
  int rep=0,sizeparamsFP;
  double *paramsFP;//stored in container
  double * T=NULL;
  FILE * popt=NULL;
  //int * typemf;//stored in container
  char *partfilename=NULL;
  int nmf;
  int * typemf=NULL;
  double nearconst=0.;

  partfilename=new char[50];
      
  // get fuzzy input characteristics 
  //(fuzzy input  was built and arrays were allocated in PrepZoning)
  sizeparamsFP=*((int *)fEvalArgs->get(4));
  paramsFP=(double*)fEvalArgs->get(5);
  // display is fEvalArgs->get(6);
  nearconst = *(double*)fEvalArgs->get(7);

  T=new double[sizeparamsFP];
  for (int i=0;i<sizeparamsFP;i++)
    T[i]=paramsFP[i];
 
  if (display)
    {
      cout <<"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"<<endl;
      cout <<"before optim,FP params to optimize:"<<endl;
      for (int i=0;i<sizeparamsFP;i++)
	cout <<paramsFP[i]<<" ";
      cout <<endl;

      cout <<"Contrainte cnear:  " << nearconst <<endl;
    }
   
  //launch optim algo, which calls GlobalOptPam
  //Key, P, fEvalArgs: algo members
  rep=SolisWetts(paramsFP,&Key,T,sizeparamsFP,P,&qual,GlobalOptPam,fEvalArgs);

  if (rep==0 )
    {
      //solution found-keep FP parameters
      cout << "solution found, Quality="<<qual<<endl;
      if (display)
	{
	  cout <<"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"<<endl;
	  cout <<"after optim, new FP params"<<endl;
	  for (int i=0;i<sizeparamsFP;i++)
	    cout <<paramsFP[i]<<" ";
	  cout <<endl;
	}
      cout<<"unnormalized fuzzy partition"<<endl;
      //rebuild FIS - throw exception if pb
      FIS*fis=FisFPOpt->Clone();
      fis->In[NumFpOpt]->UnNormalize();
      fis->In[NumFpOpt]->Print();
      sprintf( partfilename, "%s%d%s", "opt-part-input",NumFpOpt+1,".txt");
      cout << partfilename<<" created"<<endl;//add numinput to partfile name
      popt=fopen(partfilename,"wt");
      fis->In[NumFpOpt]->PrintCfgCont(popt);
    }
  else
    {
      //no solution found-go back to previous FIS-done in main
      cout << "\nno solution found"<<endl;
    }
  if (popt!=NULL)
    fclose(popt);
  delete [] T;

  return rep;
}
////////////////////////////////////////////////
double DistOptim::optPam(
			 // a pointer to an Object U try to optimize
			 void *  Item  ,
			 // which part of object U optimize
			 vkey * mykey,
			 // array of double ( values )
			 double *Array ,
			 // Array size
			 int sizeArray )
		   
{
  double qualC=0.0;
  double *myData=(double *) Item;
  int ret, nmf=0;
  int * typemf=NULL;
  double nearconst=0.;
   
  //updates MyD(Item) with values in Array 
  for (int i=0;i<sizeArray;i++)
    myData[i]=Array[i];
      
  //get elements in container
  nmf=*((int *) fEvalArgs->get(2) );
  //cout <<endl<<"in optZ nmf="<<nmf<<" ";
  typemf=(int *) fEvalArgs->get(3);
  nearconst =*(double *) fEvalArgs->get(7);
      

     
  //update fuzzy distance parameters-rebuild MFs
  // attention: try-catch
     
  try{
    ret = UpdateFuzzyInputFromParams(nmf,typemf,sizeArray,myData,Display,nearconst);
    if (Display && ret) 
      cout<<"UpdateFuzzyInputFromParams failed - Near constraint violation"<<endl;
    if(ret) return(-1);
  }
  catch(std::exception &e){
    if (Display) 
      cout<<"UpdateFuzzyInputFromParams failed - MF constraint violation"<<endl;
    return(-1);
  }
  //
  //Fis partition is modified in class DistOptim
  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //go and get FisFPOpt in class, numc, norm, mink
  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  // compute new FP-based dissim. matrix
  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  if (Display) 
    {
      printf("\ncalling distWithNormedData, with FIS input#%d:",NumFpOpt+1);
      FisFPOpt->In[NumFpOpt]->Print();
    }
  FisFPOpt->distWithNormedData(NormData,Nbex,Ninputs,(char*)"diss.out",Numc,Norm,Mink,Display,false);
  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //clustering with fuzzy distance that has been updated
  // call R and return silhouette index
  qualC=::PamCluster(Display);
  if (Display)
    cout<<"in optPam qualC="<<qualC<<endl;
  return qualC;
}

///////////////////////////////////////////////////////////////////////
double PamCluster(bool display)
{
  int irep=0;
  FILE * fqual=NULL;
  char squal[100]= "";
  double qual=0.0;
  char * ptr=NULL;
  
  //calls R with pam.R script
  //pam.R script reads normed data and diss.out file
  //performs pam cluster and computes index
  // objective=index close to 1, so return 1-index->close to 0
  //writes silhouette index into silh file
  if(display)
    printf("\nR call with script pam.r");
  irep=system("R --slave <pamk3.R");
  fqual=fopen("silh","r");
  ptr=fgets(squal, 100, fqual);
  qual=strtod(squal,NULL);
  fclose(fqual);
  if(display)
    printf("\nsilhouette cluster mean width=%g\n",qual);
  //silhouette near 1=well clustered, so use 1-silh
  return (1.0-qual);
}
//*************************************************************************
int ReadInputsOpt(const char * stri, int nbmax,double *numinputs, bool wordless)
//*************************************************************************
{

  int nbinputs=0;
  char blank=' ';

  //attention if length 1 string =pb in SearchNb
  int lenin=strlen(stri);
  if (lenin==1)
    {
      nbinputs=1;
      numinputs[0]=atol(stri);
    }
  else
    nbinputs=SearchNb(stri,numinputs,nbmax,blank);
	
  if (!wordless)
    printf("\n %d inputs found:",nbinputs);
  if (!wordless)
    printf("\ninputs to optimize: %s \n",stri);
  for (int kk=0;kk<nbinputs;kk++)
    {
      if ((int)numinputs[kk]<1 ||(int)numinputs[kk]>nbmax)
	{
	  printf("\nerror in input string\n");
	  return 1;
	}
    }
  return nbinputs;
}
