//***********************************************************************

//
//
//                              GENSAMPLE.H
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : sampling functions used by FISPRO, part of library fispro

//**********************************************************************

#ifndef __GENSAMPLE_H
#define __GENSAMPLE_H

#include <stdio.h>
#include <iostream>
#include <fstream>
#include <ctype.h>
#include <string>
#include <stdlib.h> 
#include <math.h>
#include <stdexcept>
using std::ifstream;
using namespace std;

int intcompare(const void *p1, const void *p2);
//! function to compare two integers - compatible with call by qsort
//! Arguments are pointers on the two integer values to compare
//! returns -1, 0 or 1

void genNSamples(long int seed,double percV,int classif,double eps, int outCol, char * datafile, int nfiles, int flagDisplay=0);
//! generates nfiles pairs of sample validation and sample test files
//! calls Alloc1DDoubleWorkingArray,Alloc1DIntWorkingArray, Alloc2DDoubleWorkingArray
//! calls ReadSampleFile to read data array
//! if classif flag
//!                calls SortUniq to find classes in data array outCol
//!                in that case eps is the precision to discriminate classes
//!                creates a subarray for each class in data file
//!                then calls writeVTSample (classif case) for each subarray
//! if not classif flag
//!                calls genVTSample to write samples
//! the only difference between the two is that genVTSample opens and closes the files
//! while writeVTSample appends to them
//!
//! Arguments:
//!   seed:             to reproduce the same random sequence
//! except if seed=0, a new random sequence based on machine clock at each call
//!   percV:            validation file/data file size ratio (number of rows)
//!   classif:          1 - sampling is done with respect to data file class ratio
//!   eps:              precision for finding classes in data file
//!   outCol:           column number for classes in data file
//!   datafile:         data file name
//!   nfiles:           number of samplings: 1 validation file + 1 test file created per sample
//!   flagDisplay:      1- detailed display

void genSplit(long int seed, int truncate, int classif,double eps, int outCol, char * datafile, int nfiles, int flagDisplay=0);
//! generates nfiles sample  files
//! calls Alloc1DDoubleWorkingArray,Alloc1DIntWorkingArray, Alloc2DDoubleWorkingArray
//! calls ReadSampleFile to read data array
//!                calls genVTSample to write samples
//! the only difference between the two is that genVTSample opens and closes the files
//! while writeVTSample appends to them
//!
//! Arguments:
//!   seed:             to reproduce the same random sequence
//! except if seed=0, a new random sequence based on machine clock at each call
//!   nfiles:           number of samplings: 1 validation file + 1 test file created per sample
//!   flagDisplay:      1- detailed display

void genVTSample(double ** Examples, int nrow, int ncol, int sizeV, char * fileVname,char * fileTname,int flagDisplay=0);
//! opens files 
//! and calls writeVTSample which generates 1 valid. sample (sizeV) and 1 test. sample (remaining items) from the Examples array
//! and writes them to the files fileVname and fileTname
//! Arguments:
//!   Examples:                  data array
//!   nrow,ncol:                 data array size (number of rows, number of columns)
//!   sizeV:                     number of rows of validation sample
//!   fileVname, fileTname:      file names for validation and test samples
//!   flagDisplay:               1- detailed display

void genVTSampleSplit(double ** Examples, int & nrow, int ncol, int sizeV, char * fileVname,int flagDisplay=0);
//! opens files 
//! and calls writeVTSample which generates 1 valid. sample (sizeV) and 1 test. sample (remaining items) from the Examples array
//! and writes them to the files fileVname and fileTname
//! Arguments:
//!   Examples:                  data array
//!   nrow,ncol:                 data array size (number of rows, number of columns)
//!   sizeV:                     number of rows of sample
//!   fileVname:      file name for sample
//!   flagDisplay:               1- detailed display

void writeVTSample(double ** Examples, int nrow, int ncol, int sizeV, FILE * fileV, FILE * fileT,int flagDisplay=0);
//! files must be already opened - appends values to them
//! values are drawn randomly from the Examples array by finding a series of row indexes
//! calls FisRand for random sampling
//! Arguments:
//!   Examples:                  data array
//!   nrow,ncol:                 data array size (number of rows, number of columns)
//!   sizeV:                     number of rows in the validation sample
//!                              the remaining rows of the data file will be the test sample
//!   fileV,fileT:               validation and test file pointers - must be opened before the call to writeVTSample
//!   flagDisplay:               1- detailed display
  
void writeVTSampleSplit(double ** Examples, int & nrow, int ncol, int sizeV, FILE * fileV,int flagDisplay=0);
//! files must be already opened - appends values to them
//! values are drawn randomly from the Examples array by finding a series of row indexes
//! calls FisRand for random sampling
//! Arguments:
//!   Examples:                  data array
//!   nrow,ncol:                 data array size (number of rows, number of columns)
//!   sizeV:                     number of rows in the validation sample
//!                              the remaining rows of the data file will be the test sample
//!   fileV,fileT:               validation and test file pointers - must be opened before the call to writeVTSample
//!   flagDisplay:               1- detailed display

int * Alloc1DIntWorkingArray(int n);
//! allocates a 1D int array of size n - elements initialized to 0 -   throws exception
//! returns pointer on integer array

double * Alloc1DDoubleWorkingArray(int n);
//! allocates a 1D double array of size n -  elements initialized to 0.0- throws exception
//! returns pointer on double array

double *** Alloc3DDoubleWorkingArray(int nrow,int ncol,int k);
//! allocates a 3D double array of size nrow x ncol x k -  elements initialized to 0.0- throws exception
//! 

double ** Alloc2DDoubleWorkingArray(int nrow, int ncol);
//! allocates a 2D double array of size nrow x ncol- elements initialized to 0.0 - throws exception
//! returns double pointer on double array

void del2DArray(double ** x, int nrow);
//! deletes a 2D array with nrow rows - throws exception

void del3DArray(double *** x, int i2,int i3);
//! deletes a 2D array with i2 rows, i3 columns - throws exception

FILE * openFileW(char * fileName);
//! opens file fileName for writing - throws exception
//! returns file pointer

void genSubSample(double *** & subExamples, int nrow, int ncol, int outCol, double eps, double ** Examples, double * uniq, int * occur, int nuniq, int maxcl, int flagDisplay=0);

#endif
//*************************  GENSAMPLE.H    **************************

