//*********************************************************************
//
//
//                              KM.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// Object : Test kmeans basic functions
// 
//
//*********************************************************************
#include "common.h"


void Normalize ( double &center,  double min, double max)
  //******************************************************************************
  // normalize between 0 and 1 center data which are between min and max
  // updates center array
  //******************************************************************************
{
  double ratio = max - min;
  center=(center - min)/ratio;
  return;
}

void UnNormalize ( double &center, double min, double max)
  //********************************************************************************
  // unnormalize a center value corresponding to data initially between min and max
  // and that were previously normalized between 0 and 1
  // updates center array
  //********************************************************************************
{
  double ratio = max - min;
  center=center* ratio +  min;
  return;
}

void InitCentresND(double **&C, int ng, int dim, double *min, double *max)
  //**********************************************************************
  // allocates and fills multimensional C array
  // ng C values are taken uniformly between min and max
  // in each dimension (0 to dim)
  //**********************************************************************
{
  double inc;
  int i,k;

  C = new double* [ng];
   
  for(k = 0; k < ng ;k++)
    {
      C[k]=new double [dim];
      for(i = 0; i < dim ;i++)
    	{
	  inc = (max[i] - min[i])/(ng - 1);   
	  C[k][i] = min[i] + k*inc;
	}
    }
  
}

void ResetCentresND(double **C,int ng, int dim, double *min, double *max)
{
double inc;
  int i,k;
  //***********************************************************************
  // ng number of clusters
  // dim number of dimensions
  // min,max array of boundaries for each variable
  // like InitCentresND, but C must be preallocated
  //***********************************************************************
for(k = 0; k < ng ;k++)
    {
      for(i = 0; i < dim ;i++)
    	{
	  inc = (max[i] - min[i])/(ng - 1);   
	  C[k][i] = min[i] + k*inc;
	}
    }
}
void printData(double **data,int nbrow,int nbcol)
{
//***********************************************************************
// print data array
//***********************************************************************
  
  int i,j;
  printf("\ndonnees\n");
  for(i = 0; i <nbrow; i ++)   //
    {
      printf("\nligne %d\t",i+1);
    for(j = 0; j < nbcol; j++) //
      printf("%5.2f\t",data[i][j]);  
    }
  printf("\n");
}

void CountClusters(double *T, int ni, int nc, double * C, int *&cnt, int *&clust)
//***********************************************************************
// T=data column, ni=number of elements, C = cluster centers,
// allocates and fills cnt=cluster count, clust=cluster membership
//***********************************************************************
{
  int i,cl;
  cnt = new int [nc];
  clust=new int [ni];
  

  for(i = 0; i < nc ; i++) cnt[i] = 0;

  for(i = 0; i < ni; i++) 
    {
      cl=AssignClas(T[i],C,nc);
      clust[i]=cl;
      cnt[cl]++;
    }
}

void CountClustersND(double **T, int dim, int ni, int nc, double **C, int *&cnt,int *&clust)
//*******************************************************************************
// T=data array, dim=number of variables, ni=number of rows, C = cluster centers,
// allocates and fills cnt=cluster count, clust=cluster membership
//*******************************************************************************
{
int i,cl;
  cnt = new int [nc];
  clust=new int [ni];
  

  for(i = 0; i < nc ; i++) cnt[i] = 0;

  for(i = 0; i < ni; i++) 
    {
      cl=AssignClas(T[i],dim,C,nc);
      clust[i]=cl;
      cnt[cl]++;
    }
}

void WithinSS(double *T, int ni, double *cg, int ng, double *withinss)
  //*********************************************************************
  // from T data array and ng clusters with cg centers
  // compute within sum of squares for each cluster
  // withinss array must be preallocated
  //*********************************************************************
{
  int i,k,cl;
  
  
   for(k = 0; k < ng; k++)
    {
      withinss[k] = 0.;
      for(i = 0; i < ni; i++) 
	{
	  cl=AssignClas(T[i],cg,ng);
	  if (cl==k) 
	    withinss[k] += (T[i]-cg[k])*(T[i]-cg[k]);

	}

    }
}

void WithinSSND(double **T, int dim, int ni, double **cg, int ng, double *withinss)
  //*********************************************************************
  // multidimensional version, otherwise identical to previous function
  // from T data array and ng clusters with cg centers
  // compute within sum of squares for each cluster
  // withinss array must be preallocated
  //*********************************************************************
{
  int i,k,d,cl;
  
  
   for(k = 0; k < ng; k++)
    {
      withinss[k] = 0.;
      for(i = 0; i < ni; i++) 
	{
	  cl=AssignClas(T[i],dim,cg,ng);
	  if (cl==k) 
	    for(d = 0; d < dim; d++)
	      withinss[k] += (T[i][d]-cg[k][d])*(T[i][d]-cg[k][d]);

	}

    }
}


//**************************   End of KM.CPP   **************************
