//***********************************************************************

//
//
//                              KM.H
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : Out of class functions used by FISPRO, part of library fispro

//**********************************************************************

#ifndef __KM_H
#define __KM_H
#include "common.h"

//***********************************************************************
// normalize between 0 and 1 data1d 
// which are between min and max
// updates data1d array
//**********************************************************************
void Normalize ( double &data1d, double min, double max);

//***********************************************************************
// unnormalize data1d which were initially between min and max
// and were previously normalized between 0 and 1
// updates data1d array
// used to unnormalize cluster centers in multidimensional kmeans
//**********************************************************************
void UnNormalize ( double &data1d, double min, double max);

//**********************************************************************
// allocates and fills multimensional C array
// ng C values are taken uniformly between min and max
// in each dimension (0 to dim)
//**********************************************************************
void InitCentresND(double **&C, int ng, int dim, double *min, double *max);

//************************************************************************
// ng number of clusters
// dim number of dimensions
// min,max array of boundaries for each variable
// like InitCentresND, but C must be preallocated
//***********************************************************************
void ResetCentresND(double **C,int ng, int dim, double *min, double *max);

//*************************************************************************
// print data array
//*************************************************************************
void printData(double **data,int nbrow,int nbcol);

//***********************************************************************
// T=data column, ni=number of elements, C = cluster centers,
// allocates and fills cnt=cluster count, clust=cluster membership
//***********************************************************************
void CountClusters(double *T, int ni, int nc, double * C, int *&cnt,int *&clust);

//*******************************************************************************
// T=data array, dim=number of variables, ni=number of rows, C = cluster centers,
// allocates and fills cnt=cluster count, clust=cluster membership
//*******************************************************************************
void CountClustersND(double **T, int dim, int ni, int nc, double **C, int *&cnt,int *&clust);


//*********************************************************************
// from T data array and ng clusters with cg centers
// compute within sum of squares for each cluster
// withinss array must be preallocated
//*********************************************************************
void WithinSS(double *T, int ni, double *cg, int ng, double *withinss);

//*********************************************************************
// from T data array and ng clusters with cg centers
// compute within sum of squares for each cluster
// withinss array must be preallocated
//************************************************************************************
void WithinSSND(double **T, int dim, int nrow, double **cg, int ng, double *withinss);

//*************************  KM.H    *************************************************

#endif
