//***********************************************************************
//*** FILENAME: list.h                                                ***
//*** AUTHOR: Hazal JONES (jones@ensam.inra.fr)                      ***
//*** Release Date: July 13, 2005                                     ***
//***********************************************************************
//*** DESCRIPTION:                                                    ***
//*** This is a double linked list implementation.                    ***
//***********************************************************************
//*** REVISION HISTORY                                                ***
//***********************************************************************

class POINT
{
  public :
    double x,y;

  POINT(){ x=0.; y=0.;};
  POINT(double xp,double yp){x=xp; y=yp;};
  ~POINT()
  	{
  		x=0.;
  		y=0.;
  	};
  
  void   dup( POINT* P2){x=P2->x; y=P2->y;return;};
  
  double Norme(){ return sqrt(x*x+y*y);};
  
  void   Print()
  {
  	 cout << "POINT " << x << "  " << y << "  " << endl;
  	 return;
  };
  
  bool operator== (const POINT &pt) const
    {
       	 return ((x >= pt.x-EPSILON) && (x <= pt.x+EPSILON) &&
      	         (y >= pt.y-EPSILON) && (y <= pt.y+EPSILON));
    }
    
  bool operator!= (POINT  &pt)
    {
      
      return (!(*this==pt)) ;
    }   
  	 
  POINT* operator= (POINT* pt)
    {
      x = pt->x;
      y = pt->y;
      return this;
    }
};


class LIST
{
 private:
  struct data
  {
    POINT*  pt;
    data*	next;
    data*	prev;
  };
  
  data*	head;				//Always points to first element in list
  data*	tail;				//Always points to last element in list
  data*	current;			//Always points to current element in list
  int	size;				//Holds the true(based on 1) size of list
  long	index;				//This is the current index
  
 public:

  //********************************
  //*** CONSTRUCTOR & DESTRUCTOR ***
  //********************************
  LIST()
    {
      head = tail = current = 0;
      size = 0;
      index = -1;
    }
  
  ~LIST()
    {
	  	home();
	  	while(!is_empty())
			rem_data();
    }
	
	
  //**************************
  //*** MOVEMENT FUNCTIONS ***
  //**************************
  int	next()
    {
      if(!is_empty())
	  {
		  if(current->next)
		  {
		      current = current->next;
		      index++;
		      return 1;
		  }
		  
	  }
      return 0;	
    }
  
  int	prev()
    {
      if(!is_empty())
	if(current->prev)
	  {
	    current = current->prev;
	    index--;
	    return 1;		
	  }
      return 0;
      
    }
 	
  int	home()
    {
      if(is_empty())
		return 0;
      current = head;
      index = 0;
      return 1;
    }
	
  int	end()
    {
      if(is_empty())
	return 0;
      current = tail;
      index = size-1;
      return 1;
    }
	
  void goto_index(long where)
    {
      if(where == index)
	return;
      if( where > index )
	while( where>index && next() );
      else
	while( where<index && prev() );
    }
	
	
  //**********************
  //*** DATA FUNCTIONS ***
  //**********************

  void change_point(POINT* new_point)
  {
  	current->pt = new_point;
  }
	
  void add_data(POINT* new_point)
    {
      //printf("\n in add_data\n");fflush(stdout);
      data* temp = new data;
      
      temp->next = 0;
      temp->prev = 0;
      temp->pt = new_point;

      if(end())
		{
	  	//printf("\n if end\n");fflush(stdout);

	  	current->next = temp;

	  	temp->prev = current;

		}
      else
		{	  
	 	 //printf("\n if not end\n");fflush(stdout);
	 	 head = temp;
		}
      size++;
      index = size-1;
      current = temp;
      tail = current;

      //printf("\n current->pt->x : %8.3f, current->pt->y : %8.3f \n",current->pt->x,current->pt->y);fflush(stdout);
      //printf("\n end of add_data\n");fflush(stdout);
      
      temp=NULL;
      //delete temp;
      //printf("\n delete temp just done \n");fflush(stdout); 
      //printf("\n current->pt->x : %8.3f, current->pt->y : %8.3f \n",current->pt->x,current->pt->y);fflush(stdout);     
    }
  
  void add_atpos(POINT* new_point)
    {
		
      if(is_empty())
		add_data(new_point);
      else
	{	
	  data* temp = new data;
	  temp->pt = new_point;
	  
	  if(is_home())
	    {
	      temp->next = current;
	      current->prev = temp;
	      temp->prev = 0;
	      head = temp;
	      home();
	    }
	  else
	    {
	      current->prev->next = temp;
	      temp->prev = current->prev;
	      temp->next = current;
	      current->prev = temp;
	    }
	  current = temp;
	  size++;
	}
    }

  // compare the x coordinates of 2 points
  //if x coordinates are equal, compare y coordinates
  int compare(POINT* pt1, POINT* pt2)                          
    {
      if( pt1->x  >  pt2->x )
	return 1;
      if( pt1->x  <  pt2->x )
	return -1;
      if( pt1->x == pt2->x)
	{
	  if( pt1->y  >  pt2->y )
	    return 1;
	  if( pt1->y  <  pt2->y )
	    return -1;
	}
      return 0;
    }


  void add_sort(POINT* new_point)
  {
  	//printf("\n add_sort function\n");fflush(stdout);
    //if(current!=NULL)
	//printf("\n current->pt->x : %8.3f, current->pt->y : %8.3f \n",current->pt->x,current->pt->y);fflush(stdout);

    if(is_empty())
	{
	  //  printf("\n if empty \n");fflush(stdout);
	  add_data(new_point);	  
	}
    else
	{
	  	//printf("\n if not empty \n");fflush(stdout);
	  	//printf("\n current->pt->x : %8.3f, current->pt->y : %8.3f \n",current->pt->x,current->pt->y);fflush(stdout);
		if( find_data(new_point) )
		{
			//  printf("\n if find_data \n");fflush(stdout);
			add_atpos(new_point);
		}
	  	else
	    {	     
	      	//printf("\n if not find_data \n");fflush(stdout);
	      	add_data(new_point);
	    }
	}
      //printf("\n end of add_sort function\n");fflush(stdout);
  }
	

  data* get_next()
    {
      // controler...
      if( (!is_empty()) && (current) )
	return(current->next);
      else
	return 0;
    }

  POINT* get_prev_data()
    {
      if( (!is_empty()) && (current) && (current->prev!=NULL) )
		return(current->prev->pt);
      else
		return NULL;
    }
	
  POINT* get_next_data()
    {
      if( (!is_empty()) && (current) && (current->next!=NULL) )
		return(current->next->pt);
      else
		return NULL;
    }

  POINT* get_data()
    {
      if( (!is_empty()) && (current) )
	return(current->pt);
      else
	return NULL;
    }
	
  POINT* get_atpos(long where)
    {
      if(!is_empty())
	{
	  goto_index(where);
	  return get_data();
	}
      return NULL;
    }
	
  void rem_data()
    {
      
      data* temp;
      //delete current->pt;
      if(!is_empty())
	  {
	  	if(is_home())
	    {
	      head = head->next;
	      if(head)
			head->prev = 0;
	      delete current;
	      home();
	    }
	  	else
	    {
	      temp = current->prev;
	      temp->next = current->next;
	      if(!is_end())
			current->next->prev = temp;
	      else
			tail = temp;
	      delete current;
	      current = temp;	
	      index--;
	    }
	  size--;
	  }
    }
	
	
  //************************
  //*** SEARCH FUNCTIONS ***
  //************************ 
  int find_data(POINT* search_data)
    {		
      int value;
       
      //printf("\n FIND DATA \n");fflush(stdout);

      //tester le get_data() TO DO !!!!
      //printf("get_data :  x =  %8.3f  y : %8.3f\n",get_data()->x,get_data()->y); fflush(stdout);
      //printf("search_data : x =  %8.3f  y : %8.3f\n", search_data->x, search_data->y);fflush(stdout);


      //value = compare(search_data,get_data());
      value = 1; //because value must be enter in the way of the MF
      switch(value)
	{
	case -1:				//Come in from above
	  //  printf("value = -1 : donne avant \n");fflush(stdout);
	  while( ( value = compare( search_data,get_data() ) ) < 0)
	    {
	      if(!prev())
		return -1;
	    }
	  if(value)
	    {
	      next();
	      return -1;
	    }
	  else
	    return 1;
	  break;
	case 1:						//Come in from below
	  //printf("value = 1 : donne apres \n");fflush(stdout);

	  while( /*(*/ value /*= compare( search_data,get_data() ) )*/ > 0)
	    {
	      if(!next())
		return 0;
	    }
	  if(value)
	    {
	      return -1;
	    }
	  else
	    return 1;
	  break;
	case 0:							//Exact Value
	  //printf("value = 0 : donne exacte \n");fflush(stdout);

	  return 1;
	  break;
	}      
      //printf("\n END OF FIND DATA \n");fflush(stdout);
      return 0;
      
    }
	
	
  //***********************
  //*** QUERY FUNCTIONS ***
  //***********************
  int is_empty()
    {
      if(head != 0)
		return 0;
      return 1;
    }
  
  int is_home()
    {
      if(!is_empty())
	if(current == head)
	  return 1;
      return 0;
    }
  	
  int is_end()
    {
      if(!is_empty())
	if(current == tail)
	  return 1;
      return 0;
    }
  	
  long cur_index()
    {
    return index;
    }
	
  int get_size()
    {
      return size;
    }
	
	
  //***********************************
  //*** MARKING/DEMARKING FUNCTIONS ***
  //***********************************
  void* set_mark()
    {
      if(!is_empty())
	return ((void*) current);
    }

  void goto_mark(void* mark)
    {
      current = (data*) mark;
      reindex();
    }
	
  void reindex()
    {
      data*	hold = current;
      
      home();
      while(hold != current)
	next();
      
    }
 
};
