//*********************************************************************
//
//
//                              MAIN_INTER.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 2.1
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  September 15, 2004
// Object : Test FIS basic functions
// Environnement : Uses library fispro
//
//*********************************************************************

#include "fis.h"


//! This program needs 2 arguments:  the first MFDPOSS file name  and 
//! the secund MFDPOSS file name.

int main(int argc, char ** argv)
  //******************************
{
	printf("\n*********BEGIN UNIT TESTS\n");fflush(stdout);		 		 
	printf("\n GetDeg from a trapeze\n");fflush(stdout);
	MFTRAP *tmp = new MFTRAP(0.0, 0.5, 1.0, 2.0);
	MFDPOSS *mfd = new MFDPOSS(tmp, 0.);
	delete tmp;
    double x, res;
    x = 0.0;   res =0.;
    printf("x %8.3f  res %8.3f  GetDeg %8.3f\n",x,res,mfd->GetDeg(x)); 
    x = 0.25;   res =0.5;
	printf("x %8.3f  res %8.3f  GetDeg %8.3f\n",x,res,mfd->GetDeg(x)); 
    x = 0.5;    res =1.0;
    printf("x %8.3f  res %8.3f  GetDeg %8.3f\n",x,res,mfd->GetDeg(x)); 
    x= 0.75;    res =1.0;
    printf("x %8.3f  res %8.3f  GetDeg %8.3f\n",x,res,mfd->GetDeg(x)); 
    x = 1.333;  res =0.667;
    printf("x %8.3f  res %8.3f  GetDeg %8.3f\n",x,res,mfd->GetDeg(x)); 
    x =2.5;     res =0.;
    printf("x %8.3f  res %8.3f  GetDeg %8.3f\n",x,res,mfd->GetDeg(x)); 
    
	printf("\n computeArea from a trapeze\n");fflush(stdout);
	res = 1.25;
    printf("res %8.3f  computeArea %8.3f\n", res, mfd->computeArea()); 
	delete mfd;
	try{
		char * ficres=NULL;
		char * ficgpl = NULL;
		FILE *fgres, *fgplt, *ftmp;
		MFDPOSS *mf1, *mf2, *mfinter, *mfunion, *mfjoin;
		 
		mfinter = mfunion = mfjoin = NULL;

		// construction of the first MFDPOSS
		if (argc>1)
		{
			ftmp=fopen(argv[1],"r");
			if(ftmp != NULL)
			  {
			    mf1 = new MFDPOSS(argv[1]);
			    fclose(ftmp);
			  }
			else
			{
			  printf(" the first file is not found \n");
			  exit(-1);
			}
		}
		else 
		{
			mf1 = new MFDPOSS(new MFDOOR( 1., 2.),0);
		}
		mf1->Simplify();		
		printf("\nFIRST MFDPOSS %d\n",  mf1->NbParams());fflush(stdout);	
		mf1->Print();
		
		// construction of the secund MFDPOSS
		if (argc>2)
		{
		  ftmp=fopen(argv[2],"r");
		  if(ftmp != NULL) 
		    {
		      mf2 = new MFDPOSS(argv[2]);
		      fclose(ftmp);
		    }
		  
		  else
		    {
o		      printf(" the secund file is not found \n");
		      exit(-1);
		    }
		}
		else 
		  {
		    mf2 = new MFDPOSS((MF*) new MFDOOR( 1., 2.),0);
		}
		mf2->Simplify();		
		printf("\nSECUND MFDPOSS %d\n",  mf2->NbParams());fflush(stdout);	
		mf2->Print();
		
		// computation of the intersection
		printf("\nINTERSECTION \n");fflush(stdout); 
		mfinter = mf1->Inter(mf2);
		if ((mfinter != NULL) && (mfinter->NbParams()>0)){  
			mfinter->Print();
		}
		// computation of the union 
		printf("\nUNION \n");fflush(stdout);
		mfunion = mf1->Union(mf2);
		if ((mfunion != NULL) && (mfunion->NbParams()>0)){  
			mfunion->Print();
		}
		// computation of the join
		printf("\nJOIN \n");fflush(stdout);
		mfjoin = mf1->Join(mf2);
		if ((mfjoin != NULL) && (mfjoin->NbParams()>0)){  
			mfjoin->Print();
		}	
			
		// computation of the range
		double ls,rs,lsmin,rsmax;
		mf1->Support(lsmin,rsmax);
		mf2->Support(ls,rs);
		if(ls<lsmin) lsmin =ls;
		if(rs>rsmax) rsmax =rs;
				
		// generation of the gnuplot data file
		ficres="plotfis.res";
		fgres = fopen(ficres, "wt");
		printf("le fichier donnees pour GNUPLOT est %s\n", ficres);	
					
		fprintf(fgres,"#FIRST MFDPOSS \n");	
		mf1->Print(fgres);
		fprintf(fgres, " \n\n");
		
		fprintf(fgres,"#SECUND MFDPOSS \n");	
		mf2->Print(fgres);
		fprintf(fgres, " \n\n");
				
		fprintf(fgres,"#INTERSECTION \n");	
		if ((mfinter != NULL) && (mfinter->NbParams()>0)){ 
    		mfinter->Print(fgres);
    		delete mfinter;
		} else {                            
   			fprintf(fgres,"  %8.3f  0 \n", lsmin);
   			fprintf(fgres,"  %8.3f  0 \n", rsmax);
		}
		fprintf(fgres, " \n\n");
		
		fprintf(fgres,"#UNION \n");	
		if ((mfunion != NULL) && (mfunion->NbParams()>0)){ 
    		mfunion->Print(fgres);
    		delete mfunion;
    	} else {                            
   			fprintf(fgres,"  %8.3f  0 \n", lsmin);
   			fprintf(fgres,"  %8.3f  0 \n", rsmax);    	
		}
		fprintf(fgres, " \n\n");
		
		fprintf(fgres,"#JOIN \n");	
		if ((mfjoin != NULL) && (mfjoin->NbParams()>0)){ 
    		mfjoin->Print(fgres);
    		delete mfjoin;
		} else {                            
   			fprintf(fgres,"  %8.3f  0 \n", lsmin);
   			fprintf(fgres,"  %8.3f  0 \n", rsmax);    	
		}		
		fprintf(fgres, " \n\n");		
		
		fclose(fgres);		
		delete mf1;
		delete mf2;					
		// generation of the gnuplot command file
		ficgpl = "plotfis.gpl";
		fgplt = fopen(ficgpl, "wt");
		printf("le fichier commandes pour GNUPLOT est %s\n", ficgpl);
		
		fprintf(fgplt, "set tmargin 1\n");
		fprintf(fgplt, "set bmargin 0.5\n");
		fprintf(fgplt, "set lmargin 4\n");
		fprintf(fgplt, "set rmargin 4\n");					
		fprintf(fgplt, "set yrange[0:1.1]\n");				
		fprintf(fgplt, "set xrange[%8.3f:%8.3f]\n", lsmin-0.1, rsmax+0.1);

		fprintf(fgplt, "set multiplot \n");
		fprintf(fgplt, "set size 1.0, 0.3\n");			
				
		fprintf(fgplt, "set title \"TEST INTERSECTION\" offset 0,-1\n"); 
		fprintf(fgplt, "set origin %8.3f,%8.3f \n",0.0, 0.666); 
		fprintf(fgplt, "plot \'%s\' using 1:2 index %d notitle  with filledcu,\\\n", ficres, 2);
		fprintf(fgplt, " \'%s\' using 1:2 index %d notitle  with lines lw 2,\\\n", ficres, 1);
		fprintf(fgplt, " \'%s\' using 1:2 index %d notitle  with lines lw 2, \\\n",ficres, 0);								
		fprintf(fgplt, " \'%s\' using 1:2 index %d notitle  with points ps 4\n",ficres, 2);								

		fprintf(fgplt, "set title \"TEST UNION\" offset 0,-1\n"); 
		fprintf(fgplt, "set origin %8.3f,%8.3f \n",0.0, 0.333); 
		fprintf(fgplt, "plot \'%s\' using 1:2 index %d notitle  with filledcu,\\\n", ficres, 3);
		fprintf(fgplt, " \'%s\' using 1:2 index %d notitle  with lines lw 2,\\\n", ficres, 1);
		fprintf(fgplt, " \'%s\' using 1:2 index %d notitle  with lines lw 2,\\\n",ficres, 0);								
		fprintf(fgplt, " \'%s\' using 1:2 index %d notitle  with points ps 4\n",ficres, 3);	
									
		fprintf(fgplt, "set title \"TEST JOIN\" offset 0,-1\n"); 
		fprintf(fgplt, "set origin %8.3f,%8.3f \n",0.0, 0.0); 
		fprintf(fgplt, "plot \'%s\' using 1:2 index %d notitle  with filledcu,\\\n", ficres, 4);
		fprintf(fgplt, " \'%s\' using 1:2 index %d notitle  with lines lw 2,\\\n", ficres, 1);
		fprintf(fgplt, " \'%s\' using 1:2 index %d notitle  with lines lw 2,\\\n",ficres, 0);								
		fprintf(fgplt, " \'%s\' using 1:2 index %d notitle  with points ps 4\n",ficres, 4);								
	

		fprintf(fgplt, "unset multiplot\n");
		fprintf(fgplt, "pause -1 \"Hit return to continue\"\n");
		fprintf(fgplt, "reset\n");
		fclose(fgplt);

	}
	catch(std::runtime_error &e){
		printf("catch %s\n",e.what());
		return(-1);
	}
	printf("\n******************END UNIT TESTS********************\n");fflush(stdout); 
}

//**************************   End of MAIN_INTER.CPP   *************************
