//*********************************************************************
//
//
//                              MAINFUZ2CRISP.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// Object : Call performance function
// Environnement : Uses library fispro

//*********************************************************************

#include "fis.h"
#include "common.h"

//! This program needs 2 arguments:  the FIS file name, the number of output MFs 
int main(int argc, char ** argv)
  //******************************
{
  if(argc < 3)
    {
      printf("\nThe program %s needs 3 arguments: ", argv[0]);
      printf("\n\t* the FIS file name.") ;
      printf("\n\t* the number of output MFs") ;
      printf("\nOptional arguments : ");
      printf("\n\t* -m (set max disjunction)");
      printf("\n\t* -n outputNumber (default=0 first output)");
      printf("\n\t* -a (detailed display)\n");
      return 1;
    }

  
  int i,nOut,len,outputN=0,flagDisplay=0,disjmax=0,nMF;
  double dyn, * centers=NULL;
  FIS * S;
  FILE *fg;
  char * outputFisConfigName=NULL;
  
  // main loop for catching exceptions in called functions
  try{
  
   
    if(argc > 3) 
      {
	for(i = 3; i < argc; i++)
	  {
	    if(! strncmp("-m", argv[i], 2) ) disjmax=1;
	    if(! strncmp("-o", argv[i], 2) ) outputN=atoi(argv[i] + 2);
	    if(!strcmp("-a", argv[i]) ) flagDisplay=1;
	  }
      }
    nMF=atoi(argv[2]);
    if(nMF <= 0)
      {
	printf("no MFS in output, exiting");
	return 1;
      }
    
    printf("\ncrisp output %d transformed into fuzzy output with %d MFs, using regular grid PFF\n",outputN,nMF);
   
    
   
    if (disjmax ==1)
      printf("disjunction operator set to max for output %d\n",outputN);
    else
      printf("disjunction operator set to sum for output %d\n",outputN);


   
    S = new FIS(argv[1]);  
    // check output number
    nOut=S->GetNbOut();
    if (nOut <=0)
      {
	sprintf( ErrorMsg, "~invalid~number~of~outputs~: %d", nOut);
	throw std::runtime_error( ErrorMsg );
      }
    if (outputN >(nOut-1))
      {
	sprintf( ErrorMsg, "~invalid~output~number: %d", outputN);
	throw std::runtime_error( ErrorMsg );
      }

    // compute centers
    // Generate the grid
    dyn = (S->Out[outputN]->min()) - (S->Out[outputN]->max());
   
    centers=new double[nMF];
    centers[0]=S->Out[outputN]->min();
    centers[nMF-1]=S->Out[outputN]->max();
    if (nMF>1) dyn /= (nMF - 1);
    for (int i=1;i<(nMF-1);i++)
      {
	centers[i]=centers[i-1]+ dyn;
      }
    if (flagDisplay)
      {
	for (int i=0;i<nMF;i++)
	  printf("centers[%d]=%f",i,centers[i]);
      }

    S->Crisp2Fuz(outputN,"area",centers,nMF); //defuzzification="area"
   
    if (disjmax ==1) 
      {
	S->Out[outputN]->SetOpDisj( (char *)OUT_FUZZY::DisjMax());
      }

    // Generation of a config file
    len=strlen(argv[1]) + 25;
    outputFisConfigName=new char[len];
    strcpy(outputFisConfigName,argv[1]);
  
	if (disjmax ==1)
	  strcat(outputFisConfigName,"-fuzzy-max.fis");
	else
	  strcat(outputFisConfigName,"-fuzzy-sum.fis");
      
    //
    fg = fopen(outputFisConfigName, "wt");
    S->PrintCfg(fg);
    fclose(fg);

    delete S;
    return (0);
  }
  //action in case of exception
  catch(std::exception &e){
    printf("error in mainfuz2crisp\n");
    printf("catch %s\n",e.what());
    
    return(-1);
  }
}


//**************************   End of MAINSAMPLE.CPP   *************************

