//*********************************************************************
//
//
//                              PERF.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// Object : Call performance function
// Environnement : Uses library fispro

//*********************************************************************

#include "fis.h"
#include "algo.h"
#include "distoptim.h"

//! This program needs 2 arguments:  the fis configuration file name and 
//! the data file name. Options: -a for intermediate display ; 
//! -s activity threshold for an item not to be blank (default: 0.2).
int main(int argc, char ** argv)
//******************************
{
  
  if(argc < 3)
    {
      printf("\nThe program %s needs 2 arguments : ", argv[0]);
      printf("\n\t* the fis configuration file name;");   
      printf("\n\t* the data file name.") ;
      printf("\n\t options");
      printf("\n\t -m Minkowski power (default 2)");
      printf("\n\t -c '0 1' string Euclidean/fuzzy(default fuzzy distance  in all dimensions)");
      printf("\n\t* -in \"x y \", the string of input numbers to optimize (starting at 1, order is important, default no input optimization;"); 
      //  printf("\n\t -o to optimize fuzzy partition");
      printf("\n\t -it number of iterations in sw");
      printf("\n\t -g gaussian noise range in sw");
      printf("\n\t -s random seed in sw");
      printf("\n\t -n nearconstraint between adjacent kernels in [0,1] (default 0.)");	
      printf("\n\t -wl for wordless");
      printf("\n\t -a for detailed display");
      printf("\n");
      return 1;
    }
  double mink=2.0;
  bool mixed=false,first=true;
  int argstrpos=0;
  bool wordless=false;
  bool norm=true;
  bool flagDisplay=false;
  double * numc=NULL;
  int nbc=0,nbinputs=0,nbinpdata=0,nbex=0,nbmax=0,nbopt=0;
  char blank=' ';
  double **diss=NULL;
  // param optim FP
  double gauss=0.01;
  bool OPTIM=false, display=true;
  int i=0,j=0,maxi=10;
  int * iFP=NULL;
  double *pFP=NULL;
  FISIN * ptr=NULL;
  int sizeparamsFP=0;
  vkey key;
  double qualC=0.0;
  int rep=0,nmf=0;
  double **normdata=NULL;
  FILE * nf=NULL;
  DistOptim* fpopt=NULL;
  FIS * S=NULL, *S1=NULL,*S2=NULL;
  int argstr=0,numinput=0;
  double * numopt=NULL;
  double nearconstraint=0.;
  long int seed1=0;

  if(argc > 3) 
    {
      for(int i = 3; i < argc; i++)
        {
	  if(! strcmp("-c", argv[i])) 
	    {
	      mixed=true;
	      //next argument will hold string of combinations Euclidean and FP-based
	      argstrpos=i+1;
	    }

	  if(! strncmp("-m", argv[i], 2)) 
	    {
	      mink=strtod(argv[i+1], NULL);
	      if (mink<=0)
		{
		  printf("Minkowski power must be >0\n");
		  return(1);
		}
	    }
	  if(! strncmp("-n", argv[i], 2)) 
	    {
	      nearconstraint=strtod(argv[i+1], NULL);
	      if (nearconstraint<0 || nearconstraint>1)
		{
		  printf("Constraint between adjacent kernels must be in [0, 1]\n");
		  return(1);
		}
	    }
	  if(! strcmp("-a", argv[i])) 
	    flagDisplay=true;


	  if(! strcmp("-wl", argv[i]))
	    wordless=true;
	  if(! strncmp("-in", argv[i], 3) )
	    {
	      OPTIM=true; 
	      argstr=i+1;
	    } 
	  if(! strncmp("-it", argv[i], 3) )
	    {
	      maxi=strtol(argv[i+1],NULL,10); 
	      if (maxi <=0) maxi=1;
	    } 
	  if(! strncmp("-g", argv[i], 2) )
	    {
	      gauss=strtod(argv[i+1],NULL); 
	    } 
	  if(! strncmp("-s", argv[i], 2) )
	    {
	      seed1=atol(argv[i+1]);
	    } 
	}
    }
 
  // main loop for catching exceptions in called functions
  try{
     // modif January 2013 - always set seed
    if(seed1==0) 
      seed1=time(NULL);//machine clock

    setseed(seed1);//default or argument 
    
    // Construction of FIS S
    S=new FIS(argv[1]);
   
    if (flagDisplay)
      printf("Minkowski power=%f\n",mink);

    // handle mixed (-c) argument
    // false = Euclidean distance, true = fuzzy partition-based distance
    
    // argstrpos=position of string
    //numc: integer array (length=number of inputs), 0 for Euclidean, 1 for FP
    // nbmax=number of inputs
    // nbopt=number of inputs to optimize
    nbmax=S->GetNbIn();
    numc=new double[nbmax];

    // read input string to optimize
    if (OPTIM)
      {
	numopt=new double[nbmax];
	//numopt=inputs to optimize in that order
	nbopt=ReadInputsOpt(argv[argstr],nbmax,numopt,wordless);
	if (nbopt<=0) 
	  {
	    printf("\n error in reading input numbers\n");
	    return nbopt;
	  }
      }
  
    
    // default=fp-based fuzzy partitions in all dimensions
    for (int kk=0;kk<nbmax;kk++)
      numc[kk]=1;

    if (mixed)
      {
	int lenc=strlen(argv[argstrpos]);
	// one element: shortcut for identical choices in all dimensions
	if (lenc==1)
	  {
	    numc[0]=atol(argv[argstrpos]);
	    if ((numc[0]!=0)&&(numc[0]!=1)&&(numc[0]!=2))
	      {
		printf("\nerror in c string\n");
		printf("only 0,1 or 2 are allowed");
		return 1;
	      }
	    if (!wordless)
	      {
		if (numc[0]==0)
		  printf("Euclidean distance in  all dimensions\n");
		else 
		  printf("FP-based distance in  all dimensions\n");
	      }
	    for (int pp=1;pp<nbinputs;pp++)
	      numc[pp]=numc[0];
	  }
	else
	  {
	    nbc=SearchNb(argv[argstrpos],numc,nbinputs,blank);
	
	    if (flagDisplay)
	      printf("%d choices found (0 for Euclidean, 1 for FP-based): %s\n",nbc,argv[argstrpos]);
	  
	    if ((nbc>0)&&(nbc<nbinputs))
	      {
		printf("\nerror: not enough elements in c string\n");
	      }
	    for (int kk=0;kk<nbc;kk++)
	      {
		if ((numc[kk]!=0)&&(numc[kk]!=1)&&(numc[kk]!=2))
		  {
		    printf("\nerror in c string\n");
		    printf("only 0, 1 or 2 are allowed\n");
		    return 1;
		  }
	      }
	  }
      }
    if(!wordless)
      {
	printf("Distance combination with Minkowski power=%g\n",mink);
	if(norm) 
	  printf("Multidimensional normalized distance\n");
	else
	  printf("Multidimensional unnormalized distance\n");
	if (nearconstraint>0.0)
	  printf("Near constraint=%f\n",nearconstraint);
	  
	for (int p=0;p<nbinputs;p++)
	  {
	    if (S->In[p]->IsActive())
	      {
		if (numc[p]==0)
		  printf("column %d, Euclidean\n",p+1);
		else if (numc[p]==1)
		  printf("column %d, FP-based\n",p+1);
	      }
	  }
	
	if (OPTIM)//FP to optimize in turn
	  {
	    printf("There are %d inputs to optimize:\n",nbopt);
	    printf("seed=%ld",seed1);
	    // loop on fuzzy inputs to optimize: case input with numc=2
	    for (int kk=0;kk<nbopt;kk++)
	      {
		numopt[kk]=numopt[kk]-1;
		numinput=(int)numopt[kk];//shift numbers so that input 1 is #1
		printf("%d\t,",numinput+1);
	      }
	    printf("\n");
	  }
	   
      }
    
    // read data and normalize each column, computes nbex and nbinputs
    // also normalize active input FPs
    normdata=S->NormCheckDataDist(argv[2],nbex,nbinpdata,flagDisplay,wordless);
    // write normed data to file (to be used by R if necessary)
    nf=fopen("normdata.txt", "wt");
    for (i = 0; i < nbex; i++)
      {
	for(j = 0; j < nbinpdata-1; j++)
	  fprintf(nf, "%12.7f ;", normdata[i][j]);
	fprintf(nf, "%12.7f ", normdata[i][j]);

	fprintf(nf, "\n");
      }
    fclose(nf);
    if(!wordless)
      printf("\nnormdata.txt file written\n");

    // call dist on current FIS and current data file
    // write diss matrix in file diss.out to be read by R-returns diss
    S->distWithNormedData(normdata,nbex,nbinpdata,(char*)"diss.out",numc,norm,mink,flagDisplay,wordless);
    if (!wordless)
      printf("\nfull distance matrix written to diss.out file");

    // call to R-returns qual
    //system("ls -l diss.out");
    qualC=PamCluster(false);
    if (!wordless)
      printf("\nafter call to R pam, initial qualC=%g\n",qualC);
    
    S1=S->Clone();

    if (OPTIM)//FP to optimize in turn
      {
	// loop on fuzzy inputs to optimize: case input with numc=2
	for (int kk=0;kk<nbopt;kk++)
	  {
	    numinput=(int)numopt[kk];
	    if (numc[numinput]==1)
	      {
	
		// create object derived from algo
		// with FIS ptr and num of input to optimize
		S2=S1->Clone();
		fpopt=new DistOptim(S2,numinput,numc,normdata,nbex,nbinpdata,norm,mink,flagDisplay);
		//update algoPar structure (P)
		fpopt->P->SetParGauss(gauss);
		fpopt->P->maxIte=maxi;

		//prepare optim
		//call GetFuzzyInputParams to determine sizeparamsFP 
		//pFP (MF params) and iFP (MF type:1=trap,2=tri)
		// arrays are allocated inside GetFuzzyInputParams

		ptr=fpopt->GetFuzzyInputParams(nmf,iFP, sizeparamsFP, pFP,flagDisplay);
		// store FP parameters in container to be retrieved later in launchEntry
		fpopt->AddfEvalArgs(nmf,iFP,sizeparamsFP,pFP,flagDisplay, nearconstraint);
		//update key Size
		fpopt->Size=sizeparamsFP;
		//update algoPar structure (P)
		fpopt->UpdateKeyLenght(sizeparamsFP);
		// set key
		key.setValue(sizeparamsFP);
		fpopt->setKey(&key);

		if(!wordless)
		  printf("\nOptimizing input#%d\n",numinput+1);
		
		if (first)
		  {
		    fpopt->P->Print();
		    first=false;
		  }
		if (!wordless)
		  cout<<"\narray size="<<sizeparamsFP<<",key size="<<fpopt->getKeySize()<<endl;

		//launch optim - will call SolisWetts with eval function=GlobalOptPam, calling OptPam, calling R (PamCluster C++ function, pam.R script)
		display=true;
		rep=fpopt->launchEntry(qualC,flagDisplay);
		if (rep==0)//better FP found-update S1 for next input optim.
		  {
		    delete S1;
		    S1=fpopt->GetFisFPOpt()->Clone();
		  }
		// otherwise go back to previous FIS S1
		delete fpopt;

		printf("\nfinal qualC after optimizing FP#%d =%g\n",numinput+1,qualC);
	      }//end 1 input loop
	    else
	      printf("Euclidean distance in dimension %d, no optimization will be done\n",numinput+1);
	  }//end inputs to optimize loop
      }//end OPTIM loop
    //unnormalize and write final FIS (s1)
    S1->UnNormalize(NULL,0);
    FILE * fopt=fopen("optFP.fis","wt");
    S1->PrintCfg(fopt);
    fclose(fopt);

    // delete  arrays
    delete [] numc;

    // delete diss array
    if (diss!=NULL)
      {
	for( int i=0 ; i<nbex; i++ )
	  delete [] diss[i];
	delete [] diss;
	diss=NULL;
      }
  
    // delete global variable
    if(VarNameG != NULL) 
      {
	for(int j = 0; j < NbVarG; j++)
	  if(VarNameG[j] != NULL)  delete [] VarNameG[j];
	delete [] VarNameG;
	VarNameG = NULL;
      }
    /*if (S !=NULL)
      delete S;
    */

    delete [] iFP;
    delete [] pFP;
    return (0);

    delete S;
    delete S1;
    delete S2;
  }
  //action in case of exception
  catch(std::exception &e){
    printf("error in dist\n");
    printf("catch %s\n",e.what());
    
    return(-1);
  }
}


//**************************   End of PERF.CPP   *************************

