//*********************************************************************
//
//
//                              MAINFUZ2CRISP.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// Object : Call performance function
// Environnement : Uses library fispro

//*********************************************************************

#include "fis.h"
#include "common.h"

//! This program needs 1 argument:  the FIS file name
int main(int argc, char ** argv)
  //******************************
{
  if(argc < 2)
    {
      printf("\nThe program %s needs 1 argument: ", argv[0]);
      printf("\n\t* the FIS file name.") ;
      printf("\nOptional arguments : ");
      printf("\n\t* -c (set classif flag");
      printf("\n\t* -m (set max disjunction");
      printf("\n\t* -n outputNumber (default=0 first output");
      printf("\n\t* -a (detailed display)\n");
      return 1;
    }

  
  int i,nOut,len,classif=0,outputN=0,flagDisplay=0,disjmax=0;
  FIS * S;
  FILE *fg;
  char * outputFisConfigName=NULL;
  
  // main loop for catching exceptions in called functions
  try{
  
   
    if(argc > 2) 
      {
	for(i = 2; i < argc; i++)
	  {
	    if(! strcmp("-c", argv[i]) ) classif = 1;
	    if(! strncmp("-m", argv[i], 2) ) disjmax=1;
	    if(! strncmp("-o", argv[i], 2) ) outputN=atoi(argv[i] + 2);
	    if(!strcmp("-a", argv[i]) ) flagDisplay=1;
	  }
      }
    printf("\nfuzzy output %d transformed into crisp output\n",outputN);
   
    
    if(classif == 1) 
      {
	printf("classification flag set in output %d\n",outputN);
	printf("defuzzification operator set to MaxCrisp for output %d\n",outputN);
      }
    if (disjmax ==1)
      printf("disjunction operator set to max for output %d\n",outputN);
    else
      printf("disjunction operator set to sum for output %d\n",outputN);


   
    S = new FIS(argv[1]);  
    // check output number
    nOut=S->GetNbOut();
    if (nOut <=0)
      {
	sprintf( ErrorMsg, "~invalid~number~of~outputs~: %d", nOut);
	throw std::runtime_error( ErrorMsg );
      }
    if (outputN >(nOut-1))
      {
	sprintf( ErrorMsg, "~invalid~output~number: %d", outputN);
	throw std::runtime_error( ErrorMsg );
      }
    S->Fuz2Crisp(outputN);
    if (classif ==1) 
      {
	S->Out[outputN]->Classification(classif);
	S->Out[outputN]->SetOpDefuz( (char *)OUT_CRISP::MaxCrispDefuz());
      }
    if (disjmax ==1) 
      {
	S->Out[outputN]->SetOpDisj( (char *)OUT_CRISP::DisjMax());
      }
      // Generation of a config file
    len=strlen(argv[1]) + 25;
    outputFisConfigName=new char[len];
    strcpy(outputFisConfigName,argv[1]);
    if (classif == 1)
      {
	if (disjmax ==1)
	  strcat(outputFisConfigName,"-crisp-classif-max.fis");
	else
	  strcat(outputFisConfigName,"-crisp-classif-sum.fis");
      }
    else
      {
	if (disjmax ==1)
	  strcat(outputFisConfigName,"-crisp-max.fis");
	else
	  strcat(outputFisConfigName,"-crisp-sum.fis");
      }
    //
    fg = fopen(outputFisConfigName, "wt");
    S->PrintCfg(fg);
    fclose(fg);

    delete S;
    return (0);
  }
  //action in case of exception
  catch(std::exception &e){
    printf("error in mainfuz2crisp\n");
    printf("catch %s\n",e.what());
    
    return(-1);
  }
}


//**************************   End of MAINSAMPLE.CPP   *************************

