#include "fis.h"

void printGplCommand(FIS* glob, double ** Data, FILE * fgplt, char* fichdr, char* ficout, int iout, int iin)
{

  double originx, originy, sizey;


  // compute index for the strict inputs
  int nMfIn =0;
  for(int jj =0; jj <glob->GetNbIn();jj++){
    nMfIn+=glob->In[jj]->GetNbMf();
  }
  int nMfOut = glob->Out[iout]->GetNbMf();
  int index3 = nMfOut+nMfIn+ glob->GetNbIn()*iin;
  // compute index for the fuzzy outputs
  int index2 = (glob->GetNbRule()+1)*iin;
  // get the maximum number of rows
  int maxrow = 1;
  if (maxrow < glob->GetNbIn())
    maxrow =glob->GetNbIn();
  if (maxrow < glob->GetNbRule())
    maxrow =glob->GetNbRule();

  sizey = (1./maxrow);
  // set the multiplot mode
  // fprintf(fgplt, "set title \"Inference pour le jeu d'entrees %d\"\n",iin);
  fprintf(fgplt, "set multiplot \n");
  fprintf(fgplt, "set yrange[0:1.1]\n");
  fprintf(fgplt, "set  xtics offset 0,0.5\n");
  fprintf(fgplt, "set size 0.333,%8.3f\n", sizey);
  // commands to plot the fuzzy MFs and the observed input
  int index1 = 0;
  originx =0.0;
  originy =1.0-sizey;
  double *entree;
  entree = *(Data+iin);
  for(int jj =0; jj <glob->GetNbIn();jj++){
    fprintf(fgplt, "set title \" entree %8.3f\" offset 0,-1 \n", entree[jj]);
    fprintf(fgplt, "set origin %8.3f,%8.3f \n",originx, originy);
    fprintf(fgplt, "set xrange[%8.3f:%8.3f]\n", glob->In[jj]->min(),glob->In[jj]->max());
    fprintf(fgplt, "plot \'%s\' using 1:2 index %d notitle  with lines,\\\n", fichdr, index1);
    index1++;
    for (int j= 1; j<glob->In[jj]->GetNbMf(); j++){
      fprintf(fgplt, "\'%s\' using 1:2 index %d notitle with lines,\\\n", fichdr, index1);
      index1++;
    }
    fprintf(fgplt, "\'%s\' using 1:2 index %d notitle with lines lw 2\n", fichdr, index3);
    index3++;
    fprintf(fgplt, "unset xrange\n");
    fprintf(fgplt, "unset label\n");
    originy-=sizey;

  }
  // commands to plot the fuzzy output and  global output
  fprintf(fgplt, "set xrange[%8.3f:%8.3f]\n", glob->Out[iout]->min(),glob->Out[iout]->max());
  originx =0.333;
  originy =1.0-sizey;
  for(int jj =0; jj <glob->GetNbRule();jj++){
    index1 = nMfIn;
    fprintf(fgplt, "set title \"rule %d\" offset 0,-1\n", jj);
    fprintf(fgplt, "set origin %8.3f,%8.3f \n",originx, originy);
    fprintf(fgplt, "plot \'%s\' using 1:2 index %d notitle with lines,\\\n", fichdr, index1);
    index1++;
    for (int j= 1; j<glob->Out[iout]->GetNbMf(); j++){
      fprintf(fgplt, "\'%s\' using 1:2 index %d  notitle with lines,\\\n", fichdr, index1);
      index1++;
    }
    fprintf(fgplt, "\'%s\' using 1:2 index %d notitle with filledcu\n", ficout, index2);
    index2++ ;
    originy-=sizey;
  }
  originx =0.666;
  originy =1.0-sizey;
  index1 = nMfIn;
  fprintf(fgplt, "set title \"globale %d \" offset 0,-1 \n", iin);
  fprintf(fgplt, "set origin %8.3f,%8.3f \n",originx, originy);
  fprintf(fgplt, "plot \'%s\' using 1:2 index %d notitle with lines,\\\n", fichdr, index1);
  index1++;
  for (int j= 1; j<glob->Out[iout]->GetNbMf(); j++){
    fprintf(fgplt, "\'%s\' using 1:2 index %d  notitle with lines,\\\n", fichdr, index1);
    index1++;
  }
  fprintf(fgplt, "\'%s\' using 1:2 index %d notitle with points ps 4 lw 2, \\\n", ficout, index2);
  fprintf(fgplt, "\'%s\' using 1:2 index %d notitle with filledcu\n", ficout, index2);
  index2++ ;
  originy-=sizey;

  fprintf(fgplt, "unset xrange\n");
  fprintf(fgplt, "unset multiplot\n");
  fprintf(fgplt, "unset title\n");
  fprintf(fgplt, "pause -1 \"Hit return to continue\"\n");
}

// prepares gnuplots for input MFs, output MFs and fuzzy inputs
void printGplFuzDataIn(FIS * glob, double ** Data, int iout, int nsin, char* fichdr, MFDPOSS ** fuzin,  double * KW, double * SW, double maxposs)
{
  FILE *fghdr;
  double ls,rs,lk,rk;
  // create the header part for gnuplot
  fghdr = fopen(fichdr, "wt");
  //create the curves for the input MFs
  for(int jj =0; jj < glob->GetNbIn();jj++){
    for (int i=0; i<glob->In[jj]->GetNbMf(); i++){
      glob->In[jj]->GetMF(i)->Kernel(lk,rk);
      glob->In[jj]->GetMF(i)->Support(ls,rs);
      fprintf(fghdr, "#IN%d MF%d\n", jj,i);
      fprintf(fghdr,"  %8.3f  0 \n", ls);
      fprintf(fghdr,"  %8.3f  1 \n", lk);
      fprintf(fghdr,"  %8.3f  1 \n", rk);
      fprintf(fghdr,"  %8.3f  0 \n", rs);
      fprintf(fghdr, " \n\n");
    }

  }

  //create the curves for the fuzzy output MFs
  for (int i=0; i<glob->Out[iout]->GetNbMf(); i++){
    glob->Out[iout]->GetMF(i)->Kernel(lk,rk);
    glob->Out[iout]->GetMF(i)->Support(ls,rs);
    fprintf(fghdr, "#OUT%d MF%d\n",iout, i);
    fprintf(fghdr,"  %8.3f  0 \n", ls);
    fprintf(fghdr,"  %8.3f  1 \n", lk);
    fprintf(fghdr,"  %8.3f  1 \n", rk);
    fprintf(fghdr,"  %8.3f  0 \n", rs);
    fprintf(fghdr, " \n\n");
  }

  //create the curves for the fuzzy inputs
  double *entree;
  for (int i = 0; i < nsin; i++){
    entree = *(Data+i);
    // build fuzzy input(s) fuzin array ptr for each crisp input in the data file
    // using template in mtpl[i]
    // fuzin[jj] is a poss. distrib.
    //glob->BuildFuzIn(entree, mftpl, fuzin);
    //for (int k=0;k<2;k++)
    //glob->BuildFuzIn(entree[k],k, KW[k],SW[k],fuzin[k]);
    fuzin = NULL;
    glob->BuildFuzIn(entree, KW, SW, fuzin,maxposs);
    for(int jj =0; jj <glob->GetNbIn();jj++){
      fprintf(fghdr, "#IN%d FUZZY%d\n", jj,i);
      if(fuzin[jj]!=NULL)
	if(fuzin[jj]->NbParams() >0)
	  fuzin[jj]->Print(fghdr);
      fprintf(fghdr, " \n\n");
    }
  }
  fclose(fghdr);
}

// prepares gnuplots for input MFs, output MFs and crisp inputs
void printGplDataIn(FIS * glob, double ** Data, int iout, int nsin, char* fichdr)
{
  FILE *fghdr;
  double ls,rs,lk,rk;
  // create the header part for gnuplot
  fghdr = fopen(fichdr, "wt");
  //create the curves for the input MFs
  for(int jj =0; jj <glob->GetNbIn();jj++){
    for (int i=0; i<glob->In[jj]->GetNbMf(); i++){
      glob->In[jj]->GetMF(i)->Kernel(lk,rk);
      glob->In[jj]->GetMF(i)->Support(ls,rs);
      fprintf(fghdr, "#IN%d MF%d\n", jj,i);
      fprintf(fghdr,"  %8.3f  0 \n", ls);
      fprintf(fghdr,"  %8.3f  1 \n", lk);
      fprintf(fghdr,"  %8.3f  1 \n", rk);
      fprintf(fghdr,"  %8.3f  0 \n", rs);
      fprintf(fghdr, " \n\n");
    }

  }
  //create the curves for the fuzzy output
  for (int i=0; i<glob->Out[iout]->GetNbMf(); i++){
    glob->Out[iout]->GetMF(i)->Kernel(lk,rk);
    glob->Out[iout]->GetMF(i)->Support(ls,rs);
    fprintf(fghdr, "#OUT%d MF%d\n",iout, i);
    fprintf(fghdr,"  %8.3f  0 \n", ls);
    fprintf(fghdr,"  %8.3f  1 \n", lk);
    fprintf(fghdr,"  %8.3f  1 \n", rk);
    fprintf(fghdr,"  %8.3f  0 \n", rs);
    fprintf(fghdr, " \n\n");
  }
  //create the curves for the strict inputs
  double *entree;
  for (int i = 0; i<nsin; i++){
    entree = *(Data+i);
    for(int jj =0; jj <glob->GetNbIn();jj++){
      fprintf(fghdr, "#IN%d STRICT%d\n", jj,i);
      fprintf(fghdr,"  %8.3f  0 \n", entree[jj]);
      fprintf(fghdr,"  %8.3f  1 \n", entree[jj]);
      fprintf(fghdr, " \n\n");
    }
  }
  fclose(fghdr);
}

//! This function generates the data file used
//! to show the membership functions for outputs
void printGplDataOut(FIS * glob,MFDPOSS* minf,FILE *fgres, int iout, int iin)
{
  // print output MFDPOSS for each rule
  for (int i =0; i<glob->GetNbRule(); i++){
    fprintf(fgres, "#Conclusion %d\n",i);
    if ((glob->Out[iout]->MfConc[i] != NULL) && (glob->Out[iout]->MfConc[i]->NbParams()>0)){
      glob->Out[iout]->MfConc[i]->Print(fgres);
    } else if ((glob->Out[iout]->MfConc[i] != NULL)&& (glob->Out[iout]->MfConc[i]->NbParams()==0)) {
      fprintf(fgres,"  %8.3f  0 \n", glob->Out[iout]->min());
      fprintf(fgres,"  %8.3f  0 \n", glob->Out[iout]->max());
    } else{
      fprintf(fgres,"  %8.3f  1 \n", glob->Out[iout]->min());
      fprintf(fgres,"  %8.3f  1 \n", glob->Out[iout]->max());
    }
    fprintf(fgres, " \n\n");
  }

  // print output MFDPOSS for global output
  fprintf(fgres, "#Aggregation Globale %d\n",iin);
  if ((minf != NULL) && (minf->NbParams()>0)){
    minf->Print(fgres);
  } else if ((minf != NULL)&& (minf->NbParams()==0)) {
    fprintf(fgres,"  %8.3f  0 \n", glob->Out[iout]->min());
    fprintf(fgres,"  %8.3f  0 \n", glob->Out[iout]->max());
  } else{
    fprintf(fgres,"  %8.3f  0 \n", glob->Out[iout]->min());
    fprintf(fgres,"  %8.3f  0 \n", glob->Out[iout]->max());
  }
  fprintf(fgres, " \n\n");
}


// Test of InferFati function
int main(int argc, char ** argv)
  //******************************
{
  int i=0;
  int j, nbcol, NbEx, ifuz,len=0;

  double ** Data = NULL;
  FIS *g = NULL;
  FILE *f=NULL;
  FILE *fgres=NULL;
  FILE *fgplt=NULL;
  char * ficres = NULL;
  char * ficout = NULL;
  char * suffix = NULL;
  MFDPOSS *minf, ** fuzin;

  int aff = 0;  // No display by default
  int gnuplot=0; //no gnuplot output by default
  int nalpha = 5;       // number of alpha cuts
  int fuz1 = 0;         // Is Input 1 fuzzy?
  int fuz2 = 0;         // Is Input 2 fuzzy?
  int iout = 0;         // Out number
  int ref=0; //observed output
  double symbmatch=0.0,lk,rk,mulk,murk;
  double *muOut=NULL, *muOutInf=NULL,MuSeuil=0.667;
  double maxposs=1.0,muThreshInf=0.667,muThreshObs=0.667;
  int countBl=0,NbBlancs,ValBlanc=0;
  int NbIn,nbOutMf;
  //double x =0;
  //double y =0;
  double KW[2];
  double SW[2];
  char * fichdr = NULL;
  char * ficgpl = NULL;
  char ** filet=NULL;

  double ** val=NULL;
  int nbval=0;
  double maxp=1.0;
  fuzin = NULL;
  minf = NULL  ;
  if(argc < 3)
    {
      printf("\nThe Implicative Inference test program %s needs 2 arguments : ", argv[0]);
      printf("\n\t* the fis configuration file name;");
      printf("\n\t* the data file name.") ;
      printf("\nOptions: ");
      printf("\n\t* -a for intermediate display" );
      printf("\n\t* -m for mu threshold");
      printf("\n\t* -f filename for 1st input fuzzyfier (default crisp)");
      printf("\n\t* -s filename for 2nd input fuzzyfier (default crisp)");
      printf("\n\t* -n max poss for fuzzy input (default=1)");
      printf("\n\t* -c number of alpha cuts (default %d)\n", nalpha);
      
      printf("\n\t* -p for gnuplot output" );
      printf("\n\n Implicative inference is :");
      printf("\n\t - FATI when at least one input is fuzzy,");
      printf("\n\t - FITA when all inputs are crisp.\n");
      return 1;
    }

  try
    {

      if(argc > 3)
	{
	  for(int i = 3; i < argc; i++)
	    {
	      if(!strcmp("-a", argv[i]) ) aff=1;
	      if(!strcmp("-p", argv[i]) ) gnuplot=1;
	      else if(!strncmp("-f", argv[i], 2)) fuz1=i;
	      else if(!strncmp("-s", argv[i], 2)) fuz2=i;
	      else if(!strncmp("-c", argv[i], 2)) nalpha=atoi(argv[i]+2);
	      else if(!strncmp("-m",argv[i], 2))  MuSeuil=atof(argv[i]+2);
	      else if(!strncmp("-n",argv[i], 2))  maxposs=atof(argv[i]+2);
	    }
	}
      g = new FIS(argv[1]);
      
      ifuz = (fuz1 || fuz2);  // FATI algorithm
      if (ifuz && g->GetNbIn()>2)
	{
	  sprintf( ErrorMsg, "~Invalid#InputsInferFatiLimitedTo2~:  %d", g->GetNbIn());
	  throw std::runtime_error( ErrorMsg );
	}
      if(ifuz)
	{
	  if (!(!strcmp( g->Out[iout]->GetDisj(), "igg" )||!strcmp( g->Out[iout]->GetDisj(), "irg" )||!strcmp( g->Out[iout]->GetDisj(), "igd" )))
	    {
	      sprintf( ErrorMsg,"implicative~inference~required");
	      throw std::runtime_error( ErrorMsg );
	    }
	       
	}
      Data = ReadSampleFile(argv[2], nbcol, NbEx);

      iout=0;
      suffix = new char[4];
      strcpy(suffix,"_su");
      if (! strcmp( g->Out[iout]->GetDisj(), "igg" ))  strcpy(suffix,"_gg");
      if (! strcmp( g->Out[iout]->GetDisj(), "irg" ))  strcpy(suffix,"_rg");
      if (! strcmp( g->Out[iout]->GetDisj(), "igd" ))  strcpy(suffix,"_gd");

      ficres = new char[strlen(argv[2]) + 8];
      strcpy(ficres,argv[2]);
      strcat(ficres,suffix);
      strcat(ficres,".res");
      f = fopen(ficres, "wt");

      if(gnuplot==1)
	{
	  len=strlen("plotfis.dat");
	  fichdr=new char[len+1];
	  strcpy(fichdr,"plotfis.dat");
	  printf("le fichier de donnees pour GNUPLOT est %s\n", fichdr);
	  ficgpl
=new char[len+1];
	  strcpy(ficgpl,"plotfis.gpl");
	  printf("le fichier command pour GNUPLOT est %s\n", ficgpl);
	  ficout=new char[len+1];
	  strcpy(ficout,"plotfis.res");
	  printf("le fichier sortie pour GNUPLOT est %s\n", ficout);

	  //open output file
	  fgplt = fopen(ficgpl, "wt");
	  fgres = fopen(ficout, "wt");
	}

		
      if(ifuz) // at least one input (out of 2 max) is fuzzy - build fuzzy input from data and model given in file
	{
	  //mftpl = new MFDPOSS*[g->GetNbIn()];

	  //if (fuz1) mftpl[0] = new MFDPOSS(argv[fuz1]+2);
	  //else mftpl[0] = new MFDPOSS(0.);

	  //if (g->GetNbIn() == 2)
	  //	{
	  //		if(fuz2) mftpl[1] = new MFDPOSS(argv[fuz2]+2);
	  //		else mftpl[1] = new MFDPOSS(0.);
	  //	}
	  // print template dposs for fuzzy inputs

	  // fuzin = poss. distrib. ptr
	  /*fuzin = new MFDPOSS*[g->GetNbIn()];
	    for (int ii=0;ii<g->GetNbIn();ii++)
	    {
	    printf("fuzzy input #%d template\n",ii);
	    mftpl[ii]->Print();
	    }
	  */

	  g->InferFatiPrep(iout);
	}

      // now infer
      // write header in res file
      //ref=1;
      g->WriteHeader(iout, f, ref);
      NbIn=g->GetNbIn();
      nbOutMf=g->Out[iout]->GetNbMf();
      muOut=new double[nbOutMf];
      muOutInf=new double[nbOutMf];
      // READ kw and sw FROM TEMPLATE FILES
      filet=new char* [2];
     
		
      if (ifuz)
	{
	  if (fuz1)
	    filet[0]=argv[fuz1]+2;
	  else
	    filet[0]=NULL;
	  
	  if (fuz2)
	    filet[1]=argv[fuz2]+2;
	  else
	    filet[1]=NULL;
	
		    
	  for (j = 0; j < NbIn; j++) 
	    {
	      if (filet[j]!=NULL)
		ReadTemplate(filet[j],KW[j],SW[j]);
	      else 
		KW[j]=SW[j]=0.0;
	    }
	}		
	
      for(j = 0; j < NbEx; j++) // NbEx
	{
	  if(ifuz)
	    {
		
	      // build fuzzy input fuzin[i] for each crisp input in the data file
	      // using template file
			
	      g->BuildFuzIn(Data[j], KW, SW,fuzin,maxposs);
	      // print fuzzy input(s)
	      printf("\nfuzzy input for row #%d\n",j);
	      for (int ii=0;ii<g->GetNbIn();ii++)
		{
		  printf("input #%d  crisp input %8.3f\n",ii,Data[j][ii]);
		  fuzin[ii]->Print();
		}
	      printf("InferFati\n");
	      fflush(stdout);
	      //fuzzy input inference
	      //minf = g->InferFati(fuzin, nalpha, iout, f, aff);
	      
	      g->InferFatiCheck(fuzin, iout, nalpha,val,nbval,f,aff);
	      printf("Fin InferFati\n");
	      printf("outvalue=%g\n",g->OutValue[iout]);fflush(stdout);
	      if(g->Out[iout]->MfGlob) minf = g->Out[iout]->MfGlob->Clone();
	      if (fuzin)
		{
		  for(i=0; i < g->GetNbIn(); i++) 
		    {
		      delete fuzin[i];
		      fuzin[i]=NULL;
		    }
		  delete [] fuzin;
		  fuzin=NULL;
		}
	    }
	  else
	    {
	      //crisp input
	      //g->InferCheck(Data[j]); // to test jni interface
	      //if observed output value print it in result file 
	      if(ref)
		fprintf(f, FORMAT_DOUBLE, Data[j][NbIn + iout]);
	      // implicative case
	      // if observed output value get MF membership of observed output value and use Museuil for label membership
	      if (ref)
		{
		  for(int imf = 0; imf < nbOutMf; imf++)
		    {
		      muOut[imf]=g->Out[iout]->GetADeg(imf,Data[j][NbIn + iout]);;
		      fprintf(f, FORMAT_DOUBLE,muOut[imf]);
		      if (aff)
			printf("muObs[%d]=%g\t",imf,muOut[imf]);
		    }
		}

	      g->Infer(Data[j], iout, f, aff);
	      printf("outvalue=%g\n",g->OutValue[iout]);fflush(stdout);
	      // Infer writes inferred value (middle of kernel , alarm (0), membership of inferred dposs to output MFs 
	      // compute symbolic match between inferred and observed output
	      // this is done in DEFUZ_ImpFuzzy::EvalOut
	      symbmatch=1;
	      countBl=0;
	      // compute MFGlob kernel
	      if (g->Out[iout]->MfGlob !=NULL)
		{
		  g->Out[iout]->MfGlob->AlphaKernel(lk,rk,maxp);
		  for(int imf = 0; imf < nbOutMf; imf++)
		    {
		      mulk=g->Out[iout]->GetADeg(imf,lk);
		      murk=g->Out[iout]->GetADeg(imf,rk);
		      // consider maximum membership to each mf
		      muOutInf[imf]=mulk;
		      if (murk>mulk)
			muOutInf[imf]=murk;
		      if (aff)
			printf("\noutMF#%d\tlk=%g\tmulk=%g\trk=%g\tmurk=%g\tmuInf[%d]=%g\t",imf,lk,mulk,rk,murk,imf,muOutInf[imf]);
		      if (ref)
			fprintf(f, FORMAT_DOUBLE,muOutInf[imf]);

		      if (muOutInf[imf]>muThreshInf)
			countBl++;
		    }
		  if(countBl>2)
		    {
		      ValBlanc = 1;
		      NbBlancs++;
		    }
		}// end case output dposs not empty
	  
	      if (aff) printf("\ncountBl=%d,ValBlanc=%5d \n", countBl,ValBlanc);
	      if (ref)
		{
		  symbmatch=((OUT_FUZZY *)(g->Out[iout]))->SymbMatch(muOut,muOutInf,nbOutMf,muThreshObs,muThreshInf,aff);

		  fprintf(f,FORMAT_DOUBLE,symbmatch);
		}
	      if(g->Out[iout]->MfGlob) minf = g->Out[iout]->MfGlob->Clone();
	    }
	  

	  // Process minf ...
	  if(minf)
	    {
	      if (minf->NbParams()>0)
		{
		  printf("example # %d inf. result:\n",j);
		  /*for (int ii= 0; ii< minf->NbParams();ii++)
		    {
		    minf->GetPoint(x,y,ii);
		    printf("point%d %8.3f %8.3f\n",ii,x,y); fflush(stdout);
		    }
		  */
		  minf->Print();
		  printf("\nMfConc per rule\n");
		  for (int kkk=0;kkk<g->GetNbRule();kkk++)
		    {
		      printf("rule %d\n",kkk);
		      if (g->Out[iout]->MfConc[kkk]!=NULL)
			g->Out[iout]->MfConc[kkk]->Print();
		      else
			printf("null MfConc");
		      printf("\n");
		    }
		  
		}
	      else
		printf("inference result EMPTY\n for sample # %d",j);
	      // plot sortie
	      if (gnuplot==1)
		printGplDataOut(g,minf,fgres, iout, j);
	      delete minf; minf = NULL;
	      for (int kkk=0;kkk<g->GetNbRule();kkk++)
		    {
		      if (g->Out[iout]->MfConc[kkk]!=NULL)
			delete g->Out[iout]->MfConc[kkk];
		      g->Out[iout]->MfConc[kkk]=NULL;
		    }
	    }
	}//end of loop on data items

     

      if(gnuplot==1)
	{
	  // update MFGlob
	  //g->Out[iout]->MfGlob=minf;

	  if(ifuz)
	    {
	      //entrees floues
	      printGplFuzDataIn(g, Data, iout, NbEx, fichdr, fuzin,KW, SW,maxposs);
	    }
	  else
	    //plot des entrees nettes
	    printGplDataIn(g,Data,iout, NbEx,fichdr);

	  //plot des commandes
	  fprintf(fgplt, "set tmargin 1\n");
	  fprintf(fgplt, "set bmargin 0.5\n");
	  fprintf(fgplt, "set lmargin 4\n");
	  fprintf(fgplt, "set rmargin 4\n");
	  for(int ii = 0;ii<NbEx; ii++)
	    printGplCommand(g,Data,fgplt, fichdr, ficout, iout, ii);
	}

      if (f!=NULL) fclose(f);
      if (fgres !=NULL) fclose(fgres);


      delete [] suffix;
      delete [] ficres;
      delete [] muOut;
      delete [] muOutInf;
      delete [] filet;
      delete g;
      for (i = 0; i < NbEx; i++)
	{
	  if (Data[i] != NULL) delete [] Data[i];
	}
      if (Data != NULL) delete [] Data;
    } // End of try

  catch(std::runtime_error &e)
    {
      printf("catch %s\n",e.what());
      return(-1);
    }
}// End of main



