#include "fis.h"
#include <errno.h>
#include <iostream>
#include <fstream>
#include <string>
#include <stdexcept>

int main(int argc, char ** argv)
  //******************************
{
  // Java_fis_jnifis_InitClasses 
  FIS *fis =NULL, *tmpfis=NULL;
  int ninputs=0,i,j,k,len,nsam=0,is,nbr;
  int * nmf=NULL;
  int ** nbparams=NULL;
  char * FisConfigName=NULL;
  char * snum=NULL;
  double ****par=NULL;
  double firstq, median, thirdq,min1,max1;
  double * tab=NULL;
  double ** conclu=NULL;
  MF * mf=NULL;
  bool wordless=false;
  
  // argv[1] = prefix
  // argv[2] = suffix
  
  // generate file names
 if(argc < 3)
    {
      printf("\nThe program %s needs 3 arguments : ", argv[0]);
      printf("\n\t* the fis  file name prefix;");   
      printf("\n\t* the fis  file name suffix"); 
      printf("\n\t* the number of samples\n");
      printf("\n\t* -wl for wordless.\n");
      
      return 1;
    }

  if(argc > 3) 
      {
	for(i = 3; i < argc; i++)
	  {
	    if(! strncmp("-wl", argv[i], 3) ) wordless = true;
	  }
      }

  len=strlen(argv[1])+strlen(argv[2]) + 5;
  FisConfigName=new char[len];
  snum=new char[5];

  ofstream f ("stat.txt");
  ofstream fr ("statrules.txt");

  nsam=atoi(argv[3]);

  if (!wordless)
    cout<<"nsamples="<<nsam<<endl;

  //first FIS=model
   strcpy(FisConfigName,argv[1]);
   sprintf(snum,"%i",0);
   strcat(FisConfigName,snum);
   strcat(FisConfigName,argv[2]);
   fis=new FIS(FisConfigName);
   //fis->PrintCfg();
   ninputs=fis->GetNbIn();

   nmf=new int[ninputs];
   nbparams=new int*[ninputs];
   
   par=new double ***[nsam];
   conclu=new double *[nsam];
   for (is=0;is<nsam;is++)
     par[is]=new double **[ninputs]; 

  
   //examine each input for sample 0
   for (i=0;i<ninputs;i++)
     {
       // for each MF, get MF number, type and bounds
       nmf[i]=fis->In[i]->GetNbMf();
       nbparams[i]=new int[nmf[i]];
       par[0][i]=new double *[nmf[i]];
       for (j=0;j<nmf[i];j++)
	 {
	   mf=fis->In[i]->GetMF(j);
	   nbparams[i][j]=mf->NbParams();
	   par[0][i][j]=new double[nbparams[i][j]];
	   mf->GetParams(par[0][i][j]);
	 }
     }
   nbr=fis->GetNbRule();
   conclu[0]=new double[nbr];
   // rule conclusions
   for (i=0;i<nbr;i++)
     conclu[0][i]=fis->Rule[i]->GetAConc(0);
   //other samples
   for (is=1;is<nsam;is++)
     {
       if (!wordless)
	 cout <<endl<<"sample # "<<is;
         strcpy(FisConfigName,argv[1]);
	 sprintf(snum,"%i",is);
	 strcat(FisConfigName,snum);
	 strcat(FisConfigName,argv[2]);
	 tmpfis=new FIS(FisConfigName);
	 //tmpfis->PrintCfg();
       //examine each input
       for (i=0;i<ninputs;i++)
	 {
	   // for each MF, get MF number, type and bounds
	   nmf[i]=tmpfis->In[i]->GetNbMf();
	   par[is][i]=new double *[nmf[i]];
	   for (j=0;j<nmf[i];j++)
	     {
	       mf=tmpfis->In[i]->GetMF(j);
	       nbparams[i][j]=mf->NbParams();
	       par[is][i][j]=new double[nbparams[i][j]];
	       mf->GetParams(par[is][i][j]);
	     }
	 }
       
       // rule conclusions
       conclu[is]=new double[nbr];
       for (i=0;i<nbr;i++)
	 conclu[is][i]=tmpfis->Rule[i]->GetAConc(0);
     }

   // record stats in file "stat.txt" and "statrules.txt"
   f << "input number;mf number;param number;median;firstq;thirdq";
   fr << "rule number;median;firstq;thirdq";
   
   //average each parameter
   for (i=0;i<ninputs;i++)
     {
       if (!wordless)
	 cout <<endl<<"i="<<i<<" nmf="<<nmf[i];
        for (j=0;j<nmf[i];j++)
	  {
	    //cout <<endl<<"mf="<<j<<" nbparams="<<nbparams[i][j];
	    for (k=0;k<nbparams[i][j];k++)
	      {
		tab=new double[nsam];
		for (is=0;is<nsam;is++)
		  tab[is]=par[is][i][j][k];
		StatArrayQuart(tab, nsam, 0, firstq, median, thirdq,min1,max1);
		// update par
		par[0][i][j][k]=median;
		//cout <<endl<<"k="<<k<<" median. param="<<median<<" par="<<par[0][i][j][k];
		//cout<<" firstq="<<firstq<<" thirdq="<<thirdq<<endl;
		f<<endl<<i<<";"<<j<<";"<<k<<";"<<median<<";"<<thirdq<<";"<<firstq;
		// update MF parameters
		fis->In[i]->GetMF(j)->Update(par[0][i][j]);

	      }
	  }
     }
   for (i=0;i<nbr;i++)
     {
       tab=new double[nsam];
       	for (is=0;is<nsam;is++)
	  tab[is]=conclu[is][i];
	StatArrayQuart(tab, nsam, 0, firstq, median, thirdq,min1,max1);
	fr<<endl<<i<<";"<<median<<";"<<thirdq<<";"<<firstq;
	// update rule conclusions
	fis->Rule[i]->SetAConc(0,median);
     }
  
   fis->PrintCfg();
   FILE * fg=fopen("med.fis","wt");
   fis->PrintCfg(fg);
  
   fclose(fg);

   // check
   /*cout <<endl<<"ninputs="<<ninputs;
   for (i=0;i<ninputs;i++)
     {
       cout <<endl<<"i="<<i<<" nmf="<<nmf[i];
        for (j=0;j<nmf[i];j++)
	  {
	    cout <<endl<<"mf="<<j<<" nbparams="<<nbparams[i][j];
	    for (k=0;k<nbparams[i][j];k++)
	      cout <<endl<<"k="<<k<<" param="<<par[0][i][j][k];
	  }
     }
   cout <<endl;
   */
     
   

  
   return 0;
}
