//*********************************************************************
//
//
//                              MAINSAMPLESPLIT.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// Object : Call performance function
// Environnement : Uses library fispro

//*********************************************************************

#include "gensample.h"
#include <cstring>

//! This program needs 1 argument:  the data file name
int main(int argc, char ** argv)
  //******************************
{
  if(argc < 2)
    {
      printf("\nThe program %s requires 1 argument: ", argv[0]);
      printf("\n\t* the data file name.") ;
      printf("\nOptional arguments : ");
      printf("\n\t* -n samples (total number of sample pairs)");
      printf("\n\t\tdefault value=1");
      printf("\n\t\t -> split data file into 2  files");
      printf("\n\t* -p validation file/data file size ratio (default value= 0.75)");
      printf("\n\t\t -> only used when sampling with replacement");
      printf("\n\t* -l (sampling without replacement-default=sampling done with replacement) ");
      printf("\n\t* -t truncate last file -> all sample files of equal size (default=last file used for complement");
      printf("\n\t\t -> only used when sampling without replacement");
      printf("\n\t* -c (classification sample )");
      printf("\n\t\t -> draw samples with respect to class ratio in data file");
      printf("\n\t* -o class column number (for classification sample only)");
      printf("\n\t\tdefault value=last column in data file");
      printf("\n\t* -e precision (for classification sample only)");
      printf("\n\t\t -> used to determine classes - default value=1e-2");
      printf("\n\t* -s seed (set seed to draw the same random sampling)");
      printf("\n\t\tdefault value=0");
      printf("\n\t\t -> new random sampling");
      printf("\n\t* -a (detailed display)\n");
      printf("\n\t* -wl for wordless.\n");

      return 1;
    }

  int i,classif,outCol,npairs,FlagDisplay=0,replace,truncate;
  long int seed;
  double percV,eps;
  bool wordless=false;

  // main loop for catching exceptions in called functions
  try{
  
    classif=0;
    replace=0;
    truncate=0;
    outCol=-1;
    eps=1e-6;//default precision
    percV=0.75;
    seed=0;// default=new random sequence at each call
    npairs=1;

    if(argc > 2) 
      {
	for(i = 2; i < argc; i++)
	  {
	    if(! strcmp("-c", argv[i]) ) classif = 1;
	    else if(! strcmp("-l", argv[i]) ) replace=1;
	    else if(! strncmp("-e", argv[i], 2) ) eps = strtod(argv[i] + 2, NULL);
	    else if(! strncmp("-o", argv[i], 2) ) outCol=atoi(argv[i] + 2);
	    else if(! strncmp("-p", argv[i], 2) ) percV = strtod(argv[i] + 2, NULL);
	    else if(! strncmp("-s", argv[i], 2) ) seed =  atol(argv[i] + 2);
	    else if(! strncmp("-n", argv[i], 2) ) npairs =  atoi(argv[i] + 2);
	    else if(! strcmp("-t", argv[i]) ) truncate=1;
	    else if(!strcmp("-a", argv[i]) ) FlagDisplay=1;
	    else if(! strncmp("-wl", argv[i], 3) ) wordless = true;
	  }
      }
    if(wordless)
      FlagDisplay=0;
    if((classif == 1) && !wordless) printf("sampling respects class ratio in data file, class value resolution=%f\n",eps);

    // call sample generation: 
    // seed=init random numbers (0=machine clock)
    // npairs = number of samples to generate.
    // classif=generate samples respecting data file class ratio
    // eps=precision to determine classes
    // outCol=data column number for class generation
    if (replace)
      {
	if (npairs < 1) npairs=1;
	if (!wordless)
	  {
	    if (truncate==1)
	      printf("\nGenerating %d cross validation samples of equal size\n",npairs);
	    else
	      printf("\nGenerating %d cross validation samples of approx. equal size\n",npairs);
	  }
	genSplit(seed,truncate,classif,eps,outCol,argv[1],npairs,FlagDisplay);
	printf("\n");
      }
    else
      {
	// npairs must be divided by two - if not a multiple of two it is the number of sample pairs
	if (!wordless)
	  printf("\nGenerating %d validation sample(s) + %d test sample(s)",npairs,npairs);
	if(percV>1.) percV=1.;
	if (!wordless)
	  printf("\nvalidation sample size ratio=%g percent of datafile\n",percV*100);
	genNSamples(seed,percV,classif,eps,outCol,argv[1],npairs,FlagDisplay);
	if (!wordless)
	  printf("\n");
      }

    
    return (0);
    }
  //action in case of exception
  catch(std::exception &e){
  printf("error in mainsample\n");
  printf("catch %s\n",e.what());
    
    return(-1);
   }
}


//**************************   End of MAINSAMPLE.CPP   *************************

