//******************************************************************
//
//
//                              MAINVOCREDUC.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// Object : Inference with fuzzy inputs illustration
// Environnement : Uses library fispro.
//
//******************************************************************


#include "fis.h"

int main(int argc, char ** argv)
{

  if (argc < 3) 
    {
      printf("\nThe program %s needs 2 arguments: ", argv[0]);
      printf("\n\t* the fis configuration file name;");  
      printf("\n\t* the data file name for perf. calculation");
      printf("\n\t  and voc. reduction when made from output data (option -d0)");
      printf("\n\nOptions:");
      printf("\n\t -o output number (default : first output)");
      printf("\n\t -d data used to make vocabulary reduction");
      printf("\n\t    (1:  rule conclusions -default-, 0:  output values)");
      printf("\n\t -l performance loss allowed by vocabulary reduction (default:  0.1)");
      printf("\n\t -c NumberOfConc to fix the number of elements in the reduced vocabulary ");
      printf("\n\t    (default:  incremental algorithm with a 10%% performance loss allowed)");
      printf("\n\t    (NB:  priority given to number of conclusions)");
      printf("\n\t -s activity threshold for an item not to be blank (default:  0.2)" );
      printf("\n\t -a for intermediate display (default: no display)");
      printf("\n\t* -wl for wordless.\n");
      printf("\n");
      return 1 ;    
    }
  
  // Construction of FIS S
  FIS * S;
  FILE *fg;
  double perfLoss,muthresh,r,coverage,MaxError;
  double ** Data;
  int i,ret,nbcol,nbrow,outputNumber=0, dType,FlagDisplay=0,nOut,nConc,len;
  char * outputFisConfigName=NULL;
  bool wordless=false;
  
  dType = 1;
  try
    {
      perfLoss=0.1;
      muthresh=0.2;
      nConc=0;
      for (i = 3; i < argc; i++)
	{
	    if(! strcmp("-a", argv[i]) ) FlagDisplay = 1;
	    else if(! strncmp("-o", argv[i], 2) ) outputNumber = atoi(argv[i] + 2);
	    else if(! strncmp("-l", argv[i], 2) ) perfLoss = strtod(argv[i] + 2, NULL);
	    else if(! strncmp("-c", argv[i], 2) ) nConc = atoi(argv[i] + 2);
	    else if(! strncmp("-s", argv[i], 2) ) muthresh = strtod(argv[i] + 2, NULL);
	    else if(! strncmp("-d", argv[i], 2) ) 
	      {
		if( strlen(argv[i]) == 2) dType = 1;
		else dType = atoi(argv[i] + 2);
	      }
	     else if(! strncmp("-wl", argv[i], 3) ) wordless = true;
	} 
       if(wordless)
	 FlagDisplay=0;
       
       S = new FIS(argv[1]);   

      // check output number
      nOut=S->GetNbOut();
      if (outputNumber >(nOut-1))
      {
	sprintf( ErrorMsg, "~invalid~output~number: %d", outputNumber);
	throw std::runtime_error( ErrorMsg );
      }

    

      if (S->GetNbRule()<=0)
	{
	sprintf( ErrorMsg, "~no~rules(Nbrule=%d)\n",S->GetNbRule());
	throw std::runtime_error( ErrorMsg );
	}
      //read data - allocate Data array
      Data = ReadSampleFile(argv[2], nbcol, nbrow);
      // evaluate performance before vocab. reduction
      r = S->Performance(outputNumber, argv[2], coverage, MaxError, muthresh, "perfinit.res",FlagDisplay );
      if (!wordless)
	printf("\n Performance before vocab. reduction: %f; Coverage level (Threshold %4.2f) : %4.2f MaxError : %11.3f\n", r, muthresh, coverage, MaxError);
      
      // launch vocabulary reduction with the specified parameters
      ret=S->VocReduc(outputNumber,Data,nbrow,muthresh,perfLoss,nConc,dType);
      // returns number of different rule conclusions
      
      if (!wordless)
	printf("\n after vocab. reduction %d rule conclusions",ret);
      //output file
      len=strlen(argv[1]) + 10;
      outputFisConfigName=new char[len];
      strcpy(outputFisConfigName,"red-");
      strcat(outputFisConfigName,argv[1]);
      
      fg = fopen(outputFisConfigName, "wt");
      S->PrintCfg(fg);
      fclose(fg);
      // evaluate performance after vocab. reduction
      r = S->Performance(outputNumber,argv[2] , coverage, MaxError, muthresh, "perf.res",FlagDisplay );
      if (!wordless)
	printf("\n Performance after vocab. reduction: %f; Coverage level (Threshold %4.2f) : %4.2f MaxError : %11.3f\n", r, muthresh, coverage, MaxError);

      // clean up
      if (S!=NULL) delete S;
      if (outputFisConfigName!=NULL) delete [] outputFisConfigName;
      // delete 2D data array
      for (i = 0; i < nbrow; i++)
	{
	  if (Data[i] != NULL) delete [] Data[i];
	}
      if (Data != NULL) delete [] Data;
      if(VarNameG != NULL) 
	{
	  for(int j = 0; j < NbVarG; j++)
	    if(VarNameG[j] != NULL)  delete [] VarNameG[j];
	  delete [] VarNameG;
	  VarNameG = NULL;
	}
      return 1;
    }
  catch(std::exception &e)
    {
      printf("\n%s", e.what());
    }
  
}

//**************************  Fin de MainOls.cpp ************************
