//*********************************************************************
//
//
//                              PERF.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// Object : Call performance function
// Environnement : Uses library fispro

//*********************************************************************

#include "fis.h"
//#include <windows.h>

//! This program needs 2 arguments:  the fis configuration file name and 
//! the data file name. Options: -a for intermediate display ; 
//! -s activity threshold for an item not to be blank (default: 0.2).
int main(int argc, char ** argv)
//******************************
{
  if(argc < 3)
    {
      printf("\nThe program %s needs 2 arguments : ", argv[0]);
      printf("\n\t* the fis configuration file name;");   
      printf("\n\t* the data file name.") ;
      printf("\nOption : ");
      printf("\n\t* -n output number");
      printf("\n\t* -s activity threshold for an item not to be blank (default: 0.2)");
      printf("\n\t* -i regression performance index (1: RMSE - Default, 2:  MAE, 3: PI");
      printf("\n\t* -f for fuzzy reference target" ); 
      printf("\n\t* -wk for kernel width when fuzzy reference target.\n" ); 
      printf("\n\t* -ws for support width when fuzzy reference target.\n" ); 
      printf("\n\t* -a for intermediate display.\n" ); 
      printf("\n\t* -wl for wordless.\n");
      
       
      return 1;
    }

  double ret, Couverture, MaxError, thres=0.2;
  int i, FlagDisplay, indtype = 0, outputNumber=0;
  bool wordless=false;
  
  double sw, kw;
  sw = kw = 0.;

  // main loop for catching exceptions in called functions
  try{
    FlagDisplay = 0; // No display by default
   

    // Construction of FIS S
    FIS * S;
    FILE *f;

    S = new FIS(argv[1]);  
     
    if(argc > 3) 
      {
	for(i = 3; i < argc; i++)
	  {
	    if(! strcmp("-a", argv[i]) ) FlagDisplay = 1;
	    else if(! strncmp("-o", argv[i], 2) ) outputNumber = atoi(argv[i] + 2);
	    else if(! strncmp("-s", argv[i], 2) ) thres = strtod(argv[i] + 2, NULL);
	    else if(! strncmp("-i", argv[i], 2) ) indtype = atoi(argv[i] + 2);
	    else if(! strncmp("-wl", argv[i], 3) ) wordless = true;
	    else if(! strncmp("-wk", argv[i], 3) ) kw = strtod(argv[i] + 3, NULL);
	    else if(! strncmp("-ws", argv[i], 3) ) sw = strtod(argv[i] + 3, NULL);
	  }
      }
    if(wordless) FlagDisplay=0;

   
    switch(indtype)
      {
      case 0: break;
      case 1: S->SetErrorIndex(FIS::RmseErrorIndex()); break;
      case 2: S->SetErrorIndex(FIS::MaeErrorIndex()); break;
      case 3: S->SetErrorIndex(FIS::PiErrorIndex()); break;
      default: break;
      }
     
    ret = S->Performance(outputNumber, argv[2], Couverture, MaxError, thres, "perf.res", FlagDisplay);
    if (!wordless)
      {
	printf("\n Created result files : perf.res and result.perf");
	if(! S->Out[outputNumber]->Classification()) 
	  printf("\n RMSE: %f   MAE:  %f   PI:  %f ", S->RMSE, S->MAE, S->PIn);
	printf("\n Performance return: %f; Coverage level (Threshold %4.2f): %4.2f MaxError: %11.3f\n", ret, thres, Couverture, MaxError);
      }

    f = fopen("result.perf", "wt");
    S->WriteHeaderPerfRB(outputNumber, f);
    fprintf(f, "\n");
    S->PerfRB(ret, Couverture, MaxError, outputNumber, f);
    fprintf(f, "\n");
    fclose(f);

    delete S;
   
    if(VarNameG != NULL) 
      {
	for(int j = 0; j < NbVarG; j++)
	  if(VarNameG[j] != NULL)  delete [] VarNameG[j];
	delete [] VarNameG;
	VarNameG = NULL;
      }
    return (0);
  }
  //action in case of exception
  catch(std::exception &e){
    printf("error in perf\n");
    printf("catch %s\n",e.what());
    
    return(-1);
  }
}


//**************************   End of PERF.CPP   *************************

