//*********************************************************************
//
//
//                              TESTKM.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// Object : Test kmeans basic functions
// 
//
//*********************************************************************

#include "common.h"
#include "km.h"

//! This program needs 2 arguments:  the fis configuration file name and 
//! the data file name. Option: -a for intermediate display.
int main(int argc, char ** argv)
  //******************************
{
  if(argc < 4)
    {
      printf("\nThe %s program is an example of kmeans use", argv[0]);
      printf("\nIt requires 3 arguments: ");
      printf("\n\t* the data file name;");
      printf("\n\t* (mono or multidimensional data separated by ;);");
      printf("\n\t* the number of clusters;") ;
      printf("\n\t* 0/1 to work with unnormalized/normalized data;") ;
      printf("\nOption : ");
      printf("\n\t* -a for detailed display. \n" );
             
      return 1;
    }

  double ** data= NULL, **datacol=NULL;
  int i,j,k,kk,kmin,kmax,naff,FlagDisplay,ng,norm;
  int nbrow=0,nbcol=0;
  double *Centres=NULL, *withinss=NULL;
  double **CentresND=NULL, **NormCentresND=NULL;
  double median,meanO, stdO;
  double *mini=NULL,*maxi=NULL,*minin=NULL,*maxin=NULL;
  char * charResultFile=NULL,* charResultFile2=NULL;
  int *cnt=NULL, *clust=NULL;

  FILE *fres=NULL,*fres2=NULL;
  

  try{
    FlagDisplay = 0; // No display by default

    if(argc > 4) 
      {
	for(i = 4; i < argc; i++)
	  {
	    if(! strcmp("-a", argv[i]) ) FlagDisplay = 1;
	  }
      }

    //read data from arg 1
   data = ReadSampleFile(argv[1], nbcol, nbrow);
   if (FlagDisplay) printf("\ndata:  %d variables\t%d rows\n",nbcol,nbrow);
   mini=new double [nbcol];
   maxi=new double [nbcol];
   
   
   //read number of clusters from arg 2
   ng=strtol(argv[2],NULL,10);
   
   // choice of kmeans: normalized or not
   norm=strtol(argv[3],NULL,10);
   
   
   // rearrange data by columns
   datacol = new double* [nbcol];
   for(i = 0; i <nbcol ; i ++)   // 
     datacol[i]=new double [nbrow];
   
   
     for(i = 0; i <nbcol; i ++)   //
       for(j = 0; j < nbrow; j++) //
	 datacol[i][j]=data[j][i];
     
   

   //compute mini and maxi for non normalized data and for each column
   for(i = 0; i <nbcol ; i ++)   // 
     {
       StatArray(datacol[i], nbrow, 0, median, meanO, stdO, mini[i], maxi[i], 0);
       if (FlagDisplay) 
	 printf("\nvariable %d\t median=%8.3f\tmean=%8.3f\tstd=%8.3f\tmin=%8.3f\tmax=%8.3f\n", i+1,median, meanO, stdO, mini[i], maxi[i]);
     }
   if (FlagDisplay) printf("\nkmeans with %d clusters\n",ng);

   // compute minin et maxin for normalized data
   minin=new double[nbcol];
   maxin=new double[nbcol];
   for (i=0;i<nbcol;i++)
     {
       minin[i]=0.0;
       maxin[i]=1.0;
     }
   //
   if (FlagDisplay) printf("\n--------------------------------\nkmeans 1D\n--------------------------------\n");
   for(i = 0; i <nbcol; i ++)   //
     { 
      // set initial centers
       if (FlagDisplay) printf("\n**********\nvariable %d\n**********",i+1);
       InitCentres(Centres, ng,mini[i],maxi[i]);


     if (FlagDisplay) 
       {
	 printf("\nInitial centers:");
	 for (j=0;j<ng;j++)
	   printf("\tcenter #%d:  %8.3f",j+1,Centres[j]);
	 printf("\n");
       }
     // kmeans 1d done using datacol
     // with norm =false (no difference for 1D Kmeans)
     //
     Kmeans(datacol[i],nbrow,Centres,ng,false);
     if (FlagDisplay) 
       {
	 printf("Centers found by the kmeans procedure:");
	 for (j=0;j<ng;j++)
	   printf("\t%8.3f",Centres[j]);
	 printf("\n");
       }

     // compute cluster counts and cluster membership
     CountClusters(datacol[i], nbrow,  ng, Centres, cnt, clust);
     if (FlagDisplay) 
       {
	 printf("cluster counts:\n");
	 for (j=0;j<ng;j++)
	   printf("\t%d",cnt[j]);
	 printf("\ncluster membership (row index,cluster #) :\n");
	 naff=(int)ceil((double)nbrow/10);
	 for (kk=0;kk<=naff;kk++)
	   {
	     kmin=kk+(kk*9);
	     kmax=kmin+10;
	     if(kmax>nbrow)kmax=nbrow;
	     for (k=kmin;k<kmax;k++)
	       printf("\t%d,%d",k+1,clust[k]+1);
	     printf("\n");
	   }
       }
     
     }
   
   
       if (FlagDisplay) printf("\n--------------------------------\nkmeans nD\n--------------------------------\n");
   
   
       // set initial nD centers

       InitCentresND(CentresND, ng, nbcol,mini, maxi);
     
       if (FlagDisplay) 
	 {
	   printf("\nInitial centers:\n");
	   for (i=0;i<ng;i++)
	     {
	       printf("Center #%d:\n",i+1);
	       for (j=0;j<nbcol;j++)
		 printf("%8.3f\t",CentresND[i][j]);
	       printf("\n");
	     }
	 }

     
       // call multidim kmeans
       // 

       Kmeans(data,nbrow,CentresND,ng,nbcol,false);
     
       if(!norm)
	 {
	   if (FlagDisplay) printf("\nUnnormalized multidimensional version\n------------------------------------------------\n");
	   for (i=0;i<ng;i++)
	     {
	       if (FlagDisplay) printf("Center #%d found by kmeans nD (variables 1 2 ...):\n",i+1);
	       for (j=0;j<nbcol;j++)
		 if (FlagDisplay) printf("%8.3f\t",CentresND[i][j]);
	       if (FlagDisplay) printf("\n");
	     }
	 }
       if (norm)
	 {
     
	   // choose initial centers between 0 and 1 for normalized kmeans
	   // des kmeans multidim
	   ResetCentresND(CentresND, ng, nbcol,minin, maxin);
	   if (FlagDisplay) 
	     {
	       if (FlagDisplay) printf("\nInitial centers:\n");
	       for (i=0;i<ng;i++)
		 {
		   for (j=0;j<nbcol;j++)
		     if (FlagDisplay) printf("\t%8.3f",CentresND[i][j]);
		   if (FlagDisplay) printf("\n");
		 }
	     }
	   // be careful! data will be normalized
	   Kmeans(data,nbrow,CentresND,ng,nbcol,true);

	   if (FlagDisplay) printf("\nNormalized multidimensional kmeans\n---------------------------------------------\n");
	   // backup normalized CentresND for cluster assignment
	   NormCentresND= new double* [ng];
	   for(i = 0; i < ng ;i++)
	     NormCentresND[i]=new double [nbcol];
	   for (i=0;i<ng;i++)
	     for (j=0;j<nbcol;j++)
	       NormCentresND[i][j]=CentresND[i][j];

	   for (i=0;i<ng;i++)
	     {
	       if (FlagDisplay)
		 {
		   printf("Centre #%d found by multidimensional (norm.) kmeans:\n",i+1);
		   for (j=0;j<nbcol;j++)
		     printf("\t%8.3f",CentresND[i][j]);

		   printf("\n");
	 
		   printf("after unnormalization:\n");
		 }
	     
	       for (j=0;j<nbcol;j++)
		 {
		   UnNormalize(CentresND[i][j],mini[j],maxi[j]);
		   if (FlagDisplay) printf("\t%8.3f",CentresND[i][j]);
		 }
	       if (FlagDisplay) printf("\n");
	   
	     }
	 }
       // compute cluster counts and cluster membership
       // reset cnt and clust
       cnt=NULL;
       clust=NULL;
       if (norm)
	 CountClustersND(data, nbcol, nbrow,  ng, NormCentresND, cnt, clust);
       else
	 CountClustersND(data, nbcol, nbrow,  ng, CentresND, cnt, clust);

       if (FlagDisplay) 
	 {
	   printf("cluster counts:\n");
	   for (j=0;j<ng;j++)
	     printf("\t%d",cnt[j]);
	   printf("\ncluster membership (row index,cluster #) :\n");
	   naff=(int)ceil((double)nbrow/10);
	   for (kk=0;kk<=naff;kk++)
	     {
	       kmin=kk+(kk*9);
	       kmax=kmin+10;
	       if(kmax>nbrow)kmax=nbrow;
	       for (k=kmin;k<kmax;k++)
		 printf("\t%d,%d",k+1,clust[k]+1);
	       printf("\n");
	     }
	   printf("\n");
	 }
     
       // compute sum of squares for each cluster
       withinss=new double[ng];
       if (norm)
	 WithinSSND(data, nbcol,nbrow, NormCentresND, ng, withinss);
       else
	 WithinSSND(data, nbcol,nbrow, CentresND, ng, withinss);
       if (FlagDisplay) 
	 {
	   for (j=0;j<ng;j++)
	     {
	       if (norm)
		 printf("Normalized within sum of squares for cluster %d: %8.3f",j+1,withinss[j]);
	       else
		 printf("Within sum of squares for cluster %d: %8.3f",j+1,withinss[j]);
	       printf("\n");
	     }
	 }
       // write centers to file
       charResultFile = new char [30];
       sprintf(charResultFile, "%s", "centers.km");
       fres = fopen(charResultFile, "wt");
       if (fres ==NULL)
	 throw std::runtime_error("~CannotOpenResultFile~centers.km\n");
       //fprintf(fres, "%d\n",ng);
       for (i=0;i<ng;i++)
	 {
	   for (j=0;j<nbcol;j++)
	     fprintf(fres,"%8.3f\t",CentresND[i][j]);
	   fprintf(fres, "\n");
	 }
       // write multidimensional cluster membership to file
       charResultFile2 = new char [30];
       sprintf(charResultFile2, "%s", "result.km");
       fres2 = fopen(charResultFile2, "wt");
       if (fres2 ==NULL)
	 throw std::runtime_error("~CannotOpenResultFile~result.km\n");
       //fprintf(fres2, "%d\n",nbrow);
       for (i=0;i<nbrow;i++)
	 {
	   fprintf(fres2,"\t%d\t%d",i+1,clust[i]+1);
	   fprintf(fres2, "\n");
	 }
       fclose(fres2);
       // write withinss to file
       for (k=0;k<ng;k++)
	 {
	   fprintf(fres,"%8.3f",withinss[k]);
	   fprintf(fres, "\n");
	 }
       fprintf(fres, "\n");
       fclose(fres);
       if (FlagDisplay) printf("Multidimensional centers and within ss stored in %s file\n",charResultFile);
       if (FlagDisplay) printf("item row number and membership stored in %s file\n",charResultFile2);
 
   
   return(0);
  }
  catch(std::runtime_error &e){
    printf("catch %s\n",e.what());
    return(-1);
  }
}


//**************************   End of TESTKM.CPP   *************************
