//*********************************************************************
  //
  //
  //                              TESTPCPE.CPP
  //
  //
  // Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
  // FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
  // Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
  // Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
  // Object : Call performance function
  // Environnement : Uses library fispro

  //*********************************************************************

#include "fis.h"

    //! This program needs 2 arguments:  the fis configuration file name and 
    //! the data file name. Options: -a for intermediate display ; 
    //! -s activity threshold for an item not to be blank (default: 0.2).
  int main(int argc, char ** argv)
    //******************************
{
  if(argc < 3)
    {
      printf("\nThe program %s needs 2 arguments : ", argv[0]);
      printf("\n\t* the fis configuration file name;");   
      printf("\n\t* the data file name.") ;
      
      return 1;
    }

  int i;
  double partc, parte;
  int nbcol,nbex;
  double ** Data=NULL;
  double * col=NULL;
  partc = parte = 0.;

  // main loop for catching exceptions in called functions
  try{
    // Construction of FIS S
    FIS * S;
    S = new FIS(argv[1]);  
    Data = ReadSampleFile(argv[2], nbcol, nbex);

    col = new double[nbex];

    for(i=0; i < S->GetNbIn(); i++)
    {
      GetColumn(Data, nbex, i, col);
      S->In[i]->PcPe(col, nbex, partc, parte);
      printf("i:  %d   pc:  %f   pe: %f\n", i, partc, parte);
    }

    delete S;
    delete col;
    // delete data
    if (Data !=NULL) {
      for( int i=0 ; i<nbex ; i++ )
	delete [] Data[i];
      delete [] Data;
    }
    if(VarNameG != NULL) 
      {
	for(int j = 0; j < NbVarG; j++)
	  if(VarNameG[j] != NULL)  delete [] VarNameG[j];
	delete [] VarNameG;
	VarNameG = NULL;
      }

    return (0);
  }
  //action in case of exception
  catch(std::exception &e){
    printf("error in perf\n");
    printf("catch %s\n",e.what());
    
    return(-1);
  }
}


//**************************   End of PERF.CPP   *************************

