//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************

#include "algo.h"
// algoPar members functions
// initialize with the following parameters
algoPar::algoPar( int pname ){
  if  ( pname == SOLISWETTS ){
    RangeNoise =   12;
    lamb1      =   0.4;
    lamb2      =   0.2;
    lamb3      =   0.5;
    maxIte     =   5;
    maxConstraints = 100;
    maxFail      =   100;
    EpsilonCenter  = .000001;
  };
  if  ( pname == FORWARD ){
    RangeNoise =   0.8;
    maxIte     = 100  ;
    maxFail      =   100;
  };
}
void algoPar::UpdatePar(int maxite,double amplinoise,int maxcons,int maxfail,double epsiloncenter,double l1,double l2,double l3)
{
  // update parameters
  maxIte=maxite;
  RangeNoise=amplinoise;
  maxConstraints=maxcons;
  maxFail=maxfail;
  EpsilonCenter=epsiloncenter;
  if ( l1 != -1 ) lamb1=l1;
  if ( l2 != -1 ) lamb2=l2;
  if ( l3 != -1 ) lamb3=l3;
}
void algoPar::SetParGauss(double amplinoise)
{
   RangeNoise=amplinoise;
}
void algoPar::SetMaxite(double ite)
{
  maxIte=ite;
}
// print the parameters state
void algoPar::Print( void ){
printf("\n  Parameters of the algorithm : ");
printf("\n  Solis and Wetts");
printf("\n  Range of gaussian noise   : %f   ",RangeNoise);
printf("\n  Max number of iterations : %i   ",maxIte); 	
printf("\n  Max number of fails          : %i ",maxFail); 	  	
printf("\n  Min distance optimized center : %f ",EpsilonCenter); 	  	
printf("\n  Redraw a gaussian before counting an iteration : %i   " , maxConstraints ); 	
printf("\n  Solis and Wetts constants %f %f %f  \n ",lamb1,lamb2,lamb3); 	  	
}
// print the parameters state on a line to create a CSV file
void algoPar::PrintCSV( void ){
printf(";%f",RangeNoise);
printf(";%i",maxIte);
printf(";%i",maxFail);
printf(";%f;%f;%f;%f",EpsilonCenter,lamb1,lamb2,lamb3);
}
// unused
algo::algo(void){
  P        = NULL;
  Name     = 0;
  Size     = 0;
  KeySize  = 0;
  fEvalArgs= NULL;
  V        = NULL;
}


algo::~algo( void ) { 
  if ( P != NULL )         delete P;
  if ( fEvalArgs != NULL ) delete fEvalArgs;
  if ( V != NULL )         delete V;
}
algo::algo( int pname , int pSize){
  P       = new algoPar(  pname );
  Name    = pname;
  Size    = pSize;
  KeySize = 0;
  fEvalArgs=NULL;
  V= new avect(pSize); //not used yet
}
// give number of fEvalArgsConteners item
algo::algo( int pname , int pSize , int NbPar ){
  P       = new algoPar(  pname );
  Name    = pname;
  Size    = pSize;
  KeySize = 0;
  fEvalArgs=new conteneur( NbPar );
  V = new avect(pSize); //not used yet
}

void algo::Load( double * tab ){
  for( int i=0 ; i < Size ; i++ )
    V->setValue(i,tab[i]);
  return;
}
void algo::initData( double * tab, int pSize ){
  Size    = pSize;
  if ( V != NULL ) delete V;
  V = new avect(pSize); 
  for( int i=0 ; i < Size ; i++ )
    V->setValue(i,tab[i]);
#ifdef DEBUG
  printf("\n+Algo allocated a %i size vector\n ",pSize); 
  V->print();
#endif
  return;
}

int algo::getKeySize() const {
  return Key.Size();
}

vkey * algo::getKey()  {
  vkey * keyaddr=&Key;
  return keyaddr;
}
void algo::setKey( unsigned long int value ){
  Key.setValue(value) ;
  KeySize = getKeySize();
}

void algo::setKey( const char *value ){
  Key.setValue(value) ;
  KeySize = getKeySize();
}

void algo::setKey( const vkey * pKey ){
  Key.setValue( pKey ) ;
  KeySize = getKeySize();
}

void algo::UpdateKeyLenght( int L ){
  KeySize = L;
}

int algo::launch(
                 void     * Structur,
                 double   * ArrayOfDouble ,
                 double   * gain,
                 double  (* pfEval)( void   *  ,
                                     vkey   *  ,
                                     double *  ,
                                     int       ,
                                     conteneur *)
 )
{
  if ( Name == SOLISWETTS )
    { return SolisWetts( Structur,
      & Key ,
      ArrayOfDouble ,
      Size,
      P ,
      gain ,
      pfEval ,
      fEvalArgs);
    };
  if ( Name == FORWARD )
    { return FwdIfBetter( Structur,
       & Key ,
       ArrayOfDouble ,
       Size,
       P ,
       gain ,
       pfEval ,
       fEvalArgs);
    };
  if ( Name == PERMUTATION )
    { return Permutation( Structur,
       & Key ,
       ArrayOfDouble ,
       Size,
       P ,
       gain ,
       pfEval ,
       fEvalArgs);
    };
  return(-9);
}
