//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************
/*!       \class   algo
 *        \brief   Define a language to implement any algorythm on any object.
 * 
 *  The algo class is used to applying an algorithm on a set of values using an evaluation function. Sample file are given that show the optimisation of a Point ( C++ file sampleP.cpp  ) and an Array ( C file sampleB.cpp and file sampleB2.cpp ), it's also used whith Fuzzy Inference System.\n
 *  The aim of this class is to define a generic language to write algorythm to optimize FIS, Point, Array, apple, pear .... only 2 algorythms are implement now, Hill climbing and Solis Wetts. ( Debugging FIS class code was the main obstacle to it).\n
 *  The Evaluation function have to be positive, an negative result would mean that a constraint has been violated. ( see more detail in testConstraints member function.) but don't worry if your is not, U can easily change it to an positive equivalent one ...
 *  \todo implement gradient and genetic algo whith the point sample to use it whith FIS...
 */
/*!       \class   algoPar
 *        \brief   Used as a struct to store algorythms params.
 *
 *   The algoPar, used as a struct,is used for passing parameters to the algorythm.
 *   U can easely change a Param of your optimisation :
 *   if varOpt is the address of a algo instance or sifopt instance\n
 *   The Generic notation is \n
 *            varOpt->P-><Parameter>=Value.\n
 *   An example :\n
 *            varOpt->P->RangeNoise=0.1\n
 */
#ifndef __ALGO_H
#define __ALGO_H

// commun functions
#include "common.h"
// algo vector key class
#include "vkey.h"
// algo vector class
#include "avect.h"
// used in feval store an array of address ( void ** ) used for fEval arguments.
#include "conteneur.h"
// we separate the code of each algorythm in several souces sw.cpp fib.cpp
// the declaration of these algorythm function should be in :
#include "fctalgo.h"

/*! 
   \def SOLISWETTS 1
   This pingouin's style define is used to pass to the optimizer the name of the algo\n
*/
// The name of algorythms
// Solis and wetts
#define PERMUTATION 0
// Solis and wetts
#define SOLISWETTS 1
// Forward if better Hill climbing
#define FORWARD 2
class algoPar {
public :
  double RangeNoise;  //!< range of noise for gaussian vectors
  double lamb1;   //!< used in SW's vector transformation =  0.4
  double lamb2;   //!< used in SW's vector transformation =  0.2
  double lamb3;   //!< used in SW's vector transformation =  0.5
  int    maxIte;      //!< maximum of iterations	
  int maxConstraints; //!< the maximum times we can violated a constraints before it counts as an interaion
  int maxFail;   //!< the max iteration to do without a better perf
  double EpsilonCenter;   //!< the distance for which U consider 2 optimized centers as equal
  algoPar( int );
  ~algoPar(){};
  void UpdatePar(int maxite,double amplinoise,int maxcons,int maxfail,
		 double epsiloncenter,double l1,double l2,double l3);
  void SetParGauss(double amplinoise);
  void SetMaxite(double maxite);
  void Print(void); //!< print all the params on STDOUT
  void PrintCSV(void);    //!< print trace on the output in CSV form ie data1;data2... to redirect and reopen the trace whith another program :Gnumeric ,.. 
};
class algo {
  algo& operator=(const algo&);
  algo(const algo&);
protected:
  // The key as explained in the begining
  vkey Key; //<! The key to transorfm a part of your object into a set of updatables values
  int  KeySize; //<! guess...
  //! The name as defined SOLISWETTS 1 HILLCLIMBING 2 ... 
  int       Name;

public :
  //! The Size of the Set of Values
  int       Size;
  //! a vector with the initials datas
  avect   * V;
  //! fEvalArgs contains the arguments of fEval
  conteneur * fEvalArgs;
  //! Parameters of the algorythm
  algoPar * P;
  algo( void );
  algo( int , int );
  algo( int , int , int);
  ~algo( );
//! creates the vector whith the initial values 
  void initData( double * , int);

  int launch(  void  * YO , double   * AD , double * gain , double  (* pfEval)( void * , vkey *, double * , int , conteneur * ));
/*!  \fn  int launch(  void  * YO , double   * AD , double * gain , double  (* pfEval)( void * , vkey *, double * , int , conteneur * ))
 *  \brief This Member function launch the algorihtm
 *  \param YO The adress of your object casted to void * ( For it can works whith other classes than FIS.)
 *  \param AD The set of updatables values to optimize.
 *  \param gain The Evaluation at the end of the algoryhtm.
 *  \param pfEval A pointer to a wrapped object member function or to a normal funtion.
 *  \return 0 if there where MaxIte iterations whithout a gain of Evaluation or the number of iterationsor -1 if the key were empty\n
 */
  void Load( double *);
//! Key manipulation  
  void UpdateKeyLenght( int );
//! Key manipulation  
  int getKeySize( void ) const;
//! Key manipulation  
  void setKey( unsigned long int );
//! Key manipulation  
  void setKey( const char * );
//! Key manipulation  
  void setKey( const vkey * pKey );
//! Key manipulation  
  vkey* getKey() ;

};
#endif

