//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************
/*! \class avect
    \brief  algo vector class can handle vectors,keys and gaussians.

    Declaration of algo vector class permit a simple use of vectors while writing your algorythm, extract subvectors whith keys and generate gaussian vectors.
    \todo check every functions complexity, lower it.
    \todo get more generic than double.
*/
#ifndef AVECT_H
#define AVECT_H
#include <stdio.h>
#include <math.h>
#include "vkey.h"
#include "common.h"
class avect {
  avect& operator=(const avect&);
  avect(const avect&);
      private :
//! the array of values store as C language values of double type.
              double * vectValue;
//! the number of items.
              int      nbElem;
      public  :
//! generate an empty instance.
      avect() { vectValue = NULL; nbElem = 0; };
/*! 
 * \fn avect( int Vsize, double *DoubleArray )
 * \brief generate an ready to use instance.
 * \param Vsize the size to be allocated.
 * \param DoubleArray the address of an C array of double.
 * \return the instance adress as any constructor does.
 * 
 * generate an instance whith the allocation of the array whith intial values.
 */
      avect( int Vsize, double *DoubleArray );
/*! 
 * \fn avect( int Vsize )
 * \brief generate an allocated but empty instance.
 * \param Vsize the size to be allocated.
 * \return the instance adress as any constructor does.
 * 
 * generate an instance whith an allocated array of double of Vsize size.
 */
      avect( int Vsize );
      ~avect()  { if ( vectValue != NULL ) delete [] vectValue; };
//! dupVector  duplicate a vector's value using a key, it creates a sub vector for those indices corresponding to true in the key
       void   dupVector( avect * , vkey * );
//! get value of the i'th item.
       double getValue(int i) const;
//! return the adress where C language allocated your double array.
       double *getArrayAdress(){return vectValue;};
//! get the number of allocated doubles.
       int    getNbElem(void) const;
//! set value d to the i'th item.
       void   setValue(int i, double d );
//! print the vector on stdout.
       void   print(void) const;
//! print the firsts arg item of the vector on stdout.
       void   print(int) const;
//! addVector equivalent to C language this += v
       void  addVector(avect * );
//! dupVector duplicate a vector
       void  dupVector(avect * );
//! tlambVector scalar multiplication
       void  tlambVector( double  );
//! moinsuVector equivalent to unitary minus
       void  moinsuVector( void );
//! init0 initiale vector with zeros
       void  init0( void );
//! generate a vector by adding a noise of range Range to an existing vector Average.
       void initGauss( double Range,  avect * Average);
      };
#endif
