//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************
#include "algo.h"
/*!
 * \file fib.cpp
 * \brief defines the Hill climbing algorythm ( forward if better ) using the algo language.

 This function defines the Hill climbing algorythm ( forward if better )  algorythm using the algo language, it considers that the evalalution is positive and that a constraint violation returns -1 to the evaluation. As this algo is the simpliest you know, this file might get your blank model for a new algorythm.
 */
int FwdIfBetter(
	       void     * YourObject,
	       vkey     * Key,
	       double   * ArrayOfDouble ,
	       int        SizeArray,
	       algoPar  * P,
	       double   * gain,
	       double  (* pfEval)( void   *  ,
				   vkey    *  ,
				   double *  ,
				   int       ,
				   conteneur *),
	       conteneur * fEvalArgs
	       )
{
  long int     SWite=0;  // nb of iteration
  int          KeySize=0;  // Number of true set in the key
  int          SWC=0;  // nb of constraints violation for this iteration
  double	SWE;      // Evaluation of E
  double	SWprevE;  // Previous evaluation of E
  double	SWfirstE; // first Evaluation of E
  int	SWFail=0;
  KeySize=Key->Size();
  if ( KeySize == 0 ) return -1;
///////////////////////////////////////////////
// Init
////////
// Put input datas in a vector object VectDouble
  avect *VectDouble=new avect( KeySize , ArrayOfDouble );
  SWfirstE=(*pfEval)( YourObject ,
		      Key ,
		      VectDouble->getArrayAdress() ,
		      VectDouble->getNbElem()      ,
		      fEvalArgs);

#ifdef DEBUG2
  printf("\n Forward if Better recoit ");VectDouble->print();
  printf("\n longeur vecteur %i\n" , KeySize );
#endif

  // create a new vector with fewer args
  avect *T = new avect(KeySize);
  avect *T2= new avect(KeySize);
  //  Generate a gaussian vector
  avect *G = new avect(KeySize);
  avect *AvgGauss= new avect(KeySize);
  AvgGauss->init0();
  G->initGauss(P->RangeNoise,AvgGauss);

#ifdef DEBUGS
  printf("\n vect Depart \n");
  VectDouble->print();
  printf("\n premier gaussien moy 0 bruit %f ", P->RangeNoise );
  G->print();
  printf("%d",KeySize);
#endif
// let's get in the while with : previous correct and worse than current
  SWprevE=SWfirstE+1;

#ifdef DEBUGS
  printf("\n Performence initiale %f ",SWfirstE );
#endif
///////////////////////////////////////////////
// Main Loop
////////
  while (  SWite < P->maxIte )
    {

#ifdef DEBUG3
      printf("\nIteration %d ", SWite );
#endif
      //  T <- A + G
      T->dupVector(VectDouble);
      T->addVector(G);
      //  SWE <- EVAL A+G
      SWE=(*pfEval)( YourObject ,
		     Key ,
		     T->getArrayAdress() ,
		     T->getNbElem()      ,
		     fEvalArgs);

      // we consider that the evalalution is positive
      // a constraints violated return -1 to the evaluation
       if ( SWE != -1  )
	{
  if (( SWE < SWprevE ) && ( fabs ( SWprevE - SWE ) > EPSILON ))
      {
       VectDouble->dupVector(T);
       SWprevE=SWE;
       SWFail=0;
#ifdef DEBUG2
       printf("\n On avance");
       printf("\n Nouveau vect " ); VectDouble->print();
       }else
       {
       printf("\n un coup pour rien");
       printf("\n on reste sur " ); VectDouble->print();
#endif
       SWFail++;
       };
       };
       if ( SWE == -1 )
      { SWE = SWprevE;
        SWC++;
#ifdef DEBUG2
      printf("\n colision n %i\n",SWC);
#endif
        if ( SWC > P->maxConstraints)
           {
           SWC=0;
           SWite ++;
#ifdef DEBUG2
      printf("\n max colision atteint , iteraion  %li\n",SWite);
#endif
           };
      }
      else
          {  SWite ++;
             SWC=0;};
      G->initGauss(P->RangeNoise,AvgGauss);

#ifdef DEBUG2
      printf("\n Distance au but de %f vers %f",SWprevE,SWE);
      printf("\n Nouveau gaussien " ); G->print();
#endif


    }

#ifdef DEBUG2
  printf("\n Fin au bout de %ld itrations ",SWite);
  printf("\n Distance au but de %f vers %f",SWfirstE,SWE);
  printf("\n Parametre ");P->Print();
  printf(" vecteur VectDouble " ); VectDouble->print();
#endif
  *gain=(*pfEval)( YourObject ,
              Key ,
              VectDouble->getArrayAdress() ,
              VectDouble->getNbElem()      ,
              fEvalArgs);
#ifdef DEBUG2
  printf("\n+SWFIN gain vaut %f \n",SWE );
#endif
  //  free memory
  delete AvgGauss;
  delete G;
  delete VectDouble;
  delete T;
  delete T2;

  return 0;
}
