//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************
/*! testsifopt.cpp 
 * \brief Test and interprete the params,\n buid the Samples whith the argv2 file,\n lauch the algorythm.\n 
 *
 */

#include "common.h"
#include "fis.h"
#include "sifopt.h"
#include "gensample.h"

FIS* medfis(char* prefix,char* suffix,int ns, bool wordless);
int PrintArgsLoopOS(int argc, char ** argv);
int ReadInputsOpt(const char * stri, int nbmax,double *numinputs,bool wordless);
int CallOptim(sifopt *&Opt,FIS * &S, vkey * KeyIn, double ** Exemples, char * outfile, int nOut, int &nbcol, int &nbrow, double perfinit, double muthresh, bool fisoutflag, char * outputFisConfigName, bool wordless);
sifopt* PrepareOpt(FIS * S, int * &ResClassif, double * &Lab, double ** &Exemples, int &nbrow, int &nOut, int &cnear, double &InitialCoverage, double &covloss,double &MaxError, double &muthresh,parstruct paramP);
vkey* BuildKey(sifopt*Opt,FIS* S,int keytype,int num,bool wordless);
double GenIterName(char *& outputFisConfigName, int keytype, int numi);
void GenPerfName(char *outputFisConfigName,const char* fisPath,char * optimName, char * outfile,char * basename, int isample);
void WritePerfFile(FILE* perfile, int ns, char * fisfile1, char * fisfile2, FIS * M, char * datafile, double muthresh, int nOut, bool header, bool wordless);
void computeWritePerf(FIS* M,const char* fisPath,const char * datafile,const char * perfile,const char * fisname,const char* optimName,  double muthresh, int nOut, int ns, bool wordless);
FIS* loopoptimsample(FIS* &S,const char* fisPath,const char* datafile,const char * optimName, int nsample,int nsteps,bool optimIn,bool optimOut,bool optimRules, int nbinputs,double * numinputs, int nOut,long int seed,double muthresh,double covloss,parstruct paramP,bool intermfile,bool wordless,double nearC);

  int main(int argc, char ** argv)
//******************************
{
  int ret;
  ret=PrintArgsLoopOS(argc,argv);
  if (ret !=0) return 0;
  //example
  //... iris.fis iris 
  //  Gaussian noise=very important parameter

  double ** Exemples  = NULL;   // A Pointer to sample array
  int * ResClassif = NULL;
  double * Lab = NULL;
  int nbcol=0, nbrow=0, cnear=0, numOpt=0 ;
  char *basename=NULL;
  char * fisout=NULL;;
  double * numinputs=NULL;
  char blank=' ';
  char * iter=NULL;
  double muthresh=.1;
  int i,nOut=0;
  int reponse=-1,len=0,nbinputs=0;
  int nsteps=2;
  double l1=-1,l2=-1,l3=-1;
  double gain=-1;
  char * optimName=NULL;
  char * outputFisConfigName=NULL, *finalFisConfigName=NULL;
  bool intermfile=false;
  bool optimRules=false,optimIn=false,optimOut=false;
  int argstr=0;
  char* perfile=NULL;
  double nearC=0.001;
  
  // Added by SG - March 16th, 2010
  double covloss = 0.1;
  bool wordless=false, flagDisplay=false;
  double InitialCoverage=0.0;
  double *VPos = NULL;//to hold distinct output values (last column of data file)
  double min1,max1,xnorm,vp;
  int ns=10;
  double rs=0.25;
  bool Flagcs=false;
  double perf0, perfopt,perfall,perfallmed,perfmed;
  FIS* M=NULL;
  long int seed=0;
  double ga;

  //now default parameters as in the interface
  int maxite=100,maxcons=1000,maxfail=1000;
  double gauss=0.005, epsiloncenter=1e-6;
    

  int nbmax;
  FILE * fg=NULL;
  double eps=1e-6;//default precision
  long int seed1=0;

  optimName=new char[6];
  strcpy(optimName,"optim");

  // check and transform arguments
  if(argc >=3) 
    {
      for(i = 3; i < argc; i++)
        {
	  
	  if(! strncmp("-b", argv[i],2))
	    {
	      delete [] optimName;
	      optimName=new char[strlen(argv[i+1])+1];
	      strcpy(optimName,argv[i+1]);
	    }
	  if(! strncmp("-it", argv[i],3))
	    maxite=atoi(argv[i+1]);	
	  if(! strncmp("-e", argv[i],2))
	    gauss=strtod(argv[i+1], NULL);
	  if(! strncmp("-v", argv[i],2))
	    maxcons=atoi(argv[i+1]);
	  if(! strncmp("-f", argv[i],2))
	    maxfail=atoi(argv[i+1]);
	  if(! strncmp("-d", argv[i],2))
	    epsiloncenter=strtod(argv[i+1], NULL);
	  if(! strncmp("-in", argv[i],3)) 
	    {
	      optimIn=true;//next argument will hold string of input numbers
	      argstr=i+1;
	    }
	  
	  //
	  else if(! strncmp("-ns", argv[i],3)) ns = atoi(argv[i+1]);
	  else if(! strncmp("-nc", argv[i], 3)) 
	    {
	      nearC=strtod(argv[i+1], NULL);
	      if (nearC<0 || nearC>1)
		{
		  printf("Constraint between adjacent kernels must be in [0, 1]\n");
		  return(1);
		}
	    }
	  else if(! strncmp("-cs", argv[i],3)) Flagcs = true;	 
	  else if(! strncmp("-rs", argv[i],3)) rs = strtod(argv[i+1], NULL);
	  //
	  else if(! strncmp("-r", argv[i],3) ) optimRules=true;
	  else if(! strncmp("-o", argv[i],2)) optimOut=true;

	  else if(! strncmp("-n", argv[i],2)) nOut=atoi(argv[i+1]);
	  else if(! strncmp("-m", argv[i],2)) muthresh=atof(argv[i+1]);
	 
          else if(! strncmp("-s", argv[i],2)) 
	    {
	  
	      seed1=atol(argv[i+1]);
	    }
          else if(! strncmp("-l1", argv[i],3)) l1=atof(argv[i+1]);
          else if(! strncmp("-l2", argv[i],3)) l2=atof(argv[i+1]); 
          else if(! strncmp("-l3", argv[i],3)) l3=atof(argv[i+1]);
	  else if(! strncmp("-cn", argv[i],3)) cnear=1;
	  else if(! strncmp("-u",argv[i],3)) nsteps= atoi(argv[i+1]);
	  else if(! strncmp("-c",argv[i],2)) covloss=atof(argv[i+1]);
	  else if(! strncmp("-g",argv[i],2)) intermfile=true;
	  else if(! strncmp("-wl", argv[i], 3)) wordless = true;
	  else if(! strncmp("-a", argv[i], 2)) flagDisplay = true;
	
	}
    }
  
  // modif January 2013 - always set seed
   if(seed1==0) 
      seed1=time(NULL);//machine clock

    setseed(seed1);//default or argument 
    //done also in  genNsample and loopoptimsample functions
    
  // max sample pairs
  if (ns >1000)
    {
      printf("\n ns=%d is too high (max 1000)\n",ns);
      return 1;
    }
  if (!wordless)
    {
      for (i=0;i<argc;i++)
	printf("argument %d = %s  ",i,argv[i]);
    }
 
  if (flagDisplay)
    {
      printf("\nnsteps=%d,covloss=%f,seed=%d,muthresh=%f",nsteps,covloss,seed1,muthresh);
      if (ns>0)
	printf("\nsampling with ns=%d,rs=%f,Flagcs=%s",ns,rs,Flagcs);
    }
  
  try{
    
  
    // Build Fis
    FIS * S=NULL;
    S = new FIS(argv[1]);
    nbmax=S->GetNbIn();
    if (nbmax >100)
      {
	printf("\n number of inputs=%d is too high, maximum is 100\n");
	return 1;
      }


    // Generate config file 
    fg = fopen("initial.fis", "wt");
    S->PrintCfg(fg);
    fclose(fg);
    delete S;
    
    // default=optimize all rules if no other flag is set
    if (!optimIn && !optimOut && !optimRules)
      optimRules=true;
    //

    if (optimIn)
      {
	//iter=new char[20];
	numinputs=new double[nbmax];
	nbinputs=ReadInputsOpt(argv[argstr],nbmax,numinputs,wordless);
	if (nbinputs<=0) 
	  {
	    printf("\n error in reading input numbers\n");
	    return nbinputs;
	  }
      }
  
    // generate FIS from initial FIS configuration
    S = new FIS(argv[1]); 
    // max sample=1000 (4 digits), maxinput=100
    // to protect in jni
    if(rs>0.9) rs=0.9;
    if (ns==1) ns=2;
    
    if (!wordless && (ns>0))
      printf("%s %d %s %s %s\n","\nGenerating ",ns," samples with names ",basename,"n");
    //generate samples for training/test
    // max sample=1000 (4 digits), maxinput=100
    // protection necessary in main or jni
    if(ns>0) genNSamples(seed1,rs,Flagcs,eps,nbmax,argv[2],ns);
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // loop on learning samples instead of data file
    // loop on training samples-if no sample isample=-1 and ns=0
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // returns medfis (if ns>0) or NULL (if ns=0)
    parstruct paramP={maxite,gauss,maxcons,maxfail,l1,l2,l3};
    M=loopoptimsample(S,"",argv[2],optimName,ns,nsteps,optimIn,optimOut,optimRules,nbinputs,numinputs,nOut,seed1,muthresh,covloss,paramP,intermfile,wordless,nearC);
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // build for each sample
    // perf file: initial perf on whole sample, training, initial perf on test
    //            optimized perf on whole sample, on training, optimized perf on test
    perfile=new char[strlen("perf.res")+1];
    strcpy(perfile,"perf.res");
    computeWritePerf(M,"",argv[2],perfile,argv[1],optimName,muthresh,nOut,ns,wordless);
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // loop on test samples plus learning samples, instead of data file
    //if no sample isample=-1 and ns=0
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    // delete temp arrays
    if (M!=NULL) delete M;
    if (S != NULL) delete S;
    if (optimName!= NULL) delete []optimName;
    if (numinputs != NULL) delete [] numinputs;
    if (iter!= NULL) delete [] iter;
    if (perfile!= NULL) delete []perfile;		     
  
 
    if (!wordless)
      printf("\n***********************\nFIS opt END\n");

    // delete global arrays
    if(VarNameG != NULL) 
      {
	for(int j = 0; j < NbVarG; j++)
	  if(VarNameG[j] != NULL)  delete [] VarNameG[j];
	delete [] VarNameG;
	VarNameG = NULL;
      }

    
    return reponse;  
  }
  catch(std::exception &e){
    printf("error in loopoptim\n");
    printf("catch %s\n",e.what());
    return -1;
  }

}
