//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************
/*! testsifopt.cpp 
 * \brief Test and interprete the params,\n buid the Samples whith the argv2 file,\n lauch the algorythm.\n 
 *
 */

#include "common.h"
#include "fis.h"
#include "fuztarget.h"
#include "sifopt.h"
#include "gensample.h"

FIS* medfis(char* prefix,char* suffix,int ns)
{
  FIS *fis =NULL, *tmpfis=NULL;
  int ninputs=0,i,j,k,len,is,nbr;
  int * nmf=NULL;
  int ** nbparams=NULL;
  char * FisConfigName=NULL;
  char * snum=NULL;
  double ****par=NULL;
  double firstq, median, thirdq,min1,max1;
  double * tab=NULL;
  double ** conclu=NULL;
  MF * mf=NULL;
  int wordless=1;
  

  len=strlen(prefix)+strlen(suffix)+6;
  FisConfigName=new char[len];
  snum=new char[5];
  //first FIS=model
  strcpy(FisConfigName,prefix);
  sprintf(snum,"%i",0);
  strcat(FisConfigName,snum);
  strcat(FisConfigName,suffix);
  fis=new FIS(FisConfigName);
  //fis->PrintCfg();
  ninputs=fis->GetNbIn();

  nmf=new int[ninputs];
  nbparams=new int*[ninputs];
   
  par=new double ***[ns];
  conclu=new double *[ns];
  for (is=0;is<ns;is++)
    par[is]=new double **[ninputs]; 

  
  //examine each input for sample 0
  for (i=0;i<ninputs;i++)
    {
      // for each MF, get MF number, type and bounds
      nmf[i]=fis->In[i]->GetNbMf();
      nbparams[i]=new int[nmf[i]];
      par[0][i]=new double *[nmf[i]];
      for (j=0;j<nmf[i];j++)
	{
	  mf=fis->In[i]->GetMF(j);
	  nbparams[i][j]=mf->NbParams();
	  par[0][i][j]=new double[nbparams[i][j]];
	  mf->GetParams(par[0][i][j]);
	}
    }
  nbr=fis->GetNbRule();
  conclu[0]=new double[nbr];
  // rule conclusions
  for (i=0;i<nbr;i++)
    conclu[0][i]=fis->Rule[i]->GetAConc(0);
  //other samples
  for (is=1;is<ns;is++)
    {
      if (!wordless)
	cout <<endl<<"sample # "<<is;
      strcpy(FisConfigName,prefix);
      sprintf(snum,"%i",is);
      strcat(FisConfigName,snum);
      strcat(FisConfigName,suffix);
      tmpfis=NULL;
      tmpfis=new FIS(FisConfigName);
      //tmpfis->PrintCfg();
      //examine each input
      for (i=0;i<ninputs;i++)
	{
	  // for each MF, get MF number, type and bounds
	  nmf[i]=tmpfis->In[i]->GetNbMf();
	  par[is][i]=new double *[nmf[i]];
	  for (j=0;j<nmf[i];j++)
	    {
	      mf=tmpfis->In[i]->GetMF(j);
	      nbparams[i][j]=mf->NbParams();
	      par[is][i][j]=new double[nbparams[i][j]];
	      mf->GetParams(par[is][i][j]);
	    }
	}
    
      // rule conclusions
      conclu[is]=new double[nbr];
      for (i=0;i<nbr;i++)
	conclu[is][i]=tmpfis->Rule[i]->GetAConc(0);
      delete tmpfis;
    }
   
  //average each parameter
  for (i=0;i<ninputs;i++)
    {
      if (!wordless)
	cout <<endl<<"i="<<i<<" nmf="<<nmf[i];
      for (j=0;j<nmf[i];j++)
	{
	  //cout <<endl<<"mf="<<j<<" nbparams="<<nbparams[i][j];
	  for (k=0;k<nbparams[i][j];k++)
	    {
	      tab=new double[ns];
	      for (is=0;is<ns;is++)
		tab[is]=par[is][i][j][k];
	      StatArrayQuart(tab, ns, 0, firstq, median, thirdq,min1,max1);
	      delete[] tab;
	      // update par
	      par[0][i][j][k]=median;
	      //cout <<endl<<"k="<<k<<" median. param="<<median<<" par="<<par[0][i][j][k];
	      //cout<<" firstq="<<firstq<<" thirdq="<<thirdq<<endl;
	      if (!wordless)
		cout<<endl<<i<<";"<<j<<";"<<k<<";"<<median<<";"<<thirdq<<";"<<firstq;
	      // update MF parameters
	      fis->In[i]->GetMF(j)->Update(par[0][i][j]);

	    }
	}
    }
  for (i=0;i<nbr;i++)
    {
      tab=new double[ns];
      for (is=0;is<ns;is++)
	tab[is]=conclu[is][i];
      StatArrayQuart(tab, ns, 0, firstq, median, thirdq,min1,max1);
      delete [] tab;
      if(!wordless)
	cout<<endl<<i<<";"<<median<<";"<<thirdq<<";"<<firstq;
      // update rule conclusions
      fis->Rule[i]->SetAConc(0,median);
    }
  //delete objects
  delete [] FisConfigName;
  
  for (is=0;is<ns;is++)
    delete [] conclu[is];
  delete [] conclu;

  delete [] snum;
  

  for (i=0;i<ninputs;i++)
    delete [] nbparams[i];
  delete [] nbparams;
  
  
  for (is=0;is<ns;is++)
    {
      for (i=0;i<ninputs;i++)
	{
	  for (j=0;j<nmf[i];j++)
	    delete [] par[is][i][j];
	  delete [] par[is][i];
	}
      delete [] par[is];
    }
      
  delete [] par;

  delete [] nmf;
  
  

  return fis;
}


int main(int argc, char ** argv)
//******************************
{
  if(argc < 9)
    {
      printf("\nThe programm %s needs 8 arguments : ", argv[0]);
      printf("\n\t* the name of the FIS configuration file;");   
      printf("\n\t* the name of the data file used for tuning;");   
      printf("\n\t* the basename of the resulting FIS configuration file;");   
      printf("\n\t* the number of iterations ;");   
      printf("\n\t* the range of the gaussian noise;");
      printf("\n\t* the max number of constraints violations before it counts as an iteration;");
      printf("\n\t* the max number of failure steps in an algorithm;");
      printf("\n\t* the equality center distance threshold");
      printf("\n\t* default optimize all rules");
      //sampling args
      printf("\n\t* -ns Number of samples to generate from data file");
      printf("\n\t* -cs Draw samples with respect to class ratio in data file (last column)");
      printf("\n\t* -rs Ratio");
      printf("\n\t\tdefault value=0");
      printf("\n\t\t -> new random sampling");
      //end sampling args

      printf("\n\t* -in \"x y \", the string of input numbers to optimize (starting at 1);"); 
      printf("\n\t* -r optimize all rules (conclusions)");
      printf("\n\t* -o optimize fuzzy output");

      printf("\n\t* -n   Output number to consider (default=0: first output);");
      printf("\n\t* -m   minimum membership;");
      printf("\n\t* -s   integer to set seed value for parker miller random;");
      printf("\n\t* -l1  Solis Wetts Constant 1;");
      printf("\n\t* -l2  Solis Wetts Constant 2;");
      printf("\n\t* -l3  Solis Wetts Constant 3;");
      printf("\n\t* -u   Number of loops for optimizing (default=2);");
      printf("\n\t* -cn  Activate the CNear Constraint;");
      printf("\n\t* -c relative tolerated loss loss of coverage (default: 0. ; 1.=100.0%%).");
      printf("\n\t* -g create intermediate files (default=none).");
      printf("\n\t* -f1 for fuzzy reference target" );
      printf("\n\t* -f2 for full fuzzy reference target" );
      printf("\n\t* -wk for kernel width when fuzzy reference target." ); 
      printf("\n\t* -ws for support width when fuzzy reference target." );
      printf("\n\t* -mR for max right bound.");
      printf("\n\t* -mL for min left bound."); 
      printf("\n\t* -t for centroid correction (if no MF overlap)."); 
      printf("\n\t* -ne0 for left neighbor."); 
      printf("\n\t* -ne1 for right neighbor."); 
      printf("\n\t* -r2 for ro2 calculation."); 
      printf("\n\t* -bi for distance combination."); 
      printf("\n\t* -a for intermediate display." ); 
      printf("\n\t* -wl for wordless.\n");
  
      return 1;
    }
  //example
  //... iris.fis iris out.fis 100 0.005 1000 1000 0.000001 -in '3 4' -r -out -m 0.2 -c 0.10
  // attention bruit gaussien = parametre tres important

  double ** Exemples  = NULL;   // A Pointer to sample array
  int * ResClassif = NULL;
  double * Lab = NULL;
  int nbcol=0, nbrow=0, cnear=0, numInput=0 ;
  char NomFichier[1000];
  char NomFichierB[1000];
  char * fisout=NULL;;
  double * numinputs=NULL;
  char blank=' ';
  char * iter=NULL;
  double Couverture=0.0, MaxError=0.0,museuil=.1,perfinit=0.0,perf=0.0;
  int i, ElimRedondance=0,nOut=0;
  int reponse=-1,len=0,nbinputs=0;
  int nsteps=2;
  double l1=-1,l2=-1,l3=-1;
  double gain=-1;
  char * outputFisConfigName=NULL, *itername=NULL, *outputGain=NULL;
  bool last=false,intermfile=false;
  bool optimRules=false,optimIn=false,optimOut=false;
  int argstr=0;
  
  vkey *KeyIn,*KeyIn2, *KeyRules, *KeyOut;
  
  // Added by SG - March 16th, 2010
  double cov = 0.;
  bool wordless=false;
  double InitialCoverage=0.0;
  // parameters for f1 and f2
  double sw, kw,sw1,kw1;
  // parameters for f2
  double neigh0=0.0,neigh1=0.0;
  bool n0=false,n1=false;
  double *VPos = NULL;//to hold distinct output values (last column of data file)
  double min1,max1,xnorm,vp;
  int ns=0;
  double rs=0.25;
  bool Flagcs=false;
  int isinit=-1;
  double perf0, perfopt,perfall,perfallmed,perfmed;
  FIS* M=NULL;
  long int seed=0;
  double ga;

  sw = kw = 0.;

  if(argc > 9) 
    {
      for(i = 8; i < argc; i++)
        {
	  if(! strcmp("-in", argv[i])) 
	    {
	      optimIn=true;//next argument will hold string of input numbers
	      argstr=i+1;
	    }
	  else if(! strcmp("-r", argv[i]) ) optimRules=true;
	  else if(! strcmp("-o", argv[i]) ) optimOut=true;

	  //
	  else if(! strncmp("-ns", argv[i],3)) ns = atoi(argv[i] + 3);
	  else if(! strncmp("-cs", argv[i],3)) Flagcs = true;	 
	  else if(! strncmp("-rs", argv[i],3)) rs = strtod(argv[i] + 3, NULL);
	  //
	  else if(! strcmp("-n", argv[i]) ) nOut=atoi(argv[i+1]);
	  else if(! strcmp("-m", argv[i]) ) museuil=atof(argv[i+1]);
	 
          else if(! strcmp("-s", argv[i]) ) 
	    {
	      setseed(atoi(argv[i+1]));
	      seed=atol(argv[i+1]);
	    }
          else if(! strcmp("-l1", argv[i]) ) l1=atof(argv[i+1]);
          else if(! strcmp("-l2", argv[i]) ) l2=atof(argv[i+1]); 
          else if(! strcmp("-l3", argv[i]) ) l3=atof(argv[i+1]);
	  else if(! strcmp("-cn", argv[i]) ) cnear=1;
	  else if(! strcmp("-u",argv[i]) ) nsteps= atoi(argv[i+1]);
	  //if(! strcmp("-r", argv[i]) ) ElimRedondance = 1;
	  // Added by SG - March 16th, 2010
	  else if(! strcmp("-c",argv[i]) ) cov=atof(argv[i+1]);
	  else if(! strcmp("-g",argv[i]) ) intermfile=true;
	  else if(! strncmp("-f1", argv[i],3)) FuzRefG = true;
	  else if(! strncmp("-f2", argv[i],3))  FuzRefGfull= true;
	  else if(! strncmp("-wk", argv[i], 3) ) kw = strtod(argv[i] + 3, NULL);
	  else if(! strncmp("-ws", argv[i], 3) ) sw = strtod(argv[i] + 3, NULL);
	  else if(! strncmp("-wl", argv[i], 2) ) wordless = true;
	  else if(! strncmp("-R", argv[i], 2) ) RFuzTarg = strtod(argv[i] + 2, NULL);
	  else if(! strncmp("-mR", argv[i], 3) ) maxRightFuzTarg = strtod(argv[i] + 3, NULL);
	  else if(! strncmp("-mL", argv[i], 3) ) minLeftFuzTarg = strtod(argv[i] + 3, NULL);
	  else if(! strncmp("-t", argv[i], 2) ) centroidFuzTarg = true;
	  else if(! strncmp("-ne0", argv[i], 4) ) 
	    {
	      neigh0= strtod(argv[i] + 4, NULL);
	      n0=true;
	    }
	  else if(! strncmp("-ne1", argv[i], 4) ) 
	    {
	      neigh1= strtod(argv[i] + 4, NULL);
	      n1=true;
	    }
	  else if(! strncmp("-r2", argv[i], 3) ) rsk2FuzTarg = true;
	  else if(! strncmp("-bi", argv[i], 3) ) combiFuzTarg = true;
	 
        }
    }


  if (!wordless)
    {
      for (i=0;i<argc;i++)
	printf("argument %d = %s  ",i,argv[i]);
    }
 
  
  
  try{
    
  
    // Build Fis
    FIS * S;
   
    int nbmax=0;
    
    
    int inums;
    char nums[100];
    double eps=1e-6;//default precision
    
    FILE *fg,*fs,*fgain,*final, *fperf, *fmed;
    S = new FIS(argv[1]);   
    nbmax=S->GetNbIn();
    // Generate config file 
    fg = fopen("initial.fis", "wt");
    S->PrintCfg(fg);
    fclose(fg);
    delete S;
    
    // default=optimize all rules
    if (!optimIn && !optimOut && !optimRules)
      optimRules=true;
    //

    if (optimIn)
      {
	numinputs=new double[nbmax];
    	iter=new char[20];
	//attention if length 1 string =pb in SearchNb
	int lenin=strlen(argv[argstr]);
	if (lenin==1)
	  {
	    nbinputs=1;
	    numinputs[0]=atol(argv[argstr]);
	  }
	else
	  nbinputs=SearchNb(argv[argstr],numinputs,nbmax,blank);
	
	if (!wordless)
	  printf("\n %d inputs found:",nbinputs);
	if (!wordless)
	  printf("\ninputs to optimize: %s \n",argv[argstr]);
	for (int kk=0;kk<nbinputs;kk++)
	  {
	    //printf("%d  ",(int)(1+numinputs[kk]));
	    if ((int)numinputs[kk]<1 ||(int)numinputs[kk]>nbmax)
	      {
		printf("\nerror in input string\n");
		return 1;
	      }
	  }
	if (!wordless)
	  printf("\n");
      }

    len=1000;
    itername=new char[len];

    len=strlen(argv[3]) + 100;
    outputFisConfigName=new char[len];
    outputGain=new char[len];

    //generate samples for training/test
    if (ns>0)
      {

	strcpy(NomFichierB,argv[2]);
	strcat(NomFichierB,".lrn.sample.");
	
	if (!wordless)
	  printf("%s %d %s %s %s\n","\nGenerating ",ns," samples with names ",NomFichierB,"n");
	// generate samples
	// ratio training/test
	if(rs>1.) rs=1.;
	if (ns < 2) ns=2;
	//if ((ns % 2) ==0) ns=ns/2;
	genNSamples(seed,rs,Flagcs,eps,nbmax,argv[2],ns);
      }
    if (ns >0)
      {
	isinit=0;
	strcpy(NomFichier,NomFichierB);

      }
    else
      strcpy(NomFichier,argv[2]);

    //strcpy(NomFichier,argv[2]);

    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // loop on learning samples instead of data file
    //loop on training samples-if no sample isample=-1 and ns=0
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    for (int isample=isinit;isample<ns;isample++)
      {
	// regenerate initial FIS
	S = new FIS(argv[1]);   
	// complete sample data file name
	if (isample >=0)
	  {
	    strcpy(NomFichier,NomFichierB);
	    sprintf(nums,"%d",isample);
	    strcat(NomFichier,nums);
	  }
	if (!wordless)
	  printf("\n------------\nData file name %s\n------------",NomFichier);
	// complete optimized FIS name
	strcpy(outputFisConfigName,argv[3]);
	if (isample >=0)
	  {
	    strcat(outputFisConfigName,"-lrn.sample");
	    strcat(outputFisConfigName,nums);
	  }
	strcat(outputFisConfigName,"-final.fis");
	final=fopen(outputFisConfigName,"wt");
	if (!wordless)
	  printf("\n------------\nOutput FIS file name %s\n------------",outputFisConfigName);
	if (intermfile)
	  {
	    strcpy(outputGain,outputFisConfigName);
	    strcat(outputGain,"-gain");
	    fgain=fopen(outputGain,"wt");
	    fprintf(fgain,"step %i & %s & %g & %g\n", 0,"",perfinit,0.0);
	  }

	//strcpy(outputFisConfigName,argv[3]);
    
	if (optimIn)
	  {
	    strcat(outputFisConfigName,"-input");
	    strcpy(itername,"");
	  }
    
   
    
    
	// initialise Samples
#ifdef DEBUG
	printf("\n\t* acquisition du pointer sur exemple et classif;\n\n");
#endif
	//  nbrow=0.2S->InitResClassifSamples( &Exemples , &ResClassif , NomFichier , nOut);
	//  It should be a FIS function ...

	// read data for current training sample
	Exemples = ReadSampleFile( NomFichier , nbcol, nbrow);
	if(!*Exemples) 
	  {
#ifdef DEBUG
	    printf("\nThe sample matrix cannot be built ! \n");
#endif
	    return 1 ;
	  }
	//test on column and row number
	if(nbcol < S->GetNbIn()) 
	  {
#ifdef DEBUG
	    printf("\n The number of input variables is smaller than the number of inputs in the FIS.\n");
#endif
	    return  -2;
	  }

	if(nbcol < S->GetNbIn() + S->GetNbOut())
	  {
#ifdef DEBUG
	    printf("\n The number of variables is smaller than the number of inputs plus the number of outputs.");
#endif
	  }  
	// call to Classcheck once to allocate ResClassif and updates Lab pointer
	reponse=S->ClassCheck(ResClassif,Lab,Exemples,nbrow,nOut );
#ifdef DEBUG
	printf("\n The Class Check returns %i",reponse);
#endif
    
	if(FuzRefG)
	  {
	    int outputNumber;
	    outputNumber=nOut;

	    if(strcmp(S->Out[outputNumber]->GetOutputType(), OUT_CRISP::OutputType()))
	      {
		FuzRefG = false;
		throw std::runtime_error("~FuzzTargetOnlyForCrispOutput~");
	      }
	    else
	      {
		double ** Data = NULL;
		int nbcol=0, nbex=0;
		double *VPos = NULL;
		int *pPos=NULL;
	   
		Data = ReadSampleFile(NomFichier, nbcol, nbex);
		if(nbcol < S->GetNbIn())  return -2;
		if(nbcol < S->GetNbIn() + 1 + outputNumber) 
		  throw std::runtime_error("~NotEnoughColumnsInDataFile~");
		double *tmp;
		tmp = new double [nbex];
		GetColumn(Data, nbex, S->GetNbIn()+outputNumber, tmp);
		SortUniq(tmp, nbex, &VPos, &pPos, NbFuzTargG, EPSILON);
	 
		FuzTargG = new MF * [NbFuzTargG];	   
		for(i = 0; i < NbFuzTargG; i++)
		  FuzTargG[i] = FuzNumber(VPos[i], kw, sw);
		delete [] tmp;
		delete [] VPos;
		delete [] pPos;
		for (i = 0; i < nbex; i++) delete [] Data[i];
		delete [] Data;
	      }

	    if(!wordless) 
	      printf("NbFuzTargG= %d",NbFuzTargG);
	    if(!wordless) 
	      for(i = 0; i < NbFuzTargG; i++)
		{
		  FuzTargG[i]->Print(); 
		  printf("\t kern:  %f", FuzTargG[i]->Kernel(kw,sw));
		}
	    //normalize fuzzy numbers/output range attention voir fuzrefgfull !
	    min1=S->Out[nOut]->min();
	    max1=S->Out[nOut]->max();
	    
	    if (!wordless)
	      printf("\nafter normalization with output range min=%f, max=%f",min1,max1);
	    for(i = 0; i < NbFuzTargG; i++)
	      {
		FuzTargG[i]->Normalize(min1,max1); 
		FuzTargG[i]->Print(); 
	      }
	  }// end if(FuzRefG)

	if(FuzRefGfull)
	  {
	    int outputNumber;
	    outputNumber=nOut;
	    if(strcmp(S->Out[outputNumber]->GetOutputType(), OUT_CRISP::OutputType()))
	      {
		FuzRefGfull = false;
		throw std::runtime_error("~FuzzTargetOnlyForCrispOutput~");
	      }
	  }
   
    
    
   
	if(FuzRefGfull)
	  {
	    FuzTargG=CreateFuzTarg(NomFichier, kw,sw,NbFuzTargG, VPos);// create fuzzy targets from last column of data file 
	    // distinct data values returned in VPos array

	    if(!wordless) 
	      printf("\n---------NbFuzTargG= %d-------------",NbFuzTargG);
	    if(!wordless) 
	      for(i = 0; i < NbFuzTargG; i++)
		{
		  FuzTargG[i]->Print(); 
		  printf("\t kernel center:  %f", FuzTargG[i]->Kernel(kw1,sw1));
		}

	    // use global arrays for easy access from within perf
	    // precompute neighFuzTarg array
	    neighFuzTarg=NeighborsTargG(FuzTargG,  NbFuzTargG);// automatically determine neighboring bounds
	    // if not default values read neighbors as parameters
	    if (n0)
	      neighFuzTarg[0]=neigh0;
	    if (n1)
	      neighFuzTarg[1]=neigh1;
	    //pre compute all global arrays rskFuzTarg, deltaSFuzTarg, deltaKFuzTarg, rightjm1FuzTarg, leftjFuzTarg, set global RFuzTarg
	    precomprsk(RFuzTarg, maxRightFuzTarg,  minLeftFuzTarg,  neighFuzTarg,rskFuzTarg, deltaSFuzTarg, deltaKFuzTarg, rightjm1FuzTarg, leftjFuzTarg,rsk2FuzTarg);
	    // precompute global intersec arrays
	    intersecFuzTarg(mucombiFuzTarg,muinterFuzTarg,xinterFuzTarg);

	    // print check
	    if(!wordless) 
	      {
		printFuzTargArrays();
		// print symbolic distances
		printf("\nSymbolic distances\n");
		for (int kk=-2;kk<NbFuzTargG+1;kk++)
		  printf("ds(%d,%d)=%f\n",kk,kk+1,MFdataDistCalc(kk,kk+1,minLeftFuzTarg,maxRightFuzTarg,neighFuzTarg));

	      }
	  }//end if fuzrefgfull

	// Added by SG - March 16th, 2010
	// modif for fuzzy reference
   
	perfinit=S->Performance(nOut, NomFichier, Couverture, MaxError, museuil, NULL, 0);
	if (!wordless)
	  printf("\nperfinit (before normalization)=%f\n",perfinit);
    
      	//normalize fuzzy numbers/output range and all related arrays
	min1=S->Out[nOut]->min();
	max1=S->Out[nOut]->max();
	    
	if (!wordless)
	  printf("\nafter normalization with output range min=%f, max=%f",min1,max1);
	
	if (FuzRefGfull)
	  {
	    NormalizeFuzTarg(min1,max1);
	    if(!wordless) 
	      {
		printf("\nafter normalization with output range min=%f, max=%f",min1,max1);
		for(i = 0; i < NbFuzTargG; i++)
		  {
		    FuzTargG[i]->Print(); 
		    printf("\t kern:  %f", FuzTargG[i]->Kernel(kw,sw));
		  }
		printFuzTargArrays();
	      }
	  }
	

	InitialCoverage=Couverture;
    
	// .... 
	//beginning of sifopt
	sifopt *Opt=NULL;
 
	if (!wordless)
	  printf("\n%d loop iterations for optimizing",nsteps);
   
	for (int step=0;step<nsteps;step++)
	  {
	    if (!wordless)
	      printf("\n-------------------------\nIteration %d, sample %d\n------------------------",step+1,isample+1);
	    if (optimIn)
	      {
		if (!wordless)
		  printf("\n--------------------\noptimizing inputs\n--------------------\n");
	
	  
		strcpy(itername,"");
		for (int kk=0;kk<nbinputs;kk++)
		  {
		    last=(kk==nbinputs-1);
		    //need this to do a clean copy
		    //create sifopt using updated FIS - FIS is normalized and data also
		    Opt= new sifopt(S,ResClassif,Exemples,nbrow,nOut);
	      
		    // build key corresponding to input to optimize
		    numInput=(int)numinputs[kk];
		    KeyIn=Opt->keysetIN( S , numInput-1 , 1 );
		    Opt->setKeyInit(KeyIn);
		    if (!wordless)
		      printf("\noptimizing input %d for FIS\n",numInput);
		    if (!wordless)
		      printf("Key:  ");
		    if (!wordless)
		      KeyIn->Print();

		    sprintf(iter,"-%d",numInput);
		    strcat(outputFisConfigName,iter);
		    if (kk>0)
		      sprintf(iter,"-%d",numInput);
		    else
		      sprintf(iter,"%d",numInput);
		    strcat(itername,iter);
		    // update FISOPT
		    // sets CovLoss and InitCov in FISOPT object
		    Opt->Update(ResClassif,Exemples,nOut,nbrow,cnear,InitialCoverage,cov);
		    //update P = algorithm parameters
		    Opt->P->UpdatePar(atoi(argv[4]),(double)atof(argv[5]),atoi(argv[6]),atoi(argv[7]),atof(argv[8]),l1,l2,l3);
		    // update parameters for evaluation function
		    Opt->UpdatefEvalArgs(Exemples,nbrow,Couverture,MaxError,ResClassif,museuil, Lab);

		    //launch optimizer : optimize entries or/and conclusion depending on key
		    //uses parameters defined above
		    reponse=Opt->launchEntry(gain, wordless);//starts by normalizing data !!!

		    if (fisout!=NULL)
		      delete [] fisout;
		    fisout=new char[strlen(outputFisConfigName)+5];
		    strcpy(fisout,outputFisConfigName);
		    strcat(fisout,".fis");
      
		    if (last & intermfile)
		      {
			fs = fopen(fisout, "wt");
		      }
		    if   ( ( Opt->Reponse == 1 ) || ( Opt->Reponse == 2 ) || ( Opt->Reponse == 3 )) 
		      {
			if (last & intermfile)
			  Opt->Solution->PrintCfg( fs );
		
			//update FIS for next iteration
			if (S != NULL) delete S;
		   
			S=new FIS(*(Opt->Solution));
			//S->ClassCheck(ResClassif,Lab,Exemples,nbrow,nOut );
			//Opt->Solution->In[numInput-1]->Print();
			if (FuzRefGfull)
			  UnNormalizeFuzTarg(min1,max1);
			perf=S->Performance(nOut, argv[2], Couverture, MaxError, museuil, NULL, 0);
			if (FuzRefGfull)
			  NormalizeFuzTarg(min1,max1);
			if (intermfile)
			  fprintf(fgain,"step%i & %s & %g & %g\n", 1+step,itername,perf,(perfinit-perf)/perfinit);
			if (!wordless)
			  printf("\n+ Rel. gain: %f",(perfinit-perf)/perfinit);
		      }
		    else
		      {
			if (!wordless)
			  printf("\nno better FIS\n");
			if (last & intermfile)
			  S->PrintCfg( fs );
			//warning: fuzzy data must be unnormalized before archiving performance
			if (FuzRefGfull)
			  UnNormalizeFuzTarg(min1,max1);
			perf=S->Performance(nOut, argv[2], Couverture, MaxError, museuil, NULL, 0);
			if (FuzRefGfull)
			  NormalizeFuzTarg(min1,max1);
			if (intermfile)
			  fprintf(fgain,"step%i & %s & %g & %g\n", 1+step,itername,perf,(perfinit-perf)/perfinit);
			if (!wordless)
			  printf("\n+ Rel. gain: %f",(perfinit-perf)/perfinit);
		  
		      }
		    if (last & intermfile )
		      fclose( fs );
	   
		    //option DEBUGCSV used in makefile only

#ifdef DEBUGCSV
		    Opt->P->PrintCSV(); Opt= new sifopt(S,ResClassif,Exemples,nbrow,nOut);
		    printf("\n");
#endif
		    if ((step==(nsteps-1)) && (kk==(nbinputs-1)))
		      { 
			if (!wordless)
			  Opt->P->Print(); 
			if (!wordless)
			  printf(" InitialCoverage=%f, Allowed Cov. Loss=%f",Opt->InitCov,Opt->CovLoss);
		      }
     

		    //delete FISOPT and KeyIn objects
		    if (Opt != NULL) delete Opt;
		    if (KeyIn != NULL) delete KeyIn;
		    // necessary because of exemple normalization...
		    if (Exemples !=NULL) 
		      {
			for( int i=0 ; i<nbrow ; i++ )
			  delete [] Exemples[i];
			delete [] Exemples;
			Exemples=NULL;
		      }
		    Exemples = ReadSampleFile( NomFichier , nbcol, nbrow);
	    
		  }//end loop on inputs to optimize
      
	      }// end of input optimization

	    if (optimOut)
	      {
    
		if (!wordless)
		  printf("\n\n------------\noptimizing fuzzy output\n------------\n");
		Opt= new sifopt(S,ResClassif,Exemples,nbrow,nOut);//Exemples was read again at the end of the previous loop
		KeyOut=Opt->keysetOUT(S , nOut, 1);//-1 stands for PFF, nout=output number (0)
		if (!wordless)
		  printf("Key:  ");
		if (!wordless)
		  KeyOut->Print();
		Opt->setKeyInit(KeyOut);
      
		// sets CovLoss and InitCov in FISOPT object
		Opt->Update(ResClassif,Exemples,nOut,nbrow,cnear,InitialCoverage,cov);
		//update P = algorithm parameters
		Opt->P->UpdatePar(atoi(argv[4]),(double)atof(argv[5]),atoi(argv[6]),atoi(argv[7]),atof(argv[8]),l1,l2,l3);
		// update parameters for evaluation function
		Opt->UpdatefEvalArgs(Exemples,nbrow,Couverture,MaxError,ResClassif,museuil, Lab);

		//launch optimizer : optimize entries or/and conclusion depending on key
		//uses parameters defined above
		reponse=Opt->launchEntry(gain,wordless);//starts by normalizing data !!!
		if (fisout!=NULL)
		  delete [] fisout;
		fisout=new char[strlen(outputFisConfigName)+25];
		strcpy(fisout,outputFisConfigName);
		strcat(fisout,"-fuzzyoutput.fis");
		if (intermfile)
		  fs = fopen(fisout, "wt");
		strcpy(itername,"fuzzyoutput");
		if   ( ( Opt->Reponse == 1 ) || ( Opt->Reponse == 2 ) || ( Opt->Reponse == 3 )) 
		  {
		    if (intermfile)
		      Opt->Solution->PrintCfg( fs );
		    //Opt->Solution->PrintCfg(final);
		    //Opt->Solution->PrintCfg(final);
		    //update FIS for next iteration
		    if (S != NULL) delete S;
	
		    S=new FIS(*(Opt->Solution));
		    S->ClassCheck(ResClassif,Lab,Exemples,nbrow,nOut );
		    if (FuzRefGfull)
		      UnNormalizeFuzTarg(min1,max1);
		    perf=Opt->Solution->Performance(nOut, argv[2], Couverture, MaxError, museuil, NULL, 0);
		    if (FuzRefGfull)
		      NormalizeFuzTarg(min1,max1);
		    if (intermfile)
		      fprintf(fgain,"iter & %s & %g & %g\n", "",itername,perf,(perfinit-perf)/perfinit);
		    if (!wordless)
		      printf("\n+ Rel. gain: %f",(perfinit-perf)/perfinit);
		  }
		else
		  {
		    if (!wordless)
		      printf("\nno better FIS\n");
		    if (intermfile)
		      S->PrintCfg( fs );
		    //S->PrintCfg(final);
		    if (FuzRefGfull)
		      UnNormalizeFuzTarg(min1,max1);
		    perf=S->Performance(nOut, argv[2], Couverture, MaxError, museuil, NULL, 0);
		    if (FuzRefGfull)
		      NormalizeFuzTarg(min1,max1);
		    if (intermfile)
		      fprintf(fgain,"iter & %s & %g & %g\n",itername,perf,(perfinit-perf)/perfinit);
		    if (!wordless)
		      printf("\n+ Rel. gain: %f",(perfinit-perf)/perfinit);
		  
		  }
		//delete FISOPT and KeyOut objects
		if (Opt != NULL) 
		  delete Opt;
		if (KeyOut != NULL) 
		  delete KeyOut;
		if (intermfile)
		  fclose(fs);
		// necessary because of exemple normalization...
		if (Exemples !=NULL) 
		  {
		    for( int i=0 ; i<nbrow ; i++ )
		      delete [] Exemples[i];
		    delete [] Exemples;
		    Exemples=NULL;
		  }
		Exemples = ReadSampleFile( NomFichier , nbcol, nbrow);
	      } 
  
	    // now optimize all rules if asked for - 1 step only
  
	    if (optimRules)
	      {
     

		if (!wordless)
		  printf("\n\n---------------------------\noptimizing all rules\n---------------------------\n");
		Opt= new sifopt(S,ResClassif,Exemples,nbrow,nOut);//Exemples was read again at the end of the previous loop
		KeyRules=Opt->keysetRULE(S , nOut, -1);//-1 stands for all rules
		Opt->setKeyInit(KeyRules);

		if (!wordless)
		  printf("Key:  ");
		if (!wordless)
		  KeyRules->Print();
		if (!wordless)
		  printf("\n");
      
		// sets CovLoss and InitCov in FISOPT object
		Opt->Update(ResClassif,Exemples,nOut,nbrow,cnear,InitialCoverage,cov);
		//update P = algorithm parameters
		Opt->P->UpdatePar(atoi(argv[4]),(double)atof(argv[5]),atoi(argv[6]),atoi(argv[7]),atof(argv[8]),l1,l2,l3);
		// update parameters for evaluation function
		Opt->UpdatefEvalArgs(Exemples,nbrow,Couverture,MaxError,ResClassif,museuil, Lab);
		perf=S->Performance(nOut, argv[2], Couverture, MaxError, museuil, NULL, 0);
		if (!wordless)
		  printf("in optimrules before step%i perfinit(before norm)=%g perf0(after norm)=%g\n", 1+step,perfinit,perf);
		//launch optimizer : optimize entries or/and conclusion depending on key
		//uses parameters defined above
		reponse=Opt->launchEntry(gain,wordless);//starts by normalizing data !!!
		if (!wordless)
		  printf("\nreponse=%d",reponse);
		if (fisout!=NULL)
		  delete [] fisout;
		fisout=new char[strlen(outputFisConfigName)+15];
		strcpy(fisout,outputFisConfigName);
		strcat(fisout,"-allrules.fis");
		if (intermfile)
		  fs = fopen(fisout, "wt");
      
		strcpy(itername,"all rules");
		if   ( ( Opt->Reponse == 1 ) || ( Opt->Reponse == 2 ) || ( Opt->Reponse == 3 )) 
		  {
		    if (intermfile)
		      Opt->Solution->PrintCfg( fs );
		    //Opt->Solution->PrintCfg(final);
		    //update FIS for next iteration
		    if (S != NULL) delete S;
	
		    S=new FIS(*(Opt->Solution));//S is unnormalized
		    S->ClassCheck(ResClassif,Lab,Exemples,nbrow,nOut );
		    // UnNormalize FuzTargets - required for performance calculation
		    if (FuzRefGfull)
		      UnNormalizeFuzTarg(min1,max1);
		    perf=Opt->Solution->Performance(nOut, argv[2], Couverture, MaxError, museuil, NULL, 0);
		    // renormalize FuzTargets for next step
		    if (FuzRefGfull)
		      NormalizeFuzTarg(min1,max1);
		    if (intermfile)
		      fprintf(fgain,"rules & %s & %g & %g\n", "",itername,perf,(perfinit-perf)/perfinit);
		    if (!wordless)
		      printf("\n+ Rel. gain: %f",(perfinit-perf)/perfinit);
		  }
		else
		  {
		    if (!wordless)
		      printf("\nno better FIS\n");
		    if (intermfile)
		      S->PrintCfg( fs );
		    //S->PrintCfg(final);
		    if (FuzRefGfull)
		      UnNormalizeFuzTarg(min1,max1);
		    perf=S->Performance(nOut, NomFichier, Couverture, MaxError, museuil, NULL, 0);
		    if (FuzRefGfull)
		      NormalizeFuzTarg(min1,max1);
		    if (intermfile)
		      fprintf(fgain,"rules & %s & %g & %g\n",itername,perf,(perfinit-perf)/perfinit);
		    if (!wordless)
		      printf("\n+ Rel. gain: %f",(perfinit-perf)/perfinit);
		  
		  }
		if (!wordless)
		  printf("\nin optimrules after step%i perfinit=%g perf=%g & gain rel=%g\n", 1+step,perfinit,perf,(perfinit-perf)/perfinit);
		//delete FISOPT and KeyIn objects
		if (Opt != NULL) 
		  delete Opt;
		if (KeyRules != NULL) 
		  delete KeyRules;
		if (intermfile)
		  fclose( fs );
		// necessary because of exemple normalization...
		if (Exemples !=NULL) 
		  {
		    for( int i=0 ; i<nbrow ; i++ )
		      delete [] Exemples[i];
		    delete [] Exemples;
		    Exemples=NULL;
		  }
		Exemples = ReadSampleFile( NomFichier , nbcol, nbrow);
	      }
   
	  }//end loop on steps

	// final output file
	S->PrintCfg(final);
 
	if (S != NULL) delete S;

	// delete data array
	if (Exemples !=NULL) {
	  for( int i=0 ; i<nbrow ; i++ )
	    delete [] Exemples[i];
	  delete [] Exemples;
	  Exemples=NULL;
	}
	// delete resclassif array
	if (ResClassif != NULL) delete [] ResClassif;
	// delete temp arrays
   
	if (fisout!=NULL) 
	  {
	    delete [] fisout;
	    fisout=NULL;
	  }
	if (intermfile)
	  fclose(fgain);
	fclose(final);
	// build for each sample
	// perf file: initial perf on whole sample, training, initial perf on test
	//            optimized perf on whole sample, on training, optimized perf on test
      }// end loop on samples
    // if samples exist, build medfis
    if (ns >0)
      {
	if (!wordless)
	  cout <<"building med.fis"<<endl;
	M=medfis("optim-lrn.sample","-final.fis",ns);
	fmed=fopen("med.fis","wt");
	M->PrintCfg(fmed);
	fclose(fmed);
      }
   
    //
    //generate perf array file
    //  3 columns: initial FIS, test-sample opt FIS, median FIS perf
    // 1 row/sample
    fperf=fopen("perf.res", "wt");
    if (ns>0)
      fprintf(fperf,"Dataset & initial FIS & sample k-optimized FIS & Median FIS & Rel.gain\n");
    else
      fprintf(fperf,"Dataset & initial FIS & Optimized FIS & Rel.gain\n");
    // perf on whole data set
    S=new FIS(argv[1]);
    strcpy(NomFichier,argv[2]);
    perfall=S->Performance(nOut, NomFichier, Couverture, MaxError, museuil, NULL, 0);
    delete S;
    if (ns>0)
      {
	perfallmed=M->Performance(nOut, NomFichier, Couverture, MaxError, museuil, NULL, 0);
	if (perfall>0)
	  ga=(perfall-perfallmed)/perfall;
	else
	  ga=FisMknan();
	if (!wordless)
	  printf("\n FIS=%s Dataset=%s Perf init=%g Perf med=%g",argv[1],NomFichier,perfall,perfallmed,ga);
	fprintf(fperf,"%s & %g & %g & %g & %g\n",NomFichier,perfall,FisMknan(),perfallmed,ga);
      }
    else
      {
	//build FIS from saved file
	strcpy(outputFisConfigName,argv[3]);
	strcat(outputFisConfigName,"-final.fis");
	S=new FIS(outputFisConfigName);	
	perfopt=S->Performance(nOut, NomFichier, Couverture, MaxError, museuil, NULL, 0);
	delete S;
	if (perfall>0)
	  ga=(perfall-perfopt)/perfall;
	else
	  ga=FisMknan();
	if (!wordless)
	  printf("\n FIS=%s Dataset=%s Perf init=%g Perf opt=%g Rel. gain=%g",argv[1],NomFichier,perfall,perfopt,ga);
	fprintf(fperf,"%s & %g & %g & %g\n",NomFichier,perfall,perfopt,ga);
      }
    
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // loop on test samples plus learning samples, instead of data file
    //if no sample isample=-1 and ns=0
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if (ns>0)
      {
	strcpy(NomFichierB,argv[2]);
	if (!wordless)
	  printf("\nSample perf loop\n");
	//first loop on test samples, second loop on lrn samples
	int kk=0;
	for (kk=0;kk<2;kk++)
	  {
	    if (kk==0)
	      strcat(NomFichierB,".tst.sample.");
	    else
	      {
		strcpy(NomFichierB,argv[2]);
		strcat(NomFichierB,".lrn.sample.");
	      }
	      
	    for (int isample=0;isample<ns;isample++)
	      {
	    

		//evaluate perf with appropriate data file
		strcpy(NomFichier,NomFichierB);
		sprintf(nums,"%d",isample);
		strcat(NomFichier,nums);
		// build initial FIS
		S=new FIS(argv[1]);
		perf0=S->Performance(nOut,NomFichier, Couverture, MaxError, museuil, NULL, 0);
		if (!wordless)
		  printf("\n FIS=%s Dataset=%s Perfinit=%g",argv[1],NomFichier,perf0);
		delete S;
		perfmed=M->Performance(nOut,NomFichier, Couverture, MaxError, museuil, NULL, 0);
		//build FIS from saved file
		strcpy(outputFisConfigName,argv[3]);
		sprintf(nums,"%d",isample);
		strcat(outputFisConfigName,"-lrn.sample");
		strcat(outputFisConfigName,nums);
	  
		strcat(outputFisConfigName,"-final.fis");
		S=new FIS(outputFisConfigName);			
	
		perfopt=S->Performance(nOut, NomFichier, Couverture, MaxError, museuil, NULL, 0);
	  	   
		if (perf0>0)
		  ga=(perf0-perfmed)/perf0;
		else
		  ga=FisMknan();
	   
		if (!wordless)
		  printf("\n FIS=%s Dataset=%s Perfopt=%g Perfmed=%g Rel.gain=%g\n",outputFisConfigName,NomFichier,perfopt,perfmed,ga);
		    
		fprintf(fperf,"%s & %g & %g & %g & %g\n",NomFichier,perf0,perfopt,perfmed,ga);
    
		delete S;

	      }// isample loop
	  }// for loop
      }// ns>0 loop
    fclose(fperf);
    
    if (M!=NULL)
      delete M;

    // delete global arrays
    if (outputFisConfigName!= NULL) delete [] outputFisConfigName;
    if (outputGain!= NULL) delete [] outputGain;
    if (numinputs != NULL) delete [] numinputs;
    if (itername != NULL) delete [] itername;
    if (iter!= NULL) delete [] iter;
  
    if (!wordless)
      printf("\n***********************\nFIS opt END\n");
    
    return reponse;  
  }
  catch(std::exception &e){
    printf("error in loopoptim\n");
    printf("catch %s\n",e.what());
    return -1;
  }

}
