#include "common.h"
#include "fis.h"
#include "sifopt.h"
#include "gensample.h"


//*********************************************************
FIS* medfis(const char* prefix,const char* suffix,int ns, bool wordless)
//********************************************************
{
  FIS *fis =NULL, *tmpfis=NULL;
  int ninputs=0,i,j,k,len,is,nbr;
  int * nmf=NULL;
  int ** nbparams=NULL;
  char * FisConfigName=NULL;
  char * snum=NULL;
  double ****par=NULL;
  double firstq, median, thirdq,min1,max1;
  double * tab=NULL;
  double ** conclu=NULL;
  MF * mf=NULL;
  
  if (!wordless)
    cout <<"\nBuilding med.fis"<<endl;

  len=strlen(prefix)+strlen(suffix)+6;
  FisConfigName=new char[len];
  snum=new char[5];
  //first FIS=model
  strcpy(FisConfigName,prefix);
  sprintf(snum,"%i",0);
  strcat(FisConfigName,snum);
  strcat(FisConfigName,suffix);
  if (!wordless)
    printf("FisConfigName=%s\n",FisConfigName);
  fis=new FIS(FisConfigName);
  //fis->PrintCfg();
  ninputs=fis->GetNbIn();

  nmf=new int[ninputs];
  nbparams=new int*[ninputs];
   
  par=new double ***[ns];
  conclu=new double *[ns];
  for (is=0;is<ns;is++)
    par[is]=new double **[ninputs]; 

  
  //examine each input for sample 0
  for (i=0;i<ninputs;i++)
    {
      // for each MF, get MF number, type and bounds
      nmf[i]=fis->In[i]->GetNbMf();
      nbparams[i]=new int[nmf[i]];
      par[0][i]=new double *[nmf[i]];
      for (j=0;j<nmf[i];j++)
	{
	  mf=fis->In[i]->GetMF(j);
	  nbparams[i][j]=mf->NbParams();
	  par[0][i][j]=new double[nbparams[i][j]];
	  mf->GetParams(par[0][i][j]);
	}
    }
  nbr=fis->GetNbRule();
  conclu[0]=new double[nbr];
  // rule conclusions
  for (i=0;i<nbr;i++)
    conclu[0][i]=fis->Rule[i]->GetAConc(0);
  //other samples
  for (is=1;is<ns;is++)
    {
      strcpy(FisConfigName,prefix);
      sprintf(snum,"%i",is);
      strcat(FisConfigName,snum);
      strcat(FisConfigName,suffix);
      tmpfis=NULL;
      tmpfis=new FIS(FisConfigName);
    
      //examine each input
      for (i=0;i<ninputs;i++)
	{
	  // for each MF, get MF number, type and bounds
	  nmf[i]=tmpfis->In[i]->GetNbMf();
	  par[is][i]=new double *[nmf[i]];
	  for (j=0;j<nmf[i];j++)
	    {
	      mf=tmpfis->In[i]->GetMF(j);
	      nbparams[i][j]=mf->NbParams();
	      par[is][i][j]=new double[nbparams[i][j]];
	      mf->GetParams(par[is][i][j]);
	    }
	}
    
      // rule conclusions
      conclu[is]=new double[nbr];
      for (i=0;i<nbr;i++)
	conclu[is][i]=tmpfis->Rule[i]->GetAConc(0);
      delete tmpfis;
    }
   
  //average each parameter
  for (i=0;i<ninputs;i++)
    {
      /*if (!wordless)
	cout <<endl<<"i="<<i<<" nmf="<<nmf[i];
      */
      for (j=0;j<nmf[i];j++)
	{
	  //cout <<endl<<"mf="<<j<<" nbparams="<<nbparams[i][j];
	  for (k=0;k<nbparams[i][j];k++)
	    {
	      tab=new double[ns];
	      for (is=0;is<ns;is++)
		tab[is]=par[is][i][j][k];
	      StatArrayQuart(tab, ns, 0, firstq, median, thirdq,min1,max1);
	      delete[] tab;
	      // update par
	      par[0][i][j][k]=median;
	      //cout <<endl<<"k="<<k<<" median. param="<<median<<" par="<<par[0][i][j][k];
	      //cout<<" firstq="<<firstq<<" thirdq="<<thirdq<<endl;
	      // update MF parameters
	      fis->In[i]->GetMF(j)->Update(par[0][i][j]);

	    }
	}
    }
  for (i=0;i<nbr;i++)
    {
      tab=new double[ns];
      for (is=0;is<ns;is++)
	tab[is]=conclu[is][i];
      StatArrayQuart(tab, ns, 0, firstq, median, thirdq,min1,max1);
      delete [] tab;
      
      // update rule conclusions
      fis->Rule[i]->SetAConc(0,median);
    }
  //delete objects
  delete [] FisConfigName;
  
  for (is=0;is<ns;is++)
    delete [] conclu[is];
  delete [] conclu;

  delete [] snum;
  

  for (i=0;i<ninputs;i++)
    delete [] nbparams[i];
  delete [] nbparams;
  
  
  for (is=0;is<ns;is++)
    {
      for (i=0;i<ninputs;i++)
	{
	  for (j=0;j<nmf[i];j++)
	    delete [] par[is][i][j];
	  delete [] par[is][i];
	}
      delete [] par[is];
    }
      
  delete [] par;

  delete [] nmf;
  
  

  return fis;
}
//******************************
int PrintArgsLoopOS(int argc, char ** argv)
//******************************
{
  if(argc < 3)
    {
      printf("\nThe programm %s needs 2 arguments : ", argv[0]);
      printf("\n\t* the name of the FIS configuration file;");   
      printf("\n\t* the name of the data file used for tuning;"); 
      printf("\n\t* default only optimize rule conclusions if no other option is given");
      printf("\n Options:");
      printf("\n\t* -b the basename of the resulting FIS configuration file;(default optim)");   
      printf("\n\t* -it the number of iterations; (default 100)");   
      printf("\n\t* -e the range of the gaussian noise; (default 0.005)");
      printf("\n\t* -v the max number of constraints violations before it counts as an iteration (default 1000);");
      printf("\n\t* -f the max number of failure steps in an algorithm (default 1000);");
      printf("\n\t* -d the equality center distance threshold (default 1e-6);");
     
      //sampling args
      printf("\n\t* -ns Number of sample pairs to generate from data file (default 10-put -ns 0 for no sampling);");
      printf("\n\t* -cs Draw samples with respect to class ratio in data file (class is last column, default no));");
      printf("\n\t* -rs Ratio learning/all pairs (default 0.75, maximum 0.9);");
      printf("\n\t* -s   integer to set seed value for parker miller random;(default 0:new sampling each time);");//end sampling args

      printf("\n\t* -in \"x y \", the string of input numbers to optimize (starting at 1, order is important, default no input optimization);"); 
      printf("\n\t* -r optimize  rule conclusions;");
      printf("\n\t* -o optimize fuzzy output (default false)");

      printf("\n\t* -n   Output number to consider (default 0: first output);");
      printf("\n\t* -m   minimum membership (default 0.1);");
  
      printf("\n\t* -l1  Solis Wetts Constant 1 (default 0.4);");
      printf("\n\t* -l2  Solis Wetts Constant 2 (default 0.2);");
      printf("\n\t* -l3  Solis Wetts Constant 3 (default 0.5);");
      printf("\n\t* -u   Number of loops for optimizing (default 2);");
      printf("\n\t* -c relative tolerated loss of coverage (default 0.10 ; 1.=10.0%%).");
      printf("\n\t* -nc minimum distance between MF centers (default=1e-3)");
      printf("\n\t* -g create intermediate files (default false);");
      printf("\n\t* -a for intermediate display (default false);" ); 
      printf("\n\t* -wl for wordless.(default not silent)\n");
  
      return 1;
    
    }
  return 0;
}

//*************************************************************************
int ReadInputsOpt(const char * stri, int nbmax,double *numinputs, bool wordless)
//*************************************************************************
{

  int nbinputs=0;
  char blank=' ';

  //attention if length 1 string =pb in SearchNb
  int lenin=strlen(stri);
  if (lenin==1)
    {
      nbinputs=1;
      numinputs[0]=atol(stri);
    }
  else
    nbinputs=SearchNb(stri,numinputs,nbmax,blank);
	
  if (!wordless)
    printf("\n %d inputs found:",nbinputs);
  if (!wordless)
    printf("\ninputs to optimize: %s \n",stri);
  for (int kk=0;kk<nbinputs;kk++)
    {
      if ((int)numinputs[kk]<1 ||(int)numinputs[kk]>nbmax)
	{
	  printf("\nerror in input string\n");
	  return 1;
	}
    }
  return nbinputs;
}

//*************************************************************************************************************
int CallOptim(sifopt *&Opt,FIS * &S, vkey * KeyIn,double ** Exemples, char * NomFichier, int nOut, int &nbcol, int &nbrow, double perfinit, double muthresh, bool fisoutflag, char * outputFisConfigName,  bool wordless)
//*************************************************************************************************************
{
  int reponse=0;
  char * fisout=NULL;
  FILE * fs=NULL;
  double inCov, MaxError;
  double perfnorm,perf;
 
  //launch optimizer : optimize entries or/and conclusion depending on key
  //uses parameters defined above
  // reponse=0: no better FIS, 
  //1: better FIS after input optim,
  //2: better FIS after fuzzy output optim
  //3: better FIS after rule optim
  //***************************************
  reponse=Opt->launchEntry(perfnorm, wordless);
  //***************************************
  if (!wordless)
    printf("In CallOptim reponse=%d, perf(norm)=%f",reponse,perfnorm);
  if ((Opt->Solution)==NULL) reponse=0;
  if (perfnorm <=0) reponse=0;
  
  if (fisoutflag==true)
    {
      fisout=new char[strlen(outputFisConfigName)+strlen(".fis")+1];
      strcpy(fisout,outputFisConfigName);
      strcat(fisout,".fis");
      fs = fopen(fisout, "wt");
    }
  

  if   ( reponse !=0) 
    {
      if (fisoutflag)
	Opt->Solution->PrintCfg( fs );
      //update FIS for next iteration and compute performance
      if (S != NULL) 
	{
	  delete S;
	  S=NULL;
	}
      S=new FIS(*(Opt->Solution));
      perf=S->Performance(nOut, NomFichier, inCov, MaxError, muthresh, NULL, 0);
      if (!wordless)
	printf("\n+ Rel. gain: %f",(perfinit-perf)/perfinit);
    }
  else
    {
      if (fisoutflag)
	S->PrintCfg( fs );
      if (!wordless)
	printf("\nno better FIS");
      // same S as initial FIS
      perf=S->Performance(nOut, NomFichier, inCov, MaxError, muthresh, NULL, 0);
      if (!wordless)
	printf("\n+ Rel. gain: %f",(perfinit-perf)/perfinit);
		  
    }
  if (fisoutflag)
    fclose( fs );

  if (Opt != NULL) 
    {
      delete Opt;
      Opt=NULL;
    }
  if (KeyIn != NULL) delete KeyIn;
  
  if (fisout!=NULL)
    delete [] fisout;
  
  return reponse;
 
}
//*************************************************************************
sifopt* PrepareOpt(FIS * S, int *&ResClassif, double *&Lab,double **&Exemples, int &nbrow, int &nOut, int &cnear,double nearC, double &InitialCoverage, double &covloss,double &MaxError, double &muthresh,parstruct param )
//*************************************************************************
{
  sifopt* Opt=NULL;

  Opt= new sifopt(S,ResClassif,Exemples,nbrow,nOut,nearC);
  // update FISOPT
  // sets CovLoss and InitCov in FISOPT object
  Opt->Update(ResClassif,Exemples,nOut,nbrow,cnear,InitialCoverage,covloss);// diff avec inCov
  Opt->P->UpdatePar(param.maxite,param.gauss,param.maxcons,param.maxfail, param.epsiloncenter,param.l1,param.l2,param.l3);
  // update parameters for evaluation function
  Opt->UpdatefEvalArgs(Exemples,nbrow,InitialCoverage,MaxError,ResClassif,muthresh, Lab);
  return Opt;
}
//********************************************************************
vkey* BuildKey(sifopt*Opt,FIS* S,int keytype,int num,bool wordless)
//********************************************************************
{
  vkey * key = NULL;
 
  if (keytype==1)
    {
      // build key corresponding to input to optimize
      key=Opt->keysetIN(S,num-1,1);
    }
  if (keytype==2)
    {
      key=Opt->keysetOUT(S , num, 1);
    }
  if (keytype==3)
    {
      key=Opt->keysetRULE(S , num, -1);//-1 stands for all rules
    }

  Opt->setKeyInit(key);
  if (!wordless)
    {
      if (keytype==1)
	printf("\noptimizing fuzzy input %d\n",num);
      if (keytype==2)
	printf("\noptimizing fuzzy output %d\n",num);
      if (keytype==3)
	printf("\noptimizing rules for FIS\n");

      printf("Key:  ");
      key->Print();
    }
  return key;		    
}

//*************************************************************************
void GenIterName(char *& outputFisConfigName, int keytype,int numInput)
//*************************************************************************
{
  char iter[20];

  // input iteration
  if (keytype==1)
    {
      sprintf(iter,"-%d",numInput);
      strcat(outputFisConfigName,iter);
    }
  if (keytype==2)
    strcat(outputFisConfigName,"-fuzzyoutput");
  if (keytype==3)
    strcat(outputFisConfigName,"-all-rules");
  
}
//*************************************************************************
void GenPerfName(char *outputFisConfigName,const char *fisPath,const char * optimName, char * filename,const char * basename, int isample)
{
  char nums[100];

  //evaluate perf with appropriate data file
  strcpy(filename,basename);
  sprintf(nums,"%d",isample);
  strcat(filename,nums);
  //build FIS name from saved file
  strcpy(outputFisConfigName,fisPath);
  strcat(outputFisConfigName,optimName);
  strcat(outputFisConfigName,"-lrn.sample");
  sprintf(nums,"%d",isample);
  strcat(outputFisConfigName,nums);
  strcat(outputFisConfigName,"-final.fis");
}
//*************************************************************************
//*************************************************************************
double WritePerfFile(FILE* fperf, int ns, const char * fisfile1, const char * fisfile2, FIS * M, const char * datafile, double muthresh, int nOut, bool header,bool wordless)
//*************************************************************************
{
  double inCov,MaxError,perf0=0.0,perfmed=0.0,ga=0.0,perfopt;
  FIS *S=NULL, *Opt=NULL;
    
  perfopt=FisMknan();

  //generate perf array file
  //  3 columns: initial FIS, test-sample opt FIS, median FIS perf
  // 1 row/sample
    
  if (header)
    {
      if (ns>0)
	fprintf(fperf,"Dataset & initial FIS & sample k-optimized FIS & Median FIS & Rel.gain\n");
      else
	fprintf(fperf,"Dataset & initial FIS & Optimized FIS & Rel.gain\n");
      // perf on whole data set
    }
  if (fisfile1!=NULL)
    S=new FIS(fisfile1);
  //strcpy(filename,datafile);
  perf0=S->Performance(nOut, (char*)datafile, inCov, MaxError, muthresh, NULL, 0);
  delete S;
  S=NULL;
    
  if (fisfile2!=NULL)
    {
      Opt=new FIS(fisfile2);
      perfopt=Opt->Performance(nOut, (char*)datafile, inCov, MaxError, muthresh, NULL, 0);
      delete Opt;
    }
  if (ns>0)
    {
      // median FIS performance
      if (M !=NULL)
	perfmed=M->Performance(nOut,(char*)datafile, inCov, MaxError, muthresh, NULL, 0);
      if (perf0>0)
	ga=(perf0-perfmed)/perf0;
      else
	ga=FisMknan();

      fprintf(fperf,"%s & %g & %g & %g & %g\n",datafile,perf0,perfopt,perfmed,ga);
      if (!wordless)
	printf("\n FIS=%s Dataset=%s Perf init=%g Perf med=%g gain=%g",fisfile2,datafile,perf0,perfmed,ga);
    }//end ns>0: there are samples
  else
    {// no sample
      //build FIS from saved file
      if (perf0>0)
	ga=(perf0-perfopt)/perf0;
      else
	ga=FisMknan();
      if (!wordless)
	printf("\n FIS=%s Dataset=%s Perf init=%g Perf opt=%g Rel. gain=%g",fisfile2,datafile,perf0,perfopt,ga);
      fprintf(fperf,"%s & %g & %g & %g\n",datafile,perf0,perfopt,ga);
    }

  return ga;
}


//******************************************************************************************************************
FIS* loopoptimsample(FIS*& S,const char* fisPath, const char* datafile,const char * optimName, int nsample,int nsteps,bool optimIn,bool optimOut,bool optimRules, int nbinputs,double * numinputs, int nOut,long int seed,double muthresh,double covloss, parstruct paramP,bool intermfile,bool wordless,double nearC)
//*******************************************************************************************************************
{
  char *infile=NULL,*basename=NULL,*outputFisConfigName=NULL;
  char *finalFisConfigName=NULL/*,*fisout=NULL*/;
  int len=0,isinit=-1;

  sifopt *Opt=NULL;
 
  int keytype=0,rep;
  char nums[100];
  
  FILE *final, *fmed;
  bool fisoutflag=false,flagOpt=true;

  char  * tmpchar=NULL, *cstr=NULL;
  double ** Exemples  = NULL;   // A Pointer to sample array
  int * ResClassif = NULL;
  double * Lab = NULL;
  int nbcol=0, nbrow=0, cnear=0, numOpt=0,lastinput=0;
  int reponse=-1;
  double InitialCoverage=0.0,inCov=0.0, MaxError=0.0,perfinit=0.0;
  vkey *KeyIn;
  FIS* M=NULL, *M1=NULL;
  
  //set random sampling seed
  setseed(seed);


  len=strlen(fisPath) + strlen(optimName) + 1000;
  outputFisConfigName=new char[len];
  finalFisConfigName=new char[len];
  basename=new char[strlen(datafile)+50];
  infile=new char[strlen(datafile)+50];
  //fisout=new char[len];
 
  if (nsample >0)
    {
      isinit=0;
      strcpy(basename,datafile);
      strcat(basename,".lrn.sample.");
      strcpy(infile,basename);
    }
  else
    strcpy(infile,datafile);

  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  // loop on learning samples instead of data file
  //loop on training samples-if no sample isample=-1 and nsample=0
  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  for (int isample=isinit;isample<nsample;isample++)
    {
   
      // complete sample data file name
      if (isample >=0)
	{
	  strcpy(infile,basename);
	  sprintf(nums,"%d",isample);
	  strcat(infile,nums);
	}
      //initialize intermediate configname
      strcpy(outputFisConfigName,fisPath);
      strcat(outputFisConfigName,optimName);

      // complete optimized FIS name
      strcpy(finalFisConfigName,fisPath);
      strcat(finalFisConfigName,optimName);

      if (isample >=0)
	{
	  strcat(finalFisConfigName,"-lrn.sample");
	  strcat(finalFisConfigName,nums);
	  strcat(outputFisConfigName,"-lrn.sample-");
	  strcat(outputFisConfigName,nums);
	}
      strcat(finalFisConfigName,"-final.fis");
      final=fopen(finalFisConfigName,"wt");
      if (!wordless)
	{
	  printf("\n------------\nData file name %s\n------------",infile);
	  printf("\n------------\nOutput FIS file name %s\n------------",finalFisConfigName);
	}


      if (optimIn)
	strcat(outputFisConfigName,"-input");
    
      // read data for current training sample
      Exemples = ReadSampleFile( infile , nbcol, nbrow);
      if(!*Exemples) 
	{
	  printf("\nThe sample matrix cannot be built ! \n");
	  return NULL;
	}
      //test on column and row number
      if(nbcol < S->GetNbIn()) 
	{
	  printf("\n The number of input variables is smaller than the number of inputs in the FIS.\n");
	  return NULL;
	}

      if(nbcol < S->GetNbIn() + S->GetNbOut())
	{
	  printf("\n The number of variables is smaller than the number of inputs plus the number of outputs.");
	  return NULL;
	}  
      if (nOut>S->GetNbOut()-1)
	{
	  printf("\n Invalid value for output number: %d, max is %d\n",nOut,S->GetNbOut()-1);
	  return NULL;
	}
      // call to Classcheck once to allocate ResClassif and updates Lab pointer
      reponse=S->ClassCheck(ResClassif,Lab,Exemples,nbrow,nOut );
#ifdef DEBUG
      printf("\n The Class Check returns %i",reponse);
#endif
      perfinit=S->Performance(nOut, infile, inCov, MaxError, muthresh, NULL, 0);
      if (!wordless)
	printf("\nperfinit (before normalization)=%f\n",perfinit);
    
  
	
      InitialCoverage=inCov;
      
      // .... 
      //beginning of optim
	
      if (!wordless)
	printf("\n%d loop iterations for optimizing",nsteps);
   
      for (int step=0;step<nsteps;step++)
	{
	  if (!wordless)
	    printf("\n-------------------------\nIteration %d, sample %d\n------------------------",step+1,isample+1);
	    
	  // nbinputs  iterations + 1 iteration for fuzzy output + 1 for rules
	  for (int kk=0;kk<nbinputs+2;kk++)
	    {
	      flagOpt=true;
	      if (!wordless)
		{
		  if (optimIn && (kk==0))
		    printf("\n\n--------------------\noptimizing inputs\n--------------------\n");
		  if (optimOut && (kk==nbinputs))
		    printf("\n\n--------------------\noptimizing fuzzy output\n--------------------\n");
		  if (optimRules && (kk==nbinputs+1))
		    printf("\n\n--------------------\noptimizing rules\n--------------------\n");
		}
	
	      //optimIn case
	      if (kk<nbinputs)
		{
		  keytype=1;
		  numOpt=(int)numinputs[kk];
		}
	      //optimOut case
	      if (kk==nbinputs)
		{
		  keytype=2;
		  numOpt=nOut;
		  if (!optimOut)
		    flagOpt=false;
		}
	      //optimrules case
	      if (kk==nbinputs+1)
		{
		  keytype=3;
		  numOpt=nOut;
		  if (!optimRules)
		    flagOpt=false;
		}
	      if (flagOpt)
		{ 	
		  //create sifopt using updated FIS - FIS is normalized and data also (Exemples array)
		  // nbrow = passage by reference necessary for UpdatefEvalArgs
		  Opt=PrepareOpt(S, ResClassif, Lab, Exemples, nbrow,  nOut,cnear,nearC,InitialCoverage,covloss,MaxError,muthresh,paramP);
		  if (!wordless)
		    Opt->PrintNC();
		    		  
		  KeyIn=BuildKey(Opt,S,keytype,numOpt, wordless);
		    
		  // flag for intermediate fis file
		  lastinput=(kk>=nbinputs-1);
		  fisoutflag=lastinput || intermfile;
		  GenIterName(outputFisConfigName,keytype,numOpt);
		  if (!wordless)
		    printf("\noutputFisConfigName=%s",outputFisConfigName);
		    
		  // call optim
		  // record optimized FIS and update Opt
		  rep=CallOptim(Opt,S, KeyIn, Exemples, infile,  nOut, nbrow, nbcol, perfinit, muthresh, fisoutflag, outputFisConfigName, wordless);
		  if (Exemples !=NULL) 
		    {
		      for( int i=0 ; i<nbrow ; i++ )
			delete [] Exemples[i];
		      delete [] Exemples;
		      Exemples=NULL;
		    }
		  // reread Exemples
		  Exemples = ReadSampleFile( infile , nbcol, nbrow);
		}//end flagOpt
      
	    }// end of optimization loop on input/output/rules

	    
	}//end loop on steps
	 // final output file for currentsample
      S->PrintCfg(final);
      fclose(final);
      
      if (ResClassif != NULL) 
	{
	  delete [] ResClassif;//Lab ?
	  ResClassif=NULL;
	}
      
      if (Exemples !=NULL) 
	{
	  for( int i=0 ; i<nbrow ; i++ )
	    delete [] Exemples[i];
	  delete [] Exemples;
	  Exemples=NULL;
	}
    }
  // end loop on samples
  

  // if samples exist, build medfis
  if (nsample >0)
    {
      //build median FIS
      tmpchar=new char [strlen(fisPath) + strlen(optimName)+15];
      strcpy(tmpchar,fisPath);
      strcat(tmpchar,optimName);
      strcat(tmpchar,"-lrn.sample");
      M=medfis(tmpchar,"-final.fis",nsample,wordless);
      delete [] tmpchar;
    }
  else
    M=S->Clone();//if no sample median FIS=optimized FIS

  std::string medfisName(fisPath);
  medfisName.append("med.fis");
  fmed=fopen(medfisName.c_str(),"wt");
  M->PrintCfg(fmed);
  fclose(fmed);

  // check that MFs are correct-as MF constructor is not called in medfis...
  try{
    cstr = new char [medfisName.size()+1];
    strcpy (cstr, medfisName.c_str());
    M1=new FIS(cstr);
    delete M1;
    delete [] cstr;
  }
   catch(std::exception &e){
    printf("error in building medfis\n");
    printf("catch %s\n",e.what());
    return S;
  }
  //delete temp arrays
  if (finalFisConfigName!= NULL) delete [] finalFisConfigName; 
  if (basename!= NULL) delete [] basename;
  if (outputFisConfigName!= NULL) delete [] outputFisConfigName;
  if(infile!= NULL) delete []infile;
  //if (fisout!=NULL) delete [] fisout;
  if (Exemples !=NULL) 
    {
      for( int i=0 ; i<nbrow ; i++ )
	delete [] Exemples[i];
      delete [] Exemples;
      Exemples=NULL;
    }
  if (ResClassif != NULL) delete [] ResClassif;//Lab 
    
  return M;

}//end loopoptimsample function
//***************************************
void computeWritePerf(FIS* M,const char* fisPath,const char * datafile,const char * perfile,const char * fisname,const char* optimName,  double muthresh, int nOut, int nsample, bool wordless)
//***************************************
// build for each sample
// perf file: initial perf on whole sample, training, initial perf on test
//            optimized perf on whole sample, on training, optimized perf on test
 
{
  bool header;
  FILE * fperf=NULL;
  char * outputFisConfigName=NULL,*basename=NULL,*infile=NULL;
  int kk,len;
  
  header=true;
  std::string perfileName(fisPath);
  perfileName.append(perfile);
  fperf=fopen(perfileName.c_str(), "wt");
  basename=new char[strlen(datafile)+50];
  len=strlen(fisPath) + strlen(optimName) + 50;
  outputFisConfigName=new char[len];
  infile=new char[strlen(datafile)+50];

  // write header plus first line with perf initial and perf median
  // on initial dataset
  // and perfopt if no sample or NA if there are samples
  if (nsample==0)//full perf
    {
      strcpy(outputFisConfigName,fisPath);
      strcat(outputFisConfigName,optimName);
      strcat(outputFisConfigName,"-final.fis");
      WritePerfFile(fperf,nsample,fisname,outputFisConfigName,M,datafile,muthresh,nOut,header,wordless);
    }
  else //no perfopt on full data set
    WritePerfFile(fperf,nsample,fisname,NULL,M,datafile,muthresh,nOut,header,wordless);
  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  // loop on perf for test samples plus learning samples, instead of data file
  //if no sample nsample=0
  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  if (nsample>0)
    {
      strcpy(basename,datafile);
      if (!wordless)
	printf("\nSample perf loop\n");
      //first loop on test samples, second loop on lrn samples
      header =false;
      for (kk=0;kk<2;kk++)
	{
	  if (kk==0)
	    strcat(basename,".tst.sample.");
	  else
	    {
	      strcpy(basename,datafile);
	      strcat(basename,".lrn.sample.");
	    }
	      
	  for (int isample=0;isample<nsample;isample++)
	    {
	      //generates names for infile and outputFisConfigName
	      // from basename, isample and optimName
	      GenPerfName(outputFisConfigName,fisPath,optimName,infile,basename,isample);
	      // write perf file with perf initial, perf opt and perf median
	      // on each data sample (lrn or tst)
	      WritePerfFile(fperf,nsample,fisname,outputFisConfigName,M,infile,muthresh,nOut,header,wordless);
	    }// isample loop
	}// for loop
    }// nsample>0 loop
  fclose(fperf);
    
  if (infile !=NULL) delete [] infile;
  if (basename!= NULL) delete [] basename;
  if (outputFisConfigName!= NULL) delete [] outputFisConfigName;

}//end computeWritePerf function
