//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************

/*! \file  optconstraint.cpp
 *    \brief Optimizer member function testConstraints.
 *
 This File contents the definition of the Optimizer member function testConstraints. 
*/
/*!
 * \fn int sifopt::testConstraints( double * Array , int size , vkey * pKey )
 * \brief This function check if the set of values, check some contraints, if not it can consider a FAIL case or modify the set of values. Meanwhile this function construct the InputCenters Arrays that will be sent to the FIS constructor.
 * \param Array The array U have just built whith ObjectToDouble functiun
 * \param size Te size of the array.
 * \param pKey The key to encode your object.
 * \return -1 if it's a FAIL, nothing else otherwise.
 */
#include "sifopt.h"
int sifopt::oldtestConstraints( double * Array , int size , vkey * pKey )
{
  int  indArray=0,indArrayInput=0,nbSef=0,ind=0,indKey=0,NbPar=0,cpt2,cptActiveEntry=0;
  int *pType,tmpVar;
  MF *Cour=NULL;
  double *sef_par=NULL;
  double CNearRatio=0.0;
  int NumEntry=0,indPFF=0;
  int SizeInput=0;

  //case of input
  if ( IsInput == 1 )
    {
      for ( NumEntry = 0 ; NumEntry < nbEntree ; NumEntry ++ )
	{
	  if  ( InitialFis->In[NumEntry]->IsActive() == false )  continue;
#ifdef DEBUGC
	  printf("\n+ConstraintBuildIncputCenters NumEntry  %i " ,  NumEntry );
#endif
	  indPFF= PFFIn.Value[cptActiveEntry];cptActiveEntry++;
	  pType=InputCentersType[NumEntry];
	  nbSef=InputNbMf[NumEntry];
	  SizeInput=0;indArrayInput=0;
	  tmpVar=nbSef+1;
	  CNearRatio=(float)(1.0 / (float)tmpVar);

      
#ifdef DEBUGC
	  printf( "nbSef %i " , nbSef);
	  printf( "CNearRatio %f " , CNearRatio);
	  printf( "indArrayInput %i indArray %i |" ,indArrayInput , indArray );
#endif
	  for ( int NumSef = 0 ; NumSef < nbSef ; NumSef ++ )
	    {
	      if ( Key.Value[indKey] == 1 ){
		if ( indPFF == 1 ) 
		  {
#ifdef DEBUGC
		    printf("\n+++++ PFF=1 " );
#endif
		    if ( *pType == 1 )
		      {
			InputCenters[NumEntry][indArrayInput]=Array[indArray];
			InputCentersUpdatable[NumEntry][indArrayInput]=1;
			if ( ( CNearActivate == 1 ) && ( CNearInitIn == 0 ))
			  {
			    InputCentersMax[NumEntry][indArrayInput]=Array[indArray]+CNearRatio;
			    InputCentersMin[NumEntry][indArrayInput]=Array[indArray]-CNearRatio;
			  }
#ifdef DEBUGC
			printf(" ( TR  %f %i  1 %i ) " ,  InputCenters[NumEntry][indArrayInput], indArrayInput, indArray);
#endif
			indArray++;
			indArrayInput++;
		      }
		    else
		      {
			InputCenters[NumEntry][indArrayInput]=Array[indArray];
			InputCentersUpdatable[NumEntry][indArrayInput]=1;
			InputCenters[NumEntry][indArrayInput+1]=Array[indArray+1];
			InputCentersUpdatable[NumEntry][indArrayInput+1]=1;
			if ( ( CNearActivate == 1 ) && ( CNearInitIn == 0 ))
			  {
			    InputCentersMax[NumEntry][indArrayInput]=Array[indArray]+CNearRatio;
			    InputCentersMin[NumEntry][indArrayInput]=Array[indArray]-CNearRatio;
			    InputCentersMax[NumEntry][indArrayInput+1]=Array[indArray+1]+CNearRatio;
			    InputCentersMin[NumEntry][indArrayInput+1]=Array[indArray+1]-CNearRatio;
			  }
#ifdef DEBUGC
			printf(" ( TRAP  %f %i %f %i  1 %i) " ,  InputCenters[NumEntry][indArrayInput], indArrayInput, InputCenters[NumEntry][indArrayInput+1] , indArrayInput+1 ,indArray);
#endif
			indArray=indArray+2;;
			indArrayInput=indArrayInput+2;
		      }}
		else // PFF = 0
		  {
#ifdef DEBUGC
		    printf("\n+++++ PFF=O " );
#endif
		    NbPar=InputCentersNbPar[NumEntry][NumSef];
		    for ( cpt2=0 ; cpt2 < NbPar ; cpt2++ )
		      {
			if ( ( CNearActivate == 1 ) && ( CNearInitIn == 0 )){
			  InputCentersMax[NumEntry][indArrayInput+cpt2]=Array[indArray+cpt2]+CNearRatio;
			  InputCentersMin[NumEntry][indArrayInput+cpt2]=Array[indArray+cpt2]-CNearRatio;
			}
			InputCenters[NumEntry][indArrayInput+cpt2]=Array[indArray+cpt2];
#ifdef DEBUGC
			printf(" (  %f %i 1 )", InputCenters[NumEntry][indArrayInput+cpt2], indArrayInput +cpt2 );
#endif
		      };
		    indArrayInput+=NbPar;
		    indArray+=NbPar;
		  };
	      } 
	      else
		{
		  Cour=InitialNormedFis->In[NumEntry]->GetMF(NumSef);
		  NbPar=InputCentersNbPar[NumEntry][NumSef];
		  sef_par = new double[ NbPar];
		  Cour->GetParams(sef_par);
		  if ( PFFIn.Value[NumEntry] == 1 ) 
		    {
#ifdef DEBUGC
		      printf("\nLock+ PFF=1 " );
#endif
		      if ( *pType == 1 )
			{
			  InputCenters[NumEntry][indArrayInput]=sef_par[1];
			  if ( ( CNearActivate == 1 ) && ( CNearInitIn == 0 ))
			    {
			      InputCentersMax[NumEntry][indArrayInput]=sef_par[1]+CNearRatio;
			      InputCentersMin[NumEntry][indArrayInput]=sef_par[1]-CNearRatio;
			    }
			  InputCentersUpdatable[NumEntry][indArrayInput]=0;
#ifdef DEBUGC
			  printf(" ( TR  %f %i  0 ) " ,  InputCenters[NumEntry][indArrayInput], indArrayInput);
#endif
			  indArrayInput++;
			}


		      if ( ( *pType == 2 ) ||  ( *pType == 4 ))
			{
			  if ( ( CNearActivate == 1 ) && ( CNearInitIn == 0 ))
			    {
			      InputCentersMax[NumEntry][indArrayInput]=sef_par[1]+CNearRatio;
			      InputCentersMin[NumEntry][indArrayInput]=sef_par[1]-CNearRatio;
			    }
			  InputCenters[NumEntry][indArrayInput]=sef_par[1];
			  InputCentersUpdatable[NumEntry][indArrayInput]=0;
#ifdef DEBUGC
			  printf(" ( TRAP TRAPSUP  %f %i   " ,  InputCenters[NumEntry][indArrayInput], indArrayInput);
#endif
			  indArrayInput++;
			  if ( ( CNearActivate == 1 ) && ( CNearInitIn == 0 ))
			    {
			      InputCentersMax[NumEntry][indArrayInput]=sef_par[2]+CNearRatio;
			      InputCentersMin[NumEntry][indArrayInput]=sef_par[2]-CNearRatio;
			    }
			  InputCenters[NumEntry][indArrayInput]=sef_par[2];
			  InputCentersUpdatable[NumEntry][indArrayInput]=0;
#ifdef DEBUGC
			  printf(" %f %i 0 )" ,  InputCenters[NumEntry][indArrayInput], indArrayInput);
#endif
			  indArrayInput++;

			}

		      if ( ( *pType == 3 ) )
			{
			  InputCenters[NumEntry][indArrayInput]=sef_par[0];
			  if ( ( CNearActivate == 1 ) && ( CNearInitIn == 0 ))
			    {
			      InputCentersMax[NumEntry][indArrayInput]=sef_par[0]+CNearRatio;
			      InputCentersMin[NumEntry][indArrayInput]=sef_par[0]-CNearRatio;
			    }
			  InputCentersUpdatable[NumEntry][indArrayInput]=0;
#ifdef DEBUGC
			  printf(" ( TRAPINF  %f %i   " ,  InputCenters[NumEntry][indArrayInput], indArrayInput);
#endif
			  indArrayInput++;
			  InputCenters[NumEntry][indArrayInput]=sef_par[1];
			  if ( ( CNearActivate == 1 ) && ( CNearInitIn == 0 ))
			    {
			      InputCentersMax[NumEntry][indArrayInput]=sef_par[1]+CNearRatio;
			      InputCentersMin[NumEntry][indArrayInput]=sef_par[1]-CNearRatio;
			    }
			  InputCentersUpdatable[NumEntry][indArrayInput]=0;
#ifdef DEBUGC
			  printf(" %f %i 0 )" ,  InputCenters[NumEntry][indArrayInput], indArrayInput);
#endif
			  indArrayInput++;
			}
		    }
		  else // PFF = 0
		    {
#ifdef DEBUGC
		      printf("\nLock+ PFF=0 " );
#endif
		      for ( cpt2=0 ; cpt2 < NbPar ; cpt2++ )
			{
			  InputCenters[NumEntry][indArrayInput]=sef_par[cpt2];
#ifdef DEBUGC
			  printf(" (  %f %i 1 )", InputCenters[NumEntry][indArrayInput], indArrayInput );
#endif
			  if ( ( CNearActivate == 1 ) && ( CNearInitIn == 0 ))
			    {
			      InputCentersMax[NumEntry][indArrayInput]=sef_par[cpt2]+CNearRatio;
			      InputCentersMin[NumEntry][indArrayInput]=sef_par[cpt2]-CNearRatio;
			    }
			  indArrayInput++;
			};
		    };
		  delete [] sef_par;

		};
	      pType++;
	      indKey++;
	      ind++;
	    };

	  //! The CSort ( or CSortModify ) and CSortKernel constraints are needed for inner FIS constraint
	  if  ( PFFIn.Value[cptActiveEntry-1] == 1 ){ 
	    if ( CSort( indArrayInput, NumEntry ) == -1 )
	      return -1;
	  }else{
	    if ( CSortCenter( indArrayInput, NumEntry ) == -1 )
	      return -1;
	  }
	  if ( CDomain(  indArrayInput, NumEntry ) == -1 )  
	    return -1;
	  if (CNearActivate == 1 ) { if ( CNear(  indArrayInput,  NumEntry ) == -1 )  return -1; }
	  /*
	    if ( CDomainModify(  indArrayInput, NumEntry ) == -1 )
	    return -1;
	    if ( CSortModify( indArrayInput, NumEntry ) == -1 )
	    return -1;*/
	  /* 
	   * Test de la version des contraintes qui retire un gaussien
	   */
	};

    }



  if ( IsInput == 0 )
    {
      nbSef=OutputNbMf;
      indPFF= PFFOut;
      NumEntry=nbEntree;
      pType=InputCentersType[NumEntry];
      SizeInput=0;indArrayInput=0;
      tmpVar=nbSef+1;
      CNearRatio=(float)(1.0 / (float)tmpVar);
      
#ifdef DEBUGC
      printf( "nbSef %i " , nbSef);
      printf("\n+++++ CNearRatio %f", CNearRatio);
#endif

      for ( int NumSef = 0 ; NumSef < nbSef ; NumSef ++ )
        {
	  if ( Key.Value[indKey] == 1 ){
	    if ( indPFF == 1 ) 
	      {
#ifdef DEBUGC
		printf("\n+++++ PFF=1 " );
#endif
		if ( *pType == 1 )
		  {
		    InputCenters[NumEntry][indArrayInput]=Array[indArray];
		    if ( ( CNearActivate == 1 ) && ( CNearInitOut == 0 ))
		      {
			InputCentersMax[NumEntry][indArrayInput]=Array[indArray]+CNearRatio;
			InputCentersMin[NumEntry][indArrayInput]=Array[indArray]-CNearRatio;
		      }
		    InputCentersUpdatable[NumEntry][indArrayInput]=1;
#ifdef DEBUGC
		    printf(" ( TR  %f %i  1 %i ) " ,  InputCenters[NumEntry][indArrayInput], indArrayInput, indArray);
#endif
		    indArray++;
		    indArrayInput++;
		  }
		else
		  {
		    InputCenters[NumEntry][indArrayInput]=Array[indArray];
		    InputCentersUpdatable[NumEntry][indArrayInput]=1;
		    if ( ( CNearActivate == 1 ) && ( CNearInitOut == 0 ))
		      {
			InputCentersMax[NumEntry][indArrayInput]=Array[indArray]+CNearRatio;
			InputCentersMin[NumEntry][indArrayInput]=Array[indArray]-CNearRatio;
		      }
		    InputCenters[NumEntry][indArrayInput+1]=Array[indArray+1];
		    InputCentersUpdatable[NumEntry][indArrayInput+1]=1;
		    if ( ( CNearActivate == 1 ) && ( CNearInitOut == 0 ))
		      {
			InputCentersMax[NumEntry][indArrayInput+1]=Array[indArray+1]+CNearRatio;
			InputCentersMin[NumEntry][indArrayInput+1]=Array[indArray+1]-CNearRatio;
		      }
#ifdef DEBUGC
		    printf(" ( TRAP  %f %i %f %i  1 %i) " ,  InputCenters[NumEntry][indArrayInput], indArrayInput, InputCenters[NumEntry][indArrayInput+1] , indArrayInput+1 ,indArray);
#endif
		    indArray=indArray+2;;
		    indArrayInput=indArrayInput+2;
		  }}
	    else // PFF = 0
	      {
#ifdef DEBUGC
		printf("\n+++++ PFF=O " );
#endif
		NbPar=InputCentersNbPar[NumEntry][NumSef];
		for ( cpt2=0 ; cpt2 < NbPar ; cpt2++ )
		  {
		    if ( ( CNearActivate == 1 ) && ( CNearInitOut == 0 ))
		      {
			InputCentersMax[NumEntry][indArrayInput+cpt2]=Array[indArray+cpt2]+CNearRatio;
			InputCentersMin[NumEntry][indArrayInput+cpt2]=Array[indArray+cpt2]-CNearRatio;
		      }
		    InputCenters[NumEntry][indArrayInput+cpt2]=Array[indArray+cpt2];
#ifdef DEBUGC
		    printf(" (  %f %i 1 )", InputCenters[NumEntry][indArrayInput+cpt2], indArrayInput +cpt2 );
#endif
		  };
		indArrayInput+=NbPar;
		indArray+=NbPar;
	      };
	  } 
	  else
	    {
	      Cour=InitialNormedFis->In[NumEntry]->GetMF(NumSef);
	      NbPar=InputCentersNbPar[NumEntry][NumSef];
	      sef_par = new double[ NbPar];
	      Cour->GetParams(sef_par);
	      if ( PFFIn.Value[NumEntry] == 1 ) 
		{
#ifdef DEBUGC
		  printf("\nLock+ PFF=1 " );
#endif
		  if ( *pType == 1 )
		    {
		      InputCenters[NumEntry][indArrayInput]=sef_par[1];
		      if ( ( CNearActivate == 1 ) && ( CNearInitOut == 0 ))
			{
			  InputCentersMax[NumEntry][indArrayInput]=sef_par[1]+CNearRatio;
			  InputCentersMin[NumEntry][indArrayInput]=sef_par[1]-CNearRatio;
			}
		      InputCentersUpdatable[NumEntry][indArrayInput]=0;
#ifdef DEBUGC
		      printf(" ( TR  %f %i  0 ) " ,  InputCenters[NumEntry][indArrayInput], indArrayInput);
#endif
		      indArrayInput++;
		    }


		  if ( ( *pType == 2 ) ||  ( *pType == 4 ))
		    {
		      InputCenters[NumEntry][indArrayInput]=sef_par[1];
		      if ( ( CNearActivate == 1 ) && ( CNearInitOut == 0 ))
			{
			  InputCentersMax[NumEntry][indArrayInput]=sef_par[1]+CNearRatio;
			  InputCentersMin[NumEntry][indArrayInput]=sef_par[1]-CNearRatio;
			}
		      InputCentersUpdatable[NumEntry][indArrayInput]=0;
#ifdef DEBUGC
		      printf(" ( TRAP TRAPSUP  %f %i   " ,  InputCenters[NumEntry][indArrayInput], indArrayInput);
#endif
		      indArrayInput++;
		      InputCenters[NumEntry][indArrayInput]=sef_par[2];
		      if ( ( CNearActivate == 1 ) && ( CNearInitOut == 0 ))
			{
			  InputCentersMax[NumEntry][indArrayInput]=sef_par[2]+CNearRatio;
			  InputCentersMin[NumEntry][indArrayInput]=sef_par[2]-CNearRatio;
			}
		      InputCentersUpdatable[NumEntry][indArrayInput]=0;
#ifdef DEBUGC
		      printf(" %f %i 0 )" ,  InputCenters[NumEntry][indArrayInput], indArrayInput);
#endif
		      indArrayInput++;

		    }

		  if ( ( *pType == 3 ) )
		    {
		      InputCenters[NumEntry][indArrayInput]=sef_par[0];
		      if ( ( CNearActivate == 1 ) && ( CNearInitOut == 0 ))
			{
			  InputCentersMax[NumEntry][indArrayInput]=sef_par[0]+CNearRatio;
			  InputCentersMin[NumEntry][indArrayInput]=sef_par[0]-CNearRatio;
			}
		      InputCentersUpdatable[NumEntry][indArrayInput]=0;
#ifdef DEBUGC
		      printf(" ( TRAPINF  %f %i   " ,  InputCenters[NumEntry][indArrayInput], indArrayInput);
#endif
		      indArrayInput++;
		      InputCenters[NumEntry][indArrayInput]=sef_par[1];
		      if ( ( CNearActivate == 1 ) && ( CNearInitOut == 0 ))
			{
			  InputCentersMax[NumEntry][indArrayInput]=sef_par[1]+CNearRatio;
			  InputCentersMin[NumEntry][indArrayInput]=sef_par[1]-CNearRatio;
			}
		      InputCentersUpdatable[NumEntry][indArrayInput]=0;
#ifdef DEBUGC
		      printf(" %f %i 0 )" ,  InputCenters[NumEntry][indArrayInput], indArrayInput);
#endif
		      indArrayInput++;
		    }
		};
	      delete [] sef_par;

	    };
	  pType++;
	  indKey++;
	  ind++;
	};

      if  ( PFFOut == 1 ){ 
	if ( CSort( indArrayInput, nbEntree ) == -1 )
	  return -1;
      }else{
	if ( CSortCenter( indArrayInput, nbEntree ) == -1 )
	  return -1;
      }
      if ( CDomain(  indArrayInput, nbEntree ) == -1 )  
	return -1;
      if (CNearActivate == 1 ) { if ( CNear(  indArrayInput, nbEntree ) == -1 )  return -1; }
    }
  // Init the Min and Max array once for the CNear Constraint


  if ( ( CNearActivate == 1 ) && ( CNearInitOut == 0 ) && ( IsInput == 0 )) CNearInitOut=1;
  if ( ( CNearActivate == 1 ) && ( CNearInitIn == 0 ) ) CNearInitIn=1;

  if ( IsInput == 2 )
    {
      for ( int Cpt = 0 ; Cpt < NbRule ; Cpt++ )
	{
          if  ( InitialNormedFis->Rule[Cpt]->IsActive() == false ) continue;
          if ( Key.Value[indKey] == 1 )
	    {
	      InputCenters[nbEntree+1][indArrayInput]=Array[indArray];
	      indArray++;
	    }
          indKey++;
          indArrayInput++;
	}
      if ( Name != PERMUTATION ) { if ( CDomain(  indArrayInput , nbEntree+1 ) == -1 ) return -1; }
    }
  return 1;
}

int  sifopt::oldCSort(  int indArrayInput, int nbE )
{
  for ( int cpt = 1 ; cpt < indArrayInput ; cpt ++ )
    if (( fabs ( InputCenters[nbE][cpt] - InputCenters[nbE][cpt-1] ) > EPSILON ) &&
        ( InputCenters[nbE][cpt] < InputCenters[nbE][cpt-1] ))
      {
#ifdef DEBUGC
	printf(" !  %i : %g  < %g  collision detecte par contrainte : ",
	       cpt,
	       InputCenters[nbE][cpt],
	       InputCenters[nbE][cpt-1]);
	printf(" fabs = %g EPSILON = %g ",fabs ( InputCenters[nbE][cpt] - InputCenters[nbE][cpt-1] ),EPSILON);
#endif
	return -1;
      }
#ifdef DEBUGC
    else
      {
	printf(" %i : %f  > %f ",
	       cpt,
	       InputCenters[nbE][cpt],
	       InputCenters[nbE][cpt-1]);
      }
#endif
  ;
 return 1;
}


int  sifopt::oldCSortCenter(  int indArrayInput, int nbE )
{
#ifdef DEBUGC
  printf("\n+CSortCenter  " );
#endif
  int cptMF=1;
  int cpt=1+InputCentersNbPar[nbE][0];
  double prev=InputCenters[nbE][1];
  while ( cpt < indArrayInput ){
    if ( InputCenters[nbE][cpt] < prev   )
      {
#ifdef DEBUGC
	printf("\n+Ccollision detecte par contrainte ordre sur non PFF : cpt %i valeur %f ", cpt , InputCenters[nbE][cpt] );
#endif
	return -1;
      }
    else
      {
	cpt += InputCentersNbPar[nbE][cptMF];
	cptMF++;
      }
  }
#ifdef DEBUGC
  printf(" OK" );
#endif
  return 1;
  //! \todo instead of testing the second param of the MF use the gravity center of the MF to check if it's sorted
}



/*!
 * \fn int  sifopt::CNear( int indArrayInput, int nbE )
 * \param indArrayInput the size of the In params array
 * \param nbE the number of the in.
 * \return 1 if it's OK -1 if a constaints was violated.
 * \brief check if the param move more than 1 / ( nbMF + 1 )
 */
int  sifopt::CNear( int indArrayInput, int nbE )    
{
  double centol=0.1;
  int *pType;
  int nbSef,npar/*,istart*/,i;

#ifdef DEBUGC
  printf("\n+CNear  " );
#endif
  for ( int cpt = 0 ; cpt < indArrayInput ; cpt ++ ){
#ifdef DEBUGC
    printf(" |  %f < %f < %f | ", InputCentersMin[nbE][cpt], InputCenters[nbE][cpt], InputCentersMax[nbE][cpt] );
#endif
    if ( InputCenters[nbE][cpt] < InputCentersMin[nbE][cpt] ) {
#ifdef DEBUGC
      printf("\n+C Contrainte cnear  %f < %f ", InputCenters[nbE][cpt], InputCentersMin[nbE][cpt] );
#endif
      return -1;
    }
    if ( InputCenters[nbE][cpt] > InputCentersMax[nbE][cpt] ) {
#ifdef DEBUGC
      printf("\n+C Contrainte cnear  %f > %f ", InputCenters[nbE][cpt], InputCentersMax[nbE][cpt] );
#endif
      return -1;
    }
  }
#ifdef DEBUGC
  printf(" OK\n" );
#endif
  // added by bch - 30 septembre, 2011
  // MF param not closer than tol
  // pType=1: triangular, 2 trap 3 semi-trap inf, 4 semi-trap sup
  //       5: gbell
  //       6: Gaussian
  //       7: universal
    
  pType=InputCentersType[nbE];
  nbSef=InputNbMf[nbE];
  i=0;
    
  if (pType[i]==1)
    npar=3;
  else if (pType[i]==2)
    npar=4;
  else if (pType[i]==3 || pType[i]==4)
    npar=3;
  else if (pType[i]==6 || pType[i]==7)
    npar=2;//loop not to be done
  else if (pType[i]==5)
    npar=3;//loop not to be done
  /*
    for ( int cpt = istart ; cpt < istart+npar-1; cpt ++ ){
    if(pType[i]<=4)
    if ( InputCenters[nbE][cpt+1]<InputCenters[nbE][cpt]+centol)
    {
    printf("\n Pb!centol=%f, InputCenters[%d][%d]=%f< InputCenters[%d][%d]=%f + centol\n",centol,nbE,cpt+1,InputCenters[nbE][cpt+1],nbE,cpt,InputCenters[nbE][cpt]);
    for (int kk=0;kk<indArrayInput;kk++)
    printf("\tInputCenters[%d][%d]=%f",nbE,kk,InputCenters[nbE][kk]);
    return -1;
    }
    }
    istart=istart+npar;
  */
  // does not work!!!
  for ( int cpt = i; cpt <indArrayInput-1 ; cpt ++ ){
    if(pType[0]<=4 && indArrayInput<=4)
      if ( InputCenters[nbE][cpt+1]<InputCenters[nbE][cpt]+centol)
	{
	  printf("\n Pb!centol=%f, InputCenters[%d][%d]=%f< InputCenters[%d][%d]=%f + centol\n",centol,nbE,cpt+1,InputCenters[nbE][cpt+1],nbE,cpt,InputCenters[nbE][cpt]);
	  for (int kk=0;kk<indArrayInput;kk++)
	    printf("\tInputCenters[%d][%d]=%f",nbE,kk,InputCenters[nbE][kk]);
	  return -1;
	}
  }
    
  return 1;
}
// new version bch Feb 2013-implements cnearconstraint
int sifopt::testConstraints( double * Array , int size )
{
  int indArray=0,indArrayInput=0,nbSef=0,ind=0,indKey=0,NbPar=0,cpt2=0,cptActiveEntry=0;
  int *pType;
  MF *Cour=NULL;
  double *sef_par=NULL;
  double CNearRatio=0.0;
  int NumEntry=0,indPFF=0;
  int SizeInput=0;

  //input case
  if ( IsInput == 1 )
    {
      for ( NumEntry = 0 ; NumEntry < nbEntree ; NumEntry ++ )//loop on entries
	{
	  if  ( InitialFis->In[NumEntry]->IsActive() == false )  continue;//inactive entry: nothing to do

	  indPFF= PFFIn.Value[cptActiveEntry];
	  cptActiveEntry++;
	  pType=InputCentersType[NumEntry];
	  nbSef=InputNbMf[NumEntry];
	  SizeInput=0;indArrayInput=0;
  
	  for ( int NumSef = 0 ; NumSef < nbSef ; NumSef ++ )//loop on MFs
	    {
	      if ( Key.Value[indKey] == 1 ){//MF is to be optimized
		if ( indPFF == 1 ) //entry is PFF
		  {

		    if ( *pType == 1 )
		      {
			InputCenters[NumEntry][indArrayInput]=Array[indArray];
			InputCentersUpdatable[NumEntry][indArrayInput]=1;
			indArray++;
			indArrayInput++;
		      }
		    else
		      {
			InputCenters[NumEntry][indArrayInput]=Array[indArray];
			InputCentersUpdatable[NumEntry][indArrayInput]=1;
			InputCenters[NumEntry][indArrayInput+1]=Array[indArray+1];
			InputCentersUpdatable[NumEntry][indArrayInput+1]=1;
			indArray=indArray+2;;
			indArrayInput=indArrayInput+2;
		      }
		  }
		else // not PFF
		  {
		    NbPar=InputCentersNbPar[NumEntry][NumSef];
		    for ( cpt2=0 ; cpt2 < NbPar ; cpt2++ )
		      {
 			InputCenters[NumEntry][indArrayInput+cpt2]=Array[indArray+cpt2];
		      };
		    indArrayInput+=NbPar;
		    indArray+=NbPar;
		  };
	      }//end entry is to be optimized
	      else//entry is not to be optimized
		{
		  Cour=InitialNormedFis->In[NumEntry]->GetMF(NumSef);
		  NbPar=InputCentersNbPar[NumEntry][NumSef];
		  sef_par = new double[ NbPar];
		  Cour->GetParams(sef_par);
		  if ( PFFIn.Value[NumEntry] == 1 ) //entry is PFF
		    {
		      if ( *pType == 1 )
			{
			  InputCenters[NumEntry][indArrayInput]=sef_par[1];
			  InputCentersUpdatable[NumEntry][indArrayInput]=0;
			  indArrayInput++;
			}


		      if ( ( *pType == 2 ) ||  ( *pType == 4 ))
			{
			  InputCenters[NumEntry][indArrayInput]=sef_par[1];
			  InputCentersUpdatable[NumEntry][indArrayInput]=0;
			  indArrayInput++;
			  InputCenters[NumEntry][indArrayInput]=sef_par[2];
			  InputCentersUpdatable[NumEntry][indArrayInput]=0;
			  indArrayInput++;
			}

		      if ( ( *pType == 3 ) )
			{
			  InputCenters[NumEntry][indArrayInput]=sef_par[0];
			  InputCentersUpdatable[NumEntry][indArrayInput]=0;
			  indArrayInput++;
			  InputCenters[NumEntry][indArrayInput]=sef_par[1];
			  InputCentersUpdatable[NumEntry][indArrayInput]=0;
			  indArrayInput++;
			}
		    }
		  else // entry is not PFF
		    {
		      for ( cpt2=0 ; cpt2 < NbPar ; cpt2++ )
			{
			  InputCenters[NumEntry][indArrayInput]=sef_par[cpt2];
			  indArrayInput++;
			};
		    };
		  delete [] sef_par;

		};
	      pType++;
	      indKey++;
	      ind++;
	    };
	  if ( CDomain(  indArrayInput, NumEntry ) == -1 )  //preserve domain range
	    return -1;
	  //! The CSort ( or CSortModify ) and CSortKernel constraints are needed for inner FIS constraint
	  if  ( PFFIn.Value[cptActiveEntry-1] == 1 ){ //if PFF sort centers
	    if ( CSort( indArrayInput, NumEntry ) == -1 )
	      return -1;
	  }else{//if not PFF sort centers
	    if ( CSortCenter( indArrayInput, NumEntry ) == -1 )
	      return -1;
	  }
	  
	}

    }//end input
  
  if ( IsInput == 0 )//fuzzy output case
    {
      nbSef=OutputNbMf;
      indPFF= PFFOut;
      NumEntry=nbEntree;
      pType=InputCentersType[NumEntry];
      SizeInput=0;indArrayInput=0;

      for ( int NumSef = 0 ; NumSef < nbSef ; NumSef ++ )//loop on MFs
        {
	  if ( Key.Value[indKey] == 1 ){//MF is to be optimized
	    if ( indPFF == 1 ) //output is PFF
	      {
		if ( *pType == 1 )
		  {
		    InputCenters[NumEntry][indArrayInput]=Array[indArray];
		    InputCentersUpdatable[NumEntry][indArrayInput]=1;
		    indArray++;
		    indArrayInput++;
		  }
		else
		  {
		    InputCenters[NumEntry][indArrayInput]=Array[indArray];
		    InputCentersUpdatable[NumEntry][indArrayInput]=1;
		    InputCenters[NumEntry][indArrayInput+1]=Array[indArray+1];
		    InputCentersUpdatable[NumEntry][indArrayInput+1]=1;
		    indArray=indArray+2;;
		    indArrayInput=indArrayInput+2;
		  }}
	    else // output is not PFF
	      {
		NbPar=InputCentersNbPar[NumEntry][NumSef];
		for ( cpt2=0 ; cpt2 < NbPar ; cpt2++ )
		  {
		    InputCenters[NumEntry][indArrayInput+cpt2]=Array[indArray+cpt2];
		  };
		indArrayInput+=NbPar;
		indArray+=NbPar;
	      };
	  }//end MF is to be optimized
	  else//MF is not to be optimized
	    {
	      Cour=InitialNormedFis->In[NumEntry]->GetMF(NumSef);
	      NbPar=InputCentersNbPar[NumEntry][NumSef];
	      sef_par = new double[ NbPar];
	      Cour->GetParams(sef_par);
	      if ( PFFIn.Value[NumEntry] == 1 ) //entry is PFF
		{
		  if ( *pType == 1 )
		    {
		      InputCenters[NumEntry][indArrayInput]=sef_par[1];
		      InputCentersUpdatable[NumEntry][indArrayInput]=0;
		      indArrayInput++;
		    }


		  if ( ( *pType == 2 ) ||  ( *pType == 4 ))
		    {
		      InputCenters[NumEntry][indArrayInput]=sef_par[1];
		      InputCentersUpdatable[NumEntry][indArrayInput]=0;
		      indArrayInput++;
		      InputCenters[NumEntry][indArrayInput]=sef_par[2];
		      InputCentersUpdatable[NumEntry][indArrayInput]=0;
		      indArrayInput++;

		    }

		  if ( ( *pType == 3 ) )
		    {
		      InputCenters[NumEntry][indArrayInput]=sef_par[0];
		      InputCentersUpdatable[NumEntry][indArrayInput]=0;
		      indArrayInput++;
		      InputCenters[NumEntry][indArrayInput]=sef_par[1];
		      InputCentersUpdatable[NumEntry][indArrayInput]=0;
		      indArrayInput++;
		    }
		};
	      delete [] sef_par;

	    };
	  pType++;
	  indKey++;
	  ind++;
	}//end loop on output MFs
    

      if  ( PFFOut == 1 ){ //fuzzy output is PFF
	if ( CSort( indArrayInput, nbEntree ) == -1 )
	  return -1;
      }else{ //fuzzy output is not PFF
	if ( CSortCenter( indArrayInput, nbEntree ) == -1 )
	  return -1;
      }
      if ( CDomain(  indArrayInput, nbEntree ) == -1 )  //impose domain range
	return -1;
      if (CNearActivate == 1 ) { if ( CNear(  indArrayInput, nbEntree ) == -1 )  return -1; }
    }//end fuzzy output
  
  if ( IsInput == 2 )//rules
    {
      for ( int Cpt = 0 ; Cpt < NbRule ; Cpt++ )
	{
          if  ( InitialNormedFis->Rule[Cpt]->IsActive() == false ) continue;
          if ( Key.Value[indKey] == 1 )
	    {
	      InputCenters[nbEntree+1][indArrayInput]=Array[indArray];
	      indArray++;
	    }
          indKey++;
          indArrayInput++;
	}
      if ( Name != PERMUTATION ) { if ( CDomain(  indArrayInput , nbEntree+1 ) == -1 ) return -1; }
    }//end rules
  
  return 1;
}//end testConstraints

// new cSort that implements near constraint as a real value (class member)
// called only for PFF entries
// we check that 
int  sifopt::CSort(  int indArrayInput, int nbE )
// indArrayInput is the number of MF centers including edges
// all centers must be in increasing order
{
  for (int cpt = 1;cpt < indArrayInput;cpt++)
    if (DBL_SUP(InputCenters[nbE][cpt-1],InputCenters[nbE][cpt]))
      return -1;
  // centers must be separated at least by nearConstraint except for trapezoid kernel bounds
  for (int cpt = 1; cpt < (indArrayInput-2);cpt +=2)
    {
      if (DBL_SUP((InputCenters[nbE][cpt] + NearConstraint),InputCenters[nbE][cpt+1]))
	return -1;
    }
  
 return 1;
}

//force all values between 0 and 1
// force first element=0, last one=1
int  sifopt::CDomain( int indArrayInput, int nbE )    
{
  InputCenters[nbE][0]=0.0;//left edge
  InputCenters[nbE][indArrayInput-1]=1.0;//right edge

  for ( int cpt = 1 ; cpt < (indArrayInput-1) ; cpt ++ )
    {
      if ( InputCenters[nbE][cpt] <  EPSILON)
	InputCenters[nbE][cpt]=0.0;
      if ( InputCenters[nbE][cpt] > 1.0)
	InputCenters[nbE][cpt]=1.0;
    }
  
  return 1;
}

//called for non FPP entries
int  sifopt::CSortCenter(  int indArrayInput, int nbE )
{
  int cptMF=1;
  int cpt=1+InputCentersNbPar[nbE][0];
  double prev=InputCenters[nbE][1];
  while ( cpt < indArrayInput ){
    if ( InputCenters[nbE][cpt] < prev   )
      {
	return -1;
      }
    else
      {
	prev=InputCenters[nbE][cpt];
	cpt += InputCentersNbPar[nbE][cptMF];
	cptMF++;
      }
  }
  return 1;
 
}
