//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************
/*!
 * \file optkeyutil.cpp
 * \brief This file implements methods to define the vkey instances.the file testoptkey.cpp is a simple sample that show U how to manipulate keys.
 * \remark nIn and nOut params is the indice of a C array of FIS class, as the FIS member function do U should substract 1 to the naturall language number.
 * 
 */
#include "sifopt.h"
/*! \fn  void sifopt::setKeyInit( char *value )
 *  \brief This function transmit the Key in a char * form by argv to the Key constructor and store the result in a fisopt class.
 *  \param value the key written in an allocated char * 
 *
 *  This function transmit the Key in a char * form by argv to the Key constructor and store the result in a fisopt class.
 */
void sifopt::setKeyInit( char *value ){
  InitKey.setValue(value) ;
}

/*! \fn  void sifopt::setKeyInit( vkey *pKey )
 *  \brief This function transmit the Key built with vkey instance by argv store the result in a fisopt class.
 *  \param pkey a build key instance.
 *
 *  This function transmit the Key built with vkey instance by argv store the result in a fisopt class.
 */
void sifopt::setKeyInit( vkey *pKey ){
  InitKey.setValue(pKey) ;
}
/*! \fn    void sifopt::keyallopt( int NOut , FIS * F )
 *  \brief This function generate a Full Key : yes to all, note that the function will allocate the key for u to perform logical operation whith the existing one's, this mean U should delete it further...
 *  \param NOut The number of the output to optimize.
 *  \param F a pointer to a FIS instance
 *  \return a pointer to an instance of vkey ( this function allocate it ).
 */
vkey  * sifopt::keyallopt( int NOut , FIS * F )
   {
    vkey *myKey=new vkey;
    int cpt=0;
    for ( int NumEntry = 0 ; NumEntry < F->GetNbIn(); NumEntry ++ )
    {
      if  ( F->In[NumEntry]->IsActive() == false )  continue;
      int nbSef=F->In[NumEntry]->GetNbMf();
      for ( int NumSef = 0 ; NumSef < nbSef ; NumSef ++ )
        {
        myKey->Value[cpt] = 1;
        cpt++;
        }; 
      // tell if the partition is "PFF"
        myKey->Value[cpt] = 1;
        cpt++;
    };
    // output Nout type
    if( strcmp( F->Out[NOut]->GetOutputType(), OUT_FUZZY::OutputType() ) ==0 )
      {
        // output fuzzy the next bit classification is read in the FIS the next one not used
        // great serge specification
        myKey->Value[cpt] = 1;cpt++;
        if ( F->Out[NOut]->Classification() == 0 )
            {
            myKey->Value[cpt] = 1;  cpt++;   
            myKey->Value[cpt] = 1;  cpt++;  // not used 
            for ( int Cpt = 0 ; Cpt < F->Out[NOut]->GetNbMf() ; Cpt++ )
                {
                myKey->Value[cpt] = 1;
                cpt++;
                }
            }
        else
            {
            myKey->Value[cpt] = 0;  cpt++;   
            myKey->Value[cpt] = 1;  cpt++;  // not used 
            };
      }
    else
      {
        myKey->Value[cpt] = 0;  cpt++;
        myKey->Value[cpt] = 1;  cpt++;  // Vocabularu read
        myKey->Value[cpt] = 1;  cpt++;  // not used 
      }
    
   // all rules to 1
   for ( int Cpt = 0 ; Cpt < F->GetNbRule() ; Cpt++ )
       {
       myKey->Value[cpt] = 1;
       cpt++;
       }
#ifdef DEBUG
   printf("\n longueur cles %i\n",cpt);
   myKey->Print(cpt-1);
#endif
   return myKey;
   }
/*! \fn    vkey * sifopt::keysetIN( FIS * F , int nIn , int indPFF )
 *  \brief This function generates a Key that will optimize the nIn'th in, note that the function will allocate the key for u to perform logical operation whith the existing one's, this mean U should delete it further...
 *  \param F a pointer to a FIS instance
 *  \param nIn The number of the In.
 *  \param indPFF Tell if the partition is PFF
 *  \return a pointer to an instance of vkey ( this function allocates it ).
 */
vkey  * sifopt::keysetIN( FIS * F , int nIn , int indPFF )
   {
    vkey *myKey=new vkey;
    int cpt=0;
    for ( int NumEntry = 0 ; NumEntry < F->GetNbIn(); NumEntry ++ )
    {
      if  ( F->In[NumEntry]->IsActive() == false )  continue;
      int nbSef=F->In[NumEntry]->GetNbMf();
      if ( nIn != NumEntry )
      {
      // tell if the partition is "PFF"
        myKey->Value[cpt] = 0;
        cpt++;
        for ( int NumSef = 0 ; NumSef < nbSef ; NumSef ++ )
            {
            myKey->Value[cpt] = 0;
            cpt++;
            }; 
      }
      else 
      {
      // tell if the partition is "PFF"
         myKey->Value[cpt] = indPFF;
         cpt++;
         for ( int NumSef = 0 ; NumSef < nbSef ; NumSef ++ )
             {
             myKey->Value[cpt] = 1;
             cpt++;
             };
      }; 
    };
   return myKey;
}
/*! \fn    vkey * sifopt::keysetMFIN( FIS * F , int nIn , int indPFF , int nMF )
 *  \brief This function generate a Key that will optimize the nMF 'th MF in the nIn ' th in, note that the function will allocate the key for u to perform logical operation whith the existing one's, this mean U should delete it further...
 *  \param F a pointer to a FIS instance
 *  \param nIn The number of the In.
 *  \param indPFF Tell if the partition is PFF
 *  \param nMF The number of the MF in the In.
 *  \return a pointer to an instance of vkey ( this function allocate it ).
 */
vkey  * sifopt::keysetMFIN( FIS * F , int nIn , int indPFF , int nMF)
   {
    vkey *myKey=new vkey;
    int cpt=0;
    for ( int NumEntry = 0 ; NumEntry < F->GetNbIn(); NumEntry ++ )
    {
      if  ( F->In[NumEntry]->IsActive() == false )  continue;
      int nbSef=F->In[NumEntry]->GetNbMf();
      if ( nIn != NumEntry )
      {
      // tell if the partition is "PFF"
        myKey->Value[cpt] = 0;
        cpt++;
        for ( int NumSef = 0 ; NumSef < nbSef ; NumSef ++ )
            {
            myKey->Value[cpt] = 0;
            cpt++;
            }; 
      }
      else 
      {
      // tell if the partition is "PFF"
         myKey->Value[cpt] = indPFF;
         cpt++;
         for ( int NumSef = 0 ; NumSef < nbSef ; NumSef ++ )
             {
             myKey->Value[cpt] = 0;
             if ( NumSef == nMF ) myKey->Value[cpt] = 1;
             cpt++;
             };
      }; 
    };
   return myKey;
}
/*! \fn    vkey * sifopt::keysetOUT( FIS * F , int nOut , int indPFF )
 *  \brief This function generate a Key that will optimize the nOut ' th out, note that the function will allocate the key for u to perform logical operation whith the existing one's, this mean U should delete it further...
 *  \param F a pointer to a FIS instance
 *  \param NOut The number of the Out.
 *  \param indPFF Tell if the partition is PFF
 *  \return a pointer to an instance of vkey ( this function allocate it ).
 */
vkey  * sifopt::keysetOUT( FIS * F , int nOut , int indPFF )
   {
     // does not work for multiple outputs
    vkey *myKey=new vkey;
    int cpt=0;
    for ( int NumEntry = 0 ; NumEntry < F->GetNbIn(); NumEntry ++ )
    {
      if  ( F->In[NumEntry]->IsActive() == false )  continue;
      int nbSef=F->In[NumEntry]->GetNbMf();
      // tell if the partition is "PFF"
        myKey->Value[cpt] = 0;
        cpt++;
      for ( int NumSef = 0 ; NumSef < nbSef ; NumSef ++ )
        {
        myKey->Value[cpt] = 0;
        cpt++;
        }; 
    };
    // output Nout type
    if( strcmp( F->Out[nOut]->GetOutputType(), OUT_FUZZY::OutputType() ) ==0 )
      {
        // output fuzzy the next bit classification is read in the FIS the next one not used
        // great serge specification
        myKey->Value[cpt] = 1;cpt++;
	// modif bch - July 2010
        if ( F->Out[nOut]->Classification() == 0 )
            {
            myKey->Value[cpt] = 0;  cpt++;  
	    }
	else
	  {
	    myKey->Value[cpt] = 1;  cpt++; 
	  }
            myKey->Value[cpt] = 1;  cpt++;  // not used 
            for ( int Cpt = 0 ; Cpt < F->Out[nOut]->GetNbMf() ; Cpt++ )
                {
                myKey->Value[cpt] = 1;
                cpt++;
                }
	    // modif bch - July 2010
    /*else
            {
            myKey->Value[cpt] = 0;  cpt++;   
            myKey->Value[cpt] = 1;  cpt++;  // not used 
            };
    */
      
      }
    
    else
      {
        myKey->Value[cpt] = 0;  cpt++;
        myKey->Value[cpt] = 1;  cpt++;  // Vocabulary read
        myKey->Value[cpt] = 1;  cpt++;  // not used 
      }
   return myKey;
   }
/*! \fn    vkey * sifopt::keysetMFOUT( FIS * F , int nOut , int indPFF, int nMF )
 *  \brief This function generate a Key that will optimize the nOut ' th out, note that the function will allocate the key for u to perform logical operation whith the existing one's, this mean U should delete it further...
 *  \param F a pointer to a FIS instance
 *  \param NOut The number of the Out.
 *  \param indPFF Tell if the partition is PFF
 *  \param nMF tell the MF number of the output to optimize
 *  \return a pointer to an instance of vkey ( this function allocate it ).
 */
vkey  * sifopt::keysetMFOUT( FIS * F , int nOut , int indPFF, int nMF )
   {
    vkey *myKey=new vkey;
    int cpt=0;
    for ( int NumEntry = 0 ; NumEntry < F->GetNbIn(); NumEntry ++ )
    {
      if  ( F->In[NumEntry]->IsActive() == false )  continue;
      int nbSef=F->In[NumEntry]->GetNbMf();
      // tell if the partition is "PFF"
        myKey->Value[cpt] = 0;
        cpt++;
      for ( int NumSef = 0 ; NumSef < nbSef ; NumSef ++ )
        {
        myKey->Value[cpt] = 0;
        cpt++;
        }; 
    };
    // output Nout type
    if( strcmp( F->Out[nOut]->GetOutputType(), OUT_FUZZY::OutputType() ) ==0 )
      {
        // output fuzzy the next bit classification is read in the FIS the next one not used
        // great serge specification
        myKey->Value[cpt] = 1;cpt++;
        if ( F->Out[nOut]->Classification() == 0 )
            {
            myKey->Value[cpt] = 1;  cpt++;   
            myKey->Value[cpt] = 1;  cpt++;  // not used 
            for ( int Cpt = 0 ; Cpt < F->Out[nOut]->GetNbMf() ; Cpt++ )
                {
                myKey->Value[cpt] = 0;
                if ( Cpt == nMF ) myKey->Value[cpt] = 1;
                cpt++;
                }
            }
        else
            {
            myKey->Value[cpt] = 0;  cpt++;   
            myKey->Value[cpt] = 1;  cpt++;  // not used 
            };
      }
    else
      {
        myKey->Value[cpt] = 0;  cpt++;
        myKey->Value[cpt] = 1;  cpt++;  // Vocabularu read
        myKey->Value[cpt] = 1;  cpt++;  // not used 
      }
   return myKey;
   }
/*! \fn    vkey  * sifopt::keysetRULE( FIS * F , int nOut, int nRule)
 *  \brief This function generate a Key that will optimize the nRule ' th Rule, note that the function will allocate the key for u to perform logical operation whith the existing one's, this mean U should delete it further...
 *  \param F a pointer to a FIS instance
 *  \param nOut The number of the Output that will be optimize.
 *  \param nRule The number of the Rule.
 *  \return a pointer to an instance of vkey ( this function allocate it ).
 */
vkey  * sifopt::keysetRULE( FIS * F , int nOut , int nRule )
   {
     //nRule=-1:  optimize all rules
     // otherwise optimize one rule (the nRule th one)
    vkey *myKey=new vkey;
    int cpt=0;
    for ( int NumEntry = 0 ; NumEntry < F->GetNbIn(); NumEntry ++ )
    {
      if  ( F->In[NumEntry]->IsActive() == false )  continue;
      int nbSef=F->In[NumEntry]->GetNbMf();
      for ( int NumSef = 0 ; NumSef < nbSef ; NumSef ++ )
        {
        myKey->Value[cpt] = 0;
        cpt++;
        }; 
      // tell if the partition is "PFF"
        myKey->Value[cpt] = 0;
        cpt++;
    };
    // output Nout type
    if( strcmp( F->Out[nOut]->GetOutputType(), OUT_FUZZY::OutputType() ) ==0 )
      {
        // output fuzzy the next bit classification is read in the FIS the next one not used
        // great serge specification
        myKey->Value[cpt] = 0;cpt++;
        if ( F->Out[nOut]->Classification() == 0 )
            {
            myKey->Value[cpt] = 0;  cpt++;   
            myKey->Value[cpt] = 0;  cpt++;  // not used 
            for ( int Cpt = 0 ; Cpt < F->Out[nOut]->GetNbMf() ; Cpt++ )
                {
                myKey->Value[cpt] = 0;
                cpt++;
                }
            }
        else
            {
            myKey->Value[cpt] = 0;  cpt++;   
            myKey->Value[cpt] = 0;  cpt++;  // not used 
            };
      }
    else
      {
        myKey->Value[cpt] = 0;  cpt++;
        myKey->Value[cpt] = 0;  cpt++;  // Vocabulary read
        myKey->Value[cpt] = 0;  cpt++;  // not used 
      }
    
   // nRule 'th rule to 1 or all rules to 1 if nRule=-1
   for ( int Cpt = 0 ; Cpt < F->GetNbRule() ; Cpt++ )
       {
       if (nRule == -1)
	 myKey->Value[cpt] = 1 ;
       else
	 {
	   myKey->Value[cpt] = 0;
	   if ( Cpt == nRule ) myKey->Value[cpt] =1;
	 }
       cpt++;
       }
#ifdef DEBUG
   printf("\n longueur cles %i\n",cpt);
   myKey->Print(cpt-1);
#endif
   return myKey;
   }
