//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************
/*! \file  optotd.cpp
 *  \brief This File contents the definition of the Optimizer member function ObjectToDouble.

    In order to applicate an algorithm on an object, U must first determinate what parameters of this object U will change. This is the role of this function, whith the key and the Object, it constructs and set of values stored as an C array of double. This function switches the variable IsInput which values are 1 for ins 0 for outs and 3 for rules. 
 */
/*!
\fn void sifopt::ObjectToDouble( void * Item, vkey * pKey, double *Array )
\brief This function extract the set of values U want to update in your object ( in the aim to transmit it to the algo class. ( who don t know what a FIS is ... ))
\param Item The adress of Your object
\param pkey A quiet complicated boolean string to tell what part of the FIS to update
\param Array set of double to fill
\return nothing the result is store in the 3rd param.
 */

#include "sifopt.h"
void sifopt::ObjectToDouble( void * Item, vkey * pKey, double *Array )
{
  FIS  *S=( FIS * )Item;
  int *pType=NULL;
  int  indArray=0,nbSef=0,indKey=0,indPFF=0,NbParam,cptE=0;
  double *sef_par=NULL;
  MF *Cour=NULL;
// 3 cases : input (IsInput=1), output (IsInput=0)  if first one is Fuzzy, rules (IsInput=2)
// Updates Array which has been allocated before
if ( IsInput == 1 )
 {

  for ( int NumEntry = 0 ; NumEntry < nbEntree ; NumEntry ++ )
      {
      if  ( S->In[NumEntry]->IsActive() == false ) continue;
      indPFF = PFFIn.Value[ cptE ] ;
      cptE++;
#ifdef DEBUG
printf ("\n\n+OTD entree %i indPFF %i : ", NumEntry,indPFF);
#endif
      nbSef=InputNbMf[NumEntry];
      pType=InputCentersType[NumEntry];
       for ( int NumSef = 0 ; NumSef < nbSef ; NumSef ++ )
          {
          if ( Key.Value[indKey] == 1 )
             {
             Cour=S->In[NumEntry]->GetMF(NumSef);
             NbParam=Cour->NbParams();
             sef_par = new double[NbParam];
             Cour->GetParams(sef_par);
             if ( indPFF == 1 )
                {
#ifdef DEBUG
printf (" PFF ");
#endif
// 4 a triangle just the center
 if   (  *pType == 1 )
 {
                Array[indArray]=sef_par[1];
#ifdef DEBUG
printf (" TRIANGLE  ( %f  %i ), ",Array[indArray],indArray);
#endif
                indArray++;
 }
// a trapezoid or an trapesoid sup just the center second and third parameter
 if   ( (  *pType == 2 ) || (  *pType == 4 ))
 {
                Array[indArray]=sef_par[1];
#ifdef DEBUG
printf (" TRAP OR TRAPSUP 1val %f indArray %i,  ",Array[indArray],indArray);
#endif
                indArray++;
                Array[indArray]=sef_par[2];
#ifdef DEBUG
printf ("  2 ( %i ) val %f indArray %i,  ",*pType,Array[indArray],indArray);
#endif
                indArray++;
 }
                if  (  *pType == 3 )
 {
                Array[indArray]=sef_par[0];
#ifdef DEBUG
printf (" TRAPINF 1val %f indArray %i,  ",Array[indArray],indArray);
#endif
                indArray++;
                Array[indArray]=sef_par[1];
#ifdef DEBUG
printf ("  2 ( %i ) val %f indArray %i,  ",*pType,Array[indArray],indArray);
#endif
                indArray++;
 }
                 }
              else // not PFF
                 {
#ifdef DEBUG
printf (" NOTPFF NbParam %i  ", NbParam);
#endif
                  for ( int cpt=0 ; cpt < NbParam ; cpt++ ) 
                       {
                       Array[indArray]=sef_par[cpt];
#ifdef DEBUG
printf (" valeur %f indice %i   ",Array[indArray],indArray);
#endif
                       indArray++;
                       };
              }; //endif PFF
                delete [] sef_par;
              }; // endif Key
         indKey++;
         pType++;
       }; // end for MF
      }; // end for entry
 }
// fuzzy output case
if ( IsInput == 0 )
 {
// read the key once to known PFF flag
      indPFF = PFFOut;
#ifdef DEBUG
printf ("\n+OTD sortie indPFF %i", indPFF);
#endif
      pType=InputCentersType[nbEntree];
      for ( int NumSef = 0 ; NumSef < OutputNbMf ; NumSef ++ )
          {
          if ( Key.Value[indKey] == 1 )
             {
             Cour=S->Out[numOutput]->GetMF(NumSef);
             NbParam=Cour->NbParams();
             sef_par = new double[NbParam];
             Cour->GetParams(sef_par);
             if ( indPFF == 1 )
     { 
#ifdef DEBUG
printf (" PFF ");
#endif
// 4 a triangle just the center
 if   (  *pType == 1 )
 {
                Array[indArray]=sef_par[1];
#ifdef DEBUG
printf (" TRIANGLE  ( %f  %i ), ",Array[indArray],indArray);
#endif
                indArray++;
 }
// a trapezoid or an trapesoid sup just the center second and third parameter
 if   ( (  *pType == 2 ) || (  *pType == 4 ))
 {
                Array[indArray]=sef_par[1];
#ifdef DEBUG
printf (" TRAP OR TRAPSUP 1val %f indArray %i,  ",Array[indArray],indArray);
#endif
                indArray++;
                Array[indArray]=sef_par[2];
#ifdef DEBUG
printf ("  2 ( %i ) val %f indArray %i,  ",*pType,Array[indArray],indArray);
#endif
                indArray++;
 }
                if  (  *pType == 3 )
 {
                Array[indArray]=sef_par[0];
#ifdef DEBUG
printf (" TRAPINF 1val %f indArray %i,  ",Array[indArray],indArray);
#endif
                indArray++;
                Array[indArray]=sef_par[1];
#ifdef DEBUG
printf ("  2 ( %i ) val %f indArray %i,  ",*pType,Array[indArray],indArray);
#endif
                indArray++;
 }
                 }
              else // not PFF
                 {
#ifdef DEBUG
printf (" NOTPFF NbParam %i  ", NbParam);
#endif
                  for ( int cpt=0 ; cpt < NbParam ; cpt++ ) 
                       {
                       Array[indArray]=sef_par[cpt];
#ifdef DEBUG
printf (" valeur %f indice %i   ",Array[indArray],indArray);
#endif
                       indArray++;
                   };
                 };//endif PFF
                delete [] sef_par;
              }; // endif Key
         indKey++;
         pType++;
     }; // end MF
 }; // end if output

// rule case
if ( IsInput == 2 )
 {
#ifdef DEBUG
printf ("\n+OTD regles ");
#endif
  for ( int Cpt = 0 ; Cpt < S->GetNbRule() ; Cpt++ )
    {
      if  ( S->Rule[Cpt]->IsActive() == false ) continue;
          if ( Key.Value[indKey] == 1 )
             {
             Array[indArray]=S->Rule[Cpt]->GetAConc(numOutput);
#ifdef DEBUG
printf (" Valeur de la conclusion  %f indice  %i",Array[indArray],indArray);
#endif
             indArray++;
             }
             indKey++;
    }
 }; // end if rules 


#ifdef DEBUG
printf ("\n+Fin OTD : %i  ",indArray);
for ( int i = 0 ; i< indArray ; i++ ) printf("(  %f , %i ) ",Array[i], i );
printf ("\n");
#endif
  return ;
}
