//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************
#include "algo.h"
/*!
 * \file permutation.cpp
 * \brief defines a permutation algorythm used to optimize rules' head when the vocabulary is limited.

 This function defines a permutation algorythm used to optimize rules' head when the vocabulary is limited, or when it's the number of an MF. This algo code is a particular case ( use of an possible value array ( possibles or MFs )), you must rather consider Hill climbing ( fib.cpp ) as a model to writte an algo like Sollis and Wetts ( sw.cpp ).
 */
int Permutation(
           void     * YourObject,
           vkey     * Key,
           double   * ArrayOfDouble ,
           int        SizeArray,
           algoPar  * P,
           double   * gain,
           double  (* pfEval)( void   *  ,
                   vkey    *  ,
                   double *  ,
                   int       ,
                   conteneur *),
           conteneur * fEvalArgs
           )
{
  long int  SWite=0;  // The number of iteration
  int       KeySize=0;  // Number of true set in the key
  double    SWE;      //  Current evaluation
  double    SWbestE; //   Best evaluation of E
  double    SWprevE;  // Previous evaluation of E
  double    SWfirstE; // First Evaluation of E

  //warning !!!! if new arguments increase get(9), get(10) !!!!!!!!!!!!!
  int Card=*((int * )fEvalArgs->get(9));
  double *Domaine=(double * )fEvalArgs->get(10);
  int tmpint;

#ifdef DEBUGS
  printf("\n+Permute cardinal %i",Card );
  for ( int i=0; i<  Card ; i++ ) printf(" ( %i,%f ) ",i, Domaine[i]);
#endif

  // A little check
  KeySize=Key->Size();
  if ( KeySize == 0 ) return -1;
///////////////////////////////////////////////
// Init
////////
// Put input datas in a vector object VectDouble
avect *VectDouble=new avect( SizeArray , ArrayOfDouble );
#ifdef DEBUG
  printf("\n+Permutation recoit ");VectDouble->print();
  printf("\n+Permutation longeur vecteur %i\n" , KeySize );
  printf("\n+Permutation SizeArray %i\n" , SizeArray );
  VectDouble->print();
#endif
  SWfirstE=(*pfEval)( YourObject ,
              Key ,
              VectDouble->getArrayAdress() ,
              VectDouble->getNbElem()      ,
              fEvalArgs);
  //printf("\nfirst evaluation in permutation SWfirstE=%f\n",SWfirstE);
  SWbestE=SWfirstE;

  avect *T = new avect(SizeArray);
  avect *Random= new avect(SizeArray);
  avect *BestVect= new avect(SizeArray);
  
  BestVect->dupVector(VectDouble);
#ifdef DEBUG
  printf("\n+SW vect Depart \n");
  VectDouble->print();
#endif
// let's get in the while whith : previous correct and worse than current
  SWprevE=SWfirstE+1;

#ifdef DEBUG
  printf("\n+SWInitial Evaluation %f ",SWfirstE );
#endif
///////////////////////////////////////////////
// Main Loop
////////
  while (  SWite < P->maxIte ) 
    {
#ifdef DEBUG
printf("\n+SWIteration %li ", SWite );
#endif
//////////////////
// get a random vector of possibles values ( MF 4 or fixed 1.2 conc )
  for ( int i=0 ; i< SizeArray ; i++ ) 
  { tmpint = ( int )floor ( FisRand() * Card ); 
  Random->setValue(i,Domaine[  tmpint  ]);
  }
      SWE=(*pfEval)( YourObject , Key , Random->getArrayAdress() , Random->getNbElem() , fEvalArgs);

#ifdef DEBUG
printf("\n+SWEval : %f | ",SWE);
printf("bestE : %f | ",SWbestE);
printf("SWprevE : %f",SWprevE);
#endif

      if ( ( SWE <= SWbestE ) && ( fabs ( SWbestE - SWE ) > EPSILON )) 
        {
          VectDouble->dupVector(Random);
#ifdef DEBUGTERMFAST
          printf("\n+Permute Meilleure performance -- %f \n",SWE);
#endif
#ifdef DEBUG
          printf("\n+SW Nouveau Vecteur");
          VectDouble->print();
#endif
          SWprevE=SWE;
          if ( SWE > -EPSILON ) { SWbestE=SWE;  BestVect->dupVector(VectDouble); }
        }
      SWite++;
    };
///////////////////////////////////////////////
// End Loop
////////
#ifdef DEBUG
  printf("\n+Permute Fin au bout de %ld itrations ",SWite);
  printf("\n+Permute Distance au but de %f vers %f",SWfirstE,SWbestE);
  printf("\n+Permute Parametre "); P->Print();
  printf("\n+Permute Vecteur " ); VectDouble->print();
  printf("\n+Permute FIN Performence finale %f \n",SWbestE );
#endif
  if (  SWfirstE != SWbestE )
  *gain=(*pfEval)( YourObject ,
              Key ,
              BestVect->getArrayAdress() ,
              BestVect->getNbElem()      ,
              fEvalArgs);
#ifdef DEBUG
  printf("\n+Permute gain vaut %f \n",SWbestE );
#endif
  //  free memory
  delete Random;
  delete BestVect;
  delete T;
  delete VectDouble;
#ifdef DEBUG
  printf("\n+Permute Fin\n\n");
#endif

  return 0;
}
