//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************
/*! \file fisopt.cpp
 *  \brief This File contents the delaration of the Fuzzy Inference System Optimizer.
 *
 This File contents the delaration of the Optimizer constructor and destructor. The init and lauchentry member functions. There are also two Evaluation function : one that does'nt belong to FIS class ( generic case ) who wrapps a member evaluation function
 *
 */
#include "sifopt.h"
class sifopt;
class FIS;
//! Wrap a class member evaluation function in a global function. This algorythm language works whith any Evaluation function and Datas, U will have to wrap your Evaluation function if it's a member function one.
// warning !!!! if more arguments increase get(8) !!!!!!!!!!!!!!
double FisEval( void *  Item  , vkey * pKey , double *Array ,
		int sizeArray , conteneur * fEvalArgs )
{
  sifopt *pthis    = ( sifopt * )fEvalArgs->get(8);
#ifdef DEBUG
  printf("\n %i \n",sizeArray);
#endif
  return pthis->Eval( Item  , pKey , Array , sizeArray );
}
//! unused constructor
sifopt::sifopt() : algo ( 1 , 1 )
{
  InitialFis = NULL;
  tmpFis = NULL;
  Solution = NULL;
  MfType= NULL;
  IsInited  =0;
  IsInput  =1;
  FuzzyOutput =1;
  // Added by SG - March 16th, 2010
  InitCov = 0.;
}
/*! sifopt::sifopt ( FIS     * pS  ,int ** pMissClass , double **pDouble,int nbrow ,int Sortie )
 * \param pS The adress of a fis instance
 * \param pMissClass a pointer to the Miss Classed array
 * \param pDouble a pointer to the Samples Data
 * \param nbrow the number of samples
 * \param Sortie the number of the output to optimize
 * \return an sifopt instance adress ( as a constructor usualy do )
 This construcor calls algo constructor with the following params an Solis Wetts method ( 1 ) an size of ( 20 ) for conteneur. It read the Fis data and create his own data.
*/
sifopt::sifopt( FIS     * pS  ,int *& pMissClass , double **pDouble,int nbrow ,int Sortie, double nearC) : algo ( 1, 1 , 20 )
{
  int ret=0;

  NearConstraint=nearC;//added Feb 2013 to implement near constraint between MF centers
  Reponse=0;
  pResClassif=&pMissClass;
  Exemples=pDouble;
  nbRow=nbrow;
  numOutput= Sortie;
  InitialFis   = new FIS (*pS);
  // tri2trap transformation - January 2013
  for (int i=0;i<InitialFis->GetNbIn();i++)
    InitialFis->In[i]->Tri2Trap();
  if (! strcmp(InitialFis->Out[numOutput]->GetOutputType(), OUT_FUZZY::OutputType()))
    InitialFis->Out[numOutput]->Tri2Trap();

  ret= InitialFis->ClassCheckNoAllocResClassif(Exemples,nbRow,numOutput );
#ifdef DEBUG
  printf ("\n ClassCheckNoAllocResClassif returns %i",ret);
  InitialFis->PrintCfg();
#endif
  InitialNormedFis = new FIS (*pS);
  // tri2trap transformation - January 2013
  for (int i=0;i<InitialNormedFis->GetNbIn();i++)
    InitialNormedFis->In[i]->Tri2Trap();
  if (! strcmp(InitialNormedFis->Out[numOutput]->GetOutputType(), OUT_FUZZY::OutputType()))
    InitialNormedFis->Out[numOutput]->Tri2Trap();
  ret= InitialNormedFis->ClassCheckNoAllocResClassif(Exemples,nbRow,numOutput );

  // Added by SG - March 16th, 2010
  InitCov = 0.;
  // Can't be computed at this step because many of the parameters are 
  // set after the object construction. To be updated in the next version.

  // TMP Temporaly here to check ( restore ) base class fis coherence
#ifdef DEBUGFIS
  printf ("\n ClassCheckNoAllocResClassif returns %i",ret);
#endif
  // TMP end

  tmpFis       = NULL;
  nbEntree=InitialFis->GetNbIn();
  MfType = NULL;
  Solution  = NULL;
  IsInited  = 0;
  CNearActivate = 0;
  CNearInitIn  = 0;
  CNearInitOut = 0;
  IsInput     =1;
  FuzzyOutput =1;
  SumCentersLenght=0;
  InputCenters = new double *[nbEntree+2];
  InputCentersMax = new double *[nbEntree+2];
  InputCentersMin = new double *[nbEntree+2];
  InputCentersUpdatable = new int *[nbEntree+1];
  InputCentersType          = new int *[nbEntree+1];
  InputCentersNbPar         = new int *[nbEntree+1];
  InputCentersLenght        = new int [nbEntree+1];
  InputNbMf                 = new int [nbEntree+1];
  for ( int cpt=0 ; cpt < nbEntree+1 ; cpt++ )
    {
      InputCenters[cpt] = NULL;
      InputCentersMin[cpt] = NULL;
      InputCentersMax[cpt] = NULL;
      InputCentersUpdatable[cpt] = NULL;
      InputCentersType[cpt] = NULL;
      InputCentersNbPar[cpt] = NULL;
    }
  InputCenters[nbEntree+1] = NULL;
}

sifopt::~sifopt(){
  if ( InitialFis != NULL ) delete InitialFis;
  if ( InitialNormedFis != NULL ) delete InitialNormedFis;
  if ( tmpFis != NULL ) delete tmpFis;
  if ( Solution != NULL ) delete Solution;
  if ( MfType != NULL ) delete [] MfType;
  for ( int cpt=0 ; cpt < nbEntree+1 ; cpt++ )
    {
      if ( InputCenters[cpt] != NULL) delete [] InputCenters[cpt];
      if ( InputCentersMax[cpt] != NULL) delete [] InputCentersMax[cpt];
      if ( InputCentersMin[cpt] != NULL) delete [] InputCentersMin[cpt];
      if ( InputCentersUpdatable[cpt] != NULL) delete [] InputCentersUpdatable[cpt];
      if ( InputCentersType[cpt] != NULL) delete [] InputCentersType[cpt];
      if ( InputCentersNbPar[cpt] != NULL) delete [] InputCentersNbPar[cpt];
    }
  if (InputCenters[nbEntree+1]!= NULL) delete [] InputCenters[nbEntree+1];
  delete [] InputCenters;
  delete [] InputCentersMax;
  delete [] InputCentersMin;
  delete [] InputCentersUpdatable;
  delete [] InputCentersType;
  delete [] InputCentersNbPar;
  delete [] InputCentersLenght;
  delete [] InputNbMf;
}



/*!
 * \fn   int sifopt::launchEntry(double &gain)
 * \brief This function launches the optimisation of the FIS.
 *
 * This function launches the optimisation of the FIS. First we treat the inputs, than the output ( if it's a fuzzy one ) and the rules with the same algo ( Solis and Wetts or Hill climbing ) in the general case, or with an algo of permutation for limited vocabulary to a crisp's head rule or Fuzzy head rule. The inner Data function 
 */
int sifopt::launchEntry(double &gain, bool wordless){
  ////////////////////////////////////////////////////////////
  //double gain    =-1;
  gain    =-1;
  double bestgain=INFINI;
  int        response,ret,OnePass=0,nbpossibles;
  double *TabDomaine;
  double *Possibles;
  if ( IsInited == 0 ) 
         IsInited=init();
   // init determines SumCentersLenght (input + fuzzy output) 
  // and KeyRuleLenght (rule conc. length)
  // it allocates once vector of doubles for running optim. algorithm (see initdata)
  // values of T array will be updated in ObjectToDouble and transferred to optim. by launch function
  // T is allocated here

  // error!modif bch March 2010
  //double *T= new double[SumCentersLenght];
  int maxLenght=max(SumCentersLenght,KeyRuleLenght);

  double *T= NULL;
  T=new double[maxLenght];
  //
  //
  //for ( int initT=0; initT< SumCentersLenght;initT++)T[initT]=-1;
  for ( int initT=0; initT< maxLenght;initT++)T[initT]=-1;
  MaxKey =1; OnePass=1;

#ifdef DEBUG
  printf("\n+A First FIS ");
  tmpFis->Print();
#endif
  
  ///////////////////////////////////////////////////////////////
  //  Optimisation of inputs
  ///////////////////////////////////////////////////////////////
     
  ObjectToDouble( tmpFis, &Key , T );
  response=launch(tmpFis,T,&gain,FisEval);
   
 #ifdef DEBUG
  printf("\n\n\n\n\n");
  printf("\nLaunch IN's reponse : %i \nGain %f\n",response,gain);
  printf("\n\n\n\n\n");
#endif
  
  // Is there a better FIS 
  if (( gain < bestgain ) && ( gain > -EPSILON ) )
    {
      Reponse=1;
      if ( Solution != NULL ) delete Solution;
      Solution =new FIS (*tmpFis);
      // TMP Temporaly here to check ( restore ) base class fis coherence
      ret= Solution->ClassCheckNoAllocResClassif(Exemples,nbRow,numOutput );
#ifdef DEBUGFIS
      printf ("\n ClassCheckNoAllocResClassif returns %i",ret);
#endif
      // end TMP

      // Yes there is
      bestgain=gain;
      if (! wordless)
	printf("+ There is a better FIS ( ( Normed ) perf %f ) after the input optimisation  \n+",bestgain);
#ifdef DEBUG
      Solution->PrintCfg();
      printf ("\n");
#endif
    }
  else {
    // No there isn t
    Reponse=0;
    gain = -1; 
#ifdef DEBUG
    printf("+ There is no better FIS  after the input optimisation.\n" );
#endif
  };
  ///////////////////////////////////////////////////////////////
  //  Optimisation of outputs
  ///////////////////////////////////////////////////////////////
  // if the output we optimize is fuzzy, Go for it...
  if ( FuzzyOutput == 1) 
    {
      // The flag IsInput = 0 is used in Object to double and double to object functions
      IsInput = 0;

      // build a key with output size
      Key.setValue((unsigned long int)0);
      Key=KeyOut;
#ifdef DEBUG
      printf("\n\n+Output debut Output gain = %f\n+O Key ",bestgain);
      Key.Print(OutputNbMf+1);
#endif
      // To go on with outputs we get the better FIS or the initial FIS, it depends ...
      if ( Reponse == 1 )//there was a better FIS following input optimization
        {
	  if ( tmpFis != NULL ) delete tmpFis;
	  tmpFis =new FIS (*Solution);
	  // TMP Temporaly here to check ( restore ) base class fis coherence
	  ret= tmpFis->ClassCheckNoAllocResClassif(Exemples,nbRow,numOutput );
#ifdef DEBUGFIS
	  printf ("\n ClassCheckNoAllocResClassif returns %i",ret);
#endif
	  // end TMP 
        }
      else
        {
	  if ( tmpFis != NULL ) delete tmpFis; 
	  tmpFis = new FIS ( * InitialNormedFis );
	  // TMP Temporaly here to check ( restore ) base class fis coherence
	  ret= tmpFis->ClassCheckNoAllocResClassif(Exemples,nbRow,numOutput );
#ifdef DEBUG
	  printf ("\n ClassCheckNoAllocResClassif returns %i",ret);
#endif
	  // end TMP 
	};
      ObjectToDouble( tmpFis, &Key , T );
      response=launch(tmpFis,T,&gain,FisEval);//returns -1 if no optim done
#ifdef DEBUG
      printf("\n\n\n\n\n");
      printf("\nLauch OUT's reponse : %i \nGain %f\n",response,gain);
      printf("\n\n\n\n\n");
#endif
      // Is there a better FIS 
      if ( ( ( Reponse == 0 ) && ( gain > -EPSILON ) ) 
	   || ( ( Reponse == 1 ) && ( gain < bestgain -EPSILON ) && (  gain > -EPSILON ) ))
	{
	  Reponse=2;
	  bestgain=gain;
	  if ( Solution != NULL ) delete Solution;
	  Solution =new FIS (*tmpFis);

	  // TMP Temporaly here to check ( restore ) base class fis coherence
	  ret= Solution->ClassCheckNoAllocResClassif(Exemples,nbRow,numOutput );
#ifdef DEBUG
	  printf ("\n ClassCheckNoAllocResClassif returns %i",ret);
#endif
	  // end TMP 

	  if (!wordless) 
	    printf("+ There is a better FIS ( perf %f ) after the output optimisation ( Normed ) \n+",bestgain);
	}
      else {
	// No there isnt:  keep previous solution
	gain = -1;
	// modif August 2012 bch
	// restore gain (from input optimization)
	if (Reponse==1)
	  gain=bestgain;
#ifdef DEBUG
	printf("+ There is no better FIS  after the output optimisation.\n" );
#endif

      }
    }

  ///////////////////////////////////////////////////////////////
  //  Optimisation of rules
  ///////////////////////////////////////////////////////////////
  // If there is at least 1 bit set in the key, Go for it...
  if  ( KeyRuleLenght != 0 ) {
    // The flag IsInput = 2 is used in Object to double and double to object functions
    IsInput = 2;

    // build a key with output size
    Key.setValue((unsigned long int)0);
    Key=KeyRule;
#ifdef DEBUG
    printf("\n\n+Rule debut Output gain = %f\n+O Key ",bestgain);
    Key.Print(OutputNbMf+1);
#endif

    // To go on with rules we get the better FIS or the initial FIS, it depends ...
    if ( ( Reponse == 2 ) || ( Reponse == 1 ) )
      {
	if ( tmpFis != NULL ) delete tmpFis;
	tmpFis =new FIS (*Solution);
	// TMP Temporaly here to check ( restore ) base class fis coherence
	ret= tmpFis->ClassCheckNoAllocResClassif(Exemples,nbRow,numOutput );
#ifdef DEBUG
	printf ("\n ClassCheckNoAllocResClassif returns %i",ret);
#endif
	// end TMP
      }
    else
      {
	if ( tmpFis != NULL ) delete tmpFis;
	tmpFis = new FIS ( * InitialNormedFis );
	// TMP Temporaly here to check ( restore ) base class fis coherence
	ret= tmpFis->ClassCheckNoAllocResClassif(Exemples,nbRow,numOutput );
#ifdef DEBUG
	printf ("\n ClassCheckNoAllocResClassif returns %i",ret);
#endif
	// end TMP
      };
    ObjectToDouble( tmpFis, &Key , T );// builds T=array of doubles to optimize
    ///////////////////////////////////////////////////////////////
    //  Case of Crisp head rules
    ///////////////////////////////////////////////////////////////
    if ( FuzzyOutput != 1  )
      { if ( LimitedVocabulary == 1 )
	{
	  ///////////////////////////////////////////////////////////////
	  //  Case of Limited vocabulary
	  ///////////////////////////////////////////////////////////////
	  nbpossibles= InitialNormedFis->Out[numOutput]->GetNbPossibles();
	  fEvalArgs->add( &OutputNbMf ); 
	  Possibles= InitialNormedFis->Out[numOutput]->GetPossibles();
	  TabDomaine=new double [nbpossibles];
	  for ( int i = 0 ; i< nbpossibles ; i++ ) TabDomaine[i]=Possibles[i];
	  fEvalArgs->add( TabDomaine );
	  Name=PERMUTATION;
	  response=launch(tmpFis,T,&gain,FisEval);
	  delete [] TabDomaine;
	}
      else
	{
	  ///////////////////////////////////////////////////////////////
	  //  Case of non Limited vocabulary
	  ///////////////////////////////////////////////////////////////
	  response=launch(tmpFis,T,&gain,FisEval);//optimizes T in function of reconstructed FIS performance 
	}
      }
    ///////////////////////////////////////////////////////////////
    //  Case of Fuzzy head rules
    ///////////////////////////////////////////////////////////////
    if ( FuzzyOutput == 1  )
      {
	fEvalArgs->add( &OutputNbMf ); 
	TabDomaine=new double [OutputNbMf];
	for ( int i = 0 ; i< OutputNbMf ; i++ ) TabDomaine[i]=i+1;
	fEvalArgs->add( TabDomaine );
	Name=PERMUTATION;
	response=launch(tmpFis,T,&gain,FisEval);
	delete [] TabDomaine;
      }

    // Is there a better FIS 
    if ( ((Reponse==0)  && (  gain > -EPSILON ))
	 || ( (  ( Reponse ==1  ) || ( Reponse ==2  )) && ( gain < bestgain -EPSILON ) && (  gain > -EPSILON ) ))
      {

	Reponse=3;
	bestgain=gain;
	if ( Solution != NULL ) delete Solution;
	Solution =new FIS (*tmpFis);

	// TMP Temporaly here to check ( restore ) base class fis coherence
	ret= Solution->ClassCheckNoAllocResClassif(Exemples,nbRow,numOutput );
#ifdef DEBUG
	printf ("\n ClassCheckNoAllocResClassif returns %i",ret);
#endif
	// end TMP
      }
    else {
      //restore former gain
      gain=bestgain;
    }
    
     
  }

  if ( Reponse != 0) 
    { 
      //unnormalize FIS solution and data array
      // Exemples are not properly unnormalized!!!!
      Solution->UnNormalize(Exemples,nbRow);
      ret= Solution->ClassCheckNoAllocResClassif(Exemples,nbRow,numOutput );
#ifdef DEBUG
      printf ("\n ClassCheckNoAllocResClassif returns %i",ret);
#endif
    }
 
  // end TMP

  delete [] T;
  //return 1;
  // Reponse=0: no better FIS, 
  //1: better FIS after input optim,
  //2: better FIS after fuzzy output optim
  //3: better FIS after rule optim
  return Reponse;//modif August 2012
}

/*!
 * \fn int sifopt::init(void)
 * \param void
 * \return 1 
 * This init function duplicates fis, normalizes current one, allocates arrays to store what is read in the FIS class.
 */
int sifopt::init(void){
  // Store "this" in the conteneur to use it later when wrapping the evaluation function.
  fEvalArgs->add(this);
  double *MyInitTab,*Tab2;
  int    TailleTableau=0,ind=0,ind2=0,ind3,NBPAR,cptMF,ret;
  int    cptK1=0 , cptK2=0, cptK3=0, cptK4=0 , cptK5=0 ;
  char tampon[50];
  MaxKey=0;
  SizeIn=0;
  SizeOut=0;
  MfType = new int[nbEntree+1];
  double **Exemples=  (double **)fEvalArgs->get(1)  ;
  int    nbrow     =*((int *)    fEvalArgs->get(2) );
  int colonne=0;
  int cptE,Cpt;
  InitialNormedFis->Normalize(Exemples,nbrow);
  //InitialNormedFis->PrintCfg();
  //printf("\nafter norm Ex[0][0]=%f,Ex[nbrow-1][NbIn-1]=%f",Exemples[0][0],Exemples[nbrow-1][InitialNormedFis->GetNbIn()-1])

  ret= InitialNormedFis->ClassCheckNoAllocResClassif(Exemples,nbRow,numOutput );
#ifdef DEBUG
  printf ("\n ClassCheckNoAllocResClassif returns %i",ret);
  InitialNormedFis->PrintCfg();
#endif
  NbRule=InitialNormedFis->GetNbRule();
  if ( NbRule != 0 )InputCenters[nbEntree+1] = new double [ NbRule ];
  for ( Cpt = 0 ; Cpt < NbRule ; Cpt++ )
    {
      InputCenters[nbEntree+1][Cpt]=InitialNormedFis->Rule[Cpt]->GetAConc(numOutput);
#ifdef DEBUG
      printf (" Valeur de la conclusion  %f indice  %i",InputCenters[nbEntree+1][Cpt],Cpt);
#endif
    }
  // Calcul de la Taille du tableau de double ( que entrees )
  for ( cptE=0; cptE < nbEntree ; cptE++ )
    {
      ind=0;
      // unactive entry have a column in data but no MFs
      colonne+=1 ;
      InputNbMf[cptE]=InitialFis->In[cptE]->GetNbMf();
      for ( cptMF =0 ; cptMF < InputNbMf[cptE] ; cptMF++ )
	{
	  if ( InitialFis->In[cptE]->IsActive() == true ) MaxKey++;
	  ind3=(InitialNormedFis->In[cptE]->GetMF(cptMF))->NbParams();
	  ind+=ind3;
	  TailleTableau+=ind;
	};
      InputNbMf[cptE]=InitialFis->In[cptE]->GetNbMf();

      if ( ind  != 0 )
	{
	  InputCenters[cptE]=new double[ ind ];
	  InputCentersMin[cptE]=new double[ ind ];
	  InputCentersMax[cptE]=new double[ ind ];
	  InputCentersUpdatable[cptE]=new int[ ind ];
	  InputCentersType[cptE]=new int[ ind ];
	  InputCentersNbPar[cptE]=new int[ ind ];
	  InputCentersLenght[cptE]=ind;
	  SumCentersLenght+=ind;
#ifdef DEBUG
	  printf(" \n+I create InputCenters InputCentersUpdatable InputCentersType InputCentersNbPar of %i dim on %i num entry\n",
		 ind, cptE );
#endif
	  for ( int i=0;i< ind;i++){
	    InputCenters[cptE][i]=-999;
	    InputCentersMin[cptE][i]=-999;
	    InputCentersMax[cptE][i]=-999;
	    InputCentersUpdatable[cptE][i]=-999; 
	    InputCentersType[cptE][i]=-999;  
	    InputCentersNbPar[cptE][i]=-999;
	  }
	};
    };
  OutputNbMf=InitialNormedFis->Out[numOutput]->GetNbMf();
  // Calcul de la Taille du tableau de double ( que sortie )
  ind=0;
  // unactive entry have a column in data but no MFs
  InputNbMf[nbEntree]=InitialFis->Out[numOutput]->GetNbMf();
  for ( cptMF =0 ; cptMF <  InputNbMf[nbEntree] ; cptMF ++ )
    {
      ind3=(InitialNormedFis->Out[numOutput]->GetMF(cptMF))->NbParams();
      ind+=ind3;
    }
  SumCentersLenght+=ind;//OK - size includes fuzzy output
  if ( ind  != 0 )
    {
      InputCenters[nbEntree]=new double[ ind ];
      InputCentersMin[nbEntree]=new double[ ind ];
      InputCentersMax[nbEntree]=new double[ ind ];
      InputCentersUpdatable[nbEntree]=new int[ ind ];
      InputCentersType[nbEntree]=new int[ ind ];
      InputCentersNbPar[nbEntree]=new int[ ind ];
      InputCentersLenght[nbEntree]=ind;
#ifdef DEBUG
      printf(" \n+I create InputCenters InputCentersUpdatable InputCentersType InputCentersNbPar of %i dim on %i output num entry\n",
	     ind, nbEntree );
#endif
    };
#ifdef DEBUG
  printf(" \n+I OutputNbMf =  %i \n", OutputNbMf );
#endif

#ifdef DEBUG
  printf(" \n+I SumCentersLenght %i \n",SumCentersLenght);
#endif

  if ( tmpFis != NULL ) delete tmpFis;
  tmpFis = new FIS( * InitialNormedFis );
  ret= tmpFis->ClassCheckNoAllocResClassif(Exemples,nbRow,numOutput );
#ifdef DEBUG
  printf ("\n ClassCheckNoAllocResClassif returns %i",ret);
#endif
  /* Warning MaxKey is used for tests by giving a 0 key :
     it launches all the possibilities of Keys, it works if the integer
     corresponding to the number of possibilities can be store in
     a long unsigned int ( about 32 or 64 dimension, it depend on your register size )
     this sofware can manage any dimension whithout the 0 key because of template biset */
  MaxKey=(unsigned long int)(pow(2,(double)MaxKey)-1);
#ifdef DEBUG
  printf("\n+I MaxKey : %lu \n",
	 MaxKey );
#endif
  MyInitTab=new double[SumCentersLenght];
  Tab2 =new double[SumCentersLenght];// should be max number of parameters !!!
  ind=0;
  for ( cptE=0; cptE < nbEntree ; cptE++ ){
    if ( InitialFis->In[cptE]->IsActive() != true ) continue;
    PFFIn.Value[cptK1]=InitKey.getValue(cptK2); cptK1++; cptK2++;
    for ( int cptMF =0 ; cptMF< InitialNormedFis->In[cptE]->GetNbMf() ; cptMF++ ){
      KeyIn.Value[cptK3]=InitKey.getValue(cptK2);  cptK2++;    
      if (( KeyIn.Value[cptK3] == 1 ) && ( PFFIn.Value[cptK1-1] == 1 )) SizeIn++;
      ( InitialNormedFis->In[cptE]->GetMF(cptMF) )->GetParams(Tab2);
      NBPAR=( InitialNormedFis->In[cptE]->GetMF(cptMF) )->NbParams();
      InputCentersNbPar[cptE][cptMF] = NBPAR;
      if (( KeyIn.Value[cptK3] == 1 ) && ( PFFIn.Value[cptK1-1] == 0 )) SizeIn+=NBPAR;
      strcpy( tampon , ( InitialNormedFis->In[cptE]->GetMF(cptMF) )->GetType() ) ;
      if (strcmp( "triangular\0",tampon)==0)
	InputCentersType[cptE][cptMF] = 1;
      if (strcmp( "trapezoidal\0",tampon)==0)
	{ 
	  InputCentersType[cptE][cptMF] = 2;
	  if ( KeyIn.Value[cptK3] == 1 ) SizeIn++;}
      if (strcmp( "SemiTrapezoidalInf\0",tampon)==0)
	InputCentersType[cptE][cptMF] = 3;
      if (strcmp( "SemiTrapezoidalSup\0",tampon)==0)
	InputCentersType[cptE][cptMF] = 4;
      if (strcmp( "gbell\0",tampon)==0)
	InputCentersType[cptE][cptMF] = 5;
      if (strcmp( "gaussian\0",tampon)==0)
	InputCentersType[cptE][cptMF] = 6;
      if (strcmp( "universal\0",tampon)==0)
	InputCentersType[cptE][cptMF] = 7;
      cptK3++;
#ifdef DEBUG
      printf("+I Type MF %i Input %i %i //",
             cptMF,
             cptE,
             InputCentersType[cptE][cptMF]);
#endif
      for ( int cpt=0 ; cpt < NBPAR ; cpt++ ) {
	MyInitTab[ind2]=Tab2[cpt];
	ind2++;
      };
      ind++;
    };
  };
#ifdef DEBUG
  printf("\n+I KeyIn ");
  KeyIn.Print(cptK3-1);
  printf("+I SizeIn %i",SizeIn);
  printf("+I PFFIn ");
  PFFIn.Print(cptK1-1);
#endif
  //strcpy(tampon, KeyIn.Print(cptK3));
  //setKey(KeyIn.Print(cptK3));
  Key=KeyIn;
  //setKey((long unsigned int)0);
  // if the first output does not use classification it won't be optimize
  if( strcmp( InitialFis->Out[numOutput]->GetOutputType(), OUT_FUZZY::OutputType() ) !=0 )
    {
#ifdef DEBUG
      printf("+I Crisp ");  
#endif
      cptK2++; // the flag FC is read in the FIS and not in the Key
      FuzzyOutput=-1;
      LimitedVocabulary=InitKey.getValue(cptK2); cptK2++;
      CardVocabulary=InitKey.getValue(cptK2); cptK2++;
      PFFOut=-1;
#ifdef DEBUG
      printf("+I LimitedVocabulary = %i ",LimitedVocabulary);
      printf("+I CardVocabulary = %i \n",CardVocabulary);
#endif
    }
  else
    {
#ifdef DEBUG
      printf("\n+I Fuzzy ");  
#endif
      cptK2++; // the flag FC is read in the FIS and not in the Key
      cptK2++; // the classif flag  is read in the FIS and not in the Key
      //PFFOut=InitKey.getValue(cptK2); cptK2++; // for the fuzzy output  F1 means PFF or not
      // temporaire pour permettre de travailler avant la correction
      // PFFOut=InitKey.getValue(cptK2);cptK2++;
      PFFOut=1;cptK2++;
      for ( int cptO = 0 ; cptO <OutputNbMf ; cptO++)
	{
	  KeyOut.Value[cptK4]=InitKey.getValue(cptK2);
	  // temporaire pour permettre de travailler avant la correction
	  // KeyOut.Value[cptK4]=1;
	  cptK4++; cptK2++;
	};
      for ( int cptMF =0 ; cptMF< InitialNormedFis->Out[numOutput]->GetNbMf() ; cptMF++ ){
	if (( KeyOut.Value[cptK3] == 1 ) && ( PFFOut == 1 )) SizeIn++;
	NBPAR=( InitialNormedFis->Out[numOutput]->GetMF(cptMF) )->NbParams();
	// added by bch - init correction for fuzzy output
	( InitialNormedFis->Out[numOutput]->GetMF(cptMF) )->GetParams(Tab2);
	//
	InputCentersNbPar[nbEntree][cptMF] = NBPAR;
	if (( KeyOut.Value[cptK3] == 1 ) && ( PFFOut == 0 )) SizeIn+=NBPAR;
	strcpy( tampon , ( InitialNormedFis->Out[numOutput]->GetMF(cptMF) )->GetType() ) ;
	if (strcmp( "triangular\0",tampon)==0)
	  InputCentersType[nbEntree][cptMF] = 1;
	if (strcmp( "trapezoidal\0",tampon)==0)
	  {
	    InputCentersType[nbEntree][cptMF] = 2;
	    if ( KeyOut.Value[cptK3] == 1 ) SizeIn++;}
	if (strcmp( "SemiTrapezoidalInf\0",tampon)==0)
	  InputCentersType[nbEntree][cptMF] = 3;
	if (strcmp( "SemiTrapezoidalSup\0",tampon)==0)
	  InputCentersType[nbEntree][cptMF] = 4;
	if (strcmp( "gbell\0",tampon)==0)
	  InputCentersType[nbEntree][cptMF] = 5;
	if (strcmp( "gaussian\0",tampon)==0)
	  InputCentersType[nbEntree][cptMF] = 6;
	if (strcmp( "universal\0",tampon)==0)
	  InputCentersType[nbEntree][cptMF] = 7;
	cptK3++;
#ifdef DEBUG
	printf("\n+IO Output InputCentersType[ %i ][ %i ] %i ",nbEntree,cptMF,InputCentersType[nbEntree][cptMF]);
#endif
	// added by bch - init correction for fuzzy output
	for ( int cpt=0 ; cpt < NBPAR ; cpt++ ) {
	MyInitTab[ind2]=Tab2[cpt];
	ind2++;
	};
	//
      };
    };
#ifdef DEBUG
  printf("\n+I PFFout %i ",PFFOut);
  printf("+I KeyOut ");
  KeyOut.Print(cptK4);
#endif

  // extract the key for the rules
  for (  Cpt = 0 ; Cpt < InitialFis->GetNbRule() ; Cpt++ )
    {
      KeyRule.Value[cptK5]=InitKey.getValue(cptK2);
      cptK5++; cptK2++;
    };

  KeyRuleLenght=cptK5;
#ifdef DEBUG
  printf("+I KeyRule ");
  KeyRule.Print(cptK5);
  printf("+I Normed FIS\n");
  InitialNormedFis->PrintCfg();
  printf("+I \nEnd init\n\n");
#endif


  //error! bch March 2010 - case where there were more rules than input parameters
  // allocation must be sufficient
  int maxLenght=max(SumCentersLenght,KeyRuleLenght);
  
  double * Tab=NULL;
  Tab=new double[maxLenght];
  for (int itab=0; itab<SumCentersLenght;itab++)
    Tab[itab]=Tab2[itab];
  for (int itab=SumCentersLenght;itab<maxLenght;itab++)
    Tab[itab]=0.0;
  //

  //initData(MyInitTab,SumCentersLenght);
  initData(Tab,maxLenght);//+KeyRuleLenght
  //
  if ( Tab2 != NULL) delete [] Tab2;
  if ( MyInitTab != NULL) delete [] MyInitTab;
  if ( Tab != NULL) delete [] Tab;
  return 1;
}
/*!
 * \fn void sifopt::Update(int * ResClassif,double ** Exemples);
*/
void sifopt::Update(int * ResClassif,double ** &Exemples,int &nout, int & nbrow, int & cnear, double & InitCoverage, double &cov)
{
  pResClassif=&ResClassif;
  Exemples=Exemples;
  numOutput=nout;
  nbRow=nbrow;
  CNearActivate=cnear;
  InitCov = InitCoverage; 
  if(cov > 0.)
    CovLoss = cov;
  else
    CovLoss = 0.;
}
void sifopt::UpdatefEvalArgs(double ** &exemples, int & nbrow, double & couverture, double & maxerror, 
		       int * &ResClassif, double &museuil, double * &lab )
{
  // update feval parameters
  fEvalArgs->add(exemples);
  fEvalArgs->add(&nbrow);
  fEvalArgs->add(&couverture);
  fEvalArgs->add(&maxerror);
  fEvalArgs->add(ResClassif);
  fEvalArgs->add(&museuil);
  fEvalArgs->add(&lab);
}

/*!
 * \fn double sifopt::Eval( void *  Item  , vkey  *  pKey , double *Array , int sizeArray )
 * \param Item A non typed pointer to your object.
 * \param pKey The code to extract values from your object.
 * \param Array The current double values.
 * \param sizeArray The size of the double values.
 * \return the Evaluation or -1 if a constraint was violated.
 * \brief this function get the FIS Performance params out of the conteneur and launch the evaluation of the FIS in the current state.
 */
double sifopt::Eval( void *  Item  , vkey  *  pKey , double *Array ,
                     int sizeArray )
{
  double **Exemples=  (double **)fEvalArgs->get(1)  ;
  int    nbrow     =*((int *)    fEvalArgs->get(2) );
  double Couverture=*((double *) fEvalArgs->get(3) );
  double MaxError  =*((double *) fEvalArgs->get(4) );
  int    *ResClassif=(int *) fEvalArgs->get(5);
  double    museuil     =*((double *)    fEvalArgs->get(6) );
  double *Lab       =(double *) fEvalArgs->get(7);
  FIS *MyF=(FIS *) Item;
  double resVal;
  if ( DoubleToObject ( Array , sizeArray , pKey , MyF ) == -1) return -1;
  resVal=MyF->Perf(0,
		   Exemples,
		   nbrow,
		   Couverture,
		   MaxError,
		   museuil, 
		   ResClassif,
		   Lab);
  // Added by SG - March 16th, 2010
  if(Couverture < InitCov*(1. - CovLoss)) return -1.;

  return resVal;
}
