//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************
#include "algo.h"
/*!
 * \file sw.cpp
 * \brief defines the solis and wells algorythm using the algo language.

 This function defines the solis and wells algorithm using the algo language, it considers that the evalalution is positive and that a constraint violation returns -1 to the evaluation.
 */
int SolisWetts(
           void     * YourObject,
           vkey     * Key,
           double   * ArrayOfDouble ,
           int        SizeArray,
           algoPar  * P,
           double   * gain,
           double  (* pfEval)( void   *  ,
                   vkey    *  ,
                   double *  ,
                   int       ,
                   conteneur *),
           conteneur * fEvalArgs
           )
{
  long int  SWite=0;  // The number of iteration
  int       KeySize=0;  // Number of true set in the key
  long int       SWC=0;  // A counter that counts constraints violations in this iteration
  double    SWE;      //  Current evaluation
  double    SWbestE; //   Best evaluation of E
  double    SWprevE;  // Previous evaluation of E
  double    SWfirstE; // First Evaluation of E
  int reponse=0;
  
  // A little check-very important: if all elements=0 in key return -1
  KeySize=Key->Size();
  //
  if ( KeySize == 0 ) return -1;
///////////////////////////////////////////////
// Init
////////
// Put input datas in a vector object VectDouble
avect *VectDouble=new avect( SizeArray , ArrayOfDouble );
#ifdef DEBUG2
  printf("\n+SW Solis Wells recoit ");VectDouble->print();
  printf("\n+SW longueur vecteur %i\n" , KeySize );
  printf("\n+SW SizeArray %i\n" , SizeArray );
  VectDouble->print();
#endif

  SWfirstE=(*pfEval)( YourObject ,
              Key ,
              VectDouble->getArrayAdress() ,
              VectDouble->getNbElem()      ,
              fEvalArgs);

  //printf("\n+SW Solis Wells gets ");VectDouble->print();
  //printf("\nfirst evaluation in sw SWfirstE=%f\n",SWfirstE);
  SWbestE=SWfirstE;

// create a new vector to perform operations
  avect *T = new avect(SizeArray);
  avect *T2= new avect(SizeArray);
//  Generate a gaussian vector
  avect *G = new avect(SizeArray);
  avect *AvgGauss= new avect(SizeArray);
  avect *BestVect= new avect(SizeArray);
  AvgGauss->init0();
  G->initGauss(P->RangeNoise,AvgGauss);
  BestVect->dupVector(VectDouble);
#ifdef DEBUGS
  printf("\n+SW vect Depart \n");
  VectDouble->print();
  printf("\n+SW premier gaussien moy 0 bruit %f ", P->RangeNoise );
  G->print();
#endif
// let's get in the while whith : previous correct and worse than current
  SWprevE=SWfirstE+1;

#ifdef DEBUGS
  printf("\n+SWInitial Evaluation %f ",SWfirstE );
#endif
///////////////////////////////////////////////
// Main Loop
////////
  while (  SWite < P->maxIte ) 
    {
#ifdef DEBUGS
printf("\n+SWIteration %li ", SWite );
#endif
      //  T <- A + G
      T->dupVector(VectDouble);
      T->addVector(G);
      //  SWE <- EVAL A+G
      SWE=(*pfEval)( YourObject , Key , T->getArrayAdress() , T->getNbElem() , fEvalArgs);
     
#ifdef DEBUGS
printf("\n+SWEval : %f | ",SWE);
printf("bestE : %f | ",SWbestE);
printf("SWprevE : %f",SWprevE);
#endif

       if ( SWE != -1 )
    {
      if ( ( SWE <= SWbestE ) && ( fabs ( SWbestE - SWE ) > EPSILON )) 
        {
          //      A <- A + G
          VectDouble->dupVector(T);
          //      M <- 0.4G + 0.2M        
          T->dupVector(G);
          T->tlambVector(P->lamb1);
          T2->dupVector(AvgGauss);
          T2->tlambVector(P->lamb2);
          T->addVector(T2);
          AvgGauss->dupVector(T);
#ifdef DEBUGTERMFAST
          printf("\n+SW cas 1 E ( A + G ) < E ( A )  -- %f \n",SWE);
#endif
#ifdef DEBUGS
          printf("\n+SW Nouveau Vecteur");
          VectDouble->print();
          printf("\n+SW Nouveau Gaussien Moyen");
          AvgGauss->print();
#endif
          SWprevE=SWE;
          if ( SWE > -EPSILON ) { SWbestE=SWE;  BestVect->dupVector(VectDouble); }
        }
      else {
        //  T <- A - G
        T->dupVector(G);
        T->moinsuVector();
        T->addVector(VectDouble);
        //  SWE <- EVAL A - G
        SWE=(*pfEval)( YourObject , Key , T->getArrayAdress() , T->getNbElem() , fEvalArgs);
#ifdef DEBUGS
printf("\n+SWEval : %f",SWE);
printf("\n+SWbestE : %f",SWbestE);
printf("\n+SWprevE : %f",SWprevE);
#endif
        // we consider that the evaluation is positive
        // a constraints violated return -1 to the evaluation
        if ( SWE != -1 ) 
          {
        if (( SWE <= SWbestE ) && ( fabs ( SWbestE - SWE ) > EPSILON ))
          {
          //  A <- T
          VectDouble->dupVector(T);
          //  M <- M - 0.4G
          T->dupVector(G);
          T->tlambVector(P->lamb1);
          T->moinsuVector();
          T2->dupVector(AvgGauss);
          T->addVector(T2);
          AvgGauss->dupVector(T);
#ifdef DEBUGTERMFAST
          printf("\n+SW cas 2 E ( A - G ) < E ( A )  -- %f \n",SWE);
#endif
#ifdef DEBUGS
          printf("\n+SW Nouveau Vecteur");
          VectDouble->print();
          printf("\n+SW Nouveau Gaussien Moyen");
          AvgGauss->print();
#endif
          if  ( SWE > -EPSILON )  { SWbestE=SWE;  BestVect->dupVector(VectDouble); }
          SWprevE=SWE;
        }
        else
          {
            AvgGauss->tlambVector(P->lamb3);
#ifdef DEBUGS
            printf("\n+SW cas 3 E ( A -+ G ) > E ( A ) \n+SW Nouveau VectDouble");
            VectDouble->print();
            printf("\n+SW Nouveau Gaussien Moyen");
            AvgGauss->print();
#endif
          }
          }

      };

    };

      if ( SWE == -1 )
      { SWE = SWprevE;
        SWC++;
#ifdef DEBUG2
      printf("\n+SW colision n %li\n",SWC);
#endif
        if ( SWC > P->maxConstraints)
           {
           SWC=0;
           SWite ++;
#ifdef DEBUG2
      printf("\n+SW max collision atteint , iteraion  %li\n",SWite);
#endif
           };
      }
      else
          {  SWite ++;
             SWC=0;
      };
      G->initGauss(P->RangeNoise,AvgGauss);

#ifdef DEBUG2
      printf("\n+SW  Nouveau gaussien " ); G->print();
#endif
    }
///////////////////////////////////////////////
// End Loop
////////
#ifdef DEBUG2
  printf("\n+SW Fin au bout de %ld itrations ",SWite);
  printf("\n+SW Distance au but de %f vers %f",SWfirstE,SWbestE);
  printf("\n+SW Parametre "); P->Print();
  printf("\n+SW Vecteur " ); VectDouble->print();
  printf("\n+SWFIN Performance finale %f \n",SWbestE );
#endif


  if (  SWfirstE != SWbestE )
  *gain=(*pfEval)( YourObject ,
              Key ,
              BestVect->getArrayAdress() ,
              BestVect->getNbElem()      ,
              fEvalArgs);
#ifdef DEBUG2
  printf("\n+SWFIN gain vaut %f \n",SWbestE );
#endif
  //correction bch - September 18 2012
  if (  SWfirstE <= SWbestE )
    reponse=1;//no improvement- initial values should be restored in calling pgm

  //  free memory
  delete AvgGauss;
  delete G;
  delete VectDouble;
  // correction bch - July 20 2009
  delete BestVect;
  delete T;
  delete T2; 

  return reponse;
}
