//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.1 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Contact : fispro@supagro.inra.fr
// Last modification date:December 1, 2007
//**********************************************************************
/*! testsifopt.cpp 
 * \brief Test and interprete the params,\n buid the Samples whith the argv2 file,\n lauch the algorythm.\n 
 *
 */

#include "common.h"
#include "fis.h"
#include "sifopt.h"

int main(int argc, char ** argv)
//******************************
{
  if(argc < 10)
    {
      printf("\nThe programm %s needs 9 arguments : ", argv[0]);
      printf("\n\t* the name of the configuration file;");   
      printf("\n\t* the name of the data file;");   
      printf("\n\t* the name of the resulting FIS configuration file;");   
      printf("\n\t* the key;");   
      printf("\n\t* the number of iterations;");   
      printf("\n\t* the range of the gaussian noise;");
      printf("\n\t* the max number of constraints violations before it counts as an iteration;");
      printf("\n\t* the max number of failure steps in an algorithm;");
      printf("\n\t* the equality center distance threshold");
      printf("\n\t* -o   Output to optimise;");
      printf("\n\t* -m   minimum membership;");
      printf("\n\t* -s   integer to set seed value for parker miller random;");
      printf("\n\t* -l1  Solis Wetts Constant 1;");
      printf("\n\t* -l2  Solis Wetts Constant 2;");
      printf("\n\t* -l3  Solis Wetts Constant 3;");
      printf("\n\t* -cn  Activate the CNear Constraint;\n\n");
      return 1;
    }

  double ** Exemples  = NULL;   // A Pointer to sample array
  int * ResClassif = NULL;
  double * Lab = NULL;
  int nbcol=0, nbrow=0, cnear=0 ;
  char NomFichier[100];
  strcpy(NomFichier,argv[2]);

  double Couverture, MaxError,museuil=.1;
  int i, ElimRedondance,nOut=0;
  int reponse=-1;

  ElimRedondance=0;
  double l1=-1,l2=-1,l3=-1;
  
  
  if(argc > 10) 
    {
      for(i = 10; i < argc; i++)
        {
	  if(! strcmp("-o", argv[i]) ) nOut=atoi(argv[i+1]);
	  if(! strcmp("-m", argv[i]) ) museuil=atof(argv[i+1]);
          if(! strcmp("-s", argv[i]) ) setseed(atoi(argv[i+1]));
          if(! strcmp("-l1", argv[i]) ) l1=atof(argv[i+1]);
          if(! strcmp("-l2", argv[i]) ) l2=atof(argv[i+1]); 
          if(! strcmp("-l3", argv[i]) ) l3=atof(argv[i+1]);
	  if(! strcmp("-cn", argv[i]) ) cnear=1;
	  if(! strcmp("-r", argv[i]) ) ElimRedondance = 1;

        }
    }
  // Build Fis
  FIS * S;

  try{

  S = new FIS(argv[1]);     

  FILE *fg,*fs;           // Generation d'un fichier de config 
  fg = fopen("config.sif", "wt");
  S->PrintCfg(fg);
  fclose(fg);

  // initialise Samples
#ifdef DEBUG
  printf("\n\t* acquisition du pointer sur exemple et classif;\n\n");
#endif
//  nbrow=S->InitResClassifSamples( &Exemples , &ResClassif , NomFichier , nOut);
//  It should be a FIS function ...
  Exemples = ReadSampleFile( NomFichier , nbcol, nbrow);
   if(!*Exemples) 
      {
#ifdef DEBUG
      printf("\nThe sample matrix cannot be built ! \n");
#endif
      return 1 ;
      }
//test on column and row number
      if(nbcol < S->GetNbIn()) 
      {
#ifdef DEBUG
      printf("\n The number of input variables is smaller than the number of inputs in the FIS.\n");
#endif
      return  -2;
      }

   if(nbcol < S->GetNbIn() + S->GetNbOut())
      {
#ifdef DEBUG
      printf("\n The number of variables is smaller than the number of inputs plus the number of outputs.");
#endif
      }  
   // call to Classcheck once to allocate ResClassif and updates Lab pointer
  reponse=S->ClassCheck(ResClassif,Lab,Exemples,nbrow,nOut );
#ifdef DEBUG
      printf("\n The Class Check returns %i",reponse);
#endif

// .... 
  //beginning of sifopt
  //create sifopt using sif
    sifopt *Opt;
   //need this to do a clean copy
    Opt= new sifopt(S,ResClassif,Exemples,nbrow,nOut);
    Opt->pResClassif=&ResClassif;
    Opt->Exemples=Exemples;
    //build key from 3rd argument  
    Opt->setKeyInit( argv[4] ) ;
   //choose output
    Opt->numOutput=nOut;
   //write nbrow
    Opt->nbRow=nbrow;
   //write nbrow
    Opt->CNearActivate=cnear;
   //P = algorithm parameters
    Opt->P->maxIte=atoi(argv[5]);
    Opt->P->RangeNoise=(double)atof(argv[6]);
    Opt->P->maxConstraints=atoi(argv[7]);
    Opt->P->maxFail=atoi(argv[8]);
    Opt->P->EpsilonCenter=atof(argv[9]);
    if ( l1 != -1 ) Opt->P->lamb1=l1;
    if ( l2 != -1 ) Opt->P->lamb2=l2;
    if ( l3 != -1 ) Opt->P->lamb3=l3;
   //following lines : parameters for evaluation function
   // see performance in sif.h
    Opt->fEvalArgs->add(Exemples);
    Opt->fEvalArgs->add(&nbrow);
    Opt->fEvalArgs->add(&Couverture);
    Opt->fEvalArgs->add(&MaxError);
    Opt->fEvalArgs->add(ResClassif);
    Opt->fEvalArgs->add(&museuil);
    Opt->fEvalArgs->add(Lab);
    
   //launch optimizer : optimize entries or/and conclusion depending on key
   //uses parameters defined above
   reponse=Opt->launchEntry();

  //open result config file for optimized fis
  // if Opt->Solution == NULL , no solution found
  fs = fopen(argv[3], "wt");
  if   ( ( Opt->Reponse == 1 ) || ( Opt->Reponse == 2 ) || ( Opt->Reponse == 3 )) {
       Opt->Solution->PrintCfg( fs );
      };
  fclose( fs );
  //option DEBUGCSV used in makefile only
#ifdef DEBUGCSV
  Opt->P->PrintCSV(); 
  printf("\n");
#endif
  
  //delete FIS and FISOPT objects
  if (Opt != NULL) delete Opt;
  if (S != NULL) delete S;

  // delete data array
  if (Exemples !=NULL) {
    for( int i=0 ; i<nbrow ; i++ )
      delete [] Exemples[i];
    delete [] Exemples;
  }
  // delete resclassif array
  if (ResClassif != NULL) delete [] ResClassif;

  printf("\n***********************\nFIS opt END\n");
  //! \todo correct to sifopt destructor
  }
   catch(std::exception &e){
   printf("error in fis\n");
   printf("catch %s\n",e.what());

  }
  return (reponse);
}
