//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************
/*! \file testkeyutil.cpp
    \brief This source file give a simple sample of key manipulations.

This source file give a simple sample of key manipulations.\n
First his nature : key is a template of bitset, that's why there are some stranges behaviours whith C operators OR and AND.\n
A simple way to use it is to directly store a char string into the sifopt vkey member whith setKeyInit ( but that mean u would have anderstand how U write it ... ).\n
An other more interesting way is \n
\arg \c create the key corresponding to an in an out a mf or a rule whith sifopt member function beginnig by key
\arg \c use logical operator to translate what U want

*/
#include "sifopt.h"
int main(int argc, char *argv[]){
  // needed to create FIS and sifopt 
  double ** Exemples  = NULL;
  int * ResClassif = NULL;
  int nbrow=0,nOut ;

  FIS * S;
  sifopt *Opt;
  vkey *K1,*K2,*K3,*K4;
  

  // test normal random numbers
  double pr=0.0, mag=0.25;
  int ndice=12;
  
  FILE * fdice=NULL;
  fdice=fopen("dice","wt");
  for (int ii=0;ii<100;ii++)
    {
      pr=GaussDice(mag,ndice);
      fprintf(fdice,"%5.3f ",pr);
    }
  fclose(fdice);
  
  if(argc < 3)
    {
      printf("\nThe programm %s needs 2 arguments : ", argv[0]);
      printf("\n\t* the FIS configuration file");
      printf("\n\t* the output number to be optimized");
      printf("\n\t* 0 for first output, 1 for second output ...\n\n");
      return 1;
    }

  try {
    nOut=atoi(argv[2]);
    S = new FIS(argv[1]);
    Opt= new sifopt(S,ResClassif,Exemples,nbrow,nOut);
    // beginning 
    printf("\nbeginning\n");
    /*(Opt->InitKey).Print();
    Opt->setKeyInit( "11" ) ;
    printf("\nAfter setKeyInit\( \"11\" \)\n");
    (Opt->InitKey).Print();
    */
    printf("\nAfter keyallopt\n"); 
    K1=Opt->keyallopt(nOut,S);
    K1->Print();
    /*printf("\nAfter setKeyInit\( \"0\" \)\n");
    Opt->setKeyInit( "0" ) ;
    (Opt->InitKey).Print();
    printf("\nInput 0 \n");
    K3=Opt->keysetIN( S , 0 , 1 );
    K3->Print();
    */
    printf("\nOutput  %i \n",nOut);
    K4=Opt->keysetOUT( S , nOut , 1 );
    K4->Print();
    printf("\nRule  0 \n");
    K2=Opt->keysetRULE( S , nOut , 0 );
    K2->Print();
    //printf("\nOutput 1 AND input 1 ( logical OR )\n");
    //K4->Value=K4->Value|K3->Value;
    //K4->Print();
    //printf("        Etonov Niet ?\n"); 
    // and now very complicated
    Opt->setKeyInit(K4);
    printf("\nthe key the optimizer starts with\n");
    (Opt->InitKey).Print();
    delete K1;
    delete K3;
    delete K4;
    for ( int NumEntry = 0 ; NumEntry < S->GetNbIn(); NumEntry ++ )
      {
	printf("\nInput %i PFF\n", NumEntry);
	K3=Opt->keysetIN( S , NumEntry , 1 );
	K3->Print();
	printf("\nInput %i nonPFF\n", NumEntry);
	K4=Opt->keysetIN( S , NumEntry ,0);
	K4->Print();
	delete K3;
	delete K4;
      };

    // optimize all rules
    
    delete S;
  }
  catch(std::exception &e){
    printf("error in fis\n");
    printf("catch %s\n",e.what());
  }
  printf("\n\n"); 
  return ( 1 );
}
