//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************
#ifndef VKEY_H
#define VKEY_H
#define SIZE 500
#include <iostream>
#include <cstring>
#include <bitset>
#define BOOL(x) (!(!(x)))
#define BitTst(arg,posn) BOOL((arg) & (1L << (posn)))
/*!
 * \class vkey
 * \brief vector's key class manipulate array of boolean to extract the values of an item that will be optimized.
 *
 * An example of how to use vector key is given in file reference testkeyutil.cpp
 * This class is used to manipulate boolean Array, that will determinate what part of the FIS will be sent to the optimizer.\n
 * Some keys utilities function has been writing in sifopt files (those ones who are specific to FIS ... )
 * \todo  A first version considers 500 bits store in 64 octects using bitset template, find an other implementation that is as good, but does not need to relink if there is a use of a big amount of bitset
*/
class vkey {
public:
//! The Value of the Array
std::bitset<SIZE> Value;
/*!
 * \fn int  getValue(int i)
 * \param i the indices in the array
 * \return 1 if i'th boolean is true 0 in other case.
*/
int  getValue(int i) const { return Value[i];}
/*!
 * \fn void setValue(int i,int j)
 * \param i the indices in the array
 * \param j the value 1 true 0 false
*/
void setValue(int i,int j)
{
 Value[i]=j;
 return;
}

/*!
 * \fn void setValue( unsigned long int i )
 * \brief create the 110 key whith the 6 integer
 * \param i an integer to transform into a key
 \remark beware that a too big integer will cycle

 This construcor is used is small samples to generate the key equivalent to the binary writing of an integer. Very useful to perform key loops.
*/
void setValue( unsigned long int i )
{
 int k=0;
 Value.reset();
 while (  i != 0 )
 {
  Value[k] = BitTst( i , 0 );
  i >>= 1;
  k++;
  }
 return;
}
//! C language = operator ( affectation )
void setValue( const vkey *  kPar )
{
 Value=kPar->Value;
 return;
}

/*!
 * \fn void setValue( const char * chaine )
 * \brief create the 110 key whith the "110" CString
 * \param chaine an adress of an CString containing the binary writing of the key
 \remark const,for your conviniance, not to allocate the string in a rude using like setValue("1111111111101110000000000000011...."

Reverse the string ( as U should know that we read from left to right and write figure from right to left ... ) and create the equivalent key.
 */
void setValue( const char * chaine )
{
 int i,l;
 char c;
 l=strlen(chaine);
 Value.reset();
 for ( i=0 ; i < l ; i++  ) 
 {
  c = chaine[l-i-1];
  Value[i] = (( int ) c) - (( int ) '0') ;
#ifdef DEBUGK
  std::cout<<"\n Cles Valeur indice "<<i<<" " <<Value[i];
#endif
  }
 return;
}

/*!
 * \fn void Print( void )
 * \brief print the key ommiting left zeros
 */
void Print( void ) const
 {
 int gauche=1;
 for(int i=Value.size();i>=0;i--) 
  { 
  if  ( ( gauche == 1 ) && ( Value[i] == 1 ) ) gauche = 0;
  if ( gauche == 0 ) std::cout<<Value[i];
  }
 std::cout<<std::endl;
 return;
 }

/*!
 * \fn void Print( void )
 * \brief print the arg's first values of key
 */
void Print( int vsize ) const
 {
 for(;vsize>=0;vsize--) std::cout<<Value[vsize];
 return;
 }
//! tells are many true are set in the key 
int Size( void ) const
 {
 return Value.count();
 }

};


#endif
