//*********************************************************************
//
//
//                              CRITERIA.CPP
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISTREE utility functions for fuzzy decision tree
//
//*******************************************************************
//
//
//                FISTREE split criterion (entropy or other) Functions
//                FISTREE is a derived class of FIS
//
//*******************************************************************
#include "fistree.h"

// Class FISTREE
// Entropy function
//--------------------------------------------------------
double FISTREE::Entropy(NODE * wnode,
			int dim,int nbMF,
			double ** pnDim, double * pnDimTot, 
			double &pnT, double * entroMF, 
			int flagDisplay)
{
  //computes entropy=enDim (return value) at node wnode for variable dim
  //modifies pnDim, pnDimTot, pnT and entroMF
  // pnDim : size ncl*nbMF 
  // pnDimTot, entroMF: size nbMF
  // dim=-1 : root node, do not compute previous mu (muprec)

  int nMF,numex,numclas,nbl,nbclas, outputCol;
  double munMF,muex,munc,munPrev,tempo,ex,enDim,prop,pnTupNode;
  double tempoUpNode;
  double ** exemples;
  double mumin;
  char * conj=NULL;
  FISIN ** in;
  FISOUT ** out;
  
  //-------------------------------------------------------
  //
  //-------------------------------------------------------

  //get infos from tree object
  in=(*this).In;
  out=(*this).Out;
  outputCol=(*this).NbIn + OutputN; // output column in data array
  nbclas=(*out[OutputN]).GetNbMf();
  exemples=(*this).Examples;
  nbl=(*this).Snbrow;
  mumin=(*this).Smumin;
  conj=new char[1+strlen(this->TypeConj())];
  strcpy(conj,this->TypeConj());
  
  //init variables
  enDim=0.0;
  prop=0.0;
  pnT=0.0;
  pnTupNode=0.0;
  //error catching
  try
    {
      for (nMF=0;nMF<nbMF;nMF++)
	{ 
	  // reinit arrays in loop for examined variable
	  pnDimTot[nMF]=0.0;
	  entroMF[nMF]=0.0;
	  for (numclas=0;numclas<nbclas;numclas++)
	    pnDim[numclas][nMF]=0.0;
	  // end reinit arrays
	  // item loop
	  for (numex=0;numex<nbl;numex++)
	    {
	      // munc: recursive calculation
	      // from current node up to root
	  
	      munPrev=(*wnode).CalcMuN(exemples, numex,in,conj);
	      
	      if (dim == -1)
		munMF=1.;//neutral element for root
	      else
		{
		  ex=exemples[numex][dim];
		  munMF=in[dim]->GetADeg(nMF,ex);
		}
	      munc=(*wnode).AND(munMF,munPrev,conj);
	    
	      // item matters only if munc>mumin
	      if (munc <= mumin) munc=0.;
	      //
	      // loop on output classes
	      for (numclas=0;numclas<nbclas;numclas++)
		{
		  ex=exemples[numex][outputCol]; //output value 
		  muex=out[OutputN]->GetADeg(numclas,ex);
		  tempo=(*wnode).IMPLI(munc,muex,conj);//munc if item belongs to numclas, 0 otherwise
		  if(nMF==0) 
		    {
		      tempoUpNode=(*wnode).IMPLI(munPrev,muex,conj);//munPrev if item belongs to numclas, 0 otherwise
		      pnTupNode+=tempoUpNode;//sum of node (to be split) mu for all classes
		    }
		  pnDim[numclas][nMF]+=tempo;
		  pnDimTot[nMF]+=tempo;
		      
		}// end of class loop
		     
	    } // end of item loop
	  if (flagDisplay) printf("in Entropy dim=%d, PnDimTot[%d]=%g\n",dim,nMF,pnDimTot[nMF]);
	  for (numclas=0;numclas<nbclas;numclas++)
	    {
	      if (pnDimTot[nMF] > EPSILON)
		prop=pnDim[numclas][nMF]/ pnDimTot[nMF];
	      else
		prop=0.;
	      if (flagDisplay) printf("in Entropy dim=%d,nMF=%d,numclas=%d,Fuzzy prop.=%g\n",dim,nMF,numclas,prop);
	      //  proportion could be zero for a given class
	      if (prop > EPSILON) 
		entroMF[nMF]-=prop*log(prop)/LOG2;
		      
	    }
	      
	  enDim+=entroMF[nMF]*pnDimTot[nMF];
	  pnT+=pnDimTot[nMF];
	} // end of MF loop (for the dim variable)
      if (flagDisplay) printf("end of Entropy enDim=%f\t pnT=%f\n",enDim,pnT);
      delete [] conj;
      return(enDim);

    } //end try block
  catch(...)
    {
      throw;
    }



}

