//*********************************************************************
//
//
//                              FISTREE.CPP
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISTREE classes for fuzzy decision tree
//
//*******************************************************************
//
//
//                FISTREE General use functions
//                FISTREE is a derived class of FIS
//
// get and set properties, print node, print tree, 
//*******************************************************************

#include "fistree.h"

//-----------------------------------------------------------------------------------------------
// NODE class functions
//-----------------------------------------------------------------------------------------------

int NODE::StatnbNode = 0;

int NODE::GetNChildren() const
{
  return (NChildren);
}

int NODE::GetNChildrenPrev() const
{
  return (NChildrenPrev);
}

double NODE::GetEn() const
{
  return (En);
}

int NODE::GetOrderNum() const
{
  return (OrderNum);
}

int NODE::GetVar() const
{

  return(Numvar);
}
int NODE::GetMF() const
{

  return(NumMF);
}

int NODE::GetMajClass() const
{
  return (MajClass);
}

void NODE::SetMajClass(int clm)
{
  MajClass=clm;
}


int NODE::GetNumChildC() const
{
  int ret;
  ret=NumChildC;
  return (ret);
}

void NODE::SetNumChildC(int num)
{
  NumChildC=num;
  //printf("\nin SetNumChildC NumChildC=%d\n",NumChildC);
}

int NODE::GetTLDim() const
{
  return(ListDimSize);
}

int NODE::GetLeaf() const
{
  // pure node
  return(Leaf);
}
int
 NODE::GetLeafPrev() const
{
  // was it a pure node before Pruning?
  return(LeafPrev);
}

void NODE::SetLeaf(int feu)
{
  Leaf=feu;
  if (Leaf == 1) NChildren=0;
}

void NODE::SetLeafPrev(int feu)
{
  LeafPrev=feu;
  if (LeafPrev == 1) NChildrenPrev=0;
}
int * NODE::GetListDim()
{
  return(ListDim);
}

void NODE::SetListDim(int * ListV, int NbV)
{
  int k;
  ListDimSize=NbV;
  if (NbV>0)
    {
      ListDim=new int[NbV];
      for (k=0;k<NbV;k++) ListDim[k]=ListV[k];
    }
}

NODE * NODE::GetFather()
{
  return (Father);
}

NODE * NODE::GetChild(int num)
{
  if (num < NChildren && num > -1 ) 
    return (Child[num]);
  else
    return (NULL);
}

NODE * NODE::GetChildPrev(int num)
{
  if (num < NChildrenPrev && num > -1 ) 
    return (ChildPrev[num]);
  else
    return (NULL);
}

void NODE::SetChild(int num, NODE * adrchild)
{
  if (num < NChildren && num > -1) 
    {
      Child[num]=adrchild;
      ChildPrev[num]=adrchild;
    }
}

double NODE::AND(double mu1,double mu2, char * conj)
{
  //inherits FIS AND operator
  //! implemented:  'min', 'prod', 'luka' (max(0, mu1 + mu2 - 1))
  try
    {
      if (!strcmp(conj,"min"))
	{
	  if (mu2<mu1) return (mu2);
	  else return (mu1);
	}
      else if (!strcmp(conj,"prod"))
	{
	  return(mu1*mu2);
	}
      else if (!strcmp(conj,"luka"))
	{
	  if ((mu1+mu2) >=1.0)
	    return(mu1+mu2-1.0);
	  else
	    return(0.0);
	}
      else throw runtime_error("error~unknown~AND~operator~\n");
    }
  catch(...)
    {
      throw;
    }
    

      
      
}

double NODE::IMPLI(double mu1,double mu2, char* conj)
{
  //should inherit FIS implication operator
  return(AND(mu1,mu2,conj));
}


void NODE::CreateChildren(int nbch)
{
  //Create ptr on children nodes
  NChildren=nbch;
  NChildrenPrev=NChildren;
  if (nbch>0) 
    {
      Child=new NODE * [nbch];
      ChildPrev=new NODE * [nbch];
    }
}

void NODE::RemoveChildNode(int nch, int flagDisplay)
{	
  //Set children related data
	int k=0, i;
	//look for the child node  which is going to be deleted
	while(k<NChildren && (*Child[k]).GetOrderNum()!= nch) k++;
	// delete child node
	if (flagDisplay) printf("\nremoving node %d \t Child %d of node %d ",nch,k,OrderNum);

	// if first removal of a child, keep previous info for future destruction
	if (PrExam==0)
	  {
	    NChildrenPrev=NChildren;
	    LeafPrev=Leaf;
	    PrExam=1;
	    for (i=0;i<NChildren;i++)
	      ChildPrev[i]=Child[i];
	  }


	//remove the node from the Children node array
	if (k < (NChildren-1))
	  {
	    while(k<NChildren-1)
	      {
		Child[k]=Child[k+1];
		k++;
	      }
	    Child[NChildren-1]=NULL;
	  }
	//decrement the number of children
	NChildren--;
	//case when the father node becomes a leaf
	if(NChildren == 0) Leaf = 1;
	//decrement current child node number
	NumChildC--;
	
}
void NODE::RemoveAllChildren(int flagDisplay)
{
  int i;

  // do not delete child nodes - only update father node info
  if (flagDisplay) printf("\nremoving all subnodes of node %d ",OrderNum);
  // update node info - keep previous info for future destruction
  if (PrExam==0)
    {
      NChildrenPrev=NChildren;
      LeafPrev=Leaf;
      PrExam=1;
      for (i=0;i<NChildren;i++)
	ChildPrev[i]=Child[i];
    }
  NChildren = 0;
  Leaf = 1;
}

void NODE::SetEn(double en)
{
  En=en;
}

void NODE::SetPropn(double propn)
{
  Propn=propn;
}

void NODE::SetProps(double * propnode, int nbcl)
{
Propn=0.;
if (nbcl>0)
      {
	for (int i=0;i<nbcl;i++)
	  {
	    PnClas[i]=propnode[i];
	    if (PnClas[i] >Propn)
	    Propn=PnClas[i];
	  }
      }
}


void NODE::UpdateListDim(int * listDimBefore,int taille, int nVarToRemove,int * listDimAfter)
  //*********************************

{
  // removes a variable from listedimavant
  // variables are sorted in ascending order
  // NVarToRemove=-1: by convention listDimBefore=listDimAfter
  int k=0;
  int kk;
    
  if (nVarToRemove == -1)
    for (k=0;k<taille;k++) listDimAfter[k]=listDimBefore[k];
  else
    {
      kk=0;
      for (k=0;k<taille-1;k++)
	{
	  if (listDimBefore[k] == nVarToRemove) kk++;
	  listDimAfter[k]=listDimBefore[kk];
	  kk++;
	}
    }

} // end of UpdateListDim function







//-----------------------------------------------------------------------------------------------
// FISTREE class functions
//-----------------------------------------------------------------------------------------------
//! converts crisp output (with classes in dataCol) into fuzzy output with discrete MFs
//! return the new fuzzy output
//! argument outputClasses is modified and contains the class labels (double values)
//!     trueClasses argument = 1: no tolerance on class values
//!                            0: tolerance on class values which implies normalization
//!                               in that case datacol must be reassigned to Examples array (to avoid roundoff problems)
//! because of normalization-unnormalization class labels can be slightly different from original data values
//! argument outputNumber >0: output parameters are cloned from existing output
//!                       =-1: no cloning
FISOUT * FISTREE::Crisp2DiscreteFuzClassesFromData(double * dataCol, double *& outputClasses,
						   int trueClasses, int outputNumber)
						   
  //******************************************

{
  double *tmpOutputClasses;
  FISOUT * newout;
  double minc, maxc, ratio,medianc, meanc, deltac, defv;
  int n_range,i, k,discrete, classif,res,cont;
  int * noccur;
  char * mmdefuz, *disjT, *nameO;
 

 
  outputClasses=NULL;
  tmpOutputClasses=NULL;
 
  // check if output number is correct
  // check if output is crisp
  try{
    if((outputNumber > NbOut-1))
      throw std::runtime_error("~Error~in~Crisp2FuzClassesFromData~Output~invalid~number\n~");
   
    if ( outputNumber >=0)
      {
	// use output range
	minc = Out[outputNumber]->min();
	maxc = Out[outputNumber]->max();
       
      }
    else
      {
	// use data range 
	StatArray(dataCol,Snbrow,0, medianc, meanc, deltac, minc, maxc,0);
      }
    if (trueClasses == 0)
      {
	// normalize temporary output to use a proper tolerance for unique values
	ratio = maxc - minc;
	if (ratio < EPSILON)
	  //throw std::runtime_error("~Error~in~Crisp2FuzClassesFromData~:~only~one~class~in~data\n~");
	  throw std::runtime_error("~Error~in~Crisp2FuzClassesFromData~:output~range~too~small\n");
	else
	  for ( i = 0 ; i < Snbrow ; i++ ) dataCol[i]=(dataCol[i] - minc)/ratio;
   
   
	// sort temporary output and keep unique values
	// (tolerance = EPSILON*ratio) - put result in tmpOutputClasses
	res=SortUniq(dataCol,Snbrow, &tmpOutputClasses, &noccur, Snbclass, EPSILON);
	// check if number of classes is not too high
	if (Snbclass > MAXCLASSES) 
	  throw std::runtime_error("~Error~in~Crisp2FuzClassesFromData~:~Too~many~classes~in~data\n");
	// reassign label values to datacol to take account of tolerance
	// because of roundoff errors this must be done before unnormalizing data values
	 cont=true;
	for ( i = 0 ; (i < Snbrow) && (cont==true); i++ )
	  {
	    cont=true;

	    for (k=0;k<Snbclass;k++)
	      if (fabs(dataCol[i]-tmpOutputClasses[k])<= EPSILON)
		{
		  dataCol[i]=tmpOutputClasses[k];
		  cont=false;
		}
	  }   
   
	// unnormalize classes found by SortUniq
	for ( i = 0 ; i < Snbclass ; i++ ) tmpOutputClasses[i]=(tmpOutputClasses[i] * ratio) + minc;
	// unnormalize datacol values
	for ( i = 0 ; i < Snbrow ; i++ ) dataCol[i]=(dataCol[i]* ratio) + minc;
      } //end of approximate classes
    else
      {
	// exact class values - no tolerance - no change in datacol
	InitUniq(dataCol,Snbrow ,tmpOutputClasses , Snbclass);
      }
    // keeps the outputClasses values within the output range - n_range=final number of classes
    outputClasses= new double[Snbclass];
    n_range=0;

    for( i=0 ; i<Snbclass ; i++ )
      if( (tmpOutputClasses[i] >= minc) && (tmpOutputClasses[i] <= maxc) )
	{
	  outputClasses[n_range] = tmpOutputClasses[i];
	  n_range++;
	}
  
    Snbclass=n_range;
    if (Snbclass == 0)
      throw std::runtime_error("~Error in Crisp2FuzClassesFromData~:~No~classes~in~data~within~data~range~\n");
    // create a new fuzzy output with as many discrete MF's as classes
    // assign  same disj type to new fuzzy output
    // defuz type is meanmax
 
    mmdefuz=new char[1+strlen(OUT_FUZZY::MeanMaxDefuz())];
    disjT= new char[4];
    strcpy(mmdefuz,OUT_FUZZY::MeanMaxDefuz());
    if (outputNumber >= 0)
      {
	strcpy(disjT,Out[outputNumber]->Disjunct());
	defv=Out[outputNumber]->DefaultValue();
	nameO=new char[1+strlen(Out[outputNumber]->Name)];
	strcpy(nameO,Out[outputNumber]->Name);
	classif=Out[outputNumber]->Classification();//crisp output - keep classif flag
      }
    else
      {
	// choose default parameters, disj=sum, classif=yes
	strcpy(disjT,"sum");
	defv=meanc;
	nameO=new char[7];
	strcpy(nameO,"cr2fuz");
	classif=1;
      }
   
    discrete=1;
    newout = new OUT_FUZZY(discrete,outputClasses,Snbclass, minc, maxc, 
			   mmdefuz, disjT, 
			   classif,defv);

    newout->SetName(nameO);
    // give the MF a name
    for (i=0;i<Snbclass;i++)
      {
	sprintf(nameO,"%s%d","Cl",i+1);
	newout->GetMF(i)->SetName(nameO);
      }
    if (tmpOutputClasses != NULL) delete [] tmpOutputClasses;
    if (disjT != NULL) delete [] disjT;
    if (nameO != NULL) delete [] nameO;
    if (mmdefuz != NULL) delete [] mmdefuz;
  
    return(newout);
  }
 
  catch(std::runtime_error){
   
    throw;
  }
    

}



int * FISTREE::Alloc1DIntWorkingArray(int n, int ival)
{
  //allocates 1d double array and initialize all elements to ival
  int k;
  int *ivec;
  
  ivec=new int[n];
  for (k=0;k<n;k++)
    ivec[k]=ival;
  return ivec;
}

  
double ** FISTREE::Alloc2DDoubleWorkingArray(int nrow, int ncol)
{
  //allocates 2d double array and initialize all elements to 0.
  int k,kk;
  double ** mat;

  mat=new double*[nrow];
  for (k=0;k<nrow;k++)
    {
      mat[k]=NULL;
      if (ncol >0) mat[k]=new double[ncol];
      for (kk=0;kk<ncol;kk++)
	{
	  mat[k][kk]=0.; 
	}
    }
  return mat;
        
}

double * FISTREE::Alloc1DDoubleWorkingArray(int n)
{
  //*************************************************
  //allocates 1d double array and initialize all elements to 0.
  int k;
  double *vec;

  vec=new double[n];
  for (k=0;k<n;k++)
    vec[k]=0.;
  return vec;
}

void FISTREE::SetRoot(NODE * ptrn)
{
  // updates the tree root ptr
  Sroot=ptrn;
}


NODE *  FISTREE::GetRoot(){return (Sroot);}

int FISTREE::GetNbex() const {return (Snbrow);}

double ** FISTREE::GetExamples() {return(Examples);}

int FISTREE::GetNbClass() const {return (Snbclass);}

void FISTREE::ResetExamples(char * valFile)
{
  //*************************************************
  // reset Example ptr in FISTREE, after change data file -> validation file
  int oldSnbcol;

  oldSnbcol=Snbcol;
 
      //destroy data array
  if (Examples !=NULL)
    {
      for (int i=0;i<Snbrow;i++) 
	{
	  if (Examples[i] != NULL) delete [] Examples[i];
	}
       delete [] Examples;
      Examples=NULL;
    }

     //!read sample file, update Snbcol and Snbrow fields
    Examples = ReadSampleFile(valFile,  Snbcol, Snbrow);
    if (oldSnbcol != Snbcol) throw std::runtime_error("wrong~number~of~columns~in~validation~file\n");
      
}// end ResetExamples

void FISTREE::RuleString(int * ruleArray, char * ruleStr, int ruleLen)
  //*************************************************
  // Rebuilds in ruleStr the array describing the R rule
  // can be passed as an argument of rule constructor
  // The array char ruleStr must be preallocated
  // the rule length is given in ruleLen
{
  int j;
  char * tmp;

  tmp=new char[ruleLen+1];

  sprintf(tmp, "%2d%c", ruleArray[0], SEPARE);
  strcpy(ruleStr,tmp);
  
  //set rule premise and init conclusion to ANY
  for(j = 1; j < (NbIn+NbOut); j++) 
    {
      sprintf(tmp, "%2d%c", ruleArray[j], SEPARE);
      strcat(ruleStr, tmp);
    }

   
  //end of RuleString function
  if (tmp != NULL) delete [] tmp;
}

double  FISTREE::calcAver(double *tab, int n)
{
	double s;
	int i;
	
	s = 0.0;
	if (n >0){
	  for(i=0; i<n;i++) s += *(tab+i);
	  s/=n;
	}
	return(s);
}//end calcAver

void NODE::Analyze(double ** x, int nbex, FISTREE * fistree,
		   double mumin, double muthresh,int flagDisplay)

{
  int mis,counter;
  
  //*************************************************
  // for each leaf, average and max number of attracted items nbAtItems,nbAtItemsMax
  //                average and max fuzzy cardinality,mucumAtItems,mucumAtItemsMax
  //                average standardized entropy/deviance stdED weighted with fuzzy card.
  
  if (Leaf==1)
    {
      mis=PerfClassif(x,nbex,counter,fistree,mumin,muthresh,flagDisplay); //counter=number of node attracted items
      (fistree->nbAtItems)+=counter;
      if (counter>fistree->nbAtItemsMax)
	(fistree->nbAtItemsMax)=counter;
      //Mucum=node fuzzy cardinality
      (fistree->mucumAtItems)+=MuCum;
      if (MuCum>fistree->mucumAtItemsMax)
	(fistree->mucumAtItemsMax)=MuCum;
      (fistree->stdED)+=En*MuCum; 
    }
  
}

void NODE::Print(double ** x, int nbex, FISTREE * fistree,
		 double mumin, double muthresh,int flagDisplay, int prof, FILE *f)

{
  //*************************************************
  int i,k,mis,counter,flagLabel;
  FISIN ** in;
  FISOUT ** out;
  int dataOutputCol,outputN,nbclas,classif=0;
  
  // get info from tree
  in=(*fistree).In;
  out=(*fistree).Out;
  outputN=(*fistree).GetOutputNumber();
  dataOutputCol=outputN+((*fistree).GetNbIn());
  nbclas=(*out[outputN]).GetNbMf();
  classif =(*fistree).GetClassif();
  //
  
  printf("\n");
  
  for (k=0;k<prof;k++)
    printf("    ");
  
  //print var and MF labels if we can, numbers if we can't
  flagLabel=0;
  if (Numvar>=0 && Numvar < (fistree->GetNbIn()))
      if(NumMF>=0 && NumMF < (in[Numvar]->GetNbMf()))
	flagLabel=1;

  if (flagLabel==1)
    {
      if (NChildren >0) 
 	printf("%d) node   (var %s, mf %s) ",OrderNum,in[Numvar]->Name,in[Numvar]->GetMF(NumMF)->Name);//var and MF labels
      else
	printf("%d) *leaf   (var %s, mf %s) ",OrderNum,in[Numvar]->Name,in[Numvar]->GetMF(NumMF)->Name);//var and MF labels
    }
  if (flagLabel==0)
    {
      if (NChildren >0) 
	printf("%d) node   (var %d, mf %d) ",OrderNum,Numvar+1,NumMF+1);
      else
	printf("%d) *leaf   (var %d, mf %d) ",OrderNum,Numvar+1,NumMF+1);
    }
  
  printf(" remaining variables: (");
  for (k=0;k<ListDimSize;k++)
    printf("%d ",ListDim[k]);
  printf(") ");
    
  if (Father !=NULL) 
    printf("father node=%d",(*Father).OrderNum);
  else
    printf("root");
  
  if (NChildren >0) 
    {
      if (NChildren >1) 
	printf(", %d children\n",NChildren);
      else
	printf(", %d child\n",NChildren);
    }
  else
    printf(", no children\n");
  
  for (k=0;k<prof;k++)
    printf("    ");
  //  classification case, printing maj.class for the node
  // else mean, std
  if (classif) 
    printf("majority class =%d\t Smu=%f \t Prop=%f\tEn= %f",MajClass+1,Propn*MuCum,PnClas[MajClass],En);
  else
    printf("mean=%f\t std=%f",Propn,En);
 
    
  for (k=0;k<prof;k++)
    printf("    ");
  //  classification case, printing fuzzy prop. for each class in the node
  if (classif) 
    {
      printf("\n");
      for (i=0;i<nbclas;i++)
	printf("class=%d, fuzzy prop.=%f\t",i+1,PnClas[i]);
    }
  
   mis=PerfClassif(x,nbex,counter,fistree,mumin,muthresh,flagDisplay);
   printf("\n");
   for (k=0;k<prof;k++)
     printf("    ");
   printf("attracted item count = %d\ttotal membership=%f",counter,MuCum);

  //  classification case
  if (classif)
    printf("\tmisclassified count = %d\n",mis);
 
      
}

void NODE::PrintVisu(double ** x, int nbex, FISTREE * fistree,
		   double mumin, double muthresh, char datasep, FILE *f)
{
  //*************************************************
  // print tree summary in file f
  //calls PerfClassif and calcMuSigma2
  FISIN ** in;
  FISOUT ** out;
  int outputNumber,dataOutputCol,nbclas;
  int counter=0,mis,classif=1,classifTree=1;
  double mucum,ave,sigma2;
  char * conj=NULL;
  MF * mf;

  // get info from tree
  in=(*fistree).In;
  out=(*fistree).Out;
  outputNumber=(*fistree).GetOutputNumber();
  dataOutputCol=outputNumber+((*fistree).GetNbIn());
  nbclas=(*out[outputNumber]).GetNbMf();
  conj=(*fistree).TypeConj();
  classifTree =(*fistree).GetClassif();
  classif =(*out[outputNumber]).Classification();//flag classif determines output result type
  
  
  fprintf(f,"%d%c",OrderNum,datasep);//node#

  //search for split  variable label
 
  if (Numvar>=0)
    fprintf(f,"%s",in[Numvar]->Name);//label var for split
  else
    fprintf(f," "); //blank space if no label
  fprintf(f,"%c",datasep);

  
  if (Numvar>=0 && NumMF>=0)
    {
      mf=(*in[Numvar]).GetMF(NumMF);
      fprintf(f,"%s",(*mf).Name);//MF label
    }
  else
    fprintf(f," "); //blank space if no label
  fprintf(f,"%c",datasep);

  if (Father !=NULL)
    fprintf(f,"%d%c",Father->OrderNum,datasep);
  else
    fprintf(f,"%d%c",0,datasep);

  if (Leaf)
    fprintf(f,"%d%c",1,datasep);
  else
    fprintf(f,"%d%c",0,datasep);

  

  mis=PerfClassif(x,nbex,counter,fistree,mumin,muthresh,0);//count items attracted by node
  fprintf(f,"%d%c",counter,datasep);

  

  if (classif)
    {
      fprintf(f,"%g%c",En,datasep);
      // classification - maj. class + class fuzzy proportions
      fprintf(f,"%d",MajClass+1); //majority class#
      for (int i=0;i<nbclas;i++)
	fprintf(f,"%c%g",datasep,PnClas[i]);
    }
  else
       {
	 // regression - mean + std
	 if (classifTree) // recalculate mean and std
	   {
	     fprintf(f,"%g%c",En,datasep);
	     ave=CalcMuSigma2(Numvar,NumMF,x,nbex,dataOutputCol,mumin,
			      in,conj,sigma2,mucum,counter);
	     if (sigma2 >=0) 
	       sigma2=sqrt(sigma2);
	     else
	       sigma2=0.;
				 
	     fprintf(f,"%c%g%c%g",datasep,ave,datasep,sigma2);
	   }
	 else // regression tree - mean and std already stored in node
	   {
	     fprintf(f,"%g%c",En*En,datasep); // print deviance
	     fprintf(f,"%c%g%c%g",datasep,Propn,datasep,En);
	   }
       }
 
  fprintf(f,"\n"); 

  
}



void FISTREE::Print(FILE *f) const
  //
{
  int i;
  
  printf("\nReminder: total node count in tree = %d\n",(*Sroot).StatnbNode);
  printf("------------------------------\n");
  for(i = 0; i < NbRules; i++)   Rule[i]->Print(f);
  printf("\n");
  
  
}

void FISTREE::ResetAnalyze()
  //
  //reset arrays before pruning
{
  nbAtItems=0;
  nbAtItemsMax=0;
  mucumAtItems=0.;
  mucumAtItemsMax=0.;
  stdED=0.;
}
void FISTREE::AnalyzeTreeWriteH(FILE *fres)
{
  //
  int nbact,i,p;

if (fres==NULL)
  throw std::runtime_error("~not~a~valide~file~in~AnalyzeTree\n~");
nbact=0;
 for (i=0;i<NbIn;i++)
   nbact+=In[i] -> IsActive();
// write header
 fprintf(fres, "AttItAv & AttItMax & FcardAv & FcardMax & EDweAv ");
 for (p=0;p<nbact;p++)
   fprintf(fres, "& (VarNum) & Times & avRank ");
 fprintf(fres, "&");
}

void FISTREE::AnalyzeTree(FILE *fres, int depth, int flagDisplay)
  //
{
  int i,p,compt,action,nbact;
  double avdep,enRoot,avn,mucn;

  if (fres==NULL)
    throw std::runtime_error("~not~a~valide~file~in~AnalyzeTree\n~");
  //compute tree depth
  nbact=0;
  for (i=0;i<NbIn;i++)
    nbact+=In[i] -> IsActive();
  if (depth <= 0) 
    depth=nbact;
  if (depth >nbact)
    depth=nbact;

  action=5; //browse tree and analyze tree
  if(Sroot==NULL)
    throw std::runtime_error("no~root~in~tree~\n");
  compt=UpDownTree(Sroot,action,flagDisplay,depth); // calls analyze function on each node and fills FISTREE summary fields
  
  // for each leaf, average and max number of attracted items
  //                average and max fuzzy cardinality
  //                average standardized entropy/deviance weighted with fuzzy card.
  //En=Entropy-Deviance standardized by root ED
  enRoot=Sroot->GetEn();
  if (enRoot >EPSILON)
    stdED=stdED/enRoot;
  if (mucumAtItems>EPSILON)
    stdED=stdED/mucumAtItems;
  avn=0.;
  mucn=0.;
  if(NbRules>0)
    {
      stdED=stdED/NbRules;
      avn=(double)nbAtItems/NbRules;
      mucn=mucumAtItems/NbRules;
    }
  
  // write info
  fprintf(fres, "%8.1f & %d & %8.2f & %8.2f & %8.2f ",
	  avn,nbAtItemsMax,mucn,mucumAtItemsMax,stdED);
 
  for (p=0;p<NbIn;p++)
    {
      avdep=0;
      if (countVar[p]>0)
	avdep=(double)(depthVar[p]+countVar[p])/(double)countVar[p];
      fprintf(fres,"& (%d) & %d & %8.2f ",p+1,countVar[p],avdep);
    }
  fprintf(fres, "& ");
  
  if (flagDisplay)
    {
      //!!!!verifier pour l'elagage si c'est bien NbRules (regles inactives)!!!!!!!!!!!!!!!!!!!!
      printf("\nSummary tree info\n");
      printf("\nFor leaves\n");
      printf("att.items average=%8.1f,att.items max=%d\n",avn,nbAtItemsMax);
      printf("fuzzy card. average=%8.2f,fuzzy card. max=%8.2f\n",mucn,mucumAtItemsMax);
      printf("weighted average std. entropy/deviance=%8.4f\n",stdED);
	
      printf("\nFor variables\n");
      for (p=0;p<NbIn;p++)
	{
	  avdep=0;
	  if (countVar[p]>0)
	    avdep=(double)(depthVar[p]+countVar[p])/(double)countVar[p];
	  printf("Var %d (%s), appears %d times, average rank=%8.2f\n",p,In[p]->Name,countVar[p],avdep);
	}
      
    }
}// end AnalyzeTree

int FISTREE::CalcMaxMf()
  //
  
{
  int i,maxmf;
  maxmf=0;
  for (i=0; i < NbIn; i++)
    {
      if ((In[i] -> GetNbMf()) > maxmf)
	maxmf=In[i] -> GetNbMf();
    }
  return(maxmf);
}


void  FISTREE::InitRoot(double ** w_PnDim,double * w_PnDimTot,double ** propDim, double * w_EnDim,int * clasmaj,int flagDisplay=0)
  //
{
  double pnTot=0., enTot=0., tempo=0.,avRoot,sigma2Root,mucumRoot;
  int numclas,countRoot;

  try {
    if (Sroot == NULL) throw std::runtime_error("~error~in~~InitRoot\n~");
    //init node counter
    Sroot->StatnbNode=1;

    if (Classif == 1) //classification case
      {
	//compute entropy...
	//2nd argument dim=-1 for root, 3rd argument nbMF=1 will only use propDim[0][], w_pnDimTot[0], w_EnDim[0]
	enTot=Entropy(Sroot,-1,1,w_PnDim,w_PnDimTot,pnTot,w_EnDim,flagDisplay);
	//normalize root entropy and fuzzy memberships
	if (pnTot >EPSILON) 
	  enTot/=pnTot;
	for(numclas=0;numclas<Snbclass;numclas++)
	  if (fabs(w_PnDimTot[0]) >EPSILON) 
	    w_PnDim[numclas][0]=w_PnDim[numclas][0]/w_PnDimTot[0];
	//root majority class
	tempo=0.;
	for (numclas=0;numclas<Snbclass;numclas++)
	  if (w_PnDim[numclas][0]>tempo) 
	    {
	      tempo=w_PnDim[numclas][0];
	      clasmaj[0]=numclas;
	    }
	// update root node data
	(*Sroot).SetMajClass(clasmaj[0]);
  	(*Sroot).SetEn(enTot);
	(*Sroot).SetPropn(1.);
	(*Sroot).SetMuCum(pnTot);
	// transfer for size compatibility
	for (numclas=0;numclas<Snbclass;numclas++)
	  propDim[0][numclas]=w_PnDim[numclas][0];
	(*Sroot).SetProps(propDim[0],Snbclass);
      }
    else // regression case
      {
	avRoot=(*Sroot).CalcMuSigma2(-1,0,Examples,Snbrow,OutputN+NbIn,Smumin,In,TypeConj(),sigma2Root,mucumRoot,countRoot);
	(*Sroot).SetPropn(avRoot);
	(*Sroot).SetMuCum(mucumRoot);
	if (sigma2Root >EPSILON)
	  (*Sroot).SetEn(sqrt(sigma2Root));
	else
	  (*Sroot).SetEn(0.0);
      }
  }
  catch (std::runtime_error &e)
    {
      printf("error in tree - root node could not be allocated\n");
      printf("\n%s\n", e.what());
      throw;
    }
}

void  FISTREE::SetRootVar(int flagDisplay=0)
  //
{
  int k,kk,nbact;
  int * listeV=NULL;
  try
    {
      // make list of active variables
      nbact=0;
      for (k=0;k<NbIn;k++) 
	{
	  nbact+=In[k]->IsActive();
	}
      if (flagDisplay) printf("\nin tree, %d active variables \n",nbact);
      listeV=new int[nbact];
      if (listeV==NULL) throw std::runtime_error("~error~in~SetRootVar\n~");

      kk=0;
      for (k=0;k<NbIn;k++) 
	{
	  if (In[k]->IsActive()) 
	    {
	      listeV[kk]=k;
	      if (flagDisplay) printf("\n variable %d is active\n",k);
	      kk++;
	    }
	}
      (*Sroot).SetListDim(listeV,nbact); 
      
      delete [] listeV;
    }
  catch( std::exception &e )
    {
      throw;
    }
}

int FISTREE::CreateSubNodes(int mindim,NODE * wnode, int * leaf,int * supnode,int * clasmaj,
			    double * pnDim,double ** propDimCl,double * enDim,
			    double * mucum,int depth, int flagDisplay=0)
  // create subnodes of NODE wnode for split mindim if the split generates subnodes
  // CreateChildren function allocates array of children node pointers for wnode
  // Each child node itself is created here
  // leaf,supnode, clasmaj,pnDim,enDim, mcum are arrays of size equal to the number of subnodes created
  // propDimCl = 2D array (previous size x #class)
  // mucum:       sum of cumulated mu
  // classif:     pnDim=sum of mu for maj.class, pnDim/mucum prop of mu for maj.class, propDimCl prop of mu for each class, enDim node entropy
  // regression:  pnDim=average value, EnDim=node standard deviation, other fields not filled
  // does not modify wnode pointer
  // returns leafnode: if 1 wnode is a leaf
{
  int k,kk,maxchild,nbchildren,nbc,numvm;
  int leafnode=false;
  double * dtemp=NULL;
  NODE * nodetempo;
  try
    {
      if (mindim == -1) 
	{
	  // node will not be split-no children - update Leaf field=1
	  nbchildren=0;
	  //	  (*wnode).CreateChildren(nbchildren);
	  (*wnode).SetLeaf(1);
	  (*wnode).SetLeafPrev(1);
	  leafnode=true;
	}
      else
	{
	  nbchildren=(*In[mindim]).GetNbMf();
		  	  
	  for (k=0;k<nbchildren;k++) 
	    {
	      if (flagDisplay && supnode[k]) {
		printf("_____________________________\n");
		printf("Warning !!!!!!!!!!!!!!!!!!!!!!!!!!!! node #%d \t of %d children\t will not be created ",k,nbchildren); 
		printf("\n_____________________________\n");
	      }
	    }
	  //if a node is useless (PnDim=0), it will not be built
	  maxchild=nbchildren;
	
	  for (k=0;k<maxchild;k++) 
	    if (supnode[k]) nbchildren--; 
	  if (nbchildren<=0) // case when no subnode was created
	    {
	      (*wnode).SetLeaf(1);
	      (*wnode).SetLeafPrev(1);
	      leafnode=true;
	    }
	  // update node info on children
	  // create ptr on children
	  (*wnode).CreateChildren(nbchildren);
	  //
	  nbc=-1;
	  numvm=mindim;
	  // Pn,En are deduced from return arguments of ChooseDim
	  // loop for creating child nodes - leaf[k] indicates if subnode k is a leaf (must not be split)
	  kk=-1;
	  for (k=0;k<maxchild;k++) 
	    {
	      if (!supnode[k])
		{
		  kk++;
		  nodetempo=NULL;
		  if (Classif)
		    nodetempo=new NODE(Classif,mindim,k,leaf[k],clasmaj[k],pnDim[k],propDimCl[k],enDim[k],mucum[k],wnode,Snbclass,nbc,numvm);
		  else
		    nodetempo=new NODE(Classif,mindim,k,leaf[k],clasmaj[k],pnDim[k],dtemp,enDim[k],mucum[k],wnode,Snbclass,nbc,numvm);
		  if (nodetempo == NULL) throw std::runtime_error("~error~in~~CreateSubNodes\n~");
		  if (flagDisplay) printf("\n----> creating node (%d,%d), number %d, child of %d depth %d",
					  mindim,k,(*nodetempo).GetOrderNum(),(*wnode).GetOrderNum(), depth); 
		  (*wnode).SetChild(kk,nodetempo);
		  
		}
	    }
	  // end of loop for creating child nodes
	} // end of else on mindim=-1
    } // end of try
  catch( std::exception &e )
    {
      throw;
    }
  return(leafnode);
}

void FISTREE::NextNode(NODE *& wnode, int & depth, NODE * start)
  // in tree building 
  // searches for the next node to be split (up and right)
  // from start node (default is root)
{
  int goUp=true, childToSplit=true, pureNode=false;
  int numchild, nbchildren;;
  NODE * nodetempo=NULL, * father=NULL;

  try
    {
      if (start !=NULL) father=(*start).GetFather(); // NULL if root was starting point
      else throw std::runtime_error("~error~in~NextNode\n~");
    
      while(goUp)
	{
	  if (wnode != NULL) wnode=(*wnode).GetFather();
	  depth--;
	  if (wnode != father) // if start (NULL) we are at starting point (tree root)
	    {
	      nbchildren=(*wnode).GetNChildren(); // number of children of father node
	      numchild=(*wnode).GetNumChildC(); // current node (has just been dealt with)
	      
	      childToSplit=true;
	      depth++;

	      while (childToSplit )
		{
		  // passage to next child (immediate right)
		  // except if it is pure (skipped)
		  numchild++;
		  if (numchild >= nbchildren) 
		    {
		      childToSplit =false; // no more children, go up again
		      depth--;
		    }
		  else
		    {
		      (*wnode).SetNumChildC(numchild);
		      nodetempo=(*wnode).GetChild(numchild);
		      pureNode=(*nodetempo).GetLeaf();
		      if (!pureNode) 
			{
			  childToSplit=false;
			  goUp=false; // we found the node to be split
			  // update remaining variables
			  wnode=nodetempo;
			}
		    } //end of else (remaining child nodes)
		  
		} // end of while for passing to next child
	      numchild=0; //to restart at left in upper node
	    } // end of if on wnode not null
	  else
	    goUp=false; //we are at tree root
	} //end of goUp while 
    }
  catch(std::runtime_error &e)
    {
      printf("error in NextNode\n");
      printf("\n%s\n", e.what());
      throw;
    }

}

void FISTREE::InitUpDownTree2(int & lRule, int *& tabRule, double & concRule, char*& cRule, int flagDisplay=0)
{
  // called by UpDownTree at beginning of function
  // action =2 : generate rules
  // allocates and updates NbRules Rule NodeRule LnodeRule FISTREE fields 
  // allocates, initializes and returns lRule cRule tabRule used for holding rules
  //concRule will hold the conclusion of the rule, for the chosen output 
  
  if (flagDisplay >=1) 
    printf("\n______________________________\ngenerated FIS has %d rules\n______________________________\n",Snbleaf);
  
      
  //first delete rules and rule ptr if they exist
  if (Rule != NULL)
  {
    for (int i=0;i<NbRules;i++)
    {
      if (Rule[i] != NULL) delete Rule[i];
      Rule[i]=NULL;
    }
   delete [] Rule;
   Rule=NULL;
  }

  //update number of rules in FIS
  NbRules=Snbleaf; 
  Rule=new RULE* [NbRules];
  for (int i=0;i<NbRules;i++)
    Rule[i]=NULL;
	
  // allocate NodeRule correspondence array
  // first delete it if already allocated
  if (NodeRule != NULL) delete [] NodeRule;
  NodeRule=new int[NbRules];
  LnodeRule=Snbleaf;

  //  number of MFs limited to 99 each (2 digits for integers)
  lRule=(NbIn+NbOut)*3; //rule length=MF length+separator length
  // conclusions = double numbers 
  // only one output is concerned, the others are set to 1
  
  if (cRule != NULL) delete [] cRule;
  cRule=new char[lRule+1]; //null character at end of string
  if (tabRule != NULL) delete [] tabRule;
  tabRule=new int[NbIn+NbOut];
  //init tabRule
  for (int kk=0;kk<NbIn;kk++)
    tabRule[kk]=ANY;
  for (int kk=0;kk<NbOut;kk++)
    {
      if (kk == OutputN)
	tabRule[NbIn+kk]=1; //output for learning
      else
	tabRule[NbIn+kk]=1;// other fuzzy/crisp outputs
    }
  //init concRule
  concRule=1;

   
}


void FISTREE::InitUpDownTree3(FILE * fres, NODE * wnode, int flagDisplay=0)
  //action == 3: write summary file
  {
    int classif, fuzzy;
    char * space=NULL;

  classif = Out[OutputN]->Classification(); //flag classif
  fuzzy = !strcmp(Out[OutputN]->GetOutputType(),OUT_FUZZY::OutputType()); // 1 for fuzzy
    if (classif)
      {
	//two cases: 1=discrete output classes, 2=fuzzy output classes
	//test nature, if classif - Snbclass
	fprintf(fres,"%d",-Snbclass); // class count in first line of file
	// classes have names only for a fuzzy output
	if (fuzzy)
	  {
	    for (int i=0;i<Snbclass;i++)
	      fprintf(fres,"%c %s", SEPARE, (Out[OutputN]->GetMF(i))->Name); 
	  }
	 
	fprintf(fres,"\n");
      }
    else
      fprintf(fres,"%d\n",0);//regression nbclasses=0

      
    if (flagDisplay) printf("\nwriting summary file\n");
    (*wnode).PrintVisu(Examples,Snbrow,this,Smumin,Smuthresh,SEPARE,fres);
  
  }



void FISTREE::InitUpDownTree014(int action,  NODE * wnode, int depth, int flagDisplay=0)
{
  // called by UpDownTree at beginning of function
  // action =0 : count leaf nodes
  // action =1 : print nodes
  // action =4 : destroy all nodes
  // action =5:  analyze nodes - no init action
  //
  if (action ==0)
    {
      // count leaves from starting point
      if (flagDisplay) printf("\ncounting leaves\n");
      //nbleaf=0;
    }
 
  if (action ==1)
    (*wnode).Print(Examples,Snbrow,this,Smumin,Smuthresh,flagDisplay,depth);

  
  if (action == 4 && flagDisplay)
    {
      printf("\ndestroying tree\n"); 
    }
  if (action == 5 && flagDisplay)
    printf("\nanalyzing tree\n");
}

void FISTREE::InternalNodeAction(int action, NODE * wnode, int & numvari, int * tabRule, 
				 FILE * fres,int depth, int flagDisplay=0)
{
  //action =1:   print nodes
  //action =2:   generate rule premise - tabRule et numvari are modified and returned
  //action =3:   write summary file
  //action =5:   analyze tree
  int numMFi;
  
  if (action ==1)
    (*wnode).Print(Examples,Snbrow,this,Smumin,Smuthresh,flagDisplay,depth);
  
  if (action ==2)
    {
      //update  premise relative to node
      // returns node numvar et numMF
      numvari=(*wnode).GetVar();numMFi=(*wnode).GetMF();
      //shift by one because ANY=0 for the interface
      if(numvari >=0 && numvari < (NbIn+NbOut))
	tabRule[numvari]=numMFi+1;
    }

  if (action == 3)
    (*wnode).PrintVisu(Examples,Snbrow,this,Smumin,Smuthresh,SEPARE,fres);

  if (action ==5)
    (*wnode).Analyze(Examples,Snbrow,this,Smumin,Smuthresh,flagDisplay);
	      
}

int FISTREE::LeafNodeAction(int action, NODE *& wnode, NODE * start,int & ruleCounter, int & nbleaf, int numvari, int lRule, int * tabRule, 
			     double & concRule,char * cRule, int flagDisplay=0)
{
  // action =0 count leaves
  // action =2 the corresponding rule must be added
  // first generate the string
  // modifies arguments ruleCounter, nbleaf, wnode (for node destruction)
  // action =4 delete leaf (no pruning) and sets wnode to NULL


  int ret=0;
  
  if (action ==0) nbleaf++;
	  
  if (action ==2)
    {
      ruleCounter++;
      if (Classif)
	concRule=1+(*wnode).GetMajClass();
      else
	concRule=(*wnode).GetAverageValue();
	  
      // prepare string cRule for rule premise and rule conc. init
      RuleString(tabRule,cRule,lRule);
      //printf("\napres RuleString cRule=%s\t lRule=%d\n",cRule,lRule);
      // call the FIS RULE constructor
      if (ruleCounter < NbRules) 
	{
	  Rule[ruleCounter] = new RULE( NbIn, In, NbOut, Out, cConjunction, cRule);
	  //reset rule conclusion for chosen output to estimated value
	  Rule[ruleCounter]->SetAConc(OutputN,concRule);
	  // update NodeRule corresp. array
	  NodeRule[ruleCounter] = (*wnode).GetOrderNum();
	}
      	  
      // update array element for the next rule
      concRule=1;
      tabRule[numvari]=ANY;

    }
   if (action == 4 ) 
     {
       // all info will be lost after node destruction
       ret=DeleteLeaf(wnode,start,flagDisplay);
     }
   
   return(ret);
      
}

int FISTREE::NextNodePlusAction(NODE *& wnode, NODE * upNode, int & depth, NODE * start, int action,
				int & ruleCounter, int & nbleaf, int & numvari, int lRule, 
				int * tabRule, double & concRule,char * cRule, 
				FILE * fres, int flagDisplay=0 )
  // in tree  going through
  // searches for the next node to be split (up and right)
  // from start node (default is root)
  // action =4: node destruction - prev values are used for NChildren and Leaf instead of current ones
{
  int goUp=true, nextChildHasChildren=true, leafNode=false;
  int numchild, nbchildren;
  int  ret=0;
  NODE * nodetempo=NULL, * father=NULL, * curnode=NULL;

  try
    {
      if (start !=NULL) father=(*start).GetFather(); // NULL if root was starting point
      else throw std::runtime_error("~error~in~NextNode\n~");

      while(goUp)
	{
	  curnode=wnode;
	  // go up
	  if (wnode != NULL) 
	    wnode=(*wnode).GetFather();
	  else
	    {
	    if (action==4)
	      wnode=upNode; // case when curnode has just been deleted, no info on Father
	    }
	  depth--;
	  // action=4 - delete node if leaf node or not leaf node when all children have been examined
	  // curnode=NULL: do not delete node
	  if (action == 4) 
	    {
	      if (curnode != NULL) 
		{
		  ret=DeleteLeaf(curnode,start,flagDisplay);
		  if (ret==1) return(ret); // we came back to starting point - exit function
		}
	    }
	  
	  //
	  if (wnode != father) // if father of start (NULL) we are at starting point (tree root)
	    {
	      if (action == 4)
		nbchildren=(*wnode).GetNChildrenPrev(); // number of children of father node
	      else
		nbchildren=(*wnode).GetNChildren();
	      numchild=(*wnode).GetNumChildC(); // current node (has just been dealt with)
	      
	      nextChildHasChildren=true;
	      depth++;

	      while (nextChildHasChildren )
		{
		  // passage to next child (immediate right)
		  // except if it is pure (skipped)
		  numchild++;
		  if (numchild >= nbchildren) 
		    {
		      nextChildHasChildren =false; // no more children, go up again
		      depth--;
		      //case when node has no more children to examine
		      //action =4 : destroy node once we will have gone up 
		      //even if it is not a leaf
		      if (action ==2) UpdateTabRule(wnode,tabRule);
		    }
		  else
		    {
		      nodetempo=wnode;
		      (*wnode).SetNumChildC(numchild);
		      if (action == 4)
			{
			wnode=(*wnode).GetChildPrev(numchild);//next child
			leafNode=(*wnode).GetLeafPrev();
			}
		      else
			{
			  wnode=(*wnode).GetChild(numchild);
			  leafNode=(*wnode).GetLeaf();
			}
		      // action on internal/leaf node
		      InternalNodeAction(action,wnode,numvari,tabRule,fres,depth,flagDisplay);
		      // action on leaf node
		      if (leafNode) 
			{
			  ret=LeafNodeAction(action,wnode,start,ruleCounter,nbleaf,numvari,lRule,tabRule,concRule,cRule,flagDisplay);
			  if (ret==1) return(ret); // we came back to starting point - exit function
			  wnode=nodetempo; // return to father
		
			}
		      else
			{
			  nextChildHasChildren=false;
			  goUp=false; // we found a node that has descendants - go back to main loop
			}
		    } //end of else (remaining child nodes)
		  
		} // end of while for passing to next child
	      //numchild=0; //to restart at left in upper node
	    } // end of if on wnode not null
	  else
	    goUp=false; //we are at tree root
	} //end of goUp while 
      return(0);
    }
  catch(std::runtime_error &e)
    {
      printf("error in NextNode\n");
      printf("\n%s\n", e.what());
      throw;
    }

}

int FISTREE::DeleteLeaf(NODE *& curnode, NODE * start, int flagDisplay=0)
{
  // arguments
  //          curnode = node to delete
  //          start = tree starting point
  //          curnode=NULL: node has already been deleted - do not delete again
  //          else delete node and set curnode=NULL
  int numdel,numstart;
  
  try
    {
      // do not delete twice the same node
      if (curnode!=NULL) 
	{
	  numdel=(*curnode).GetOrderNum();
	  numstart=(*start).GetOrderNum();
	  if (flagDisplay) printf("deleting node %d\n",numdel);
	  delete curnode;
	  curnode=NULL; // set ptr to NULL - will serve as protection
	  if (numdel == numstart) return(1); //back to starting point (or root)
	}
     return(0); 
    }
  
catch(std::runtime_error &e)
{
  printf("error~in~DeleteLeaf\n");
  printf("\n%s\n", e.what());
  throw;
}

}
void FISTREE::UpdateTabRule(NODE * wnode, int * tabRule)
//
{
  int numvari;
  numvari=(*wnode).GetVar();
  //protection for root
  if (numvari >=0 && numvari < (NbIn+NbOut))
    tabRule[numvari]=ANY;
}
  //*********************    End of FISTREE.CPP   ********************
