//*********************************************************************
//
//
//                              TESTFISTREE.CPP

//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISTREE main program for testing fuzzy decision tree
//*******************************************************

#include "fistree.h"

int main(int argc, char ** argv)
{
  
  if(argc < 3)
    {
      printf("\nThe %s program needs 2 arguments : ", argv[0]);
      printf("\n\t* the FIS configuration file name");   
      printf("\n\t* the data file name") ;
      
      printf("\n\tOPTIONS (no blank space after the option name)");
      printf("\n\t* -o the output number  (default=first output=0)") ;
      printf("\n\t* -s the membership significance threshold for data items (default=0.2)") ;
       printf("\n\t* -x the minimum leaf cardinality (default=min(10,#rows/10))") ;
      printf("\n\t* -t the tolerance in case of ambiguous classification at node (default=0.1)") ;
      printf("\n\t* -d the maximum tree depth (default or 0=maximum depth)");
      printf("\n\t* -g the minimum relative entropy/deviance gain for splitting nodes (default=1e-6)");
      printf("\n\t* -e0 for absolute entropy gain, -e1 for relative entropy gain (default=absolute)");
      printf("\n\t* -p0 (default=no pruning), -p1 (tree will be pruned using a performance criterion-full split removal), -p2 (tree will be pruned using a performance criterion-leaf pruning)");
      printf("\n\t* -lthe relative performance loss tolerated during pruning (default=0.0)");
      printf("\n\t* -cthe minimum coverage level required during pruning (default=0.90)");
      printf("\n\t* -vthe validation file (default=data file)");
      printf("\n\t* -zthe validation file for pruning (default=data file)");
      printf("\n\t* -a for detailed display" );
      printf("\n\t* -a0 for scarse display \n" );
      printf("Three cases:");
      printf("\n\t1-classif=yes and (discrete MFs or crisp output): crisp classes in tree");
      printf("\n\t2-(classif=yes or no) and fuzzy output with non discrete MFs: fuzzy classes in tree");
      printf("\n\t3-classif=no and crisp output: node purity criterion is based on output standard deviation instead of misclassified items\n");
      printf("\n\t* -wl for wordless.\n");
      
      return 1;
    }
  

  double mumin=0.2,muthresh=0.1,muperf,relPerfLoss=0.0,minEDgain=0., covThresh=0.90;// threshold  values
  int minleafcard=0; 
  double pf, coverage, maxError;
  double * outputClasses=NULL;
  int  FlagDisplay=0,FlagScarseDisplay=0, compt=0,maxdep=0, len, pruning=0,nbrem=0;
  //int compt1=0;
  char * outputFisConfigName=NULL;
  char * outputPFisConfigName=NULL;
  char * sumTreeName=NULL;
  char * sumPTreeName=NULL;  
  char * validFile=NULL;  
  char * prvalidFile=NULL;
  char * charResult=NULL;
  double ** Examples = NULL;   // Ptr on data
  double * Item = NULL;    // one item
  int * misClass=NULL; // holds classif results
  double * Lab=NULL;// labels for classif results
  int i,j,nbcol,nbrow,grel=0,maxMF;
  int outputNumber=0,dataOutputCol,split;
  int nbactrules, nbtotrules;
  FILE *fres=NULL;
  FISOUT * crispout=NULL;
  FISTREE  * S=NULL;
  bool wordless=false;
  

  // get and transform arguments -assign default values

  if(argc > 3) {
    for (i=3; i<argc;i++)
      {
	//display
	if(! strcmp("-a", argv[i]) )
	  {
	    FlagDisplay = 1;
	    FlagScarseDisplay=2;
	  }
	else if(! strcmp("-a0", argv[i]) )
	  {
	    FlagDisplay = 0;
	    FlagScarseDisplay=1;
	  }
	// wordless
	else if(! strncmp("-wl", argv[i], 3) ) wordless = true;
	//output number - default=0
	else if(! strncmp("-o", argv[i], 2)) outputNumber=strtol(argv[i]+2,NULL,10);
	//significant mu
	else if(! strncmp("-s", argv[i], 2)) mumin=strtod(argv[i] + 2, NULL);
	// minimum leaf cardinality
	else if(! strncmp("-x", argv[i], 2)) minleafcard=strtol(argv[i] + 2, NULL,10);
	//tolerance threshold
	else if(! strncmp("-t", argv[i], 2)) muthresh=strtod(argv[i] + 2, NULL);
	//max depth
	else if(! strncmp("-d", argv[i], 2)) maxdep=strtol(argv[i]+2,NULL,10);
	//rel or abs gain
	else if(! strncmp("-e", argv[i], 2)) grel=strtol(argv[i]+2,NULL,10);
	//pruning
	else if(! strncmp("-p", argv[i], 2)) pruning=strtol(argv[i] + 2, NULL,10);
	//rel. loss
	else if (! strncmp("-l", argv[i], 2)) relPerfLoss=strtod(argv[i] + 2, NULL);
	else if (! strncmp("-c", argv[i], 2)) covThresh=strtod(argv[i] + 2, NULL);
	// validation files
	else if(! strncmp("-v", argv[i], 2) )  validFile = argv[i] + 2;
	else if(! strncmp("-z", argv[i], 2) )  prvalidFile = argv[i] + 2;
	//entropy/deviance minimum gain
	else if(! strncmp("-g", argv[i], 2)) minEDgain=strtod(argv[i] + 2, NULL);
	
      }
  }
  if(wordless)
    {
      FlagDisplay=0;
      FlagScarseDisplay=0;
    }
  if(outputNumber<0) outputNumber=0;
  if(mumin<=0.) mumin=0.;
  if(mumin>1.) mumin=1.;
  if(muthresh<=0.) muthresh=0.;
  if(muthresh>1.) muthresh=1.;
  if (maxdep <=0) maxdep=0; 
  if (grel >1 || grel <0) grel=0;
  if (pruning <0) pruning=0;
  if (minEDgain <=0.) minEDgain=EPSILON;
  if (minEDgain >1.) minEDgain=1.;
    
  if (!wordless)
    {
      printf("\n________________________________________________________________________________________________________________\n");
      if (maxdep >0) 
	printf("Fuzzy tree paremeters:   mu min.=%g\tmu tol.=%g\tmaximum depth of tree=%d\nminimum Entropy/Deviance gain=%g\n",mumin,muthresh,maxdep,minEDgain);
      else
	printf("Fuzzy tree paremeters:   mu min.=%g\tmu tol.=%g\tmaximum depth of tree=number of active input variables\nminimum Entropy/Deviance gain=%g\n",mumin,muthresh,minEDgain);
  
      printf("chosen output=#%d\t",outputNumber+1);
      if(validFile != NULL)
	printf("\t validation file for performance%s=",validFile);
      if (pruning==0) 
	printf("\t no pruning");
      else
	if(prvalidFile != NULL)
	  printf("\t validation file for pruning%s=",prvalidFile);

      if (pruning==1){
	printf("\t full split pruning with perf. criterion");
	printf("\nrelative performance loss allowed for pruning=%g",relPerfLoss);
	printf("\nminimum coverage level required for pruning=%g",covThresh);
      }
      if (pruning==2){
	printf("\t leaf pruning with perf. criterion");
	printf("\trelative performance loss allowed for pruning=%g",relPerfLoss);
      }
  

      if (FlagDisplay==1) printf("\nDetailed Display");
      printf("\n_________________________________________________________________________________________________________________\n");
    }

  // generate file names
  // same length for suffixes
  len=strlen(argv[1]) + 10;
  outputFisConfigName=new char[len];
  strcpy(outputFisConfigName,argv[1]);
  strcat(outputFisConfigName,".tree.fis");
  
  sumTreeName=new char[len];
  strcpy(sumTreeName,argv[1]);
  strcat(sumTreeName,".sum.tree");

  if (FlagDisplay) {
    printf("\ngenerated file names:\n");
    printf("\toutputFisConfigName: %s \t",outputFisConfigName);
    printf("sumTreeName: %s \n",sumTreeName);
  }

  // generate file names for pruned tree  
  
  if (pruning >0)
    {
      len=strlen(argv[1]) + 16;
      outputPFisConfigName=new char[len];
      strcpy(outputPFisConfigName,argv[1]);
      strcat(outputPFisConfigName,".prunedtree.fis");
      
      sumPTreeName=new char[len];
      strcpy(sumPTreeName,argv[1]);
      strcat(sumPTreeName,".prunedsum.tree");
      if (FlagDisplay) {
	printf("\ngenerated file names for pruned tree:\n");
	printf("\toutputPFisConfigName: %s \t",outputPFisConfigName);
	printf("sumPTreeName: %s \n",sumPTreeName);
      }
    }
  
  
  // first build tree from data
  // main loop for catching exceptions in called functions
  try
    {
      muperf=mumin; // mu for perf calculation - can be different from mumin-extra argument to add
      // Construct FISTREE S from config file and data file
      // with a ptr on root node
      // warning!!!
      // S will be built without rules 
      // (even if rules are given in config file argv[1])
      // to use it as a sif after rules will be added, initpossibles must be called.
      // tree is based on inputs and first output
      // output must have classes
      // for fuzzy output, classes are MF labels
      // for crisp output, classes are taken from data
      // in both cases, classif=yes will make a classification tree
      //              classif=no will make a regression tree   
      
      // add minleafcard - July 2010
      S=new FISTREE(argv[1],argv[2],mumin,muthresh,outputNumber,outputClasses, crispout,covThresh,minleafcard);
      if (!wordless)
	{
	  printf("\nMinimum leaf cardinality:%d",S->GetSminleafcard());
	  printf("\n-----------------------------------");
	  if (S -> GetClassif()) 
	    {
	      printf("\nClassification tree\t");
	      if (grel==0) printf("absolute entropy criterion");
	      if (grel==1) printf("relative entropy criterion");
	      printf("\n");
	    }
	  else
	    printf("\nRegression tree\n");
	}

      maxMF=S->CalcMaxMf();
      if (maxMF >99) throw std::runtime_error("~error~too~many~MFs~in~input~partitions\n~");
      if (S->GetNbClass()>99) throw std::runtime_error("~error~too~many~classes~in~output\n~");
      // crispout is NULL, except if output was crisp
      // in that case the crisp output is turned into a fuzzy discrete output
      // and saved in crispout, for perf. calculations
      // (printing or pruning)
      // tree building is base on the fuzzy output
     
      nbcol =  S->GetNbIn() + S->GetNbOut();
      Item = new double [nbcol];
      Examples=S ->GetExamples();
      nbrow=S->GetNbex();
      
 
      if(FlagScarseDisplay || FlagDisplay) 
	{
	  // print some infos on fistree
	  printf("\n %d input variables\t %d rows \t %d classes in dataset\n",
		 S->GetNbIn(),nbrow,S->GetNbClass() );
	  if(outputClasses != NULL)
	    {
	      printf("Class labels (taken from dataset):\t");
	      for(i = 0; i <S->GetNbClass();i++)
		printf("%30.20f\t",outputClasses[i]);
	    }
	  printf("\n");
	}

      //print data
     
      if(FlagDisplay) 
	{
	  printf("\nData file: %d columns and %d rows\n", nbcol, nbrow);
	  if ((nbcol*nbrow)<=200)
	    {
	      for(j = 0; j <nbrow ; j ++)   // print data
		{
	      
		  Item = Examples[j];
	      
		  printf("\nItem n %d", j+1);  
		  for(i = 0; i < (S->GetNbIn())-1; i++) //input
		    {
		      if(NbVarG >= i  && VarNameG != NULL && VarNameG[i] != NULL)
			printf("\n (%s) %8.3f", VarNameG[i],Item[i]); 
		      else 
			printf("\n (Variable %d) %8.3f", i+1,Item[i]);  
		      for (int sef=0; sef < S->In[i]->GetNbMf();sef++)
			printf("\tsef=%d,mu=%8.3f",sef,S->In[i]->GetADeg(sef,Item[i]));
		    }
		  dataOutputCol=(S->GetNbIn()) + outputNumber; //chosen output
		  // i is now the first output in the VarNameG array
		  if(NbVarG >= i  && VarNameG != NULL && VarNameG[i] != NULL)
		    printf("\n (%s) %8.3f", VarNameG[i],Item[i]); 
		  else 
		  printf("\n (Output   %d) %8.3f", 1,Item[dataOutputCol]);  
		  for (int sef=0; sef < S->Out[outputNumber]->GetNbMf();sef++)
		    printf("\tMF=%d,mu=%8.3f",sef,S->Out[outputNumber]->GetADeg(sef,Item[dataOutputCol]));
		  
		}
	    }
	  else
	    printf("no detailed display as file is too big\n");
	}

    
	
     
      // build full tree with depth maxdep
      // and write config file outputFisConfigName
      // for new FIS with generated rules from tree
      // output defuzzification type will be changed from area to sugeno
      // if output MFs are discrete
      // MAXIMUM DEPTH=NUMBER OF ACTIVE VARIABLES - computed in BuildFullTree

      compt=S->BuildFullTree(outputFisConfigName,sumTreeName,maxdep,minEDgain,
			     grel,crispout,outputClasses,FlagScarseDisplay);

      if (FlagScarseDisplay || FlagDisplay) 
	{
	  printf("\n-------------------------------------------------------------------------------------------------------");
	  if (compt >1) printf("\nFull tree has %d leaves and %d nodes",compt, NODE::StatnbNode);
	  if(compt==1) printf("\nFull tree has %d leaf and %d node",compt, NODE::StatnbNode);
	}
    
      //first allocates Classes  array (from Examples array) in the classification case
      //call only once if Examples does not change
      S->ClassifCheck(Examples,nbrow,outputNumber );
      //then allocates misclass array and updates lab pointer for classification results
      //note: classcheck could do it all in one pass, but is also calls checkconsistency
      //which is useless here
      //call only once if Examples does not change
      S->ResClassifAlloc(misClass, Lab, outputNumber);
  
      // compute performance of generated fis
      pf=S->Perf(outputNumber,  Examples, nbrow,
		      coverage,
		      maxError,
		      muperf, 
		      misClass,
		      Lab,true,NULL,0);
      if (FlagScarseDisplay || FlagDisplay ) 
	  printf("\nEquivalent FIS Performance: %f; Coverage level (Threshold %g) : %4.2f MaxError : %11.3f\n", pf, muperf,coverage, maxError);

      // store perf. on validation file and RB characteristics in result.fistree
      if (validFile==NULL) validFile=argv[2];
      pf = S->Performance(outputNumber, validFile,coverage , maxError,muperf , "perf.fistree",FlagDisplay );
      // warning: next calls to Perf will have to first reassign Lab ptr to Classes array
      // because Performance function reassigns Classes array (in defuz object) but does not reassign Lab ptr
     
      charResult = new char [30];
      sprintf(charResult, "%s", "result.fistree");
      fres = fopen(charResult, "wt");
      if (fres ==NULL)
	throw std::runtime_error("~CannotOpenResultFile~result.fistree\n");
      //fprintf(fres, "fisname & Perf & Coverage & Max Error & MuMin & ");
      S->WriteHeaderPerfRB(outputNumber, fres);
      S->AnalyzeTreeWriteH(fres);
       fprintf(fres, "\n");
      S->PerfRB(pf,coverage,maxError,outputNumber,fres);
      // warning: next calls to Perf will have to first reassign Lab ptr to Classes array
      // because PerRB function reassigns Classes array (in defuz object) but does not reassign Lab ptr

      // print specific tree characteristics in result.fistree
      S->AnalyzeTree(fres,maxdep,FlagScarseDisplay);
      fprintf(fres, "\n");
      
      //tree pruning
    
      if (pruning >0)
	{
	  if (pruning==1) 
	    split=1;
	  else
	    split=0;
	  if (FlagScarseDisplay || FlagDisplay) 
	    {
	      printf("\n-------------------------------------------------------------------------------------------------------\nPruning tree");
	      if (pruning ==1) printf(" (Full split pruning)\n");
	      else
		 printf(" (Leaf pruning)\n");
	      
	    }
	  // Lab ptr reassigned in PrunetreePerf
	  compt=S->PruneTreePerf(outputPFisConfigName,sumPTreeName,prvalidFile,
				 misClass,Lab,muperf, relPerfLoss, nbrem,
				 crispout,split,FlagDisplay);
	  if (compt == -1)
	    throw std::runtime_error("~error~in~pruning~tree\n~");
	  else 
	    {
	      Examples=S->GetExamples();
	      nbrow=S->GetNbex();

	
	      pf=S->Perf(outputNumber,  Examples, nbrow,
			 coverage,
			 maxError,
			 muperf, 
			 misClass,
			 Lab,true,NULL,0);
	      if (FlagScarseDisplay|| FlagDisplay ) //print tree performance
		{
		  if (compt >1) printf("Pruned tree has %d leaves and %d nodes\n",compt, (NODE::StatnbNode)-nbrem);
		  if(compt==1)  printf("Pruned tree has %d leaf and %d nodes\n",compt, (NODE::StatnbNode)-nbrem);
		  printf ("Nodes keep their original numbers\n");
		
		
		  printf("Equivalent FIS Performance: %f; Coverage level (Threshold %g) : %4.2f MaxError : %11.3f", pf, muperf,coverage,maxError);
		  printf("\n-------------------------------------------------------------------------------------------------------");
		  //compt1=S->UpDownTree(S->GetRoot(),1,FlagDisplay,maxdep);
		}
	      // store RB characteristics in result.fistree file
	      // temporary reset Number of rules=number of active rules
	      // keep total number of rules in nbtotrules for correct destruction
	      nbactrules=S->ComputeNbActRule();
	      nbtotrules= S->GetNbRule();
	      S->SetNbRule(nbactrules);
	      // compute performance on validation file for result.fistree
	      pf = S->Performance(outputNumber, validFile,coverage , maxError,muperf , "perf.fistree",FlagDisplay );
	      // store specific tree characteristics in result.fistree
	      S->PerfRB(pf,coverage,maxError,outputNumber,fres);
	      S->ResetAnalyze();
	      S->AnalyzeTree(fres,maxdep,FlagScarseDisplay);
	      fprintf(fres, "\n");
	      //Irb.Print(fres);
	  
	      S->SetNbRule(nbtotrules);
      
	    }
	}// end of tree pruning
	  
      fclose(fres); //close result.fistree file

      // destroy tree
      if (S != NULL) 
	{
	  if(crispout!=NULL) 
	    S->SetNbOut((S->GetNbOut())+1); // for clean destruction of both outputs
	  if (FlagScarseDisplay || FlagDisplay) printf ("Destruction of full tree\n");
	  delete S; 
	}
      // note:  do not delete crispout (it is only a ptr) - misclass and lab should also be deleted
      if (outputFisConfigName != NULL) delete [] outputFisConfigName;
      if ( sumTreeName != NULL) delete [] sumTreeName;
      if ( sumPTreeName != NULL) delete [] sumPTreeName;
      if (outputPFisConfigName != NULL) delete [] outputPFisConfigName;
      if (outputClasses != NULL) delete [] outputClasses;
      if (misClass != NULL) delete [] misClass;
      if (Item != NULL) delete [] Item;
      // Lab is only a ptr - do not delete it
      if (charResult!=NULL) delete [] charResult;
   
      return(0);
    } //end of try

  catch(std::runtime_error &e)
    {
      printf("error in fistree\n");
      printf("\n%s\n", e.what());
      if (S != NULL) 
	{
	  if(crispout!=NULL) 
	    S->SetNbOut((S->GetNbOut())+1); // for clean destruction of both outputs
	  if (FlagScarseDisplay || FlagDisplay) printf ("Destruction of full tree\n");
	  delete S; 
	}
      // note:  do not delete crispout (it is only a ptr) - misclass and lab should also be deleted
      if (outputFisConfigName != NULL) delete [] outputFisConfigName;
      if ( sumTreeName != NULL) delete [] sumTreeName;
      if (outputClasses != NULL) delete [] outputClasses;
      if (misClass != NULL) delete [] misClass;
      if (Item != NULL) delete [] Item;

      exit(-1);
    }
  //-------------------------  End of main   -------------------------
}



