//*********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISTREE functions for pruning fuzzy decision tree
// definitions are in fistree.h
//*******************************************************

#include "fistree.h"

//-----------------------------------------------------------------------------------------------
// FISTREE class functions
//-----------------------------------------------------------------------------------------------

int FISTREE::PruneTree(char* cfgOut, char* viewTree, int flagDisplay)
{
  //prune tree and create new tree
  // entropy criterion - validation file
  
  int compt,fonc=0,depth;

  // protection against problem in building tree
  // test if root node exists
  if (Sroot == NULL) return(-3);
  depth=GetNbIn();
      
  // follow and prune fuzzy decision tree from starting node depart 
  // according to naive criterion
  // on maxdepth maxdepth
  // return node count on followed tree
  
  compt=PrTree(Sroot,fonc,flagDisplay,depth);
      
  if (compt == -2) 
    return (compt);
  // Count leaves
  fonc=0;
  compt=UpDownTree(Sroot,fonc,flagDisplay,depth);
      
  // Create fuzzy rules-fonc=2
  fonc=2;
  compt=UpDownTree(Sroot,fonc,flagDisplay,depth);
  try
    {
      //create new fis config file
      FILE *fg;           
      fg = fopen(cfgOut,"wt");
      if( fg == NULL ){
	char s[150];
	sprintf( s, "~CannotOpenFisFile~: %.100s~", cfgOut);
	throw std::runtime_error(s);
      }
      else{
	PrintCfg(fg);
	fclose(fg);
      }

      //  Create summary view file-fonc=3
      fonc=3;
      fg = fopen( viewTree, "wt");
      if( fg == NULL ){
	char s[150];
	sprintf( s, "~CannotOpenSummaryFile~: %.100s~", viewTree );
	throw std::runtime_error(s);
      }
      else{
	compt=UpDownTree(Sroot,fonc,flagDisplay,depth,fg);
	fclose(fg);
      }
      return (compt);
    }
  catch(...)
    {
      throw;
    }
}


//-----------------------------------------------------------------------------
// begin of PrTree
//-----------------------------------------------------------------------------

int FISTREE::PrTree(NODE * depart, int action, int aff, int maxdepth)
  //*********************************
  // called by PruneTree
  // follows and prunes fuzzy decision tree from starting node depart 
  // uses naive criterion
  // on maxdepth maxdepth
  // returns node count on followed tree
  
  // used variables from FIS class: In, Out, NbIn, NbOut, Rule, NbRules
  
{
  int counter=0,depth=0,numfils=0;
  int dimr,prune,classif;
  int vars,leafnode,goUp,nchildren,end;
  NODE * wnode, * nodetempo, * upNode, * gdupNode;
  
  
  printf("\n-----------------\nPruning...\n-----------------\n");
  classif=(*Out[OutputN]).Classification();
  printf("during pruning classif=%d\n",classif);
  if (! classif) return (-2);
  

  vars=NbIn; // FIS output (number)
  //starting point
  wnode=depart;
  gdupNode=(*depart).GetFather(); //father of starting point= NULL if starting from root
  // update remaining variables in node
  dimr=(*wnode).GetTLDim();
  leafnode=(*wnode).GetLeaf();
  (*wnode).SetNumChildC(numfils);
   
   
  while(wnode !=gdupNode)
    {
      //left subtree
      while(dimr >=1 && (!leafnode) && wnode !=NULL && depth <maxdepth)
      	{
	  
	  
	  numfils=0;
	  //go down left

	  if ((*wnode).GetChild(numfils) != NULL) 
	    {
	      
	      (*wnode).SetNumChildC(numfils);
	      wnode=(*wnode).GetChild(numfils); 
	      (*wnode).SetNumChildC(numfils);
	      depth++;
	      //
	      //  performance of father 
	      // classif case...
	      if(aff)
		printf("...examining node number %d...\n",(*wnode).GetOrderNum());
	      prune=0;
	      //call criterion for father
	      upNode=(*wnode).GetFather();
	      prune=(*upNode).NaiveCriterion(Examples,Snbrow,this,Smumin,Smuthresh,0);
	      //correction for root (node 1)
	      if ((*upNode).GetOrderNum() == 1) prune=0;
	      
	      if(prune)
		{
		  (*upNode).SetLeaf(1);
		  (*upNode).RemoveChildNode(0);
		  leafnode=1;
		  if(aff)
		    printf("transform node %d into a leaf\n",(*upNode).GetOrderNum());
		}
	      else
		{
		  leafnode=(*wnode).GetLeaf();
		}
	    }
	    
	}//end of the second while (left subtree)
      // we came to a leaf
      

      // go up (if possible, i.e. till upNode=root or last child)
      // to deal with immediate right node
      // if no more children go up again
      goUp=true;
      
      while(goUp)
	{
      	  
	  wnode=(*wnode).GetFather(); //father

	   
	  depth--; // go up
	  if (wnode != gdupNode) // we are at root node
	    {
	      nchildren=(*wnode).GetNChildren(); // number of children of father
	      numfils=(*wnode).GetNumChildC(); // current child just been examined
	      
	      end=false;
	      depth++;
	      
	      while (!end )
		{
		  // go to next child (immediate right)
		  // except if it is a leaf, skip it
		  numfils++;
		  if (numfils >= nchildren) 
		    {
		      end =true; // no more children go up again
		      depth--;
		    }
		  else
		    {
		      (*wnode).SetNumChildC(numfils);
		      nodetempo=wnode;
		      wnode=(*wnode).GetChild(numfils);
		      leafnode=(*wnode).GetLeaf(); 
		      if (aff) 
			printf("...examining node number %d...\n",(*wnode).GetOrderNum());
		      
		      upNode=(*wnode).GetFather();
		      //call criterion for father of current node
		      prune=(*upNode).NaiveCriterion(Examples,Snbrow,this,Smumin,Smuthresh,0);
		      //correction for root
		      if ((*upNode).GetOrderNum() == 1) prune=0;
		      // prune: father becomes a leaf
		      if(prune)
			{
			  (*upNode).SetLeaf(1);
			  (*upNode).RemoveChildNode(0);
			  leafnode=1;
			  if(aff)
			    printf("transform node %d into a leaf\n",(*upNode).GetOrderNum());
			}
		      
		      
		      //
		      if (!leafnode) 
			{
			  end=true;
			  goUp=false; // no more going up
			  // update remaining variable list
			  dimr=(*wnode).GetTLDim();
			}
		      else
			{
			  wnode=nodetempo;
			  dimr=(*wnode).GetTLDim();
			}
		    } //end of else (children remain)
		  
		} // end of while (passage to next child)
	      numfils=0; //always start left
	      
	    } // end of if wnode==null
	  else
	    goUp=false; //we are at root, the end
	} //end of while (up to grandfather)
      
      
    }//end of the first while
  
  return(counter);
   
  
}//end of PrTree
////------------------------------------------------------------------------------------------------
//
// begin NaiveCriterion function
int NODE::NaiveCriterion(double ** exemples,int nbl,FISTREE * sif,double mumin,double muthresh,int aff)
{
  NODE * upNode, * fils;
  FISOUT * out;
  int i,nchildren,misfils=0,n=0,misupNode=1,counter=0,classif,outputN;
  outputN=(*sif).GetOutputNumber();
  out=(*sif).Out[outputN];
  classif = (*out).Classification(); //flag classif
  

  upNode=this;
  //classif case
  if (classif)
    misupNode=(*upNode).PerfClassif(exemples,nbl,counter,sif,mumin,muthresh,0);

  // father examines immediate children performances
  nchildren=(*upNode).GetNChildren(); // number of children
  misfils=0;
  for (i=0;i<nchildren;i++)
    {
      fils=(*upNode).GetChild(i);
      // classif case
      if (classif)
	n=(*fils).PerfClassif(exemples,nbl,counter,sif,mumin,muthresh,0);
      misfils+=n;
      if(aff)
	printf("node number %d misclassified=%d\n",(*fils).GetOrderNum(),n);
    }
  // if children improve perf we keep them
  // if not father becomes a leaf
  if (aff)
    printf("recap : node number %d father=%d, misclassified of father=%d, children misclassified sum=%d\n",
	   (*this).GetOrderNum(), (*upNode).GetOrderNum(), misupNode,misfils);
	      
	      
  if(misfils>=misupNode)
    return(1);
  else
    return(0);
}
//------------------------------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
// begin of PruneTreePerf function
//-----------------------------------------------------------------------------
int FISTREE::PruneTreePerf(char* cfgOut, char* viewTree, char * validFile,
			   int*& misClass, double*& lab, double muperf, double relPerfLoss, 
			   int & nbrem, FISOUT * oldcrispout, int split,int flagDisplay)
{
  //prune tree and create new tree
  //pruning is done according to fis perf criterion
  // follow and prune fuzzy decision tree from starting node depart 
  // according to perf criterion
  // on maxdepth maxdepth
  // return node count on followed tree
  
  int compt,compt1,fonc,depth,ir, noact;
  

  // protection against problem in building tree
  // test if root node exists
  if (Sroot == NULL) return(-3);
  depth=GetNbIn();
      
  // use validation file
  if (validFile != NULL) 
    ResetExamples(validFile); // updates Examples array
  
  // reallocates Classes array in the classification case (previous call to AnalyzeRB has changed Classes)
  ClassifCheck(Examples,Snbrow,OutputN);
  // in any case:  reallocate misclass array and update lab pointer for classification results
  // necessary because of previous call to Performance in maintree, which has reallocated Classes
  // and results in lab ptr addressing a non valid address
  ResClassifAlloc(misClass, lab, OutputN);

  fonc=1; // call to perf pruning
  nbrem=PrTreePerf(Sroot,fonc,misClass,lab,muperf,relPerfLoss,depth,split,flagDisplay);

  if (nbrem <0) 
    {
      printf("PrTreePerf returns=%d\n",nbrem);  
      return (nbrem); 
    }
  if (flagDisplay) printf("\n\n%d nodes were removed\n",nbrem);
  //print tree
  if (flagDisplay) compt=UpDownTree(Sroot,1,flagDisplay,depth);
  // Count leaves - update Leaf parameter
  fonc=0;
  compt=UpDownTree(Sroot,fonc,flagDisplay,depth);
  if (flagDisplay) printf("%d leaves in pruned tree\n",compt);
  
  // Generate fuzzy rules for pruned tree
  // Count deactivated rules
  noact=0;
  for (ir=0;ir<NbRules;ir++)
    {
      if (!Rule[ir]->IsActive())
	{
	  noact++;
	}
    }
  if (flagDisplay) printf("\n______________________________\nPruned FIS has %d rules\n______________________________\n",NbRules-noact);
  
      //create new fis config file
      FILE *fg;           
      fg = fopen(cfgOut,"wt");
      if( fg == NULL ){
	char s[150];
	sprintf( s, "~CannotOpenFisFile~: %.100s~", cfgOut);
	throw std::runtime_error(s);
      }
      PrintCfg(fg);
      fclose(fg);

     
      // for crisp output - classif flag: summary uses equivalent discrete output stored in Out[NbOut]
      if (oldcrispout !=NULL)
	{
	  // transfer disc. fuzzy output into Out[OutputN] for summary
	  // NbOut was reduced by one in FISTREE constructor
	  // update Num fields - required for InitPossibles
	  Out[OutputN]=Out[NbOut]; 
	  Out[NbOut]=oldcrispout;
	}
       //  Create summary view file-fonc=3
      fonc=3;
      fg = fopen( viewTree, "wt");
      if( fg == NULL ){
	char s[150];
	sprintf( s, "~CannotOpenFisFile~: %.100s~", cfgOut);
	throw std::runtime_error(s);
      }
      compt1=UpDownTree(Sroot,fonc,flagDisplay,depth,fg);
      fclose(fg);

      if (oldcrispout !=NULL)
	{
	  // restore crisp output in Out[outputnumber]
	  // update Num fields - required for InitPossibles
	  // NbOut was reduced by one in FISTREE constructor
	  Out[NbOut]=Out[OutputN];
	  Out[OutputN]=oldcrispout;
	}
      // reinit possible conclusions for following treatments (java, C)
      Out[OutputN]->InitPossibles(Rule,GetNbRule(),OutputN);
      return (compt);
    
 
}

int FISTREE::PrTreePerf(NODE * start, int action, int * misClass, double * lab, double muperf, double relPerfLoss, 
			int maxdepth, int split,int Flagdisplay)
  //*********************************
  
  // used variables from FIS class In, Out, NbIn, NbOut, Rule, NbRules
  // follows and prunes fuzzy decision tree from starting node start 
  // uses perf criterion
  // on maxdepth = maxdepth
  // returns node count on followed tree
  
{
  int count=-1,depth=0, ret=0;
  NODE * wnode,  * father , * upnode;
  int leaf,nchildren;
  int nbrem=0;
  int classif,prune=0, kk;
  int numChild=0;
  double perfInit=0., coverage, maxError;
  
  

  classif =(*Out[0]).Classification(); //flag classif
  
  if (Flagdisplay) {
    printf("-------------------------------------\n");
    printf("\nNoderule array \nlength=%d\n",LnodeRule);
    for (kk=0;kk<LnodeRule;kk++)
      {
	printf("Rule %d \tNodeRule[%d]=%d\n",kk,kk,NodeRule[kk]); 
      }
  }

  // starting point
  //init algorithm
  wnode=start;
  if (wnode == NULL) return(-1);//case when tree was not built
  father=(*start).GetFather(); //father of starting point : NULL if starting from root
  upnode=father;
   leaf=(*wnode).GetLeaf();
  (*wnode).SetNumChildC(numChild);
  // print initial tree performance
  if (Flagdisplay) (*wnode).Print(Examples,Snbrow,this,Smumin,Smuthresh,0,depth);
  perfInit=Perf(OutputN,  Examples, Snbrow,
		coverage,
		maxError,
		muperf, 
		misClass,
		lab,true,NULL,0);
  if (Flagdisplay) printf("\n Initial Tree Performance: %f; Coverage level (Threshold 0.2) : %4.2f MaxError : %11.3f\n", 
			  perfInit, coverage, maxError);
   
  //--------------------------------------------------------------------
  //end initialization
  //--------------------------------------------------------------------
  prune=0;
  if (Flagdisplay) {
    (*wnode).Print(Examples,Snbrow,this,Smumin,Smuthresh,0,depth);
    printf("\n First leaf node - Performance return : %f; Coverage level (Threshold 0.2) : %4.2f MaxError : %11.3f\n",
	   perfInit, coverage, maxError);
  }
  ;

  while(wnode !=father)
    {
      // left branch
      while((!leaf) && wnode !=NULL && depth <maxdepth)
	{
	  numChild=0; // always start with left child
	  //down to bottom left

	  if ((*wnode).GetChild(numChild) != NULL) 
	    {
	      (*wnode).SetNumChildC(numChild);
	      upnode=wnode;
	      wnode=(*wnode).GetChild(numChild); 
	      (*wnode).SetNumChildC(numChild);
	      depth++;
	      leaf=(*wnode).GetLeaf();
	    }
	    
	}//end of second while, the one on the left subtree
      //-----------------------------------------------------------------
      // we came down to a leaf node -  try pruning
      //-----------------------------------------------------------------
      //(*wnode).SetPrExam(1);
	prune=TryPruning(wnode,upnode,perfInit,muperf,relPerfLoss,misClass,lab,split,
			 Flagdisplay,numChild, nchildren,nbrem,depth-1);
	if (prune==-1) throw std::runtime_error("error~in~trypruning");
	// if upnode was pruned it has no more children and wnode has been deleted
	// in that case wnode was set to upnode in TryPruning
	// nchildren is upnode's number of children and was set to zero - we must go up
	// other loop 
	// to deal with immediate right child
	// if no more children keep going up
	// pruning is also done in function NextNodePlusPrune for nodes that became leaves at this step
	ret=NextNodePlusPruneLeaf(wnode,depth,start,action,
				  perfInit,muperf,relPerfLoss,misClass,lab,nbrem,split,Flagdisplay);
	if (ret==1) return(ret); // we came back to starting point - exit function
	if (wnode !=NULL) 
	  leaf=(*wnode).GetLeaf();
    }//end of first while
  //-------------------------------------------
  // prepare to exit
  count=nbrem;
    //destroy all dynamically allocated objects
   fflush(stdout);
  return(count);

}//end of PrTreePerf

int FISTREE::FindRule(NODE * wnode)
{
  // search for rule number corresponding to node number
  // in NodeRule array
  int i,ind=-1;

  for (i=0;i<LnodeRule;i++)
    {
      if ((*wnode).GetOrderNum() == NodeRule[i]) ind=i;
    }
  if (ind <0)
    throw std::runtime_error("~error~in~find~rule~");
  return(ind);
}


int FISTREE::UpdateRule(int indR, NODE * wnode, NODE * upNode, int split, int *& subR, double * lab, int Flagdisplay)
  // updates rule indR
  // split=0 leaf pruning:  rule corresponding to leaf is deactivated if there are other children, 
  //                        or modified if leaf is unique child
  //       1 split pruning: rules corresponding to other children of upnode are deactivated, 
  //                        rule corresponding to leaf is modified
{
  int deac=0,sub,nbc,ind, numRule;
  NODE * tmpNode;
  
  ind=(*upNode).GetOrderNum();
  nbc=(*upNode).GetNChildren();
  if (nbc > 1)  //wnode is a leaf and upnode has other children
    {
      deac=1;
      if (!split)
	{
	  // the corresponding rule must be deactivated
	  if (Flagdisplay) printf("\n\tRule %d corresponding to leaf node %d deactivated\n",indR,(*wnode).GetOrderNum());
	  Rule[indR]->Deactivate();
	}
      else
	{
	  
	  subR=new int[nbc-1];
	  // this rule must be modified, and the rules corresponding to other children deactivated
	  // suR holds node numbers corresponding to modified rules
	  // to be used by Backrule if Pruning is not done
	  ind=0;
	  for (sub=0;sub<nbc;sub++)
	    {
	      tmpNode=(*upNode).GetChild(sub);
	      numRule=FindRule(tmpNode); //FindRule throws an exception if no node found in NodeRule array
	      if (numRule !=indR)
		{
		  if (Flagdisplay) printf("\n\tRule %d corresponding to subnode %d must be deactivated\n",numRule,(*tmpNode).GetOrderNum());
		  Rule[numRule]->Deactivate();
		  //protection against memory error
		   if (ind < (nbc-1)) 
		     {
		       subR[ind]=numRule;
		       ind++;
		     }
		   else
		     printf("\nerror~in~UpdateRule nbc=%d,ind=%d\n",nbc,ind);
		}
	      else
		{
		  if (Flagdisplay) printf("\n\tRule %d corresponding to subnode %d must be modified\n",indR,(*wnode).GetOrderNum());
		  PruneRule(wnode,upNode,indR,lab, Flagdisplay);
		}
	    }
	}
      
    }
  else //wnode is a leaf and the only child of upnode
    {
      // split pruning or (leaf pruning when leaf is the only child)
      // update  premise and conclusion relative to upnode
      // returns node numvar and numMF
      if (Flagdisplay) printf("\n\tRule %d corresponding to subnode %d must be modified\n",indR,(*wnode).GetOrderNum());
      PruneRule(wnode,upNode,indR,lab,Flagdisplay);
    }
return(deac);
}


void FISTREE::PruneRule(NODE * wnode,NODE * upNode, int indR, double * lab, int Flagdisplay)

{
  //
  int numvari,numMFi,crisp;
  numvari=(*wnode).GetVar();
  numMFi=(*wnode).GetMF();
  crisp=strcmp(Out[OutputN]->GetOutputType(),OUT_FUZZY::OutputType());
  if (Flagdisplay) printf("\n\tRule %d  \n\t", indR);
  if (Flagdisplay) Rule[indR]->Print();
  //set corresponding premisse MF to any
  Rule[indR]->SetAProp(ANY,numvari);
  // set rule conclusion to father node maj. class if fuzzy output, to class label if crisp output and classification
  //                     to average value if crisp output and regression
  if (Classif)
    {
      if (!crisp)
	Rule[indR]->SetAConc(OutputN,1+(*upNode).GetMajClass());
      else
	Rule[indR]->SetAConc(OutputN,lab[(*upNode).GetMajClass()]);//Lab does the correspondence between class numbers and class values
    }
  else
    Rule[indR]->SetAConc(OutputN,(*upNode).GetAverageValue());
  // update correspondence rule-node array
  NodeRule[indR]=(*upNode).GetOrderNum();	
  if (Flagdisplay) 
    {
      // modif Huaque Chen to print true variable names
      if(NbVarG >= numvari  && VarNameG != NULL && VarNameG[numvari] != NULL)
	printf("\tRule %d modified \t%s -> ANY\t", indR, VarNameG[numvari]);
      else
	printf("\tRule %d modified \tVariable%d -> ANY\t", indR,numvari);
   
    }
  if (Flagdisplay) printf("\tNodeRule[%d]=%d\n",indR, NodeRule[indR]);
  if (Flagdisplay) Rule[indR]->Print();
 }


int FISTREE::BackRule(int indR, NODE * wnode, NODE * upnode,  int split, int * subR, double * lab, int FlagDisplay)
  // updates rule indR
{
  int deac,nbc,ind;
  
  nbc=(*upnode).GetNChildren();
 
  deac=0;
  if (nbc > 1)
    {
      deac=1;
      if (!split)
	{// leaf is not the only child
	  // corresponding rule must be reactivated
	  if (FlagDisplay) printf("\n\tRule %d corresponding to leaf node %d must be reactivated\n",indR,(*wnode).GetOrderNum());
	  Rule[indR]->Activate();
	}
      else
	{
	  UnPruneRule(wnode,indR,lab,FlagDisplay); // restore main rule (upnode -> wnode)
	  ind=0;
	  for (ind=0;ind<(nbc-1);ind++) // reactivate other rules (upnode -> other subnodes)
	    {
	      if (FlagDisplay)
		    printf("\n\tRule %d corresponding to subnode %d must be reactivated\n",subR[ind],NodeRule[subR[ind]]);
	      Rule[subR[ind]]->Activate();
	    }
	}
      
    }
  else
    {
      UnPruneRule(wnode,indR,lab,FlagDisplay);

    }
  return(deac);
}

void FISTREE::UnPruneRule(NODE * wnode,int indR,double * lab, int FlagDisplay)
{
  // called by BackRule
  // restore previous state of rule that was pruned by UpdateRule
  // leaf is the only child			 
  // update  premise relative to node
  // returns node numvar et numMF
  int numvari,numMFi,crisp;
  numvari=(*wnode).GetVar();
  numMFi=(*wnode).GetMF();
  crisp=strcmp(Out[OutputN]->GetOutputType(),OUT_FUZZY::OutputType());
  if (FlagDisplay) printf("\n\tRule %d  \n\t", indR);
  if (FlagDisplay) Rule[indR]->Print();
  //set corresponding premisse MF to original values
  Rule[indR]->SetAProp(numMFi+1,numvari);
  // set rule conclusion to node maj. class
  if (Classif)
    {
    if (!crisp)
      Rule[indR]->SetAConc(OutputN,1+(*wnode).GetMajClass());
    else
      Rule[indR]->SetAConc(OutputN,lab[(*wnode).GetMajClass()]);
    }
  else
    Rule[indR]->SetAConc(OutputN,(*wnode).GetAverageValue());
  if (FlagDisplay) printf("\tRule %d back to prior state \t Variable%d -> %d\t", indR,numvari,numMFi+1);
  // update correspondence rule-node array
  NodeRule[indR]=(*wnode).GetOrderNum();
  if (FlagDisplay) printf("\tNodeRule[%d]=%d\n",indR, NodeRule[indR]);
  if (FlagDisplay) Rule[indR]->Print();
}

int FISTREE::TryPruning(NODE * wnode,NODE * upnode,double perfInit,  double muperf,double relPerfLoss,int * misClass,double * lab,
			int split,int FlagDisplay, int & numChild, int & nchildren, int &nbrem, int depth )
{
  // computes new perf
  // updates or deactivates rule
  // indR= index of rule corresponding to current leaf
  // if no pruning restores previous state
  // returns integer : -1 if problem, 0 if no pruning was done, 1 if pruning was done
  // modifies the last two arguments 
  int indR,deac,numx,sub,leaf;
  //int exam;
  int prune=0,OK=false,nbc,mindim;
  int * subR;//array for subnode deactivated rules
  double maxError, coverage,ret0, ret, tesperf;

  if (upnode != NULL)
    {
      // FindRule returns the rule number corresponding to the rule
      indR=FindRule(wnode);
      if (indR <0)
	{
	  throw std::runtime_error("~error~rule~not~found~for~node");
	  //node number# %d\n",(*wnode).GetOrderNum());
	}
      
      else
	{
	  // examine upnode children
	  subR=NULL; // will be allocated in UpdateRule
	  nbc=(*upnode).GetNChildren();
	  if(nbc >0)
	    {
	      OK=0;
	      for (sub=0;sub<nbc;sub++)
		{
		  //OK if all subnodes are leaves or have been examined
		  leaf=upnode ->GetChild(sub)->GetLeaf();
		  //OK if all subnodes are leaves 
		  OK+=leaf;
		}
	    }
	  
	  if (OK<nbc)
	    {
	      prune=0;
	      if (FlagDisplay)
		printf("\nIn TryPruning Leaf node number %d-father node number %d is not OK for pruning, OK=%d, nbc=%d", 
		   (*wnode).GetOrderNum(),(*upnode).GetOrderNum(),OK,nbc);
		return(prune);
	    }
	  //OK we can try perf pruning
	  // set exam flag to true anyway
	  //upnode ->SetPrExam(1);
	  //
	  if (FlagDisplay){
	    printf("\nIn TryPruning Leaf node number %d-father node number %d is OK for pruning, OK=%d,nbc=%d", 
		   (*wnode).GetOrderNum(),(*upnode).GetOrderNum(),OK,nbc);
	    ret0=Perf(OutputN,  Examples, Snbrow,
		      coverage,
		      maxError,
		      muperf, 
		      misClass,
		      lab);
	    printf("\n Leaf node - Performance return : %f; Coverage level (Threshold 0.2) : %4.2f MaxError : %11.3f\n", 
		   ret0, coverage, maxError);
	  }
	  // modify rule corresponding to leaf node (wnode with father=upnode)
	  deac=UpdateRule(indR, wnode, upnode,split,subR,lab,FlagDisplay);
	  ComputeNbActRule(); //update active rule number
	  //Out[OutputN]->UpdatePossibles(Rule,GetNbRule(),indR); //update possible conclusion array - error does not work
	  Out[OutputN]->InitPossibles(Rule,GetNbRule(),OutputN);
   
	  ret=Perf(OutputN,  Examples, Snbrow,
		   coverage,
		   maxError,
		   muperf,
		   misClass,
		   lab);
	  if (FlagDisplay) printf("\n if pruning -> Tree performance  : %f; Coverage level (Threshold 0.2) : %4.2f MaxError : %11.3f\n",
				  ret, coverage, maxError);
	  // pruning or not pruning? test on perf. loss plus enough coverage
	  // special case perfInit=0, consider absolute loss instead of relative loss.
	  tesperf=ret-perfInit;
	  if (perfInit>EPSILON) 
	    tesperf=tesperf/perfInit;
	  if (((fabs(tesperf)) <= relPerfLoss) && (coverage >= CovThresh) )
	    {
	      prune=1;
	      // keep rule modifications
	      //update father node
	      numx = (*wnode).GetOrderNum();
	      mindim=(*wnode).GetVar();
	       if (!split) 
		{
		  (*upnode).RemoveChildNode(numx,FlagDisplay);
		  if (FlagDisplay) printf("           -> pruning leaf node (%d)\n",numx);
		  nbrem++;
		}
	       else
		 {
		   nbrem=nbrem+nbc;
		   (*upnode).RemoveAllChildren(FlagDisplay);
		   // upnode has no more children
		   if (FlagDisplay) printf("           -> pruning split node (%d)\n",(*upnode).GetOrderNum());
		 }
	      
	      //get new current node child number
	      numChild=(*upnode).GetNumChildC();
	      //get new number of children
	      nchildren=(*upnode).GetNChildren();
	      // update summary arrays for variables if father node is now a leaf
	      
	      if ((mindim>=0) && ((*upnode).GetLeaf()==1))
		{
		  countVar[mindim]-=1;
		  depthVar[mindim]-=depth;
		}
	    }//end pruning case
	  else
	    {
	      prune=0;
	      // go back to prior state
	      if (FlagDisplay) {
		numx = (*wnode).GetOrderNum();
		if (!split)
		  printf("            -> no pruning of leaf node %d)\n",numx);
		else
		  printf("            -> no pruning of split node %d)\n",(*upnode).GetOrderNum());
	      }
	      deac=BackRule(indR, wnode, upnode,split,subR,lab,FlagDisplay);
	      ComputeNbActRule(); //update active rule number                                                                                     
	      //Out[OutputN]->UpdatePossibles(Rule,GetNbRule(),indR); //update possible conclusion array - error does not work                    
	      Out[OutputN]->InitPossibles(Rule,GetNbRule(),OutputN);
	    }
	  // delete temporary array
	  if (subR != NULL) delete [] subR;
	}// end else (case indR >=0)
    } // end if (case upnode !=NULL)
  return(prune);
} //end TryPruning function

int FISTREE::NextNodePlusPruneLeaf(NODE *& wnode, int & depth, NODE * start, int action,
				   double perfInit,double muperf,double relPerfLoss,
				   int *misClass,double * lab, int & nbrem,
				   int split,int flagDisplay=0 )
  // in tree going through
  // searches for the next node to be split (up and right)
  // from start node (default is root)
{
  int goUp=true, moreChildren=true, leafNode=false;
  int numChild, nchildren;
  int prune=0;

  NODE * nodetempo=NULL, * father=NULL, * curnode=NULL,* upnode=NULL ;

  try
    {
      if (start !=NULL) father=(*start).GetFather(); // NULL if root was starting point
      else throw std::runtime_error("~error~in~NextNode\n~");
    
      while(goUp)
	{
	  curnode=wnode;
	  if (wnode != NULL) wnode=(*wnode).GetFather();
	  depth--;
	  	  
	  //
	  if (wnode != father) // if father of start (NULL) we are at starting point (tree root)
	    {
	      nchildren=(*wnode).GetNChildren(); // number of children of father node
	      numChild=(*wnode).GetNumChildC(); // current node (has just been dealt with)
	      
	      moreChildren=true;
	      depth++;

	      while (moreChildren )
		{
		  // passage to next child (immediate right)
		  numChild++;
		  if (numChild >= nchildren) //case when node has no more children to examine
		    {
		      moreChildren =false; // no more children, go up again
		      depth--;
		      // cas when node just became a leaf (was pruned in PrTreePerf)
		      if ((*wnode).GetLeaf()==1)
			{
			  if (flagDisplay) 
			    printf("_________node (%d) became a leaf as all its children were removed________\n",(*wnode).GetOrderNum());
			  // reexamine pruning which may be possible now
			  upnode=(*wnode).GetFather();
			  //(*wnode).SetPrExam(1);
			  prune=TryPruning(wnode,upnode,perfInit,muperf,relPerfLoss,misClass,lab,split,
					   flagDisplay,numChild,nchildren,nbrem,depth-1);
			  if (prune==-1) throw std::runtime_error("error~in~trypruning");
			  // if pruning was done wnode has been deleted and set to pnode
			  // nchildren has been set to 0
			}
		    }
		  else
		    {
		      nodetempo=wnode;
		      (*wnode).SetNumChildC(numChild);
		      wnode=(*wnode).GetChild(numChild);//next child
		      leafNode=(*wnode).GetLeaf();
		      
		      if (leafNode) // we came to a leaf node - try pruning 
			{
			  // examine leaf pruning
			  upnode=(*wnode).GetFather();
			  //(*wnode).SetPrExam(1);
			  prune=TryPruning(wnode,upnode,perfInit,muperf,relPerfLoss,misClass,lab,split,
					   flagDisplay,numChild,nchildren,nbrem,depth-1); 
			  curnode=wnode;
			  wnode=nodetempo; // return to father and go up again
			}
		      else // we found a node that has descendants - exit function and go back to main loop
			{
			  moreChildren=false;
			  goUp=false; 
			}
		    } //end of else (remaining child nodes)
		  
		} // end of while for passing to next child
	    } // end of if on wnode not father (null)
	  else
	    goUp=false; //we are at tree root
	} //end of goUp while 
      return(0);
    }
  catch(std::runtime_error &e)
    {
      printf("error in NextNodePlusPruneLeaf\n");
      printf("\n%s\n", e.what());
      throw;
    }

}

