//*********************************************************************
//
//
//                              TREE.CPP
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISTREE classes for fuzzy decision tree
//
//*******************************************************************
//
//
//                FISTREE General use functions
//                FISTREE is a derived class of FIS
//
//*******************************************************************

#include "fistree.h"

//-----------------------------------------------------------------------------------------------
// NODE class function
//-----------------------------------------------------------------------------------------------
NODE * FISTREE::Tree(int flagDisplay, int maxdepth, double gainEntromin, int grel)
  //*********************************
  // builds the fuzzy decision tree 
  // maximum depth maxdepth if greater than zero
  // allocates and returns the root node
  // leaf node field : 1 if node is pure (must not be split), 0 if not 
{
  int maxMF=0; // should be computed for classification case only 
  int k=0,mf=0;
  int nodeCounter=0;
  int depth=0;
  int leafnode = false;
  int dimr = NbIn;
  int mindim=-1;
  int numchild;
  int nbchildren=-1;
  int numvm=-1;
  int counter=0;
  int numex=0;
  NODE * wnode=NULL, *father=NULL;
  double ** w_PnDim=NULL, ** propDim=NULL;
  double * pnDim=NULL,  * w_PnDimTot=NULL;
  double * enDim=NULL,  * w_EnDim=NULL;
  double * muCum=NULL,  * w_muCum=NULL;
  double * ptrNULL=NULL;
  double  enAfter=0.,sigma2After=0.0,entrodevBefore=0.0;
  int * leaf=NULL, *clasmaj=NULL, *supnode=NULL;
  double pnTot=0., enTot=0.;
  double eGain,munode;
  //uses the following variables in FIS class: In , Out, NbIn, in FISTREE Sroot, Snbclass
  

  //  mu_nDim array size= item# x maxMF#
  // it is only allocated once

  // the  w... arrays are working arrays allocated here
  // init is done here
  try
    {
      maxMF=CalcMaxMf(); // max number of MFs in FIS input variables
      
      if (flagDisplay) 
	{
	  printf("\nbuilding tree with a maximum depth of %d\n",maxdepth);
	  printf("\nmaximum number of membership functions in input variables=%d\n",maxMF);
	}

      w_PnDim=NULL;pnDim=NULL;propDim=NULL;w_PnDimTot=NULL;enDim=NULL;
      w_EnDim=NULL;leaf=NULL;supnode=NULL;clasmaj=NULL;
  
      if (Classif)
	{
	  w_PnDim=Alloc2DDoubleWorkingArray(Snbclass,maxMF);
	  propDim=Alloc2DDoubleWorkingArray(maxMF,Snbclass); 
	} 

      pnDim=Alloc1DDoubleWorkingArray(maxMF);
      w_PnDimTot=Alloc1DDoubleWorkingArray(maxMF);
      enDim=Alloc1DDoubleWorkingArray(maxMF);
      w_EnDim=Alloc1DDoubleWorkingArray(maxMF);

      muCum=Alloc1DDoubleWorkingArray(maxMF);
      w_muCum=Alloc1DDoubleWorkingArray(maxMF);

      leaf=Alloc1DIntWorkingArray(maxMF,false); // init 0
      supnode=Alloc1DIntWorkingArray(maxMF,false); // init 0
      clasmaj=Alloc1DIntWorkingArray(maxMF,-1); // init -1
  
      //end of allocation and init
      //___________________________________
      // build root node and compute root entropy, enTot, pnTot, clasmaj and fuzzyprop(class)
      //___________________________________
      
      // create and initialize root node - will allocate PnClas and initialize elements to 0
      father=NULL;
      k=-1;
      if (Classif)
	Sroot=new NODE(Classif,mindim,k,leaf[0],clasmaj[0],pnTot,propDim[0],enTot,muCum[0],
		       father,Snbclass,nbchildren,numvm);
      else
	Sroot=new NODE(Classif,mindim,k,leaf[0],clasmaj[0],pnTot,ptrNULL,enTot,muCum[0],
		       father,Snbclass,nbchildren,numvm);
      if (Sroot == NULL) throw std::runtime_error("~error~in~~tree\n~");
      // init root node parameters - common to classif. and reg. cases
      InitRoot(w_PnDim, w_PnDimTot, propDim, w_EnDim, clasmaj,flagDisplay);
      SetRootVar(flagDisplay);
      //-------------------------------------------------------------------------
      //root node is complete
      //-------------------------------------------------------------------------

      //-------------------------------------------------------------------------
      //begin tree construction 
      //___________________________________
      wnode=Sroot; // point de dpart
      numchild=0;
  
  
      // two embedded while loops:
      //     the first loop generates a whole branch (the leftmost one from the current node)
      //                    and updates the current node
      //     the second one (function NextNode) goes up to find the next node to be split, starting from the new current  node
      // leafnode is true if a node is pure (must not be split
      // function called for choosing the split depends on classification-regression case
      while(wnode !=NULL)
	{
	  // generate leftmost subtree
	  while(dimr >=1 && (!leafnode) && wnode !=NULL)
	    {
	      // select dimension for splitting node
	      if (depth >= maxdepth) 
		mindim=-1;
	      else
		if (Classif)
		  mindim=ChooseDim(wnode,enAfter,clasmaj,pnDim,propDim,enDim,muCum,
				   leaf,supnode,
				    w_PnDim,w_PnDimTot,w_EnDim,grel,
				   flagDisplay);
		else
		  mindim=ChooseDimReg(wnode,sigma2After,pnDim,enDim,muCum,leaf,supnode,
				      w_PnDimTot,w_EnDim,w_muCum,
				      flagDisplay);
	      
	      
	      // check if entropy or deviance reduction is > minimum allowed
	      entrodevBefore=(*wnode).GetEn();
	      if (entrodevBefore>EPSILON)
		{
		  if (Classif)
		    eGain=(entrodevBefore-enAfter)/entrodevBefore;
		  else
		    eGain=(entrodevBefore*entrodevBefore-sigma2After)/(entrodevBefore*entrodevBefore);//beware of square (std is stored in node, not deviance)
		}
	      else
		eGain=0.;
	      if (eGain<gainEntromin)
		{
		  // split wont be done - not worth it
		  mindim=-1;
		}
	      // add minleafcard test - July 2010
	      // compute number of examples attracted by node to be splitted
	      counter=0;
	      for (numex=0;numex<Snbrow;numex++) //item loop
		{
		  munode=(*wnode).CalcMuN(Examples,numex,In,TypeConj() ); // item node membership
		  if(munode > Smuthresh) // significant membership
		    counter++; // regression or classification - # attracted examples
		}
	      if (counter <Sminleafcard)
		{
		  // split wont be done - not worth it
		  mindim=-1;
		}
	      //printf("node %d, number of attracted examples with mu>muthresh=%d\n",(*wnode).GetOrderNum(),counter);
	      
	      if(flagDisplay) 
		{
		  if (Classif)
		    {
		      printf("\nReturn from ChooseDim mindim=%d\t Entropy before=%g\t Entropy after=%g\tRelative entropy reduction=%g\n",
			     mindim,(*wnode).GetEn(),enAfter,eGain);
		      if (eGain<gainEntromin)
			printf("no split of node %d - relative entropy reduction  %g smaller than %g (minimum)\n",
			       (*wnode).GetOrderNum(),eGain,gainEntromin);
		      if (counter <Sminleafcard)
			printf("no split of node %d -   number of attracted examples %d smaller than %d (minimum)\n",
			       (*wnode).GetOrderNum(),counter,Sminleafcard);
		    }
		  else
		    {
		      printf("\nReturn from ChooseDimReg mindim=%d\tDev before=%g\t Dev after=%g\tRelative deviance reduction=%g\n",
			     mindim,entrodevBefore*entrodevBefore,sigma2After,eGain);
		      if (eGain<gainEntromin)
			printf("no split of node %d - relative deviance reduction  %g smaller than %g (minimum)\n",
			       (*wnode).GetOrderNum(),eGain,gainEntromin);
		      if (counter <Sminleafcard)
			printf("no split of node %d -   number of attracted examples %d smaller than %d (minimum)\n",
			       (*wnode).GetOrderNum(),counter,Sminleafcard);
		    }
		}
	     
	      // create subnodes - if mindim =-1 no subnode creation and returns leafnode=1
	      leafnode=CreateSubNodes(mindim, wnode, leaf, supnode, clasmaj, pnDim, propDim, enDim, muCum,depth,flagDisplay);
	       // for each active input variable update depthVar and countVar arrays
	      if (mindim>=0 && !leafnode)		{
		  countVar[mindim]+=1;
		  depthVar[mindim]+=depth;
		}
	      // dimr indicates if there are remaining variables
	      // NumChildC field = last child node dealt with
	      // passage to leftmost next child node if possible
	      // update current node = last child node dealt with
	      if (!leafnode) // there has been a split
		{
		  numchild=(*wnode).GetNumChildC();
		  if ((*wnode).GetChild(numchild) != NULL) 
		    {
		      depth++;
		      wnode=(*wnode).GetChild(numchild); 
		      dimr=(*wnode).GetTLDim();
		      leafnode=(*wnode).GetLeaf();
		    }
		}
	    }//end of second while, the one on the left subtree
	  //
	  // Find next node - immediate right child node of father node
	  // or if father node has no more children to split, go up again
	  // updates wnode and depth
	  NextNode(wnode,depth,Sroot);
	  if (wnode!= NULL) {
	    leafnode=(*wnode).GetLeaf();
	    dimr=(*wnode).GetTLDim();
	  }
	        
    }//end of first while
      nodeCounter=(*Sroot).StatnbNode;
      if (flagDisplay) printf("\ntree has  %d nodes\n",nodeCounter);
      
      /* be careful with delete ...*/
      if (Classif)
	{
	  for (k=0;k<Snbclass;k++)
	    delete [] w_PnDim[k];
	  for (mf=0;mf<maxMF;mf++)
	    delete [] propDim[mf];
   
	  delete [] w_PnDim;
	  delete [] propDim;
	}

      delete [] pnDim;
      delete [] w_PnDimTot;
      delete [] enDim;
      delete [] w_EnDim;
      delete [] leaf;
      delete [] supnode;
      delete [] clasmaj;
      delete [] muCum;
      delete [] w_muCum;
  /*...*/
 
      return (Sroot);

    } //end of try
  //---------------------------------
  // freeing memory
  //---------------------------------
  catch(std::exception &e)
    {
      for (k=0;k<Snbclass;k++)
	delete [] w_PnDim[k];
      for (mf=0;mf<maxMF;mf++)
	delete [] propDim[mf];
      delete [] w_PnDim;
      delete [] pnDim;
      delete [] propDim;
      delete [] w_PnDimTot;
      delete [] enDim;
      delete [] w_EnDim;
      delete [] leaf;
      delete [] supnode;
      delete [] clasmaj;
      delete [] muCum;
      delete [] w_muCum;
      throw;
    }
  
}//end of Tree function

//-----------------------------------------------------------------------------------------------
// FISTREE class functions
//-----------------------------------------------------------------------------------------------
// begin of  UpDownTree function
//-----------------------------------------------------------------------------------------------

int FISTREE::UpDownTree(NODE * start, int action, int flagDisplay, int maxdepth, FILE * fres)
  //*********************************
  // follows decision tree from given node start down to given depth=maxdepth
  // returns the node# of the followed tree
  // action =0 : count leaf nodes
  // action =1 : print nodes
  // action =2 : generate rules
  // action =3 : generate FIS summary
  // action =4 : destroy all nodes
  //             Prev values are used instead of current ones for number of children and leaf flag
  // action =5 : analyze tree
  
  // used variables from FIS class In, Out, NbIn, NbOut, Rule, NbRules
  
{
  int dimr,leafnode,classif,fuzzy;
  int nbleaf=0, ret=0, numchild=0,depth=0;
  int lRule=0,numvari=0,ruleCounter=-1;
  int * tabRule=NULL;
  double concRule=0.0;
  char * cRule=NULL;
  NODE * wnode, * father, * subnode, *upNode;

  
  
  try
    {
      // starting point
      //init algorithm
      ret=0;
      if (start == NULL) throw(std::runtime_error("~error~in~UpDownTree~no~tree\n~"));//case when tree was not built
  
      wnode=start;
      father=(*start).GetFather(); //starting point father: NULL if starting from root
      // update remaining dims in starting node
      dimr=(*wnode).GetTLDim();
      if (action == 4)
	leafnode=(*wnode).GetLeafPrev();
      else
	leafnode=(*wnode).GetLeaf();

      (*wnode).SetNumChildC(numchild);

      classif = Out[OutputN]->Classification(); //flag classif
      fuzzy = !strcmp(Out[OutputN]->GetOutputType(),OUT_FUZZY::OutputType()); // 1 for fuzzy
      //------------------------------------------------------------------------------
      // Init function
      //------------------------------------------------------------------------------
  
      if (action ==2) InitUpDownTree2(lRule,  tabRule, concRule,cRule, flagDisplay);
      else
	if (action ==3) InitUpDownTree3(fres,wnode,flagDisplay);
	else
	  if (action ==0 || action ==1 || action ==4|| action ==5)
	    InitUpDownTree014(action,wnode,depth,flagDisplay);
	  else
	    throw std::runtime_error("~action~unknown\n~");
      //--------------------------------------------------------------------
      //end initialization
      //--------------------------------------------------------------------
  
      while(wnode !=father)
	{
	  // left branch
	  while((!leafnode) && wnode !=NULL && depth <maxdepth)
	    {
	      numchild=0;
	      // go down to first left child
	      if (action==4)
		subnode=(*wnode).GetChildPrev(numchild);
	      else
		subnode=(*wnode).GetChild(numchild);
	      
	      if (subnode != NULL) 
		{
		  (*wnode).SetNumChildC(numchild);
		  wnode=subnode;
		  (*wnode).SetNumChildC(numchild);
		  depth++;
		  // action for internal nodes - updates numvari tabRule - writes summary in fres - prints node
		  InternalNodeAction(action,wnode,numvari,tabRule,fres,depth,flagDisplay);
		  //
		  if (action == 4)
		    leafnode=(*wnode).GetLeafPrev();
		  else
		    leafnode=(*wnode).GetLeaf();
		}
	    
	    }//end of second while, the one on the left subtree
	  //-----------------------------------------------------------------
	  // we came down to a leaf node
	  // the corresponding action must be done
	  // generate the rule ou incrementing the number of leaves
	  // updates nbleaf numvari tabRule
	  //-----------------------------------------------------------------
	  upNode=(*wnode).GetFather();
	  ret=LeafNodeAction(action,wnode,start,ruleCounter,nbleaf,numvari,lRule,tabRule,concRule,cRule,flagDisplay);
	  if (ret==1) return(ret); // we came back to starting point - exit function
	  //-------------------------------------------------------------------
	  // other loop 
	  // to deal with immediate right brother node
	  // if no more children keep going up
	  // performs some actions on each internal node and other actions on leaf nodes
	  //-------------------------------------------------------------------
	  ret=NextNodePlusAction(wnode, upNode, depth, start, action,ruleCounter,nbleaf,numvari,lRule,tabRule,concRule,cRule,fres,flagDisplay);
	  //-------------------------------------------------------------------
	  //
	  if (ret==1) return(ret); // we came back to starting point - exit function
	  if (wnode !=NULL) 
	    {
	      if (action == 4)
		leafnode=(*wnode).GetLeafPrev();
	      else
		leafnode=(*wnode).GetLeaf();
	    }
	 	
	}//end of first while - finished

      // prepare to exit
      if (action==0) 
	{
	  Snbleaf=nbleaf;
	  ruleCounter=nbleaf;
	}
  
  
      //destroy all dynamically allocated objects
      if (action==2)
	{
	  if (cRule !=NULL) delete [] cRule;
	  if (tabRule !=NULL) delete [] tabRule;  
	}
      fflush(stdout);
      return(ruleCounter);
    }
  catch (std::runtime_error &e)
    {
      printf("error in UpDownTree\n");
      printf("\n%s\n", e.what());
      throw;
    }
}//end of UpDownTree

//-----------------------------------------------------------------------------------------------
// beginning of BuildFullTree
//-----------------------------------------------------------------------------------------------

int FISTREE::BuildFullTree(char* cfgOut, char* viewTree, int depth, double minEDgain, int grel, 
			   FISOUT * oldout, double * outputClasses, int flagScarseDisplay)
  // char* dataFile argument sup?
{
  NODE * root;
  int compt=0,nbleaves=0,fonc,nbactr,mfConc,nr,nbact;
  int ii,flagDisplay=0;

  OUT_FUZZY * ptrfisoutfuz=NULL;
  OUT_CRISP * ptrfisoutcr=NULL;
  FILE *fg;   
   
  // compute number of active inputs
  nbact=0;
  for (ii=0;ii<GetNbIn();ii++)
    nbact+=In[ii] -> IsActive();
  
  if (depth <= 0) 
    depth=nbact;
  if (depth >nbact)
    depth=nbact;
  if (flagScarseDisplay==0||flagScarseDisplay==1) 
    flagDisplay=0;
  else
    flagDisplay=1;
  try
    {
      // exception handling
      // construct root node and then
      // construct full tree (with given depth)
      if (flagDisplay) 
	{
	  printf("starting to build full tree\n");
	  printf("\nNUMBER OF ACTIVE VARIABLES=%d\n",nbact);  
	}
 
      root =  Tree(flagDisplay,depth,minEDgain,grel);
      
      if (flagDisplay) printf("\nfinished building full tree\n");
      //print tree
      if (flagDisplay) compt=UpDownTree(root,1,flagDisplay,depth);
      // Count leaves
      if (flagDisplay) printf("\nstarting to count leaves in full tree");
      fonc=0;
      nbleaves=UpDownTree(root,fonc,flagDisplay,depth);
      if (flagDisplay) {
	printf("finished counting leaves in full tree\n");
	printf("%d leaves in full tree\n",nbleaves);
      }
      
      // Create fuzzy rules-fonc=2
      fonc=2;
      if (flagDisplay) printf("starting to create fuzzy rules\n");
      compt=UpDownTree(root,fonc,flagDisplay,depth);
      if (flagDisplay) printf("finished creating fuzzy rules\n");

      // update active rule number in fis
      nbactr=ComputeNbActRule();
      if (flagDisplay) printf("\n%d active rules in fis\n",nbactr);

      // Create summary view file-fonc=3
      fonc=3;
      if (flagDisplay) printf("starting to write summary file\n");
      fg = fopen( viewTree, "wt");
      if (fg ==NULL) {
	char s[250];
	sprintf( s, "~CannotOpenSummaryFile~: %.100s~", viewTree);
	throw std::runtime_error( s );
      }
      else {
	compt=UpDownTree(root,fonc,flagDisplay,depth,fg);
	fclose(fg);
	if (flagDisplay) printf("summary file written\n");
      }

      
      // print tree 
      if (flagScarseDisplay >=1) 
	compt=UpDownTree(root,1,flagDisplay,depth);

     
      // init possible values for fuzzy output - call to InitPos or CheckConsistency
      // necessary because rules were added to initial fis
      if(!strcmp(Out[OutputN]->GetOutputType(),OUT_FUZZY::OutputType()))
	{
	  ptrfisoutfuz=(OUT_FUZZY *) Out[OutputN];
	  ptrfisoutfuz->InitPossibles(Rule,GetNbRule(),OutputN);


	  // if discrete MFs in output, area defuzzification is meaningless
	  // change it for sugeno
	  // other possibility is MeanMax
	  // first delete defuz object in output, then build new one
	  if (!strcmp(ptrfisoutfuz->Defuzzify(), ptrfisoutfuz->AreaDefuz()) )
	    {
	      // check if first MF in output is discrete
	      if (ptrfisoutfuz->GetNbMf() >=0)
		{
		  if (!strcmp(ptrfisoutfuz->GetMF(0)->GetType(),"discrete"))
		    {
		   
		      if (flagDisplay) printf("\n\n output before conversion area-sugeno\n\n\n");
		      if (flagDisplay) ptrfisoutfuz->Print();
		      ptrfisoutfuz->SetOpDefuz(ptrfisoutfuz->SugenoDefuz());
		      printf("\n\n Output defuzzification type changed to sugeno because of discrete MFs");
		      if (flagDisplay) printf("\n\n output after conversion area-sugeno\n\n\n");
		      if (flagDisplay) 
			{
			  printf("Fuzzy output after area-sugeno defuz. transf.\n");
			  ptrfisoutfuz->Print();
			  printf("\n");
			}
		    }
		}
	    } // end area defuz -> sugeno defuz for discrete MFs
	}//end fuzzy output
      else
	{//crisp output
	  ptrfisoutcr=(OUT_CRISP *) Out[OutputN];
	  ptrfisoutcr->InitPossibles(Rule,GetNbRule(),OutputN);
	}
    
      // if output was modified (crisp into discrete fuzzy MFs) go back to original output configuration
      // reset rule conclusions to class labels instead of class numbers
      // pruning is based on perf and works with original output
      // oldout=original crisp output
      // store new discrete fuzzy output into Out[NbOut]
      if ( oldout != NULL )
	{
	  Out[NbOut]=Out[OutputN];//NbOut has been reduced by one in FISTREE constructor
	  Out[OutputN]=oldout;// crisp output
	  ptrfisoutcr=(OUT_CRISP *) Out[OutputN];//outputnumber
	  if (flagDisplay) 
	    {
	      printf("Crisp output after InitPossibles\n");
	      ptrfisoutcr->Print();
	      printf("\n");
	    }
	  //rules
	  for (nr=0;nr<GetNbRule();nr++)
	    {
	      mfConc= (int) Rule[nr]->GetAConc(OutputN);
	      Rule[nr]->SetAConc(OutputN,outputClasses[mfConc-1]);
	    }
	  ptrfisoutcr->InitPossibles(Rule,GetNbRule(),OutputN); // in case of pruning
	  
	}
	  
      //create new fis config file - in all cases with original output
              
      fg = fopen(cfgOut,"wt");
      if (flagDisplay) printf("starting to write config file\n");
      if (fg ==NULL) {
	char s[250];
	sprintf( s, "~CannotOpenFISFile~: %.100s~", cfgOut);
	throw std::runtime_error( s );
      }
      else{
	
	PrintCfg(fg);
	if (flagDisplay) printf("config file written\n");
	fclose(fg);
      }
      
     
      
      //julie ... to be completed with mu parameters, etc.
      //create a summary view file with more details
      //name.sum
      /*
	char * sumviewTree=NULL;
	int len=strlen(viewTree);
			
	sumviewTree = new char[len];
	len = len-5;
	strncpy(sumviewTree,viewTree,len);
	sumviewTree[len]='\0';
	sumviewTree = strcat(sumviewTree, ".log");
	printf("fic = %s", sumviewTree);
			
	FILE * fich = fopen(sumviewTree, "wt");
	int Classif;
	//classification or regression tree?
	Classif = (*Out[OutputN]).Classification();
	if(Classif==1) fprintf(fich,"classification tree\n\n");
	else fprintf(fich,"regression tree\n\n");
			
	//data file
	//fprintf(fich,"data file = %s\n\n", dataFile);
			
	//copy of fis config file
	fprintf(fich,"fis config file : \n");
			
	FILE* fisF = fopen(cfgOut, "rt");
	int c;
	while ((c=getc(fisF)) != EOF)
	putc(c, fich);
			
	fclose(fisF);
	fclose(fich);
	//julie
	*/

      return(nbleaves);
    }
  catch(...)
    {
      throw;
    }
}

