//***********************************************************************
//
//
//                              MAINRULEH.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : Test of FISFPA class functions 
//        requires library fispro, fpa (Fast Prototype Algorithm)

//**********************************************************************

#include "fpa.h"

int main(int argc, char ** argv)
  //****************************
{

  if(argc < 2)
    {
      printf("\nThe program %s needs 2  arguments : ", argv[0]);
      printf("\n\t* fis config file;");   
      printf("\n\t* the data set file.\n") ;
      printf("\nOptions: ");
      // printf("\n\t* -a for intermediate display. \n" );
      // printf("\n\t* -fconfout: the name of the output config file (default:  config.fpa)"); 
      printf("\n\t* -dy : y minimum of the matching degree (default:  0.1)");
      printf("\n\t* -n output number (default:  0)");
      // printf("\n\t* -ez : z is the minimum cardinality (default:  1)");
      // printf("\n\t* -r remove temporary files (default:  false)"); 
      printf("\n");
      return 1;
    }

   FISFPA *S = NULL;
  FILE *f = NULL;
  int outputNumber, nr, k, i, * num;
  double dmin, tmp, * index, * weight;
  double median, min, max, meani, stdi, meanw, stdw;
 
  dmin = 0.1;
  outputNumber = 0;

  if(argc > 3) 
    {
      for(i = 3; i < argc; i++)
	{	
	  if(! strncmp("-d", argv[i], 2) )  dmin = strtod(argv[i] + 2, NULL);
	  else if(! strncmp("-n", argv[i], 2) ) outputNumber = atoi(argv[i] + 2);
	}
    }

  try
    {
      S = new FISFPA(argv[1], argv[2], 0, 1, dmin);
      nr = S->GetNbRule();
      index = new double [nr];
      weight = new double [nr];
      num = new int [nr];
      k = 0;
    
      for(i = 0; i < nr; i++) 
	{
	  tmp = S->FpaARule(i, outputNumber, true);
	  if (tmp != NOT_ENOUGH)
	    {
	      index[k] = tmp;
	      weight[k] = S->Rule[i]->Weight;
	      num[k] = i;
	      k++;
	    }
	}
      
     f = fopen("result.ruleh","wt");
      if(f == NULL)
	{
	  printf("\nUnable to open config file (writing mode):  %s\n", "result.ruleh");
	  return 1;
	}

      for(i = 0; i < k; i++)
	fprintf(f, "%d & %f & %f \n", num[i] + 1, index[i], weight[i]);

      StatArray(index, k, 0, median, meani, stdi, min, max);
      StatArray(weight, k, 0, median, meanw, stdw, min, max);

      fprintf(f, "-1 & %f & %f \n", meani, meanw);
      fprintf(f, "-1 & %f & %f \n", stdi, stdw);

      fclose(f);

      delete S;
      delete [] weight; 
      delete [] index;
      delete [] num;
   }
  
  catch(std::exception &e)
    {
      printf("\n%s", e.what());
    }

  return 0;
}

//**********************   End of MAINRULEH.CPP   ***************
