//***********************************************************************
//
//
//                              OPTIMISE.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : Test of FISFPA class functions 
//        requires library fispro, fpa (Fast Prototype Algorithm)

//**********************************************************************

#include "fpa.h"

int main(int argc, char ** argv)
  //****************************
{

  if(argc < 3)
    {
      printf("\nThe program %s needs 2  arguments : ", argv[0]);
      printf("\n\t* fis config file;");   
      printf("\n\t* the data set file.\n") ;
      printf("\nOptions: ");
      printf("\n\t* -a for intermediate display. \n" );
      printf("\n\t* -fconfout: the name of the output config file (default:  configfpa.fis)"); 
      printf("\n\t* -sx : x is the strategy for computing the set of items to be used  0 for MIN, 1 for DEC (default)"); 
      printf("\n\t* -dy : y minimum of the matching degree (default:  0.3)");
      printf("\n\t* -ez : z is the minimum cardinality (default:  3)");
      printf("\n\t* -tfile data file name used for performance computation (default : data file)");
      printf("\n\t* -u activity threshold for performance computation (default: 0.1)" );
      printf("\n\t* -r remove temporary files (default:  false)"); 
      printf("\n");
       printf("\n\t* -wl for wordless.\n");
      
       return 1;
    }

  int i, FlagDisplay = 0, rm_files = 0; 
  char * ConfigSif, * testperffile;
  int strat = DEC;
  int emin = 3;
  double dmin = 0.3;
  double ret, Cov, MaxErr, thres;
  bool flagconfig=false;
  FISFPA *S = NULL;
  FIS *T = NULL;
  FILE *f = NULL;
  bool wordless=false;
  
  ConfigSif = testperffile = NULL;
  thres = 0.1;

  if(argc > 3) 
    {
      for(i = 3; i < argc; i++)
	{	
	  if(! strcmp("-a", argv[i]) ) FlagDisplay = 1;
	  else if(! strncmp("-f", argv[i], 2) )  ConfigSif = argv[i] + 2;
	  else if(! strncmp("-s", argv[i], 2) )  strat = atoi(argv[i] + 2);
	  else if(! strncmp("-d", argv[i], 2) )  dmin = strtod(argv[i] + 2, NULL);
	  else if(! strncmp("-e", argv[i], 2) )  emin = atoi(argv[i] + 2);
	  else if(! strncmp("-t", argv[i], 2) ) testperffile = argv[i] + 2;
	  else if(! strncmp("-u", argv[i], 2) ) thres = strtod(argv[i] + 2, NULL);
	  else if(! strncmp("-r", argv[i], 2) )  
	    {
	      if(strlen(argv[i]) > 2) rm_files = atoi(argv[i] + 2);
	      else rm_files = true;
	    }
	  else if(! strncmp("-wl", argv[i], 3) ) wordless = true;
	}
    }

  if(wordless)
     FlagDisplay=0;

  try
    {
      S = new FISFPA(argv[1], argv[2], strat, emin, dmin);

      for(i = 0; i < S->GetNbOut(); i++) S->FpaRules(i);

      if (! S->ComputeNbActRule()) 
	{
	  delete S;	
	  sprintf( ErrorMsg, "~NoRuleInitializedUsingThisConfiguration~\n");
	  throw std::runtime_error( ErrorMsg );
	}

      if(FlagDisplay)  S->Print();

      if(! ConfigSif) 
	{
	  flagconfig=true;
	  ConfigSif = new char [15];
	  sprintf(ConfigSif, "configfpa.fis");
	}
 
      f = fopen(ConfigSif,"wt");
      if(f == NULL)
	{
	  printf("\nUnable to open config file (writing mode):  %s\n", ConfigSif);
	  return 1;
	}
      S->PrintCfg(f);
      fclose(f);
      delete S;

      if (!wordless)
	printf("\nNew configuration file:  %s was written\n", ConfigSif);
      if(strcmp(ConfigSif, "configfpa.fis") && rm_files)  remove("configfpa.fis");

      if(testperffile == NULL) testperffile = argv[2];
      
      T = new FIS(ConfigSif);
      ret = T->Performance(0, testperffile, Cov, MaxErr, thres, "perffpa.res", false);
      f = fopen("result.fpa", "wt");
      T->WriteHeaderPerfRB(0, f);
      fprintf(f, "\n");
      T->PerfRB(ret, Cov, MaxErr, 0, f);
      fprintf(f, "\n");
      fclose(f);
      delete T;
      if (flagconfig) delete [] ConfigSif;
    }
  
  catch(std::exception &e)
    {
      printf("\n%s", e.what());
    }

  return 0;
}

//**********************   End of OPTIMISE.CPP   ***************
