//****************************************************************
//
//
//                              GENERATE.CPP
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//         handling of inactive inputs
// File : starting function for the rule generation module

// Environnment : requires library fispro and  genfis package
//****************************************************************

#include "genfis.h"

int main(int argc, char ** argv)
//******************************
{

  if(argc < 2)
    {
      printf("\nThe program %s needs 1 argument:  ", argv[0]);
      printf("\n\t* the FIS file configuration;");   
      printf("\nOptions: ");
      printf("\n\t* -a for intermediate display. \n" );
      printf("\n\t* -eSampleFile:  the sample file name.") ;
      printf("\n\t* -s:  sort the rules according to their cumulated weight over the sample.") ;
      printf("\n\t* -tVal:  cumulated weight threshold (rule whose value is lower than Val are removed (default:  0.1).");
      printf("\n\t* -fFisConfig:  the output fis config file (default:  config.fis)."); 
      printf("\n\t* -r remove temporary files (default:  false)"); 
      printf("\n");

      printf("\nThree possibilities for rule sorting:");
      printf("\n\t* lexicographic order (if no sample file)");
      printf("\n\t* according to sample order (when -e but not -s)");
      printf("\n\t* according to their cumulated weight (when -e and -s)");
      printf("\n\t* -wl for wordless.\n");
      
      printf("\n");

      return 1;
    }

  int i, alloc = 0, FlagDisplay = 0, TriCumul = 0, rm_files = 0; 
  char * ConfigSif = NULL;
  char * FichEx = NULL;
  double thres;
  FILE *f;
  bool wordless=false;
  
  

  thres = 0.1;

  if(argc > 2) 
    {
      for(i = 2; i < argc; i++)
	{	
	  if(! strcmp("-a", argv[i]) ) FlagDisplay = 1;
	  else if(! strcmp("-s", argv[i]) ) TriCumul = 1;
	  else if(! strncmp("-f", argv[i], 2) )  ConfigSif = argv[i] + 2;
	  else if(! strncmp("-e", argv[i], 2) )  FichEx = argv[i] + 2;
	  else if(! strncmp("-t", argv[i], 2) )  thres = strtod(argv[i] + 2, NULL);
	  else if(! strncmp("-r", argv[i], 2) )  
	    {
	      if(strlen(argv[i]) > 2) rm_files = atoi(argv[i] + 2);
	      else rm_files = true;
	    }
	  else if(! strncmp("-wl", argv[i], 3) ) wordless = true;
	}
     }
  
  if(wordless)
    FlagDisplay=0;
 
  GENFIS *S = NULL;
  try
    {
      S = new GENFIS(argv[1], FichEx, TriCumul, thres);
   

      if(FlagDisplay)  S->Print();

      if(! ConfigSif) 
	{
	  ConfigSif = new char [15];
	  alloc = 1;
	  sprintf(ConfigSif, "config.fis");
	}

      f = fopen(ConfigSif,"wt");
      if(f == NULL)
	{
	  printf("\nUnable to open configuration file (writing mode): %s\n", ConfigSif);
	  return 1;
	}
      S->PrintCfg(f);
          
      fclose(f);
      if(strcmp(ConfigSif, "config.fis") || !rm_files)
	 if (!wordless)
	   printf("\nConfiguration file: %s written\n", ConfigSif);

      if(!rm_files) 
	{
	  f = fopen("info.gen","wt");
	  S->Print(f);
	  fclose(f);
	  if (!wordless)
	    printf("\n'info.gen' file written\n");
	}
      else
	{
	  if(!strcmp(ConfigSif, "config.fis") )  remove("config.fis");
	  remove("info.gen");
	}
      delete S; 
      S = NULL;
      if(alloc) delete [] ConfigSif;
    }
  
  catch(std::exception &e)
    {
      delete S; 
      printf("\n%s", e.what());
    }

  return 0;
}
//**************************   End of GENERATE.CPP   ***************
