//***********************************************************************
//
//
//                              HFPFIS.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISHFP class functions used by FISPRO, 
//        part of package hfp (Hierarchical fuzzy partitioning)
//        requires library fispro
//
//**********************************************************************


#include "hfpfis.h"
#include "genfis.h"
#include "fpa.h"
#include "fiswm.h"

void FISHFP::Init(void)
  //*******************
{
  // Pointers set to NULL
  Out = NULL;
  In = NULL;
  Name = NULL;
  strDistanceType = NULL;
  fData = NULL;
  fFisCfg = NULL;
  fTest = NULL;
  fVertex = NULL;
  cConjunction = NULL;
  strHierarchy = NULL;
  OutVal = NULL;
  MisClass = NULL;
  Labels = NULL;

  // Attributes init
  NbIn = 0;
  NbOut = 0;
  NbEx = 0;
  NSort = 0;
  CumWeThres = 0.;  
  CardStrategy = DEC;
  MinCard = 3;
  MuMin = 0.3;
  NbClasses = 0;
  NmfInit = 1;
  BlankThres = 1.0;
  IterMax = 10;
  Coverage = 0.0;
  cRuleInductionMethod = new char [10];
  strcpy(cRuleInductionMethod, FISHFP::RuleInductionFpa());
}

void FISHFP::InitSystem(char * InConfig, char *dataf, char * OutConfig)
  //*******************************************************************
{
  ifstream f(InConfig);
  int i, bsize;

  if (f.fail())     
    {
      sprintf( ErrorMsg, "~CannotOpenFISFile~: %.100s~", InConfig);
      throw std::runtime_error( ErrorMsg );
    }
  
  Init();

  bsize = MaxLineSize(f);
  
  Name = new char[bsize];
  strDistanceType = new char[bsize];
  strHierarchy = new char[bsize];
  cConjunction = new char[bsize];

  Name[0] = 0;  
  strDistanceType[0] = 0; 

  ReadHdrHfp(f, bsize);

  In = new INHFP * [NbIn];
  for(i = 0 ; i < NbIn ; i++ ) In[i] = NULL;
  Out = new FISOUT * [NbOut];
  for(i = 0 ; i < NbOut ; i++ ) Out[i] = NULL;

  for(i = 0; i < NbIn; i++)  ReadInHfp(f, bsize, i);
  
  for(i = 0; i < NbOut; i++)  ReadOut(f, bsize, i);

  fData = dataf;

  if((OutConfig != NULL) && ((unsigned int) bsize < strlen(OutConfig)) ) 
    bsize = strlen(OutConfig) + 1; 
  fFisCfg = new char [bsize + 4];
  if(OutConfig == NULL) sprintf(fFisCfg, "%s.fis", Name);
  else strcpy(fFisCfg, OutConfig);
}

void FISHFP::InitSystem(char * dataf, char * config)
  //************************************************
{
  int i, col, lig;
  double ** Data;
  double * tmp;
  int bsize;

  Data = NULL;

  ifstream f(config);

  if (f.fail())     
    {
      sprintf( ErrorMsg, "~CannotOpenFISFile~: %.100s~", config);
      throw std::runtime_error( ErrorMsg );
    }

  Init();
  
  bsize = MaxLineSize(f);
  
  Name = new char[bsize];
  strDistanceType = new char[bsize];
  strHierarchy = new char[bsize];
  cConjunction = new char[bsize];

  
  Name[0] = 0;  
  strDistanceType[0] = 0; 
  strHierarchy[0] = 0;
  cConjunction[0] = 0;

  ReadHdrHfp(f, bsize);

  In = new INHFP * [NbIn];
  Out = new FISOUT * [NbOut];
  
  fData = dataf;
  Data = ReadSampleFile(dataf, col, lig);  
  NbEx = lig;
   
  tmp = new double [NbEx];

 for(i = 0; i < NbIn; i++)  ReadInHfp(f, bsize, i);
  
  for(i = 0; i < NbIn; i++)  
    {
      GetColumn(Data, NbEx, i, tmp);
      In[i]->InitCont(tmp, NbEx, Tol);
    }
 
  for(i = 0; i < NbOut; i++)  ReadOut(f, bsize, i);
  
  if(NbOut)
    {
      double vmin, vmax, diff;
      OutVal = new double [lig];
      vmin = Out[NSort]->min();
      vmax = Out[NSort]->max();
      diff = vmax - vmin;
      if(diff < EPSILON) diff = 1.0;
      
      for(i = 0; i < lig; i++) OutVal[i] = (Data[i][NbIn + NSort] - vmin) / diff;
    }
  delete [] tmp;
  for(i = 0; i < NbEx; i ++) delete [] Data[i];
  delete [] Data;
} // End of InitSystem(dataf, config)

void FISHFP::InitSystem(char * dataf, int ns, char * ocfg)
  //******************************************************
{
  int i, j, col, lig;
  double ** Data;
  double min, max; 
  
  Data = NULL;
  Init();     
  
  Data = ReadSampleFile(dataf, col, lig);
  
  Name = new char[strlen(dataf)+10];

  fData = dataf;

  In = new INHFP * [col - ns];
  if(ns) Out = new FISOUT * [ns];

  // Compute range limits
  for(j = 0; j < col; j++)
    {
      max = min = Data[0][j];
      for(i = 1; i < lig; i ++)
	{
	  if(Data[i][j] < min) min = Data[i][j];
	  else if (Data[i][j] > max) max = Data[i][j];
	}
      if(j < (col-ns)) In[j] = new INHFP(j, min, max);
      else 
	{
	  Out[col-ns-j] = new OUTHFP(col-ns-j, min, max);
	  if(NbVarG >= (col-ns)  && VarNameG != NULL && VarNameG[(col-ns)] != NULL)
	    Out[col-ns-j]->SetName(VarNameG[(col-ns)]);
	}
    }

  NbIn = col-ns;
  NbOut = ns;
  NbEx = lig;

  DefaultValues();

  for(i = 0; i < NbEx; i ++) delete [] Data[i];
  delete [] Data;
}

void FISHFP::ReadHdrHfp(ifstream &f, int bufsize)
  //*********************************************
{
  char *tmp, *buf;

  tmp = buf = NULL;

  try
    {
      tmp = new char[bufsize];
      buf = new char[bufsize];

      do{                           // Skip empty lines
	f.getline(buf, bufsize);
      }while((strlen(buf) == 0) || (buf[0]==0x0D));  // ignore Dos file ^M 

      sprintf( tmp, "[Interface]" );
      if( !strncmp(tmp, buf, strlen(tmp)) ) //Obsolete config file
	{
	  f.getline(buf, bufsize);
	  
	  do{                           // skip empty lines
	    f.getline(buf, bufsize);
	  }while((strlen(buf) == 0) || (buf[0]==0x0D));  // ignores the ^M of MS-DOS files read in Linux-Unix
	}
      
       sprintf( tmp, "[System]" );
      if( strncmp(tmp, buf, strlen(tmp)) )
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
    
      f.getline(buf, bufsize);
      sprintf( tmp, "Name=" );
      if( strncmp(tmp, buf, strlen(tmp)) )
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
      if(SearchStr(buf, Name))
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~~StringSeparatorNotFoundInString~: %.50s~", buf );
	  throw std::runtime_error( ErrorMsg );
	}
    
      f.getline(buf, bufsize);
      sprintf( tmp, "Ninputs=" );
      if( strncmp(tmp, buf, strlen(tmp)) )
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
      NbIn = atoi(buf + strlen(tmp));
      if(NbIn < 0)
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~~InvalidNumberOfInputs~: %-3d~", NbIn );
	  NbIn = 0;
	  throw std::runtime_error( ErrorMsg );
	}

      f.getline(buf, bufsize);
      sprintf( tmp, "Noutputs=" );
      if( strncmp(tmp, buf, strlen(tmp)) )
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
      NbOut = atoi(buf + strlen(tmp));
      if(NbOut < 0)
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~~InvalidNumberOfOutputs~: %-3d~", NbOut );
	  NbOut = 0;
	  throw std::runtime_error( ErrorMsg );
	}

      f.getline(buf, bufsize);
      sprintf( tmp, "Conjunction=" );
      if( strncmp(tmp, buf, strlen(tmp)) )
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
      if(SearchStr(buf, tmp))
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~~StringSeparatorNotFoundInString~: %.50s~", buf );
	  throw std::runtime_error( ErrorMsg );
	}
      SetConjunction(tmp);

      f.getline(buf, bufsize);
      sprintf(tmp,"Hierarchy=");
      if( strncmp(tmp, buf, strlen(tmp)) )
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
      if(SearchStr(buf, strHierarchy))
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~~StringSeparatorNotFoundInString~: %.50s~", buf );
	  throw std::runtime_error( ErrorMsg );
	}

      f.getline(buf, bufsize);              
      sprintf(tmp,"ToleranceThresh=");
      if( strncmp(tmp, buf, strlen(tmp)) )
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
      Tol = strtod(buf + strlen(tmp), NULL);
			
      f.getline(buf, bufsize);
      sprintf(tmp,"DistanceType=");
      if( strncmp(tmp, buf, strlen(tmp)) )
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
      if(SearchStr(buf, strDistanceType))
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~~StringSeparatorNotFoundInString~: %.50s~", buf );
	  throw std::runtime_error( ErrorMsg );
	}
      if(!strcmp(strDistanceType, FISHFP::DistanceSymbNum())) Symbolic = -1;
      else if(!strcmp(strDistanceType, FISHFP::DistanceNumerical())) Symbolic = 0;
      else if(!strcmp(strDistanceType, FISHFP::DistanceSymbolic())) Symbolic = 1;
      else  
	{  
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~~UnknownParameterType~: %.50s~", strDistanceType );
	  throw std::runtime_error( ErrorMsg );
	}
    }

  catch( std::exception &e )
    {
      delete [] tmp;
      delete [] buf;
      throw;
    }

  delete [] tmp;
  delete [] buf;
} // End of ReadHdrHfp()


void FISHFP::ReadInHfp(ifstream & f, int bufsize, int num)  
  //******************************************************
{
  char *tmp=NULL, *buf=NULL;

  try
    {
      tmp = new char[bufsize+20];   
      buf = new char[bufsize];

      do{                           
	f.getline(buf, bufsize);
      }while((strlen(buf) == 0) || (buf[0]==0x0D));  

      sprintf(tmp,"[Input%d]", num + 1);    
      if( strncmp(tmp, buf, strlen(tmp)) )
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
  
      In[num] = new INHFP(f, bufsize, num + 1, strHierarchy, Symbolic, NbGroup);     
      // i + 1 for error messages
    }
  catch( std::exception &e )
    {
      delete [] tmp;
      delete [] buf;
      throw;
    }
  delete [] tmp;
  delete [] buf;
}


void FISHFP::PrintCfgHfp(char * dataf, char * of)
  //*********************************************
{
  int i; 
  FILE *f;
  
  char * Chaine=NULL, * ChaineHfp=NULL;
  Chaine = ChaineHfp = NULL;
  i = FileNameIndex(dataf);
  Chaine = new char [strlen(dataf + i) + 1];
  ChaineHfp = new char [strlen(dataf + i) + 5];
  sprintf(Chaine, "%s", dataf + i);
  sprintf(ChaineHfp, "%s.hfp", dataf + i);

  if(of == NULL)  f = fopen(ChaineHfp, "wt");
  else f = fopen(of, "wt");

  if(!f)
    {
      if(of) sprintf( ErrorMsg, "~CannotOpenFISFile~: %.100s~", of);
      else sprintf( ErrorMsg, "~CannotOpenFISFile~: %.100s.hfp~", dataf);
      throw std::runtime_error( ErrorMsg );
    }

  fprintf(f, "[System]\n");
  fprintf(f, "Name=%c%s%c\n", STRING_SEP, Chaine, STRING_SEP);
  fprintf(f, "Ninputs=%d\n", NbIn);  
  // default:  the last variable is the output
  fprintf(f, "Noutputs=%d\n", NbOut);   
  fprintf(f, "Conjunction=%c%s%c\n", STRING_SEP, cConjunction, STRING_SEP); 
  fprintf(f, "Hierarchy=%c%s%c\n", STRING_SEP, strHierarchy, STRING_SEP);
  fprintf(f, "ToleranceThresh=%9.6f\n", Tol);
  fprintf(f, "DistanceType=%c%s%c\n", STRING_SEP, strDistanceType, STRING_SEP);

  for(i = 0; i < NbIn; i++)  	In[i]->PrintCfgHfp(i + 1, f);
  for(i = 0; i < NbOut; i++)  Out[i]->PrintCfg(i + 1, f);
  
  fclose(f);

  delete [] Chaine;
  delete [] ChaineHfp;
} 


void FISHFP::Hfp(void)
  //******************
{
  int i;

  for(i = 0; i < NbIn; i++)
    if(In[i]->IsActive()) In[i]->InMerge();
}

void FISHFP::HRegular(void)
  //***********************
{
  int i;

  for(i = 0; i < NbIn; i++)
    if(In[i]->IsActive()) In[i]->HRegular();
}

void FISHFP::HKmeans(void)
  //**********************
{
  int i;
  
  for(i = 0; i < NbIn; i++)
    if(In[i]->IsActive()) In[i]->HKmeans();
}

void FISHFP::Hierarchy(char * fi)
  //*****************************
{
  int i;
  FILE * fres;

  if(!strcmp(strHierarchy, FISHFP::HierarchyHFP())) Hfp();
  else if(!strcmp(strHierarchy, FISHFP::HierarchyRegular()))  HRegular();
  else if(!strcmp(strHierarchy, FISHFP::HierarchyKMeans())) HKmeans();
      
  else
    { 
      sprintf( ErrorMsg, "~UnknownHierarchyType~: %.100s~", strHierarchy);
      throw std::runtime_error( ErrorMsg );
    }

  if(fi == NULL)
    {
      char ch[25];
      
      sprintf(ch, "vertices.%s", strHierarchy);
      fres = fopen(ch, "wt");
    }
  else fres = fopen(fi, "wt");
  
  for(i = 0; i < NbIn; i++)  In[i]->StoreVertices(fres);
  
  fclose(fres);
}


void FISHFP::GenereCfgFis(int Strap)
  //********************************
{
  int *config, i;

  config = NULL;
  config = new int [NbIn];

  ReadVertices(fVertex);

  for(i = 0; i < NbIn; i++)     config[i] = In[i]->SetNmf();

  FisBase(config, fFisCfg, Strap); 
  delete [] config;
  
  if(! NbOut || NSort < 0 || NSort > NbOut)
    {
      sprintf( ErrorMsg, "~UnknownOutput~: NbOut=%d  NumOut%d~", NbOut, NSort);
      throw std::runtime_error( ErrorMsg );
    }
  RuleInduction();
}

void FISHFP::ReadVertices(char * fic)
  //*********************************
{
  ifstream * f;
  int i, bsize;
  char * chaine;

  bsize = 30; 
  if((fic != NULL) && ((unsigned int) bsize < strlen(fic)) ) 
    bsize = strlen(fic) + 1; 
  chaine = new char [bsize];
  if(fic == NULL) sprintf(chaine, "vertices.%s", strHierarchy);
  else strcpy(chaine, fic);

  f = new ifstream(chaine);

  bsize = MaxLineSize(*f);

  if(f->fail() || bsize == 0) 
    {
      sprintf( ErrorMsg, "~CannotOpenFISFile~: %.100s~", chaine);
      delete [] chaine;
      throw std::runtime_error( ErrorMsg );
    }
   
  for(i = 0; i < NbIn; i++)  In[i]->ReadVertices(f, bsize);
  f->close();
  delete [] chaine;
  delete f;
}

void FISHFP::FisBase(int * n, char *fic, int Strap)
  //***********************************************
{
  FILE * f;
  int i, *act;

  act = NULL;
  act = new int [NbIn];

  if( (f = fopen(fic, "wt")) == NULL)
    {
      sprintf( ErrorMsg, "~CannotOpenFISFile~: %.100s~", fic);
      throw std::runtime_error( ErrorMsg );
    }     
     
  for(i = 0; i < NbIn; i++)  
    {
      In[i]->MfInstall(n[i], Strap); // From the vertex array
      act[i] = In[i]->IsActive();
      if(act[i] && n[i] < 2) In[i]->Deactivate();
    }

  PrintCfgFis(f);

  for(i = 0; i < NbIn; i++)  
      if(act[i] && n[i] < 2) In[i]->Activate();

  fclose(f);
  delete [] act;
}    

void FISHFP::PrintCfgFis(FILE * f)
  //******************************
{
  int i;

  fprintf(f, "[System]\n");
  fprintf(f, "Name=%c%s%c\n", STRING_SEP, Name, STRING_SEP);
  fprintf(f, "Ninputs=%d\n", NbIn);
  fprintf(f, "Noutputs=%d\n", NbOut);
  fprintf(f, "Nrules=0\n");
  fprintf(f, "Nexceptions=0\n");
  fprintf(f, "Conjunction=%c%s%c\n", STRING_SEP, cConjunction, STRING_SEP);
  fprintf(f, "MissingValues=%c%s%c\n", STRING_SEP, "random", STRING_SEP);


  for(i = 0; i < NbIn; i++)   In[i]->PrintCfg(i + 1, f);

  for(i = 0; i < NbOut; i++)  Out[i]->PrintCfg(i + 1, f);

  fprintf(f, "\n[Rules]\n");
 
  fprintf(f, "\n");

  fprintf(f, "\n[Exceptions]\n");

} 

void FISHFP::GenereRules(void) 
  //**************************
{   
  GENFIS * Gen; 
  FILE * f;

  Gen = new GENFIS(fFisCfg, fData, true, CumWeThres);
  f = fopen(fFisCfg,"wt");
  if(f == NULL)
    {
      sprintf( ErrorMsg, "~CannotOpenFISFile~: %.100s~", fFisCfg);
      throw std::runtime_error( ErrorMsg );
    }

  Gen->PrintCfg(f);
  fclose(f);
  
  delete Gen;
}

void FISHFP::FpaThis(void)
  //**********************
{
  FISFPA * Opt;   
  FILE *f;

  Opt = new FISFPA(fFisCfg, fData, CardStrategy, MinCard, MuMin);
  if((f = fopen(fFisCfg,"wt")) == NULL)
    {
      sprintf( ErrorMsg, "~CannotOpenFISFile~: %.100s~", fFisCfg);
      throw std::runtime_error( ErrorMsg );
    }
  Opt->FpaRules(NSort);
  Opt->PrintCfg(f);
  fclose(f);
  delete Opt;
}  

void FISHFP::WmThis(void)
  //*********************
{
  FISWM * Opt;   

  Opt = new FISWM(fFisCfg, fData, fFisCfg);
  Opt->SetNoLimit(true);
  if(Opt->wm())
    {
      sprintf(ErrorMsg, "\n~NoOutputDefinedInFis~:  %s\n", fFisCfg); 
      throw std::runtime_error( ErrorMsg );
    }
  delete Opt;
}  

void FISHFP::RuleInduction(void)
  //****************************
{
  if(!strcmp(cRuleInductionMethod, RuleInductionFpa()))
    {
      GenereRules();
      FpaThis();
    }
  else if(!strcmp(cRuleInductionMethod, RuleInductionWm()))
    WmThis();
  else
    {
      sprintf(ErrorMsg, "~UnknownRuleInductionMethod~in~FISHFP~: %.50s\n", cRuleInductionMethod);
      throw std::runtime_error( ErrorMsg );
    }
}

double FISHFP::EvalThis(const char * fArchive, int append)
  //******************************************************
{
  int i, j, classif;
  double valperf, maxError;
  char *config;
  char * ficvalid;
  double ** Data;
  FIS * Test;

  Test = NULL;
  config = NULL;

  valperf = 0;

  Test = new FIS(fFisCfg);
  
  if(! NbOut || NSort < 0 || NSort > NbOut)
    {
      sprintf( ErrorMsg, "~UnknownOutput~: NbOut=%d  NumOut%d~", NbOut, NSort);
      throw std::runtime_error( ErrorMsg );
    }

  Data = NULL;

  if(fTest) ficvalid = fTest;
  else ficvalid = fData;

  
  Data = ReadSampleFile(ficvalid, i, j);
  if(i < NbIn+NSort+1)
    {
      sprintf( ErrorMsg, "~NoObservedDataForOutput~: %d\n", NSort);
      throw std::runtime_error( ErrorMsg );
    }

  NbEx = j;

  classif = Test->Out[NSort]->Classification();

  if(classif) 
    {
    Test->ClassifCheck(Data, NbEx, NSort);
    Test->ResClassifAlloc(MisClass, Labels, NSort);
    }
  valperf = Test->Perf(NSort, Data, NbEx, Coverage, maxError, MuMin, MisClass, Labels, 1);

  config = new char [NbIn * 4];
  config[0] = 0;

  for(i = 0; i < NbIn; i ++)
    if(In[i]->IsActive()) 
	sprintf(config + strlen(config), "%d ", In[i]->GetNbMf());
    else sprintf(config + strlen(config), "0 ");
 
  FILE * fres;
  char mode[3];

  if(append) sprintf(mode, "at");
  else sprintf(mode, "wt");

  if(!fArchive)     
    fres = fopen("result", mode);
  else 
    fres = fopen(fArchive, mode);

  if(!append)
    {
      Test->WriteHeaderPerfRB(NSort, fres);
      fprintf(fres, "\n");
    }
  Test->PerfRB(valperf, Coverage, maxError, NSort, fres);
  fprintf(fres, "\n");
  fclose(fres);	  

  if (config != NULL) delete [] config;
  if (Data != NULL) 
    {
      for (i = 0; i < NbEx; i++) delete [] Data[i];
      delete [] Data;
    }

  if (Test != NULL) delete Test;

  return valperf;
}

void FISHFP::StoreRes(char * sc, double p, double me, int cl, InfoRB & Irb, int Ap, char * fic)
  //*******************************************************************************************
{ 
  FILE * fres;
  char mode[3];
  int m;

  if(Ap) sprintf(mode, "at");
  else sprintf(mode, "wt");

  if(!fic)     
    fres = fopen("result", mode);
  else 
    fres = fopen(fic, mode);

  if(! Ap)
    {
      fprintf(fres, "config & NbEx & Max Error & Coverage & MuMin & Perf & ");
      if(cl) 
	for(m = 0; m < NbClasses; m ++)
	  fprintf(fres, " MisCl %d & ", m+1);
      Irb.WriteHeader(fres);
    } 

  fprintf(fres, "%s & %d & %11.5f &  %4.2f & (%3.2f) & ", sc, NbEx, me, Coverage, MuMin);

  if(cl) 
    {
      fprintf(fres, "%d & ", (int) p);         // number of misclassified
      for(m = 0; m < NbClasses; m ++)
	fprintf(fres, "%d & ", MisClass[m]);
    }

  else fprintf(fres, "%11.5f &", p);  // root mean square of errors

  Irb.Print(fres);
  fprintf(fres, "\n");
  fclose(fres);	  
}

void FISHFP::SelectFis(bool wl)
  //***************************
{
  int *config, i, k, arret, i_min, first_res, first_resmin;
  double perf, valmin;

  if(! NbOut || NSort < 0 || NSort > NbOut)
    {
      sprintf( ErrorMsg, "~UnknownOutput~: NbOut=%d  NumOut%d~", NbOut, NSort);
      throw std::runtime_error( ErrorMsg );
    }

  config = NULL;
  config = new int [NbIn];

  ReadVertices(fVertex);

  for(i = 0; i < NbIn; i++)     
    {
      In[i]->InitNmfNvertices();
      config[i] = NmfInit;
    }

  k = 0;
  arret = false;

  first_res = first_resmin = 1;

  do  
    {
      valmin = INFINI_HFP;
      i_min=-1;
      
      for(i = 0; i < NbIn; i++)
	{
	  if(! In[i]->IsActive()) continue;
	  if(config[i] == In[i]->GetNvertices()) continue; 
	  config[i]++;
	  FisBase(config, fFisCfg, true); // 
	  RuleInduction();
	  if(first_res)
	    {
	      perf = EvalThis("result", 0);
	      first_res = 0;
	    }
	  else  perf = EvalThis("result", 1);
	  if((perf < valmin) && (Coverage >= BlankThres)) 
	    { valmin = perf; i_min = i; }

	  config[i]--; 
	}

      if(i_min == -1)
	{
	  for(i = 0; i < NbIn; i++) 
	    if(config[i] < In[i]->GetNvertices()) break; 
	  if (! wl)
	    {
	      if(i == NbIn) 
		printf("\nMaximum number of MF reached on each input\n");
	      else
		printf("\nToo much blank examples\n");
	    }
	  arret = true; 
	}
      else
	{
	  config[i_min]++;
	  FisBase(config, fFisCfg, true);  
	  RuleInduction();
	  if(first_resmin)
	    {
	      perf = EvalThis("result.min", 0);
	      first_resmin = 0;
	    }
	  else EvalThis("result.min", 1);
	}
	  
      if(k++ == IterMax) arret = true;
         
    } while(arret == false);

  delete [] config;

}
//**************************************************************************************
//
//                              FISHFP::ReadOut
//
//          Identical to FIS::ReadOut but FISHFP does not inherit from FIS
//
//**************************************************************************************

void FISHFP::ReadOut(ifstream &f, int bsize, int num) 
  //*************************************************
{
  double vDefault;
  char *tmp=NULL, *buf=NULL, *nature=NULL, *cDefuz=NULL, *cDisj=NULL;
  int cclas;

  try
    {
      tmp = new char[bsize];
      buf = new char[bsize];
      nature = new char[bsize];
      cDefuz = new char[bsize];
      cDisj = new char[bsize];

      do{                               
	f.getline(buf, bsize);
      }while((strlen(buf) == 0) || (buf[0]==0x0D));  
  

      sprintf(tmp,"[Output%d]", num + 1);    // Tag
      if( strncmp(tmp, buf, strlen(tmp)) )
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}

      f.getline(buf, bsize);
      sprintf(tmp,"Nature=");    // Tag
      if( strncmp(tmp, buf, strlen(tmp)) ) 
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~Output~: %-3d\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", num+1, tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
      if(SearchStr(buf, nature))
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~Output~: %-3d\n~StringSeparatorNotFoundInString~: %.50s~", num+1, buf );
	  throw std::runtime_error( ErrorMsg );
	}

      // read the defuzzification type
      f.getline(buf, bsize);
      sprintf(tmp,"Defuzzification=");
      if( strncmp(tmp, buf, strlen(tmp)) ) 
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~Output~: %-3d\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", num+1, tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
      if(SearchStr(buf, cDefuz))
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~Output~: %-3d\n~StringSeparatorNotFoundInString~: %.50s~", num+1, buf );
	  throw std::runtime_error( ErrorMsg );
	}

      // read the disjunction type
      f.getline(buf, bsize);
      sprintf(tmp,"Disjunction=");
      if( strncmp(tmp, buf, strlen(tmp)) ) 
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~Output~: %-3d\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", num+1, tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
      if(SearchStr(buf, cDisj))
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~Output~: %-3d\n~StringSeparatorNotFoundInString~: %.50s~", num+1, buf );
	  throw std::runtime_error( ErrorMsg );
	}


      // read the default value
      f.getline(buf, bsize);
      sprintf(tmp,"DefaultValue=");
      if( strncmp(tmp, buf, strlen(tmp)) ) 
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~Output~: %-3d\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", num+1, tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
      vDefault = strtod(buf + strlen(tmp), NULL);
  
      f.getline(buf, bsize);
      sprintf(tmp,"Classif=");
      if( strncmp(tmp, buf, strlen(tmp)) ) 
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~Output~: %-3d\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", num+1, tmp, buf );
	  throw std::runtime_error( ErrorMsg );
	}
      if(SearchStr(buf, tmp))
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~Output~: %-3d\n~StringSeparatorNotFoundInString~: %.50s~", num+1, buf );
	  throw std::runtime_error( ErrorMsg );
	}
      if( ! strncmp(tmp, "no", 4) )  cclas = false;
      else  if( ! strncmp(tmp, "yes", 4) ) cclas = true;
      else
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~Output~: %-3d\n~ExpectedString~: Classif=yes or no\n~ReadString~: %.50s~", num+1, tmp );
	  throw std::runtime_error( ErrorMsg );
	}

      if(!strcmp(nature, OUT_CRISP::OutputType()))
	Out[num] = new OUT_CRISP(f, bsize, num + 1, cDefuz, cDisj, cclas, vDefault);
      // i+1 for interface
      else if(!strcmp(nature, OUT_FUZZY::OutputType()))
	Out[num] = new OUT_FUZZY(f, bsize, num + 1, cDefuz, cDisj, cclas, vDefault);
      else
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~Output~: %-3d\n~UnknownNature~:~%.50s~", num+1, nature );
	  throw std::runtime_error( ErrorMsg );
	}
      if(NbVarG >= num+NbIn  && VarNameG != NULL && VarNameG[num+NbIn] != NULL)
	Out[num]->SetName(VarNameG[num+NbIn]);
    }
  catch( std::exception &e )
    {
      delete [] tmp;
      delete [] buf;
      delete [] nature;
      delete [] cDefuz;
      delete [] cDisj;
      throw;
    }
  delete [] tmp;
  delete [] buf;
  delete [] nature;
  delete [] cDefuz;
  delete [] cDisj;
}

void FISHFP::GenereCfgFisSr(int nmfout, int hout, char *def, char *dij, double tol, int clas)
  //*****************************************************************************************
{
  int i, activestate = 0;
  FISOUT * O;
  FILE *f;

  ReadVertices(fVertex);

  for(i = 0; i < NbIn; i++)    
    {
      In[i]->MfInstall(In[i]->SetNmf(), true);
      //! Avoid the Generation of a universal MF (useless)
      activestate=In[i]->IsActive();
      if(In[i]->GetNbMf() < 2 && activestate) In[i]->Deactivate();   
    }

  if(nmfout) O = InitOutHfp(hout, nmfout, -1, tol, def, dij);
  else  O = new OUT_CRISP(Out[0]->min(), Out[0]->max(), def, dij, clas);
  if(NbVarG >= NbIn  && VarNameG != NULL && VarNameG[NbIn] != NULL)
    O->SetName(VarNameG[NbIn]);
  
  f = fopen(fFisCfg, "wt");
 
  fprintf(f, "[System]\n");
  fprintf(f, "Name=%c%s%c\n", STRING_SEP, Name, STRING_SEP);
  fprintf(f, "Ninputs=%d\n", NbIn);
  fprintf(f, "Noutputs=%d\n", NbOut);
  fprintf(f, "Nrules=0\n");
  fprintf(f, "Nexceptions=0\n");
  fprintf(f, "Conjunction=%c%s%c\n", STRING_SEP, cConjunction, STRING_SEP);
  fprintf(f, "MissingValues=%c%s%c\n", STRING_SEP, "random", STRING_SEP);


  for(i = 0; i < NbIn; i++)   
    {
      In[i]->SetStdMfNames();
      In[i]->PrintCfg(i + 1, f);
      //! REACTIVATE the universal MF
      if(In[i]->GetNbMf() < 2 && activestate) In[i]->Activate();   
    }

  O->PrintCfg(1, f);

  fprintf(f, "\n[Rules]\n");
  fprintf(f, "\n");
  fprintf(f, "\n[Exceptions]\n");

  fclose(f);
  
  delete O;
}


FISOUT * FISHFP::InitOutHfp(int h, int n, int o, double tol, char *def, char *dij)
  //******************************************************************************
{
  double ** Data;
  double * T, * centres;
  int i, j, col, lig;
  INHFP * I;
  FISOUT * O;

  Data = NULL;
  Data = ReadSampleFile(fData, col, lig);

  T = new double [lig];
  
  j = col - 1;  
  if( o > -1 && o < col) j = o;

  for(i = 0; i < lig; i ++)   T[i] = Data[i][j];

  for(i = 0; i < lig; i ++)  delete [] Data[i];   
  delete [] Data;

  I = new INHFP(T, lig, h, n, tol);

  delete [] T;

  col = I->GetNbMf();
  centres = new double [col];
  I->GetMfCenters(centres);

  O = new OUT_FUZZY(centres, col,  I->min(), I->max(), 0, def, dij, 0); 
  O->SetName("OutHfp");
  O->SetStdMfNames();

  delete I;
  delete [] centres;
  return O; 
}

void FISHFP::SetConjunction(const char *conjunction)
  //************************************************
{
  delete [] cConjunction;
  cConjunction = new char[strlen(conjunction)+1];
  sprintf( cConjunction, "%s", conjunction );
}

void FISHFP::SetHierarchy(const char *hierarchy)
  //********************************************
{
  delete [] strHierarchy;
  strHierarchy = new char[strlen(hierarchy)+1];
  sprintf( strHierarchy, "%s", hierarchy );
}

void FISHFP::SetDistanceType(const char *distance)
  //**********************************************
{
  delete [] strDistanceType;
  strDistanceType = new char[strlen(distance)+1];
  sprintf( strDistanceType, "%s", distance );
}

void FISHFP::DefaultValues(void)
  //****************************
{
  SetConjunction("min");
  SetHierarchy(FISHFP::HierarchyKMeans());
  SetToleranceThres(TOLERANCE);
  // SetDistanceType(FISHFP::DistanceNumerical());
  SetDistanceType(FISHFP::DistanceSymbNum());
}

void FISHFP::ReplaceOutput(int output_number, FISOUT *output)
  //*********************************************************
{
  delete Out[output_number];
  Out[output_number] = output;
}


//************************    End of HFPFIS.CPP    ************************
