//***********************************************************************
//
//
//                              HFPIN.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : INHFP class functions used by FISPRO, 
//        part of package hfp (Hierarchical fuzzy partitioning)
//        requires library fisprob
//
//**********************************************************************

#include "hfpin.h"

#define NOT_INIT 147.852369

INHFP::INHFP(double *t, int n, int hier, int nmf, double tol)
  //********************************************************* 
  : FISIN(0)
{
  int nb;
 
  Init();

  HierType = hier;

  ValInf = NOT_INIT;
  ValSup = NOT_INIT;

  NmfMax = ((nmf < MAX_MF) ? nmf : MAX_MF);

  InitCont(t, n, tol);

  switch(HierType)
    {
    case T_HFP    : InMerge(); break;
    case T_KMEANS : HKmeans();  break;
    case T_REG    : HRegular(); break;
    default       : HRegular(); break;
    }

  nb = ((Nmf < Nvertices) ? Nmf : Nvertices);
  MfInstall(nb, true); // Vertices are used as MF centres

}

INHFP::INHFP(ifstream & f, int bufsize, int num, char * s_hier, int DistSymb, int ng) 
  //**********************************************************************
  : FISIN(0)
{
  char * tmp=NULL,  * buf=NULL;
  double Tab[2];

  tmp = new char[bufsize];
  buf = new char[bufsize];

  Init();

  do{                           // Skip empty lines
    f.getline(buf, bufsize);
  }while((strlen(buf) == 0) || (buf[0]==0x0D));  // Ignore ^M (DOS file end of line)


  sprintf(tmp,"Active=");              
  if( strncmp(tmp, buf, strlen(tmp)) )
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
      throw std::runtime_error( ErrorMsg );
    }

  if(SearchStr(buf, tmp))
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~~StringSeparatorNotFoundInString~: %.50s~", buf );
      throw std::runtime_error( ErrorMsg );
    }
  if( ! strncmp(tmp, "no", 4) )  active = false;   
  else  if( ! strncmp(tmp, "yes", 4) ) active = true;
  else 
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: yes or no\n~ReadString~: %.50s~", tmp);
      throw std::runtime_error( ErrorMsg );
    } 

  f.getline(buf, bufsize);              // Name
  sprintf(tmp,"Name=");
  if( strncmp(tmp, buf, strlen(tmp)) )
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
      throw std::runtime_error( ErrorMsg );
    }
  if(SearchStr(buf, tmp)) 
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~~StringSeparatorNotFoundInString~: %.50s~", buf );
      throw std::runtime_error( ErrorMsg );
    }
  SetName(tmp);

  f.getline(buf, bufsize);             
  sprintf(tmp,"Range=");
  if( strncmp(tmp, buf, strlen(tmp)) )
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
      throw std::runtime_error( ErrorMsg );
    }

  if((strlen(buf + strlen(tmp)) == 0) || (* (buf + strlen(tmp))==0x0D))
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~~EmptyString~: %.50s~", tmp);
      throw std::runtime_error( ErrorMsg );
    }

  if((SearchNb(buf, Tab, 2, SEPARE, START_NB, END_NB)) != 2) 
    {
      ValInf = FisMknan();
      ValSup = FisMknan(); 
    }
  else 
    {
      ValInf = Tab[0];
      ValSup = Tab[1];
    }

  f.getline(buf, bufsize);             
  sprintf(tmp,"NMFs=");
  if( strncmp(tmp, buf, strlen(tmp)) ) 
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
      throw std::runtime_error( ErrorMsg );
    }
  NmfMax = atoi(buf + strlen(tmp));
  if(NmfMax > MAX_MF) NmfMax = MAX_MF;
  Symbolic = DistSymb;
  SetHierType(s_hier);

  delete [] tmp;    
  delete [] buf;
} 

void  INHFP::InitCont(double * T, int nb, double thres)
  //***************************************************
{
  if(! active) return;

  int i, j, cut;
  double diff;
  double * reg;

  reg = NULL;
  cut = false;
  j = 0;
  
  MinObserve = MaxObserve = T[0]; 
  
  for(i = 0; i < nb; i++)
    {
      if(T[i]< MinObserve) MinObserve = T[i];
      else if(T[i] > MaxObserve) MaxObserve = T[i];
    }

  cut = false;
  if(ValInf == NOT_INIT) ValInf = MinObserve;
  else if(ValInf > MinObserve) 
    {
      MinObserve = ValInf;
      cut = true;
    }
  if(ValSup == NOT_INIT) ValSup = MaxObserve;
  else if(ValSup < MaxObserve) 
    {
      MaxObserve = ValSup;
      cut = true;
    }

  diff = ValSup - ValInf;
  if(diff < EPSILON) diff = 1.0;

  if(cut)
    for(i = 0; i < nb; i++)
      if((T[i] < ValInf) || (T[i] > ValSup)) j++;
  // j is the number of values to be cut

  Val = new double [nb - j];
  Size = nb - j;  
  j = 0;    

  for(i = 0; i < nb; i++)
    {
      if((T[i] < ValInf) || (T[i] > ValSup)) continue;
      Val[j++] = (T[i] - ValInf) / diff;
    }

 if(thres > 1.)  KmeansInitPart((int) floor(thres));
  else UniqInitPart(thres);
  
  // Generate missing  values 
  for(i = 0; i < Size; i++)
    if(FisIsnan(Val[i]))  Val[i] = FisRand();

  Dist = new double * [NbUnique];
  for(i = 0; i < NbUnique; i++) Dist[i] = new double [NbUnique];
  
  if(HierType == T_REG) Nmf = NmfMax; // No limitation
  else if (HierType == T_KMEANS) 
    Nmf = ((NbUnique < NmfMax) ? NbUnique : NmfMax);
  else Nmf = NbUnique;

  Nvertices = ((NmfMax < Nmf) ? NmfMax : Nmf);

  Vertices = new TVertex * [Nvertices];
  for(i = 0; i < Nvertices; i++) 
    Vertices[i] = new TVertex  [i + 2];

  if(HierType == T_REG) 
    {
      reg = new double [Nvertices];
      if(Nvertices > 1)
	for(i = 0; i < Nvertices; i++)
	  reg[i] = (double) i/(Nvertices-1);
      else reg[0] = ValInf;
    }

  switch(HierType)
    {
    case T_HFP    : InitMf(Unique); break;
    case T_KMEANS : InitMf(Unique); break;
    case T_REG    : InitMf(reg); break;
    default       : InitMf(reg); break;
    }

  if(HierType == T_REG) delete [] reg;
} 

void INHFP::UniqInitPart(double equal)
  //**********************************
{
  if(equal < EPSILON) equal = EPSILON;
 
  SortUniq(Val, Size, &Unique, &Occur, NbUnique, equal);

  // Remove Nan from unique
  if(FisIsnan(Unique[NbUnique - 1])) NbUnique --;

 }


void INHFP::KmeansInitPart(int n)
  //*****************************
{
  int i, empty, k;
  int * eff;
  double *Centres;
  double iter;

  if(n < 1) 
    {
      sprintf(ErrorMsg,  "KmeansInitPart-~InvalidNumberOfGroups~: %d~\n~", n);
      throw std::runtime_error( ErrorMsg );
    }

  Centres = new double [n];
  eff = new int [n];
 
  if(n == 1) 
    {
      Centres[0] = 0.;
      eff[0]=0;
    }
  else   
    for(i = 0; i < n ; i++) 
      {	
	eff[i] = 0;
	Centres[i] = (double) i/(n-1);
      }

  iter = Kmeans(Val, Size, Centres, n);
       	
  for(i = 0; i <Size;i++) eff[AssignClas(Val[i],Centres,n)]++;
      
  empty = 0;
  for(i = 0; i < n ; i++) 
    if (!eff[i]) empty++;
  NbUnique = n - empty;
  // printf("\n init using kmeans %d non empty groups\n", NbUnique);

  Unique = new double [NbUnique];
  Occur = new int [NbUnique];
  k=0;
  for(i = 0; i < n ; i++)
    if (eff[i])
      {
	Occur[k] = eff[i];
	Unique[k] = Centres[i];
	k++;
      }
  
  delete [] Centres;
  delete [] eff;
}


void INHFP::PrintCfgHfp(int num, FILE *f)
  //*************************************
{
  char Chaine[4];
  if(active) sprintf(Chaine, "yes");
  else sprintf(Chaine, "no");
      
  FISIN::PrintCfgTag(num, f);
  fprintf(f, "Active=%c%s%c\n", STRING_SEP, Chaine, STRING_SEP);
  fprintf(f, "Name=%c%s%c\n", STRING_SEP, Name, STRING_SEP);
  fprintf(f, "Range=%c%f%c%f%c\n", START_NB, ValInf, SEPARE, ValSup, END_NB);
  fprintf(f, "NMFs=%d\n", NmfMax);
}


void INHFP::PrintDist(int num) 
  //***************************
{
  int i, j;

  printf("Input %d : \n", num);
  for(i = 0; i < NbUnique; i++)
    {
      for(j = 0; j < NbUnique; j++)
	printf("%10.5f", Dist[i][j]);
      printf("\n");
    }
  printf("\n");
}


double INHFP::ExtDist(int a, int b)
  //*******************************
{
  if(b == a) return 0.0;
 
  CSeg an, bn;

  if(! Symbolic) // Distance type = numerical
    {
      Fp[a]->Kernel(an);
      Fp[b]->Kernel(bn);
    }

  if(b > a)   
    {
      if(Symbolic) return 0.5 + (double) (b - a) / (Nmf-1);
      else return bn.l - an.l + 0.5;   
    }
  
  if(b < a)   
    {
      if(Symbolic) return 0.5 + (double) (a - b) / (Nmf-1);
      else return an.l - bn.l + 0.5; 
    }

  return 0.0;
}


double INHFP::DistSum(int s, int before)
  //************************************
{
  int i, j;
  double somme = 0.0;
  int Tous = false;

  if(before) Nmf++;

  FISIN I(* this); // ensures the update of MF linearization

  if(Symbolic == -1) // distance type = symbnum
    {
      for(int i = 0; i < NbUnique; i++)
	{
	  Dist[i][i] = 0.0;
	  for(int j = i + 1; j < NbUnique; j++) 
	    Dist[j][i] = Dist[i][j] = I.Distance(Unique[i], Unique[j], true);
	}
   }
  else Distance(0, Nmf-1);

  if(before) Nmf--;

  for(i = 0; i < NbUnique; i++)
    for(j = i; j < NbUnique; j++)
      somme += Dist[i][j] * Occur[i] * Occur[j];
  if(Tous) somme /= Size * (Size - 1) * 2.;
  return somme;
}

void INHFP::StoreVertices(FILE * f)
  //*******************************
{
  int i, j;
  double diff;
  
  diff = ValSup - ValInf;
  if(diff < EPSILON) diff = 1.0;
  
  fprintf(f, "%d\n", Nvertices);
  for(i = 1; i < Nvertices; i++) 
    {
      fprintf(f, "%f%c%f", Vertices[i][0].l,  SEPARE, Vertices[i][0].r); 
      
      for(j = 1; j < i + 2; j ++)  
	fprintf(f, "%c%f", SEPARE, Vertices[i][j].l * diff + ValInf);
      fprintf(f, "\n");
    } 
}

void INHFP::ReadVertices(ifstream *f, int bufsize)
  //**********************************************
{
  int i, j;
  double diff;
  char *buf;
  double *Tab;

  Tab = new double [5];
  buf = new char[bufsize];
  
  diff = ValSup - ValInf;
  if(diff < EPSILON) diff = 1.0;

  f->getline(buf, bufsize);
  Nvertices = atoi(buf);
  if(Nvertices == 0) return;
  if(Nvertices < 0) 
    {
      sprintf(ErrorMsg, "~UnexpectedNumberOfValues~ in the vertices file : %d~", Nvertices);
      delete [] Tab; delete [] buf;
      throw std::runtime_error( ErrorMsg );
    }    
  
  delete [] Tab;
  Tab = new double [Nvertices + 2];

  Vertices = new TVertex * [Nvertices];
  for(i = 0; i < Nvertices; i++) 
    Vertices[i] = new TVertex  [i + 2];

  for(i = 1; i < Nvertices; i++) 
    {
      f->getline(buf, bufsize);
      if((strlen(buf) == 0) || (buf[0] == 0x00D) || 
	 (SearchNb(buf, Tab, i+3, SEPARE) != i+3) )
	{
	  sprintf( ErrorMsg, "~UnexpectedNumberOfValues~ in the vertices file : %d values expected ", i+3);
	  delete [] Tab; delete [] buf;
	  throw std::runtime_error( ErrorMsg );
	}

      Vertices[i][0].l = Tab[0]; 
      Vertices[i][0].r = Tab[1]; 

      for(j = 1; j < i + 2; j ++)  
	Vertices[i][j].l = Vertices[i][j].r = (Tab[j+1] - ValInf) / diff ;
    } 
  
  delete [] buf;
  delete []Tab;
}

int INHFP::HRegular(void)
  //*********************
{
  int i, j;
  double *Centres, partc, parte;

  Centres = new double [Nvertices];
  partc = parte = 0.;
 
  for(j = 1; j < Nvertices; j++)
    {
      for(i = 1; i < (j+2); i++)
	Vertices[j][i].l = Vertices[j][i].r = Centres[i-1] = (double) (i-1)/j;

      ComputeIndices(Centres, j+1, partc, parte);
      Vertices[j][0].l = partc;
      Vertices[j][0].r = parte;
    }
  delete [] Centres;
  return 0;
}
  
int INHFP::HKmeans(void)
  //********************
{
  int i, j, iter;
  double *Centres, partc, parte;

  Centres = new double [Nvertices];
  partc = parte = 0.;

  for(j = 1; j < Nvertices; j++)
    {
       
      for(i = 0; i < j+1; i++) 	Centres[i] = (double) i/j;

      iter = Kmeans(Val, Size, Centres, j+1);
 	
      for(i = 1; i < (j+2); i++)
	Vertices[j][i].l = Vertices[j][i].r = Centres[i-1];

      ComputeIndices(Centres, j+1, partc, parte);
      Vertices[j][0].l = partc;
      Vertices[j][0].r = parte;
    }

  delete [] Centres;
  return 0;
}

void INHFP::ComputeIndices(double *cent, int n, double &pc, double &pe)
  //*******************************************************************
// Both FISIN and data are normalized
{
  double deg=0.;
  int i, j;
 
  FISIN I(cent, n, 0, 1); // build a FISIN with semitrapezoidal edges
  pc = pe = 0.;
  for(i = 0; i < Size; i++)
    {
      I.GetDegsV(Val[i]); 
      for(j = 0; j < n; j++)
	{
	  deg = I.Mfdeg()[j];
	  pc += deg * deg;
	  if(deg > EPSILON && deg < (1.0 - EPSILON)) 
	    pe += deg * log(deg);
	}
    }
  pc /= Size;
  pe /= Size;
  pe *= -1.;
}

//*************************************************************************
//
//
//                           PROTECTED FUNCTIONS
//
//
//*************************************************************************
int INHFP::InitNmfNvertices(void)
  //*****************************
{
  int j; 

  NmfMax = 0;
  if(Nvertices == 0) active = false;
  if(active == false) return 0;
   
  NmfMax = Nvertices;

  Fp = new MF * [NmfMax];
  for(j = 0; j < NmfMax; j++) Fp[j] = NULL;
  return NmfMax;
}

int INHFP::SetNmf(void)
  //*******************
{
  int j, nb; 

  if(active == false) return 0;
   
  if(NmfMax > Nvertices)
    {
      nb = Nvertices;
      printf("\nNumber of MF requested (%d) greater than number of vertices (%d) for input %s",  NmfMax, Nvertices, Name);
    }
  else  nb = NmfMax;

  Fp = new MF * [nb];
  for(j = 0; j < nb; j++) Fp[j] = NULL;
  return nb;
}


void INHFP::Distance(int smin, int smax)
  //************************************
{
  int i, j, k, ki, kj, OKi, OKj;
  double *VaI, SumI, SumJ, Div;

  FISIN I(* this); // ensures the update of MF linearization

  VaI = new double [Nmf];
 
  for(i = 0; i < NbUnique; i++)
    {
      Dist[i][i] = 0.0;
      I.GetDegsV(Unique[i]); 
      OKi = false;

      for(k = 0; k < Nmf ; k++) 
	VaI[k] = I.Mfdeg()[k];

      for(k = smin; k <= smax ; k++) 
	if(VaI[k] > EPSILON) OKi = true;
      // Point i has at least a non null degree to one of the MF of interest
    
      for(j = i + 1; j < NbUnique; j++) 
	{
	  Dist[i][j] = Dist[j][i] = 0.;
	  if(OKi == false) continue;

	  I.GetDegsV(Unique[j]); 
	  OKj = false;
      
	  for(k = smin; k <= smax ; k++) 
	    if(I.Mfdeg()[k] > EPSILON) OKj = true;

	  if(OKj == false) continue;

	  // All the degrees are taken into account
	  SumI = 0.; SumJ = 0.;
	  
	  for(ki = 0; ki < Nmf ; ki++)
	    {
	      if(VaI[ki] < EPSILON) continue;
	      SumI += VaI[ki];
	      SumJ = 0.;
	      for(kj = 0; kj < Nmf ; kj++)
		{
		  if(I.Mfdeg()[kj] < EPSILON) continue;
		  SumJ += I.Mfdeg()[kj];
		  
		  Dist[i][j] += VaI[ki] * I.Mfdeg()[kj] * IntDist(VaI[ki], I.Mfdeg()[kj]); 

		  if(ki != kj)
		    Dist[i][j] += VaI[ki] * I.Mfdeg()[kj] * ExtDist(ki, kj);
		}
	    }
	  Div = SumI * SumJ;
	  if(Div) Dist[i][j] /= Div;
	  Dist[j][i] = Dist[i][j];
	}
    }
  delete [] VaI;
} 


double INHFP::EvalMass(int f)
  //*************************
{
  int i;
  double masse =  0.0;

  FISIN I(* this);  
  for(i = 0; i < NbUnique; i++)
    {
      I.GetDegsV(Unique[i]); 
      masse += I.Mfdeg()[f]  * Occur[i];
    }
  // masse += Fp[f]->GetDeg(Unique[i]) * Occur[i];

  return masse;
}


void INHFP::InMerge(void)
  //*********************
  // Examine l'ensemble des possibilites de fusion et memorise la meilleure
  // combinaison a chaque etape
{
  if(Nmf == Nvertices)
    {
      int i;
      double partc, parte;
      double *tab;
      tab = new double[Nmf];
      GetMfCenters(tab);
      partc = parte = 0.;
      ComputeIndices(tab, Nmf, partc, parte);
      delete [] tab;

      Vertices[Nmf-1][0].l = partc;
      Vertices[Nmf-1][0].r = parte;
      for(i = 0; i < Nmf; i++)
	Fp[i]->Kernel(Vertices[Nmf-1][i+1].l, Vertices[Nmf-1][i+1].r);
    }
  
  SelectMf();

}    
//*****************        End of HFPIN.CPP    *********************************

