//**********************************************************************
//
//
//                              MAINHFPFIS.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : Main file for generating a FIS configuration file for hfp data,
//        (configuration and vertex files) 
//        Fispro, part of package hfp (Hierarchical fuzzy partitioning)
//        requires library fispro
//
//**********************************************************************

#include "hfpfis.h"

int main(int argc, char ** argv)
//******************************
{
  if(argc < 3)
    {
      printf("\nThe program %s needs 2 argument(s) : ", argv[0]);
      printf("\n\t* the data file name;\n") ;
      printf("\n\t* the hfp config file name;") ;

      printf("\n\t\tOptions:");
      printf("\n\t\t  -r rule induction method set to wm (default:  fpa)" );
      printf("\n\t\t  -s minimum cumulated weight (default:  0.0)" );
      printf("\n\t\t  -t cardinality strategy:  0 for MIN or 1 for DEC (default:  1)" );
      printf("\n\t\t  -m minimum matching degree (default:  0.3)" );
      printf("\n\t\t  -e minimum cardinality (default:  3)" );
      printf("\n\t\t  -o the result output file name (default:  'Name'.fis)");
      printf("\n\t\t  -p the output number (default:  0)");
      printf("\n\t\t  -l vertices file name (default:  vertices.'Hierarchy, ex .hfp)");
      printf("\n");
      return 1;
    }

  FISHFP *S=NULL;
  char * Config = NULL;
  
  int i;

  try
    {
      for(i = 3; i < argc; i++)
	if(! strncmp("-o", argv[i], 2) )  
	  if(strlen(argv[i]) > 3) Config = argv[i] + 2;

      S = new FISHFP (argv[2], argv[1], Config);
      
      for(i = 3; i < argc; i++)
	{	
	  if     (! strncmp("-r", argv[i], 2) )  
	    S->SetRuleInductionMethodWm();
	   
	  else if(! strncmp("-s", argv[i], 2) )  S->SetCumWeThres(strtod(argv[i] + 2, NULL));
	  else if(! strncmp("-t", argv[i], 2) )  S->SetCardStrategy(atoi(argv[i] + 2));
	  else if(! strncmp("-m", argv[i], 2) )  S->SetMuMin(strtod(argv[i] + 2, NULL));
	  else if(! strncmp("-e", argv[i], 2) )  S->SetMinCard(atoi(argv[i] + 2));
	  else if(! strncmp("-p", argv[i], 2) )  S->SetNSort(atoi(argv[i] + 2));
	  else if(! strncmp("-l", argv[i], 2) )  S->SetVertexFile(argv[i] + 2);

	}
      
      S->GenereCfgFis(true);
      // true for semi trapezoidal ended partitions
      delete S;
      S=NULL;
      if(VarNameG != NULL) 
	{
	  for(int j = 0; j < NbVarG; j++)
	    if(VarNameG[j] != NULL)  delete [] VarNameG[j];
	  delete [] VarNameG;
	  VarNameG = NULL;
	}
    }      

  catch(std::exception &e)
    {
      printf("\n%s", e.what());
      delete S;
      S=NULL;
    }

  delete S;
  return 0;
}


