//**********************************************************************
//
//
//                              MAINHFPFISSR.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : Main file for generating a FIS configuration file for hfp data,
//        (configuration and vertex files) 
//        Fispro, part of package hfp (Hierarchical fuzzy partitioning)
//        requires library fispro
//
//**********************************************************************

#include "hfpfis.h"

int main(int argc, char ** argv)
//******************************
{
  if(argc < 8)
    {
      printf("\nThe program %s needs 7 argument(s) : ", argv[0]);
      printf("\n\t* the data file name;\n") ;
      printf("\n\t* the hfp config file name;") ;
      printf("\n\t* vertices file name;");
      printf("\n\t* the output number of mf (if nmf=0 output crisp);");
      printf("\n\t* the output hierarchy type : 1 hfp, 2 kmeans, 3 regular (default)");
      printf("\n\t* the defuzzification type (valid strings depend on the output type)"); 
      printf("\n\t* the disjunction type (valid strings : sum, max)");

      printf("\n\t\tOptions:");
      printf("\n\t\t  -o the result output file name (default:  'Name'.fis)");
      printf("\n\t\t  -c set classification flag for the output (default:  false)");
      printf("\n\t\t     (Valid only when output number of mf = 0)");
      printf("\n");
      return 1;
    }

  FISHFP *S=NULL;
  char * Config = NULL;
  int i, clas = false;

  try
    {
      for(i = 8; i < argc; i++)
	{
	  if(! strncmp("-o", argv[i], 2) && strlen(argv[i]) > 3)
	    Config = argv[i] + 2;
	  else if(! strncmp("-c", argv[i], 2)) clas = true;
	}

      S = new FISHFP (argv[2], argv[1], Config);

      S->SetVertexFile(argv[3]);
      S->GenereCfgFisSr(atoi(argv[4]), atoi(argv[5]), argv[6], argv[7], TOLERANCE, clas);
    }      

  catch(std::exception &e)
    {
      printf("\n%s", e.what());
    }

  delete S;
  if(VarNameG != NULL) 
    {
      for(int j = 0; j < NbVarG; j++)
	if(VarNameG[j] != NULL)  delete [] VarNameG[j];
      delete [] VarNameG;
      VarNameG = NULL;
    }
  return 0;
}
