//**********************************************************************
//
//
//                              MAINHFPSELECT.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : Main file for building and selecting the best FIS using families of 
//        monodimensional partitions 
//        Fispro, part of package hfp (Hierarchical fuzzy partitioning)
//        requires library fispro
//
//**********************************************************************

#include "hfpfis.h"

int main(int argc, char ** argv)
//******************************
{
  if(argc < 3)
    {
      printf("\nThe program %s needs 2 argument(s) : ", argv[0]);
      printf("\n\t* the data file name;\n") ;
      printf("\n\t* the hfp config file name;") ;

      printf("\n\t\tOptions:");
      printf("\n\t\t  -r rule induction method set to wm (default:  fpa)" );
      printf("\n\t\t  -s minimum cumulated weight (0.0 par defaut)" );
      printf("\n\t\t  -t cardinality strategy:  0 for MIN or 1 for DEC (default:  1)" );
      printf("\n\t\t  -m minimum matching degree (default:  0.3)" );
      printf("\n\t\t  -e minimum cardinality (default:  3)" );
      //      printf("\n\t\t  -o the result output file name (default:  'Name'.fis)");
      printf("\n\t\t  -b minimum coverage (default:  1.0, 100 %%)"); 
      printf("\n\t\t  -n initial number of mf for each input (default:  1)");
      printf("\n\t\t  -i maximum number of iterations (default:  10)");
      printf("\n\t\t  -l vertices file name (default:  vertices.'Hierarchy, ex .hfp)");
      printf("\n\t\t  -p the output number (default:  0)");
      printf("\n\t\t  -v validation file name (default: data file name)");
      printf("\n\t* -wl for wordless.\n");
      printf("\n");
      return 1;
    }


  FISHFP *S=NULL;
  int i;
  char * OutCfg = NULL;
  bool wordless=false;
  bool vf=false;
  
  
  OutCfg = new char [30];
  strcpy(OutCfg, "fisfile.tmp");

  try
    {
      
      S = new FISHFP (argv[2], argv[1], OutCfg);
      
      if(argc > 3)
	{
	  for(i = 3; i < argc; i++)
	    {	
	      if     (! strncmp("-r", argv[i], 2) )  
		S->SetRuleInductionMethodWm();
	      
	      else if(! strncmp("-s", argv[i], 2) )  S->SetCumWeThres(strtod(argv[i] + 2, NULL));
	      else if(! strncmp("-t", argv[i], 2) )  S->SetCardStrategy(atoi(argv[i] + 2));
	      else if(! strncmp("-m", argv[i], 2) )  S->SetMuMin(strtod(argv[i] + 2, NULL));
	      else if(! strncmp("-e", argv[i], 2) )  S->SetMinCard(atoi(argv[i] + 2));
	      else if(! strncmp("-o", argv[i], 2) )  S->SetConfigFile(argv[i] + 2);
	      else if(! strncmp("-b", argv[i], 2) )  S->SetBlankThres(strtod(argv[i] + 2, NULL));
	      else if(! strncmp("-n", argv[i], 2) )  S->SetNmfInit(atoi(argv[i] + 2));
	      else if(! strncmp("-i", argv[i], 2) )  S->SetIterMax(atoi(argv[i] + 2));
	      else if(! strncmp("-l", argv[i], 2) )  S->SetVertexFile(argv[i] + 2);
	      else if(! strncmp("-p", argv[i], 2) )  S->SetNSort(atoi(argv[i] + 2));
	      else if(! strncmp("-v", argv[i], 2) )  
		{
		  S->SetValidationFile(argv[i] + 2);
		  vf=true;
		}
	      else if(! strncmp("-wl", argv[i], 3) ) wordless = true;
	    }
	}
      if(vf & !wordless)
	 printf("\nValidation file : %s\n", argv[i] + 2);
      if (!wordless)
	{
	  printf("\nData file : %s", argv[1]);
	  printf("\nInput configuration file : %s", argv[2]);
	}
	  
      
      S->SelectFis(wordless);
      if (OutCfg!=NULL) 
	{
	  remove(OutCfg);
	  delete [] OutCfg;
	}
    }

  catch(std::exception &e)
    {
      printf("\n%s", e.what());
    }
  delete S;
  return 0;
}


