//****************************************************************
//
//
//                              SETHFPFIS.CPP
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : one function = modifies a HFP file (MF  per input0)
// hierarchical fuzzy partitioning
// Environnment : requires library fispro
//****************************************************************


#include "sethfpfis.h"

void SetHfpFis(char *In, char *Cfg, char *Out)
  //******************************************
{
  ifstream f(In);
  FILE * fo;
  int i, bufsize, nb, Nin;
  char * buf=NULL, *tmp=NULL;
  double * v;

  if ( f.fail() )
    {
      sprintf( ErrorMsg, "~CannotOpenFISFile~: %.100s~", In);
      throw std::runtime_error( ErrorMsg );
    }

  if((fo = fopen(Out, "wt")) == NULL)
    {
      sprintf( ErrorMsg, "~CannotOpenFile~: %.100s~", Out);
      throw std::runtime_error( ErrorMsg );
    }

  bufsize = MaxLineSize(f);
  buf = new char [bufsize];
  tmp = new char [20];
  
  // Read the number of inputs
  buf[0] = 0;
  sprintf(tmp, "Ninputs=");
  while(((strlen(buf) < strlen(tmp)) || strncmp(tmp, buf, strlen(tmp))) && ! f.eof() ) 
    {
      f.getline(buf, bufsize);
      fprintf(fo, "%s\n", buf);
    } ;
  
  if( f.eof() )
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedStringNotFound~: %.50s\n~", tmp);
      throw std::runtime_error( ErrorMsg );
    }
  Nin = atoi(buf + strlen(tmp));

  // Set v size and fill v array
  v = new double [Nin];
  // particular case of single input-add one space at the end of chain
  // correction bch July 20 2009 - protection Nin=0 or 1
  if (Nin<=0) 
    {
      sprintf( ErrorMsg, "~invalid #MFs");
      throw  std::runtime_error( ErrorMsg );
    }
  
  nb = SearchNb(Cfg, v, Nin, ' ');

  if(FisIsnan(v[nb-1])) nb--; // the string is terminated by a separator


  // Set the nb values in the hfp file
  sprintf(tmp, "NMFs=");

  for(i = 0; i < nb; i++)
    {
      f.getline(buf, bufsize);
      while(((strlen(buf) < strlen(tmp)) || strncmp(tmp, buf, strlen(tmp))) && ! f.eof() ) 
 	{
	  fprintf(fo, "%s\n", buf);
	  f.getline(buf, bufsize);
	} ;
  
      if( f.eof() )
	{
	  sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedStringNotFound~: %.50s\n~", tmp);
	  throw std::runtime_error( ErrorMsg );
	}
      sprintf(buf, "%s%d", tmp, (int) v[i]);
      fprintf(fo, "%s\n", buf);
    }

  // Achive file copy
  while (!f.eof())
    {
      f.getline(buf, bufsize);
      fprintf(fo, "%s\n", buf);
    }

  fclose(fo);

  delete [] buf;
  delete [] tmp;
  delete [] v;
}
