//****************************************************************
//
//
//                              FIS_JNIFIS.CPP
//
//*****************
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC, Jean-Luc Lablee
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//
// File : C++ Functions to be called by the java interface
//****************************************************************

#include "fis_jnifis.h"
#include "fis.h"
#include "fpa.h"
#include "genfis.h"
#include "fistree.h"
#include "linkfis.h"
#include "hfpfis.h"
#include "fisimple.h"
#include "fisols.h"
#include "sethfpfis.h"
#include "sifopt.h"
#include "fiswm.h"
#include "hfpfis.h"
#include "gensample.h"
#include <errno.h>
#include <stdexcept>
#include <locale.h>


#include <cstring> // memcpy function
#include <memory>

int ReadInputsOpt(const char * stri, int nbmax,double *numinputs,bool wordless);
FIS* loopoptimsample(FIS* &S,const char* fisPath,const char* datafile,const char * optimName, int nsample,int nsteps,bool optimIn,bool optimOut,bool optimRules, int nbinputs,double * numinputs, int nOut,long int seed,double muthresh,double covloss,parstruct paramP,bool intermfile,bool wordless,double nearC);
void computeWritePerf(FIS* M,const char* fisPath,const char * datafile,const char * perfile,const char * fisname,const char* optimName,  double muthresh, int nOut, int ns, bool wordless);

static jclass class_java_lang_string = NULL;

static void throw_by_name(JNIEnv *env, const char *name, const char *msg) {

	jclass cls = env->FindClass(name);
	if(!env->ExceptionCheck())
		env->ThrowNew(cls, msg);
	env->DeleteLocalRef(cls);
}

static void find_class_java_lang_string(JNIEnv* env) {

	if(!class_java_lang_string) {
		jclass local_class_java_lang_string = env->FindClass("java/lang/String");
		if(!env->ExceptionCheck()) {
			class_java_lang_string = (jclass)env->NewGlobalRef(local_class_java_lang_string);
			env->DeleteLocalRef(local_class_java_lang_string);
			if(!class_java_lang_string)
				throw_by_name(env, "java/lang/OutOfMemoryError", NULL);
		}
	}
}

char *get_native_string(JNIEnv* env, jstring jstr) {

	find_class_java_lang_string(env);

	static jmethodID mid_string_getBytes = NULL;
	if(!mid_string_getBytes) {
		mid_string_getBytes = env->GetMethodID(class_java_lang_string, "getBytes", "()[B");
		if(env->ExceptionCheck())
			return NULL;
	}

	if(env->EnsureLocalCapacity(2) < 0) { return NULL; } // out of memory error

	jbyteArray bytes = (jbyteArray) env->CallObjectMethod(jstr, mid_string_getBytes);
	if(!env->ExceptionCheck()) {
		jint len = env->GetArrayLength(bytes);
		char *result = new char[len + 1];
		if(result == 0) {
			throw_by_name(env, "java/lang/OutOfMemoryError", 0);
			env->DeleteLocalRef(bytes);
			return 0;
		}
		env->GetByteArrayRegion(bytes, 0, len, (jbyte *)result);
		env->DeleteLocalRef(bytes);
		result[len] = 0; // NULL-terminate
		return result;
	}
	return NULL;
}

void release_native_string(const char *native_string) { delete [] native_string; }

const char *UserHomeFisproPath = NULL; /**< Chemin d'installation de Fispro, renseigne par la fonction #Java_fis_jnifis_UserHomeFisproPath */

//---------------------------------------------------------------------------
// Sends an exception e to JAVA
void ExceptionToJava( JNIEnv *env, std::exception &e )
{
	jclass new_ex = env->FindClass( "fis/JNIException" );
	if( new_ex != 0 )
		env->ThrowNew( new_ex, e.what() );
	env->DeleteLocalRef( new_ex );
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_setLocale(JNIEnv *, jclass) {
	setlocale(LC_NUMERIC, "C");
}

//---------------------------------------------------------------------------
/**
 * Renseigne la variable #UserHomeFisproPath.
 */
JNIEXPORT void JNICALL Java_fis_jnifis_setUserHomeFisproPath(JNIEnv *env, jclass, jstring jpath)
{

	if(UserHomeFisproPath != NULL)
		release_native_string(UserHomeFisproPath);
	UserHomeFisproPath = get_native_string(env, jpath);

}


//---------------------------------------------------------------------------
/**
 * Cree un nom unique de fichier temporaire dans le repertoire #UserHomeFisproPath \n
 * si #UserHomeFisproPath est NULL, le fichier est cree selon les regles de la fonction tmpnam, donc dans P_tmpdir.
 * @return une chaine de caractere contenant le nom du fichier temporaire (il faut la detruire par un delete lorsqu'elle n'est plus utile).
 */
char *TempFileName()
{
	char *name = tmpnam(NULL);
	if( UserHomeFisproPath == NULL ) return name;
	char *tmp = new char[strlen(UserHomeFisproPath) + strlen(name) + 1];
	if( tmp != NULL )
	{
		strcpy(tmp, UserHomeFisproPath);
		strcat(tmp, &name[FileNameIndex(name)]);
	}
	return tmp;
}

// sauvegarde le fis dans un fichier temporaire
const char *TemporaryFisName(FIS *fis)
{
	const char *tmp = TempFileName();
	FILE *f = fopen(tmp, "wt");
	if(f == NULL)
		throw std::runtime_error("cannot open temporary file");
	fis->PrintCfg(f);
	fclose(f);
	return tmp;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewFIS__ (JNIEnv *env, jclass)
{
	try { return (jlong) new FIS(); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	return (jlong)0;
}
//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewFIS__Ljava_lang_String_2Z (JNIEnv *env, jclass, jstring file_name, jboolean cover)
{
	const char *sif_file = NULL;
	FIS *fis = NULL;
	try
	{
		sif_file = get_native_string(env, file_name);
		fis = new FIS( (char *)sif_file, cover );  // ouverture d'un nouveau sif
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	release_native_string(sif_file);
	return (jlong)fis;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewFIS__J (JNIEnv *env, jclass, jlong sif_ptr)
{
	FIS *fis = (FIS *)sif_ptr;
	try { return (jlong) new FIS(*fis); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	return (jlong)0;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_InitClasses (JNIEnv *env, jclass, jlong jfis_ptr, jint output_number, jstring jdata_file)
{
	FIS *fis = (FIS *)jfis_ptr;
	char *data_file = NULL;
	double **data = NULL;
	int nbcol, nbex,rep;

	try
	{
		data_file = get_native_string(env, jdata_file);
		data = ReadSampleFile(data_file, nbcol, nbex);
		if( nbcol >= (fis->GetNbIn() + 1 + output_number) )
			// Observed output is part of the sample file{
		{
			rep=fis->ClassifCheck(data, nbex, 0);
		}

	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }

	release_native_string(data_file);

	//delete objects
	if (data != NULL){
		for(int i = 0; i < nbex; i ++) delete [] data[i];
		delete [] data;
	}

}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_InitPossibles (JNIEnv *env, jclass, jlong jfis_ptr)
{
	FIS *fis = (FIS *)jfis_ptr;
	try
	{
		int i;
		for(i = 0; i < fis ->GetNbOut(); i++) fis ->Out[i]->InitPossibles(fis->Rule, fis ->GetNbRule(),i);
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_AgregationResult (JNIEnv *env, jclass, jlong jfis_ptr, jint output_number)
{
	FIS *fis = (FIS *)jfis_ptr;
	jarray resulta = NULL;
	jobjectArray result = NULL;
	jclass objectCls = NULL;
	try
	{
		objectCls = env->FindClass( "java/lang/Object" );
		if( objectCls != NULL )
		{
			resulta = env->NewObjectArray( 2, objectCls, NULL );
			if( resulta != NULL )
			{
				result=(jobjectArray) resulta;
				int i = fis->Out[output_number] -> GetNbPossibles();
				const double * val = fis->Out[output_number] -> GetPossibles();
				double ret = 0.;  // le total pour normaliser
				for( int j=0 ; j<i ; j++ )
					//        ret += fis->MuInfer[output_number][j];
					ret += fis->Out[output_number] -> MuInfer[j];

				// output values
				jdoubleArray joutput_values = env->NewDoubleArray( i );
				if( joutput_values != NULL )
					env->SetDoubleArrayRegion( joutput_values, 0, i, (double *)val );
				env->SetObjectArrayElement( result, 0, joutput_values );
				// Normalized weight
				jdoubleArray jweight = env->NewDoubleArray( i );
				if( jweight != NULL )
				{
					double *weight = new double[i];
					for( int j=0 ; j<i ; j++ )
						//          weight[j] = fis->MuInfer[output_number][j];// / ret;
						weight[j] = fis->Out[output_number]->MuInfer[j];// / ret;

					env->SetDoubleArrayRegion( jweight, 0, i, weight );
					delete [] weight;
				}
				env->SetObjectArrayElement( result, 1, jweight );
			}
		}
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	if( objectCls != NULL ) env->DeleteLocalRef( objectCls );
	return result;
}

//---------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_fis_jnifis_AgregationImpliResult (JNIEnv *env, jclass, jlong jfis_ptr, jint output_number, jobject myVector )
{
	FIS *fis = (FIS *)jfis_ptr;
	jclass doubleCls = NULL, vectorCls = NULL;
	jmethodID constructorDoubleID, addElem;
	jint nbPoint = 0;
	double x, y = 0;

	try
	{
		vectorCls = env->FindClass( "java/util/Vector" );
		doubleCls = env->FindClass( "java/lang/Double" );
		addElem = env->GetMethodID(vectorCls, "addElement", "(Ljava/lang/Object;)V");
		constructorDoubleID = env->GetMethodID(doubleCls, "<init>", "(D)V" );

		if( vectorCls == NULL || doubleCls == NULL || addElem == NULL || constructorDoubleID == NULL)
			return (jint)NULL;
		else
		{
			if ( (fis->Out[output_number]->MfGlob!=NULL) && fis->Out[output_number]->MfGlob->NbParams()>0)
			{
				nbPoint = fis->Out[output_number]->MfGlob->NbParams();
				for (int i= 0; i<nbPoint ;i++)
				{
					fis->Out[output_number]->MfGlob->GetPoint(x,y,i);
					env->CallVoidMethod(myVector, addElem, env->NewObject( doubleCls, constructorDoubleID, x ));
					env->CallVoidMethod(myVector, addElem, env->NewObject( doubleCls, constructorDoubleID, y ));
				}
			}
		}
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	if(vectorCls  != NULL ) env->DeleteLocalRef( vectorCls );
	if( doubleCls != NULL ) env->DeleteLocalRef( doubleCls );
	return nbPoint;
}

//---------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_fis_jnifis_GetImpliConc (JNIEnv *env, jclass, jlong jfis_ptr, jint output_number, jint rule_number, jobject myVector )
{
	FIS *fis = (FIS *)jfis_ptr;
	jclass doubleCls = NULL, vectorCls = NULL;
	jmethodID constructorDoubleID, addElem;
	jint nbPoint = 0;
	double x, y = 0;

	try
	{
		vectorCls = env->FindClass( "java/util/Vector" );
		doubleCls = env->FindClass( "java/lang/Double" );
		addElem = env->GetMethodID(vectorCls, "addElement", "(Ljava/lang/Object;)V");
		constructorDoubleID = env->GetMethodID(doubleCls, "<init>", "(D)V" );

		if( vectorCls == NULL || doubleCls == NULL || addElem == NULL || constructorDoubleID == NULL)
			return (jint)NULL;
		else
		{
		  if ( (fis->Out[output_number]->MfConc) !=NULL )
			if ( (fis->Out[output_number]->MfConc[rule_number]!=NULL))
			  if (fis->Out[output_number]->MfConc[rule_number]->NbParams()>0)
			{
				nbPoint = fis->Out[output_number]->MfConc[rule_number]->NbParams();
				for (int i= 0; i<nbPoint ;i++)
				{
					fis->Out[output_number]->MfConc[rule_number]->GetPoint(x,y,i);
					env->CallVoidMethod(myVector, addElem, env->NewObject( doubleCls, constructorDoubleID, x ));
					env->CallVoidMethod(myVector, addElem, env->NewObject( doubleCls, constructorDoubleID, y ));
				}
			}
		}
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	if(vectorCls  != NULL ) env->DeleteLocalRef( vectorCls );
	if( doubleCls != NULL ) env->DeleteLocalRef( doubleCls );
	return nbPoint;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewFISFPA (JNIEnv *env, jclass, jlong fis_ptr, jstring data_file, jint strategy, jint min_card, jdouble min_deg)
{
	FIS *fis = (FIS *)fis_ptr;
	char *tmp = NULL;
	FISFPA *fisfpa = NULL;
	try
	{
		// sauvegarde le fis actuel dans un fichier temporaire
		tmp = TempFileName();
		FILE *f = fopen( tmp, "wt" );
		if( f == NULL )
			return (jlong)0;
		fis->PrintCfg(f);
		fclose(f);
		// genere le FIS FPA
		char *data = get_native_string(env, data_file);
		fisfpa = new FISFPA(tmp, data, strategy, min_card, min_deg );
		for( int i=0 ; i<fisfpa->GetNbOut() ; i++ )
			fisfpa->FpaRules(i);
		release_native_string(data);
		if (! fisfpa->ComputeNbActRule())
		{
			sprintf( ErrorMsg, "~NoRuleInitializedUsingThisConfiguration~\n");
			throw std::runtime_error( ErrorMsg );
		}

		// sauvegarde du fis genere dans un fichier temporaire
		f = fopen( tmp, "wt" );
		if( f == NULL )
			return (jlong)0;
		fisfpa->PrintCfg(f);
		fclose(f);
		delete fisfpa;
		// ouverture du fis
		FIS *new_fis = new FIS(tmp);
		char *name = new char [strlen(new_fis->Name)+5];
		strcpy( name, new_fis->Name );
		strcat( name, ".fpa" );
		new_fis->SetName(name);
		if(tmp != NULL) remove(tmp);
		delete [] tmp;
		delete [] name;
		return (jlong)new_fis;
	}
	catch( std::exception &e )
	{
		delete fisfpa;
		ExceptionToJava( env, e );
		if(tmp != NULL) remove(tmp);
		delete [] tmp;
	}
	return (jlong)0;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewGENFIS (JNIEnv *env, jclass, jlong fis_ptr, jstring data_file, jboolean sort, jdouble thres)
{
	FIS *fis = (FIS *)fis_ptr;
	// sauvegarde le fis actuel dans un fichier temporaire
	char *tmp = NULL;
	tmp = TempFileName();
	FILE *f = fopen( tmp, "wt" );
	if( f == NULL )
		return (jlong)0;
	fis->PrintCfg(f);
	fclose(f);
	try
	{
		char *data = NULL;
		if( data_file != NULL )
			data = get_native_string(env, data_file);
		FIS *genfis = new GENFIS(tmp, data, sort, thres);
		release_native_string(data);
		// sauvegarde du fis genere dans un fichier temporaire
		f = fopen( tmp, "wt" );
		if( f == NULL )
			return (jlong)0;
		genfis->PrintCfg(f);
		fclose(f);
		delete genfis;
		// ouverture et renommage du fis
		genfis = new FIS(tmp);
		char *name = new char [strlen(genfis->Name)+5];
		strcpy( name, genfis->Name );
		strcat( name, ".gen" );
		genfis->SetName(name);
		delete [] name;
		// suppression du fichier temp
		if(tmp != NULL) remove(tmp);
		delete [] tmp;
		return (jlong)genfis;
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	return (jlong)0;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewFISSIMPLE (JNIEnv *env, jclass, jlong fis_ptr, jstring jdata, jstring jresult_file, jdouble homogeneity_threshold, jdouble loss_performance, jint output_number, jdouble min_matching_degree, jdouble loss_coverage, jdouble coverage_min, jboolean keep_last, jdouble max_err, jboolean ruleRemoval, jboolean variableRemoval, jstring j_validation_file)
{
	FIS *fis = (FIS *)fis_ptr;
	FIS *fisimple = NULL;
	FISIMPLE *temp = NULL;
	char *data = NULL;
	char *validation_file = NULL;
	char *result_file = NULL;
	char *name = NULL;
	// sauvegarde le fis actuel dans un fichier temporaire
	char *tmp = NULL;
	tmp = TempFileName();
	FILE *f = fopen( tmp, "wt" );
	if( f == NULL )
		return (jlong)0;
	fis->PrintCfg(f);
	fclose(f);
	try
	{
		data = get_native_string(env, jdata);
		validation_file = get_native_string(env, j_validation_file);
		result_file = get_native_string(env, jresult_file);
		temp = new FISIMPLE(tmp, data, result_file, homogeneity_threshold, loss_performance, output_number, min_matching_degree, coverage_min, loss_coverage, keep_last, max_err);
		temp->SetRuleRemove(ruleRemoval);
		temp->SetVarRemove(variableRemoval);
		temp->InitTestData(validation_file);
		char *newfis = temp->Complete();
		temp->RemoveTmpFiles(newfis);
		if( newfis == NULL )
		{
			sprintf( ErrorMsg, "~NoSimplicationAchieved~\n");
			throw std::runtime_error( ErrorMsg );
		}
		// ouverture et renommage du fis
		fisimple = new FIS(newfis);
		remove(newfis);
		name = new char [strlen(fisimple->Name)+8];
		strcpy( name, fisimple->Name );
		strcat( name, ".simple" );
		fisimple->SetName(name);
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	release_native_string(data);
	release_native_string(validation_file);
	release_native_string(result_file);
	delete [] name;
	delete temp;
	if( tmp != NULL ) remove(tmp);
	delete [] tmp;
	return (jlong)fisimple;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewFISOLS (JNIEnv *env, jclass, jlong fis_ptr, jstring jdata, jint OutNum, jboolean keepExistingRules, jdouble standardDeviation, jstring jConjunction, jdouble totalUnexplainedVariance, jdouble maximumNumberOfRules, jboolean reduceVocabulary, jdouble degradationAllowed, jdouble numberOfConc, jboolean useRuleConc, jboolean outputFuzzification, jboolean removeTemporaryFiles, jstring jTestDataFile)
{
	FIS *fis = (FIS *)fis_ptr;
	FIS *fisOLS = NULL;
	FISOLS *S = NULL;
	char *data = NULL;
	char *testDataFile = NULL;
	char *finalFisFile = NULL;
	char *conjunction = NULL;
	char *name = NULL;
	// sauvegarde le fis actuel dans un fichier temporaire
	char *tmp = NULL;
	if(fis != NULL) {
		tmp = TempFileName();
		FILE *f = fopen( tmp, "wt" );
		if( f == NULL )
			return (jlong)0;
		fis->PrintCfg(f);
		fclose(f);
	}

	try
	{
		data = get_native_string(env, jdata);
		testDataFile = (char *)env->GetStringUTFChars( jTestDataFile, NULL );
		conjunction = (char *)env->GetStringUTFChars( jConjunction, NULL );
		if(fis != NULL)
		{
			S = new FISOLS(data, tmp, keepExistingRules, data);
			S->SetNumS(OutNum);
			S->SetNbOfOut(S->GetNbOut());
		}
		else
		{
			S = new FISOLS(data);
			S->SetUniqTol(standardDeviation);
			S->SetOriginalGauss(0);
			S->SetStd(1); // SFP partitions
			S->GenerateFIS(data);
		}
		if( !keepExistingRules ) {
			S->SetLoss(totalUnexplainedVariance);
			S->SetNbOfRules((jint)maximumNumberOfRules);
			S->SetConjunction(conjunction);
			S->SetTestFileName(testDataFile);
		}
		S->SetVocRed(reduceVocabulary);
		if(reduceVocabulary == 1) {
			S->SetAllowedLoss(degradationAllowed);
			S->SetNVocItem((jint)numberOfConc);
			S->SetVocData(useRuleConc);
			S->SetFuzzyOut(outputFuzzification);
		}
		S->SetRmTmpFile(removeTemporaryFiles);
		finalFisFile = TempFileName();
		S->SetResultFileName(finalFisFile);
		S->Run(data);
		delete S;

		// ouverture et renommage du fis
		fisOLS = new FIS(finalFisFile);
		name = new char [strlen(fisOLS->Name)+8];
		strcpy( name, fisOLS->Name );
		strcat( name, ".ols" );
		fisOLS->SetName(name);
	}
	catch( std::exception &e ) {
		ExceptionToJava( env, e );
	}
	release_native_string(data);
	env->ReleaseStringUTFChars( jTestDataFile, testDataFile );
	env->ReleaseStringUTFChars( jConjunction, conjunction );
	delete [] name;
	if( tmp != NULL ) remove(tmp);
	delete [] tmp;
	if( finalFisFile != NULL ) remove(finalFisFile);
	delete [] finalFisFile;
	return (jlong)fisOLS;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_outputVocabularyReduction (JNIEnv *env, jclass, jlong fis_ptr, jstring jdata, jint outputNumber, jdouble muthresh, jdouble perfLoss, jdouble nConc, jboolean dType)
{
	FIS *fis = (FIS *)fis_ptr;
	int nOut, nbcol, nbrow;
	char *data = NULL;
	double ** Data = NULL;
	data = get_native_string(env, jdata);
	try
	{
		// check output number
		nOut=fis->GetNbOut();
		if (outputNumber >(nOut-1))
		{
			sprintf( ErrorMsg, "~invalid~output~number: %d", (int)outputNumber);
			throw std::runtime_error( ErrorMsg );
		}
		//read data - allocate Data array
		Data = ReadSampleFile(data, nbcol, nbrow);
		// launch vocabulary reduction with the specified parameters
		fis->VocReduc(outputNumber,Data,nbrow,muthresh,perfLoss,(jint)nConc,dType);
	}
	catch( std::exception &e ) {
		ExceptionToJava( env, e );
	}
	release_native_string(data);
}

//---------------------------------------------------------------------------

JNIEXPORT jlong JNICALL Java_fis_jnifis_NewCustomFISOPT (JNIEnv *env, jclass, jlong fis_ptr, jstring jdata, jstring jkey, jint iterations, jdouble std_gauss, jint constraints, jint fails, jdouble center_thres, jint output_number, jboolean withcNear, jint seed_value, jdouble sw1, jdouble sw2, jdouble sw3, jint cardinal_vocabulary, jdouble mu_thres, jdouble loss_coverage)
{

	FIS *fis = (FIS *)fis_ptr;

	FIS *fisopt = NULL;
	char *data = NULL;
	double ** Exemples = NULL;
	int *ResClassif = NULL;
	double * Lab = NULL;
	sifopt *Opt = NULL;
	char *key = NULL;
	char *name = NULL;
	double Couverture, MaxError, museuil = mu_thres;
	int nbcol, nbrow, rep, cnear;
	double err=0.0;

	try
	{
		// check that the fis has at least one rule
		if (fis->GetNbRule() >0) {
			// open data file
			data = get_native_string(env, jdata);


			// optimization constraint (to be interfaced) - default=no
			cnear=0;
			if (withcNear) cnear=1;

			// seed for random numbers
			setseed(seed_value);

			Exemples = ReadSampleFile(data, nbcol, nbrow);

			//initialize resclassif array if necessary
			rep=fis->ClassCheck(ResClassif,Lab,Exemples,nbrow,output_number );

			fis->Performance(output_number, data, Couverture, MaxError, museuil, NULL, 0);
		    double InitialCoverage = Couverture;

			// construct sifopt
			Opt= new sifopt(fis,ResClassif,Exemples,nbrow,output_number);
			//check if the fis has rules

			Opt->pResClassif=&ResClassif;
			Opt->Exemples=Exemples;
			//build key from jkey (string argument)
			key = (char *)env->GetStringUTFChars( jkey, NULL );
			Opt->setKeyInit(key);
			//choose output
			Opt->numOutput = output_number;
			//write nbrow
			Opt->nbRow=nbrow;
			//write nbrow
			Opt->CNearActivate=cnear;
			//P = algorithm parameters
			Opt->P->maxIte = iterations;
			Opt->P->RangeNoise = std_gauss;
			Opt->P->maxConstraints = constraints;
			Opt->P->maxFail = fails;
			Opt->P->EpsilonCenter = center_thres;
			Opt->P->lamb1 = sw1;
			Opt->P->lamb2 = sw2;
			Opt->P->lamb3 = sw3;
			//following lines : parameters for evaluation function
			// see performance in sif.h
			//changer parametres ?
			Opt->fEvalArgs->add(Exemples);
			Opt->fEvalArgs->add(&nbrow);

			Opt->fEvalArgs->add(&Couverture);
			Opt->fEvalArgs->add(&MaxError);
			Opt->fEvalArgs->add(ResClassif);
			Opt->fEvalArgs->add(&museuil);
			Opt->fEvalArgs->add(Lab);

		    Opt->InitCov = InitialCoverage;
		    // replace the one computed within the constructor by the oiginal value
		    // to make sure the loss is kept within the tolerance with respect to the initial FIS
		    if(loss_coverage > 0.) Opt->CovLoss = loss_coverage;

			//launch optimizer : optimize entries or/and conclusion depending on key
			//uses parameters defined above
		    rep=Opt->launchEntry(err,false);

			// result FIS
			//if( Opt->Solution != NULL )
			// Reponse =1 : solution trouvee pour les entrees
			// Reponse =2 : solution trouvee pour les sorties
			if   ( ( Opt->Reponse == 1 ) || ( Opt->Reponse == 2 )|| ( Opt->Reponse == 3 ))
			{
				fisopt = new FIS(*Opt->Solution);
				name = new char[strlen(fisopt->Name)+5];
				sprintf( name, "%s.opt", fisopt->Name);
				fisopt->SetName(name);
			}
			// delete objects
			if (Exemples != NULL) {
				for(int i = 0; i < nbrow; i ++)
					delete [] Exemples[i];
				delete [] Exemples;
			}

			if (ResClassif !=NULL) delete [] ResClassif;
			delete [] name;
			if (Opt != NULL) delete Opt; // probleme !!!!
		}

	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }

	release_native_string(data);


	env->ReleaseStringUTFChars( jkey, key );

	return (jlong)fisopt;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewFISOPT (JNIEnv *env, jclass, jlong fis_ptr, jstring jfis_path, jstring jdata_file, jstring joptimzation_order, jboolean output, jboolean rules, jint number_of_loops, jdouble loss_coverage, jint iterations, jdouble std_gauss, jint constraints, jint fails, jint output_number, jint seed_value, jdouble sw1, jdouble sw2, jdouble sw3, jdouble mu_thres, jboolean cross_validation, jdouble validation_data_ratio, jboolean classif, jdouble precision, jint column_number, jint sample_number, jdouble cnear) {

	try
	{
		FIS *fis = ((FIS *)fis_ptr)->Clone();
		FIS *fis_result = NULL;
		const char *data_file = get_native_string(env, jdata_file);
		const char *fis_path = get_native_string(env, jfis_path);
		const char *optimzation_order = env->GetStringUTFChars(joptimzation_order, NULL);
		bool optim_in = strlen(optimzation_order) > 0;

		if(!optim_in && !(bool)rules && !(bool)output)
			return (jlong)NULL;

		double *num_inputs = new double[fis->GetNbIn()];
		int nb_inputs = ReadInputsOpt(optimzation_order, fis->GetNbIn(), num_inputs, true);
		if(optim_in && nb_inputs <= 0)
			return (jlong)NULL;

	    if((bool)cross_validation)
	    	genNSamples(seed_value, validation_data_ratio, classif, precision, column_number, (char *)data_file, sample_number);

	    const char *optim_name = "optim";
	    parstruct paramP = { iterations, std_gauss, constraints, fails, sw1, sw2, sw3 };
	    fis_result = loopoptimsample(fis, fis_path, data_file, optim_name, sample_number, number_of_loops, optim_in, (bool)output, (bool)rules, nb_inputs, num_inputs, output_number, seed_value, mu_thres, loss_coverage, paramP, false, true, cnear);

		const char *temp_fis = TemporaryFisName(fis);
	    computeWritePerf(fis_result, fis_path, data_file, "perf.res", temp_fis, optim_name, mu_thres, output_number, sample_number, true);

	    remove(temp_fis);
	    delete [] temp_fis;
	    if(num_inputs != NULL) delete [] num_inputs;
		release_native_string(fis_path);
		release_native_string(data_file);
		env->ReleaseStringUTFChars(joptimzation_order, optimzation_order);
	    delete fis;
	    string name(fis_result->Name);
	    name.append(".optim");
	    fis_result->SetName(name.c_str());
		return (jlong)fis_result;
	}
	catch( std::exception &e ) {
		ExceptionToJava( env, e );
		return (jlong)NULL;
	}
}

//---------------------------------------------------------------------------
JNIEXPORT jlongArray JNICALL Java_fis_jnifis_NewTreeFIS (JNIEnv *env, jclass, jlong fis_ptr, jstring jdata_file, jstring jtree_file,jstring jvalid_file, jstring jresult_file, jstring jperf_file,
		jint maxdepth, jdouble min_ed_gain, jdouble min_significant, jdouble ambiguity_thres, jint outputNumber,
		jboolean prune, jboolean split, jdouble relPerfLoss, jboolean grel, jboolean display, jdouble covThresh, jint minleafCard)
  //last argument is new - July 2010



{
	char *name = NULL;
	char * pruned_file = NULL;
	char *data_file = NULL;
	char *tree_file = NULL;
	char *valid_file = NULL;
	char *result_file=NULL;
	char *perf_file=NULL;
	FILE *new_stdout = NULL;
	FILE *fres=NULL;
	FIS *fis = (FIS *)fis_ptr;
	FISTREE *tree_fis = NULL;
	jlongArray result = NULL;
	jlong generated_fis[2];
	generated_fis[0] = 0;
	generated_fis[1] = 0;
	char *tmp = NULL;
	char *tmp_pruned = NULL;
	int res,nbrow, nbleaves=0,nbRemovedNodes=0,maxMF;
	int idisplay;
	int nbactrules, nbtotrules;
	int * misClass=NULL; // holds classif results
	double ** Examples = NULL;   // Ptr on data
	double * Lab=NULL;// labels for classif results
	double muperf,pf,coverage,maxError;
	FISOUT * crispout=NULL;
	double * outputClasses=NULL;
	InfoRB Irb;

	//-----------------------------------------------------------------------------------------
	// returns an array of one or two FIS ( if pruning)
	// outputNumber, split (leaf pruning or full split pruning): new arguments (java interface)
	//-----------------------------------------------------------------------------------------
	try
	{
		//  array for results
		result = env->NewLongArray(2);
		// lecture des noms de fichier
		data_file = get_native_string(env, jdata_file);
		tree_file = get_native_string(env, jtree_file);
		valid_file = get_native_string(env, jvalid_file);
		result_file = get_native_string(env, jresult_file);
		perf_file = get_native_string(env, jperf_file);
		// save current fis in a temporary file
		tmp = TempFileName();
		FILE *f = fopen( tmp, "wt" );
		if( f == NULL )
			throw std::runtime_error("~CannotOpenFISFile~");
		fis->PrintCfg(f);
		fclose(f);

		// transform boolean
		if(display)
			idisplay=2;
		else
			idisplay=0;


		// create a new fistree object
		tree_fis = new FISTREE(tmp, data_file, min_significant, ambiguity_thres,
				outputNumber, outputClasses, crispout, covThresh,minleafCard); //outputNumber is used here - warning - outputnumber shifted by minus 1
		// July 2010 - new parameter to be interfaced minleafcard (integer)
		// default value min(10,nb of examples/10)

		maxMF=tree_fis->CalcMaxMf();
		if (maxMF >99) throw std::runtime_error("~error~too~many~MFs~in~input~partitions\n~");

		// redirecting stdout into file treefis.out
		if( display == true )
			new_stdout = freopen("treefis.out", "a", stdout);
		// build tree with maximum depth

		nbleaves=tree_fis->BuildFullTree(tmp, tree_file, maxdepth, min_ed_gain, grel,
				crispout,outputClasses,idisplay);


		//allocation of arrays for perf
		Examples=tree_fis->GetExamples();
		nbrow=tree_fis->GetNbex();
		//first allocates Classes array in the classification case
		//call only once if Examples does not change
		tree_fis->ClassifCheck(Examples,nbrow, outputNumber);
		//then allocates misclass array and updates lab pointer for classification results
		//note: classcheck could do it all in one pass, but is also calls checkconsistency
		//which is useless here
		//call only once if Examples does not change
		tree_fis->ResClassifAlloc(misClass, Lab, outputNumber);

		// compute performance of generated fis
		pf=tree_fis->Perf(outputNumber,  Examples, nbrow,
				coverage,
				maxError,
				min_significant,
				misClass,
				Lab,true,NULL,0);
		// store RB characteristics in result.fistree

		fres = fopen(result_file, "wt");
		if (fres ==NULL)
			throw std::runtime_error("~CannotOpenResultFile~result.fistree\n");
		fprintf(fres, "fisname & Perf & Coverage & Max Error & MuMin & ");
		/*
		  tree_fis->AnalyzeRB(Irb,outputNumber);
		Irb.WriteHeader(fres);
		fprintf(fres, "%s &  %11.5f & %4.2f & %11.5f & %4.2f & ", tmp, pf, coverage,maxError , min_significant);
		Irb.Print(fres);
		*/
		// leak memory correction - bch Jul 21, 2009
		tree_fis->PerfRB(pf,coverage,maxError,outputNumber,fres);
		// warning: next calls to Perf will have to first reassign Lab ptr to Classes array
		// because PerRB function reassigns Classes array (in defuz object) but does not reassign Lab ptr

		// print specific tree characteristics in result.fistree
		tree_fis->AnalyzeTree(fres,maxdepth,idisplay);
		fprintf(fres, "\n");
		// rename fis
		generated_fis[0] = (jlong)new FIS(tmp);
		if( name != NULL )
			delete [] name;
		name = new char [strlen(((FIS *)generated_fis[0])->Name)+6];
		sprintf( name, "%s.tree", ((FIS *)generated_fis[0])->Name );
		((FIS *)generated_fis[0])->SetName(name);

		// pruning

		if( prune )
		{
			pruned_file = new char[strlen(tree_file)+10];
			sprintf( pruned_file, "%s.pruned", tree_file );
			tmp_pruned = TempFileName();
			muperf=min_significant;

			//tmp_pruned: FILENAME for fis equivalent to pruned tree
			//pruned_file: FILENAME for summary of pruned tree

			res = tree_fis->PruneTreePerf(tmp_pruned, pruned_file,valid_file,
					misClass,Lab,muperf, relPerfLoss,
					nbRemovedNodes, crispout, split,idisplay); //split is used here
			if( res <=0 )
				throw std::runtime_error("~ErrorInPruning~");

			Examples=tree_fis->GetExamples();
			nbrow=tree_fis->GetNbex();
			pf=tree_fis->Perf(outputNumber,  Examples, nbrow,
					coverage,
					maxError,
					muperf,
					misClass,
					Lab);

			// store RB characteristics in result.fistree file
			// temporary reset Number of rules=number of active rules
			// keep total number of rules in nbtotrules for correct destruction
			nbactrules=tree_fis->ComputeNbActRule();
			nbtotrules= tree_fis->GetNbRule();
			tree_fis->SetNbRule(nbactrules);
			/*
			tree_fis->AnalyzeRB(Irb,outputNumber);
			fprintf(fres, "%s &  %11.5f & %4.2f & %11.5f & %4.2f & ",tmp_pruned , pf, coverage,maxError , min_significant);
			Irb.Print(fres);
			*/
			// leak memory correction - bch Jul 21, 2009
			// compute performance on validation file for result.fistree
			pf = tree_fis->Performance(outputNumber, valid_file,coverage , maxError,muperf , perf_file,display );
			// store specific tree characteristics in result.fistree
			tree_fis->PerfRB(pf,coverage,maxError,outputNumber,fres);
			tree_fis->ResetAnalyze();
			tree_fis->AnalyzeTree(fres,maxdepth,idisplay);
			fprintf(fres, "\n");

			tree_fis->SetNbRule(nbtotrules);

			// generate pruned fis
			generated_fis[1] = (jlong)new FIS(tmp_pruned);
			if( name != NULL )
				delete [] name;
			name = new char [strlen(((FIS *)generated_fis[1])->Name)+16];
			sprintf( name, "%s.tree.pruned", ((FIS *)generated_fis[1])->Name );
			((FIS *)generated_fis[1])->SetName(name);

		}


		//delete tree
		if(crispout!=NULL)
			tree_fis->SetNbOut(tree_fis->GetNbOut()+1);

		if (outputClasses != NULL) delete [] outputClasses;
		if (misClass != NULL) delete [] misClass;
		// Lab and crispout are only ptrs - do not delete them
		if (tree_fis != NULL) delete tree_fis;

	}
	catch( std::exception &e )
	{
		delete (FIS *)generated_fis[0];
		delete (FIS *)generated_fis[1];
		generated_fis[0] = 0;
		generated_fis[1] = 0;
		ExceptionToJava( env, e );
	}


	release_native_string(data_file);
	release_native_string(tree_file);
	release_native_string(valid_file);
	release_native_string(result_file);
	release_native_string(perf_file);

	if( name != NULL ) delete [] name;
	if (pruned_file != NULL) delete [] pruned_file;
	if( display == true )
		fclose(new_stdout);
	if( fres != NULL ) fclose(fres);
	if( tmp != NULL ) remove(tmp);
	delete [] tmp;
	if( tmp_pruned != NULL ) remove(tmp_pruned);
	delete [] tmp_pruned;
	env->SetLongArrayRegion( result, 0, 2, generated_fis );
	return result;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_Links (JNIEnv *env, jclass, jlong fis_ptr, jstring jdata_file, jdouble threshold, jboolean sort, jstring jfile_prefix)
{
	FIS *fis = (FIS *)fis_ptr;
	char *data_file = NULL;
	char *file_prefix = NULL;
	FIS *fis_link = NULL;
	char *tmp = NULL;

	try

	{
		// read file names
		data_file = get_native_string(env, jdata_file);
		if( jfile_prefix != NULL ) file_prefix = get_native_string(env, jfile_prefix);
		// save current fis in temporary file
		tmp = TempFileName();
		FILE *f  = fopen( tmp, "wt" );
		if( f == NULL )
			throw std::runtime_error("~CannotOpenFISFile~: temp.fis~");
		fis->PrintCfg(f);
		fclose(f);
		// create a FISLINK object
		fis_link = new FISLINK( tmp, data_file, threshold, sort, file_prefix );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }

	release_native_string(data_file);
	release_native_string(file_prefix);
	if( tmp != NULL ) remove(tmp);
	delete [] tmp;
	if (fis_link != NULL) delete fis_link;
}
//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_DeleteFIS (JNIEnv *, jclass, jlong ptr)
{
	FIS *fis = (FIS *)ptr;
	if (fis != NULL) delete fis;
}

//---------------------------------------------------------------------------
JNIEXPORT jboolean JNICALL Java_fis_jnifis_EqualFIS (JNIEnv *, jclass, jlong sif1_ptr, jlong sif2_ptr)
{
	FIS *sif1 = (FIS *)sif1_ptr;
	FIS *sif2 = (FIS *)sif2_ptr;
	return (*sif1 != *sif2) ? false: true;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SaveFIS (JNIEnv *env, jclass, jlong ptr, jstring file_name)
{
	FIS *sif = (FIS *)ptr;
	const char *sif_file = get_native_string(env, file_name);

	try
	{
		FILE *f = fopen( sif_file, "wt" );    // ouverture du fichier
		if( f == NULL )                       // erreur d'ouverture
		{
			char s[250];
			sprintf( s, "~CannotSaveFISFile~: %.100s~\n %.100s~", sif_file, strerror(errno) );
			throw std::runtime_error( s );
		}
		sif->PrintCfg( f );              // ecriture de la config
		fclose( f );                          // fermeture du fichier
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	release_native_string(sif_file);
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SetFISName(JNIEnv *env, jclass, jlong ptr, jstring name)
{
	const char *sif_name = env->GetStringUTFChars( name, NULL );
	FIS *sif = (FIS *)ptr;

	try {
		// caution!
		// if name has an embedded file separator many procedures won't work!!!!
		sif->SetName( sif_name );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( name, sif_name );
}

//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_GetFISName(JNIEnv *env, jclass, jlong ptr)
{
	// caution!
	// if name has an embedded file separator many procedures won't work!!!!
	FIS *sif = (FIS *)ptr;
	return env->NewStringUTF( sif->Name );
}


//---------------------------------------------------------------------------

JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_TypeConjunction (JNIEnv *env, jclass)

{
	jobjectArray result = NULL;
	jarray resulta = NULL;

	// creation du tableau de retour
	jclass objectCls = env->FindClass( "java/lang/Object" );
	if( objectCls == NULL )
		return NULL;
	resulta = env->NewObjectArray( 3, objectCls, NULL );
	if( resulta != NULL )
	{
		result=(jobjectArray) resulta;
		env->SetObjectArrayElement( result, 0, env->NewStringUTF( RULE::PremiseProd() ) );
		env->SetObjectArrayElement( result, 1, env->NewStringUTF( RULE::PremiseMin() ) );
		env->SetObjectArrayElement( result, 2, env->NewStringUTF( RULE::PremiseLuka() ) );
	}
	env->DeleteLocalRef( objectCls );
	return result;
}

//---------------------------------------------------------------------------

JNIEXPORT void JNICALL Java_fis_jnifis_SetConjunction(JNIEnv *env, jclass, jlong ptr, jstring conjunction)
{
	FIS *sif = (FIS *)ptr;
	const char *sif_conjunction = env->GetStringUTFChars( conjunction, NULL );

	try { sif->SetConjunction( sif_conjunction ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( conjunction, sif_conjunction );
}

//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_GetConjunction(JNIEnv *env, jclass, jlong ptr)
{
	FIS *sif = (FIS *)ptr;
	return env->NewStringUTF( sif->TypeConj() );
}



//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    TypeMissingValues
 * Signature: ()[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_TypeMissingValues (JNIEnv *env, jclass)
{
	jobjectArray result=NULL;
	jarray resulta=NULL;

	// creation du tableau de retour
	jclass objectCls = env->FindClass( "java/lang/Object" );
	if( objectCls == NULL )
		return NULL;
	resulta = env->NewObjectArray( 2, objectCls, NULL );
	if( resulta != NULL )
	{
		result=(jobjectArray) resulta;
		env->SetObjectArrayElement( result, 0, env->NewStringUTF( FIS::RandomMissingValues() ) );
		env->SetObjectArrayElement( result, 1, env->NewStringUTF( FIS::MeanMissingValues() ) );
	}
	env->DeleteLocalRef( objectCls );
	return result;
}

//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    NbInputs
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_fis_jnifis_NbInputs (JNIEnv *, jclass, jlong ptr)
{
	FIS *sif = (FIS *)ptr;
	return sif->GetNbIn();
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    NbOutputs
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_fis_jnifis_NbOutputs (JNIEnv *, jclass, jlong ptr)
{
	FIS *sif = (FIS *)ptr;
	return sif->GetNbOut();
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    NbRules
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_fis_jnifis_NbRules  (JNIEnv *, jclass, jlong ptr)
{
	FIS *sif = (FIS *)ptr;
	return sif->GetNbRule();
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    GetInput
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_fis_jnifis_GetInput (JNIEnv *, jclass, jlong ptr, jint input_number)
{
	FIS *sif = (FIS *)ptr;
	return (jlong)sif->In[input_number];
}


//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    AddInput
 * Signature: (JJ)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_AddInput (JNIEnv *env, jclass, jlong ptr, jlong input_ptr)
{
	FIS *sif = (FIS *)ptr;
	FISIN *entree = (FISIN *)input_ptr;
	try { sif->AddInput( entree ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    RemoveInput
 * Signature: (JI)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_RemoveInput (JNIEnv *env, jclass, jlong ptr, jint input_number)
{

	FIS *sif = (FIS *)ptr;
	try { sif->RemoveInput( input_number ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}


//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    ReplaceInput
 * Signature: (JIJ)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_ReplaceInput (JNIEnv *env, jclass, jlong ptr, jint input_number, jlong input_ptr)
{

	FIS *sif = (FIS *)ptr;
	FISIN *input = (FISIN *)input_ptr;
	try { sif->ReplaceInput( input_number, input ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    ReplaceOutput
 * Signature: (JIJ)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_ReplaceOutput (JNIEnv *env, jclass, jlong ptr, jint output_number, jlong output_ptr)
{
	FIS *sif = (FIS *)ptr;
	FISOUT *output = (FISOUT *)output_ptr;
	try { sif->ReplaceOutput( output_number, output ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}
//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    AddOutput
 * Signature: (JJ)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_AddOutput (JNIEnv *env, jclass, jlong ptr, jlong output_ptr)
{
	FIS *sif = (FIS *)ptr;
	FISOUT *sortie = (FISOUT *)output_ptr;
	try { sif->AddOutput( sortie ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    RemoveOutput
 * Signature: (JI)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_RemoveOutput (JNIEnv *env, jclass, jlong ptr, jint output_number)
{

	FIS *sif = (FIS *)ptr;
	try { sif->RemoveOutput( output_number ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------

/*
 * Class:     fis_JNIFIS
 * Method:    GetInput
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_fis_jnifis_GetOutput (JNIEnv *, jclass, jlong ptr, jint output_number)
{
	FIS *sif = (FIS *)ptr;
	return (jlong)sif->Out[output_number];
}

//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    GetRule
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_fis_jnifis_GetRule (JNIEnv *, jclass, jlong ptr, jint rule_number)
{
	FIS *sif = (FIS *)ptr;
	return (jlong)sif->Rule[rule_number];
}

//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    AddRule
 * Signature: (JJ)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_AddRule (JNIEnv *env , jclass, jlong ptr, jlong rule_ptr)
{
	FIS *sif = (FIS *)ptr;
	RULE *regle = (RULE *)rule_ptr;
	try { sif->AddRule( regle ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}


//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    RemoveRule
 * Signature: (JI)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_RemoveRule (JNIEnv *env, jclass, jlong sif_ptr, jint rule_number)
{
	FIS *sif = (FIS *)sif_ptr;
	try { sif->RemoveRule( rule_number ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}


//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    NewInterface
 * Signature: ()V
 */
/*JNIEXPORT void JNICALL Java_fis_jnifis_NewInterface (JNIEnv *env, jclass)
  {
  try { NewInterface(); }
  catch( std::exception &e ) { ExceptionToJava( env, e ); }
  }  */

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    NewInput
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewInput__ (JNIEnv *env, jclass)
{
	// creation d'une entree vide
	try
	{
		//FISIN *input = new FISIN();
		//input->SetRange( 0, 1 );
		return (jlong) new FISIN();
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	return (jlong)0;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewInput__J (JNIEnv *env, jclass, jlong ptr)
{
	// creation d'une entree par recopie
	FISIN *entree = (FISIN *)ptr;
	try { return (jlong) new FISIN(*entree); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	return (jlong)0;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewInput___3DDDZ (JNIEnv *env, jclass, jdoubleArray jarrayvalues, jdouble jmin, jdouble jmax, jboolean jsort)
{

	FISIN *input = NULL;

	jdouble *jvalues = NULL;

	try
	{
		int nb_values = env->GetArrayLength( jarrayvalues );
		jvalues = env->GetDoubleArrayElements( jarrayvalues, NULL );
		input = new FISIN(jvalues, nb_values, jmin, jmax, jsort);
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseDoubleArrayElements( jarrayvalues, jvalues, 0 );
	return (jlong)input;
}


//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewRegularInput (JNIEnv *env, jclass, jint nb_sef, jdouble min, jdouble max)
{
	// creation d'une entree reguliere
	try { return (jlong) new FISIN( nb_sef, min, max ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	return (jlong)0;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewIrregularInput (JNIEnv *env, jclass, jdoubleArray jsommets, jint nb_sef, jdouble min, jdouble max)
{
	// creation d'une entree irreguliere
	double *sommets  = NULL;
	FISIN *new_input = NULL;
	try
	{
		sommets = env->GetDoubleArrayElements( jsommets, NULL );
		new_input = new FISIN( sommets, nb_sef, min, max );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseDoubleArrayElements( jsommets, sommets, 0 );
	return (jlong)new_input;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_DeleteInput (JNIEnv *, jclass, jlong ptr)
{
	FISIN *entree = (FISIN *)ptr;
	delete entree;
}

//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_GetInputName (JNIEnv *env, jclass, jlong ptr)
{
	FISIN *entree = (FISIN *)ptr;
	return env->NewStringUTF( entree->Name );
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    GetInputActive
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_fis_jnifis_GetInputActive (JNIEnv *, jclass, jlong ptr)
{

	FISIN *entree = (FISIN *)ptr;
	return (jboolean)entree->IsActive();
}

//---------------------------------------------------------------------------
/*

 * Class:     fis_JNIFIS
 * Method:    SetInput
 * Signature: (JLjava/lang/String;DDZ)V
 */
/*JNIEXPORT void JNICALL Java_fis_jnifis_SetInput (JNIEnv *env, jclass, jlong ptr, jstring name, jdouble domain_inf, jdouble domain_sup, jboolean active)
  {
  FISIN *entree = (FISIN *)ptr;
  const char *input_name = env->GetStringUTFChars( name, NULL );

  try
  {
  entree->SetName( (char *)input_name );
  if( active == true )
  entree->Active();
  else
  entree->Desactive();
  entree->SetDomain( domain_inf, domain_sup );
  }
  catch( std::exception &e ) { ExceptionToJava( env, e ); }

  env->ReleaseStringUTFChars( name, input_name );
  }    */

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    SetInputName
 * Signature: (JLjava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_SetInputName (JNIEnv *env, jclass, jlong ptr, jstring name)
{
	FISIN *entree = (FISIN *)ptr;
	const char *input_name = env->GetStringUTFChars( name, NULL );

	try { entree->SetName( (char *)input_name ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }

	env->ReleaseStringUTFChars( name, input_name );
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    SetInputRange
 * Signature: (JDD)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_SetInputRange (JNIEnv *env, jclass, jlong ptr, jdouble range_inf, jdouble range_sup)
{
	FISIN *entree = (FISIN *)ptr;

	try { entree->SetRange( range_inf, range_sup ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    SetInputActive
 * Signature: (JZ)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_SetInputActive (JNIEnv *env, jclass, jlong ptr, jboolean active)
{
	FISIN *entree = (FISIN *)ptr;

	try
	{
		if( active == true )
			entree->Activate();
		else
			entree->Deactivate();
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    GetInputRange
 * Signature: (J)[D
 */
JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_GetInputRange (JNIEnv *env, jclass, jlong ptr)
{

	jdoubleArray result=NULL;

	FISIN *entree = (FISIN *)ptr;

	result = env->NewDoubleArray( 2 );
	if( result != NULL )
	{
		jdouble domain[2];
		domain[0] = entree->min();
		domain[1] = entree->max();
		env->SetDoubleArrayRegion( result, 0, 2, domain );
	}
	return result;
}


//---------------------------------------------------------------------------

JNIEXPORT jint JNICALL Java_fis_jnifis_GetNbMF (JNIEnv *, jclass, jlong ptr)

{
	FISIN *entree = (FISIN *)ptr;
	return entree->GetNbMf();
}

//---------------------------------------------------------------------------

JNIEXPORT jlong JNICALL Java_fis_jnifis_GetMF (JNIEnv *, jclass, jlong ptr, jint sef_number)

{
	FISIN *entree = (FISIN *)ptr;
	return (jlong)entree->GetMF(sef_number);
}

//---------------------------------------------------------------------------

JNIEXPORT jlong JNICALL Java_fis_jnifis_GetMFGlob (JNIEnv *, jclass, jlong ptr)

{
	FISOUT *sortie = (FISOUT *)ptr;
	//printf("GetMfGlob %s \n",sortie->GetMfGlob()->GetType());
	return (jlong)sortie->GetMFGlob();
}


//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_GetMFConc (JNIEnv *, jclass, jlong ptr, jint rule_number)

{
	FISOUT *sortie = (FISOUT *)ptr;
	return (jlong)sortie->GetMFConc(rule_number);
}

//---------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_fis_jnifis_GetBuildFuzIn (JNIEnv *env, jclass, jlong ptr, jint input_number, jdoubleArray kw, jdoubleArray sw, jdoubleArray inputs_values, jobject myVector)
{
	FIS *sif = (FIS *)ptr;
	MFDPOSS ** fuzval=NULL;
	double *sw_val = NULL;
	double *values = NULL;
	double *kw_val = NULL;
	jclass doubleCls, vectorCls = NULL;
	jmethodID constructorDoubleID, addElem;
	jint nbPoint = 0;
	double x, y = 0;

	// Lecture des valeurs des sw et kw et values
	kw_val = new double[sif->GetNbIn()];
	sw_val = new double[sif->GetNbIn()];
	values = new double[sif->GetNbIn()];
	jdouble *jkw = env->GetDoubleArrayElements( kw, NULL );
	jdouble *jsw = env->GetDoubleArrayElements( sw, NULL );
	jdouble *jvalues = env->GetDoubleArrayElements( inputs_values, NULL );
	for( int i=0 ; i < sif->GetNbIn() ; i++ )
	{
		kw_val[i] = (double)jkw[i];
		sw_val[i] = (double)jsw[i];
		values[i] = (double)jvalues[i];
	}
	env->ReleaseDoubleArrayElements( kw, jkw, 0 );
	env->ReleaseDoubleArrayElements( sw, jsw, 0 );
	env->ReleaseDoubleArrayElements( inputs_values, jvalues, 0 );

	// BuildFuzIn
	sif->BuildFuzIn(values, kw_val, sw_val, fuzval);

	// Resultat
	vectorCls = env->FindClass( "java/util/Vector" );
	doubleCls = env->FindClass( "java/lang/Double" );
	addElem = env->GetMethodID(vectorCls, "addElement", "(Ljava/lang/Object;)V");
	constructorDoubleID = env->GetMethodID(doubleCls, "<init>", "(D)V" );
	if( vectorCls == NULL || doubleCls == NULL || addElem == NULL || constructorDoubleID == NULL)
		return (jint)NULL;
	else
	{
		if ( fuzval[input_number]->NbParams() > 0)
		{
			nbPoint = fuzval[input_number]->NbParams();
			for (int i= 0; i<nbPoint ;i++)
			{
				fuzval[input_number]->GetPoint(x,y,i);
				env->CallVoidMethod(myVector, addElem, env->NewObject( doubleCls, constructorDoubleID, x ));
				env->CallVoidMethod(myVector, addElem, env->NewObject( doubleCls, constructorDoubleID, y ));
			}
		}
	}
	if(vectorCls  != NULL ) env->DeleteLocalRef( vectorCls );
	if( doubleCls != NULL ) env->DeleteLocalRef( doubleCls );
	delete [] values;
	delete [] kw_val;
	delete [] sw_val;
	return nbPoint;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_AddMF (JNIEnv *env, jclass, jlong input_ptr, jlong sef_ptr)
{
	FISIN *entree = (FISIN *)input_ptr;
	MF *sef = (MF *)sef_ptr;
	try { entree->AddMF( sef ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------
/*JNIEXPORT void JNICALL Java_fis_jnifis_RemoveMF (JNIEnv *env, jclass, jlong input_ptr, jint sef_number)
  {
  FISIN *entree = (FISIN *)input_ptr;
  try { entree->RemoveMF( sef_number ); }
  catch( std::exception &e ) { ExceptionToJava( env, e ); }
  }  */

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_RemoveMFInInput (JNIEnv *env, jclass, jlong fis_ptr, jint input_number, jint mf_number)
{
	FIS *fis = (FIS *)fis_ptr;
	try { fis->RemoveMFInInput( input_number, mf_number ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_RemoveMFInOutput (JNIEnv *env, jclass, jlong fis_ptr, jint output_number, jint mf_number)
{
	FIS *fis = (FIS *)fis_ptr;
	try { fis->RemoveMFInOutput( output_number, mf_number ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_ReplaceMF (JNIEnv *env, jclass, jlong input_ptr, jint mf_number, jlong new_sef_ptr)
{


	FISIN *entree = (FISIN *)input_ptr;
	MF *new_sef = (MF *)new_sef_ptr;
	try {
	  entree->ReplaceMF( mf_number, new_sef );

	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_ReplaceMFOutput (JNIEnv *env, jclass, jlong input_ptr, jint mf_number, jlong new_sef_ptr)
{
	FISOUT *sortie = (FISOUT *)input_ptr;
	MF *new_sef = (MF *)new_sef_ptr;
	try { sortie->ReplaceMF( mf_number, new_sef ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_GetMFName (JNIEnv *env, jclass, jlong ptr)
{
	MF *sef = (MF *)ptr;
	return env->NewStringUTF( sef->Name );
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SetMFName (JNIEnv *env, jclass, jlong ptr, jstring name)
{
	const char *sef_name = env->GetStringUTFChars( name, NULL );
	MF *sef = (MF *)ptr;
	try { sef->SetName( (char *)sef_name ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( name, sef_name );
}

//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_GetMFType (JNIEnv *env, jclass, jlong ptr)
{
	MF *sef = (MF *)ptr;
	return env->NewStringUTF( sef->GetType() );
}

//---------------------------------------------------------------------------
JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_GetMFParam (JNIEnv *env, jclass, jlong ptr)
{
	MF *sef = (MF *)ptr;
	jdoubleArray result=NULL;

	result = env->NewDoubleArray( sef->NbParams() );
	if( result != NULL )
	{
		double *params = new double[sef->NbParams()];
		sef->GetParams( params );
		env->SetDoubleArrayRegion( result, 0, sef->NbParams(), params );
		delete [] params;
	}
	return result;
}


//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_DeleteMF (JNIEnv *, jclass, jlong ptr)
{
	MF *sef = (MF *)ptr;
	delete sef;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewOutput (JNIEnv *env, jclass, jlong output_ptr)
{
	FISOUT *output = (FISOUT *)output_ptr;
	try { return (jlong) output->Clone(); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }

	return (jlong)0;

}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    NewRegularOutput
 * Signature: (IDDLjava/lang/String;ZDZ)J
 */
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewRegularOutput (JNIEnv *env, jclass, jint nb_sef, jdouble min, jdouble max, jstring jdefuz, jboolean classif, jdouble default_value, jstring jdisj)
{
	const char *defuz = env->GetStringUTFChars( jdefuz, NULL );
	const char *disj = env->GetStringUTFChars( jdisj, NULL );
	OUT_FUZZY *new_output = NULL;
	int *sorted=NULL;

	try
	  {
	    new_output = new OUT_FUZZY( nb_sef, min, max, (char *)defuz,( char *)disj, classif, default_value );
	   // implicative output-QSP regular grid
	    if(!strcmp(defuz, OUT_FUZZY::ImpFuzzyDefuz()))
	      new_output->Sfp2Qsp(sorted);
	  }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( jdefuz, defuz );
	env->ReleaseStringUTFChars( jdisj, disj );
	return (jlong)new_output;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewIrregularOutput (JNIEnv *env, jclass, jdoubleArray jsommets, jint nb_sef, jdouble min, jdouble max, jstring jdefuz, jboolean classif, jdouble default_value, jstring jdisj)
{
	const char *defuz = env->GetStringUTFChars( jdefuz, NULL );
	const char *disj = env->GetStringUTFChars( jdisj, NULL );
	double *sommets  = NULL;
	OUT_FUZZY *new_output = NULL;
	try
	{
		sommets = env->GetDoubleArrayElements( jsommets, NULL );
		new_output = new OUT_FUZZY( sommets, nb_sef, min, max, 1, (char *)defuz,( char *)disj, classif, default_value );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( jdefuz, defuz );
	env->ReleaseDoubleArrayElements( jsommets, sommets, 0 );
	env->ReleaseStringUTFChars( jdisj, disj );
	return (jlong)new_output;
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    NewOutputFloue
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewOutputFloue__ (JNIEnv *env, jclass)
{
	try { return (jlong) new OUT_FUZZY( 0, 0, 1, (char *)OUT_FUZZY::AreaDefuz(), (char *)OUT_FUZZY::DisjSum(), 0 ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }

	return (jlong)0;

}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    NewOutputFloue
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewOutputFloue__J (JNIEnv *env , jclass, jlong output_ptr)
{
	FISOUT *output = (FISOUT *)output_ptr;
	try { return (jlong) new OUT_FUZZY( *output ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }

	return (jlong)0;
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    NewOutputNette
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewOutputNette__ (JNIEnv *env, jclass)
{
	try
	{
		FISOUT *new_output = new OUT_CRISP();
		new_output->SetRange( 0, 1 );
		return (jlong) new_output;
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }

	return (jlong)0;

}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    NewOutputNette
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewOutputNette__J (JNIEnv *env, jclass, jlong output_ptr)
{
	FISOUT *output = (FISOUT *)output_ptr;
	try { return (jlong) new OUT_CRISP( *output ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }

	return (jlong)0;
}


//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    SetOutputClassif
 * Signature: (JZ)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_SetOutputClassif (JNIEnv *env, jclass, jlong fis_ptr, jint output_number, jboolean classif)
{
	FIS *fis = (FIS *)fis_ptr;
	FISOUT *output = fis->Out[output_number];
	try { output->Classification( classif ); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------
JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_GetOutputClassesLabels (JNIEnv *env, jclass, jlong fis_ptr, jint output_number)
{
	FIS *fis = (FIS *)fis_ptr;
	jdoubleArray result=NULL;


	//int nb_classes = fis->Def[output_number]->NbClasses();
	int nb_classes = fis->Out[output_number]->Def->NbClasses();

	result = env->NewDoubleArray( nb_classes );
	//DEFUZ_SugenoClassif *def = (DEFUZ_SugenoClassif *)fis->Def[output_number];
	DEFUZ_SugenoClassif *def = (DEFUZ_SugenoClassif *)fis->Out[output_number]->Def;

	env->SetDoubleArrayRegion( result, 0, nb_classes, (double *)def->ClasLabel() );

	try
	{

	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	return result;
}

/*jdoubleArray result;
RULE *regle = (RULE *)ptr;

result = env->NewDoubleArray( regle->GetNbConc() );
if( result != NULL )
  for( int i=0 ; i<regle->GetNbConc() ; i++ )
    {
    jdouble action = regle->GetAConc( i );
    env->SetDoubleArrayRegion( result, i, 1, &action );
    }
return result;   */



//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    SetOutput
 * Signature: (JLjava/lang/String;DZ)V
 */
/*JNIEXPORT void JNICALL Java_fis_jnifis_SetOutput (JNIEnv *env, jclass, jlong ptr, jstring defuz, jdouble default_value, jboolean classif)
{

FISOUT *sortie = (FISOUT *)ptr;
const char *op_defuz = env->GetStringUTFChars( defuz, NULL );

try
  {
  sortie->SetOpDefuz( (char *)op_defuz );
  sortie->ValeurDefaut( default_value );
  sortie->Classification( (int)classif );
  }
catch( std::exception &e ) { ExceptionToJava( env, e ); }

env->ReleaseStringUTFChars( defuz, op_defuz );
}     */

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    GetOutputNature
 * Signature: (J)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_fis_jnifis_GetOutputNature (JNIEnv *env, jclass, jlong ptr)
{
	FISOUT *sortie = (FISOUT *)ptr;
	return env->NewStringUTF( sortie->GetOutputType() );
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    SetOutputDefaultValue
 * Signature: (JD)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_SetOutputDefaultValue (JNIEnv *, jclass, jlong output_ptr, jdouble default_value)
{
	FISOUT *output = (FISOUT *)output_ptr;
	output->DefaultValue( default_value );
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    SetOutputDefuz
 * Signature: (JLjava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_SetOutputDefuz (JNIEnv *env, jclass, jlong fis_ptr, jint output_number, jstring defuz_string)
{
	FIS *fis = (FIS *)fis_ptr;
	FISOUT *output = fis->Out[output_number];
	const char *defuz = env->GetStringUTFChars( defuz_string, NULL );
	try
	{
		output->SetOpDefuz( (char *)defuz );
		if((output->GetNbMf() > 1)  && !(strcmp(output->GetMF(0)->GetType(), MFTRAPINF::Type())) && !(strcmp(output->GetMF(output->GetNbMf()-1)->GetType(), MFTRAPSUP::Type())))
		{
			double p[3];
			double pe[3];
			output->GetMF(0)->GetParams(p);
			output->GetMF(output->GetNbMf()-1)->GetParams(pe);
			if((output->min() <= p[1]) && (output->max() >= pe[1]))
				((OUT_FUZZY *)output)->OutCoverage();
		}
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( defuz_string, defuz );
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    SetOutputImplicatif
 * Signature: (JZ)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_SetOutputImplicatifFuzzy (JNIEnv *env, jclass, jlong fis_ptr, jint output_number, jboolean impli)
{
	FIS *fis = (FIS *)fis_ptr;
	FISOUT *output = fis->Out[output_number];
	try
	{
		if(impli)
		{
			output->SetOpDefuz( (char *)OUT_FUZZY::ImpFuzzyDefuz());//defuzzyfication
			fis->Out[output_number]->SetOpDisj(OUT_FUZZY::DisjIrg());//disjunction
		}
		else
		{
			output->SetOpDefuz( (char *)OUT_FUZZY::AreaDefuz());//defuzzyfication
			fis->Out[output_number]->SetOpDisj(OUT_FUZZY::DisjSum());//disjunction
		}
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_jnifis
 * Method:    SetOutputDisjunction
 * Signature: (JILjava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_fis_jnifis_SetOutputDisjunction(JNIEnv *env, jclass, jlong fis_ptr, jint output_number, jstring disj_string)
{
	FIS *fis = (FIS *)fis_ptr;
	const char *disj = env->GetStringUTFChars( disj_string, NULL );
	try { fis->Out[output_number]->SetOpDisj(disj); }
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( disj_string, disj );
}


//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    GetOutputDefuz
 * Signature: (J)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_fis_jnifis_GetOutputDefuz (JNIEnv *env, jclass, jlong ptr)
{
	FISOUT *sortie = (FISOUT *)ptr;
	return env->NewStringUTF( sortie->Defuzzify() );
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_jnifis
 * Method:    GetOutputDisjunction
 * Signature: (J)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_fis_jnifis_GetOutputDisjunction(JNIEnv *env, jclass, jlong ptr)
{
	FISOUT *sortie = (FISOUT *)ptr;
	return env->NewStringUTF( sortie->Disjunct() );
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    GetOutputDefaultValue
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_fis_jnifis_GetOutputDefaultValue (JNIEnv *, jclass, jlong ptr)
{
	FISOUT *sortie = (FISOUT *)ptr;
	return (jdouble)sortie->DefaultValue();
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    GetOutputClassif
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_fis_jnifis_GetOutputClassif (JNIEnv *, jclass, jlong ptr)
{
	FISOUT *sortie = (FISOUT *)ptr;
	return (jboolean)sortie->Classification();
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_jnifis
 * Method:    GetTypeImplicatif
 * Signature: (J)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_fis_jnifis_GetTypeImplicatif  (JNIEnv *env, jclass, jlong ptr)
{
	FISOUT *sortie = (FISOUT *)ptr;
	return env->NewStringUTF( sortie->GetOutputType() );
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    TypeDefuzzificationImplicatif
 * Signature: ()[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_TypeDefuzzificationImplicatif (JNIEnv *env, jclass)
{
	jobjectArray result = NULL;
	jarray resulta = NULL;

	// creation du tableau de retour
	jclass objectCls = env->FindClass( "java/lang/Object" );
	if( objectCls == NULL )
		return NULL;
	resulta = env->NewObjectArray( 1, objectCls, NULL );
	if( resulta != NULL )
	{
		result=(jobjectArray) resulta;
		env->SetObjectArrayElement( result, 0, env->NewStringUTF( OUT_FUZZY::ImpFuzzyDefuz() ) );
	}
	env->DeleteLocalRef( objectCls );
	return result;
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_jnifis
 * Method:    TypeDisjunctionImplicatif
 * Signature: ()[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_TypeDisjunctionImplicatif(JNIEnv *env, jclass)
{
	jobjectArray result = NULL;
	jarray resulta;

	jclass objectCls = env->FindClass( "java/lang/Object" );
	if( objectCls == NULL )
		return NULL;

	resulta = env->NewObjectArray( 3, objectCls, NULL );
	if( resulta != NULL )
	{
		result=(jobjectArray) resulta;
		env->SetObjectArrayElement( result, 0, env->NewStringUTF( OUT_FUZZY::DisjIrg() ) );
		env->SetObjectArrayElement( result, 1, env->NewStringUTF( OUT_FUZZY::DisjIgg() ) );
		env->SetObjectArrayElement( result, 2, env->NewStringUTF( OUT_FUZZY::DisjIgd() ) );
	}
	env->DeleteLocalRef( objectCls );
	return result;
}

//---------------------------------------------------------------------------

/*
 * Class:     fis_JNIFIS
 * Method:    TypeDefuzzificationFloue
 * Signature: ()[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_TypeDefuzzificationFloue (JNIEnv *env, jclass)
{
	jobjectArray result = NULL;
	jarray resulta = NULL;

	// creation du tableau de retour
	jclass objectCls = env->FindClass( "java/lang/Object" );
	if( objectCls == NULL )
		return NULL;
	resulta = env->NewObjectArray( 3, objectCls, NULL );
	if( resulta != NULL )
	{
		result=(jobjectArray) resulta;
		env->SetObjectArrayElement( result, 0, env->NewStringUTF( OUT_FUZZY::AreaDefuz() ) );
		env->SetObjectArrayElement( result, 1, env->NewStringUTF( OUT_FUZZY::MeanMaxDefuz() ) );
		env->SetObjectArrayElement( result, 2, env->NewStringUTF( OUT_FUZZY::SugenoDefuz() ) );
	}
	env->DeleteLocalRef( objectCls );
	return result;
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    TypeDefuzzificationNette
 * Signature: ()[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_TypeDefuzzificationNette (JNIEnv *env, jclass)
{
	jobjectArray result = NULL;
	jarray resulta = NULL;

	// creation du tableau de retour
	jclass objectCls = env->FindClass( "java/lang/Object" );
	if( objectCls == NULL )
		return NULL;
	//modif taille pour intro defuzmeanmaxcrisp
	resulta = env->NewObjectArray( 2, objectCls, NULL );
	if( resulta != NULL )
	{
		result=(jobjectArray) resulta;
		env->SetObjectArrayElement( result, 0, env->NewStringUTF( OUT_CRISP::SugenoDefuz() ) );
		//modif pour intro defuzmeanmaxcrisp
		env->SetObjectArrayElement( result, 1, env->NewStringUTF( OUT_CRISP::MaxCrispDefuz() ) );
	}
	env->DeleteLocalRef( objectCls );
	return result;
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_jnifis
 * Method:    TypeDisjunctionFloue
 * Signature: ()[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_TypeDisjunctionFloue(JNIEnv *env, jclass)
{
	jobjectArray result = NULL;
	jarray resulta;

	jclass objectCls = env->FindClass( "java/lang/Object" );
	if( objectCls == NULL )
		return NULL;

	resulta = env->NewObjectArray( 2, objectCls, NULL );
	if( resulta != NULL )
	{
		result=(jobjectArray) resulta;
		env->SetObjectArrayElement( result, 0, env->NewStringUTF( OUT_FUZZY::DisjSum() ) );
		env->SetObjectArrayElement( result, 1, env->NewStringUTF( OUT_FUZZY::DisjMax() ) );
	}
	env->DeleteLocalRef( objectCls );
	return result;
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_jnifis
 * Method:    TypeDisjunctionNette
 * Signature: ()[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_TypeDisjunctionNette(JNIEnv *env, jclass)
{
	jobjectArray result = NULL;
	jarray resulta;

	jclass objectCls = env->FindClass( "java/lang/Object" );
	if( objectCls == NULL )
		return NULL;

	resulta = env->NewObjectArray( 2, objectCls, NULL );
	if( resulta != NULL )
	{
		result=(jobjectArray) resulta;

		env->SetObjectArrayElement( result, 0, env->NewStringUTF( OUT_CRISP::DisjSum() ) );
		env->SetObjectArrayElement( result, 1, env->NewStringUTF( OUT_CRISP::DisjMax() ) );
	}
	env->DeleteLocalRef( objectCls );
	return result;
}

//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    TypeNature
 * Signature: ()[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_TypeNature (JNIEnv *env, jclass)
{
	jobjectArray result = NULL;
	jarray resulta = NULL;

	// creation du tableau de retour
	jclass objectCls = env->FindClass( "java/lang/Object" );
	if( objectCls == NULL )
		return NULL;
	resulta = env->NewObjectArray( 2, objectCls, NULL );
	if( resulta != NULL )
	{
		result=(jobjectArray) resulta;
		env->SetObjectArrayElement( result, 0, env->NewStringUTF( OUT_FUZZY::OutputType() ) );
		env->SetObjectArrayElement( result, 1, env->NewStringUTF( OUT_CRISP::OutputType() ) );
	}
	env->DeleteLocalRef( objectCls );
	return result;
}



//---------------------------------------------------------------------------

JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_MFType (JNIEnv *env, jclass)

{
	jobjectArray result = NULL;
	jarray resulta = NULL;

	// creation du tableau de retour
	jclass objectCls = env->FindClass( "java/lang/Object" );
	if( objectCls == NULL )
		return NULL;
	resulta = env->NewObjectArray( 12, objectCls, NULL );
	if( resulta != NULL )
	{
		result=(jobjectArray) resulta;
		env->SetObjectArrayElement( result, 0, env->NewStringUTF( MFTRI::Type() ) );
		env->SetObjectArrayElement( result, 1, env->NewStringUTF( MFTRAP::Type() ) );
		env->SetObjectArrayElement( result, 2, env->NewStringUTF( MFTRAPINF::Type() ) );
		env->SetObjectArrayElement( result, 3, env->NewStringUTF( MFTRAPSUP::Type() ) );
		env->SetObjectArrayElement( result, 4, env->NewStringUTF( MFDISCRETE::Type() ) );
		env->SetObjectArrayElement( result, 5, env->NewStringUTF( MFUNIV::Type() ) );
		env->SetObjectArrayElement( result, 6, env->NewStringUTF( MFGAUSS::Type() ) );
		env->SetObjectArrayElement( result, 7, env->NewStringUTF( MFDOOR::Type() ) );
		env->SetObjectArrayElement( result, 8, env->NewStringUTF( MFSINUS::Type() ) );
		env->SetObjectArrayElement( result, 9, env->NewStringUTF( MFSINUSINF::Type() ) );
		env->SetObjectArrayElement( result, 10, env->NewStringUTF( MFSINUSSUP::Type() ) );
		env->SetObjectArrayElement( result, 11, env->NewStringUTF( MFDPOSS::Type() ) );
	}
	env->DeleteLocalRef( objectCls );
	return result;
}



//---------------------------------------------------------------------------

JNIEXPORT jlong JNICALL Java_fis_jnifis_NewMFTriangular (JNIEnv *env, jclass, jstring name , jdouble a, jdouble b, jdouble c)

{
	MF *sef  = NULL;
	const char *sef_name = env->GetStringUTFChars( name, NULL );
	try
	{
		sef = new MFTRI( a, b, c );
		sef->SetName( (char *)sef_name );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( name, sef_name );
	return (jlong)sef;
}

//---------------------------------------------------------------------------

JNIEXPORT jlong JNICALL Java_fis_jnifis_NewMFSinus (JNIEnv *env, jclass, jstring name , jdouble a, jdouble b)

{
	MF *sef  = NULL;
	const char *sef_name = env->GetStringUTFChars( name, NULL );
	try
	{
		sef = new MFSINUS( a, b);
		sef->SetName( (char *)sef_name );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( name, sef_name );
	return (jlong)sef;
}

//---------------------------------------------------------------------------

JNIEXPORT jlong JNICALL Java_fis_jnifis_NewMFSinusInf (JNIEnv *env, jclass, jstring name , jdouble a, jdouble b)

{
	MF *sef  = NULL;
	const char *sef_name = env->GetStringUTFChars( name, NULL );
	try
	{
		sef = new MFSINUSINF( a, b);
		sef->SetName( (char *)sef_name );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( name, sef_name );
	return (jlong)sef;
}

//---------------------------------------------------------------------------

JNIEXPORT jlong JNICALL Java_fis_jnifis_NewMFSinusSup (JNIEnv *env, jclass, jstring name , jdouble a, jdouble b)

{
	MF *sef  = NULL;
	const char *sef_name = env->GetStringUTFChars( name, NULL );
	try
	{
		sef = new MFSINUSSUP( a, b);
		sef->SetName( (char *)sef_name );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( name, sef_name );
	return (jlong)sef;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewMFTrapezoidal (JNIEnv *env, jclass, jstring name, jdouble a, jdouble b, jdouble c, jdouble d)
{
	MF *sef  = NULL;
	const char *sef_name = env->GetStringUTFChars( name, NULL );
	try
	{
		sef = new MFTRAP( a, b, c, d );
		sef->SetName( (char *)sef_name );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( name, sef_name );
	return (jlong)sef;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewMFSemiTrapezoidalInf (JNIEnv *env, jclass, jstring name , jdouble a, jdouble b, jdouble c)
{
	MF *sef  = NULL;
	const char *sef_name = env->GetStringUTFChars( name, NULL );
	try
	{
		sef = new MFTRAPINF( a, b, c );
		sef->SetName( (char *)sef_name );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( name, sef_name );
	return (jlong)sef;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewMFSemiTrapezoidalSup (JNIEnv *env, jclass, jstring name, jdouble a, jdouble b, jdouble c)
{
	MF *sef  = NULL;
	const char *sef_name = env->GetStringUTFChars( name, NULL );
	try
	{
		sef = new MFTRAPSUP( a, b, c );
		sef->SetName( (char *)sef_name );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( name, sef_name );
	return (jlong)sef;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewMFDiscrete (JNIEnv *env, jclass, jstring name, jdoubleArray values_array)
{
	MF *sef  = NULL;
	const char *sef_name = env->GetStringUTFChars( name, NULL );
	try
	{
		int nb_values = env->GetArrayLength( values_array );
		jdouble *jvalues = env->GetDoubleArrayElements( values_array, NULL );
		//double * values = new double[nb_values];
		sef = new MFDISCRETE(nb_values, jvalues);
		sef->SetName( (char *)sef_name );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( name, sef_name );
	return (jlong)sef;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewMFUniversal (JNIEnv *env, jclass, jstring jname, jdouble a, jdouble b)
{
	MF *sef  = NULL;
	const char *name = env->GetStringUTFChars( jname, NULL );
	try
	{
		sef = new MFUNIV(a, b);
		sef->SetName( (char *)name );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( jname, name );
	return (jlong)sef;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewMFDoor (JNIEnv *env, jclass, jstring jname, jdouble a, jdouble b)
{
	MF *sef  = NULL;
	const char *name = env->GetStringUTFChars( jname, NULL );
	try
	{
		sef = new MFDOOR(a, b);
		sef->SetName( (char *)name );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( jname, name );
	return (jlong)sef;
}


//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewMFGaussian (JNIEnv *env, jclass, jstring jname, jdouble std, jdouble mean)
{
	MF *sef  = NULL;
	const char *name = env->GetStringUTFChars( jname, NULL );
	try
	{
		sef = new MFGAUSS(std, mean);
		sef->SetName( (char *)name );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( jname, name );
	return (jlong)sef;
}

//---------------------------------------------------------------------------
JNIEXPORT jintArray JNICALL Java_fis_jnifis_GetRuleProps (JNIEnv *env, jclass, jlong ptr)
{
	jintArray result;
	RULE *regle = (RULE *)ptr;

	result = env->NewIntArray( regle->GetNbProp() );
	if( result != NULL )
		for( int i=0 ; i<regle->GetNbProp() ; i++ )
		{
			int facteur = 0;
			regle->GetAProp( facteur, i );
			jint jfacteur = facteur;
			env->SetIntArrayRegion( result, i, 1, &jfacteur );
		}
	return result;
}

//------------------------------------------------------------------------------
JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_GetRulesWeight (JNIEnv *env, jclass, jlong ptr, jdoubleArray inputs_values)
{
	//return the rule weights from an array of active inputs values. THe values in this
	//array must follow the FIS inputs order

	FIS* sif = (FIS*)ptr; //the fis to be infered
	jdoubleArray rule_weights_result ;
	rule_weights_result = env->NewDoubleArray(sif->GetNbActRule()); //the return value

	jdouble* jfixedinputs_values = env->GetDoubleArrayElements(inputs_values, NULL ); //jdouble* conversion from jdoubleArray
	double* rule_weights = new double[sif->GetNbRule()] ;
	sif->RuleWeights(jfixedinputs_values, rule_weights) ;

	//we keep only active rule weight
	double* act_rule_weights = new double[sif->GetNbActRule()] ;
	int k = 0 ;
	for (int i=0 ; i< sif->GetNbRule() ; i++)
	{
		if (rule_weights[i] == -1)
			continue ;
		else
		{
			act_rule_weights[k] = rule_weights[i] ;
			k++ ;
		}
	}
	env->SetDoubleArrayRegion( rule_weights_result, 0, sif->GetNbActRule(), act_rule_weights );
	return rule_weights_result ;

}
//---------------------------------------------------------------------------
JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_GetRuleConcs (JNIEnv *env, jclass, jlong ptr)
{
	jdoubleArray result;
	RULE *regle = (RULE *)ptr;

	result = env->NewDoubleArray( regle->GetNbConc() );
	if( result != NULL )
		for( int i=0 ; i<regle->GetNbConc() ; i++ )
		{
			jdouble action = regle->GetAConc( i );
			env->SetDoubleArrayRegion( result, i, 1, &action );
		}
	return result;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SetRuleProp (JNIEnv *env, jclass, jlong rule_ptr, jint input_number, jint prop)
{

	RULE *rule = (RULE *)rule_ptr;

	try { rule->SetAProp( prop, input_number ); }

	catch( std::exception &e ) { ExceptionToJava( env, e ); }

}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SetRuleConc (JNIEnv *env, jclass, jlong rule_ptr, jint output_number, jdouble conc)
{

	RULE *rule = (RULE *)rule_ptr;

	try { rule->SetAConc( output_number, conc ); }

	catch( std::exception &e ) { ExceptionToJava( env, e ); }

}

//---------------------------------------------------------------------------
JNIEXPORT jboolean JNICALL Java_fis_jnifis_GetRuleActive (JNIEnv *, jclass, jlong rule_ptr)
{
	RULE *rule = (RULE *)rule_ptr;
	return (jboolean)rule->IsActive();
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SetRuleActive (JNIEnv *, jclass, jlong rule_ptr, jboolean active)
{
	RULE *rule = (RULE *)rule_ptr;
	(active == true) ? rule->Activate() : rule->Deactivate();
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_NewRule (JNIEnv *env, jclass, jlong ptr, jboolean active, jintArray facteurs_array, jdoubleArray actions_array)
{
	FIS *sif = (FIS *)ptr;

	RULE *regle = NULL;
	int *facteurs = NULL;
	double *actions = NULL;
	try
	{
		regle = new RULE();
		// lecture des facteurs
		facteurs = new int[sif->GetNbIn()];
		if( env->GetArrayLength( facteurs_array ) != sif->GetNbIn() )
			throw std::runtime_error("");
		jint *jfacteurs = env->GetIntArrayElements( facteurs_array, NULL );
		if( jfacteurs == NULL )
			throw std::runtime_error("");
		for( int i=0 ; i<sif->GetNbIn() ; i++ )
			facteurs[i] = jfacteurs[i];
		env->ReleaseIntArrayElements( facteurs_array, jfacteurs, 0 );
		// lecture des actions
		actions = new double[sif->GetNbOut()];
		if( env->GetArrayLength( actions_array ) != sif->GetNbOut() )
			throw std::runtime_error("");
		jdouble *jactions = env->GetDoubleArrayElements( actions_array, NULL );
		if( jactions == NULL )
			throw std::runtime_error("");
		for( int i=0 ; i<sif->GetNbOut() ; i++ )
			actions[i] = jactions[i];
		env->ReleaseDoubleArrayElements( actions_array, jactions, 0 );
		// Init de l'objet regle
		regle->SetPremise( sif->GetNbIn(), sif->In, sif->TypeConj() );
		regle->SetConclusion( sif->GetNbOut(), sif->Out );
		regle->SetAProps( facteurs );
		regle->SetConcs( actions );
		if( active == true )
			regle->Activate();
		else
			regle->Deactivate();
		delete [] facteurs;
		delete [] actions;
	}
	catch( std::exception &e )
	{
		delete regle;
		delete [] facteurs;
		delete [] actions;
		ExceptionToJava( env, e );
	}
	return (jlong)regle;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_DeleteRule(JNIEnv *, jclass, jlong rule_ptr)
{
	RULE *rule = (RULE *)rule_ptr;
	delete rule;
}


//---------------------------------------------------------------------------
JNIEXPORT jboolean JNICALL Java_fis_jnifis_RulePremisseEqual(JNIEnv *, jclass, jlong rule1_ptr, jlong rule2_ptr)
{
	RULE *rule1 = (RULE *)rule1_ptr;
	RULE *rule2 = (RULE *)rule2_ptr;
	return rule1->ComparePremises(rule2) == 0 ? true : false;
}

//------------------------------------------------------------------------------
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_Infer3DSurface (JNIEnv *env , jclass, jlong ptr, jobject obj)
{
	//calculate a one-dimensional array of values that can be directly used by the InferenceSurfacePlot java class
	//array size is (number of matrix elements of the surface (X resolution* Y resolution) +2) ; two first additional values are the x and y resolutions of the matrix
	//the matrix elements correspond to different inputs values for X an Y inputs (other are fixed) whose positions must be calulated, knowing (Xmax-Xmin), (Ymax-Ymin) and the X and Y resolutions.
	//for each position in the matrix (the order is the one used by FISSurfacePlot : for one X element from Xmin, go through the corresponding Y elements from Ymin), inference is calculated : it is the Z value, stored in this order in an array to be returned

	FIS* sif = (FIS*)ptr; //the fis to be infered
	jfieldID fid ; //used to retrieve fields from the SurfacePlotParameters java class

	//the surface plot parameters to be retrieved
	double Xmax, Xmin, Ymax, Ymin ;
	int Xresolution, Yresolution ;
	int selected_X_input_number, selected_Y_input_number ;
	jintArray fixedinputs_numbers ;
	jdoubleArray fixedinputs_values ;
	int output_number ;

	//jdoubleArray infersurface_result ; //the return value
	jobjectArray infersurface_result   = NULL; //the return value : two dimensional array for x and y input data for the plot
	jarray infersurface_resulta = NULL;

	try
	{
		//retrieving of the parameters values

		jclass cls = env->GetObjectClass (obj); //obj is the Parameters object

		fid = env ->GetFieldID(cls, "Xmax", "D") ;
		Xmax = env->GetDoubleField(obj, fid) ;

		fid = env ->GetFieldID(cls, "Xmin", "D") ;
		Xmin = env->GetDoubleField(obj, fid) ;

		fid = env ->GetFieldID(cls, "Ymax", "D") ;
		Ymax = env->GetDoubleField(obj, fid) ;

		fid = env ->GetFieldID(cls, "Ymin", "D") ;
		Ymin = env->GetDoubleField(obj, fid) ;

		fid = env ->GetFieldID(cls, "Xresolution", "I") ;
		Xresolution = env->GetIntField(obj, fid) ;

		fid = env ->GetFieldID(cls, "Yresolution", "I") ;
		Yresolution = env->GetIntField(obj, fid) ;

		fid = env ->GetFieldID(cls, "selected_X_input_number", "I") ;
		selected_X_input_number = env->GetIntField(obj, fid) ;

		fid = env ->GetFieldID(cls, "selected_Y_input_number", "I") ;
		selected_Y_input_number = env->GetIntField(obj, fid) ;

		fid = env ->GetFieldID(cls, "fixedinputs_numbers", "[I") ;
		fixedinputs_numbers = (jintArray)env->GetObjectField(obj, fid) ;

		fid = env ->GetFieldID(cls, "fixedinputs_values", "[D") ;
		fixedinputs_values = (jdoubleArray)env->GetObjectField(obj, fid) ;

		fid = env ->GetFieldID(cls, "output_number", "I") ;
		output_number = env->GetIntField(obj, fid) ;

		// the fixedinputs_numbers and fixedinputs_values arrays need to be copied in a native buffer
		jint* jfixedinputs_numbers = env->GetIntArrayElements( fixedinputs_numbers, NULL );
		jdouble* jfixedinputs_values = env->GetDoubleArrayElements( fixedinputs_values, NULL );

		//calcul des positions ds la matrice ; calculate the matrix elements positions
		double Xstep = (Xmax-Xmin)/(Xresolution-1) ;
		double* const Xsteps = new double[Xresolution] ;
		for (int i=0 ;i<Xresolution ; i++)
			Xsteps[i] = Xmin + (i*Xstep) ;

		double Ystep = (Ymax-Ymin)/(Yresolution-1) ;
		double* const Ysteps = new double[Yresolution] ;
		for (int i=0 ;i<Yresolution ; i++)
			Ysteps[i] = Ymin + (i*Ystep) ;

		//calculate the values array to be infered
		jsize jActiveInputsNumber = (env->GetArrayLength (fixedinputs_numbers))+2 ; //add X and Y inputs
		int ActiveInputsNumber = jActiveInputsNumber ;
		double* input_values = new double [ActiveInputsNumber] ; //the indexes in this array must follow the active inputs indexes in the fis

		//the values for the fixed inputs are retrieved and will be fixed
		for (int i=0 ; i<ActiveInputsNumber-2 ; i++)
			input_values[jfixedinputs_numbers[i]] = jfixedinputs_values[i] ;

		int matrix_items_lenght = Xresolution*Yresolution ; //number of points to infer
		double** matrix_items = new double*[matrix_items_lenght] ; //matrix_items' array size = [Xresolution*Yresolution][ActiveInputsNumber] ; we need ActiveNumberInputs values for each Xresolution*Yresolution points to infer

		//create the matrix_items array from the X and Y positions in the surface matrix, given in the Xsteps and Ysteps arrays
		int point = 0 ;
		for (int i=0 ; i<Xresolution ; i++)
		{
			int k=0 ;
			input_values[selected_X_input_number] = Xsteps[i];
			for (k=0 ; k<Yresolution ;k++)
			{
				input_values[selected_Y_input_number] = Ysteps[k]; //at this point, all the inputs values for one position to infer are retrieved in the inputs_values array. They must be copied in the matrix_items array.
				matrix_items[point] = new double [ActiveInputsNumber] ;
				memcpy(matrix_items[point], input_values, ActiveInputsNumber * sizeof(double)); //copy is done
				point++ ;
			}
		}

		//the inactive inputs are removed from the fis after a copy (for efficiency)
		FIS* fiscopy = new FIS(*sif) ;
		int k=0 ;
		for (int i=0 ; i<sif->GetNbIn() ; i++)
		{
			if ((sif->In[i])->IsActive() == 0)
			{
				fiscopy->RemoveInput(i-k) ;
				k++ ;
			}
		}

		jdouble *jdresult = new jdouble[matrix_items_lenght+2]; //to receive the inference results
		jdresult[0] = Xresolution ; jdresult[1] = Yresolution ;  //two first positions must be the matrix dimensions
		//infersurface_result = env->NewDoubleArray(matrix_items_lenght+2); //will receive a copy of jdresult to be sent back to the java code

		//find class for the double array
		//jclass doubleArrayClass = env->FindClass("[D") ;
		jclass objectCls = env->FindClass( "java/lang/Object" );
		if( objectCls != NULL )
		{
			// NewObjectArray returns a jarray
			infersurface_resulta = env->NewObjectArray( 2, objectCls, NULL); //will receive a copy of jdresult to be sent back to the java code
			if( infersurface_resulta != NULL )
			{

				// transform this array into an objectArray
				infersurface_result=(jobjectArray) infersurface_resulta;

				double onepoint_infer_result ; //the inference for one point
				double maximal_infered_val =  -((fiscopy->Out[output_number])->min())  ; //over estimate
				double minimal_nondefault_val = ((fiscopy->Out[output_number])->max())*10000  ; //over estimate
				int* min_tresh_indices = NULL;
				bool default_val_present = false ;
				int default_val_count=0 ;
				for (int i=0 ; i< matrix_items_lenght; i++)
				{
					double* onepoint_inputsvalues = matrix_items[i] ; // the inputs values fot one point to infer
					fiscopy->InferCheck(onepoint_inputsvalues);
					/*    if (fiscopy -> OutValue[output_number] == (fiscopy->Out[output_number])->DefaultValue() ) //we don't want to display the default value
          {
          if (!default_val_present)
             {
             //min_tresh = ((fiscopy->Out[output_number])->min()) - (((sif->Out[output_number])->min())*(5/100)) ;
             min_tresh_indices = new int[matrix_items_lenght] ;
             //onepoint_infer_result = min_tresh ; // we change the inference to : (minimum observed) - 5%
           //  min_tresh_indices[default_val_count++] = i ;
             default_val_present = true ;
             }
       //   else
             min_tresh_indices[default_val_count++] = i ;
          //   onepoint_infer_result = -10 ; // we change the inference to : (minimum observed) - 5%
          }
    else
          {     */
					// min_tresh_indices[i] = false ;
					onepoint_infer_result = fiscopy->OutValue[output_number] ;
					if (onepoint_infer_result < minimal_nondefault_val)
						minimal_nondefault_val = onepoint_infer_result ;
					if (onepoint_infer_result > maximal_infered_val)
						maximal_infered_val = onepoint_infer_result ;
					jdresult[i+2] = onepoint_infer_result ; //+2 because the txo first positions must contain the X and Y resolutions

					//   }
				}

				// env->SetDoubleArrayRegion(infersurface_result, 0, matrix_items_lenght+2, jdresult ); //copy of the native type result in a jdoubleArray

				double min_tresh ;
				//fill the retained indices with min_tresh
				if (default_val_present)
				{
					min_tresh = minimal_nondefault_val-(minimal_nondefault_val*((double)10/(double)100)) ;
					for (int i=0 ; i< default_val_count; i++)
						//jdresult[i+2] = (min_tresh_indices[i]==true) ? min_tresh : jdresult[i+2] ;
						//if (min_tresh_indices[i]==true)
						jdresult[min_tresh_indices[i]+2] = min_tresh ;
				}
				else
					min_tresh = minimal_nondefault_val ;

				//printf("min_tresh %lf \n", min_tresh) ;
				jdoubleArray inferedsurface = env->NewDoubleArray(matrix_items_lenght+2) ;
				//jdoubleArray darr_xdata = env->NewDoubleArray(Xresolution) ;
				env->SetDoubleArrayRegion(inferedsurface, 0, matrix_items_lenght+2, jdresult ); //copy of the native type result in a jdoubleArray

				env->SetObjectArrayElement(infersurface_result, 0, inferedsurface); //copy of the native type result in a jdoubleArray

				/*double min_displayed  = min_tresh ;
min_displayed = (default_val_present) ? min_tresh : minimal_nondefault_val ;*/

				//conversion of minimal_nondefault_val to a relative value between 0-1
				float fminimal_nondefault_val = (float)((minimal_nondefault_val-min_tresh)/(maximal_infered_val-min_tresh)) ;
				// Double default value zone creation
				jclass floatCls = env->FindClass( "java/lang/Float" );
				jmethodID constructorID;
				if( floatCls != NULL )
				{
					constructorID = env->GetMethodID( floatCls, "<init>", "(F)V" );
					if( constructorID != NULL )
					{
						env->SetObjectArrayElement( infersurface_result, 1, env->NewObject( floatCls, constructorID, fminimal_nondefault_val ) );
					}
					env->DeleteLocalRef( floatCls );
				}


				//clean up
				for(int i = 0; i < ActiveInputsNumber; i ++)
					delete []  matrix_items[i];
				delete []  matrix_items;
				delete [] jdresult;
				delete [] Xsteps ;
			}
		}
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }

	//done
	return infersurface_result ;

}

//------------------------------------------------------------------------------
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_Infer2DSurface (JNIEnv *env , jclass, jlong ptr, jobject obj)
{
	//calculate a array of infered Y values to use with DataXYFrame java class
	//array size is Xresolution, the infered values follow the same order than the Xs' one

	FIS* sif = (FIS*)ptr; //the fis to be infered
	jfieldID fid ; //used to retrieve fields from the SurfacePlotParameters java class

	//the surface plot parameters to be retrieved
	double Xmax, Xmin  ;
	int Xresolution ;
	int selected_X_input_number ;
	jintArray fixedinputs_numbers ;
	jdoubleArray fixedinputs_values ;
	int output_number ;

	jobjectArray infersurface_result  = NULL; //the return value : two dimensional array for x and y input data for the plot
	jarray infersurface_resulta; //the return value : two dimensional array for x and y input data for the plot
	try
	{

		//retrieving of the parameters values

		jclass cls = env->GetObjectClass (obj); //obj is the Parameters object

		fid = env ->GetFieldID(cls, "Xmax", "D") ;
		Xmax = env->GetDoubleField(obj, fid) ;

		fid = env ->GetFieldID(cls, "Xmin", "D") ;
		Xmin = env->GetDoubleField(obj, fid) ;

		fid = env ->GetFieldID(cls, "Xresolution", "I") ;
		Xresolution = env->GetIntField(obj, fid) ;

		fid = env ->GetFieldID(cls, "selected_X_input_number", "I") ;
		selected_X_input_number = env->GetIntField(obj, fid) ;

		fid = env ->GetFieldID(cls, "fixedinputs_numbers", "[I") ;
		fixedinputs_numbers = (jintArray)env->GetObjectField(obj, fid) ;

		fid = env ->GetFieldID(cls, "fixedinputs_values", "[D") ;
		fixedinputs_values = (jdoubleArray)env->GetObjectField(obj, fid) ;

		fid = env ->GetFieldID(cls, "output_number", "I") ;
		output_number = env->GetIntField(obj, fid) ;

		// the fixedinputs_numbers and fixedinputs_values arrays need to be copied in a native buffer
		jint* jfixedinputs_numbers = env->GetIntArrayElements( fixedinputs_numbers, NULL );
		jdouble* jfixedinputs_values = env->GetDoubleArrayElements( fixedinputs_values, NULL );

		//calcul des positions ds l'intervalle Xmin-Xmax ; calculate the positions to infer in the Xmin-Xmax range
		double Xstep = (Xmax-Xmin)/(Xresolution-1) ;
		double* const Xsteps = new double[Xresolution] ;
		for (int i=0 ;i<Xresolution ; i++)
			Xsteps[i] = Xmin + (i*Xstep) ;

		//calculate the values array to be infered
		jsize jActiveInputsNumber = (env->GetArrayLength (fixedinputs_numbers))+1 ; //add X input
		int ActiveInputsNumber = jActiveInputsNumber ;
		double* input_values = new double [ActiveInputsNumber] ; //the indexes in this array must follow the active inputs indexes in the fis

		//the values for the fixed inputs are retrieved and will be fixed
		for (int i=0 ; i<ActiveInputsNumber-1 ; i++)
			input_values[jfixedinputs_numbers[i]] = jfixedinputs_values[i] ;

		//int matrix_items_lenght = Xresolution ; //number of points to infer
		double** matrix_items = new double*[/*matrix_items_lenght*/Xresolution] ; //matrix_items' size = [Xresolution][ActiveInputsNumber] ; we need ActiveNumberInputs values for each Xresolution points to infer

		//create the matrix_items array from the X and Y positions in the surface matrix, given in the Xsteps and Ysteps arrays
		for (int i=0 ; i<Xresolution ; i++)
		{
			input_values[selected_X_input_number] = Xsteps[i];
			matrix_items[i] = new double [ActiveInputsNumber] ;
			memcpy(matrix_items[i], input_values, ActiveInputsNumber * sizeof(double)); //copy is done
		}

		//the inactive inputs are removed from the fis after a copy (for efficiency)
		FIS* fiscopy = new FIS(*sif) ;
		int k=0 ;
		for (int i=0 ; i<sif->GetNbIn() ; i++)
		{
			if ((sif->In[i])->IsActive() == 0)
			{
				fiscopy->RemoveInput(i-k) ;
				k++ ;
			}
		}

		jdouble *jdresult = new jdouble[Xresolution]; //to receive the inference results

		//find class for the double array
		jclass doubleArrayClass = env->FindClass("[D") ;

		infersurface_resulta = env->NewObjectArray( 2, doubleArrayClass, NULL); //will receive a copy of jdresult to be sent back to the java code
		infersurface_result=(jobjectArray) infersurface_resulta; //produces a jobjectArray

		double onepoint_infer_result ;
		for (int i=0 ; i< Xresolution; i++)
		{
			double* onepoint_inputsvalues = matrix_items[i] ; // the inputs values fot one point to infer
			fiscopy->InferCheck(onepoint_inputsvalues);
			/*if (fiscopy -> OutValue[output_number] == (fiscopy->Out[output_number])->DefaultValue() ) //we don't want to display the default value
      onepoint_infer_result = ((fiscopy->Out[output_number])->min()) - (((fiscopy->Out[output_number])->min())*(5/100)) ; // we change the inference to : minimum - 5%
     else        */
			onepoint_infer_result = fiscopy->OutValue[output_number] ;

			jdresult[i] = onepoint_infer_result ;
		}

		jdoubleArray darr_xdata = env->NewDoubleArray(Xresolution) ;
		env->SetDoubleArrayRegion(darr_xdata, 0, Xresolution, Xsteps ); //copy of the native type result in a jdoubleArray
		jdoubleArray darr_ydata = env->NewDoubleArray(Xresolution) ;
		env->SetDoubleArrayRegion(darr_ydata, 0, Xresolution, jdresult ); //copy of the native type result in a jdoubleArray

		env->SetObjectArrayElement(infersurface_result, 0, darr_xdata); //copy of the native type result in a jdoubleArray
		env->SetObjectArrayElement(infersurface_result, 1, darr_ydata); //copy of the native type result in a jdoubleArray


		//clean up
		for(int i = 0; i < ActiveInputsNumber; i ++)
			delete []  matrix_items[i];
		delete []  matrix_items;
		delete [] jdresult;
		delete [] Xsteps ;

	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }

	//done
	return infersurface_result ;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_InferFuzzyInput (JNIEnv *env, jclass, jlong ptr, jdoubleArray kw, jdoubleArray sw, jint nbAlphaCut,jdoubleArray inputs_values)
{
	FIS *sif = (FIS *)ptr;
	MFDPOSS ** fuzval=NULL;
	int output_number = 0;
	double *sw_val = NULL;
	double *values = NULL;
	double *kw_val = NULL;

	try
	{
		// 0 - lecture des valeurs des sw et kw et values
		kw_val = new double[sif->GetNbIn()];
		sw_val = new double[sif->GetNbIn()];
		values = new double[sif->GetNbIn()];
		jdouble *jkw = env->GetDoubleArrayElements( kw, NULL );
		jdouble *jsw = env->GetDoubleArrayElements( sw, NULL );
		jdouble *jvalues = env->GetDoubleArrayElements( inputs_values, NULL );
		for( int i=0 ; i < sif->GetNbIn() ; i++ )
		{
			kw_val[i] = (double)jkw[i];
			sw_val[i] = (double)jsw[i];
			values[i] = (double)jvalues[i];
		}
		env->ReleaseDoubleArrayElements( kw, jkw, 0 );
		env->ReleaseDoubleArrayElements( sw, jsw, 0 );
		env->ReleaseDoubleArrayElements( inputs_values, jvalues, 0 );

		// 1 -BuildFuzIn
		sif->BuildFuzIn(values, kw_val, sw_val, fuzval);

		if(!strcmp(sif->Out[output_number]->Defuzzify(), OUT_FUZZY::ImpFuzzyDefuz()))
		{
			sif->InferFatiPrep(output_number);
			sif->InferFatiCheck(fuzval, output_number, nbAlphaCut);
		}
		else // conjonctif
			sif->InferCheck((MF**)fuzval);
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	delete [] values;
	delete [] kw_val;
	delete [] sw_val;
	if (fuzval !=NULL)
	  for(int i = 0; i < sif->GetNbIn(); i++)
	    delete fuzval[i];
	delete [] fuzval;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_Infer(JNIEnv *env, jclass, jlong ptr, jdoubleArray inputs_values)
{
	FIS *sif = (FIS *)ptr;
	double *values = NULL;

	try
	{
		/*if( env->GetArrayLength( inputs_values ) != sif->GetNbIn() )
	throw std::runtime_error("Array length != sif->NbEntrees()");*/

		// lecture des valeurs des entrees
		values = new double[sif->GetNbIn()];
		jdouble *jvalues = env->GetDoubleArrayElements( inputs_values, NULL );
		/*if( jvalues == NULL )
	throw std::runtime_error("jvalues == NULL");    */
		for( int i=0 ; i<sif->GetNbIn() ; i++ )
			values[i] = (double)jvalues[i];
		env->ReleaseDoubleArrayElements( inputs_values, jvalues, 0 );

		// calcul de l'inference - first set classes to rule conclusions (in classif case)
		sif->InferCheck( values );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	delete [] values;
}


//---------------------------------------------------------------------------
JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_DataFileInfer(JNIEnv *env, jclass, jlong fis_ptr, jint output_number, jstring data_file, jstring result_file, jdouble blank_thres, jboolean display)
{
	jdoubleArray infer_result = NULL;
	FIS *fis = (FIS *)fis_ptr;
	double perf;
	int rep;
	char *data = get_native_string(env, data_file);
	char *result = get_native_string(env, result_file);
	FILE *new_stdout = NULL;
	try
	{
		double coveradge=0, max_err=0;
		// redirection de stdout
		if( display == true )
			new_stdout = freopen("infer.out", "wt", stdout);
		// calcul de la performance - verifier la coherence du sif
		rep = fis -> CheckConsistency();
		perf = fis->Performance( output_number, data, coveradge, max_err, blank_thres, result, display );
		// fermeture de la redirection de stdout
		if( display == true )
			fclose(new_stdout);

		// resultats de performance
		infer_result = env->NewDoubleArray( 3 );
		if( result != NULL )
		{
			jdouble temp[3];
			temp[0] = perf;
			temp[1] = coveradge;
			temp[2] = max_err;
			env->SetDoubleArrayRegion( infer_result, 0, 3, temp );
		}
			}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	release_native_string(data);
	release_native_string(result);
	return infer_result;
}

//---------------------------------------------------------------------------

JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_InferErrorRegression (JNIEnv *env, jclass, jlong fis_ptr) {

	jdoubleArray error = NULL;
	FIS *fis = (FIS *)fis_ptr;
	error = env->NewDoubleArray( 3 );
	jdouble temp[3];
	temp[0] = fis->PIn;
	temp[1] = fis->RMSE;
	temp[2] = fis->MAE;
	env->SetDoubleArrayRegion( error, 0, 3, temp );
	return error;
}

//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    SortiesObtenues
 * Signature: (J)[D
 */
JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_SortiesObtenues (JNIEnv *env, jclass, jlong ptr)
{
	jdoubleArray result = NULL;
	FIS *sif = (FIS *)ptr;
	jdouble *jsorties = NULL;

	try
	{
		result = env->NewDoubleArray( sif->GetNbOut() );
		jsorties = new jdouble[sif->GetNbOut()];
		if( result != NULL )
			for( int i=0 ; i<sif->GetNbOut() ; i++ )
				jsorties[i] = sif->OutValue[i];
		env->SetDoubleArrayRegion( result, 0, sif->GetNbOut(), jsorties );
		delete [] jsorties;
	}
	catch( std::exception &e )
	{
		delete [] jsorties;
		ExceptionToJava( env, e );
	}
	return result;
}

//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    InputAppartenance
 * Signature: (J)[D
 */
JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_InputAppartenance (JNIEnv *env, jclass, jlong ptr)
{
	FISIN *entree = (FISIN *)ptr;
	jdoubleArray result = NULL;
	jdouble *jappart = NULL;

	try
	{
		result = env->NewDoubleArray( entree->GetNbMf() );
		jappart = new jdouble[entree->GetNbMf()];
		if( result != NULL )
			for( int i=0 ; i<entree->GetNbMf() ; i++ )
				jappart[i] = entree->Mfdeg()[i];
		env->SetDoubleArrayRegion( result, 0, entree->GetNbMf(), jappart );
		delete [] jappart;
	}
	catch( std::exception &e )
	{
		delete [] jappart;
		ExceptionToJava( env, e );
	}
	return result;
}

//------------------------------------------------------------------------------
JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_GetBreakPoints (JNIEnv *env, jclass, jlong ptr)

{
	int  NbBp;
	double* Bp=NULL;
	jdoubleArray breakpoints_result;

	FISIN* entree = (FISIN*)ptr;
	entree->GetBreakPoints(Bp, NbBp);
	breakpoints_result = env->NewDoubleArray(NbBp);
	jdouble *temp = new jdouble[NbBp];
	for (int i=0 ; i<NbBp ; i++)
		temp[i] = Bp[i] ;
	delete [] Bp;
	env->SetDoubleArrayRegion( breakpoints_result, 0, NbBp, temp );
	delete [] temp;
	return breakpoints_result ;

}
//------------------------------------------------------------------------------
JNIEXPORT jboolean JNICALL Java_fis_jnifis_IsSfp (JNIEnv *env, jclass, jlong ptr)
{
	FISIN* entree = (FISIN*)ptr;
	int *sorted = NULL;
	return (jboolean)entree->IsSfp(sorted);
}

//------------------------------------------------------------------------------
JNIEXPORT jboolean JNICALL Java_fis_jnifis_IsQsp (JNIEnv *env, jclass, jlong ptr)
{
	OUT_FUZZY* sortie = (OUT_FUZZY*)ptr;
	return (jboolean)sortie->IsQsp();
}

//---------------------------------------------------------------------------
JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_MFSupport (JNIEnv *env, jclass, jlong ptr)

{
	jdoubleArray result = NULL;
	MF *sef = (MF *)ptr;

	try
	{
		result = env->NewDoubleArray( 2 );
		jdouble jvalues[2];
		if( result != NULL )
			sef->Support( jvalues[0], jvalues[1] );
		env->SetDoubleArrayRegion( result, 0, 2, jvalues );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	return result;
}

//---------------------------------------------------------------------------
JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_MFAlphaKernel (JNIEnv *env, jclass, jlong ptr, jdouble alpha)
{
	jdoubleArray result = NULL;
	MF *sef = (MF *)ptr;

	try
	{
		result = env->NewDoubleArray( 2 );
		jdouble jvalues[2];
		if( result != NULL )
			sef->AlphaKernel( jvalues[0], jvalues[1], alpha );
		env->SetDoubleArrayRegion( result, 0, 2, jvalues );
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	return result;
}



//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    RulePoids
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_fis_jnifis_RulePoids (JNIEnv *, jclass, jlong ptr)
{
	RULE *regle = (RULE *)ptr;
	return regle->Weight;
}

//---------------------------------------------------------------------------
/*
 * Class:     fis_JNIFIS
 * Method:    TrapezeNoyau
 * Signature: (JI)[D
 */
/*JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_TrapezeNoyau (JNIEnv *env, jclass, jlong ptr, jint output_number)
  {
  jdoubleArray result = NULL;
  RULE *regle = (RULE *)ptr;

  try
  {
  result = env->NewDoubleArray( 4 );
  jdouble jvalues[4];
  if( result != NULL )
  {
  jvalues[0] = regle->TrapCoupes[output_number].nd;
  jvalues[1] = regle->TrapCoupes[output_number].nf;
  jvalues[2] = regle->TrapCoupes[output_number].sd;
  jvalues[3] = regle->TrapCoupes[output_number].sf;
  }
  env->SetDoubleArrayRegion( result, 0, 4, jvalues );
  }
  catch( std::exception &e ) { ExceptionToJava( env, e ); }
  return result;
  }         */


//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    ValAction
 * Signature: (JI)D
 */
/*JNIEXPORT jdouble JNICALL Java_fis_jnifis_ValAction (JNIEnv *, jclass, jlong ptr, jint output_number)
{
RULE *regle = (RULE *)ptr;
//return regle->ValActions[output_number];
return 0;
}        */



//---------------------------------------------------------------------------

/*

 * Class:     fis_JNIFIS
 * Method:    DataFile
 * Signature: (Ljava/lang/String;)[[D
 */
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_DataFile (JNIEnv *env, jclass, jstring file_name)
{
	jobjectArray result = NULL;
	jarray resulta=NULL;
	const char *name = get_native_string(env, file_name);
	int colonnes, lignes;
	double ** data = NULL;
	try
	{
		data = ReadSampleFile((char *)name, colonnes, lignes );

		// creation du tableau
		jclass double_class = env->FindClass( "[D" );
		if( double_class != NULL )
		{
			resulta = env->NewObjectArray( lignes, double_class, NULL );
			if( resulta != NULL )
			{
				result=(jobjectArray) resulta;
				for( int i=0 ; i<lignes ; i++ )
				{
					jdoubleArray da = env->NewDoubleArray( colonnes );
					env->SetDoubleArrayRegion( da, 0, colonnes, data[i] );
					env->SetObjectArrayElement( result, i, da );
					env->DeleteLocalRef( da );
				}
			}
		}
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	release_native_string(name);
	if( data != NULL ) // data est NULL si une exception a lieu dans LectureFichierExemples
	{
		for( int i=0 ; i<lignes ; i++ )
			delete [] data[i];
		delete [] data;
	}
	return result;
}

//---------------------------------------------------------------------------
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_HFPHierarchyType (JNIEnv *env, jclass)
{
	jobjectArray result  = NULL;
	jarray resulta  = NULL;

	// creation du tableau de retour
	jclass objectCls = env->FindClass( "java/lang/Object" );
	if( objectCls == NULL )
		return NULL;
	resulta = env->NewObjectArray( 3, objectCls, NULL );
	if( resulta != NULL )
	{
		result=(jobjectArray) resulta;
		env->SetObjectArrayElement( result, 0, env->NewStringUTF( FISHFP::HierarchyHFP() ) );
		env->SetObjectArrayElement( result, 1, env->NewStringUTF( FISHFP::HierarchyKMeans() ) );
		env->SetObjectArrayElement( result, 2, env->NewStringUTF( FISHFP::HierarchyRegular() ) );
	}
	env->DeleteLocalRef( objectCls );
	return result;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_HFPConfig (JNIEnv *env, jclass, jstring jdata_file, jint ns, jstring jhfp_file)
{
	FISHFP *hfp = NULL;
	char *data_file = get_native_string(env, jdata_file);
	char *hfp_file = get_native_string(env, jhfp_file);
	try
	{
		hfp = new FISHFP(data_file, ns, hfp_file);
	}
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	release_native_string(data_file);
	release_native_string(hfp_file);
	return (jlong)hfp;
}

//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_GetHFPHierarchy (JNIEnv *env, jclass, jlong hfp_ptr)
{

	FISHFP *hfp = (FISHFP *)hfp_ptr;

	return env->NewStringUTF( hfp->GetHierarchy() );
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SetHFPHierarchy (JNIEnv *env, jclass, jlong hfp_ptr, jstring jhierarchy)
{
	FISHFP *hfp = (FISHFP *)hfp_ptr;
	char *hierarchy = (char *)env->GetStringUTFChars( jhierarchy, NULL );
	try { hfp->SetHierarchy(hierarchy); }
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( jhierarchy, hierarchy );
}

//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_GetHFPConjunction (JNIEnv *env , jclass, jlong hfp_ptr)
{

	FISHFP *hfp = (FISHFP *)hfp_ptr;

	return env->NewStringUTF( hfp->GetConjunction() );
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SetHFPConjunction (JNIEnv *env, jclass, jlong hfp_ptr, jstring jconjunction)
{
	FISHFP *hfp = (FISHFP *)hfp_ptr;
	char *conjunction = (char *)env->GetStringUTFChars( jconjunction, NULL );
	try { hfp->SetConjunction(conjunction); }
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( jconjunction, conjunction );
}

//---------------------------------------------------------------------------
JNIEXPORT jdouble JNICALL Java_fis_jnifis_GetHFPToleranceThreshold (JNIEnv *env, jclass, jlong hfp_ptr)
{
	FISHFP *hfp = (FISHFP *)hfp_ptr;
	return hfp->GetToleranceThres();
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SetHFPToleranceThreshold (JNIEnv *env, jclass, jlong hfp_ptr, jdouble tolerance)
{

	FISHFP *hfp = (FISHFP *)hfp_ptr;

	hfp->SetToleranceThres(tolerance);
}

//---------------------------------------------------------------------------
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_HFPDistanceType (JNIEnv *env, jclass)
{
	jobjectArray result = NULL;
	jarray resulta = NULL;

	// creation du tableau de retour
	jclass objectCls = env->FindClass( "java/lang/Object" );
	if( objectCls == NULL )
		return NULL;
	resulta = env->NewObjectArray( 3, objectCls, NULL );
	if( resulta != NULL )
	{
		result=(jobjectArray) resulta;
		env->SetObjectArrayElement( result, 0, env->NewStringUTF( FISHFP::DistanceNumerical() ) );
		env->SetObjectArrayElement( result, 1, env->NewStringUTF( FISHFP::DistanceSymbolic() ) );
		env->SetObjectArrayElement( result, 2, env->NewStringUTF( FISHFP::DistanceSymbNum() ) );
	}
	env->DeleteLocalRef( objectCls );
	return result;
}

//---------------------------------------------------------------------------
JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_GetHFPParameters (JNIEnv *env, jclass, jlong hfp_ptr)
{

	FISHFP *hfp = (FISHFP *)hfp_ptr;

	jobjectArray result = NULL;
	jarray resulta;

	// creation du tableau de retour
	jclass objectCls = env->FindClass( "java/lang/Object" );
	if( objectCls == NULL )
		return NULL;
	resulta = env->NewObjectArray( 1, objectCls, NULL );
	if( resulta != NULL )
	{
		result=(jobjectArray) resulta;
		env->SetObjectArrayElement( result, 0, env->NewStringUTF( hfp->GetDistanceType() ) );
	}
	env->DeleteLocalRef( objectCls );
	return result;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SetHFPParameters (JNIEnv *env, jclass, jlong hfp_ptr, jstring jdistance)
{
	FISHFP *hfp = (FISHFP *)hfp_ptr;
	char *distance = (char *)env->GetStringUTFChars( jdistance, NULL );
	try
	{
		hfp->SetDistanceType(distance);
	}
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( jdistance, distance );
}

//---------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_fis_jnifis_GetHFPNbInput (JNIEnv *, jclass, jlong hfp_ptr)
{
	FISHFP *hfp = (FISHFP *)hfp_ptr;
	return hfp->GetNbInput();
}

//---------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_fis_jnifis_GetHFPNbOutput (JNIEnv *, jclass, jlong hfp_ptr)
{

	FISHFP *hfp = (FISHFP *)hfp_ptr;

	return hfp->GetNbOutput();

}


//---------------------------------------------------------------------------

JNIEXPORT jlong JNICALL Java_fis_jnifis_GetHFPInput (JNIEnv *, jclass, jlong hfp_ptr, jint input_number)

{

	FISHFP *hfp = (FISHFP *)hfp_ptr;

	return (jlong)hfp->GetInput(input_number);

}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_GetHFPOutput (JNIEnv *, jclass, jlong hfp_ptr, jint output_number)
{
	FISHFP *hfp = (FISHFP *)hfp_ptr;
	return (jlong)hfp->GetOutput(output_number);
}


//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_HFPDelete (JNIEnv *, jclass, jlong hfp_ptr)
{
	FISHFP *hfp = (FISHFP *)hfp_ptr;
	delete hfp;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_HFPSave (JNIEnv *env, jclass, jlong hfp_ptr, jstring jdata_file, jstring jhfp_file)
{
	FISHFP *hfp = (FISHFP *)hfp_ptr;
	char *data_file = get_native_string(env, jdata_file);
	char *hfp_file = get_native_string(env, jhfp_file);
	try { hfp->PrintCfgHfp(data_file, hfp_file); }
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	release_native_string(data_file);
	release_native_string(hfp_file);
}

//---------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_fis_jnifis_GetHFPInputNumberOfMF (JNIEnv *, jclass, jlong hfp_input_ptr)
{
	INHFP * input = (INHFP *)hfp_input_ptr;
	return input->GetNumberOfMF();
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SetHFPInputNumberOfMF (JNIEnv *, jclass, jlong hfp_input_ptr, jint number)
{
	INHFP * input = (INHFP *)hfp_input_ptr;
	input->SetNumberOfMF( number );
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_HFPVertex (JNIEnv *env, jclass, jstring jdata_file, jstring jhfp_file, jstring jvertex_file)
{

	FISHFP *hfp = NULL;

	char *data_file = get_native_string(env, jdata_file);
	char *hfp_file = get_native_string(env, jhfp_file);
	char *vertex_file = get_native_string(env, jvertex_file);
	try
	{
		hfp = new FISHFP(data_file, hfp_file);
		hfp->Hierarchy(vertex_file);
		delete hfp;
	}
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	release_native_string(data_file);
	release_native_string(hfp_file);
	release_native_string(vertex_file);
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_HFPOpen (JNIEnv *env, jclass, jstring jdata_file, jstring jhfp_file)
{
	FISHFP *hfp = NULL;
	char *data_file = get_native_string(env, jdata_file);
	char *hfp_file = get_native_string(env, jhfp_file);
	try { hfp = new FISHFP(data_file, hfp_file); }
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	release_native_string(data_file);
	release_native_string(hfp_file);
	return (jlong)hfp;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_HFPSelect (JNIEnv *env, jclass, jstring jdata_file, jstring jhfp_file, jboolean ruleInductionMethodFPA, jdouble min_cumul_weight, jint cardinality_strategy, jdouble min_matching_degree, jint min_cardinality, jstring jresult_file, jdouble loss_coverage, jint initial_number_of_mf, jint max_number_of_iter, jstring jvertex_file, jint output_number, jstring jvalidation_file)
{

	FISHFP *hfp = NULL;

	char *data_file = get_native_string(env, jdata_file);
	char *hfp_file = get_native_string(env, jhfp_file);
	char *result_file = get_native_string(env, jresult_file);
	char *vertex_file = get_native_string(env, jvertex_file);
	char *validation_file = get_native_string(env, jvalidation_file);
	try
	{
		hfp = new FISHFP(hfp_file, data_file, (char *)NULL);

		if( ruleInductionMethodFPA )
			hfp->SetRuleInductionMethodFpa();
		else
			hfp->SetRuleInductionMethodWm();
		hfp->SetCumWeThres(min_cumul_weight);
		hfp->SetCardStrategy(cardinality_strategy);
		hfp->SetMuMin(min_matching_degree);
		hfp->SetMinCard(min_cardinality);
		hfp->SetConfigFile(result_file);
		hfp->SetBlankThres(loss_coverage);
		hfp->SetNmfInit(initial_number_of_mf);
		hfp->SetIterMax(max_number_of_iter);
		hfp->SetVertexFile(vertex_file);
		hfp->SetNSort(output_number);
		hfp->SetValidationFile(validation_file);
		hfp->SelectFis();
	}
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	delete hfp;
	release_native_string(data_file);
	release_native_string(hfp_file);
	release_native_string(result_file);
	release_native_string(vertex_file);
	release_native_string(validation_file);
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_HFPFIS   (JNIEnv *env, jclass, jstring jdata_file, jstring jhfp_file, jboolean ruleInductionMethodFPA, jdouble min_cumul_weight, jint cardinality_strategy, jdouble min_matching_degree, jint min_cardinality, jint output_number, jstring jvertex_file)
//JNIEXPORT void JNICALL Java_fis_jnifis_HFPFIS    (JNIEnv *env, jclass, jstring jdata_file, jstring jhfp_file, jdouble min_cumul_weight, jint cardinality_strategy, jdouble min_matching_degree, jint min_cardinality, jstring jresult_file, jint output_number, jstring jvertex_file)
{
	FISHFP *hfp = NULL;
	FIS *fis = NULL;
	char *new_name = NULL;
	char *tmp = NULL;
	char *data_file = get_native_string(env, jdata_file);
	char *hfp_file = get_native_string(env, jhfp_file);
	//char *result_file = (char *)env->GetStringUTFChars( jresult_file, NULL );
	char *vertex_file = get_native_string(env, jvertex_file);
	try
	{
		tmp = TempFileName();

		hfp = new FISHFP(hfp_file, data_file, tmp);
		if( ruleInductionMethodFPA )
			hfp->SetRuleInductionMethodFpa();
		else
			hfp->SetRuleInductionMethodWm();
		hfp->SetCumWeThres(min_cumul_weight);
		hfp->SetCardStrategy(cardinality_strategy);
		hfp->SetMuMin(min_matching_degree);
		hfp->SetMinCard(min_cardinality);
		hfp->SetVertexFile(vertex_file);
		hfp->SetNSort(output_number);
		hfp->GenereCfgFis(true);
		fis = new FIS(tmp);
		new_name = new char[strlen(fis->Name)+10];
		sprintf( new_name, "%s.gen", fis->Name);
		fis->SetName(new_name);
	}
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	delete hfp;
	delete new_name;
	if( tmp != NULL ) remove(tmp);
	delete [] tmp;
	release_native_string(data_file);
	release_native_string(hfp_file);
	//env->ReleaseStringUTFChars( jresult_file, result_file );
	release_native_string(vertex_file);
	return (jlong)fis;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_ReplaceHFPOutput (JNIEnv *env, jclass, jlong hfp_ptr, jint output_number, jlong output_ptr)
{
	FISHFP *hfp = (FISHFP *)hfp_ptr;
	FISOUT *output = (FISOUT *)output_ptr;
	try { hfp->ReplaceOutput( output_number, output); }
	catch(std::exception &e) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_Fuz2Crisp (JNIEnv *env, jclass, jlong fis_ptr, jint output_number)
{
	FIS *fis = (FIS *)fis_ptr;
	try { fis->Fuz2Crisp(output_number); }
	catch(std::exception &e) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_Crisp2Fuz (JNIEnv *env, jclass, jlong fis_ptr, jstring jdefuz, jint output_number)
{
	FIS *fis = (FIS *)fis_ptr;
	try { fis->Crisp2Fuz(output_number, (char *)OUT_FUZZY::AreaDefuz()); }
	catch(std::exception &e) { ExceptionToJava( env, e ); }
}

//---------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_fis_jnifis_Conj2Imp (JNIEnv *env, jclass, jlong fis_ptr, jint output_number, jboolean transfPart)
{
	FIS *fis = (FIS *)fis_ptr;
	int error = -1;
	try { error = fis->Conj2Imp(output_number, NULL, (bool)transfPart); }
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	return error;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_Imp2Conj (JNIEnv *env, jclass, jlong fis_ptr, jint output_number, jboolean transfPart)
{
	FIS *fis = (FIS *)fis_ptr;
	int error = -1;
	try { error = fis->Imp2Conj(output_number, NULL, NULL, (bool)transfPart); }
	catch(std::exception &e) { ExceptionToJava( env, e ); }
}


//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SetHFPFIS (JNIEnv *env, jclass, jstring jhfpin, jstring jconf, jstring jhfpout)
{
	char *hfpin = (char *)env->GetStringUTFChars( jhfpin, NULL );
	char *conf  = (char *)env->GetStringUTFChars( jconf, NULL );
	char *hfpout = (char *)env->GetStringUTFChars( jhfpout, NULL );
	try { SetHfpFis(hfpin, conf, hfpout); }
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	env->ReleaseStringUTFChars( jhfpin, hfpin );
	env->ReleaseStringUTFChars( jconf, conf );
	env->ReleaseStringUTFChars( jhfpout, hfpout );
}



//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SetDataFilePath (JNIEnv *env, jclass, jstring jnewpath)
{
	char *newpath = get_native_string(env, jnewpath);
	strncpy(DataFilePath, newpath, DATAFILEMAXPATH-1);
	DataFilePath[DATAFILEMAXPATH-1] = '\0';
	release_native_string(newpath);
}


//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_WangMendel (JNIEnv *env, jclass, jstring jfis, jstring jdata_file, jstring jfisout)
{
	char *fis = get_native_string(env, jfis);
	char *data_file  = get_native_string(env, jdata_file);
	char *fisout = get_native_string(env, jfisout);
	FISWM *S = NULL;
	FIS *new_fis = NULL;
	try
	{
		S = new FISWM(fis, data_file, fisout);
		S->wm();
		new_fis = new FIS(fisout);
	}
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	delete S;
	release_native_string(fis);
	release_native_string(data_file);
	release_native_string(fisout);
	return (jlong)new_fis;
}

//---------------------------------------------------------------------------
JNIEXPORT jlong JNICALL Java_fis_jnifis_HFPSR (JNIEnv *env , jclass, jstring jdata_file, jstring jinputs_mfs, jint input_hierarchy, jdouble input_tol, jint joutput_mfs, jint joutput_hierarchy, jstring joutput_defuz, jstring joutput_disj, jdouble output_tol, jstring jconjunction, jboolean classif)
{
	char *data_file = get_native_string(env, jdata_file);
	char *inputs_mfs  = (char *)env->GetStringUTFChars( jinputs_mfs, NULL );
	char *output_defuz = (char *)env->GetStringUTFChars( joutput_defuz, NULL );
	char *output_disj = (char *)env->GetStringUTFChars( joutput_disj, NULL );
	char *conjunction = (char *)env->GetStringUTFChars( jconjunction, NULL );
	FIS *new_fis = NULL;
	FISHFP *S = NULL;
	char *ch2 = NULL;
	char *ch1 = NULL;
	char *ch3 = NULL;
	char *temp_fis = NULL;
	try
	{
		double t = input_tol;
		if( t < 0. ) t = 0.01;

		ch1 = TempFileName();
		S = new FISHFP( data_file, 1, ch1, input_hierarchy );
		S->SetToleranceThres(t);
		S->SetConjunction(conjunction);
		S->PrintCfgHfp( data_file, ch1 );
		delete S; S = NULL;

		ch3 = TempFileName();
		SetHfpFis( ch1, inputs_mfs, ch3 );

		// hfpvertex
		S = new FISHFP( data_file, ch3 );
		ch2 = TempFileName();
		S->Hierarchy(ch2);
		delete S; S = NULL;

		t = output_tol;
		if( t < 0. ) t = 0.01;

		temp_fis = TempFileName();
		S = new FISHFP( ch3, data_file, temp_fis );
		S->SetVertexFile(ch2);
		S->GenereCfgFisSr( joutput_mfs, joutput_hierarchy, output_defuz, output_disj, t, classif );
		delete S; S = NULL;

		// open FIS
		new_fis = new FIS(temp_fis);
	}
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	if(ch2 != NULL) remove(ch2);
	delete [] ch2;
	if(ch1 != NULL) remove(ch1);
	delete [] ch1;
	if(ch3 != NULL) remove(ch3);
	delete [] ch3;
	if(temp_fis != NULL) remove(temp_fis);
	delete [] temp_fis;
	delete S;
	release_native_string(data_file);
	env->ReleaseStringUTFChars( jinputs_mfs, inputs_mfs );
	env->ReleaseStringUTFChars( joutput_defuz, output_defuz );
	env->ReleaseStringUTFChars( joutput_disj, output_disj );
	env->ReleaseStringUTFChars( jconjunction, conjunction );
	return (jlong)new_fis;
}

//---------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_fis_jnifis_GetAlarm(JNIEnv *, jclass, jlong output_ptr)
{
	FISOUT *output = (FISOUT *)output_ptr;
	return output->Def->Alarm;
}

//---------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_fis_jnifis_SetDefuzThres(JNIEnv *, jclass, jlong output_ptr, jdouble thres)
{
	FISOUT *output = (FISOUT *)output_ptr;
	output->Def->SetThres(thres);
}

//---------------------------------------------------------------------------
JNIEXPORT jdouble JNICALL Java_fis_jnifis_GetDefuzThres(JNIEnv *, jclass, jlong output_ptr)
{
	FISOUT *output = (FISOUT *)output_ptr;
	return output->Def->GetThres();
}

//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_PerfHeaderObserved (JNIEnv * env, jclass) { return env->NewStringUTF( OBSERVED ); } /**< Entete de la colonne observe du fichier performance @return String entete observe */
//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_PerfHeaderInferred (JNIEnv * env, jclass) { return env->NewStringUTF( INFERRED ); } /**< Entete de la colonne infere du fichier performance @return String entete infere */
//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_PerfHeaderError (JNIEnv *env , jclass) { return env->NewStringUTF( ERROR ); } /**< Entete de la colonne erreur du fichier performance @return String entete erreur */
//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_PerfHeaderClassInferred (JNIEnv *env, jclass) { return env->NewStringUTF( CLASINF ); } /**< Entete de la colonne classe infere du fichier performance @return String entete classe infere */
//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_PerfHeaderBlank (JNIEnv *env, jclass) { return env->NewStringUTF( BLANK ); } /**< Entete de la colonne blanc du fichier performance @return String entete blanc */
//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_PerfHeaderMinKernel (JNIEnv *env, jclass) { return env->NewStringUTF( MINKERNEL ); }
//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_PerfHeaderMaxKernel (JNIEnv *env, jclass) { return env->NewStringUTF( MAXKERNEL ); }
//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_PerfHeaderMinSupport (JNIEnv *env, jclass) { return env->NewStringUTF( MINSUPPORT ); }
//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_PerfHeaderMaxSupport (JNIEnv *env, jclass) { return env->NewStringUTF( MAXSUPPORT ); }
//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_PerfHeaderInferredSymbMatch (JNIEnv *env, jclass) { return env->NewStringUTF( INFERREDSYMBMATCH ); }
//---------------------------------------------------------------------------
/**
 * Renvoie les labels des classes pour une sortie classif
 * @param output_ptr pointeur de la sortie
 * @return pointeur sur tableau des labels des classes ou null si la sortie n'est pas classif
 */
JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_GetClasses (JNIEnv *env, jclass, jlong output_ptr)
{
	FISOUT *output = (FISOUT *)output_ptr;
	double *classes = NULL;
	jdoubleArray result = NULL;

	try
	{
		if( output->Classification() )
		{
			if(! strcmp(output->Defuzzify(), OUT_CRISP::SugenoDefuz()))
				classes = ((DEFUZ_SugenoClassif *)output->Def)->ClasLabel();
			else if( !strcmp(output->Defuzzify(), OUT_CRISP::MaxCrispDefuz()))
				classes = ((DEFUZ_MaxCrisp *)output->Def)->ClasLabel();
			result = env->NewDoubleArray( output->Def->NbClasses() );
			if( result != NULL )
				env->SetDoubleArrayRegion( result, 0, output->Def->NbClasses(), classes );
		}
	}
	catch(std::exception &e) { ExceptionToJava( env, e ); }
	return result;
}

//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_LinksItemsRulesExtension (JNIEnv *env, jclass) { return env->NewStringUTF( F_ITEM ); } /**< Extension du fichier items rules */
//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_LinksRulesItemsExtension (JNIEnv *env, jclass) { return env->NewStringUTF( F_RULE ); } /**< Extension du fichier rules items */
//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_LinksRulesLinksExtension (JNIEnv *env, jclass) { return env->NewStringUTF( F_LINK ); } /**< Extension du fichier rules links */
//---------------------------------------------------------------------------
JNIEXPORT jstring JNICALL Java_fis_jnifis_LinksRulesSortedExtension (JNIEnv *env, jclass) { return env->NewStringUTF( F_SORTED ); } /**< Extension du fichier rules sorted */

JNIEXPORT void JNICALL Java_fis_jnifis_GenerateSample (JNIEnv *env, jclass, jstring dataFile, jint sampleNumber, jdouble validation_data_ratio, jint seed, jboolean classif, jint columnNumber, jdouble precision, jboolean replace, jboolean truncate)
{
	try
	{
		char *data_file = get_native_string(env, dataFile);

		// 2 cases : generate n pairs of validation-test samples if replace=F
		//           generate n subsamples if replace=T
		if (replace)
			genNSamples(seed, validation_data_ratio, classif, precision, columnNumber, data_file, sampleNumber, 0);
		else
			genSplit(seed, truncate, classif, precision, columnNumber, data_file, sampleNumber, 0);

		release_native_string(data_file);
	}
	catch(std::exception &e) { ExceptionToJava( env, e ); }
}

JNIEXPORT jobjectArray JNICALL Java_fis_jnifis_DataName (JNIEnv *env, jclass)
{
	if(NbVarG == 0)
		return NULL;
	else {
		jobjectArray result = NULL;
		jarray resulta = NULL;

		// creation du tableau de retour
		jclass objectCls = env->FindClass( "java/lang/Object" );
		if( objectCls == NULL )
			return NULL;
		resulta = env->NewObjectArray( NbVarG, objectCls, NULL );
		if( resulta != NULL )
		{
			result=(jobjectArray) resulta;
			for( int i=0 ; i<NbVarG ; i++ )
				env->SetObjectArrayElement( result, i, env->NewStringUTF( VarNameG[i] ) );
		}
		env->DeleteLocalRef( objectCls );
		return result;
	}
}

JNIEXPORT void JNICALL Java_fis_jnifis_WriteTemplate (JNIEnv *env, jclass, jstring file, jdouble kw, jdouble sw)
{
	char *file_name = get_native_string(env, file);
	WriteTemplate (file_name, kw, sw);
	release_native_string(file_name);
}


JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_ReadTemplate (JNIEnv *env, jclass, jstring file)
{
        jdoubleArray result = NULL;
	jdouble jvalues[2];
	char *file_name = NULL;
	try
	{
		file_name = get_native_string(env, file);
		result = env->NewDoubleArray( 2 );
		if( result != NULL )
			ReadTemplate (file_name, jvalues[0], jvalues[1]);
		env->SetDoubleArrayRegion( result, 0, 2, jvalues );
		release_native_string(file_name);
	}
	catch( std::exception &e ) { ExceptionToJava( env, e ); }
	return result;
}

JNIEXPORT void JNICALL Java_fis_jnifis_SetTemplate   (JNIEnv *env, jclass, jlong input_ptr, jdouble kw, jdouble sw)
{
	FISIN *input = (FISIN *)input_ptr;
	input->SetTemplate(kw,sw);
}

JNIEXPORT jdouble JNICALL Java_fis_jnifis_GetKernelWeightTemplate (JNIEnv *env, jclass, jlong input_ptr)
{
	FISIN *input = (FISIN *)input_ptr;
	return (double)(input->GetKernelWeightTemplate());
}

JNIEXPORT jdouble JNICALL Java_fis_jnifis_GetSupportWeightTemplate (JNIEnv *env, jclass, jlong input_ptr)
{
	FISIN *input = (FISIN *)input_ptr;
	return (double)(input->GetSupportWeightTemplate());
}

JNIEXPORT void JNICALL Java_fis_jnifis_SetExpertWeight (JNIEnv *env, jclass, jlong rule_ptr, jdouble weight)
{
	RULE *rule = (RULE *)rule_ptr;
	rule->SetExpertWeight(weight);
}

JNIEXPORT jdouble JNICALL Java_fis_jnifis_GetExpertWeight (JNIEnv *env, jclass, jlong rule_ptr)
{
	RULE *rule = (RULE *)rule_ptr;
	return rule->GetExpertWeight();
}

JNIEXPORT void JNICALL Java_fis_jnifis_DataDistance (JNIEnv *env, jclass, jlong jfis_ptr, jstring jdata_file, jboolean normalize, jdoubleArray jdistances, jdouble power, jstring joutput_file) {
	std::auto_ptr<FIS> fis(((FIS *)jfis_ptr)->Clone());
	char *data_file  = get_native_string(env, jdata_file);
	jdouble *distances = env->GetDoubleArrayElements(jdistances, NULL);
	char *output_file  = get_native_string(env, joutput_file);

	try {
		int nbex = 0;
		fis->dist(data_file, nbex, output_file, distances, (bool)normalize, power, false, true);
	} catch(std::exception &e) { ExceptionToJava( env, e ); }

	release_native_string(data_file);
	release_native_string(output_file);
}

JNIEXPORT jdoubleArray JNICALL Java_fis_jnifis_PcPe (JNIEnv *env, jclass, jlong jinput_ptr, jdoubleArray jdatas, jint jdata_size) {
	FISIN* input = (FISIN*)jinput_ptr;
	jdouble *datas = env->GetDoubleArrayElements(jdatas, NULL);
    jdoubleArray result = env->NewDoubleArray( 2 );
    jdouble jvalues[2];
	input->PcPe(datas, jdata_size, jvalues[0], jvalues[1]);
	env->SetDoubleArrayRegion(result, 0, 2, jvalues);
	env->ReleaseDoubleArrayElements(jdatas, datas, 0);
	return result;
}

JNIEXPORT void JNICALL Java_fis_jnifis_CheckFuzDist (JNIEnv *env, jclass, jlong jinput_ptr) {
	FISIN* input = (FISIN*)jinput_ptr;
	try {
		input->CheckFuzDist();
	} catch(std::exception &e) { ExceptionToJava( env, e ); }
}

JNIEXPORT jdouble JNICALL Java_fis_jnifis_InputDistance (JNIEnv *env, jclass, jlong jinput_ptr, jdouble x, jdouble y, jboolean normalize) {
	FISIN* input = (FISIN*)jinput_ptr;
	try {
		return input->Distance(x, y, normalize);
	} catch(std::exception &e) {
		ExceptionToJava( env, e );
		return 0;
	}
}












