//***********************************************************************
//
//
//                               FISLINK.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISLINK class functions used by FISPRO, 
//        part of package links 
//        requires library fispro

//**********************************************************************



#include "linkfis.h"

//! Global variable used by qsort
double *Tosort; 

int CmpTosort(const void * a, const void *b)
  //****************************************
{
  if(Tosort[*(int *)a] > Tosort[*(int *)b]) return -1;
  if(Tosort[*(int *)a] < Tosort[*(int *)b]) return 1;
  return 0;
}

//****************************************************************************
//
//
//              Building of crossed maSortedces rules/items
//
//
//****************************************************************************

void FISLINK::CrossedFiles(char * data)
  //*************************************
{
  int i, k, NbCol, MaxNb, l, bufsize;
  char * buf=NULL, * ch=NULL, s;
  double * Tab=NULL;
  int * index=NULL;
  FILE * Ex, *fTab;

  MaxNb = -1;

  NbCol = NbIn + NbOut;
  Tab = new double [NbCol];

  ifstream g(data); 
  if(g.fail()) 
    {
      sprintf( ErrorMsg, "~CannotOpenSampleFile~: %.100s~", data);
      throw runtime_error( ErrorMsg );
    }
  bufsize = MaxLineSize(g);

  buf = new char [bufsize];

  ListInt ** ItemsList=NULL;
  ItemsList = new ListInt * [NbRules];

  for(i = 0; i < NbRules; i++)
    ItemsList[i] = new ListInt;
 
  index = new int [NbRules];

  l = 0;
  if(PrefixResult) l = strlen(PrefixResult);
  ch = new char [20+l];

  if(PrefixResult) sprintf(ch, "%s.%s", PrefixResult, F_ITEM);
  else sprintf(ch, "%s", F_ITEM);
  Ex = fopen(ch, "wt");

  if(PrefixResult) sprintf(ch, "%s.%s", PrefixResult, F_TAB);
  else sprintf(ch, "%s", F_TAB);
  fTab = fopen(ch, "wt");


  for(i = 0; i < NbRules; i++)  
    {
      Rule[i]->Weight = 0.0;
      Rule[i]->CumWeight = 0.0;
      Rule[i]->NbOccur = 0;
    }

  for (i = 0; i < NbIn; i++) 
    {
      In[i]->MinObserve = 1e30;
      In[i]->MaxObserve = -1e30;
    }
 
  k = 1; // item number

  // read header
  int hdr;
  s = ReadSeparator(data, hdr);
  if( hdr )
        g.getline(buf, bufsize);
        
  while (1)
    {
      g.getline(buf, bufsize);
      if(g.eof()) break;
    
      if((strlen(buf) == 0) || (buf[0] == 0x00D) || 
	 ((l = SearchNb(buf, Tab, NbCol, s)) < NbIn) )
	{
	  sprintf( ErrorMsg, "~UnexpectedNumberOfValues~ in the %s file \n %d values expected , %d values read\n", data, NbIn, l );
	  delete [] Tab; delete [] buf; delete [] index;
	  for(i = 0; i < NbRules; i++)   delete ItemsList[i];	
	  delete [] ItemsList;
	  throw runtime_error( ErrorMsg );
	}

      for(i = 0; i < NbIn; i++)     // membership degrees
	{
	  if(FisIsnan(Tab[i]))      // missing value : equal degrees to all the MF
	    In[i]->SetEqDegs(Tab[i]);
	  else 
	    {
	      In[i]->GetDegsV(Tab[i]); 
	      if(Tab[i] < In[i]->MinObserve) In[i]->MinObserve = Tab[i];
	      if(Tab[i] > In[i]->MaxObserve) In[i]->MaxObserve = Tab[i];
	    }
	}
    
      // list of the rules fired by the item
      fprintf(Ex, "%d", k); 
      fprintf(fTab, "%d", k); 

      for(i = 0; i < NbRules; i++)   // Rule matching degree
	{
	  if(Rule[i]->IsActive() == false) continue;
	  Rule[i]->ExecRule();
	  fprintf(fTab, "%c %5.3f", SEPARE, Rule[i]->Weight); 

	  if(Rule[i]->Weight > Thres)                
	    {
	      Rule[i]->CumWeight += Rule[i]->Weight; 
	      Rule[i]->NbOccur ++;
	      fprintf(Ex, "%c %d", SEPARE, i + 1); 
	      ItemsList[i]->Add(k); // item k fires rule i
	      // Add makes a new allocation - element will have to be removed in destructor
	      Rule[i]->Weight = 0.0; 
	    }
	}   
      fprintf(Ex, "\n");   
      fprintf(fTab, "\n");   
     k++;
    } // end of while()

  fclose(Ex);
  fclose(fTab);
 
  for(i = 0; i < NbRules; i++)   // Maximal number of items that fire a rule
    if(ItemsList[i]->nb > MaxNb) MaxNb = ItemsList[i]->nb;

  SortRules(index);
 
  // Write the list of items that fire each rule (a line per rule)
  FILE * Reg;
  ListIntItem *ptr, *ptr_s;
 
  if(PrefixResult) sprintf(ch, "%s.%s", PrefixResult, F_RULE);
  else sprintf(ch, "%s", F_RULE);

  Reg=fopen(ch,"wt");

  if(! Reg)
    {
      sprintf( ErrorMsg, "~CannotOpenFile~: %.100s~", ch);
      throw runtime_error( ErrorMsg );
    }
 
  fprintf(Reg,"%d\n%d\n",NbRules,MaxNb);

  for(i = 0; i < NbRules; i++)   
    {
      k = index[i];
      fprintf(Reg,"%d%c %8.5f%c %d",k+1, SEPARE, Rule[k]->CumWeight, SEPARE, Rule[k]->NbOccur);
      ptr = ItemsList[k]->begin;
      while((ptr_s = ptr->Next) != NULL)
	{
	  fprintf(Reg,"%c %d", SEPARE, ptr->value);
	  ptr = ptr->Next;
	}
      fprintf(Reg,"%c %d\n", SEPARE, ptr->value);
    } 
  fclose(Reg);
 
  delete [] index;     
  delete [] Tab;    
  delete [] buf;    
  delete [] ch;   
  if (ItemsList !=NULL)
    for(i = 0; i < NbRules; i++) 
      delete ItemsList[i];
  delete [] ItemsList;
  

} // End of CrossedFiles



void FISLINK::LkRatios(double * a, double * b, double & Lab, double & Lba)
  //**********************************************************************
{
  int n, i, j;
  n = 0;
  
  if(a == b) { Lab = Lba = 1.0; return; }
  
  for(i = 1; i <= a[0]; i++)         
    // the number of items firing the rule is stored in a[0] 
    for(j = 1; j <= b[0]; j++)
      if(a[i] == b[j]) { n++; break;}
  
  Lab = n / a[0]; 
  Lba = n / b[0];
} 


void FISLINK::LkBetweenRules(void)
  //******************************
{
  int i, j, nrules, maxcol, l, bufsize;
  double * lien=NULL;
  double ** Tab=NULL;
  char *buf, *ch;

  l = 0;
  if(PrefixResult) l = strlen(PrefixResult);   
  ch = new char [20+l];

  if(PrefixResult) sprintf(ch, "%s.%s", PrefixResult, F_RULE);
  else sprintf(ch, "%s", F_RULE);
  ifstream f(ch);

  bufsize = MaxLineSize(f);
  
  if(f.fail())
    {
      sprintf( ErrorMsg, "~CannotOpenFile~: %.100s~", ch);
      throw runtime_error( ErrorMsg );
    }
 
  buf = new char [bufsize];
  
  f.getline(buf, bufsize);
  nrules = atoi(buf);
  f.getline(buf, bufsize);
  maxcol = atoi(buf) + 3;    // Rule number, cumulated weight and number of items 
 
  lien = new double [nrules * nrules];
  Tab = new double * [nrules];
  for(i = 0; i < nrules; i++)
    Tab[i] = new double [maxcol];

  for(i = 0; i < nrules; i++)
    {
      f.getline(buf, bufsize);

      if((strlen(buf) == 0) || (buf[0] == 0x00D) || 
	 (SearchNb(buf, Tab[i], maxcol) < Tab[i][2] + 3) ) 
	// number of values different from expected
	// Only the inequality is checked :
	// If the number of items is zero, the end of list value (-1) is printed
	{
	  sprintf( ErrorMsg, "~UnexpectedNumberOfValues~ in the rules.items file : %d values minimum expected ", (int) Tab[i][2] + 3);
	  if (Tab !=NULL)
	    for(i = 0; i < nrules; i++) delete [] Tab[i];
	  delete [] Tab; 
	  delete [] lien;
	  throw runtime_error( ErrorMsg );
	}
    }

  for(i = 0; i < nrules; i++)
    for(j = i;  j < nrules; j++)
      LkRatios(&Tab[i][2] , &Tab[j][2], lien[i * nrules +j],lien[j * nrules +i]);
 
  FILE * g;
  if(PrefixResult) sprintf(ch, "%s.%s", PrefixResult, F_LINK);
  else sprintf(ch, "%s", F_LINK);

  if( (g = fopen(ch, "wt")) == NULL ) 
    {
      sprintf( ErrorMsg, "~CannotOpenFile~: %.100s~", ch);
      throw runtime_error( ErrorMsg );
    }

  for(i = 0; i < nrules; i++)
    {
      for(j = 0;  j < nrules - 1; j++)
	fprintf(g, "%4.2f%c", lien[i * nrules +j], SEPARE);
      fprintf(g, "%4.2f\n", lien[i * nrules +j]);
    }
  fclose(g);
  
  if (Tab !=NULL)
    for(i = 0; i < nrules; i++) delete [] Tab[i]; 
  delete [] Tab;
  delete [] lien; 
  delete [] buf;
  delete [] ch;
}


void FISLINK::SortRules(int * index)
  //********************************
{
  int i;

  for(i = 0; i < NbRules; i++)   index[i] = i;
  
  if (! Sorted) return;
  
  Tosort = new double [NbRules]; 
  for(i = 0; i < NbRules; i++)   Tosort[i] = Rule[i]->CumWeight;

  qsort(index, NbRules, sizeof(int), CmpTosort); 

  delete [] Tosort; 

  int l = 0;
  if(PrefixResult) l = strlen(PrefixResult);
  char *ch = new char [20+l];
  if(PrefixResult) sprintf(ch, "%s.%s", PrefixResult, F_SORTED);
  else sprintf(ch, "%s", F_SORTED);
  FILE * f;
  if((f = fopen(ch, "wt")) == NULL )
    {
      sprintf( ErrorMsg, "~CannotOpenFile~: %.100s~", ch);
      throw runtime_error( ErrorMsg );
    }

  for(i = 0; i < NbRules; i++) Rule[index[i]]->Print(f);
  fclose(f);
  delete []ch;
}


//*******************    End of FISLINK.CPP   ***********************************
