//***********************************************************************
//
//
//                               FISLINK.H
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISLINK class functions used by FISPRO, 
//        part of package links 
//        requires library fispro

//**********************************************************************

#ifndef __FISLINK_H
#define __FISLINK_H

#include "fis.h"

#define F_ITEM "items.rules"
#define F_TAB  "items.murules"
#define F_RULE "rules.items"
#define F_LINK "rules.links"
#define F_SORTED "rules.sorted"


//*************************************************************************************
//
//
//                             CLASS FISLINK
//
//
//*************************************************************************************

//! Provide the files 'rules.items', 'items.rules', 'rules.links' and, 
//! optionally 'rules.sorted'. See user documentation for details.
class FISLINK : public FIS
//************************ 
{
  
 protected:

  //! Minimum degree for a rule to be considered as fired. 
  //! Default value : EPSILON
  double Thres; 

  //! If true, the rules are sorted according to their cumulated weight
  int Sorted;

  //! This prefix is used to build the result filenames: 
  //! PrefixResult'.items.rules', PrefixResult'.rules.items', , PrefixResult'.rules.links.
  char * PrefixResult;

  //! Write of the crossed files: rules.items, items.rules and items.murules
  void CrossedFiles(char * p_data); 
  
  //! Write the rules.links file from the rules.items one
  void LkBetweenRules(void);
  
  //! Lab and Lba are the link levels between a and b for one hand, b and a for the other
  //! Usually, the minimum is used
  void LkRatios(double * a, double * b, double & Lab, double & Lba);

  //! Sort the rules according their cumulated weight
  void SortRules(int * index);

 public: 

  FISLINK(char *fichier, char * p_data, double s, int OptTri, char * Prefix) : FIS(fichier)
    {
      Sorted = OptTri;
      Thres = s;
      PrefixResult = Prefix;

      CrossedFiles( p_data );
      LkBetweenRules();
    }
};


//******************************************************************************
//
//
//                   A small list of integer manager
//
//
//******************************************************************************

//! A 'ListInt' element
class ListIntItem
{
  public :
    int value;
  ListIntItem * Next;

  ListIntItem(int v = -1)
    {
      value = v;
      Next = NULL;
    }
};

//! A small list of integer manager
class ListInt
{
  ListInt& operator=(const ListInt&);
  ListInt(const ListInt&);
  public : 
    int nb; 
  ListIntItem *begin;
  
  ListInt()
    {
      nb = 0;
      begin = new ListIntItem;
    }
  
  ~ListInt()
    {
      ListIntItem *ptr = begin;
      ListIntItem *ptr_s = ptr->Next;

      while((ptr_s = ptr->Next) != NULL)
	{
	  delete ptr; 
	  ptr = ptr_s;
	}
      //delete ptr_s;
      //leak memory correction - bch July 20, 2009
      delete ptr;
    }

  void Add(int v)
    {
    
      if(! nb) begin->value = v;
      else
	{
	  ListIntItem *ptr = begin;
	  while (ptr->Next != NULL)   ptr = ptr->Next; 
	  ptr->Next  = new ListIntItem(v);
	}
      nb ++;
    }
};

#endif

//*********************      End of FISLINK.H      ***********************
