//******************************************************************
//
//
//                              FISOLS.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC
// FISPRO Version : 3.1 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Contact : fispro@supagro.inra.fr
// Last modification date:December 1, 2007
//******************************************************************


#include "fisols.h"
#include "fis.h"

//**********************************************************************
//
//
//                              CLASSE FISOLS
//
//
//**********************************************************************

FISOLS::FISOLS(char * Data1, char * FisConfig, int opti, const char *result_path)
//******************************************************
{
  int i,j;
  double min, max;
  char *ch;
  //initialise the system
  Init(Data1);
  InitSystem(FisConfig);
  Opti=opti;
  ch = new char[20];
  OutSave = new FISOUT * [NbOut];
  for(i=0;i<NbOut;i++) OutSave[i]=NULL;
  //erase the rule base if the aim is to select and not only to optimize conclusions
  if(Opti==false)
    {
      if(Rule!=NULL)
	{
	  for(i=0;i<NbRules;i++)
	      if(Rule[i]!=NULL) delete Rule[i];
	  delete [] Rule;
	}
      //check if one of the output is a fuzzy one (improvement to come : output save and rehabilitation), and if one is, replace it
      for(i=0;i<NbOut;i++)
	{
	  if(! strcmp(Out[i]->GetOutputType(), OUT_FUZZY::OutputType()))
	    {
	      OutSave[i]=Out[i]->Clone();
	      delete Out[i];
	      Out[i]= new OUT_CRISP();

	      min = INFINI;
	      max = -INFINI;

	      for(j = 0; j < Nrows; j++)
		{
		  if(pData[j][NbIn+i] < min) min = pData[j][NbIn+i];
		  if(pData[j][NbIn+i] > max) max = pData[j][NbIn+i];
		}
	      sprintf(ch, "Output%d", i);
	      if(fabs(max-min)>EPSILON) Out[i]->SetRange(min, max);
	      Out[i]->SetName(ch);
	    }
	}
    }
  // check for input and output number coherence
  if(Ncols<NbIn+NbOut)
    {
      sprintf( ErrorMsg, "~InvalidNumberOfColumn : %d Read %d Awaited~\n", Ncols, NbIn+NbOut);
      throw std::runtime_error( ErrorMsg );
    }
  // set the ols number of outputs
  SetNbOfOut(NbOut);

  // rebuilt an empty rule base if selection must be done
  if(Opti==false)
    {
      NbExceptions = 0;
      NbRules = NbActRules = Nrows;

      if(NbRules!=0)
	{
	  Rule = new RULE * [NbRules];
	  for(i=0;i<NbRules;i++) Rule[i]=NULL;
	}
    }

  // generate rules from samples if selection must be done
  if(Opti==false)
    {
      if(Display)
	{
	  printf("\n***************Rule generation****************\n");
	}
      GenerateRules();
    }
  if(Display)
    {
      printf("\n***************Matrix Generation****************\n");
    }
  // generate matrix of FBF (fuzzy basis functions)
  GenerateMatrix(pData, result_path ? result_path : Name, Nrows, NbRules);

  delete [] ch;
}

FISOLS::FISOLS(char * Data1)
//**************************
{
  Init(Data1);
  OutSave = new FISOUT * [1];
  OutSave[0]=NULL;
}

void FISOLS::Run(const char *result_path)
//********************
{
  int i;
  char * buf=NULL;
  char * tmp=NULL;

  if(NumS>NbOut || NumS<0)
    {
      sprintf( ErrorMsg, "~InvalidOutputNumber : %d~\n", NumS);
      throw std::runtime_error( ErrorMsg );
    }

  if(CfgFile == NULL)
    {
      rm = true;
      CfgFile = new char [strlen(fData) + 10];
      sprintf(CfgFile, "%s.fis", fData);
    }
  // make first pass if selection must be done
  if(Opti==false)
    {
      if(Display)
	  printf("\n*****************First Pass in progress*******************\n");
      FirstPass(result_path);
    }

  if(Display)
      printf("\n*****************Second Pass in progress*******************\n");
  if(NbRules!=0)  SecondPass(result_path);      // optimize conclusions
  // when no rule is selected (all inputs are inactive) create an universal rule
  else
    {
      tmp = new char [5];
      buf = new char [5*(NbIn+NbOut)];
      //      sprintf(buf,"");
      buf[0]=0;
      for(i=0;i<NbIn+NbOut-1;i++)
	{
	  sprintf(tmp,"0%c",SEPARE);
	  strcat(buf,tmp);
	}
      sprintf(tmp,"0");
      strcat(buf,tmp);
      // Addrule (fis.cpp) reallocates the Rule array
      AddRule(new RULE(NbIn,In,NbOut,Out,cConjunction, buf));
      delete [] tmp;
      delete [] buf;
    }

  // reinitialize conclusions
  for(i=0;i<NbOut;i++) Out[i]->InitPossibles(Rule,NbRules,i);
  //vocabulary reduction (function from fis.h)
  if(VocRed!=0 && OriginalGauss==0 && OutSave[NumS]==NULL)
    {
      if(Display)
	  printf("\n*****************Vocabulary Reduction*******************\n");
      // launch vocabulary reduction with the specified parameters
      VocReduc(NumS,pData,Nrows,0.0,AllowedLoss,NVocItem,VocData);
      if(FuzzyOut!=0) // fuzzify the output, if precised
	{
	  if(Display)
	      printf("\n*****************Output Fuzzification*******************\n");
	  FIS::Crisp2Fuz(NumS, Out[NumS]->Defuzzify());
	}
    }

  for(i=0;i<NbOut;i++)
    if(OutSave && OutSave[i])
      {
	ReplaceCrispOutput(Out[i]);
	Out[i]->InitPossibles(Rule,NbRules,i);
      }

  PrintCfgFis();    // print final fis file

  if(RmTmpFile!=0)    // remove tmp files
    {
      if(Display)
	  printf("\n*****************Remove temporary files*******************\n");
      RemoveTmpFiles();
    }

  if(Display)
      printf("\nNumber of selected rules : %i", NbRules);

}

void  FISOLS::ReplaceCrispOutput(FISOUT *& Outtmp)
  //**********************************************
{
  int i,j, index;
  double conc, max;

  for(i=0;i<NbRules;i++)
    {
      conc=Rule[i]->GetAConc(NumS);
      index=0;
      max=OutSave[NumS]->GetADeg(0,conc);
      for(j=1;j<OutSave[NumS]->GetNbMf();j++)
	if(OutSave[NumS]->GetADeg(j,conc)>max)
	  {
	    max=OutSave[NumS]->GetADeg(j,conc);
	    index=j;
	  }
      Rule[i]->SetAConc(NumS,index+1);
    }
  delete Outtmp;
  Outtmp=OutSave[NumS];
}


void FISOLS::Init(char *Data1)
//****************************
{
  // OLS initialisation
  NumS=0;
  Opti=0;
  NbOfOut=1;
  UniqTol=0.25;
  OriginalGauss=0;
  Loss = 0.01;
  NbOfRules = 10000 ;
  Std = 0;
  VocRed = 0 ;
  AllowedLoss = 0.1 ;
  NVocItem = 0 ;
  FuzzyOut = 0 ;
  RmTmpFile = 0 ;
  fData=Data1;
  rm = false;
  Display = false;
  tData=Data1;
  CfgFile = NULL;
  pData=ReadSampleFile(Data1,Ncols,Nrows);
  // save original number of examples for memory cleaning at the end of program
  Ninitex=Nrows;
  //
  SetConjunction(RULE::PremiseMin());
  OutSave = NULL;
}

void FISOLS::FirstPass(const char *result_path)
//**************************
{
  double ** P;
  char * FiMat;
  int i=0, nsel=0;
  double * centres;
  char * nom=NULL;

  int * indices=NULL;
  double * conseq;
  double * Y;

  P = new double * [Nrows];
  for(i = 0; i < Nrows; i++)  P[i] = new double [Nrows];

  FiMat= new char[strlen(result_path ? result_path : Name) + 10];
  sprintf(FiMat, "%s.mat", result_path ? result_path : Name);

  ReadMatrix(P, NbRules, NbRules, FiMat);

  Y = new double[Nrows];

  for(i = 0; i < Nrows; i++) Y[i] = pData[i][NbIn+NumS];

  nsel=0;

  if(!result_path) {
	  i = FileNameIndex(fData);
	  nom = new char [strlen(fData + i) + 1];
	  sprintf(nom, "%s", fData + i);
  }
  //Computation of OLS first pass
  ols(P, Y, Loss, NbOfRules, Nrows, indices, conseq, nsel, result_path ? result_path : nom);

  centres = new double [nsel];
  //Update of the rules and of the conclusions
  UpdateRules(nsel, indices);
  for(i=0;i<nsel;i++) Rule[i]->SetAConc(NumS,conseq[i]);

  delete [] FiMat;
  delete [] centres;
  delete [] Y;

  for(i = 0; i < Nrows; i++) delete [] P[i];
  delete [] P;

  delete [] indices;
  delete [] conseq;
  delete [] nom;

}


void FISOLS::SecondPass(const char *result_path)
//***************************
{

  int i, NbLig;
  double ** P=NULL, * Y=NULL;
  char * FiMat=NULL;
  double *t=NULL;

  //t = new double [NbRules];// t is allocated inside the ols2 function

  if (tData == NULL)
    tData = fData;
  if (pData!=NULL)
	{
	  for(i = 0; i < Ninitex; i++)
	    delete [] pData[i];
	  delete [] pData;
	  pData=NULL;
	}

  pData=ReadSampleFile(tData, Ncols, NbLig);
  Ninitex=NbLig;

  Nrows = NbLig;

  // P construction
  GenerateMatrix(pData, result_path ? result_path : tData, Nrows, NbRules);


  P = new double * [Nrows];
  for(i = 0; i < Nrows; i++)  P[i] = new double [NbRules];

  FiMat = new char[strlen(result_path ? result_path : tData) + 10];
  sprintf(FiMat, "%s.mat", result_path ? result_path : tData);

  ReadMatrix(P, Nrows, NbRules, FiMat);

  Y = new double [Nrows];

  for(i = 0; i < Nrows; i++)   Y[i]=pData[i][NbIn+NumS];
  //ols second pass that'll compute optimal conclusions
  ols2(P, Y, Nrows, NbRules, t);
  //conclusions update
  for(i=0;i<NbRules;i++) Rule[i]->SetAConc(NumS,t[i]);

  delete [] FiMat;
  delete [] Y;
  for(i = 0; i < Nrows; i++)  delete [] P[i];
  delete [] P;
  delete [] t;
}


void FISOLS::StdFP(void)
//**********************
{
   int NbMf, i, j;
   double * tab, * centres, vinf, vsup;
   char * ch;

   tab = new double[3];
   ch = new char[100];

   for (i = 0; i < NbIn; i++)
     {
       // check for sef centers
       if((In[i]->IsActive())==false) continue;
       NbMf = In[i]->GetNbMf();
       centres = new double [NbMf];
       for (j = 0; j < NbMf; j++)
	 {
	   (In[i]->GetMF(j))->GetParams(tab);
	   centres[j]=tab[1];
	 }
       vinf = In[i]->min();
       vsup =  In[i]->max();
       delete In[i];
       // build new partitions
       In[i] = new FISIN(centres, NbMf, vinf, vsup, false);
       sprintf(ch,"Var%i",i+1);
       In[i]->SetName(ch);
       In[i]->SetStdMfNames();
       delete [] centres;
     }

   delete [] tab;
   delete [] ch;
}


void FISOLS::GenerateFIS(const char *result_path)
//****************************
{
  char *ch;
  char *nom;
  double min, max;
  int i,j=0;

  ch = new char [30];

  NbIn = Ncols-NbOfOut;
  NbOut = NbOfOut;

  // initialise In and Out
  if(NbIn!=0)
    {
      In = new FISIN * [NbIn];
      for(i=0;i<NbIn;i++) In[i]=NULL;
    }

  if(NbOut!=0)
    {
      Out = new FISOUT * [NbOut];
      for(i=0;i<NbOut;i++) Out[i]=NULL;
      OutValue = new double [NbOut];
      OutErr = new double [NbOut];
    }

  //MF inputs generation
  for(i=0;i<NbIn;i++)
    {
      In[i] = new INPUTOLS(Nrows, pData, i, j++, UniqTol, OriginalGauss);
      In[i]->SetStdMfNames();
      if(strcmp(In[i]->GetMF(0)->GetType(),"universal")==0) In[i]->Deactivate();
    }
  
  // Output generation
  for(i = 0; i < NbOut; i++)
    {
      Out[i]= new OUT_CRISP();

      min = INFINI;
      max = -INFINI;

      for(j = 0; j < Nrows; j++)
	{
	  if(pData[j][NbIn+i] < min) min = pData[j][NbIn+i];
	  if(pData[j][NbIn+i] > max) max = pData[j][NbIn+i];
	}
      sprintf(ch, "Output%d", i);
      if(fabs(max-min)>EPSILON) Out[i]->SetRange(min, max);
      Out[i]->SetName(ch);
    }
 
  NbExceptions = 0;
  NbRules = NbActRules = Nrows;

  if(NbRules!=0)
    {
      Rule = new RULE * [NbRules];
      for(i=0;i<NbRules;i++) Rule[i]=NULL;
    }

  if(Std!=0 && OriginalGauss==0) StdFP(); // SEF standardization

  i = FileNameIndex(fData);
  nom = new char [strlen(fData + i) + 1];
  sprintf(nom, "%s", fData + i);

  SetName(nom);
  strcpy(strMissingValues,FIS::MeanMissingValues());
  
  // generate rules from samples and matrix of FBF
  if(Display)
      printf("\n***************Rule generation****************\n");
  
  GenerateRules();
  
  if(Display)
      printf("\n***************Matrix generation****************\n");
  GenerateMatrix(pData, result_path ? result_path : Name, Nrows, NbRules);
  
  delete [] ch;
  delete [] nom;
 
}


void FISOLS::GenerateRules(void)
//******************************
{
  int i, j, ind;

  char *tmp, *buf;

  buf = new char [6*NbIn+6*NbOut+20];
  tmp = new char [6+30];

  for(j = 0; j < Nrows; j++) // For each sample
    {
      buf[0] = 0;

      for(i = 0; i < NbIn; i++) // For each input, get the premise
	{
	  if(In[i]->IsActive())
	    {
	      ind = ((INPUTOLS *)In[i])->MaxDeg(pData[j][i]);
	      sprintf(tmp, "%d%c", ind + 1, SEPARE);
	      strcat(buf, tmp);
	    }
	  else
	    {
	      sprintf(tmp, "%d%c", ANY, SEPARE);
	      strcat(buf, tmp);
	    }
	}

      for(i = 0; i < NbOut; i++) // For each output, set conclusion to default value
	{
	  sprintf(tmp,"%f%c", Out[i]->DefaultValue(), SEPARE);
	  strcat(buf, tmp);
	}

      //rule building
      Rule[j] = new RULE(NbIn, In, NbOut, Out, cConjunction, buf);
    }

  delete [] buf;
  delete [] tmp;
}

void FISOLS::GenerateMatrix(double ** M, const char * Oname, int nl, int nr)
//********************************************************************
{
int i, j;
double TotalWeight;

char *FiMat;
FILE *f;

 FiMat = new char [strlen(Oname) + 10];
 sprintf(FiMat, "%s.mat", Oname);

 f = fopen(FiMat,"wt");

 for(j = 0; j < nl; j++)
   {
     TotalWeight = 0.0;

     for(i = 0; i < NbIn; i++)
       {
	 if(!In[i]->IsActive()) continue;

	 if(FisIsnan(M[j][i]))            // missing value
	   {
	     if(!strcmp(strMissingValues, FIS::RandomMissingValues()))
	       In[i]->GetRandDegs(M[j][i]);
	     else
	       In[i]->SetEqDegs(M[j][i]);
	   }
	 else In[i]->GetDegsV(M[j][i]);
       }
     //compute first part of the FBF
     for(i = 0; i < nr; i++)
       {
	 Rule[i]->ExecRule();
	 TotalWeight += Rule[i]->Weight;
       }

     // Write one matrix line
     for(i = 0; i < nr; i++)
       {
	 if(TotalWeight > EPSILON)
	   fprintf(f,"%f%c ",Rule[i]->Weight/TotalWeight, SEPARE); //compute total p(i)
	 else fprintf(f,"%f%c ",Rule[i]->Weight, SEPARE);
      }
     fprintf(f,"\n");

   }

 fclose(f);
 delete [] FiMat;
}

void FISOLS::UpdateRules(int & Nb, int * v)
//*****************************************

{
  int i, j;
  RULE **tmp;

  tmp = new RULE * [Nb];
  // update rule base with the selected rules
  for(i = 0; i < Nb; i++)
    {
      tmp[i] = new RULE (*(Rule[v[i]]),In,Out);
      for(j = 0; j < NbIn; j++)
	{
	  if(!In[j]->IsActive()) continue;
	  tmp[i]->SetAProp((((INPUTOLS *)In[j])->MaxDeg(pData[v[i]][j]))+1 ,j);
	}
    }

  for(i = 0; i < NbRules; i++)    delete Rule[i];
  delete [] Rule;

  Rule = tmp;
  NbRules = Nb;

}

//*************************************************************************
//
//
//                           CLASSE INPUTOLS
//
//
//*************************************************************************

// Main constructor (see fisols.h for details)
INPUTOLS::INPUTOLS(int Nb, double **Tab, int NumCol, int NumInput, double & tol, int gauss)  : FISIN (0)
//***************************************************************************

{
  double *Unique, *buffer;
  int *Occur;
  int i, NbUnique;
  char * ch;
  ch = new char [20];

  Unique=NULL;
  buffer=NULL;
  Occur=NULL;

  sprintf(ch,"V%i", NumCol);
  SetName(ch);

  ValInf = INFINI;
  ValSup = -INFINI;

  buffer = new double [Nb];
  // check domain variation
  for (i = 0; i < Nb; i++)
    {
    if(Tab[i][NumCol] < ValInf) ValInf = Tab[i][NumCol];
    if(Tab[i][NumCol] > ValSup) ValSup = Tab[i][NumCol];
    buffer[i] = Tab[i][NumCol];
    }
  // case where domain contain one value
  if(fabs(ValSup-ValInf)<EPSILON)
    {
      ValSup=ValSup*1.1;
      ValInf=ValInf*0.9;
    }
  // original method, one sef per distinct value
  if(gauss!=0)
    {
      if(SortUniq(buffer, Nb, &Unique, &Occur, NbUnique, 0.0) < 0)
	Nmf = NbUnique - 1;
      else Nmf = NbUnique;
    }
  // modified version of the OLS
  else
    {
      if(SortUniq(buffer, Nb, &Unique, &Occur, NbUnique, (ValSup - ValInf) *  tol) < 0)
	Nmf = NbUnique - 1;
      else Nmf = NbUnique;
    }

  Fp = new MF *[Nmf];
  //  Mfdeg = new double [Nmf];
  // Universal Sef
  if(Nmf==1)
    {
      Fp[0] = new MFUNIV(-INFINI, INFINI);
      sprintf(ch, "Var%dMf%d", NumInput, 1);
      Fp[0]->SetName(ch);
    }
  else
    {
      for(i = 0; i < Nmf; i++)
	{
	  // original algorithm, use of tol in standard deviation computation
	  if(gauss!=0)
	    {
	      if(tol!=0) Fp[i] = new MFGAUSS((ValSup - ValInf) * tol, Unique[i]);
	      else Fp[i] = new MFGAUSS((ValSup - ValInf)/20, Unique[i]);
	      sprintf(ch, "Var%dMf%d", NumCol, i+1);
	      Fp[i]->SetName(ch);
	    }
	  // modified version of the OLS method
	  else
	    {
	      if(tol!=0)  Fp[i] = new MFTRI(Unique[i], (ValSup - ValInf) * tol);
	      else Fp[i] = new MFTRI(Unique[i], (ValSup - ValInf)/8);
	      sprintf(ch, "Var%dMf%d", NumCol, i+1);
	      Fp[i]->SetName(ch);
	    }
	}
    }


  if(buffer!=NULL) delete [] buffer;
  if(Unique!=NULL) delete [] Unique;
  if(Occur!=NULL) delete [] Occur;
  delete [] ch;
}

int INPUTOLS::MaxDeg(double v)
//****************************
{
  int k, indice;
  double max;

  GetDegsV(v);

  max=Mfdeg()[0];
  indice = 0;

  for(k = 1; k < Nmf; k++)
    {
      if(Mfdeg()[k] > max)
	{
	  max = Mfdeg()[k];
	  indice = k;
	}
    }
  return indice;
}

void FISOLS::PrintCfgFis(void)
//****************************
{
  FILE *fi;
  char * ch;

  ch = new char [strlen(CfgFile) + 10];
  sprintf(ch, "%s", CfgFile);
  fi = fopen(ch, "wt");

  PrintCfg(fi);
  fclose(fi);

  delete [] ch;
}

void FISOLS::RemoveTmpFiles(void)
//*******************************
{
  char * ch;
  // tData and fData can be NULL if every input is inactive
  if(tData!=fData && tData!=NULL)
    {
      ch = new char [strlen(tData) + 10];
      sprintf(ch, "%s.mat", tData);
      remove(ch);
      delete [] ch;
    }
  if(fData!=NULL)
    {
      ch = new char [strlen(fData) + 10];
      sprintf(ch, "%s.mat", fData);
      remove(ch);
      delete [] ch;
    }
}

//*********************  Fin de FISOLS.CPP ****************************
