
//******************************************************************
//
//
//                              FISOLS.H
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC
// FISPRO Version : 3.1 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Contact : fispro@supagro.inra.fr
// Last modification date:December 1, 2007

//******************************************************************

#ifndef __FISols_H
#define __FISols_H


#include "fis.h"
#include "ols.h"


//***********************
class FISOLS : public FIS
//***********************
{
  FISOLS& operator=(const FISOLS&);
  FISOLS(const FISOLS&);
public :

  //! Data1:  Data file name used for the first pass
  FISOLS(char * Data1);

  //! Data1:  Data file name used for the first pass
  //! CfgFis:  the fis configuration output file name
  //! opti:
  FISOLS(char * Data1, char * FisConfig, int opti, const char *result_path = NULL);

  //! FISOLS destructor
  ~FISOLS()
    {
      if(rm) delete [] CfgFile;
      if(pData!=NULL)
	{
	  for (int i=0;i<Ninitex;i++)
	    delete [] pData[i];
	  delete [] pData;
	  pData=NULL;
	}
      if(OutSave) delete [] OutSave;
    }
  //! every paramater can be changed or called by a Get or SetParameter function

  void SetNumS(int num) {NumS=num;}
  void SetNbOfOut(int num) {NbOfOut=num;}
  void SetUniqTol(double val) {UniqTol=val;}
  void SetOriginalGauss(int num) {OriginalGauss=num;}
  void SetLoss(double val) {Loss=val;}
  void SetVocData(int d) {VocData=d;}
  void SetNbOfRules(int num) {NbOfRules=num;}
  void SetStd(int num) {Std=num;}
  void SetVocRed(int num) {VocRed=num;}
  void SetAllowedLoss(double val) {AllowedLoss=val;}
  void SetNVocItem(int num) {NVocItem=num;}
  void SetFuzzyOut(int num) {FuzzyOut=num;}
  void SetRmTmpFile(int num) {RmTmpFile=num;}
  void SetTestFileName(char *TestData) {tData=TestData;}
  void SetResultFileName(char *ResultFis) {CfgFile=ResultFis;}
  void SetDisplay(int num) {Display=num;}

  int GetNumS(void) const {return NumS;}
  int GetNbOfOut(void) const {return NbOfOut;}
  double GetUniqTol(void) {return UniqTol;}
  int GetOriginalGauss(void) const {return OriginalGauss;}
  double GetLoss(void) const {return Loss;}
  int GetVocData(void) const {return VocData;}
  int GetNbOfRules(void) const {return NbOfRules;}
  int GetStd(void) const {return Std;}
  int GetVocRed(void) const {return VocRed;}
  double GetAllowedLoss(void) const {return AllowedLoss;}
  int GetNVocItem(void) const {return NVocItem;}
  int GetFuzzyOut(void) const {return FuzzyOut;}
  int GetRmTmpFile(void) const {return RmTmpFile;}
  char* GetTestFileName(void) const {return tData;}
  char* GetResultFileName(void) const {return CfgFile;}
  int GetDisplay(void) const {return Display;}

  //! Function called to launch the OLS algorithm after initialisation
  void Run(const char *result_path = NULL);

   /*!
    \fn GenerateFIS(ifstream f)
    \param f file stream pre-defined
    Generate a FIS correponding to the configuration file
  */
  //! Function called to generate a FIS if no pre-defined FIS file is specified
  void GenerateFIS(const char *result_path = NULL);

  FISOUT ** OutSave;

protected :


  //! Data Matrix and file names
  char * fData;
  char * tData;
  char * CfgFile;
  double ** pData;
  double ** pfData;

  //! Nrows : number of rows (or example) in the weighted Matrix
  //! Ncols : number of columns (or rules) in the weighted Matrix
  int Nrows, Ncols, NbOfOut, NumS,Ninitex;
  int Display, rm;
  //! Flag used, if false, to run the first pass and select rules.
  //! If true, only the second pass, rule conclusion optimization is achieved.
  int Opti;


  //! Std : Standardization parameter (1=standardized MF, 0=unstandardized MF)
  //! Updt : Update parameter (1=updated MF after 1st pass, 0=not updated)
  //! Scd : Second pass parameter (1=Second pass is made, 0=no second pass)
  //! Conjunction : min(0) or prod(1) (default : min)
  //! FuzzyOut : fuzzyfication of the output (from reduced voc)
  //! Each parameter can be found in the OLS config file.
  //! VocData:  to select the data used to make the vocabulary reduction
  //! (1 for rule conclusions, o for output learning data).
  int Std, NbOfRules, VocRed, NVocItem, VocData, FuzzyOut, OriginalGauss, RmTmpFile;

  //! HalfWidth : Width of the future triangular membership functions
  //! (in percentage of the input range).
  //! Uniqtol : Tolerance to consider two values as a unique one
  //! (in percnetage of the input range). This value is used to generate MF
  //! when they are not given.
  //! Loss : Treshold of unexplained variance who's accepted
  double UniqTol, Loss, AllowedLoss;

  /*!
    Intialize the whole SIF : build inputs MF (standardize to strong fuzzy sets if Std = 1),
    generate the rules based on examples and the weighted Matrix.
    \fn InitSystem (char * NameData, char * CfgMF, char * CfgFis)
    \param NameData the name of the matrix containing the DATA
    \param CfgMF the name of the file containing the pre-defined MF
    \param CfgFis the name of the file where the FIS config will be written
  */

  /*!
    \fn FirstPass()
    Procede to the first pass of the OLS and select the most important rules
  */
  void FirstPass(const char *result_path);

  //! Initialise the parameters and attributes of the FISOLS object
  void Init(char *Data1);

  //! Remove the temporary files created during process (data.mat)
  void RemoveTmpFiles(void);

  /*!
    \fn GenerateRules(void)
    Generate the Rules who'll be selected during the first pass
  */
  void GenerateRules(void);
  /*!
    \fn GenerateMatrix(double **M, char * Oname, int nl, int nr)
    \param ** M The Data matrix
    \param * Oname Name of the wighted matrix file
    \param nl number of line (examples) in the DATA Matrix
    \param nr number of column (rules, selected or not) of the weighted Matrix
    generate the weighted matrix P of the OLS algorithm
  */
  void GenerateMatrix(double ** M, const char * Nom, int nl, int nr);
  /*!
    \fn SecondPass (char *fi2, double **t)
    \param fi2 the name of the file containing the DATA for the second pass (default : same as for the first pass)
    \param t new consequents (Theta) computed for the second pass
    Make the second pass of the OLS, compute the conclusion (made only if Scd = 1, see config file for details)
  */

  void ReplaceCrispOutput(FISOUT *& Outtmp);

  void SecondPass(const char *result_path);
  /*!
    \fn StdFP()
    standardize (make them strong) the MF during their construction (used if Std = 1)
  */
  void StdFP(void);
  /*!
    \fn UpdateRules(int & Nb, int * v)
    \param &Nb Number of rules to be updated
    \param *v Vector containing the indexes of the selected rules who'll be updated
    Update all the rules who've been selected during the passes of the OLS
  */
  void UpdateRules(int & Nb, int * v);
  /*!
    Print the FIS configuration file named CfgFile.Fis(int i)
  */
  void PrintCfgFis(void);

};


//***************************
class INPUTOLS : public FISIN
//***************************
{

protected :



public:

 //! Main constructor used to build MF during the program
 //! Nb : number of examples (Data matrix number of lines)
 //! Tab : Data Matrix
 //! NumCol : Input Number
 //! NumInput : Active Input Number
 //! tol : Unique Treshold (see CreateOlsCfgFile() for details)
 //! width : MF Width (see CreateOlsCfgFile() for details)
 INPUTOLS(int Nb, double **Tab, int NumCol, int NumInput, double & tol, int gauss);

 //! Constructor used for inactive inputs
 INPUTOLS() : FISIN(0, 0)    // Input inactive
   {
     sprintf(Name,"Inactive");
   }

 //! Destructor of INPUTOLS
 ~INPUTOLS() {}

 /*!
   \fn MaxDeg(double v)
   \param v the value of the variable
   \return the number of the MF for which v has the highest membership function
 */
 int MaxDeg(double v);

};
#endif

