//******************************************************************
//
//
//                              MAINOLS.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.1 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Contact : fispro@supagro.inra.fr
// Last modification date:December 1, 2007
//******************************************************************


#include "fisols.h"

int main(int argc, char ** argv)
{

  if (argc < 2) 
    {
      printf("\nThe program %s needs 1 arguments: ", argv[0]);
      printf("\n\t* the data file name for the first pass");
      printf("\n---------------------------------------------");
      printf("\nOptions are :");
      printf("\n\t* -nNumber of the Output to consider (default:  first output)");
      printf("\n**************************FIS configuration***********************");
      printf("\n\t* -hconjunction  Two types are allowed: 'prod' and 'min' (default)");
      printf("\n\t* -fFisfile the fis configuration file containing MF description");
      printf("\n\t*  (default:  none)");
      printf("\n\t* -x to run only the second pass and optimize conclusions ");
      printf("\n\t*  without rule selection (default:  first pass done)");
      printf("\n\t   (NB : Only works if a fis configuration file is given)");
      printf("\n\t\t OR ");
      printf("\n\t* -j to use original algorithm (Hohensohn and Mendel) (default:  not original algo)");
      printf("\n\t   (NB : if original algorithm, there will be no reduced vocabulary, ");
      printf("\n\t   no strong partitions and no fuzzification)");
      printf("\n\t* -tTolerance tolerance threshold (default:  0.25) or ");
      printf("\n\t*  Standard deviance factor if original algorithm (default:  0.05)");
      printf("\n\t* -s not to make strong fuzzy partitions (default:  yes)");
      printf("\n\t* -oNumberOfOutput the number of outputs (default:  1)");
      printf("\n**************************OLS options****************************");
      printf("\n -------> STOP CRITERIONS (algorithm stops when any criterion is met)");
      printf("\n\t* -gTotal unexplained variance stop condition (default:  1 percent not explained)");
      printf("\n\t* -qNumber of selected rule stop condition (default:  10000)");
      printf("\n -------> OUTPUT REDUCTION and FUZZIFICATION");	
      printf("\n\t* -v to reduce vocabulary (default:  no reduction)");
      printf("\n\t* -i to select the data to be clustered (1:  rule conclusion, 0:  output data, default:  1)");

      printf("\n\t* -dDegradationAllowed to set performance degradation allowed by vocabulary reduction (default:  0.1)");
      printf("\n\t* -mNumberOfConc to fix the number of elements in the reduced vocabulary ");
      printf("\n\t*  (default:  incremental algorithm with a 10 percent performance degradation allowed)");
      printf("\n\t  (NB : priority to number of conclusions in vocabulary reduction)");
      printf("\n\t* -l to fuzzify the ouput (works only if vocabulary is reduced)");	
      printf("\n -------> FILES CONFIGURATION");						
      printf("\n\t* -pTestFile the data file name for the second pass");
      printf("\n\t  (default:  the same than for the 1st pass)");
      printf("\n\t* -cOutputFileName final fis file name (default:  'DataFileName'.fis)");
      printf("\n\t* -r to remove temporary files ('data'.mat and summary.ols)");
      printf("\n -------> DISPLAY TYPE");
      printf("\n\t* -a for intermediate display (default:  no display)");
      printf("\n --------> PERF OPTIONS");
      printf("\n\t* -btestfile used to compute performance (default:  'DataFileName')");
 printf("\n\t* -e activity threshold to compute performance (default:  0.2)");
 printf("\n");
      return 1 ;    
    }
  
  int i; 
  char * FisFile = NULL;
  char * TestFile = NULL;
  char * FinalFisFile = NULL;
  char * perftestfile = NULL;
  char * conj = NULL;

  double Tol=0.25;
  double PerfDeg = 0.1;
  double GlobExVar = 0.01;

  int NumOfRules = 10000;
  int NumOfOutputs = 1;
  int OutNum = 1;
  int StrongPart = 1;
  int ReducedVoc = 0;
  int NbOfConc = 0;
  int Fuzzify = 0;
  int OriginalAlgo = 0;
  int TolChange = 0;
  int Remove = 0;
  int dsp = 0;
  int opti = 0;
  int dType = 1;

  FISOLS *S=NULL;
  FILE *f;

  double pf, coverage, maxerror, muperf = 0.2;

  try
    {
      for (i = 2; i < argc; i++)
	{
	  if (!strncmp("-p",argv[i],2)) TestFile = argv[i]+2;
	  else if (!strncmp("-f",argv[i],2)) FisFile = argv[i]+2;
	  else if (!strncmp("-t",argv[i],2)) 
	    {
	      Tol = atof(argv[i]+2);
	      TolChange = 1 ;
	    }
	  else if (!strncmp("-o",argv[i],2)) NumOfOutputs  = atoi(argv[i]+2);
	  else if (!strncmp("-n",argv[i],2)) OutNum = atoi(argv[i] + 2);
	  else if (!strncmp("-s",argv[i],2)) StrongPart = 0;
	  else if (!strncmp("-v",argv[i],2)) ReducedVoc = 1;
	  else if (!strncmp("-d",argv[i],2)) PerfDeg = atof(argv[i] + 2);
	  else if (!strncmp("-m",argv[i],2)) NbOfConc = atoi(argv[i] + 2);
	  else if (!strncmp("-l",argv[i],2)) Fuzzify = 1;
	  else if (!strncmp("-c",argv[i],2)) FinalFisFile = argv[i]+2;
	  else if (!strncmp("-g",argv[i],2)) GlobExVar = atof(argv[i]+2);
	  else if (!strncmp("-i",argv[i],2)) dType = atoi(argv[i]+2);
	  else if (!strncmp("-q",argv[i],2)) NumOfRules = atoi(argv[i]+2);
	  else if (!strncmp("-j",argv[i],2)) OriginalAlgo = 1;
	  else if (!strncmp("-r",argv[i],2)) Remove = 1;
	  else if (!strncmp("-a",argv[i],2)) dsp = 1;
	  else if (!strncmp("-x",argv[i],2)) opti = 1;
	  else if (!strncmp("-b",argv[i],2)) perftestfile = argv[i] +2;
	  else if (!strncmp("-e",argv[i],2)) muperf = atoi(argv[i]+2);
	  else if (!strncmp("-h",argv[i],2)) conj = argv[i]+2;
	} 
      // intialisatin, using get and set functions (see fisols.h)
      if(perftestfile==NULL) perftestfile=argv[1];

      if(FisFile!=NULL)
	{
	  S=new FISOLS(argv[1],FisFile, opti);
	  S->SetNumS(OutNum-1);
	}
      else
	{
	  if(OriginalAlgo && (TolChange == 0)) Tol=0.05;
	  S=new FISOLS(argv[1]);
	  S->SetUniqTol(Tol);
	  S->SetNumS(OutNum-1);
	  S->SetNbOfOut(NumOfOutputs);
	  S->SetOriginalGauss(OriginalAlgo);
	  S->SetStd(StrongPart);
	  S->SetDisplay(dsp);
	  S->GenerateFIS();
	
	}
      S->SetDisplay(dsp);
      S->SetLoss(GlobExVar);
      S->SetNbOfRules(NumOfRules);
      S->SetVocRed(ReducedVoc);
      S->SetAllowedLoss(PerfDeg);
      S->SetNVocItem(NbOfConc);
      S->SetVocData(dType);
      S->SetFuzzyOut(Fuzzify);
      S->SetRmTmpFile(Remove);
      S->SetTestFileName(TestFile);
      S->SetResultFileName(FinalFisFile);
      if(conj) S->SetConjunction(conj);
 
      S->Run();
      // computing performances and printing result.ols 
      // use the -b and -e options
      // These options are not needed if this part is absent (see fis.h for perfRB details)
      pf = S->Performance(OutNum-1, perftestfile, coverage , maxerror, muperf, "perf.ols", false );
      f = fopen("result.ols", "wt");
      S->WriteHeaderPerfRB(OutNum-1, f);
      fprintf(f, "\n");
      S->PerfRB(pf, coverage, maxerror, OutNum-1, f);
      fprintf(f, "\n");
      fclose(f);
		         
      delete S;

      return 1;
    }
  catch(std::exception &e)
    {
      printf("\n%s", e.what());
    }
  
}

//**************************  Fin de MainOls.cpp ************************
