//****************************************************************
//
//                       OLS.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC
// FISPRO Version : 3.1 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Contact : fispro@supagro.inra.fr
// Last modification date:December 1, 2007

//****************************************************************

#include "ols.h"

void PrintResult(int *index, double *erreur, int nsel, const char *Name)
//****************
{
  FILE * f;
  char *fi;
  int i;
  double errcum=0;

  fi= new char [strlen(Name) + 10];
  sprintf(fi, "%s.ols", Name);

  if((f = fopen(fi,"wt")) == NULL)
    {
      sprintf( ErrorMsg, "~CannotCreateFile~: %s\n~", fi);
      throw std::runtime_error( ErrorMsg );
    }
  //fourth column : rule number, sample index in data file, rule explained variance, cumulated explained variance
  fprintf(f, "Num\tIndex\tVarExp\t\tVarCum\n\n");
  for(i=0;i<nsel;i++)
    {
      errcum=errcum+erreur[i];
      fprintf(f, "%i%c\t%i%c\t%f%c\t%f%c\t\n",i+1,SEPARE,index[i]+1,SEPARE,erreur[i],SEPARE,errcum,SEPARE);
    }

  fclose(f);
  delete [] fi;

}

int MaxIndex(double a[], int size)
//*******************************
{
    int maxIndex = 0;

    for (int i=1; i<size; i++) {
        if (a[i] > a[maxIndex]) {
            maxIndex = i;
        }
    }
    return maxIndex;
}

void ReadMatrix(double ** N, int nl,int nc, char *Fi)
//***************************************************
{

  int i, j, bufsize;
  double *tmp;
  char *buf;

  ifstream f(Fi);

  bufsize = MaxLineSize(f);


  buf = new char [bufsize];
  tmp = new double [nl];

  if(f.bad())
    {
      sprintf( ErrorMsg, "~CannotOpenFile~: %s\n~", Fi);
      throw std::runtime_error( ErrorMsg );
    }

  for(i = 0; i < nl; i ++)
    {
      f.getline(buf, 20 * nc -1);
      SearchNb(buf, tmp, nc);
      for(j = 0; j < nc; j ++) N[i][j]= tmp[j];
    }

  delete [] buf;
  delete [] tmp;
}

gsl_matrix * Dbl2Mat ( double **T ,int nl ,int nc)
//************************************************
{
  gsl_matrix * m = gsl_matrix_alloc(nl,nc);
  //  gsl_vector * v1 = gsl_vector_alloc(nc);
  gsl_vector * v1; // will be allocated in Dbl2Vec
  //first make a line vector, then put it in the matrix
  int i;
  for (i=0;i<nl;i++)
    {
      v1 = Dbl2Vec(T[i],nc);
      gsl_matrix_set_row(m,i,v1);
      gsl_vector_free(v1);
    }

  return m;
}

gsl_vector * Dbl2Vec(double * p,int s)
//************************************
{
  gsl_vector * v = gsl_vector_alloc(s);
  int i;
  //set each vector value
  for (i=0;i<s;i++)
    gsl_vector_set(v,i,p[i]);
  return v;
}

double ** Mat2Dbl (gsl_matrix * n)
//********************************
{
  int l=n->size1;
  int c=n->size2;

  gsl_vector * tab = gsl_vector_alloc(c);

  double ** q;
  int i,j;

  q=new double * [l];
  for (i=0;i<l;i++)
    q[i]=new double [c];

  //get line then line elements
  for(i=0;i<l;i++)
    {
      gsl_matrix_get_row(tab,n,i);
      for (j=0;j<c;j++)
	q[i][j]=gsl_vector_get(tab,j);
    }
  gsl_vector_free(tab);
  return q;
}


void ols(double **  Mat, double *  Vect, double seuil, int nregle, int ncol,int *& IndexSel, double *& Theta, int & nsel, const char *Name)
//**************************************************************************

{

  // Data conversion to GSL type
  gsl_matrix * P= Dbl2Mat(Mat,ncol,ncol); // P=WA
  gsl_vector * Y= Dbl2Vec(Vect,ncol);

  // initialzsation
  double element=0, norme2=0, inertie=0, norme2w=0, aij=0, alphajkk=0, varexp=0;
  int i=0, j=0, k=0, kk=0, indice=0;
  int *candidats, *permu;
  double *errex, *errext;

  gsl_matrix * w = gsl_matrix_calloc(ncol,ncol);
  gsl_matrix * A = gsl_matrix_alloc(ncol,ncol); //Alphajkk Matrix
  gsl_matrix_set_identity(A);
  gsl_matrix * W = gsl_matrix_calloc(ncol,ncol); // W orthogonal matrix
  gsl_vector * g =gsl_vector_calloc(ncol); // g=w'd/w'w
  gsl_vector * gg = gsl_vector_calloc(ncol); //gg=max(g)
  gsl_vector * d = gsl_vector_calloc(ncol); // d=Y norm
  gsl_vector_memcpy(d,Y); //Outputs copy

  // 3 temporary vectors
  gsl_vector * tmp = gsl_vector_calloc(ncol);
  gsl_vector * vect = gsl_vector_calloc(ncol);
  gsl_vector * add_tmp =gsl_vector_calloc(ncol); //temporary vector used for Orthogonalization

  errext = new double [ncol]; // errext=max(errex) --> Variance explained by the selected rules
  permu = new int [ncol]; // Indexes of the selected rules
  candidats= new int[ncol]; // Rules that have not already been selected (set to -1 if selected)
  errex = new double[ncol]; //errex = g.gw'w/d'd --> temporary vector of explained variance
  //init
  for (k=0;k<ncol;k++)
    {
      errex[k]=0.0;
      errext[k]=0.0;
    }

  for (k=0;k<ncol;k++)
    candidats[k]=k; //candidates initialization

  // Inertie computation
  gsl_blas_ddot(Y,Y,&inertie); // Vector product


  //*************** general loop (with stop condition)*********
  nsel = 0; //initialization of the number of selectetd rules

  // correction bch - bug ols 24 september 2009
  for(kk=0;kk<ncol;kk++)
  //for(kk=0;kk<ncol-1;kk++)
    {

      //************ loop on the candidates ***********
      for (i=0;i<ncol;i++)
	{
	  if (candidats[i]==DEJA_SELECTIONNE)  continue;//avoid selected candidates

	  gsl_matrix_get_col(tmp,P,i);
	  gsl_matrix_set_col(w,i,tmp);
	  gsl_vector_set_zero(add_tmp);
	  //******* loop on j : vector orthogonalization*******
	  for (j=0;j<=kk-1;j++)
	    {

	      // w coefficient computation
	      gsl_matrix_get_col(vect,W,j);

	      gsl_blas_ddot(vect,tmp,&alphajkk); // vector product
	      gsl_blas_ddot(vect,vect,&norme2);

	      if (fabs(norme2)>EPSILON)
		alphajkk=alphajkk/norme2;
	      else
		alphajkk=0;

	      gsl_vector_scale(vect,alphajkk); // vector product with a constant
	      gsl_vector_add(add_tmp,vect);
	    }

	  gsl_vector_sub(tmp,add_tmp); // vectors substraction (result in tmp)
	  gsl_matrix_set_col(w,i,tmp);

	  // g and errex coefficient computation
	  gsl_matrix_get_col(vect,w,i);
	  gsl_blas_ddot(vect,vect,&norme2w);
	  gsl_blas_ddot(vect,d,&element);

	  if (fabs(norme2w)>EPSILON)
	    element=element/norme2w;
	  else element = 0;

	  gsl_vector_set(g,i,element);

	  errex[i]=element*element*(norme2w/inertie);

	}
      //************* loop i end  *****************

      indice = MaxIndex(errex,ncol);//best error correction index
      
      if(errex[indice]<EPSILON) break; // Modif 20-12-2017 SG
      
      gsl_matrix_get_col(tmp,w,indice);
      gsl_matrix_set_col(W,kk,tmp); // W update
      gsl_vector_set(gg,kk,gsl_vector_get(g,indice)); // g update
      errext[kk]=errex[indice]; //errext update (contains the explained variance of the selectetd candidates

      permu[kk]=candidats[indice];

      // Elements who are already selected are set to -1
      candidats[indice]=DEJA_SELECTIONNE; // candidates
      errex[indice]=DEJA_SELECTIONNE;  //exaplined variance

      varexp+=errext[kk]; // Cumulative explained variance (for stop condition)

      // A computation (Upper triangle, Unit diagonal)
      for (j=0;j<=kk-1;j++)
	{
	  indice=(int)permu[kk];

	  gsl_matrix_get_col(vect,W,j);
	  gsl_blas_ddot(vect,vect,&norme2);
	  gsl_matrix_get_col(tmp,P,indice);
	  gsl_blas_ddot(vect,tmp,&aij);

	  if (fabs(norme2)>EPSILON)
	    aij=aij/norme2;
	  else
	    aij=0;

	  gsl_matrix_set(A,j,kk,aij);


	}

       if((1-varexp-seuil)<EPSILON) break;//stop condition
       else if(kk==(nregle-1)) break;

    }
  // **************** general loop end *******************


  //nsel = kk + 1; // number of selected rules update
  nsel = kk; // number of selected rules update


  PrintResult(permu,errext,nsel, Name);

 // outputs vector allocation
 Theta=new double [nsel];
 IndexSel=new int [nsel];

 gsl_matrix_free(W);
 gsl_matrix_free(w);
 gsl_vector_free(g);
 gsl_vector_free(vect);
 gsl_vector_free(add_tmp);
 delete [] errext;
 delete [] errex;
 delete [] candidats;



 // IndexSel update
 for (j=0;j<nsel;j++)
   {
     IndexSel[j]=(int)(permu[j]);
   }
 // Theta computation
 gsl_vector_memcpy(tmp,gg);
 gsl_blas_dtrsv(CblasUpper,CblasNoTrans,CblasUnit,A,tmp);

 for (j=0;j<nsel;j++)
   {
     Theta[j]=gsl_vector_get(tmp,j);
   }

 gsl_matrix_free(A);
 gsl_matrix_free(P);
 gsl_vector_free(gg);
 gsl_vector_free(d);
 gsl_vector_free(Y);
 gsl_vector_free(tmp);

 delete [] permu;

} // End of ols()

//***************************  OLSBASE.CPP  *******************************

void ols2(double ** M,double * v, int nl,int nc, double *& Theta)

{

  //initialization
  gsl_matrix * P=Dbl2Mat(M,nl,nc);
  gsl_vector * Y=Dbl2Vec(v,nl);

  gsl_matrix * w = gsl_matrix_calloc(nl,nc);
  gsl_vector * g =gsl_vector_calloc(nc);
  gsl_matrix * A = gsl_matrix_alloc(nc,nc);
  gsl_matrix_set_identity(A);
  gsl_vector * b =gsl_vector_alloc(nl);

  // temporary vectors
  gsl_vector * tmp =gsl_vector_calloc(nl);
  gsl_vector * tmp2 =gsl_vector_calloc(nl);

  int k=0,i=0;
  double norme=0, elmtg=0, elmtA=0;

  // W(k) and g(k) computation
  // conclusions least-square optimisation : g is the final vector, and theta the final array containing the conclusions
  for (k=0;k<nc;k++)
    {
      gsl_matrix_get_col(tmp2,P,k);
      gsl_vector_set_zero(b);
      for(i=0;i<k;i++)
	{
	  gsl_matrix_get_col(tmp,w,i);
	  gsl_blas_ddot(tmp,tmp,&norme);
	  gsl_blas_ddot(tmp,tmp2,&elmtA);

	  if (fabs(norme)>EPSILON)
	    elmtA=elmtA/norme;
	  else
	    elmtA=0;

	  gsl_matrix_set(A,i,k,elmtA);
	  gsl_vector_scale(tmp,elmtA);
	  gsl_vector_add(b,tmp);
	}

      gsl_vector_sub(tmp2,b);
      gsl_matrix_set_col(w,k,tmp2);
      gsl_blas_ddot(tmp2,tmp2,&norme);

      gsl_blas_ddot(tmp2,Y,&elmtg);
      if (fabs(norme)>EPSILON)
	elmtg=elmtg/norme;
      else
	elmtg=0;
      gsl_vector_set(g,k,elmtg);
    }

   gsl_blas_dtrsv(CblasUpper,CblasNoTrans,CblasUnit,A,g);


   Theta=new double [nc];
   for (i=0;i<nc;i++)
     {
    Theta[i] =(gsl_vector_get(g,i));
     }

 gsl_matrix_free(w);
 gsl_vector_free(g);
 gsl_matrix_free(A);
 gsl_vector_free(b);
 gsl_vector_free(tmp);
 gsl_vector_free(tmp2);
 gsl_matrix_free(P);
 gsl_vector_free(Y);
}
