//***********************************************************************
//
//
//                              OLS.H
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC
// FISPRO Version : 3.1 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Contact : fispro@supagro.inra.fr
// Last modification date:December 1, 2007
//***********************************************************************


#ifndef __OLSBASE_h
#define __OLSBASE_h

//! Require The GNU GSL library in order to make the matrix operations

#include "common.h"
#include "gsl/gsl_matrix.h"
#include "gsl/gsl_blas.h"

//! DEJA_SELECTIONNE : used in the rule selection process
#define DEJA_SELECTIONNE -1

/*!
  \fn PrintResult(Index,Erreur,nsel)
  \param index array containing the row index of the selected rules
  \param erreur array containing the variance explained by each rule
  \param nsel number of rules selected
  \param Name is the data file name without the directory path
  Print a file called "Name.ols" and containing the number of the rules, the part of variance explained by them and the total variance explained
*/
void PrintResult(int * index, double * erreur, int nsel, char *Name);
/*!
  \fn max(double a[], int size)
  \param a array of values
  \param size size of the array of value
  \return the index of the amx value of an array
*/
int MaxIndex(double a[], int size);
/*!
  \fn ReadMatrix(double **N, int nl, int nc, char *Fich)
  \param N array where the values will be stocked
  \param nl number of line of the matrix
  \param nc number of columns of the matrix
  \param Fich Name of the file from which the value will be read
  read a matrix of number and put it in an array
*/
void ReadMatrix(double ** N, int nl,int nc, char *Fich);
/*!
  \fn Dbl2Mat(double **t, int nl, int nc)
  \param t the original array
  \param nl number of lines of the matrix
  \param nc number of columns of the matrix
  \return a gsl_matrix coming from the original array
  Convert en array into a gsl_matrix
*/
gsl_matrix * Dbl2Mat(double **t, int nl, int nc);
 /*!
   \fn Dbl2Vec(double *P, int s)
   \param *P Vector to be converted in gsl format
   \param s vector size
   \return a conversion of P in gsl format
*/
gsl_vector * Dbl2Vec(double * p,int s);
/*!
  \fn Mat2Dbl (gsl_matrix *n)
  \param *n gsl_matrix to be converted
  \return a double ** converted from a gsl_matrix
*/
double ** Mat2Dbl (gsl_matrix * n,int l,int c);
//!  OLS first pass
//!
//!  Function inputs :
//!             -  Mat : matrix of vectors derived from the FBF (P)
//!             -  Vect : outputs Vector
//!             -  seuil : threshold of accepted unexplained variance
//!             -  ncol : matrix number of columns and lines
//!             -  Name : File name used to create the summary file name.ols (data file without directory path)
//!
//!  Function outputs:
//!             -  IndexSel : indexes of the selected rules in the matrix Mat
//!             -  ThetaSel : vector Theta (consequent of the rules)
//!             -  kk : number of selected rules
void ols(double **  Mat, double *  Vect, double seuil, int nregle, int ncol,int *& IndexSel, double *& ThetaSel,int & kk, const char *Name);
//! OLS second pass
//!
//! Function inputs : - M : reduced matrix derived from the previously selected rules
//!                   - v : outputs vector
//!                   - nl : number of examples in the second training data set
//!                   - nc : number of columns (= number of selected rules in the first pass) of M
//!
//! Function output : - Theta : vector Theta (consequent of the rules)
void ols2(double ** M,double * v, int nl,int nc, double *& Theta);

#endif
