//**********************************************************************
//
//
//                              FISIMPLE.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISIMPLE class functions used by FISPRO, 
//        part of package simple (Simplify a FIS)
//        requires library fispro
//
//**********************************************************************

#include "fisimple.h"


//*************************************************************************
//
//
//                               COMBINATION
//
//
//*************************************************************************

char * FISIMPLE::Complete(bool wl)
 //*******************************
{
  int old, loop;
  double step;

  loop = 0;
  it = old = 1;  // 0 is used for the reference (see constructor and Prepare)
  step = EPSILON;
  StopCrit = 0;

  if(sc > 1.) GroupMerging(1., 1.);

  else
    {
      // loop on group merging
      while(step < 1.0 + EPSILON*2)
	{
	  old = it;
	  GroupMerging(step, step);
	  if (StopCrit == 4) break;
	  
	  if(!wl)
	    printf("\nIt: %3d  Old: %3d  Loop: %3d step: %f", it, old, loop, step); 
	  // Increment step
	  loop ++;
	  step = STEP_LOSS * loop;
	}
    }

  if(Display)
    printf("\nRule: %d   Var: %d  CurPI: %f  MaxPI: %f  CurCov: %f", RuleRemove, VarRemove, CurPi, MaxPi, CurCov);

  // loop on rule and/or variable removal
  step = EPSILON;
  loop = 0;
  if (RuleRemove||VarRemove)
    while(CurPi < MaxPi &&  step < 1.0)
      {  
	old = it;
	if(RuleRemove) RuleRemoval(step);
	if(VarRemove) VariableRemoval(step);

	// Increment sLoss
	loop ++;
	step = STEP_LOSS * loop;
	if(!wl)
	  printf("\nIt: %3d  Old: %3d  Loop: %3d sLoss: %f", it, old, loop, step); 
      }
 
  if(Display)
    printf("\nRule: %d   Var: %d  CurPI: %f  MaxPI: %f  CurCov: %f", RuleRemove, VarRemove, CurPi, MaxPi, CurCov);

  if(!strcmp(fConfig, fSimple)) return NULL;
  return fSimple;
}


//*************************************************************************
//
//
//                        THREE SIMPLIFICATION LEVELS
//
//
//*************************************************************************

int FISIMPLE::GroupMerging(double loss, double het)
  //***********************************************
{
  int arret, RelaxConstraint, NoFusion, Mergingdone;
  double ErreurCourante, ErreurRef, Cover, ErrMaxi; 
  
  //static int k;  
  // k is the real number of iterations. It corresponds and to maximum number of ANY
  // labels in a rule description.

  arret = false;
  RelaxConstraint = false;
  NoFusion = false;
  ErreurRef = InitPi;
  ErreurCourante = CurPi; 
  Cover = ErrMaxi = -1.;

  while(arret == false)
    {
      if(NoFusion == false) RelaxConstraint = false;
      else NoFusion = false;

      Mergingdone=false;

      // Group identification
      if(CreateGroups(RelaxConstraint) < 0) return -1;
      if((G.size() == 0) && (RelaxConstraint == false)) 
	{
	  RelaxConstraint = true; // Relax the distance constraint 
	  if(CreateGroups(RelaxConstraint) < 0) return -1; 
	}
      
      if(G.size() == 0)  // No more merging possibilities
	{
	  break; 
	  arret = true;
	  StopCrit = 4;
	}

      if(TestGroupMerging(het) == NO_FUSION) 
	{
	  NoFusion = true;
	  if(RelaxConstraint == true) arret = true;
	  else RelaxConstraint = true;
	  Mergingdone=false;
	  if(RelaxConstraint && loss >= 0.95) StopCrit = 2;
	}
      else
	{
	  RemoveUnvalidGroups();
	  OrderGroupsByPi();
	  Mergingdone = MakeValidMergings(loss);
	  if(RelaxConstraint && Mergingdone== false && loss >= 0.95) 
	    StopCrit = 3;
	}

      WriteFis(it);
      ResetSave();
      RemoveGroups();

      if(NoFusion==false && Mergingdone!=false) 
	{
	  ComputePI(ErreurRef, ErreurCourante, Cover, ErrMaxi);
	  
	  CurPi = ErreurCourante;
	  CurCov = Cover;
	  StoreResult(ErreurCourante, Cover, ErrMaxi, fResult);
	  sprintf(fSimple, "%s.%d", fConfig, it);
	  if(Display) printf("\nIteration %i : Group Merging",it);
	  it++;
	}

      if(NoFusion==false && Mergingdone==false)
	{
	  arret=true;
	}
    }

return it;
} // End of GroupMerging(double loss)

int FISIMPLE::RuleRemoval(double loss)
  //**********************************
{
  int i, j, index;
  double ErreurCourante, ErreurRef, VarRel, VarCov, Cover, ErrMaxi, conc; 

  index = -1;
  Cover =  ErrMaxi = -1.;
  ErreurRef=InitPi;

  for(i = 0; i < NbSave; i++)
    {
      if(KeepLast)
	{
	  conc = Save[i]->GetAConc(NSort);
	  for(j = 0; j < Cla; j++)
	    if(RuleLab[j] == conc) index = j;
	  // Do not remove the last one 
	  if((index > -1) && (RuleClass[index] == 1)) continue; 
	}
      Save[i]->Deactivate();
      WriteFis(it);
      VarRel = ComputePI(ErreurRef, ErreurCourante, Cover, ErrMaxi);
      VarCov = (InitCov-Cover)/InitCov;

      if((VarRel > se*loss) || ( ErreurCourante > MaxPi) || ( (1. - Cover) > PropBlank) || (VarCov > CovLoss))
	{
	  Save[i]->Activate();
	  WriteFis(it);
	}
      else 
	{
	  // The rule is removed
	  if(KeepLast && (index > -1)) (RuleClass[index]) --;
	  StoreResult(ErreurCourante, Cover, ErrMaxi, fResult);
	  CurPi = ErreurCourante;
	  CurCov = Cover;
	  sprintf(fSimple, "%s.%d", fConfig, it);
	  if(Display) printf("\nIteration %i : Rule removal",it);
	  it++;
	}
    }
  ResetSave();

  return it;
}

int FISIMPLE::VariableRemoval(double loss)
  //**************************************
{
  int i, j, v, pos;
  double VarRel, ind, ErreurCourante, ErreurRef, Cover, ErrMaxi; 
  FISFPA * Ff;

  char *ch;

  ch = new char [strlen(fConfig) + 10];

  v = 0;
  Cover = ErrMaxi = -1.;
  VarRel = 0.;
  ErreurCourante = CurPi;
  ErreurRef = InitPi;

  for(i = 0; i < NbSave; i++)
    {
      if(! Save[i]->IsActive() ) continue;

      for(j = 0; j < nbI; j++)
	{
	  if(Save[i]->GetAProp(v, j) == ANY) continue;
	  if(AllAny(Save[i], j)) continue;

	  Save[i]->SetAProp(ANY, j);

	  pos = Exists(Save[i]);
	  if((pos > -1) && (pos < i)) 
	    {
	      Save[i]->SetAProp(v, j);
	      continue;
	    }
	  sprintf(ch, "%s.%d", fConfig, it);
	  WriteFis(it);
	  ResetSave();

	  Ff = new FISFPA(ch, fApp, DEC, 1, 0.02);
	  ind = Ff->FpaARule(i, NSort);
	  delete Ff;

	  if(ind < sc)    // OK for output homogeneousness
	    {
	      VarRel = ComputePI(ErreurRef, ErreurCourante, Cover, ErrMaxi);
	      if(( VarRel > se*loss) || (ErreurCourante > MaxPi))
		// Not good for performance 
		{
		  Save[i]->SetAProp(v, j);  // restoration     
		  WriteFis(it-1);
		  ResetSave();
		}
	      else
		{
		  StoreResult(ErreurCourante, Cover, ErrMaxi, fResult);
		  CurPi = ErreurCourante;
		  CurCov = Cover;
		  sprintf(fSimple, "%s.%d", fConfig, it);
		  if(Display) printf("\nIteration %i : Variable removal",it);
		  it++;
		}
	    }
	  else // restore previous state
	    {
	       Save[i]->SetAProp(v, j);  // restoration     
	       WriteFis(it-1);
	       ResetSave();
	    }
	}
    }

  ResetSave();

  delete [] ch;
  
  return it;
} // End of VariableRemoval()


//*************************************************************************
//
//
//                               Gestion GROUP
//
//
//*************************************************************************

int FISIMPLE::CreateGroups(int relax)
  //*********************************
  // Return: the number of groups
  // relax:  if false, only the pairs of rule whose distance is 1, without ANY labels 
  // are considered, else at most 2 partial distance with ANY are accepted.
 
{
  int i, j, dc, num, NbIndiff, ng;      
  list<GROUP *>::iterator u;

  ng = 0;
  for(i = 0; i < NbSave; i++)
    {
      if(! Save[i]->IsActive()) continue;
      for(j = i + 1; j < NbSave; j++)
	{
	  if(! Save[j]->IsActive()) continue;
	  dc = Save[i]->Distance(Save[j], num, NbIndiff) ;

	  // An interesting pair of rules
	  if( ((dc == 1) && (NbIndiff == 0))
	      // All labels are the same except for one variable
	      // x x A 
	      // x x B
	      || ((relax) && (dc == 1) && (NbIndiff == 1)) 
	      // All the labels are the same except for two variables. 
	      // They are different for one of them and one is ANY for the other.
	      // A x C 
	      // B x 0
	      || ((relax) && (dc == 0) && (NbIndiff < 3)) 
	      // All the labels are the same except for at most 
	      // two variables whose labels are ANY in one of the rules
	      // A x x  Or  A x B   Or x A 0
	      // 0 x x      0 x 0      x 0 B
	      )
	    {
	      // Exists the corresponding group?
	      if((u = ExistingGroup(j, num)) != G.end())
		// Yes? Add rule j to the list
		(*u)->add(j);
	      
	      // No? Create new group with the pair
	      else 
		{
		  G.push_back(new GROUP(i, j, num));
		  ng ++;
		}
	    }
	}
    }
  return ng;
}

int FISIMPLE::MakeValidMergings(double loss)
  //****************************************
{
  int mergingmade;
  list<GROUP *>::iterator p,q;
  double perfloss, perfcompa;
  double Cov, Err;
  
  FISFPA * Ff;
  FILE * fc; 
  mergingmade=false;


  for(p=G.begin();p!=G.end();p++)
    {
      q=++p;
      p--;
      UpdateSave(G.begin(),q);
      fc = fopen("temp.fis", "wt");
      if(!fc)
	{
	  sprintf( ErrorMsg, "~CannotOpenFile:%s~", "temp.fis");
	  throw std::runtime_error( ErrorMsg );
	}
      PrintCfgFis(fc);
      fclose(fc);
      
      Ff = new FISFPA("temp.fis", fApp, DEC, 1, 0.02);
      
      perfcompa=Ff->Perf(NSort, Data, NbEx, Cov, Err, MuMin, MisClass, RuleLab);     
      if(Cov < (InitCov-EPSILON))
	{
	  sprintf( ErrorMsg, "~LossOfCoverage~ in MakeValidMergings%f", Cov);
	  throw std::runtime_error( ErrorMsg );
	}
  
	
      perfloss = (perfcompa-InitPi)/InitPi;
     
      delete Ff;
      
      ResetSave();
      
      if(perfloss < se*loss && mergingmade==false)
	mergingmade=true;	 

      if(perfloss > se*loss  && mergingmade==true)
	{
	  UpdateSave(G.begin(),p);
	  return 1;
	}
      if(perfloss > se*loss  && mergingmade==false)
	return 0;
    }
  // Remaining case:  perfloss < se*loss and no more groups to merge
  if(mergingmade==true)
    {
      UpdateSave(G.begin(),G.end());
      return 1;
    }
  else return 0;
}

int FISIMPLE::RemoveUnvalidGroups(void)
  //***********************************
{
  list<GROUP *> Temp;
  list<GROUP *>::iterator q;
  // transfer valid groups from G to Temp
  while(!G.empty())
    {	
      q = G.begin();
      if ( (*q)->fusion ) 
	Temp.push_front((*q));//valid group
      else
	delete (*q);//delete invalid groups-removal done here - bch July 20, 2009
      G.erase(q);
    }      
  // correction memory leak-bch July 20, 2009
  // next two lines were never reached as there was nothing left in G
  /*  
      for(q = G.begin(); q != G.end(); q++)  
      delete (*q);
  */
  // retransfer Temp into G
  G.clear();  
  G=Temp;
  return 0;
}

int FISIMPLE::OrderGroupsByPi(void)
  //*******************************
{
  double maxvalue;
  list<GROUP *> Temp;
  list<GROUP *>::iterator q;
  while(!G.empty())
    {	
      q = G.begin();
      maxvalue=(*q)->perfloss;
      for(q = G.begin(); q != G.end(); q++)
	if((*q)->perfloss > maxvalue) maxvalue=(*q)->perfloss;	  
      for(q = G.begin(); q != G.end(); q++)
	if(fabs((*q)->perfloss - maxvalue) < EPSILON) break;
      Temp.push_front((*q));
      G.erase(q);
    }
  for(q = G.begin(); q != G.end(); q++)     delete (*q);
  G.clear();  
  G=Temp;
  return 0;
}


list<GROUP *>::iterator FISIMPLE::ExistingGroup(int r, int var)
  //***********************************************************
{
  list<GROUP *>::iterator p;
  int num, dist, ind;

  for(p = G.begin(); p != G.end(); p++)
    {
      dist = Save[(*p)->r[0]]->Distance(Save[r], num, ind);
      if((var == num) && (var == (*p)->n)) 
	if(((dist == 1) && (ind < 2)) || ((dist == 0) && (ind < 3)))
	  {
	    dist = Save[(*p)->r[1]]->Distance(Save[r], num, ind);
	    if((var == num) && (((dist == 1) && (ind < 2)) || ((dist == 0) && (ind < 3))))
		return p;
	  }
    }
   return p; // G.end() when the group does not exist
}

int FISIMPLE::TestGroupMerging(double het)
  //**************************************
{
  int fusion, vIn, rulenum;      
  list<GROUP *>::iterator p;
  double ind, vOut, perfcompa, Cov, Err;
  FISFPA * Ff;
  FILE * fc;

  vIn = 0;         // Save removed input label 
  vOut = 0.;       // Save removed output value
  fusion = 0;      // nb of fusion

  for(p = G.begin(); p != G.end(); p++)    
    {
      if(BuildRule((*p), vIn, vOut)) continue;
 
      fc = fopen("temp.fis", "wt");
      if(!fc)
	{
	  sprintf( ErrorMsg, "~CannotOpenFile:%s~", "temp.fis");
	  throw std::runtime_error( ErrorMsg );
	}
      PrintCfgFis(fc);
      fclose(fc);

      Ff = new FISFPA("temp.fis", fApp, DEC, 1, 0.02);
      // As BuildRule invert and deactivate rules, find the correct position
      // of the generic rule in the rule base of the fis (temp.fis)
      rulenum = Ff->RulePos(Save[(*p)->r[0]]);

      ind = Ff->FpaARule(rulenum, NSort);

      // Save the generic rule output value
      (*p)->outval = Ff->Rule[rulenum]->GetAConc(NSort);

      perfcompa=Ff->Perf(NSort, Data, NbEx, Cov, Err, MuMin, MisClass, RuleLab);

      (*p)->perfloss = (perfcompa-InitPi)/InitPi;

      delete Ff;

      (*p)->indice = ind;

      if(ind < sc*het) 
	{
	  if(! KeepLast || (KeepLast && ! IsLast(*p)))
	    {
	      if(KeepLast) UpdateRuleClass(*p);
	      (*p)->fusion = true;
	      fusion ++;
	    }
	}
      RestoreGroup((*p), vIn, vOut); 
    }

  if(! fusion) return NO_FUSION;

  return 0;
}

void FISIMPLE::RemoveGroups(void)
  //*****************************
{
  list<GROUP *>::iterator p;

  for(p = G.begin(); p != G.end(); p++)     delete (*p);
  G.clear();
}

void FISIMPLE::PrintGroups(FILE *f)
  //*******************************
{
  list<GROUP *>::iterator p;

  for(p = G.begin(); p != G.end(); p++)  (*p)->Print(f);
}


int FISIMPLE::BuildRule(GROUP * g, int &OldInput, double &OldOutput)
  //****************************************************************
{
  int i, j, jj, num, ind, dist;
  RULE *tmp1=NULL, *tmp2=NULL;
  char *ch=NULL;
  int allany=false;

  j = 0;
  for(i = 0; i < S->GetNbIn(); i++)
    {
      Save[g->r[0]]->GetAProp(OldInput, i);
      if(OldInput != ANY) j++;
    }
  if(j < 2) return 1;

  ch = new char[nbI * 4 + nbO * 15 ];
  RecString(Save[g->r[0]], ch);
  tmp1 = new RULE(nbI, S->In, nbO, S->Out, S->TypeConj(), ch);
  tmp1->SetAProp(ANY, g->n);

  // Exist a generic rule which includes all the others?
  // For the rules of the group

  for(j = 1; j < g->eff; j++)
    {
      RecString(Save[g->r[j]], ch);
      tmp2 = new RULE(nbI, S->In, nbO, S->Out, S->TypeConj(), ch);
      tmp2->SetAProp(ANY, g->n);
      dist = tmp1->Distance(tmp2, num, ind);
      if(dist > 0 || ind > 1)
	{
	  printf("BuildRule - distance : %d, j=%d  ind=%d \n", dist, j, ind);
	  for(j = 0; j < g->eff; j++) Save[g->r[j]]->Print();
	  PrintGroups();
	  exit(1);
	}
      if(ind)
	{
	  if(tmp1->GetAProp(dist, num) == ANY); // Nothing to do
	  else if (tmp2->GetAProp(dist, num) == ANY) 
	    // Use this rule as the generic rule
	    {
	      // check if the generic rule is not all anys
	      jj = 0;
	      for(i = 0; i < S->GetNbIn(); i++)
		{
		  Save[g->r[j]]->GetAProp(OldInput, i);
		  if(OldInput != ANY) jj++;
		}
	      if(jj < 2)
		{
		  allany=true;
		  break;
		}
	      num = g->r[j];
	      g->r[j] = g->r[0];
	      g->r[0] = num;
	    }
	}
      delete tmp2;
    }
  delete tmp1;
  // correction bug delete -> delete [] - bch July 20 2009
  delete [] ch;

  // rules were not exchanged because the more generic was all anys.
  if (allany) return 1;

  Save[g->r[0]]->GetAProp(OldInput, g->n);
  OldOutput = Save[g->r[0]]->GetAConc(NSort);

  for(i = 1; i < g->eff; i++)  Save[g->r[i]]->Deactivate();

  Save[g->r[0]]->SetAProp(ANY, g->n);
  return 0;
}

void FISIMPLE::RestoreGroup(GROUP * g, int valI, double valO)
  //*********************************************************
{
  int i;

  for(i = 1; i < g->eff; i++)  Save[g->r[i]]->Activate();

  Save[g->r[0]]->SetAProp(valI, g->n);
  Save[g->r[0]]->SetAConc(NSort, valO);

}

int FISIMPLE::IsLast(GROUP * g)
  //***************************
{
  int i, j, last, index;
  double conc;

  last = false;
  index = -1;

  for(i = 0; i < g->eff; i++)  
    {
      conc = Save[g->r[i]]->GetAConc(NSort);
      for(j = 0; j < Cla; j++)
	{
	  if(RuleLab[j] == conc) index = j;
	  if((index > -1) && (RuleClass[index] == 1) && (g->outval != conc) ) 
	    { last = true; break; }
	}
    }
  return last;
}

void FISIMPLE::UpdateRuleClass(GROUP * g)
  //*************************************
{
  int i, j;
  double conc;

  if(! KeepLast) return;
  
  for(i = 0; i < g->eff; i++)  
    {
      conc = Save[g->r[i]]->GetAConc(NSort);
      for(j = 0; j < Cla; j++)
	if(RuleLab[j] == conc) RuleClass[j] --;
    }
  for(j = 0; j < Cla; j++)
    if(RuleLab[j] == g->outval) RuleClass[j] ++;
}

//*************************************************************************
//
//
//                               Gestion Simplification
//
//
//*************************************************************************

int FISIMPLE::AllAny(RULE *R, int f)
  //********************************
{
  int i;
  int v = 0;

  for(i = 0; i < S->GetNbIn(); i++)
    {
      if(i == f) continue;
      if(R->GetAProp(v, i) != ANY) return false;
    }

  return true;
}

int FISIMPLE::Exists(RULE *R)
  //*************************
{
  int i, j, v, vr, egal;
  v = vr = 0;
  egal = false;
  j  = 0; 
  while(egal == false && j < NbSave)
    {
      if(Save[j]->IsActive())
	{
	  egal = true;
	  for(i = 0; i < nbI; i++)
	    if(R->GetAProp(v, i) != Save[j]->GetAProp(vr, i)) 
	      {
		egal = false;
		break;
	      }
	}
      if(egal == false) j++;
    }
  if(j == NbSave) return -1;
  return j;
}


//*************************************************************************
//
//
//                               SAUVEGARDES
//
//
//*************************************************************************

void FISIMPLE::PrintCfgFis(FILE * f)
  //********************************
{
  int i, n;
  n = 0;

  for(i = 0; i < NbSave; i++)  if(Save[i]->IsActive()) n++;
  
  fprintf(f, "[System]\n");
  fprintf(f, "Name=%c%s%c\n", STRING_SEP, S->Name, STRING_SEP);
  fprintf(f, "Ninputs=%d\n", S->GetNbIn());
  fprintf(f, "Noutputs=%d\n", S->GetNbOut());
  fprintf(f, "Nrules=%d\n", n);
  fprintf(f, "Nexceptions=0\n");
  fprintf(f, "Conjunction=%c%s%c\n", STRING_SEP, S->TypeConj(), STRING_SEP);
  fprintf(f, "MissingValues=%c%s%c\n", STRING_SEP, S->MissingValues(), STRING_SEP);


  for(i = 0; i < S->GetNbIn(); i++)   S->In[i]->PrintCfg(i + 1, f);

  for(i = 0; i < S->GetNbOut(); i++)  S->Out[i]->PrintCfg(i + 1, f);

  fprintf(f, "\n[Rules]\n");
  for(i = 0; i < NbSave; i++)  if(Save[i]->IsActive()) Save[i]->PrintCfg(f);

  fprintf(f, "\n");

  fprintf(f, "\n[Exceptions]\n");

} 

void FISIMPLE::RecString(RULE * R, char * str)
  //******************************************
{
  int j, ValIn;
  char tmp[15];

  str[0] = 0; 
  ValIn = 0;
  
  for(j = 0; j < S->GetNbIn(); j++) 
    {
      sprintf(tmp, "%d%c ", R->GetAProp(ValIn, j), SEPARE);
      strcat(str, tmp);
    }
      
  for(j = 0; j < S->GetNbOut(); j++) 
    {
      sprintf(tmp, "%8.3f%c ", R->GetAConc(j), SEPARE);
      strcat(str, tmp);
    }
}

void FISIMPLE::UpdateSave(list <GROUP *>::iterator begin,list <GROUP *>::iterator end)
  //**********************************************************************************
{
  list <GROUP *>::iterator p;
  int i, k, ValIn, diff, nTemp;
  char * ch;
  RULE ** Temp;

  nTemp = 0;
  Temp = new RULE* [NbSave];
  
  ch = new char[nbI * 4 + nbO * 15 ];

  // For each group, build the generic rule from r[0]
  for(p = begin; p != end; p++)
    {
      if((*p)->indice == NOT_ENOUGH) 
	{
	  for(i = 0; i < (*p)->eff; i++)  Save[(*p)->r[i]]->Deactivate();
	  continue;
	}
      // no merging for inactive rules because deactivated in previous groups
      for(i = 0; i < (*p)->eff; i++) 
	if (!Save[(*p)->r[i]]->IsActive()) 
	  {
	    (*p)->fusion=false;
	    break;
	  }

      if(! (*p)->fusion) continue;
      if(nTemp ==  NbSave)  
	{
	  sprintf( ErrorMsg, "~too many rules generated in UpdateSave:%d~",nTemp );
	  throw std::runtime_error( ErrorMsg );//protection
	}

      RecString(Save[(*p)->r[0]], ch);      
      Temp[nTemp] = new RULE(nbI, S->In, nbO, S->Out, S->TypeConj(), ch);

      Temp[nTemp]->SetAProp(ANY, (*p)->n);
      Temp[nTemp]->SetAConc(NSort, (*p)->outval);

      // Deactivate all rules in the group
      for(i = 0; i < (*p)->eff; i++)  Save[(*p)->r[i]]->Deactivate();

      nTemp ++;
      for(k = 0; k < nTemp - 1; k++) 
	// We obtained two identical rules using two different ways
	if((Temp[nTemp-1]->Distance(Temp[k], ValIn, diff) == 0) && (diff == 0))
	  {
	    // Rule conclusions are different?
	    if(Temp[nTemp-1]->GetAConc(NSort) != Temp[k]->GetAConc(NSort))
	      {
		// Build a fis in order to call FpaArule
		FIS * F;
		FILE *fc;
		F = new FIS(fConfig);
		F->RemoveAllRules();
		F->AddRule(Temp[nTemp-1]);
		fc = fopen("temp.fis", "wt");
		if(!fc)
		  {
		    sprintf( ErrorMsg, "~CannotOpenFile:%s~", "temp.fis");
		    throw std::runtime_error( ErrorMsg );
		  }
		F->PrintCfg(fc);
		fclose(fc);
		delete F;
		FISFPA *Ff;
		Ff = new FISFPA("temp.fis", fApp, DEC, 1, 0.02);
		Ff->FpaARule(0, NSort);
		Temp[k]->SetAConc(NSort, Ff->Rule[0]->GetAConc(NSort));
		delete Ff;
	      }
	    // Remove the last one
	    delete Temp[nTemp-1];
	    nTemp --;
	  }
    }
  
  // Add isolated rules
  for(i = 0; i < NbSave; i++) 
    {
      if(! Save[i]->IsActive() ) continue;     
      if(nTemp ==  NbSave) 
	{
	  sprintf( ErrorMsg, "~too many rules generated in UpdateSave:%d~",nTemp );
	  throw std::runtime_error( ErrorMsg );//protection
	}
      
      RecString(Save[i], ch);
      Temp[nTemp++] = new RULE(nbI, S->In, nbO, S->Out, S->TypeConj(), ch);
    }
 
  //for(i = 0; i > NbSave; i++) delete Save[i];
  // ??? - correction leak memory bch - Jul 20, 2009
  // Temp has been allocated
  for(i = 0; i < NbSave; i++) delete Save[i];
  delete [] Save;
  
  Save = Temp;
  NbSave = nTemp;
  
  for(i = 0; i < NbSave; i ++)
    {
      k = Exists(Save[i]);
      if((k > -1) && (k < i)) Save[i]->Deactivate();
    }

  delete [] ch;

} // End of UpdateSave

int FISIMPLE::ResetSave(void)
  //*************************
{
  int i, nbR;
  char *ch;

  if(Save) 
    {
      for(i = 0; i < NbSave; i++)    delete Save[i];
      delete [] Save;
      Save = NULL;
      NbSave = 0;
    }

  nbR = S->GetNbRule();

  ch = new char[nbI * 4 + nbO * 15 ];
  Save = new RULE *[nbR];

  for(i = 0; i < nbR; i++)
    {
      RecString(S->Rule[i], ch);
      Save[i] = new RULE(nbI, S->In, nbO, S->Out, S->TypeConj(), ch);
    }

  NbSave = nbR;
  if(KeepLast) ResetRuleClass();

  delete [] ch;

  return 0;
}

void FISIMPLE::ResetRuleClass(void)
  //*******************************
{
  int i, c;
  double conc;


  for(i = 0; i < Cla; i++) RuleClass[i] = 0;

  for(c = 0; c < NbSave; c++)
    {
      conc = Save[c]->GetAConc(NSort);
      for (i = 0; i < Cla; i++)
	if(conc == RuleLab[i]) 
	  { (RuleClass[i])++; break; } 
    }
}

void FISIMPLE::WriteFis(int i)
  //**************************
{
  FILE * f=NULL;
  char * ch=NULL;

  ch = new char [strlen(fConfig)+10];
  sprintf(ch, "%s.%d", fConfig, i);

  f = fopen(ch, "wt");
  if(!f)
    {
      sprintf( ErrorMsg, "~CannotOpenFile:%s~", ch );
      throw std::runtime_error( ErrorMsg );
    }

  PrintCfgFis(f);
  fclose(f);
  
  delete S;
  S = new FIS(ch);
  
  delete [] ch;
}  

void FISIMPLE::ReadFis(int i)
  //*************************
{
  char * ch;

  ch = new char [strlen(fConfig)+10];
  sprintf(ch, "%s.%d", fConfig, i);
  
  delete S;
  S = new FIS(ch);
  ResetSave();

  delete [] ch;
}  

double FISIMPLE::ComputePI(double PerfRef, double & PerfCur, double & Cov, double & Err)
  //************************************************************************************
{
  double pf, variation;
  int c;

  Cov = Err = -1.;

  if(ClassifType)
    {
      if(! strcmp(S->Out[NSort]->Defuzzify(), OUT_CRISP::SugenoDefuz()))
	((DEFUZ_SugenoClassif *) (S->Out[NSort]->Def))->SetClasses(RuleLab, Cla);
      else if(! strcmp(S->Out[NSort]->Defuzzify(), OUT_CRISP::MaxCrispDefuz()))
	((DEFUZ_MaxCrisp *) (S->Out[NSort]->Def))->SetClasses(RuleLab, Cla);

      for(c = 0; c < Cla; c++) MisClass[c] = 0;
    }

  pf = S->Perf(NSort, vData, NbTestEx, Cov, Err, MuMin, MisClass, RuleLab);

  if(PerfRef > EPSILON)
    variation = (pf - PerfRef) / PerfRef; // Error (perf) variation 
  else variation = 0.;

  PerfCur = pf;

  return variation;
}

void FISIMPLE::StoreResult(double PerfCur, double Cov, double Err, char * fic)
  //**************************************************************************
{ 
  FILE * fres;

  if(!fic) 
    {
      sprintf( ErrorMsg, "~NoResultFileName~");
      throw std::runtime_error( ErrorMsg );
    }
   
  fres = fopen(fic, "at");
  if(fres == NULL) 
    {
      sprintf( ErrorMsg, "~CannotOpenFile:%s~", fic);
      throw std::runtime_error( ErrorMsg );
    }

  S->PerfRB(PerfCur, Cov, Err, NSort, fres);
  fprintf(fres, "\n");

  fclose(fres);	  
}

//*************************************************************************
//
//
//                               INITIALISATION
//
//
//*************************************************************************

FISIMPLE::FISIMPLE(const char *ffis, const char *fdat, const char *fres, double c, double e, int ns, double mmin, double pblank, double cov, int kl, double me)
{

  FILE *f;
  S = new FIS(ffis);

  Save = NULL;
  NbSave = 0; 
  Data = NULL;
  RuleClass = NULL;
  RuleLab = NULL;
  MisClass = NULL;
  Cla = 0;
  StdDev = 0.;

  sc = c; 
  se = e;
  pabs = me;
  NSort = ns;
  MuMin = mmin;

  if(cov)  // Priority to relative loss of coverage
    {
      PropBlank = 1.;
      CovLoss = cov;
    }
  else    // Priority to absolute nivel of coverage
    {
      PropBlank = pblank;
      CovLoss = 1.;
    }

  if( ! S->Out[NSort]->Classification()) sc = 1000.; 
  // Set inactive heterogeneity threshold

  KeepLast = kl;
  fApp = fdat;

  nbI = S->GetNbIn();
  nbO = S->GetNbOut();

  // Open the data file and init all the requested arrays and variables;
  InitData(); 

  // Default test data = training data
  fTest = NULL;
  vData = Data;
  NbTestEx= NbEx;

  S->SortRules(Data, NbEx, -1);

  fConfig=fSimple=NULL;
  fConfig = new char [strlen(ffis) + 10];
  sprintf(fConfig, "%s.jb", ffis);
  f = fopen(fConfig, "wt");
  if(!f)
    {
      sprintf( ErrorMsg, "~CannotOpenFile:%s~",  fConfig);
      throw std::runtime_error( ErrorMsg );
    }
  S->PrintCfg(f);

  delete S;
  fclose(f);

  S = new FIS(fConfig);

  fResult = new char [strlen(fres) + 1];
  sprintf(fResult, "%s", fres);

  fSimple = new char [strlen(fConfig) + 10];
  sprintf(fSimple, "%s", fConfig);

  FILE *tmp;
  tmp = fopen(fResult, "wt");
  S->WriteHeaderPerfRB(NSort, tmp);
  fprintf(tmp, "\n");
  fclose(tmp);

  Prepare();

  ResetSave();  // Called before because WriteFis uses the Save rule base.
  WriteFis(0);  // WriteFis will delete S and recreate it
  ResetSave();  // Because RULE contructor uses INPUT **.


  InitIndices();
}


void FISIMPLE::InitIndices(void)
  //****************************
{
  double cc;
  cc = -1.;
  ComputePI(0., InitPi, InitCov, cc);
  StoreResult(InitPi, InitCov, cc, fResult);


  // The loss of performance is relative and some classifiers are perfect 
  // (overparametrized?)
  if(InitPi < EPSILON) 
    {
      // StdDev = 0 for classif (see InitData)
      if(StdDev > EPSILON) InitPi = RATE_ERR * StdDev; // Regression case
      else InitPi = 1.; // Classif case
    }
  CurPi = InitPi;
  if(pabs < 0)
    MaxPi = InitPi * (1. + se)  + EPSILON;
  else
    {
      MaxPi = pabs;
      if(MaxPi > InitPi) se = (MaxPi - InitPi) / InitPi;
      else se = EPSILON;
    }
}

void FISIMPLE::InitData(void)
  //*************************
{
  double *tmp, val, mean;
  int i;

  Data = ReadSampleFile(fApp, i, NbEx);
  if(i < nbI+NSort+1)
    {
      sprintf( ErrorMsg, "~NoObservedDataForOutput~ %d in file %s\n", NSort, fApp);
      throw std::runtime_error( ErrorMsg );
    }

  ClassifType = S->Out[NSort]->Classification() && 
    (! strcmp(S->Out[NSort]->GetOutputType(), OUT_CRISP::OutputType())) && 
    ( (! strcmp(S->Out[NSort]->Defuzzify(), OUT_CRISP::SugenoDefuz()))	 ||
      (! strcmp(S->Out[NSort]->Defuzzify(), OUT_CRISP::MaxCrispDefuz())));

  
  // call to InitClass function of some DEFUZ objets
  S->ClassifCheck(Data, NbEx, NSort);
  // RuleClass and RuleLab array allocation
  S->ResClassifAlloc(RuleClass, tmp, NSort);

  if(RuleClass) 
    // RuleClass is not NULL only when c is true
    {
      Cla = S->Out[NSort]->Def->NbClasses();
      RuleLab = new double [Cla];
      for(i = 0; i < Cla; i++) RuleLab[i] = tmp[i];
      MisClass = new int [Cla];
      for(i = 0; i < Cla; i++) MisClass[i] = 0;
    }      

  if(KeepLast && (! strcmp(S->Out[NSort]->GetOutputType(), OUT_FUZZY::OutputType())) )
    {
      Cla = S->Out[NSort]->GetNbMf();
      RuleClass = new int [Cla];
      for(i = 0; i < Cla; i++) RuleClass[i] = 0;
      RuleLab = new double [Cla];
      for(i = 0; i < Cla; i++) RuleLab[i] = (double) i + 1;
    }

  StdDev = 0.;

  if(! ClassifType) 
    {
      val = 0.;
      for(i = 0; i < NbEx; i++)    val += Data[i][nbI+NSort];
      mean = val / NbEx;
      val = 0.;
      for(i = 0; i < NbEx; i++)    
	val += (Data[i][nbI+NSort] - mean) * (Data[i][nbI+NSort] - mean);
      StdDev = sqrt(val / NbEx);
      if(StdDev < EPSILON)  StdDev = 0.;
      if (! strcmp(S->Out[NSort]->GetOutputType(), OUT_CRISP::OutputType()))
	//! Crisp output, regression case 
	KeepLast = false;
     }
}

void FISIMPLE::InitTestData(char *TestData)
  //***************************************
{
  int i;

  fTest=TestData;
  vData = ReadSampleFile(fTest, i, NbTestEx);
  InitIndices();
}


void FISIMPLE::Prepare(void)
  //************************
{
  int i, j;

  for(i = 0; i < S->GetNbIn(); i++)
    {
      if(S->In[i]->GetNbMf() < 2) S->In[i]->Deactivate();
      if(S->In[i]->IsActive() == false) 
	for(j = 0; j < S->GetNbRule(); j++)
	  S->Rule[j]->SetAProp(ANY, i);
    }	  

} // End of Prepare()


//*************************************************************************
//
//
//                               REMOVE TMP FILES
//
//
//*************************************************************************

void FISIMPLE::RemoveTmpFiles(char * s)
  //***********************************
{
  char * frem;
  int i;

  frem = new char [strlen(fConfig) + 4];

  sprintf(frem, "%s", fConfig);
  remove(frem);

  for (i = 0; i <= it; i++)
    {
      sprintf(frem, "%s.%d", fConfig, i);
      if(s && ! strcmp(s, frem)) continue;
      remove(frem);
    }

  remove("temp.fis");

  delete [] frem;
}

//********************* End of FISIMPLE.CPP   ************************
