//****************************************************************
//
//
//                              MAINSIMPLE.CPP
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : test simplification
// Environnment : requires library fispro
//****************************************************************


#include "fisimple.h"

int main(int argc, char ** argv)
  //****************************
{

  if(argc < 2)
    {
      printf("\nThe program %s needs 2 argument(s) : ", argv[0]);
      printf("\n\t* the fis config file name;") ;
      printf("\n\t* the data file name;") ;

      printf("\n\nOptions:");
      printf("\n\t  -t performance test file name (default:  data file name)");
      printf("\n\t  -p the output number (default:  0)");
      printf("\n\t  -s the activation threshold for an item not to be blank ");
      printf("\n\t     (default:  0.2)");
      printf("\n\t  -b absolute coverage level:  tolerated percentage of ");
      printf("\n\t     blank samples (default:  0.1, 10%%)"); 
      printf("\n\t  -c relative tolerated loss loss of coverage ");
      printf("\n\t     during rule removal (default: 0.0%%)");
      printf("\n\t   -b and -c are exclusive, priority is given to -b if -c=0.");
      printf("\n\t   except when -c is positive"); 
      printf("\n\t   default = priority to absolute coverage level");
      printf("\n\t  -d absolute maximum performance index (default:  -1.0) ");
      printf("\n\t  -e tolerated loss of performance, percentage of ");
      printf("\n\t     initial error (default:  0.1, 10%%)");
      printf("\n\t   -d and -e are exclusive, priority is given ");
      printf("\n\t    to -e except when -d is positive");
      printf("\n\t  -h the output homogeneity threshold, classification only ") ;
      printf("\n\t     (default:  0.8)") ;
      printf("\n\t  -k do not keep the last rule with a given class or ");
      printf("\n\t     MF label as conclusion (default:  true)"); 
      printf("\n\t  -r remove temporary files (default:  false)"); 
      printf("\n\t  -q for rule removal part");
      printf("\n\t  -l for variable removal");
      printf("\n\t  -a for detailed display");
      printf("\n\t* -wl for wordless.\n");
            printf("\n"); 						       
      return 1;
    }

  int i, numsort = 0, k_last = 1, rm_files = 0, FlagDisplay = 0, rulerem = 0, varrem = 0; 
  double seuil_h = 0.8, seuil_e = 0.1, blancs = 0.1, maxpi = -1.0, covloss = 0.0, muperf = 0.2, pf, coverage, maxerror; 
  char * perftestfile = NULL;
  bool wordless=false;
  
 
  if(argc > 3) 
    {
      for(i = 3; i < argc; i++)
	{	
	  if(! strncmp("-p", argv[i], 2) )  numsort = atoi(argv[i] + 2);
	  else if(! strncmp("-s", argv[i], 2) )  muperf = atof(argv[i] + 2);
	  else if(! strncmp("-b", argv[i], 2) )  blancs = atof(argv[i] + 2);
	  else if(! strncmp("-c", argv[i], 2) )  covloss = atof(argv[i] + 2);
	  else if(! strncmp("-d", argv[i], 2) )  maxpi = atof(argv[i] + 2);
	  else if(! strncmp("-e", argv[i], 2) )  seuil_e = atof(argv[i] + 2);
	  else if(! strncmp("-h", argv[i], 2) )  seuil_h = atof(argv[i] + 2);
	  else if(! strncmp("-a", argv[i], 2) )  FlagDisplay = 1;
	  else if(! strncmp("-q", argv[i], 2) )  rulerem = true;
	  else if(! strncmp("-l", argv[i], 2) ) varrem = true;
	  else if(! strncmp("-t", argv[i], 2) )  perftestfile = argv[i] + 2;
	  else if(! strncmp("-k", argv[i], 2) )  k_last = true;
	  else if(! strncmp("-r", argv[i], 2) ) rm_files = true;
	  else if(! strncmp("-wl", argv[i], 3) ) wordless = true;
	}
    }
if(wordless)
  FlagDisplay=0;

      
  FISIMPLE *S = NULL;
  FIS *T = NULL;
  char * simple;
  FILE *f;

  try
    {
      S = new FISIMPLE(argv[1], argv[2], "result.simple", seuil_h, seuil_e, numsort, muperf, blancs, covloss, k_last, maxpi);
      S->SetDisplay(FlagDisplay);
      S->SetRuleRemove(rulerem);
      S->SetVarRemove(varrem);
      if(perftestfile != NULL) S->InitTestData(perftestfile);
      
      simple = S->Complete(wordless);
      
      if (rm_files) S->RemoveTmpFiles(simple);

      if(! simple) 
	{
	  if(!wordless)
	    printf("\n~NoSimplicationAchieved~\n");
	  delete S;
	  return 1;
	}

      if(!wordless)
	printf("\nThe most simple FIS : %s\n", simple);
      
      T = new FIS(simple);
      if(perftestfile != NULL) 
	pf = T->Performance(numsort, perftestfile, coverage , maxerror, muperf, "perf.simple", false );
      else 
	pf = T->Performance(numsort, argv[2], coverage , maxerror, muperf, "perf.simple", false );

      f = fopen("result.simple", "wt");
      T->WriteHeaderPerfRB(numsort, f);
      fprintf(f, "\n");
      T->PerfRB(pf, coverage, maxerror, numsort, f);
      fprintf(f, "\n");
      fclose(f);  
    
      if(FlagDisplay)
	printf("\n Performance return: %f; Coverage level (Threshold %4.2f): %4.2f MaxError: %11.3f\n", pf, muperf, coverage, maxerror);

      delete S;
      delete T;
    }

  catch(std::exception &e)
    {
      printf("\n%s", e.what());
    }
  
  return 0;
}
