//***********************************************************************
//
//
//                              MAIN.CPP
//
//
// Author(s) : Serge Guillaume
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Contact : 
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : Test file for superfis classes
//**********************************************************************

#include <iostream>
#include <sstream>
#include <string>
using namespace std;
#include <math.h>
#include "superfis.h"

int main(int argc, char ** argv)
  //******************************
{
  if(argc < 3)
    {
      printf("\nThe program %s needs 2 arguments : ", argv[0]);
      printf("\n\t* the superfis configuration file name;");   
      printf("\n\t* the data file name.") ;

      printf("\nOption: -a to display system configuration.") ;
      printf("\n        -r to display inference results.") ;
      printf("\n" ); 
      return 1;
    }
	
  SUPERFIS * fis = NULL;
  double ** val = NULL;
  int i, j, nRows=0, nCols=0, dispc=0, dispr=0;

  if(argc > 3) 
      for(i = 3; i < argc; i++)
	{
	  if(! strcmp("-a", argv[i]) ) dispc = 1;
	  if(! strcmp("-r", argv[i]) ) dispr = 1;
	}

  val = ReadSampleFile(argv[2], nCols, nRows);

  try
    {
      fis = new SUPERFIS(val, nCols, nRows, argv[1]);	
    }
  catch ( std::exception &e )
    {
      cerr << e.what()<< endl << flush;
    }
  
  fis->Infer();

  if(dispc) fis->ToString();   // Display config

  if(dispr) // Display results
    {
      // Header
      char * hdr;
      int nnodes = (int)fis->pvclGetSNodes()->size();
      int ninputs = 0;
      for(i = 0; i < nnodes; i++) ninputs += fis->pvclGetSNodes()->at(i)->nGetNbCols();
      hdr = new char [(nnodes+ninputs)*15];
      hdr[0] = 0;

      for(i = 0; i < nnodes; i++) 
	{
	  sprintf(hdr  + strlen(hdr), "Name \t Out \t");
	  for(j = 0; j < fis->pvclGetSNodes()->at(i)->nGetNbCols(); j++)
	    sprintf(hdr + strlen(hdr), "In%d \t", j);
	}
      printf("\n%s\n", hdr);
      delete [] hdr;

      // Table
      for(j = 0; j < nRows; j++)
	{
	  for (i = 0; i < (int)fis->pvclGetSNodes()->size(); i++)
	    {
	      printf("%s\t%f", fis->pvclGetSNodes()->at(i)->pclGetFis()->Name, fis->pvclGetSNodes()->at(i)->pvdGetOutValues()->at(j));
	      
	      
	      for (int k = 0; k < fis->pvclGetSNodes()->at(i)->nGetNbCols(); k++)
		printf("\t%f", (fis->pvclGetSNodes()->at(i)->adGetInputDatas())[j][k]);
	      printf("\t");
	    }
	  printf("\n");
	}
    }
  for (i = 0; i < (int)fis->pvclGetSNodes()->size(); i++)
    delete fis->pvclGetSNodes()->at(i);

  /*delete fis->pviSelectedColumns;
    delete fis->pvclGetSNodes;*/
  
  if (fis != NULL) 
    {delete fis; fis = NULL;}
  
  for(i = 0; i < nRows; i++)  delete [] val[i];
  if(val) delete [] val;
	
  return 0;
}

