//***********************************************************************
//
//
//                              SLEAFNODE.CPP
//
//
// Author(s) : Vincent Therry
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Contact : vincent.therry@envilys.com
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : SUPERFIS LEAF NODE class functions used by FISPRO, part of library fispro

//**********************************************************************

//#include "fis.h"
#include "sleafnode.h"

SLEAFNODE::SLEAFNODE() : SNODE(){}

SLEAFNODE::SLEAFNODE(const SLEAFNODE & s) : SNODE(s){}

SLEAFNODE::SLEAFNODE(char * sFisFileName, double ** adDataArray, int nNbRows, vector <int> * pviSelectedColumns)
: SNODE(sFisFileName)
{
	char ErrorMsg[300];
	try
	{
		// Verify that all selected columns are greater than 1
		for (int i = 0; i < (int)pviSelectedColumns->size(); i++){
			if (pviSelectedColumns->at(i) < 1) 
			{
				sprintf( ErrorMsg, "~WrongValue (%d): Value must be an integer greater than 1~\n", pviSelectedColumns->at(i));
   				throw std::runtime_error( ErrorMsg );
			}
		}
		m_nNbCols = pviSelectedColumns->size();
		m_nNbRows = nNbRows;
		m_adInputDatas = new double*[m_nNbRows];
		for (int i = 0; i < m_nNbRows; i++){
			m_adInputDatas[i] = new double[m_nNbCols];
			for (int j = 0; j < m_nNbCols; j++)
			{
				// Building the internal array with selected columns of adDataArray
				m_adInputDatas[i][j] = adDataArray[i][pviSelectedColumns->at(j) - 1];
			}
		}
	}
	catch( std::exception &e )
    {
   		sprintf( ErrorMsg, "~ErrorInSLeafNodeConstructor~\n%s\n", e.what() );
   		throw std::runtime_error( ErrorMsg );
    }
}

void SLEAFNODE::ToString()
{
	SNODE::ToString();
	cout << "(I am a leaf node)" << endl;
}

//**************************    SLEAFNODE.CPP   *******************************
