//***********************************************************************
//
//
//                              SNODE.CPP
//
//
// Author(s) : Vincent Therry
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Contact : vincent.therry@envilys.com
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : SUPERFIS NODE class functions used by FISPRO, part of library fispro

//**********************************************************************

#include <iostream>
using namespace std;
#include "snode.h"

SNODE::SNODE()
{
	m_pclFis = NULL;
	m_nNbCols = 0;
	m_nNbRows = 0;
	m_pvdOutValues = new vector <double>;
	m_adInputDatas = NULL;
	m_piResClassif = NULL;
	m_pdLab = NULL;
}

SNODE::SNODE(const SNODE & s)
{
	m_pclFis = new FIS(*(s.m_pclFis));
	m_nNbCols = s.m_nNbCols;
	m_nNbRows = s.m_nNbRows;
	m_pvdOutValues = new vector <double>;
	for (int i = 0; i < (int)s.m_pvdOutValues->size(); i++)
	{
		m_pvdOutValues->push_back(s.m_pvdOutValues->at(i));
	}
	m_adInputDatas = new double *[m_nNbRows];
	for (int i = 0; i < m_nNbRows; i++){
		m_adInputDatas[i] = new double[m_nNbCols];
		for (int j = 0; j < m_nNbCols; j++)
		{
			// Building the internal array with selected columns of adDataArray
			m_adInputDatas[i][j] = s.m_adInputDatas[i][j];
		}
	}
	m_piResClassif = new int(*(s.m_piResClassif));
	m_pdLab = new double(*(s.m_pdLab));
	
}

SNODE::SNODE(char * sFisFileName)
{
	m_pclFis = new FIS(sFisFileName);
	m_nNbCols = 0;
	m_nNbRows = 0;
	m_pvdOutValues = new vector <double>;
	m_adInputDatas = NULL;
	m_piResClassif = NULL;
	m_pdLab = NULL;
}

SNODE::~SNODE()
{
	if (m_pclFis != NULL) 
	{
		delete m_pclFis;
		m_pclFis = NULL;
	}
	if (m_adInputDatas != NULL)
	{
		for(int i = 0 ; i < m_nNbRows ; i++ )
			delete [] m_adInputDatas[i];
      delete [] m_adInputDatas;
      m_adInputDatas = NULL;
	}
	if (m_pvdOutValues != NULL) 
	{
		delete m_pvdOutValues;
		m_pvdOutValues = NULL;
	}
	if (m_piResClassif != NULL) 
	{
		delete m_piResClassif;
		m_piResClassif = NULL;
	}
	if (m_pdLab != NULL) 
	{
		delete m_pdLab;
		m_pdLab = NULL;
	}
}

FIS * SNODE::pclGetFis()
{
	return m_pclFis;
}

double ** SNODE::adGetInputDatas()
{
	return m_adInputDatas;
}

int SNODE::nGetNbCols()
{
	return m_nNbCols;
}

int SNODE::nGetNbRows()
{
	return m_nNbRows;
}

vector <double> * SNODE::pvdGetOutValues()
{
	return m_pvdOutValues;
}

void SNODE::SetFis(FIS * pclFis)
{
	m_pclFis = pclFis;
}

void SNODE::SetInputDatas(double ** adInputDatas)
{
	m_adInputDatas = adInputDatas;
}

void SNODE::SetNbCols(int nNbCols)
{
	m_nNbCols = nNbCols;
}

void SNODE::SetNbRows(int nNbRows)
{
	m_nNbRows = nNbRows;
}

void SNODE::SetOutValues(vector <double> * pvdOutValues)
{
	m_pvdOutValues = pvdOutValues;
}

void SNODE::Infer()
{
	if (m_pclFis != NULL)
	{
		m_pclFis->ClassifCheck(m_adInputDatas,m_nNbRows, 0);
    	m_pclFis->ResClassifAlloc(m_piResClassif, m_pdLab, 0);
    
    	for (int i = 0; i < m_nNbRows; i++)
		{
			m_pclFis->Infer(m_adInputDatas[i], 0, NULL, 0);
			m_pvdOutValues->push_back(m_pclFis->OutValue[0]);
		} 
		delete m_piResClassif;
		m_piResClassif = NULL;
		delete m_pdLab;
		m_pdLab = NULL;
	}
}

void SNODE::ToString()
{
	cout << "******************************************************************" << endl;
	cout << "Node associated with FIS : " << m_pclFis->Name << endl;
	cout << "Datas : " << endl;
	for (int i = 0; i < m_nNbRows; i++)
	{
		for (int j = 0; j < m_nNbCols; j++)
		{
			cout << m_adInputDatas[i][j] << "\t";
		}
		cout << endl;
	}	
	cout << "Number of columns : " << m_nNbCols << endl;
	cout << "Number of lines : " << m_nNbRows << endl;
	cout << "Results : " << endl;
	for (int i = 0; i < (int) m_pvdOutValues->size(); i++)
	{
		cout << m_pvdOutValues->at(i) << "\t";
	}
	cout << endl;
}

//**************************    SNODE.CPP   *******************************
