//***********************************************************************
//
//
//                              SNODE.H
//
//
// Author(s) : Vincent Therry
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Contact : vincent.therry@envilys.com
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : SUPERSIF NODE functions used by FISPRO, part of library fispro

//**********************************************************************
#ifndef SNODE_H_
#define SNODE_H_

#include "fis.h"

class SNODE
{
	protected :
		// Pointer on the FIS contained in the node
		FIS * m_pclFis;
		
		// Data Input array
		double ** m_adInputDatas;
		
		// Informations about the array
		int m_nNbCols;
		int m_nNbRows;
		
		// Out Values
		vector <double> * m_pvdOutValues;
		
		// For classification methods
		int * m_piResClassif;
		double * m_pdLab;
	
	public :	
		// Constructors
		SNODE();
		SNODE(const SNODE & s);
		SNODE(char * sFisFileName);
		virtual ~SNODE();
		
		// Getters
		FIS * pclGetFis();
		double ** adGetInputDatas();
		int nGetNbCols();
		int nGetNbRows();
		vector <double> * pvdGetOutValues();
		
		// Setters
		void SetFis(FIS * pclFis);
		void SetInputDatas(double ** adInputDatas);
		void SetNbCols(int nNbCols);
		void SetNbRows(int nNbRows);
		void SetOutValues(vector <double> * pvdOutValues);
		
		// Methods
		virtual void Infer();
		virtual void ToString();			
};

#endif /*SNODE_H_*/
