//***********************************************************************
//
//
//                              SPARENTNODE.CPP
//
//
// Author(s) : Vincent Therry
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Contact : vincent.therry@envilys.com
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : MF class functions used by FISPRO, part of library fispro

//**********************************************************************

#include <iostream>
using namespace std;
#include "sparentnode.h"
		
SPARENTNODE::SPARENTNODE() : SNODE()
{
	m_pvclSNodes = new vector <SNODE *>;
}

SPARENTNODE::SPARENTNODE(const SPARENTNODE & s)
{
	m_pvclSNodes = new vector <SNODE *>;
	for (int i = 0; i < (int)s.m_pvclSNodes->size(); i++)
	{
		m_pvclSNodes->push_back(new SNODE(*(s.m_pvclSNodes->at(i))));
	}
}
		
SPARENTNODE::SPARENTNODE(char * sFisFileName, double ** adDataArray, int nNbRows, vector <int> * pviSelectedColumns, vector <SNODE *> * pvclSNodes)
: SNODE(sFisFileName)
{
	char ErrorMsg[300];
	try
	{
		m_pvclSNodes = new vector<SNODE *> (*pvclSNodes);
		
		// Verify that all selected columns are greater than 1
		for (int i = 0; i < (int)pviSelectedColumns->size(); i++){
			if (pviSelectedColumns->at(i) < 1) 
			{
				sprintf( ErrorMsg, "~WrongValue (%d): Value must be an integer greater than 1~\n", pviSelectedColumns->at(i));
   				throw std::runtime_error( ErrorMsg );
			}
		}
		m_nNbCols = pviSelectedColumns->size() + pvclSNodes->size();
		m_nNbRows = nNbRows;
		m_adInputDatas = new double*[m_nNbRows];
		for (int i = 0; i < m_nNbRows; i++){
			m_adInputDatas[i] = new double[m_nNbCols];
			for (int j = 0; j < (int)pviSelectedColumns->size(); j++)
			{
				// Building the internal array with selected columns of adDataArray
				m_adInputDatas[i][j] = adDataArray[i][pviSelectedColumns->at(j) - 1];
			}
			for (int j = pviSelectedColumns->size(); j < m_nNbCols; j++)
			{
				m_adInputDatas[i][j] = 0;
			}
		}
	}
	catch( std::exception &e )
    {
   		sprintf( ErrorMsg, "~ErrorInSParentNodeConstructor~\n%s\n", e.what() );
   		throw std::runtime_error( ErrorMsg );
    }
}

SPARENTNODE::~SPARENTNODE()
{
	if (m_pvclSNodes != NULL) 
	{
		delete m_pvclSNodes;
		m_pvclSNodes = NULL;
	}
}

vector <SNODE *> * SPARENTNODE::pvclGetSNodes()
{
	return m_pvclSNodes;
}

void SPARENTNODE::SetSNodes(vector <SNODE *> * pvclSNodes)
{
	m_pvclSNodes = pvclSNodes;
}

void SPARENTNODE::Infer()
{
	// Sub snodes Inference folowed by embedded sif inference
	for (int j = 0; j < (int)m_pvclSNodes->size(); j++)
	{		
		// Sub-SNode's Inference
		m_pvclSNodes->at(j)->Infer(); 
		
		// Sub-SNode Inference's OutValues copy 
		for (int i = 0; i < m_nNbRows; i++)
		{
			m_adInputDatas[i][m_nNbCols-m_pvclSNodes->size()+j] = m_pvclSNodes->at(j)->pvdGetOutValues()->at(i);
		}
	}
	
	if (m_pclFis != NULL)
	{
		m_pclFis->ClassifCheck(m_adInputDatas,m_nNbRows, 0);
	    m_pclFis->ResClassifAlloc(m_piResClassif, m_pdLab, 0); 
	    
		// This SNode's Inference
		for (int i = 0; i < m_nNbRows; i++)
		{
			m_pclFis->Infer(m_adInputDatas[i], 0, NULL, 0);
			m_pvdOutValues->push_back(m_pclFis->OutValue[0]);
		} 
		delete m_piResClassif;
		m_piResClassif = NULL;
		delete m_pdLab;
		m_pdLab = NULL;
	}
}

void SPARENTNODE::ToString()
{
	SNODE::ToString();
	cout << "List of my child nodes :" << endl;
	for (int i = 0; i < (int)m_pvclSNodes->size(); i++)
	{
		m_pvclSNodes->at(i)->ToString();
	}
}

//**************************    SPARENTNODE.CPP   *******************************
