//***********************************************************************
//
//
//                              SUPERFIS.CPP
//
//
// Author(s) : Vincent Therry
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Contact : vincent.therry@envilys.com
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : SUPERFIS class functions used by FISPRO, part of library fispro

//**********************************************************************

#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <vector>
using namespace std;
#include "snode.h"
#include "superfis.h"
#include <map>

SUPERFIS::SUPERFIS()
{
	m_pclRoot = NULL;
	m_pvclSNodes = new vector <SNODE *>;
	m_adInputs = NULL;
	m_nNbCols = 0;
	m_nNbRows = 0; 
}

SUPERFIS::SUPERFIS(const SUPERFIS & s)
{
	m_nNbCols = s.m_nNbCols;
	m_nNbRows = s.m_nNbRows;
	m_adInputs = new double *[m_nNbRows];
	for (int i = 0; i < m_nNbRows; i++){
		m_adInputs[i] = new double[m_nNbCols];
		for (int j = 0; j < m_nNbCols; j++)
		{
			m_adInputs[i][j] = s.m_adInputs[i][j];
		}
	}
	m_pvclSNodes = new vector <SNODE *>;
	for (int i = 0; i < (int)s.m_pvclSNodes->size(); i++)
	{
		if (typeid(s.m_pvclSNodes->at(i)) == typeid(SNODE*))
		{
			m_pvclSNodes->push_back(new SNODE(*(s.m_pvclSNodes->at(i))));
		}
		else if (typeid(s.m_pvclSNodes->at(i)) == typeid(SLEAFNODE*))
		{
			// ici on sait que le pointeur sur le noeud est un sleafnode donc on le cast en sleafnode * puis on le traverse
			// pour avoir la rfrence sur le sleafnode pour la passer au contructeur par recopie
			m_pvclSNodes->push_back(new SLEAFNODE(*dynamic_cast<SLEAFNODE *>(s.m_pvclSNodes->at(i))));
		}
		else if (typeid(s.m_pvclSNodes->at(i)) == typeid(SPARENTNODE*))
		{
			m_pvclSNodes->push_back(new SPARENTNODE(*dynamic_cast<SPARENTNODE *>(s.m_pvclSNodes->at(i))));
		}
	}
	m_pclRoot = m_pvclSNodes->at(m_pvclSNodes->size()-1); // we suppose that the root is the last SNode in the vector 
}

SUPERFIS::SUPERFIS(char * sDataFilename, char *sConfigFileName)
{
	m_pclRoot = NULL;
	m_pvclSNodes = new vector <SNODE *>;
	m_adInputs = ReadSampleFile(sDataFilename, m_nNbCols, m_nNbRows); 
	Init(sConfigFileName);
}

SUPERFIS::SUPERFIS(double ** adInputs, int nNb_Cols, int nNb_Rows, char *sConfigFileName)
{
	m_pclRoot = NULL;
	m_pvclSNodes = new vector <SNODE *>;
	m_adInputs = adInputs;
	m_nNbCols = nNb_Cols;
	m_nNbRows = nNb_Rows; 
	Init(sConfigFileName);
}

SUPERFIS::~SUPERFIS()
{
	if (m_pvclSNodes != NULL) 
	{
		delete m_pvclSNodes;
		m_pvclSNodes = NULL;
	}
	
}

double ** SUPERFIS::adGetInputs()
{
	return m_adInputs;
}

int SUPERFIS::nGetNbCols()
{
	return m_nNbCols;
}

int SUPERFIS::nGetNbRows()
{
	return m_nNbRows;
}

vector <SNODE *> * SUPERFIS::pvclGetSNodes()
{
	return m_pvclSNodes;
}

SNODE * SUPERFIS::pclGetRoot()
{
	return m_pclRoot;
}

void SUPERFIS::SetInputs(double ** adInputs)
{
	m_adInputs = adInputs;
}

void SUPERFIS::SetNbCols(int nNbCols)
{
	m_nNbCols = nNbCols;
}

void SUPERFIS::SetNbRows(int nNbRows)
{
	m_nNbRows = nNbRows;
}

void SUPERFIS::SetSNodes(vector <SNODE *> * pvclSNodes)
{
	m_pvclSNodes = pvclSNodes;
}

void SUPERFIS::SetRoot(SNODE * pclRoot)
{
	m_pclRoot = pclRoot;
}

void SUPERFIS::Infer()
{
	if (m_pclRoot != NULL)
	{
		m_pclRoot->Infer();
	}
}

void SUPERFIS::ToString()
{
	cout << "Object SuperFis : " << endl;
	cout << "Datas : " << endl;
	for (int i = 0; i < m_nNbRows; i++)
	{
		for (int j = 0; j < m_nNbCols; j++)
		{
			cout << m_adInputs[i][j] << "\t";
		}
		cout << endl;
	}
	cout << "Number of lines : " << m_nNbRows << endl;
	cout << "Number of columns : " << m_nNbCols << endl;
	cout << "Overview of my architecture : " << endl;
	m_pclRoot->ToString();
}

void SUPERFIS::Init(char * sConfigFileName)
{
	char ErrorMsg[300];
	istringstream iss;
	string sLine="";
	int nLine = 0;
	char sPath[300];
	int nOffset1 = 0;
	int nOffset2 = 0;
	int nLastOffset = 0;
	int nNum=0;
	string sChaine="";
	vector <int> * pviSelectedColumns = new vector <int>;
	vector <SNODE *> * pvclSNodes = new vector <SNODE *>;
	map<string,SNODE *> NameToSNodeMap;
	
	ifstream f(sConfigFileName);

	try
	{
		if (f.good())
		{
			while (std::getline(f, sLine)) 
			{   
				nLine++;
				if (string::npos != sLine.find("SLEAFNODE"))
				{
					nLastOffset = strlen("SLEAFNODE") - 1; // We initialise nLastOffset with the position of the last letter of "SLEAFNODE" (we begin at 0)
					nOffset1 = sLine.find_first_of("(");
					if (string::npos == (unsigned int)nOffset1 || nOffset1 != nLastOffset + 1) // If not found or if the "(" isn't just after "SLEAFNODE"
					{
						sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : '(' is missing at Col%d after the node declaration : %s~\n~%s~", nLine, nLastOffset + 1, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
					}
					
					nOffset2 = sLine.find_first_of(";");
					if (string::npos == (unsigned int)nOffset2)
					{
						sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : ';' is missing after the path declaration : %s~\n~%s~", nLine, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
					}
					strcpy(sPath, (sLine.substr(nOffset1 + 1, nOffset2-nOffset1-1)).c_str());
					nLastOffset = nOffset2;
					nOffset1 = sLine.find_first_of("[");
					if (string::npos == (unsigned int)nOffset1 || nOffset1 != nLastOffset + 1) // Si non trouv, ou si le "[" ne se trouve pas juste apres le ";"
					{
						sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : '[' is missing at Col%d after the ';' : %s~\n~%s~", nLine, nLastOffset + 1, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
					}
					nOffset2 = sLine.find_first_of("]");
					if (string::npos == (unsigned int)nOffset2)
					{
						sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : ']' is missing after columns declaration : %s~\n~%s~", nLine, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
					}
					// We extract selected columns
					iss.str(sLine.substr(nOffset1+1, nOffset2-nOffset1-1)); 
					while (std::getline(iss, sChaine, ','))
     				{
     					nNum = strtol(sChaine.c_str(), NULL, 10);
     					pviSelectedColumns->push_back(nNum);      					
     				}
     				iss.clear();
     				nLastOffset = nOffset2;
     				nOffset1 = sLine.find_last_of(";");
     				if (string::npos == (unsigned int)nOffset1 || nOffset1 != nLastOffset + 1)
     				{
     					sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : ';' is missing at Col%d : %s~\n~%s~", nLine, nLastOffset + 1, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
     				}
     				nOffset2 = sLine.find_last_of(")");
    				if (string::npos == (unsigned int)nOffset2 ||  (unsigned int)nOffset2 <  (unsigned int)nOffset1 )
     				{
     					sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : ')' is missing at Col%d after the fis name declaration : %s~\n~%s~", nLine, nOffset1 + 1, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
     				}
     				iss.str(sLine.substr(nOffset1+1, nOffset2-nOffset1-1)); // Here we have the Name of the FIS : Must be unique !
					
					// We look into the map to find out if the name to ad as FIS child exists
					if (NameToSNodeMap.find(iss.str()) != NameToSNodeMap.end())
					{
						sprintf( ErrorMsg, "~Fictive FIS Name already exists~\n~Line %d : The name '%s' is already used by another FIS : %s~\n~%s~", nLine, iss.str().c_str(), sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
					}
					// We create the leaf node
					SLEAFNODE * n = new SLEAFNODE(sPath, m_adInputs, m_nNbRows, pviSelectedColumns);
					m_pvclSNodes->push_back(n);
					NameToSNodeMap[iss.str()] = n;
					pviSelectedColumns->clear();					
				}
				else if (string::npos != sLine.find("SPARENTNODE"))
				{
					
					nLastOffset = strlen("SPARENTNODE") - 1; // We initialise nLastOffset with the position of the last letter of "SPARENTNODE" (we begin at 0)
					nOffset1 = sLine.find_first_of("(");
					if (string::npos == (unsigned int)nOffset1 || nOffset1 != nLastOffset + 1) // If not found or if the "(" isn't just after "SPARENTNODE"
					{
						sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : '(' is missing at Col%d after the node declaration : %s~\n~%s~", nLine, nLastOffset + 1, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
					}
					nOffset2 = sLine.find_first_of(";");
					if (string::npos == (unsigned int)nOffset2)
					{
						sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : ';' is missing after the path declaration : %s~\n~%s~", nLine, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
					}
					strcpy(sPath, (sLine.substr(nOffset1 + 1, nOffset2-nOffset1-1)).c_str());
					nLastOffset = nOffset2;
					nOffset1 = sLine.find_first_of("[");
					if (string::npos == (unsigned int)nOffset1 || nOffset1 != nLastOffset + 1) // If not found or if the "[" isn't just after the ";"
					{
						sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : '[' is missing at Col%d after the ';' : %s~\n~%s~", nLine, nLastOffset + 1, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
					}
					nOffset2 = sLine.find_first_of("]");
					if (string::npos == (unsigned int)nOffset2)
					{
						sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : ']' is missing after columns declaration : %s~\n~%s~", nLine, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
					}
					iss.str(sLine.substr(nOffset1+1, nOffset2-nOffset1-1));
					while (std::getline(iss, sChaine, ','))
     				{
     					nNum = strtol(sChaine.c_str(), NULL, 10);
     					pviSelectedColumns->push_back(nNum);
     				}
     				iss.clear();
     				nLastOffset = nOffset2;
     				nOffset1 = sLine.find_first_of(";", nLastOffset);
     				if (string::npos == (unsigned int)nOffset1 || nOffset1 != nLastOffset + 1)
     				{
     					sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : ';' is missing at Col%d : %s~\n~%s~", nLine, nLastOffset + 1, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
     				}
     				nLastOffset = nOffset1;
     				nOffset1 = sLine.find_last_of("[");
     				if (string::npos == (unsigned int)nOffset1 || nOffset1 != nLastOffset + 1)
     				{
     					sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : '[' is missing at Col%d : %s~\n~%s~", nLine, nLastOffset + 1, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
     				}
					nOffset2 = sLine.find_last_of("]");
					if (string::npos == (unsigned int)nOffset2)
					{
						sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : ']' is missing after child node's declaration : %s~\n~%s~", nLine, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
					}
					iss.str(sLine.substr(nOffset1+1, nOffset2-nOffset1-1));
					while (std::getline(iss, sChaine, ','))
     				{
     					// sChaine here contains FIS's name
     					// We look into the map to find out if the name to ad as FIS child exists
						if (NameToSNodeMap.find(sChaine) == NameToSNodeMap.end())
						{
							sprintf( ErrorMsg, "~Fictive FIS Name Doesn't exists~\n~Line %d : The name '%s' doesn't exist in the name's map, it can't be added to the parent node's children : %s~\n~%s~", nLine, sChaine.c_str(), sLine.c_str(), sConfigFileName);
		      				throw std::runtime_error( ErrorMsg );
						}
						else
						{
     						pvclSNodes->push_back(NameToSNodeMap[sChaine]);
						}
     				}
     				nLastOffset = nOffset2;
     				nOffset1 = sLine.find_last_of(";");
     				if (string::npos == (unsigned int)nOffset1 || nOffset1 != nLastOffset + 1)
     				{
     					sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : ';' is missing at Col%d : %s~\n~%s~", nLine, nLastOffset + 1, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
     				}
     				nOffset2 = sLine.find_last_of(")");
    				if (string::npos == (unsigned int)nOffset2 ||  (unsigned int)nOffset2 <  (unsigned int)nOffset1 )
     				{
     					sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : ')' is missing at Col%d after the fis name declaration : %s~\n~%s~", nLine, nOffset1 + 1, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
     				}
     				iss.str(sLine.substr(nOffset1+1, nOffset2-nOffset1-1)); // Here we have the Name of the FIS : Must be unique !
					
					// We look into the map to find out if the name exists
					if (NameToSNodeMap.find(iss.str()) != NameToSNodeMap.end())
					{
						sprintf( ErrorMsg, "~Fictive FIS Name already exists~\n~Line %d : The name '%s' is already used by another FIS : %s~\n~%s~", nLine, iss.str().c_str(), sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
					}
     				iss.clear();
					SPARENTNODE * n = new SPARENTNODE(sPath, m_adInputs, m_nNbRows, pviSelectedColumns, pvclSNodes);
					m_pvclSNodes->push_back(n);
					NameToSNodeMap[iss.str()] = n;
					pviSelectedColumns->clear();
					pvclSNodes->clear();		
				}
				else if (string::npos != sLine.find("ROOT"))
				{
					nOffset1 = sLine.find_first_of("(");
					if (string::npos == (unsigned int)nOffset1)
					{
						sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : '(' is missing after the 'ROOT' :%s~\n~%s~", nLine, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
					}
					nOffset2 = sLine.find_first_of(")");
					if (string::npos == (unsigned int)nOffset1)
					{
						sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : ')' is missing after the 'ROOT' node :%s~\n~%s~", nLine, sLine.c_str(), sConfigFileName);
	      				throw std::runtime_error( ErrorMsg );
					}
					iss.str(sLine.substr(nOffset1+1, nOffset2-nOffset1-1));
					if (std::getline(iss, sChaine))
     				{
     					// sChaine here contains FIS's name
     					// We look into the map to find out if the name exists
						if (NameToSNodeMap.find(iss.str()) == NameToSNodeMap.end())
						{
							sprintf( ErrorMsg, "~Fictive FIS Name doesn't exists~\n~Line %d : The name '%s' doesn't exist in the name's map, it can't be set as root node : %s~\n~%s~", nLine, iss.str().c_str(), sLine.c_str(), sConfigFileName);
		      				throw std::runtime_error( ErrorMsg );
						}
						m_pclRoot = NameToSNodeMap[iss.str()];
     				}
     				iss.clear();
				}
				else
				{
					sprintf( ErrorMsg, "~UnknownLineFormat~\n~Line %d : %s~\n~sLine : %s~", nLine, sConfigFileName,sLine.c_str());
	      			throw std::runtime_error( ErrorMsg );
				}
			}     
		}
		else
		{
	      sprintf( ErrorMsg, "~CannotOpenDataFile~: %.100s~", sConfigFileName);
	      throw std::runtime_error( ErrorMsg );
		}
		delete pviSelectedColumns;
   		delete pvclSNodes;

   }
   catch( std::exception &e )
   {
   		delete pviSelectedColumns;
   		delete pvclSNodes;
   		sprintf( ErrorMsg, "~ErrorInConfigFile~\n~ErrorInInitSuperfis~\n%s\n", e.what() );
   		throw std::runtime_error( ErrorMsg );
   }
}

//**************************    SUPERFIS.CPP   *******************************
