//**********************************************************************
//
//
//                              FISWM.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISWM class description, used with fispro, 
//        part of package wm (Wnag & Mendel induction method)
//        requires library fispro and package hfp
//
//**********************************************************************

#include "fiswm.h"

int FISWM::wm(void)
  //***************
{
  int i, j;
  int * turn;
  OUT_CRISP ** SaveOut;

  turn = NULL;
  SaveOut = NULL;

  if(NbOut < 1) return -1;

  // set NbRules=0 -  WM procedure only depends on existing data
  // does not depend on existing rules
  // delete MFConc arrays
  for(i = 0; i < NbOut; i++) 
    {
      Out[i]->DeletePossibles(NbRules);
    }

  if(NbRules)
    {
      if (Rule != NULL)
	{
	  for(i = 0; i < NbRules; i++) 
	    if (Rule[i] != NULL) delete Rule[i];
	  delete [] Rule;
          Rule = NULL;
	}
      NbRules=0;
    }
  
  turn = new int [NbOut];
  SaveOut = new OUT_CRISP *[NbOut];

  for(i = 0; i < NbOut; i++) 
    {
      turn[i] = false;
      SaveOut[i] = NULL;

      if(!strcmp(Out[i]->GetOutputType(), OUT_CRISP::OutputType()) )
	{
	  turn[i] = true;
	  SaveOut[i] = new OUT_CRISP(*Out[i]);
	  
	  double * v, *u;
	  int nu, *p;
	  v = new double [lig];
	  GetColumn(Data, lig, NbIn + i, v);
	  SortUniq(v, lig, &u, &p, nu, EPSILON);

	  if(NoLimit == false && nu > MAX_CLASS)
	    {
	      sprintf( ErrorMsg, "~TwoManyDistinctValues~(%d max): %d\n~InOutput~%d\n", MAX_CLASS, nu, i+1);
	      throw std::runtime_error( ErrorMsg );
	    }

	  else
	    {
	      if( nu > MAX_MF)
		{
		  double * centres, step;
		  int nc = MAX_MF;
		  centres = new double [MAX_MF];
		  step = (Out[i]->max() - Out[i]->min()) / MAX_MF;
		  for(j = 0; j < MAX_MF; j++)
		    centres[j] =  Out[i]->min() + step*j;
		  Kmeans(u, nu, centres, nc);
		  KmeansNE(u, nu, centres, nc);
		  Crisp2Fuz(i, (char *) OUT_FUZZY::MeanMaxDefuz(), centres, nc);
		  delete [] centres;
		}
	      else
		Crisp2Fuz(i, (char *) OUT_FUZZY::MeanMaxDefuz(), u, nu);
	    }
	      delete [] v;
	      delete [] u;
	      delete [] p;
	}
      if(Out[i]->GetNbMf() < 2) 
	{
	  if(turn[i])
	    sprintf( ErrorMsg, "~LessThan2DistinctValuesForOutput~: %.50s", Out[i]->Name);
	  else
	    sprintf( ErrorMsg, "~LessThan2MFDefinedForOutput~: %.50s", Out[i]->Name);
	  throw std::runtime_error( ErrorMsg );
	}
    }  

  GenerateRules(true);
  ConflictManagement();

  for(i = 0; i < NbOut; i++) 
    if(turn[i])  
      {
	Fuz2Crisp(i);
	Out[i]->SetOpDisj(SaveOut[i]->Disjunct());
	Out[i]->SetOpDefuz(SaveOut[i]->Defuzzify());
	Out[i]->Classification(SaveOut[i]->Classification());
      }
  delete [] turn; 
  for(i = 0; i < NbOut; i++) 
    if (SaveOut[i]!=NULL) delete SaveOut[i];
  if (SaveOut !=NULL) delete [] SaveOut;
  Print();
  return 0;
}

void FISWM::GenerateRules(int out)
  //******************************
{
  int i, j, mf;
 
  char * tmp, *buf;

  buf = new char [6*NbIn+20];
  tmp = new char [30];

  if(NbRules)
    {
      if (Rule != NULL)
	{
	  for(i = 0; i < NbRules; i++) 
	    if (Rule[i] != NULL) delete Rule[i];
	  delete [] Rule;
	}
    }
 
  Rule = new RULE *[lig];
  NbRules = lig;  

  for(j = 0; j < lig; j++) // For each example
    {
      buf[0] = 0;  

      for(i = 0; i < NbIn; i++) // For each input
	{
	  if(In[i]->IsActive())
	    {
	      mf = MaxDeg(Data[j][i], i);
	      sprintf(tmp, "%d%c", mf + 1, SEPARE); 
	      strcat(buf, tmp);
	    }
	  else 
	    {
	      sprintf(tmp, "%d%c", ANY, SEPARE); 
	      strcat(buf, tmp);
	    }
	}

      if(out)
	{
	  for(i = 0; i < NbOut; i++) 
	    {
	      if(Out[i]->IsActive())
		{
		  mf = MaxDegOut(Data[j][i+NbIn], i);
		  sprintf(tmp, "%d%c", mf + 1, SEPARE); 
		  strcat(buf, tmp);
		}
	      else 
		{
		  sprintf(tmp, "%d%c", ANY, SEPARE); 
		  strcat(buf, tmp);
		}
	    }
	}

      else // if(! out)
	for(i = 0; i < NbOut; i++) 
	  {
	    sprintf(tmp,"%f%c", Out[i]->DefaultValue(), SEPARE);
	    strcat(buf, tmp);
	  }

      Rule[j] = new RULE(NbIn, In, NbOut, Out, cConjunction, buf);
    }

  delete [] buf;
  delete [] tmp;
}

int FISWM::MaxDeg(double v, int i) const
  //******************************
{
  int k, index;
  double max;

  In[i]->GetDegsV(v);

  max=In[i]->Mfdeg()[0];
  index = 0;

  for(k = 1; k < In[i]->GetNbMf(); k++)
    {
      if(In[i]->Mfdeg()[k] > max)
	{
	  max = In[i]->Mfdeg()[k];
	  index = k;
	}
    }
  return index;
}

int FISWM::MaxDegOut(double v, int i) const
  //*********************************
{
  int k, index;
  double max;

  Out[i]->GetDegsV(v);

  max=Out[i]->Mfdeg()[0];
  index = 0;

  for(k = 1; k < Out[i]->GetNbMf(); k++)
    {
      if(Out[i]->Mfdeg()[k] > max)
	{
	  max = Out[i]->Mfdeg()[k];
	  index = k;
	}
    }
  return index;
}


double FISWM::CorrespondingWeight(int e)
  //************************************
{
  int i;

  for(i = 0; i  < NbIn; i++)      
    {
      if(In[i]->IsActive() == false) continue;
     
      if(FisIsnan(Data[e][i]))            // value manquante
	{
	  if(!strcmp(strMissingValues, FIS::RandomMissingValues()))
	    In[i]->GetRandDegs(Data[e][i]);
	  else if(!strcmp(strMissingValues, FIS::MeanMissingValues()))
	    In[i]->SetEqDegs(Data[e][i]);
	  else
	    {  
	      sprintf( ErrorMsg, "~UnknownMissingValueStrategy~: %.50s", strMissingValues );
	      throw std::runtime_error( ErrorMsg );
	    }
	}
      In[i]->GetDegsV(Data[e][i]);
    }
  Rule[e]->ExecRule();
  return Rule[e]->Weight;
}

void FISWM::ConflictManagement(void)
  //********************************
{
  int i, j;

  for(i = 0; i < lig; i++)
    {
      if( ! Rule[i]->IsActive() ) continue;
      for(j = i+1; j < lig; j++)
	{
	  if( ! Rule[j]->IsActive() ) continue;
	  if ( ! Rule[i]->ComparePremises(Rule[j])) // Identical premises
	    {
	      if(CorrespondingWeight(i) > CorrespondingWeight(j))
		Rule[j]->Deactivate(); 
	      else Rule[i]->Deactivate(); 
	    }
	  if( ! Rule[i]->IsActive() ) break;

	}
    }
}


void FISWM::Print(void) const
  //*******************
{
  FILE * f;

  f = fopen(FileOut, "wt");
  if(f == NULL) return;

  PrintCfg(f);
  fclose(f);
}

