//**********************************************************************
//
//
//                              FISWM.H
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISWM class description, used with fispro, 
//        part of package wm (Wnag & Mendel induction method)
//        requires library fispro and package hfp
//
//**********************************************************************


#ifndef __FISWM_H
#define __FISWM_H

#include "fis.h"


//! Wang and Mendel induction algorithm implementation.
class FISWM : public FIS
//**********************
{ 
  FISWM& operator=(const FISWM&);
  FISWM(const FISWM&);

protected:

 double ** Data;
 int col, lig, NoLimit, Falloc;
 char * FileOut;

 //! Compute the matching degree of rule e for example e
 double CorrespondingWeight(int e);

 void ConflictManagement(void);

  //! If 'out' is false rule conclusions are set to the output default
  //! value else they are initialized in the same way than the premises, 
  //! using the MaxDegOut function.
  void GenerateRules(int out = true);

  //! Print a fis configuration file with the name 'FileOut'.
  void Print(void) const;
 
public:
  
  //! 'cfg' is the input fis configuration file. The rules mentionned 
  //! in this file are not taken into account. 
  //! 'fdata' is the data file name. 
  //! 'fout' is the output fis configuration file. If NULL the file is 
  //! 'Name'wm.fis
  //! This Wang and Mendel implementation works only with fuzzy outputs. 
  //! Convert crisp output to a fuzzy one and use discrete MF 
  //! discrete MF build with crips labels.
  FISWM(char * cfg, char * fdata, char *fout) : FIS(cfg)
    {
      Data = ReadSampleFile(fdata, col, lig);
      FileOut = fout;
      Falloc = false; 
      NoLimit = true;

      if( ! FileOut) 
	{
	  FileOut = new char [strlen(Name) + 10];
	  sprintf(FileOut, "%swm.fis", Name);
	  Falloc = true;
	}
    }

  ~FISWM()
    {
      int i; 
      for(i = 0; i < lig; i++) delete [] Data[i];
      delete [] Data;
      if (Falloc) delete [] FileOut;
    }

  //! The parameter NoLimit is used to control the number of distinct 
  //! output values. When false this number is limited to MAX_CLASS, 
  //! otherwise by the constant MAX_MF.
  void SetNoLimit(int l) { NoLimit = true; }
  void UnSetNoLimit(int l) { NoLimit = false; }
  int  GetNoLimit(void)  { return NoLimit; }

  //! Return 0 or (100 + out number) when it has less than 2 MF. 
  int wm(void);

  //! Return the mf number of input i for which the degree of v is max 
  int MaxDeg(double v, int i) const;
  
  //! Return the mf number of output i for which the degree of v is max 
  int MaxDegOut(double v, int i) const;
};

#endif
