//**********************************************************************
//
//
//                             MAINFISWM.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : Sample file to use FISWM class 
//        part of package wm (Wnag & Mendel induction method)
//        requires library fispro and package hfp
//
//**********************************************************************

#include "fis.h"
#include "fiswm.h"

int main(int argc, char ** argv)
//******************************
{
  if(argc < 3)
    {
      printf("\nThe programm %s needs 2 arguments : ", argv[0]);
      printf("\n\t* the fis configuration file name;");   
      printf("\n\t* the data file name.") ;
      printf("\nOption:");
      printf("\n\t* -tfile data file name used for performance computation (default : data file)");
      printf("\n\t* -ofile the output fis configuration file name (default 'Name'wm.fis)");
      printf("\n\t* -s activity threshold for performance computation (default: 0.2)" );
      printf("\n\t* -l 'NoLimit' parameter (default:  true)" );
      printf("\n");
      printf("\nThe program creates two more files: perfwm.res and result.wm");
      printf("\n");

      return 1;
    }
  
  FISWM *S;
  FIS *T;
  FILE *f;
  int i, limit;
  double ret, Cov, MaxErr, thres;
  char * OutConfig, * testperffile;
  bool allocConfig=false;

  OutConfig = testperffile = NULL;
  thres = 0.2;
  limit = false;

  try
    {
      if(argc > 2)   
	{
	  for(i = 2; i < argc; i++)
	    {
	      if(! strncmp("-s", argv[i], 2) ) thres = strtod(argv[i] + 2, NULL);
	      else if(! strncmp("-o", argv[i], 2) ) OutConfig = argv[i] + 2;
	      else if(! strncmp("-t", argv[i], 2) ) testperffile = argv[i] + 2;
	      else if(! strncmp("-l", argv[i], 2) ) 
		{ 
		  if(strlen(argv[i]) > 2)  limit = atoi(argv[i] + 2);
		  else limit = true;
		}
	    }
	}

      if( ! OutConfig) 
	{
	  OutConfig = new char [strlen(argv[2]) + 10];
	  sprintf(OutConfig, "%swm.fis", argv[2]);
	  allocConfig=true;
	}
      S = new FISWM(argv[1], argv[2], OutConfig);

      if(limit) S->UnSetNoLimit(limit);
     
      if(S->wm())
	{
	  sprintf(ErrorMsg, "\n~NoOutputDefinedInFis~:  %s\n", argv[1]); 
	  throw std::runtime_error( ErrorMsg );
	}
      delete S;

      if(testperffile == NULL) testperffile = argv[2];
      
      T = new FIS(OutConfig);
      ret = T->Performance(0, testperffile, Cov, MaxErr, thres, "perfwm.res", false);
      f = fopen("result.wm", "wt");
      T->WriteHeaderPerfRB(0, f);
      fprintf(f, "\n");
      T->PerfRB(ret, Cov, MaxErr, 0, f);
      fprintf(f, "\n");
      fclose(f);
      delete T;
      if (allocConfig)
	delete [] OutConfig;
    }

  catch(std::exception &e)
    {
      printf("\n%s\n", e.what());
    }


  return 0;
}

//**************************   Fin de MAINFISWM.CPP   *************************

