//***********************************************************************
//
//
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.4 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//**********************************************************************


#include "common.h"
#include "fis.h"

void GenereDataTable(FIS *S)
//**************************
{
  int i, nbin=S->GetNbIn();
  double *cen;
  FILE *f;
  f = fopen("centre.dat", "wt");
  cen=new double[2];
  for (i=0; i<nbin; i++)//input number
    {
      S->In[i]->GetMfCenters(cen);
      fprintf(f, " %f; %f\n", cen[0], cen[1]);
    }
  fclose(f);
  delete [] cen;
}


int main(int argc, char ** argv)
//******************************
{
  if(argc < 3)
    {
      printf("\nThe programm %s needs 2 arguments: ", argv[0]);
      printf("\n\t* the name of the FIS configuration file;");   
      printf("\n\t* the name of the data file;");
      printf("\nOptions : ");
      printf("\n\t\t* -in 'n m etc.' the string of input numbers to consider (default=consider all inputs)\n");
      printf("\n\t\t* -wl for wordless.\n");
      printf("\n\t\t* -t for centre.dat file generation");
      
      return 1;
    }
  //example
  //fisensi iris.fis iris 

  double ** Exemples  = NULL;   // A Pointer to sample array
  double *cen=NULL;
  int i=0,nbrow=0, numvar=0, num=0,nbin=0;	
  int len=0,maxval,vari,vari0;
  double c,cref;
  double overlap=0.0;
  char * snum=new char[100];
  char * outputFisConfigName=NULL, * outputFisBaseName=NULL, *outputF=NULL,*outputF2=NULL,*fisout=NULL;
  double coverage=0., MaxError=0.;
  double params[3];
   bool wordless=false;
   bool tablegen=false;
   FILE *fout=NULL;
   FIS * S=NULL;
   double tmp=0.0;
   double * numinputs=NULL;
   int nbinputs=0,lenin=0,n=0;
   int argstr=0;
   bool partial=false;
  char blank=' ';

  if(argc > 2) 
    {
      for(i = 2; i < argc; i++)
	{
	  if(! strncmp("-wl", argv[i], 2) ) wordless = true;
	  if(! strncmp("-t", argv[i], 2) ) tablegen = true;
	  if(! strncmp("-in", argv[i],2)) 
	    {
	      partial=true;//next argument will hold string of input numbers
	      argstr=i+1;
	    }
	}
    }
  
  if (!wordless)
    {
      for (i=0;i<argc;i++)
	printf("argument %d = %s  ",i,argv[i]);
      printf("\n");
    }

  

  try{
    S = new FIS(argv[1]);     
    nbin=S->GetNbIn();

    // Generate file centre.dat
    if (tablegen)
      {
	GenereDataTable(S);
	printf("centers have been generated in file centre.dat");
	return 0;
      }

    //put input numbers to consider into array

    numinputs=new double[nbin];//max number

    if (partial)
      // some inputs only
      {
	lenin=strlen(argv[argstr]);
	if (lenin==1)
	  {
	    nbinputs=1;
	    numinputs[0]=atol(argv[argstr]);
	  }
	else
	  nbinputs=SearchNb(argv[argstr],numinputs,nbin,blank);
	
	if (!wordless)
	  printf("\n %d inputs found:",nbinputs);
	if (!wordless)
	  printf("\ninputs to consider: %s \n",argv[argstr]);
	for (int kk=0;kk<nbinputs;kk++)
	  {
	    //printf("%d  ",(int)(numinputs[kk]));
	    if ((int)numinputs[kk]<1 ||(int)numinputs[kk]>nbin)
	      {
		printf("\nerror in input string\n");
		return 1;
	      }
	  }
	if (!wordless)
	  printf("\n");
      }
    else
      {
	//all inputs
	nbinputs=nbin;
	for (int kk=0;kk<nbinputs;kk++)
	  {
	    numinputs[kk]=kk+1;
	  }
	
      }

    len=strlen(argv[1]) + 100;
    outputFisBaseName=new char[len];
    strcpy(outputFisBaseName,argv[2]);
    strcat(outputFisBaseName,"-V");

    outputFisConfigName=new char[len];
    outputF=new char[len];
    outputF2=new char[len];
    fisout=new char[len];
    cen=new double[2];
    maxval=5; // number of values to study+1
  
    // infer with initial FIS
    S->Performance(0, argv[2],coverage , MaxError,0.1 ,"init.perf",0);

    maxval=3;  // 0 will be skiped
    
    //build FIS and infer
    //for (numvar=0;numvar<nbin;numvar++)//input number
    for (n=0;n<nbinputs;n++)//input number
      {
	numvar=(int)numinputs[n]-1;
	//generate output file basename
	strcpy(outputFisConfigName,outputFisBaseName);
	sprintf(snum,"%d",numvar+1);
	strcat(outputFisConfigName,snum);
	
	S->In[numvar]->GetMfCenters(cen);
	if(!wordless) 
	  {
	    printf("numvar=%d\n",numvar+1);
	  
	    for (i=0;i<2;i++)
	      printf("cen[%d]=%g ",i+1,cen[i]);
	    printf("\n");
	  }

	strcat(outputFisConfigName,"-c");
	
	overlap=cen[1]-cen[0];

	for (num=0;num<2;num++)//MF number
	  {
	    strcpy(outputF,outputFisConfigName);//to keep for next iteration
	    sprintf(snum,"%d",num+1);
	    strcat(outputF,snum);

	    cref=cen[num];
	    vari0=-20;
	    i=0;
	    while(i<maxval)
	      {
		strcpy(outputF2,outputF);
		vari=vari0+i*20;
		i=i+1;
		if (vari !=0)
		  {
		    c=cref+(vari*overlap/100);
		    if(vari <0)	  
		      sprintf(snum,"%dpercent",vari);
		    else
		      sprintf(snum,"+%dpercent",vari);
		    strcat(outputF2,snum);
		
		    strcpy(fisout, outputF2);
		    strcat(fisout,".fis");
		
		    if(!wordless)
		      printf("\nnumvar=%d, c=c%d, valc=%g,outputFile=%s\n",numvar+1,num+1,c,outputF2);
		      
		    //update FIS (input numvar, mf num, param c)
		    S->In[numvar]->GetMF(num)->GetParams(params);
		    if (num==0)//first MF
		      {
			tmp = max(c, params[0]);
			params[1]=tmp;
			S->In[numvar]->GetMF(0)->Update(params);
			S->In[numvar]->GetMF(1)->GetParams(params);
			params[0]=tmp;
			S->In[numvar]->GetMF(1)->Update(params);
		      }
		    if (num==1)//second MF
		      {
			tmp = min(c, params[2]);
			params[1]=tmp;
			S->In[numvar]->GetMF(1)->Update(params);
			S->In[numvar]->GetMF(0)->GetParams(params);
			params[2]=tmp;
			S->In[numvar]->GetMF(0)->Update(params);
		      }
		
		    if(!wordless) S->In[numvar]->Print();

		    // create perf file and FIS file
		    fout=fopen(fisout,"wt");   
		    S->PrintCfg(fout);
		    fclose(fout);

		    /*
		    printf("\nnumvar:  %d, num:  %d, i:  %d", numvar, num, i);
		    printf("\nfisout:   %s, outputF:  %s, outputF2:  %s", fisout, outputF, outputF2);	
		    */	


		    FIS *T;
		    T = new FIS(fisout);
		    T->Performance(0, argv[2], coverage , MaxError, 0.1 ,outputF2,0);
		    delete T;

		    // S->Performance(0, argv[2], coverage , MaxError, 0.1 ,outputF2,0);
		    
		  }
	      }
	    // Retore config
	    S->In[numvar]->GetMF(0)->GetParams(params);
	    params[1]=cen[0];
	    params[2]=cen[1];
	    S->In[numvar]->GetMF(0)->Update(params);
	    S->In[numvar]->GetMF(1)->GetParams(params);
	    params[0]=cen[0];
	    params[1]=cen[1];
	    S->In[numvar]->GetMF(1)->Update(params);
	  }
      }
    printf("\n");

    delete S;

    if (Exemples !=NULL) 
      {
	for( int i=0 ; i<nbrow ; i++ )
	  delete [] Exemples[i];
	delete [] Exemples;
	Exemples=NULL;
      }
    
    delete [] outputFisConfigName;
    delete [] outputFisBaseName;
    delete [] outputF;
    delete [] outputF2;
    delete [] fisout;
    delete [] cen;
    delete [] snum;
  
    return 0;  
}
  catch(std::exception &e){
    printf("error in sensi\n");
    printf("catch %s\n",e.what());
    return -1;
  }

}


